/* Utility for mask cutting - Header file */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2009 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/* ---------------------------------------------------------------------- */

#ifndef	INCLUDE_CUTUTIL_H
#define	INCLUDE_CUTUTIL_H

/* INCLUDEs ------------------------------------------------------------- */

#include	MGUTILH
// mgutils.h  includes optutils.h
// optutils.h includes maskdat.h
// maskdat.h  includes ioutils.h, kdcutil.h


/* DEFINEs -------------------------------------------------------------- */

//#define  SIMPLECURVE
#undef  SIMPLECURVE
/*  Symbol SIMPLECURVE allows use of Curve global variable to
  shorten execution time very slightly.  If there is any problem,
  it may be easily undefined to do full lookup.  */
/*  SIMPLECURVE is permanently undef to allow GISMO and others
  to have a cylindrical mask case.  */


/* TYPEDEFs ------------------------------------------------------------- */


struct	hole_adjust {
	double	angle_adjust;
	double	beam_diam;
	double	clearance;
	};
typedef  struct hole_adjust	adjustment;


enum	Laser_modes {
	    Laser_on,
	    Laser_off,
	    Autofocus_on,
	    Autofocus_off,
	    Autofocus_hold,
	    Autofocus_release,
	    Autofocus_restore,
	    Laser_cutting,
	    Laser_cutting2,
	    Laser_roughcut,
	    Laser_engraving,
	    Exact_stop_on,
	    Exact_stop_off,
	    Pause
	};
// The above should be in cututil.h, actually...


/* GLOBALs -------------------------------------------------------------- */


/* Acknowledging that it is bad form, nonetheless these global values
  are assumed by the generalizations in this library and get located
  here.  The variables shoule be called "extern" if this is included
  from a file which is not cututil.c  */

#ifdef IS_CUTUTIL_C
#define EXT_CUTUTIL 
#else
#define EXT_CUTUTIL extern
#endif

#ifdef IS_CUTUTIL_C
double	SF = 1.0;	// Scale factor
/* SF will be set to 1/25.4 to produce inch output */
int	inchout = 1;	// Logical flag for inch output
#else
double	SF;
int	inchout;
#endif

#ifdef IS_CUTUTIL_C
/* Global line number for reference and debug */
int	Lnum=0;

/* Long slit splitting parameters (millimeters) */
double	LongSlit = 40.0;
double	SlitGap  = 0.500;
// These are modifiable by configuration file in maskcut
#else
int	Lnum;
double	LongSlit;
double	SlitGap;
#endif

//#ifdef  USE_CTE
//// #define  CTE  12.5E-6
//#define  CTE  17.0E-6
//// Per Tyson 05/10-5, using 304 stainless, and standard reference
//#else
//#define  CTE  0.0
//#endif
/* CTE is coeficient of thermal expansion of the mask material */
//#define  Tcut  20.0
/* Tcut is temperature for cutting masks */
/* The CTE and Tcut are now read in maskcut.cfg file */

// Parameters for all cases
EXT_CUTUTIL  double	bwid;
EXT_CUTUTIL  double	quick;
EXT_CUTUTIL  double	slow;
EXT_CUTUTIL  double	vslow;
EXT_CUTUTIL  double	avtool;
#ifdef IS_CUTUTIL_C
double  Afocus = 12.7;	// Autofocus focal length from Z reference
double	Zoff = 0.0;	// Offset in Z direction, including autofocus.
int	Autofocus = 0;
int	Astop = 0;
int	Laseron = 0;
int	SetSlew = 0;
#else
double  Afocus;	// Autofocus focal length from Z reference
double	Zoff;	// Offset in Z direction, including autofocus.
int	Autofocus;
int	Astop;
int	Laseron;
int	SetSlew;
#endif
EXT_CUTUTIL  double	flange;
EXT_CUTUTIL  vect2	lastloc;
EXT_CUTUTIL  double	lastz;
#ifdef  SIMPLECURVE
EXT_CUTUTIL  double	Curve;
#else
EXT_CUTUTIL  vect2	Curve;
#endif

#ifdef  USE_ACON
EXT_CUTUTIL  double	Acon;
// EXT_CUTUTIL  double	Ctime;
// EXT_CUTUTIL  double	fa;
// EXT_CUTUTIL  double	da;
// EXT_CUTUTIL  double	dr;
// EXT_CUTUTIL  double	ax;
// EXT_CUTUTIL  double	feed;
// EXT_CUTUTIL  int	na;
// EXT_CUTUTIL  int	xac;
#endif

// INSTRUMENTS should agree with Instruments in mgutils.h
// #define  INSTRUMENTS  4
// #define  INSTRUMENTS  Instruments
// Above is not used.  Should be replaced by Masks anyway!
#ifdef IS_CUTUTIL_C
int	inst = -1;	// start with unknown
#else
int	inst;
#endif

#ifdef  USE_CDAT
EXT_CUTUTIL  cdata	*CDhead;  // Global, should be parameter
#endif
#ifdef  USE_GSTO
EXT_CUTUTIL  gsto	*CDhead;  // Global here, later parameter?
#endif
#ifdef  USE_BSTO
EXT_CUTUTIL  BNODE	*CDhead;  // Global, should be parameter
#endif

/* FUNCTION PROTOTYPEs -------------------------------------------------- */

int	nsides (double R, double Delt, double a);

double	vhigh (double r, double c);
double	dvhi (vect2 p, vect2 c);

/*  ---  Cylinder metric -- for cutting flat mask on cylinder  ---  */
//double	cproj (double s, const double c, const double SF);
//double	cunproj (double x, const double c, const double SF);

/*  ---  Common Output Formatting  ---  */
void	Feol (FILE* f);
void	Gpause (FILE* f, double s);
void	IDent (FILE* f);
void	setdisp (double disp, double ydisp, double angle, int cfflag);
void	XYdp (FILE* f, double x, double y, int e);
void	VXYdp (FILE* f, vect2 v, int e);
void	XYout (FILE* f, double x, double y, int e);
void	VXYout (FILE* f, vect2 v, int e);
void	Zout (FILE* f, double z);
void	Feedrate (FILE* f, char* s, double r);

/*  ---  Laser Interface code  ---  */
void	lasermode (int mode, FILE* f);
int	onflange (vect2 p);
double	maskz (vect2 p);
double  newloc (vect2 a, double h, vect2 b, FILE* u);
// int	in_avoid (vect2 p);
void	pltimacs (vect2 v, int pen, void* u);
double	chole (vect2 c, double d, double h, double fr, int rc, FILE* u);
// NOTE -- move define of "adjustment" from cututil.c to above when
// needed to have these prototypes defined.
double	pvhole (double r, double a, double d, double h, adjustment w, FILE* u);
void	a_slot (double r, double a, double d, double b, adjustment w, FILE* u);

double	cutabs (vect2* last, vect2 v, double h, FILE* u);
double	cutarc (vect2* last, vect2 cd, double f, FILE* u);

double	crect (vect2 p, vect2 a, double w, double hp, double fr, int rc, FILE* u);

double	carve_edge (vect2* next, vect2 e, double z, FILE* u);
double	carea (vect2 p, vect2 a, double w, double hp, double fr, FILE* u, 
		avoids* aqp, double rt);

/* ---------------------------------------------------------------------- */
/* End of cututil.h package */

#endif	/* INCLUDE_CUTUTIL_H */

/* ---------------------------------------------------------------------- */
