/* Include file for cwplot.c and dbplot.c containing common
  code to be used by both.  */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                  Copyright (C) 2007 by                            **
**     Ken Clardy,  Carnegie Observatories,  Pasadena California.    **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include	<sys/time.h>

/*  Time utility program */

double	tsince (struct timeval *then, struct timeval *now)
/* Returns time in seconds since then, and stores current time in the
  provided timeval location = now */
{
double	t;
    gettimeofday (now, NULL);
    t = (double) (now->tv_sec - then->tv_sec)
	+ (double) (now->tv_usec - then->tv_usec) / 1.0e6;
    return  t;
}


GC	gcsetup (GC gc, u_long color)
/* Set foreground and line attribute for color */
{
//GC	r = mwin.gfs.gc;	/* return value */
GC	r = gc;
u_long	clr;
u_long	dot;

/* Separate attributes */
    dot = color & DOTTED;
    clr = color - dot;

/* Install attributes into the return gc */
//    r = gc;
//    XSetForeground (display, r, clr);
    XSetForeground (mwin.disp, r, clr);
// We may need to set that 3rd attribute, line width, with a
// field in the extended color...
    if (dot)
	XSetLineAttributes(mwin.disp,r,1,LineOnOffDash,CapRound,JoinRound);
    else
	XSetLineAttributes(mwin.disp,r,1,LineSolid,CapRound,JoinRound);
    return  r;
}

// NOTE THAT DOTTED IS NOW 0; WE NEED TO DEFINE DOTTED USING A
// SEPARATE WORD, NOT THE COLOR PROPERTY.

// OR, MAYBE, INSTEAD OF COLOR BEING A RGB VALUE, IT COULD BE AN INDEX
// TO A TABLE OF DEFINED VALUES FOR COLORS, AND DOT COULD BE A HIGHER BIT...
// WE COULD HAVE A TABLE OF 256 POSSIBLE COLORS, INTS INDEXED BY AN INT.

/* The Set_GC is from dbplot.c and not used in cwplot.c */

#ifdef  ISDBPLOT
/* Local gc set/create utility, replacing CreateGC */
GC	Set_GC (int fg, int bg, char* fnam, char* fwid, int size)
{
GC	rgc;
Font font;
char fontname[128];

    rgc = gc;
    XSetForeground (mwin.disp, rgc, fg);
    XSetBackground (mwin.disp, rgc, bg);
    sprintf(fontname,"-*-%s-%s-r-*-*-%d-*-*-*-*-*-*-*",fnam,fwid,size);
    font = XLoadFont(mwin.disp,fontname);
    XSetFont(mwin.disp,rgc,font);
//  gfs.f_struct = XQueryFont(disp,font);
// From CBX_CreateGfC_Ext in cxtfiles/cxt_main.c
// We are NOT doing the lock/unlock things, or storing the font structure
// that is in the GfC structure.

    return  rgc;
}
#endif


static	int	xloci (int sx)
/* Inverse xloc function from guiutil stuff */
{
int	k;
    k = (sx - 4) / Hunit;
    return k;
}

void	set_sizes (XWindowAttributes	*xwa)
/* Set up window */
{
    wx = xwa->width;
    wy = xwa->height;

/* Set up button locations */
    dbfield.xloc = xloci (nint(wx - 10.0));
    MoveButton (&dbfield);	// "Done" tracks right side

#ifdef  ISDBPLOT
//    rzfield.xloc = xloci (nint((wx - Hunit*Xp*(rzfield.width+6))/2.0) );
// above is centered button.
    rzfield.xloc = xloci (nint((wx - Hunit*Xp*(0))/2.0) );
#else
    rzfield.xloc = xloci (nint((wx - Hunit*Xp*(rzfield.width+6))/2.0) );
// The actual button width is width + text size (6 here)
#endif
    MoveButton (&rzfield);	// "ReZero" is centered

/* Find the viewport center for zooming */
    cwx = wx / 2;
    cwy = wy - (wy-50)/2;
}

// See EnableDMenu and DisableDMenu in dbplot.c04 and earlier
// These were not used.


// CHANGE THE ABOVE TO REMOVE THE MASK/DETECTOR STUFF; THERE IS ONLY
// ONE VIEW V...  WE MAY NOT HAVE VMENU AT ALL.
/* The handle vmenu things are from dbplot and not used in cwplot */

#ifdef  ISDBPLOT
#ifndef  MASKONLY
//int handle_vmenu(void* param)
void handle_vmenu(void* param)
{
int	j;
//int	k;
    if (param == NULL) bpressed++;	// Mostly to satisfy lint
    vp[w_active] = V;
//    j = CBX_HandlePMenu (&vmenu);
    j = CBX_HandleMenu (&vmenu);
//    k = (w_active == j) ? 0 : 1;	// change in window
    if (j == MASK || j == DETECTOR) w_active = j;
    V = vp[w_active];
    vmenu.sel = w_active;
    CBX_UpdateMenu(&vmenu);	/* Update needed for MacOSX */
    XSetWindowBackground (mwin.disp, win, V.color);
//    return 1;
//    bpressed++;
//    return  k;
}
#endif  // on MASKONLY
#endif  // on ISDBPLOT

void	setup_window ()
{
// #ifdef  ISDBPLOT
double	dt;
int	pb;
/* Track the time, may want to skip if quick update? */
    dt = tsince (&previous, &current);
    pb = bpressed;
    previous = current;
    bpressed = 0;
    if (dt < 0.080 && !pb) return;	// Skip if quick redraw
// The .08 is long enough to get solaris redraw time
    if (dt < 1.0 && !pb) {
	printf (" *Redraw window after %.3f seconds\n", dt);
    }
// #endif

/* Clear the window first, then start drawing in it. */
    XClearWindow (mwin.disp, win);

/* Update the buttons; We don't yet have a utility to draw
  the button's label fields as DrawMenu does */
    CBX_UpdateButton (&dunbut);
    CBX_UpdateButton (&zibut);
    CBX_UpdateButton (&zobut);
    CBX_UpdateButton (&rzbut);
#ifdef  ISDBPLOT
//    CBX_UpdateSButton (&tupbut);
//    CBX_UpdateSButton (&tlfbut);
//    CBX_UpdateSButton (&trtbut);
//    CBX_UpdateSButton (&tdnbut);

/* vmenu is the view menu; it is a p-menu */
/* tmenu, rmenu, xmenu are all unused this version */
/* Update the menus within the window */
/* tmenu is a d-menu; xmenu is a p-menu */

/* Update all the menus in the window; The general method is to
  call the "DrawMenu" utility from guiutil.c with the menu field
  pointer as the argument.  It calls CBX_UpdateMenu also.  */

#ifndef  MASKONLY
    DrawMenu (&vfield);		// general task in guiutil.c
#endif  // on MASKONLY
#endif  // on ISDBPLOT
/* Update all the menus in the window; The general method is to
  call the "DrawMenu" utility from guiutil.c with the menu field
  pointer as the argument.  It calls CBX_UpdateMenu also.  */

//  Currently, no menus in this plot applicaton.  (in cwplot)

/*	printf (" Expose event on main window.\n");
	fflush (stdout);	*/
//- 	printf (" Draw the window.\n"); fflush (stdout);  //debug
#ifdef  ISDBPLOT
#ifdef  LOOPDEBUG
    printf (" Draw window.\n"); fflush(stdout); //debug
#endif
#endif
}


/*  --------------- The special plotting stuff  --------------- */


int	handle_scale (int d, int b)
/* Change scale; direction d, button b */
{
double	m;

    switch (b) {
    case  1:
	m = 1.25;
	break;
    case  2:
	m = 2.00;
	break;
    case  3:
	m = 4.00;
	break;
    default:
	return 0;
    }

    if (d) {
	V.BX *= m;
	V.BY *= m;
    } else {
	V.BX /= m;
	V.BY /= m;
    }

//#ifdef  ISDBPLOT
    bpressed++;
//#endif
// NOTE -- ISDBPLOT IS NOT REQURED AROUND bpressed STATEMENTS.
    return 1;	/* flag for re-display */
}

#ifdef  ISDBPLOT
int	handle_translate (int d, int b)
/* Translate; direction d, button b */
{
double	m;

    switch (b) {
    case  1:
	m = 10.0;
	break;
    case  2:
	m = 40.0;
	break;
    case  3:
	m = 100.0;
	break;
    default:
	return 0;
    }

    switch (d) {
    case UP:
	V.ZY += m/V.BY; break;
    case DOWN:
	V.ZY -= m/V.BY; break;
    case LEFT:
	V.ZX += m/V.BX; break;
    case RIGHT:
	V.ZX -= m/V.BX; break;
    default:
	return 0;
    }
    bpressed++;
    return 1;	/* flag for re-display */
// If dflag is kept local to main, set it on button release...
}
#endif


int	scale (double slope, double zp, double v, int c)
/* Change floating space into integer plotting space */
{
int	k;
double	x;
/* A value equal to zp appears at the center of the viewport at
  a coordinate of c.  Other locations are linearly related at a
  slope which is given.  */
	x = slope * (v - zp);
/* We limit values of x to greater than the display size, yet smaller
  than the range of integers, so high magnifications will not cause the
  image to wrap around due to conversion. */
#define  LARGE  3.0e4
// 5.0e5
	if (x < -LARGE) x = -LARGE;
	if (x >  LARGE) x =  LARGE;
	k = (int) (x + 0.5 + (float)c);
	return  k;
}


int	handle_location (int x, int y, int button)
/* If center button pressed, change location to center */
{
int	df = 0;
double	fx, fy;

/* Find floating location.  x = B * (p - Z) + cw, so the inverse is
  p = (x - cw)/B + Z; the floating point coordinate. */

    fx = ( (double) (x - cwx) )/V.BX;
    fy = ( (double) (y - cwy) )/V.BY;

    switch (button) {
	case  2:
/* Update zero point to center indicated point on screen */
	V.ZX += fx;
	V.ZY += fy;
	df = 1;	/* Flag causes re-plot of window */
	break;

	case  3:
/* Display the cursor location in lower left */
/*	sprintf (locbuf, "%10.3g (%d) %10.3g (%d)%c",  */
	sprintf (locbuf, "%10.3f (%d) %10.3f (%d)%c",
		fx+V.ZX, x,  fy+V.ZY, y, AN);
	df |= 2;	/* Special flag to display (global) locbuf */
	break;

	case  1:
/* Null action here */
	break;

    }	// end switch on button value
//#ifdef  ISDBPLOT
    bpressed++;
//#endif
    return  df;
}

// NOTE THAT WE WANT TO ADD SHIFT-CLICK ABOVE TO MOVE TO NEAREST INTEGER
// LOCATION -- SEE HOW TO GET SHIFT STATUS.

//int	recenter (int button)
int	recenter (void)
/* Re-center the map. */
{
int	df = 0;

/* Don't worry too much yet about which button... */
    df = (V.ZX != 0.0) || (V.ZY != 0.0);
    V.ZX = 0.0;
    V.ZY = 0.0;
//    df = 1;		/* re-plot */
//#ifdef  ISDBPLOT
    bpressed++;
//#endif
    return df;
}


void	vect (u_long color, double x1, double y1, double x2, double y2)
/* Plot vector from x1,y1 to x2,y2 using color */
{
long	ex, ey;
long	px, py;
#ifndef  ISDBPLOT
GC	lgc;
#endif
// Put point on the display
    px = scale (V.BX, V.ZX, x1, cwx);
    ex = scale (V.BX, V.ZX, x2, cwx);
    py = scale (V.BY, V.ZY, y1, cwy);
    ey = scale (V.BY, V.ZY, y2, cwy);
// Change gc foreground from existing to color, then change back...
#ifdef  ISDBPLOT
    lgc = gcsetup (gc, color);
#else
    lgc = gcsetup (mwin.gfs.gc, color);
#endif
    XDrawLine (mwin.disp, win, lgc, px, py,  ex, ey);
}

void	v2line (u_long color, vect2 a, vect2 b)
/* Draw line between points a and b */
{
    vect (color, a.x, a.y, b.x, b.y);
}


void	cross (u_long color, double x, double y, int r, int type)
/* x, y in real coordinates, r in pixels, type is 0 = cross, 1 = x */
{
int	jx, jy;
#ifndef  ISDBPLOT
GC	lgc;
#endif

    jx = scale (V.BX, V.ZX, x, cwx);
    jy = scale (V.BY, V.ZY, y, cwy);
#ifdef  ISDBPLOT
    lgc = gcsetup (gc, color);
#else
    lgc = gcsetup (mwin.gfs.gc, color);
#endif
    if (type == 0) {		/* cross */
	XDrawLine (mwin.disp, win, lgc, jx, jy-r, jx, jy+r);
	XDrawLine (mwin.disp, win, lgc, jx-r, jy, jx+r, jy);
    } else {			/* X  */
	XDrawLine (mwin.disp, win, lgc, jx-r, jy-r, jx+r, jy+r);
	XDrawLine (mwin.disp, win, lgc, jx-r, jy+r, jx+r, jy-r);
    }
}

// NOTE here ISDBPLOT is using a global gc and lgc, and this may
// not be the best thing to do.  But also, we don't restore gc...

void	scircle (u_long color, vect2 p, double r, int type)
/* Scaled circle at p, r is scaled, type is 0 = open,
  1 = filled circle */
{
int	jx, jy;
int	d;
#ifndef  ISDBPLOT
GC	lgc;
#endif

    jx = scale (V.BX, V.ZX, p.x, cwx);
    jy = scale (V.BY, V.ZY, p.y, cwy);
    d = scale (V.BX, 0.0, r, 0);	/* Radius of circle */
#ifdef  ISDBPLOT
    lgc = gcsetup (gc, color);
#else
    lgc = gcsetup (mwin.gfs.gc, color);
#endif
    if (type == 0) {
	XDrawArc (mwin.disp, win, lgc,
	    jx-d, jy-d, 2*d, 2*d, 0, 23040);
    } else {
	XFillArc (mwin.disp, win, lgc,
	    jx-d, jy-d, 2*d, 2*d, 0, 23040);
    }
}


// sarc is not used in dbplot.c so it is bypassed if true
#ifndef  ISDBPLOT
void	sarc (u_long color, vect2 c, vect2 r, double a1, double a2)
/* Draw arc about center c, radius vectors r, starting at angle a1,
  for a2 degrees (positive ccw).  Radius is scaled.  */
{
int	jx, jy;   // Scaled center location
int	rx, ry;   // Scaled radius sizes
int	c1, c2;   // Scaled angles
GC	lgc;

    jx = scale (V.BX, V.ZX, c.x, cwx);
    jy = scale (V.BY, V.ZY, c.y, cwy);

    rx = scale (V.BX, 0.0, r.x, 0);
    ry = scale (V.BX, 0.0, r.y, 0);  // don't invert scale here.

    lgc = gcsetup (mwin.gfs.gc, color);

    c1 = a1 * 64.0;
    c2 = a2 * 64.0;

    XDrawArc (mwin.disp, win, lgc,
	jx-rx, jy-ry,  2*rx, 2*ry,  c1, c2);

}
#endif

void	fcircle (u_long color, vect2 p, int r, int type)
/* Fixed size circle at p, r in pixels, type is 0 = open,
  1 = filled circle */
{
int	jx, jy;
#ifndef  ISDBPLOT
GC	lgc;
#endif

    jx = scale (V.BX, V.ZX, p.x, cwx);
    jy = scale (V.BY, V.ZY, p.y, cwy);
#ifdef  ISDBPLOT
    lgc = gcsetup (gc, color);
#else
    lgc = gcsetup (mwin.gfs.gc, color);
#endif
    if (type == 0) {
	XDrawArc (mwin.disp, win, lgc,
	    jx-r, jy-r, 2*r, 2*r, 0, 23040);
    } else {
	XFillArc (mwin.disp, win, lgc,
	    jx-r, jy-r, 2*r, 2*r, 0, 23040);
    }
}


void	rectangle (u_long color, double x1, double x2, double y1, double y2)
/* Plot a rectangle aligned with axis, given 4 sides */
{
long	px, ex;
long	py, ey;
#ifndef  ISDBPLOT
GC	lgc;
#endif

/* Get the scaled positions */
    px = scale (V.BX, V.ZX, x1, cwx);
    ex = scale (V.BX, V.ZX, x2, cwx);
    py = scale (V.BY, V.ZY, y1, cwy);
    ey = scale (V.BY, V.ZY, y2, cwy);
/* Set up the color */
#ifdef  ISDBPLOT
    lgc = gcsetup (gc, color);
#else
    lgc = gcsetup (mwin.gfs.gc, color);
#endif
/* Draw the lines */
    XDrawLine (mwin.disp, win, lgc, px, py,  px, ey);
    XDrawLine (mwin.disp, win, lgc, px, ey,  ex, ey);
    XDrawLine (mwin.disp, win, lgc, ex, ey,  ex, py);
    XDrawLine (mwin.disp, win, lgc, ex, py,  px, py);
}

void	quadralat (u_long color, vect2 p1, vect2 p2, vect2 p3, vect2 p4)
/* Plot a quadralateral around 4 points */
{
    vect (color, p1.x, p1.y, p2.x, p2.y);
    vect (color, p2.x, p2.y, p3.x, p3.y);
    vect (color, p3.x, p3.y, p4.x, p4.y);
    vect (color, p4.x, p4.y, p1.x, p1.y);
}


void	axis (u_long color, double angle, int n)
/* Plot the given axis from 0,0 at angle for n units */
// This is an old debug routine for historical purposes only
{
double	a, s, c;
double	x;
int	gx, gy, tx, ty;
int	px, py, ex, ey;
int	k;
#ifndef  ISDBPLOT
GC	lgc;
#endif

// Obtain unit vectors along and perpendicular
    a = angle / (180.0 / M_PI);
    s = sin (a);
    c = cos (a);
// Vector along axis is c,s and perpendicular is -s,c

// Obtain vector offsets for segment and tick mark
    gx = scale (V.BX, 0.0, c, 0);
    gy = scale (V.BY, 0.0, s, 0);
    tx = scale (V.BX, 0.0, -s*0.1, 0);
    ty = scale (V.BY, 0.0, c*0.1, 0);

#ifdef  ISDBPLOT
    lgc = gcsetup (gc, color);
#else
    lgc = gcsetup (mwin.gfs.gc, color);
#endif

// Loop to draw axis -- segment followed by tick
    for (x=0.0,k=0; k<n; k++,x+=1.0) {
// x is distance along axis, k is count of intervals
	px = scale (V.BX, V.ZX, x*c, cwx);
	py = scale (V.BY, V.ZY, x*s, cwy);
	ex = px + gx;
	ey = py + gy;
// px, py is starting point
	XDrawLine (mwin.disp, win, lgc, px, py,  ex, ey);
	XDrawLine (mwin.disp, win, lgc, ex-tx, ey-ty, ex+tx, ey+ty);
    }

}
