/*  datutil.c == utility programs introduced to gmap */
/*  --- Ken Clardy -----------------------------------------------
 *
 *  datutil.c
 *
 *  Project:  PMAP 
 *
 *  V0.5  2003-08-28
 *  V0.5a 2008-11-11
 *
 *  --------------------------------------------------------------*/

/* DEFINEs -------------------------------------------------------------- */

/* #define DEBUG */

#ifdef SOLARIS
#ifndef SVR4
#define SVR4
#endif
#endif

/* INCLUDES ------------------------------------------------------------- */

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<unistd.h>	// access

#include	"datutil.h"

#include	"utils.h"
 
 
/* GLOBALs -------------------------------------------------------------- */
 
/* none */
 
 
/* EXTERNs -------------------------------------------------------------- */
 
/* none */
 


/* ---------------------------------------------------------------------- */

/* NOTE:  These first 2 programs do not use the structures defined in
  datutil.h; they are a general I/O program which should be placed in the
  ioutil.c library and linked from there.  */
/* They DO however require defines of WHITESPACE and AN */

/* ---------------------------------------------------------------------- */

/* The "get_data" program from here is now in general i/o utilities. */

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */


section*  newsection (char* name)
/* Obtain, initialize and return a section structure */
/* Put name given into the new structure */
{
section	*p;	// return value

    p = (section*) malloc (sizeof (section));
    p->last = p->next = p;	// rather than NULL, make sure it's circular
//    p->name = name;	// Maybe, make a copy?  yeah, a copy, thats it...
    if (name != NULL) {
	p->name = malloc (strlen(name)+1);
	strcpy (p->name, name);  // includes term. null
    } else p->name = NULL;
    p->flag = 0;
    p->head = NULL;

    return  p;
}

/* ---------------------------------------------------------------------- */

void*	gdatstore (char* line, void* head)
//gdata*	gdatstore (char* line, gdata* head)
/* Data storage for gmap.  Put data from "line" into dynamic structures
  pointed by "head".  User program to get_data. */
{
#define  KEYSIZE  80
static  section*  cursec=NULL;
static  datlist*  curend=NULL;
gdata	*ghead;
char	keyword[KEYSIZE];
char	*c, *cl;
section	**s=NULL, *p;
datlist	*d;
int	key;
// int	i;
int	j;
int	k;
// static	int	dc = 0;		// debug count -- debug only

enum  keywords { NONE, TEL, CAM, GDR, DEF };

/* If line is NULL, return head unconditionally */
    if (line == NULL) {
//	printf (" (Processed %d lines.)\n", dc);  //debug
	return  head;
    }

/* If head is NULL, obtain our base head, and use it forevermore */
    if (head == NULL) {
	ghead = (gdata*) malloc (sizeof (gdata));
	ghead->telescope = NULL;
	ghead->camera    = NULL;
	ghead->guider    = NULL;
	head = (void*) ghead;
    } else {
	ghead = (gdata*) head;
    }

/* The void* is actually a gdata* pointer here.  It is "void" in the
  calling program, the general reading program.  */

/* For DEBUG, we just echo the input line, with some stuff around
  it to show what we have to work with... */
//	dc++;	// debug

/* Remember where we are - current section is static storage;
  start with NULL, and if NULL, ignore as being not in a section. */
/* Also, remember the current last entry in the section datlist */


/* Set char *c to the "rest" of the line and copy a whitespace
  delimited keyword to the "keyword" buffer... */

    c = strpbrk (line, WHITESPACE);
    if (c == NULL) {		// No whitespace in line
	j = strlen (line);
	c = line + j;
    } else {		// keyword and rest
	j = c - line;	// length of keyword
	while (strchr(WHITESPACE, *c) && *c != AN) c++;
    }
    strncpy (keyword, line, KEYSIZE);
    if (j < KEYSIZE) keyword[j] = AN;
    else keyword[KEYSIZE-1] = AN;

// Replace above section by the below; needs ioutils library.
//    c = parse (line, keyword, WHITESPACE);

//	if (dc < 10) printf (" >> \"%s\": Keyword \"%s\", rest \"%s\"\n",
//		line, keyword, c);	// debug


// Put the result in last place in the appropriate head structure...
// use the push-section for that.

// Don't yet have push/pop section; also need a kill section code.


/* Translate keyword values into key values for later case statements */
    if (! strcasecmp(keyword, "telescope")) 
	key = TEL;
    else if (! strcasecmp(keyword, "camera")) 
	key = CAM;
    else if (! strcasecmp(keyword, "guider")) 
	key = GDR;
    else if (! strcasecmp(keyword, "default")) 
	key = DEF;
    else 
	key = NONE;


/* Select the section if any... */
    switch (key) {
	case  TEL:  s = &(ghead->telescope); break;
	case  CAM:  s = &(ghead->camera); break;
	case  GDR:  s = &(ghead->guider); break;
    }

//    if (dc < 10) printf (".key = %d\n", key);  //debug
//    fflush (stdout);

/* Process the line, based on keyword, previously decoded */
    switch (key) {
	case  TEL:
	case  CAM:
	case  GDR:
/* Start a new section if we have a starting keyword */
	    cursec = newsection (c);
	    if (*s == NULL) {
		*s = cursec;
	    } else {
		p = (*s)->last;  // can be null?
		if (p == NULL) p = *s;	// for safety
// Above added as debug.  List should now always be circular.
		p->next = cursec;
		cursec->last = p;
		cursec->next = (*s);
		(*s)->last = cursec;
	    }
	    curend = NULL;
	    break;
	case  DEF:
/* Keyword "default" just turns on the flag in current section */
	    if (cursec == NULL) break;	// can't add to unknown section
	    cursec->flag |= 1;  // or symbolic default bit
	    break;
	default:
/* Anything else is not a keyword; copy entire line to a datlist element,
  and then append it at curend, update curend. */
	    if (cursec == NULL) break;	// can't add to unknown section
	    k = strlen(line) + 1;
	    cl = malloc (k+2);
// Added +2 above to avoid Linux segmentation fault...
	    strcpy (cl, line);	// cl is copy of line
	    cl[k] = AN;
	    d = (datlist*) malloc (sizeof(datlist));
// Sometimes segmentation fault in above statement...
	    d->next = NULL;
	    d->data = cl;
	    if (curend == NULL) cursec->head = d;
	    else		curend->next = d;
	    curend = d;
	    break;
    }

/* Return the head pointer */
    return  head;
}

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */

/* Following are replacement programs for those used to read individual
  files for this data... */

char* get_gstring (section* sec, char* key, char* value, char* defstring)
{
char	buffer[1024],val[128];
char	*h,*r=NULL;
datlist	*dp;
//  FILE   *fp;

    if (defstring) { strcpy(value,defstring); r=value; }

    if (sec == NULL) return r;
    for (dp = sec->head; dp != NULL; dp = dp->next) {
	strcpy (buffer, dp->data);
	cut_spaces(buffer);                    /* remove <CR> */
	h = strchr(buffer,';');                /* comment */
	if (h) *h='\0';
	h = strchr(buffer,'=');                /* delimiter */
	if (h==NULL) continue;
	strcpy(val,h+1);
	*h = '\0';
	cut_spaces(buffer);                    /* leading and trailing */
	if (!strcasecmp(buffer,key)) {         /* key found */
	    cut_spaces(val);
	    strcpy(value,val); r=value;
	    break;
	}
    }
    return(r);
}

/* ---------------------------------------------------------------------- */

long get_glong (section* sec, char* key, long defvalue)
{
char   buf[128],buf2[128];
long   value,l;

    if (get_gstring(sec,key,buf,NULL) == NULL) return(defvalue);
    if (!strncmp(buf,"0x",2)) {                      /* hex value */
        sscanf(buf,"%lx",&value); 
        return(value);
    }
    if ((l=strlen(buf)) > 1) {
        if ((buf[l-1] == 'h') || (buf[l-1] == 'H')) {  /* hex value */
            buf[l-1] = '\0'; 
            sprintf(buf2,"0x%s",buf);
            sscanf(buf2,"%lx",&value);
            return(value);
        }
    }
    return(atol(buf));                               /* decimal value */
}

/* ---------------------------------------------------------------------- */

double get_gdouble(section* sec, char* key, double defvalue)
{
char   buf[128];
 
    if (get_gstring(sec,key,buf,NULL) == NULL) return(defvalue);
    return (atof(buf));                               /* decimal value */
}

/* ---------------------------------------------------------------------- */

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */

int	checkfile (char* spec)
/* See if file exists.  Return sum of 1=exist, 2=readable, 3=writable,
  4=executable. */
{
int	r=0;	// return value
    if (access (spec, F_OK) == 0) {	// exists
	r |= 1;
	if (access (spec, R_OK) == 0) r |= 2;
	if (access (spec, W_OK) == 0) r |= 4;
	if (access (spec, X_OK) == 0) r |= 8;
    }
// fprintf (stderr, "  %d ..File \"%s\"\n", r, spec);	//debug only
    return  r;
}

/* ---------------------------------------------------------------------- */

/* For old findfile, see datutil.c.01 comments.  */

/* ---------------------------------------------------------------------- */

/* --UDP special extension----------------------------------------------- */

/* All comments.  See datutil.c.01 for those. */

/* ---------------------------------------------------------------------- */


