/*   GUISUBS == guisubs.c is a portion of the intgui.c code which
  is used in parallel by xmap.c while the latter program is being
  made ready to integrate into intgui.c
  When fully integrated, xmap ceases to exist and the code here
  will be returned to intgui.c    */
/*  Copyright (C) 2011 by Ken Clardy, Carnegie Observatories  */

/* Several of the more general routines have been placed in guiutil.c,
  to form a basis of an extended general library */

/* Commented out those things which have been generalized into
  the new guiutil.c code... */

/* ----- Function prototypes ----- */
/* (temporarily?) moved to intgui.h - for ease of editing */

static void	pgui		(Obf*, int);
// Moved here as it requires definition of Obf, not defined at time
// of original include of intgui.h way back above...

/*  --- Debug help  ---  */

/* Display message; ask to continue; return or exit. */
void	dpause (char* mes)
{
char	buf[32];
    if (dodebug < 3) return;
    fprintf (stderr, " --%s-- continue? ", mes);
//    gets (buf);
    fgets (buf, 30, stdin);
    if (strncasecmp(buf,"no", 1)) return;
    exit (0);
}

/*  -----  Below is the code for Xwindow stuff  ----- */

/*  ------------------------------------------------------------  */
void	ToggleButton(void* param)
// Callback for create D-button; toggle the status
{
Button *b=(Button*)param;
butblock	*bb;

/* Debug - find what button has been toggled */
// Would call BlockOfBut (b) to get button block;
// We have to move this down slightly to make the call work!
// Use the data pointer in button block;
// Set that data from the b->status value

    bb = BlockOfBut (butlist, b);
// if (b == &slitype) printf (" -- Toggling slit type.");
// else  printf (" -- Toggling some other button.");
// if (bb != NULL) printf (" (%s)\n", bb->label);
// else  printf ("\n");

    if (b->status) b->status = 0;
    else           b->status = 1;
    Gob->changed |= 1;

    if (bb != NULL) {
	if (bb->data != NULL) *(bb->data) = b->status;
    }

/* If we have the adc button, set/clear the rotbq queue... */
    if (b == &noadc) {
	rotbq = qkill (rotbq);
	if (b->status) {	// Turned off; get the queue
//    pd->width = 45.0;	// May parameterize?
//    pd->target = RotOff;	// May parameterize?
//	    pdx.width = 45.0;
//	    pdx.target = RotOff;
// Above moved from intgui.
	    rotbq = qwarn (&pdx, RotOff - 45.0, RotOff + 45.0, NULL);
	}
	warnflag = 1;	// Need to re-draw it.
    }

#ifdef  NSROT
    if (b == &nsrot) {
double	bias;
	bias = (nrbiask == 0) ? 0.0 : 90.0;
	if (fabs(Gob->obs->nrbias - bias * Degree) > NSrlim) Gob->changed |= 1;
	Gob->obs->nrbias = bias * Degree;
	slitpaf.offset = 90.0 - bias;
// need to handle slitpaf at this point...
	HandleTextWindow (winlist, slitpaf.ew);
    }
#endif

#ifdef  DIFREF
    if (b == &drefbut) {
/* Enable or Disable the Hour Angle window depending on whether
  the differential refraction feature is turned on or off.    */
	if (b->status) CBX_EnableEditWindow (&hawin);
	else  CBX_DisableEditWindow (&hawin, CBX_DISABLED); 
    }
#endif  // on DIFREF

    CBX_ReleaseButton(b);
}


/*  ------------------------------------------------------------  */
void	TextBack (void* param)
// General callback for automatic text window
{
EditWindow *w=(EditWindow*)param;
winblock	*wb;
double		val;

int	k;
int	n;
char	*p;

// Get the winblock
    wb = BlockOfWin (winlist, w);
#ifdef  IFUM
    if (wb == NULL) wb = BlockOfWin (IFUwinlist, w);
#endif

// Check for null case
    if (wb == NULL) return;


// Debug to see when this thing is actually called...
// if (wb == &tmpfield) printf (" ..Processing temperature\n");
// if (wb == &titlefld) printf (" ..Processing title\n");
// fflush(stdout);  //debug

// Decode text into the data (similar to handle_text)


/* Get the data from the window text block into the data field
  specified by winblock pointer wb.  */
/* This routine is usually called when we process a carriage return
  while in an active window with a green border */

/* The binary data field is primary.  This program is being called to
  set that data from the window text field, which has been entered by
  a user and may not be fully formed.  Thus after getting the proper
  data here, we need to call HTW at the end below to properly form
  the window text data field, and also to do UTW for updates.    */

    switch (wb->type) {
    case  TEXT:
	n = wb->width;
/* Get up to n characters for data field. */
	k = strlen (wb->ew->text);
	if (k > n) wb->ew->text[n] = AN;
	Gob->changed |= strcmp (wb->data, wb->ew->text);
	strncpy (wb->data, wb->ew->text, n);
	break;
    case  DTEXT:
	p = *(char**)wb->data;
	{		// Block for local data
    int	change;
	    if (p == NULL) change = 1;
	    else {
		if ((change=strcmp (p, wb->ew->text) )) free (p);
// Only reference p if it is not NULL, and
// only free, restore  p if it has changed
	    }
	    if (change) {
		p = dynamstr (wb->ew->text);
		*(char**)wb->data = p;
	    }
	    Gob->changed |= change;
	}
	break;
    case  ITEXT:
	*(int*)wb->data = wb->scale * atoi (wb->ew->text);
	Gob->changed |= 1;
	break;
    case  FTEXT:
	val = atof (wb->ew->text);
//	if (wb == &slitpaf) val -= 90.0;
	val -= wb->offset;
	*(double*)wb->data = wb->scale * val;
//	sprintf (wb->ew->text, "%.*f", 
//	    wb->dp, *(double*)wb->data/wb->scale);
	Gob->changed |= 1;
	break;
    case  DMS:
	*(double*)wb->data = wb->scale * pdecode (wb->ew->text);
//	strcpy (wb->ew->text,  
//	    sexig( *(double*)wb->data/wb->scale, 3, wb->dp) );
	Gob->changed |= 1;
	break;
    }
//    CBX_UpdateEditWindow(wb->ew);
// or reset edit window?  the reset also sets it to normal if not disabled.
//    CBX_ResetEditWindow (wb->ew);

/* ===== Move this section from here ===== to the HandleTextWindow
  program so as to not duplicate things when done.    */

/* ===== The application dependent window updates from here =====
  down should be put in Update Text Window, called by TextBack (here)
  and by HandleTextWindow.  It should return change status.    */

// Call the general update for the window
// That will put the data into the text window,
// and then call CBX_ResetEditWindow (w)

/*  =====  From here, the code now in Update Text Window (above) was
  extracted to allow it to be called both from here and from the
  HandleTextWindow program in guiutil.c.      */

/*  =====  Things down to here are extracted into a Update Text Window
  program, which is called by TextBack (here) and HandleTextWindow.  It may
  call HandleTextWindow for another window not itself if needed.  Be careful
  not to get in a loop, especially a->b->a changes.  When this is done,
  change the calls to HandleTextWindow below to similar calls to the
  Update Text Window program.  Perhaps use wb as the sole parameter in
  the new program just to make it easier.      =====      */

/* Make sure that HandleTextWindow is actually done, including special updates
  based on the data field being updated.      */

/* We [TB(w)] have gotten the assigned binary data set from the edit
  window, and now need HTW to normalize things, and to call UTW for
  needed updates to other windows and globals.   */

    Gob->changed |= HandleTextWindow (winlist, w);
#ifdef  IFUM
    Gob->changed |= HandleTextWindow (IFUwinlist, w);
#endif

/* Now, call UpdateTextWindow, which will in turn call the
  HandleTextWindow as needed; 
  this replaces the former calls to HandleTextWindow.      */
// Except that UTW does not call HTW now, it goes other way;
// So we continue to call HTW knowing it will call UTW

//    Gob->changed |= UpdateTextWindow (wb);
}


/*  ------------------------------------------------------------  */

//int	UpdateTextWindow (winblock* wb)
int	TextUpdate (winblock* wb)
/* Update any application dependent items in the indicated window
  presuming it has changed.  (We call this only when a change has
  been detected or made).  If a change is needed to another window,
  call HandleTextWindow (in external moduel guiutil.c) which will
  then call this module for that window.  It is important to NOT
  call for this same window, so test first.
  The point here is to put all window update dependencies in this
  one place.
  Return change status, 0 = no change, 1 = changes made.      */
{
int	change = 0;

/* Most of this code is from TextBack above, which was the former place
  to concentrate consequences of window updating.    */

// Usual test on null case
    if (wb == NULL) return change;

/*  =====  This program contains all the application dependent code
  for updating text windows (i.e. not buttons/menus ) for changes in
  the data or window text value.  This means that a change in a window
  MAY necessitate a change to another window or to a global value,
  and such changes are handled here.      =====    */

/*  The binary data is primary.  When we change it, we must call
  HTW(w) to handle the window's display.  We DO NOT call HTW for
  our own window, as we are called by HTW for its window, and this
  would cause a call loop.  But we DO call HTW for other windows.    */

// Check for some change in IFU position
#ifdef  IFUM
    if (wb == &URAfield || wb == &UDECfield) {
	IFUfindcenter (Gob->obs);
	change |= HandleTextWindow (winlist, RAfield.ew);
	change |= HandleTextWindow (winlist, DECfield.ew);
// Note, NOT updating the called windows with this.
    }
#endif

// HERE WE TEST FOR DECLINATION OR POSITION ANGLE WINDOWS, AND IF SO,
// RE-COMPUTE THE ROTATOR WARNING ANGLE...
    if (wb == &DECfield || wb == &slitpaf
#ifdef  IFUM
	|| wb == &URAfield || wb == &UDECfield
// Also need to test for IFU mode windows being changed too
#endif
	) {
// This is the only computation of rotator warning angle
double	dec, ipa;
//anglist   *al;
//bandlist  *bl;
int	doqflip = 1;	// driving variable for logic
int	Indx;
//#ifdef  IFUM
//	IFUfindcenter (Gob->obs);
//	HandleTextWindow (winlist, RAfield.ew);
//	HandleTextWindow (winlist, DECfield.ew);
//#endif
// The above section was actually done up about 24 lines
	dec = *(double*)DECfield.data / Degree;
	ipa = *(double*)slitpaf.data / Degree;
/* In the above we should probably use xxx.data / xxx.scale, but
  for this case it is the same thing.    */
#ifdef  INTGUI
//	hrf = gflip (dec, lat, ipa, 1);	// Side is 1
// Compute hrf and both queues...
// rotaq and rotbq, use pointers al and bl;
// struct is pdx...

// Set up the current pdx, using angles in degrees
	pdx.dec = dec;
	pdx.lat = lat;   // this is a global, in degrees
	pdx.pa  = ipa;
	pdx.side = 1;	// IMACS side is 1
// Find the new angle queue
//	while (rotaq != NULL) {  // kill the rotaq queue
//	    rotaq = G_pop_l (al = rotaq);
//	    free (al);
//	}
/* Move the above functionality into the new rotator queue subroutine */

/* Set the doqflip based on instrument.  */
	Indx = InstNdx (Gob->obs);
	if (Indx == I_imacs) doqflip = 1;
	if (Indx == I_lds2 || Indx == I_lds3) doqflip = 2;
	if (Indx == I_cfg) doqflip = 0;
	if (Indx == I_wfcd) doqflip = 0;

/* The rotator flip angle computation is only done for instruments
  on the Magellan telescopes, omit for other telescopes.  */
#ifdef  WFCCD	// this may not be needed; omit if extending instrument list
////	if (InstNdx (Gob->obs) != I_wfcd)  
//	if (InstNdx (Gob->obs) == I_wfcd)  doqflip = 0;
/* Above omits qflip calls when wfccd instrument is being used.  */
/* And, add other instruments if needed...  When the above conditional is
  false, rotaq stays NULL here.  */
/* NOTE that there may be a problem if instrument is changed without
  changing the position or angle values causing this computation to
  not be done.. Should maybe set a flag when instrument is changed?
  Or detect the flag which is set then?   */
#endif
#ifdef  GISAP
	if (FeatIndx(Gob->obs) == F_gisap) doqflip = 0;
/* Omit qflip call when gismo aperture feature is in effect. */
#endif

	find_rotator (doqflip, 0);

// #ifndef  LDSS_NEW
//	rotbq = qkill (rotbq);
// 	if (NoADC) rotbq = qwarn (&pdx);
// #endif
// Above not needed, so it is removed...

// Queues get displayed in redraw_warnings; we set the flag for that...
#else  // on INTGUI
	hrf = 30.0;	// No flip if not in the gui
#endif  // on INTGUI
	warnflag = 1;
//	printf ("  ..Set hrf to %.2f; dec=%.2f, ipa=%.2f\n", 
//		hrf, dec, ipa);		// debug
// We can substitute the expressions for dec and ipa in the call to
// gflip, and drop the defined variables above, making this more compact;
// After removing the debug print.
    }
// lat is an initialized constant above

#ifdef  GUIDATE
/* We raise warnflag if RA of object changes to possibly change the
  plotting of the twilight graphic if enabled. */
    if (wb == &RAfield) warnflag = 1;
// Might also test for a valid date present too.
#endif

#ifdef  SLXB
/* Slit extension feature.  If either uncut length is non-zero,
  we cannot use the feature, so disable the button.  */
    if (wb == &dcafield || wb == &dcbfield) {
//-	dec = *(double*)DECfield.data / Degree;
	if ( *(double*)dcafield.data == 0.0 &&
	     *(double*)dcbfield.data == 0.0 ) {   // button is OK
	    CBX_EnableButton (&slext);
	} else {				  // button is disabled
	    CBX_DisableButton (&slext, CBX_DISABLED);
// Could also turn Gob->obs->slext off (to 0) here; not necessary.
	}
    }
#endif


#ifdef  DWAV
// Remove dependence on DWAV when this feature is documented.
// The DWAV feature is defined in mgfeats.h now.
// The DWAV symbol is defined in intgui.c which <include>s this file.
/* When passband changes, update detector limits to be equal. */
    if (wb == &blupbfld) {
	*(double*)bludbfld.data = *(double*)blupbfld.data;
	change |= HandleTextWindow (winlist, bludbfld.ew);
    }
    if (wb == &redpbfld) {
	*(double*)reddbfld.data = *(double*)redpbfld.data;
	change |= HandleTextWindow (winlist, reddbfld.ew);
/* Might change above to include " Gob->changed |= " if needed */
    }
#endif

/* We specifically DO NOT call HandleTextWindow here, but we are
  called by it, and called by TextBack also.  */

/*  =====  Return the change found.  =====  */
    return  change;
}



/*  ------------------------------------------------------------  */
void handle_eqmenu(void)
{
int    new;
double	eqnx;
char   buf[128];

//    new = CBX_HandlePMenu(&eqmenu);
    new = CBX_HandleMenu(&eqmenu);
    if (new < 0) return;
    if ((new != EQU_SET) && (new == eqmenu.sel)) return;
    eqnx = Gob->obs->equinox;	// copy old equinox

    switch (new) {
    case EQU_1950:
	eqnx = 1950.;
	break;
    case EQU_2000:
	eqnx = 2000.;
	break;
    case EQU_TODAY:
	eqnx = get_epoch(0);
	break;
    case EQU_SET:
	sprintf(buf,"%.2f",eqnx);
	if (CBX_ParameterBox("Equinox",buf)) {     /* ok */
	    eqnx = fabs(atof(buf));
	    while (eqnx > 3000.) eqnx -= 3000.;
	    while (eqnx < 1000.) eqnx += 1000.;
	}
	sprintf(eqmenu.entry[EQU_SET].text,"%.2f",eqnx);
	CBX_SendExpose_Ext (mwin.disp, mwin.win);
	break;
    }
    eqmenu.sel = new;

//    Gob->changed = Gob->obs->equinox != eqnx;
    if (Gob->obs->equinox != eqnx) {
	Gob->obs->equinox = eqnx;	// restore global value
	wepoch = eqnx;	// Set the global
	Gob->changed = 1;
	warnflag = 1;
    }
}

/* === Location of handle_instmenu */
/* === Location of handle_dispmenu */
/* === Location of handle_filtmenu */


/*  ------------------------------------------------------------  */
#ifdef  WDEBUG
void  vgridline (int x, int color)
{
GC	lgc = mwin.gfs.gc;
int	xx, yf;

    XSetForeground (mwin.disp, lgc, color);
    xx = Xloc (x);
    yf = Yloc (Vsize);
    XDrawLine (mwin.disp,mwin.win,lgc, xx,0, xx,yf);
}
#endif

/*  ------------------------------------------------------------  */
#ifdef  WDEBUG
void  hgridline (int y, int color)
{
GC	lgc = mwin.gfs.gc;
int	xr, yy;

    XSetForeground (mwin.disp, lgc, color);
    xr = Xloc (Hsize);
    yy = Yloc (y);
    XDrawLine (mwin.disp,mwin.win,lgc, 0,yy, xr,yy);
}
#endif

/*  ------------------------------------------------------------  */
#ifdef  WDEBUG
void  redraw_grid(void)
// Draw a grid as a debug aid to alignment
{
int	ca = app->green;
int	cb = app->blue;

/* Draw some horizontal lines */
    hgridline (Line1, ca);
    hgridline (Line2, ca);
    hgridline (Line3, ca);
    hgridline (Line4, cb);

    hgridline (Line5, ca);
    hgridline (Line6, ca);
    hgridline (Line7, ca);
    hgridline (Line8, cb);

    hgridline (Line9, ca);
    hgridline (Line10, ca);
    hgridline (Line11, cb);
    hgridline (Line11a, ca);

    hgridline (Line12, ca);
    hgridline (Line13, ca);
    hgridline (Line14, cb);

    hgridline (Line15, ca);
    hgridline (Line16, ca);
    hgridline (Line17, ca);
    hgridline (Line18, cb);

    hgridline (Line19, ca);

/* Draw some vertical lines */
    vgridline (COL1, ca);
    vgridline (COL1a, app->yellow);
    vgridline (COL2, ca);
    vgridline (COL2a, app->yellow);
    vgridline (COL3, ca);
    vgridline (COL4, cb);

    vgridline (COL5, ca);
    vgridline (COL6, ca);
    vgridline (COL7, cb);

    XSetForeground (mwin.disp, mwin.gfs.gc, app->black);

}
#endif

/*  ------------------------------------------------------------  */
// NOTE -- Add pointer to a menu list as a parameter, ultimaely
// all menus will need to be in that list, and the kind of update
// they get will be determined by the menu block data, specifically
// the menu type will drive whether to draw labels in main window.
//void redraw_menus(void)
void redraw_menus(mble* list)
{
mble	*p;
//u_long	clr;

dpause ("Start redraw_menus");

    C_loop (p, list)  DrawMenu (p->blk); 
//    C_loop (p, list) {
/* Now we define a color for the menu label (to support the gui date
  window which needs to plot black on date choice).  Look at the
  flag field of the menu block to see what color to plot.  */
////	clr = (p->blk->flag & MBred) ? app->red : app->black;
//	if      (p->blk->flag & MBred)	clr = app->red;
//// The red is chosen if both flags are set...
//	else if (p->blk->flag & MByel)	clr = app->yellow;
//	else if (p->blk->flag & MBgrn)	clr = app->green;
//	else 				clr = app->black;
//// The MBred flag bit is set to plot the title color as red.
//	DrawMenu (p->blk, clr);
////	clr = app->black;	// Reset if changed
//    }

dpause ("End redraw_menus");
}

/*  ------------------------------------------------------------  */
/* Items transfered from redraw_menus */
void	redraw_text (void)
{

//    XSetBackground(display,gc,CBX_MAINCOLOR);

/* Draw any text or labels associated with menus */
dpause ("Start redraw_text");
//    draw_label (&mwin, Col1, Line9+TH, "Instrument Details:");
    draw_label (&mwin, COL1, Line9+TH, "Instrument Details:");


//    draw_label (&mwin, Col1, Line12+TH, "Slit Specifications: (ArcSeconds)");
// #ifndef  EXORDX
    draw_label (&mwin, COL1, Line12+TH, "Slit Specifications: (ArcSeconds)");
// #else
//     draw_label (&mwin, COL1, Line12+TH, "Slit Specific'ns: (ArcSeconds)");
// #endif


dpause ("End redraw_text");
}
/*  ------------------------------------------------------------  */
#define  AIRMASS2
//#undef  AIRMASS2
/* Symbol AIRMASS2 used to draw a different color for airmass
  over 2.0 in graphic bar. */

//#define  NAMECOLOR
#undef  NAMECOLOR
/* Symbol NAMECOLOR used to get a named color in the low color
  for the air mass feature.  */

void	redraw_warnings (void)
{
char	text[128];
double	hk;
GC      gc= mwin.gfs.gc;
/* Draw warning messages and flags */

#ifdef  CONSTEL
    {
char	cbuf[64];
    ConstellNames (cbuf, rafc, defc, eqfc);
    sprintf (text, "Position in: %s", cbuf);
    draw_label (&mwin, COL1+2*Xp, Line19-2, text);
    }
#endif  // on CONSTEL

/* Rotator flip warning message */
    if (hrf > 24.0) {
	sprintf (text, "(OK)");
//	XSetForeground (mwin.disp, gc, app->green);	// looks poor
    } else {
	sprintf (text, "!ROTATOR!");
	hk = gsda (30.0, defc, pdx.lat);
//	if (hrf > -6.0 && hrf < 6.0)
//	if (hrf > -hk && hrf < hk)
	if (fabs(hrf) < hk)
		XSetForeground (mwin.disp, gc, app->red);
// We leave the warning black unless above +/- 6 hours.
// Now, leave black unless hrf is above 30 degree altitude...
    }
    draw_label (&mwin, COL6-3*Xp, Line6+TH, text);

/* Draw a graphic warning of rotator flip angles and guider occultation
  areas, at scale... */
    {	// start graphics section
XWindowAttributes	xwa;
GC	lgc = mwin.gfs.gc;
int	xl, xc, xr, xd;
int	kl, kr;
int	y;
int	wid=2*LH;
int	xwd=1;
int	k, j;
int	w;
double	hp, h, s;
//double	sds;
double	hs, hr;
double	hl, hk;
double	ra;
bandlist  *bl;
anglist   *al;
#ifdef  NAMECOLOR
static	u_long	lowcolor = 0;
#endif

// find the scale
    hp = sda (defc, pdx.lat);
//    hp = sda (pdx.dec, pdx.lat);
//    if (h < 1.0) h = 1.0;
    h = (hp < 1.0) ? 1.0 : hp;
// start is at COL1a converted to pixels -- Xloc (COL1a)
    xl = Xloc (COL1a);

// center is at x= 1/2 screen width, y = Yloc(Warnl)
// find width with XGetWindowAttributes(disp, win, &xwa)
// and then use xwa.width
    XGetWindowAttributes (mwin.disp, mwin.win, &xwa);
    xc = xwa.width / 2;
    xd = xc - xl;
    s = xd / h;
// Now, s * ha + xc is x location of that ha.
    xr = xc + xd;
    y = Yloc (Line6 + 16);

/* Set line width and location based on number of lines plotted... */
    if (rats > -24.0) {		// 2 or 3 lines
#ifdef  MOONLINE
	wid = Vunit*Yp;			// 18 (3 * 6)
	y = Yloc (Line7);		// +0 ( 0 + 2*6/3 = 6.0 )
#else
	wid = Vunit*1.3333*Yp;		// 16 (2 * 8.0)
	y = Yloc (Line7 + 2);		// +2 ( 2 + 8.0/2 = 6.0 )
#endif
    } else {			// Single line
	wid = Vunit*1.5*Yp;		//  9 (1 * 9)
	y = Yloc (Line7 + Yp);		// +6 ( 0 + 0 * 9/2 = 6.0 )
    }

/*
  Line7 is Line6 + 15, and Line8 is Line7 + 18.  The graphic space runs
  from Line6 + WH = Line6 + 8 to Line8.  That's Line7 - 7 to Line7 + 18.
  Halfway is Line7 + 5.5.  One line with width 9 needs to be centered at
  Line7 + 5.5, and +6 is close.  Two lines with width 8.0 each need to
  have the first at Line7 + 5.5 - 4.0 = Line7 + 1.5 and +2 is close.
  Three lines with width 6 each need to have the first start at 6 up
  from halfway, or Line7 -0.5, Line7 is close.
  Thus all lines are 0.5 low.  We could drop Line8 by one to center
  better, but having the graphic slightly closer to the date windows
  is a way to associate the graphic with the date values.
.. */

#ifdef  AIRMASS2
// plot the background (beige)
#ifdef  NAMECOLOR
    if (lowcolor == 0)
	lowcolor = get_named_color (mwin.disp, "lightskyblue", app->lgrey);
    XSetForeground (mwin.disp, lgc, lowcolor);
#else  // on NAMECOLOR
    XSetForeground (mwin.disp, lgc, app->lgrey);
#endif  // on NAMECOLOR
    XSetLineAttributes (mwin.disp, lgc, wid, LineSolid,CapButt,JoinMiter);
// or CapButt and JoinMiter
    kl = xc + Nint (s * -hp);
    kr = xc + Nint (s *  hp);
    XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);
// plot the background (green)
    hk = gsda (30.0, defc, pdx.lat);
//    hk = gsda (30.0, pdx.dec, pdx.lat);
    XSetForeground (mwin.disp, lgc, app->green);
    XSetLineAttributes (mwin.disp, lgc, wid, LineSolid,CapButt,JoinMiter);
// or CapButt and JoinMiter
    kl = xc + Nint (s * -hk);
    kr = xc + Nint (s *  hk);
    XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);
#else  // on AIRMASS2
// plot the background (green)
    XSetForeground (mwin.disp, lgc, app->green);
    XSetLineAttributes (mwin.disp, lgc, wid, LineSolid,CapButt,JoinMiter);
// or CapButt and JoinMiter
    kl = xc + Nint (s * -hp);
    kr = xc + Nint (s *  hp);
    XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);
#endif  // on AIRMASS2

// add in occultation warnings (yellow)
    XSetForeground (mwin.disp,lgc, app->yellow);
    for (bl=rotbq; bl!=NULL; bl=bl->next) {
	kl = xc + Nint (s * bl->a);
	kr = xc + Nint (s * bl->b);
	XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);
    }

// add in rotator warnings (red)
    XSetForeground (mwin.disp,lgc, app->red);
    C_loop (bl, rotcq) {
	kl = xc + Nint (s * bl->a);
	kr = xc + Nint (s * bl->b);
	if ( (kr-kl) < wid/2) {
// Use round ends on short line for visibility (special feature)
    XSetLineAttributes (mwin.disp, lgc, wid, LineSolid,CapRound,JoinRound);
	XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);
    XSetLineAttributes (mwin.disp, lgc, wid, LineSolid,CapButt,JoinMiter);
	} else
	XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);
    }

// add in debug rotator warnings (orange)
// these are usually empty; debug is to show any which might be non-empty.
  if (listcount(rotdq) > 0) {
    XSetForeground (mwin.disp,lgc, app->orange);
    C_loop (bl, rotdq) {
	kl = xc + Nint (s * bl->a);
	kr = xc + Nint (s * bl->b);
	XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);
    }
  }
//    C_loop (bl, roteq) {
//	kl = xc + Nint (s * bl->a);
//	kr = xc + Nint (s * bl->b);
//	XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);
//    }

// plot any flip angles (red)
// These might still exist for some instruments...
    XSetForeground (mwin.disp,lgc, app->red);
    XSetLineAttributes (mwin.disp, lgc, wid, LineSolid,CapRound,JoinRound);
    C_loop (al, rotaq) {
	kl = xc + Nint (s * al->v);
// Draw 0 length line with round ends to get a dot...
// Could also draw a dot directly.
	XDrawLine (mwin.disp,mwin.win, gc, kl, y, kl, y);
    }

// Add a warning for near the zenith (blue)
    if (fabs(defc - pdx.lat) < 0.6666666) {
//    if (fabs(pdx.dec - pdx.lat) < 0.6666666) {
	XSetForeground (mwin.disp,lgc, app->blue);
	XDrawLine (mwin.disp,mwin.win, gc, xc, y, xc, y);
    }

// Return the width attribute...
    XSetLineAttributes (mwin.disp, gc, xwd, LineSolid,CapRound,JoinRound);
    XSetForeground (mwin.disp,lgc, app->black);

// draw and label an axis
    XDrawLine (mwin.disp,mwin.win,lgc, xl, y, xr, y);
    k = h;
    for (j=-k; j<=k; j++) {	// Tick marks on axis
	w = 2;
	if ( (j%5) == 0) w = 4;
	kl = xc + Nint (s * j);
	XDrawLine (mwin.disp,mwin.win,lgc, kl, y-w, kl, y+w);
    }
    w = wid/2;
    for (j=xl; j<=xr; j+=xd) {	// Center and ends
	XDrawLine (mwin.disp,mwin.win,lgc, j, y-w, j, y+w);
    }

/* Add section for sst line if a date is present */
// Check for date, and if present find dec and ra of true sun
// Those ra/dec values are put in global values by the date
// window routine, default of -25h ra when not present.
// if date not present, whole thing is skipped...
    if (rats > -24.0) {
// draw background in yellow, just below the above line
// i.e. y += wid to get new line center
// Set line attributes to square ends...
	y += wid;
	XSetForeground (mwin.disp,lgc, app->yellow);
	XSetLineAttributes (mwin.disp, lgc, wid, LineSolid,CapButt,JoinMiter);
	kl = xc + Nint (s * -hp);
	kr = xc + Nint (s *  hp);
	XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);

/* --
  compute twilight hour angles of sun by finding solar sda
  and adding to solar ra, then differencing from object ra,
  with proper set to -12/+12 range.
  Limit twilight to current hp range and plot as blue.
  May have two parts; start with (limited) sunset and continue to
  either high limit or sunrise.  If stopped at high limit, do another
  starting at low limit and continue to sunrise, skip if limited so
  both would be same.       -- */

//    hp = sda (pdx.dec, pdx.lat);
//	dec = *(double*)DECfield.data / Degree;
	ra = rafc;
//	ra = *(double*)RAfield.data / Hour;

// compact method...
// Plot the twilight line in blue, use sunset/rise
	XSetForeground (mwin.disp,lgc, app->blue);
//	sds = gsda (-1.0, dets, pdx.lat) * Sidereal;	// Values in deg, hour
//// Better estimate is to use altitude of -1.0 for sunset and 0 for rise.
//	hs = fdifmod (rats+sds, ra, 24.0);	// ha sunset
//	sds = gsda (0.0, dets, pdx.lat) * Sidereal;	// Values in deg, hour
//// Eastern horizon is much higher than western one...
//	hr = fdifmod (rats-sds, ra, 24.0);	// ha sunrise

	hs = fdifmod (stm+sst, ra, 24.0);
	hr = fdifmod (stm+srt, ra, 24.0);

	hl = (hs > -hp) ? hs : -hp;	// left limit of sunset
	hk = (hr <  hp) ? hr :  hp;	// Right limit of sunrise
	if (hl < hk) {
	    kl = xc + Nint (s * hl);
	    kr = xc + Nint (s * hk);
	    XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);
	} else {
	    if (hk < -hp) hk = -hp;
	    kl = xc + Nint (s * -hp);
	    kr = xc + Nint (s *  hk);
	    XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);
	    if (hl > hp) hl = hp;
	    kl = xc + Nint (s *  hl);
	    kr = xc + Nint (s *  hp);
	    XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);
	}

// Do similar comp. for darkness using solar/twilight sda,
// and plot in black.
	XSetForeground (mwin.disp,lgc, app->black);
//	sds = gsda (-18.0, dets, pdx.lat) * Sidereal;	// Values in deg, hour
//// -18 is astronomical twilight independent of horizon
//	hs = fdifmod (rats+sds, ra, 24.0);	// ha sunset
//	hr = fdifmod (rats-sds, ra, 24.0);	// ha sunrise

	hs = fdifmod (stm-twit, ra, 24.0);	// ha sunset
	hr = fdifmod (stm+twit, ra, 24.0);	// ha sunrise

	hl = (hs > -hp) ? hs : -hp;	// left limit of sunset
	hk = (hr <  hp) ? hr :  hp;	// Right limit of sunrise
	if (hl < hk) {
	    kl = xc + Nint (s * hl);
	    kr = xc + Nint (s * hk);
	    XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);
	} else {
	    if (hk < -hp) hk = -hp;
	    kl = xc + Nint (s * -hp);
	    kr = xc + Nint (s *  hk);
	    XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);
	    if (hl > hp) hl = hp;
	    kl = xc + Nint (s *  hl);
	    kr = xc + Nint (s *  hp);
	    XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);
	}

// plot a short (5pixel?) line centered on solar midnight, which
// is ha of anti-solar ra point; if it is on graph.
	hs = fdifmod (rats+12.0, ra, 24.0);	// ha midnight
	if (hs > -hp && hs < hp) {
	    XSetForeground (mwin.disp,lgc, app->green);
	    kl = xc + Nint (s * hs) - 2;
	    kr = kl + 4;
	    XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);
	}

#ifdef  MOONLINE
// Plot the moon line stuff here...

// draw background in blue, just below the above line
// i.e. y += wid to get new line center
// Set line attributes to square ends...
	y += wid;
	XSetForeground (mwin.disp,lgc, app->blue);
	XSetLineAttributes (mwin.disp, lgc, wid, LineSolid,CapButt,JoinMiter);
	kl = xc + Nint (s * -hp);
	kr = xc + Nint (s *  hp);
	XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);

// Plot the moon above horizon in white or light grey
	XSetForeground (mwin.disp,lgc, app->offwhite);

	hs = fdifmod (ltt+stlm+mrt, ra, 24.0);	// ha moonrise
	hr = fdifmod (ltt+stlm+mst, ra, 24.0);	// ha moonset

	hl = (hs > -hp) ? hs : -hp;	// left limit of moonrise
	hk = (hr <  hp) ? hr :  hp;	// Right limit of moonset
	if (hl < hk) {
	    kl = xc + Nint (s * hl);
	    kr = xc + Nint (s * hk);
	    XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);
	} else {
	    if (hk < -hp) hk = -hp;
	    kl = xc + Nint (s * -hp);
	    kr = xc + Nint (s *  hk);
	    XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);
	    if (hl > hp) hl = hp;
	    kl = xc + Nint (s *  hl);
	    kr = xc + Nint (s *  hp);
	    XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);
	}

// plot a short (3pixel?) line centered on lunar transit, if it
// is on the graph...
	hs = fdifmod (ltt+stlm, ra, 24.0);	// ha moon transit
	if (hs > -hp && hs < hp) {
	    XSetForeground (mwin.disp,lgc, app->brown);
	    kl = xc + Nint (s * hs) - 1;
	    kr = kl + 2;
	    XDrawLine (mwin.disp,mwin.win, lgc, kl, y, kr, y);
	}

// Add the text of the pci to the right of the line...

	{	// data section
char	buf[8];
int	lb;
//	    sprintf (buf, "%02d%%\0", pci);
	    sprintf (buf, "%02d%%", pci);
	    lb = strlen (buf);
	    XSetForeground (mwin.disp,lgc, app->black);
	    XDrawString (mwin.disp,mwin.win,mwin.gfs.gc,
		xr+(int)(Hunit*3.0), y + (int)(Vunit*2.0), buf, lb);

/* and add adm if requested */
#ifdef  MOONANG
	    sprintf (buf, "%3d", adm);
	    lb = strlen (buf);
	    XSetForeground (mwin.disp,lgc, app->black);
	    XDrawString (mwin.disp,mwin.win,mwin.gfs.gc,
		xl-(2*lb-1)*(int)(Hunit*3.0), 
		y + (int)(Vunit*2.0), buf, lb);
// Hunit is 1/6 size of a character...
#endif  // on MOONANG

	}	// end section

#endif  // on MOONLINE

    }	// end if on valid rats value

// Return the width attribute...
    XSetLineAttributes (mwin.disp, gc, xwd, LineSolid,CapRound,JoinRound);
    XSetForeground (mwin.disp,lgc, app->black);

    }	// End graphics section

}



/*  ------------------------------------------------------------  */
