/*  guiutil.c  == Utilities for gui development  */
/* .. An extension to the libcxt.a library .. */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2010 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Many of these programs were developed for the intgui.c program, but
  have a general utility beyond that project.  By operating on a common
  set of data structures, we can extend the generality of the handling
  of windows, menus and buttons.  */

/*   GUISUBS == guisubs.c is a portion of the intgui.c code which
  is used in parallel by xmap.c while the latter program is being
  made ready to integrate into intgui.c
  When fully integrated, xmap ceases to exist and the code here
  will be returned to intgui.c    */
/* The above is technically wrong -- this code was removed from intgui
  into a separate library module.   */

/* Defines */
#define  IS_GUIUTIL_C

/* Includes */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include	"mgutils.h"	// for lint only
#include	"guiutil.h"

/* Externs */

/* Largely removed as application dependent. */

// extern	MainWindow	mwin;

// STICKY PART -- the mwin pointer is needed for most creation of
// subwindows, and must be passed as a parameter.  That means lots
// of subroutine calls need this passed to them.
// We could make this pointer a member of the structure of
// winblock, and thus run things in multiple windows from a single
// list, possibly.
// So, we make *mwin a part of each winblock, butblock, etc.
// That means all the data statements defining all the windows need
// to be updated, but they also need offset, callback too.

// Where mwin is not passed in a block, as in draw_label, it is
// passed as an explicit pointer.



/*  -----  Dynamic Positioning support  -----  */
/*  THE FONT SIZE LOCATION ROUTINES MAY BE APP DEPENDENT, THEY REQUIRE
  THE PREVIOUS DEFINITION OF HUNIT AND VUNIT AS GLOBALS... */
int	Xsize (int c)
{
double	hs;
int	v;
// Input is horizontal units, output is rasters horizontally
// There are Xp units per character
//    hs = Hunit * Xp * (0.5 + (double)c);  // Used c in characters
    hs = Hunit * (0.5 + (double)c);
    v = Nint (hs);
    return  v;
}

int	Xloc (int x)
{
double	hp;
int	v;
    hp = 4.0 + Hunit * (double)x;
    v = Nint (hp);
    return  v;
}

int	Yloc (int y)
{
double	vp;
int	v;
    vp = 2.0 + Vunit * (double)y;
    v = Nint (vp);
    return  v;
}


/*  -----  List support code  -----  */

/*  ------------------------------------------------------------  */
wble*	wlpush (wble* a, wble* b)
{
wble	*p, *r;
    if (b == NULL) return a;
    if (a == NULL) return b;
    p = b->last;
    if (p == NULL) p = b;       /* ...p - b... */
    r = a->last;
    if (r == NULL) r = a;       /* ...r - a... */
    a->last = p;
    p->next = a;
    r->next = b;
    b->last = r;        /* ...p - a...r - b... */
    return  b;
}

/*  ------------------------------------------------------------  */
wble*	wlpop (wble* a)
{
wble	*p, *q;
    if (a == NULL) return a;
    p = a->next;
    if (p == a) p = NULL;
    q = a->last;
    if (q == a) q = NULL;
    if (p == NULL) p = q;
    if (p != NULL) p->last = q;
    if (q != NULL) q->next = p;
    a->last = a->next = a;
    return p;
}

/*  ------------------------------------------------------------  */
wble*	wladd (wble* a, winblock* b)
{
wble	*d;
    if (b == NULL) return a;
    d = (wble*) malloc (sizeof(wble));
    d->last = d->next = NULL;
    d->blk = b;
    return  wlpush (a, d);
}

/*  ------------------------------------------------------------  */
bble*	blpush (bble* a, bble* b)
{
bble	*p, *r;
    if (b == NULL) return a;
    if (a == NULL) return b;
    p = b->last;
    if (p == NULL) p = b;       /* ...p - b... */
    r = a->last;
    if (r == NULL) r = a;       /* ...r - a... */
    a->last = p;
    p->next = a;
    r->next = b;
    b->last = r;        /* ...p - a...r - b... */
    return  b;
}

/*  ------------------------------------------------------------  */
bble*	blpop (bble* a)
{
bble	*p, *q;
    if (a == NULL) return a;
    p = a->next;
    if (p == a) p = NULL;
    q = a->last;
    if (q == a) q = NULL;
    if (p == NULL) p = q;
    if (p != NULL) p->last = q;
    if (q != NULL) q->next = p;
    a->last = a->next = a;
    return p;
}


/*  ------------------------------------------------------------  */
bble*	bladd (bble* a, butblock* b)
{
bble	*d;
    if (b == NULL) return a;
    d = (bble*) malloc (sizeof(bble));
    d->last = d->next = NULL;
    d->blk = b;
    return  blpush (a, d);
}

/*  ------------------------------------------------------------  */
mble*	mladd (mble* a, mnublock* b)
{
mble	*d;
    if (b == NULL) return a;
    d = (mble*) malloc (sizeof(mble));
    d->last = d->next = NULL;
    d->blk = b;
    return  G_push_cl (a, d);
}


/*  -----  Below is the code for Xwindow stuff  ----- */

/*  ------------------------------------------------------------  */
double	pdecode (char* field)
/* Recursively decode an entered value as sexigisimal;
  extension of posvalue in ioutils.  Takes fields separated
  by either colon or space using parse... */
{
char	buf[16];	// The current string found
char	*p;		// The "next" field
double	a;		// Value being found
int	k;

// Take out any null cases
    if (field == NULL) return 0.0;
    if (field == AN) return 0.0;

// Get rid of (recursively) leading whitespace and signs
    if (*field == ' ' || *field == '\t') return pdecode (field+1);
    if (*field == '-') return  -pdecode (field+1);

// Parse the current value
    p = parse (field, buf, ": \t");
    if (*buf == '-') return  pdecode (field+1);
// note that if *field were the - it gets negated above

    k = sscanf (buf, "%lf", &a);
    if (k != 1) a = 0.0;
    a += pdecode (p) / 60.0;
    return a;
}


/*  ------------------------------------------------------------  */
winblock*	BlockOfWin (wble* list, EditWindow* ew)
// Return the window block (if any) pointing to given edit window ew
{
wble	*lp;
winblock  *wb;


// Actually list is "winlist" in all cases, a global...
// we could use "list = winlist" here...

// Find pointer ew in a window block, return that block's pointer
    for (lp=list; lp != NULL; lp = lp->next) {
	wb = lp->blk;
	if (wb->ew == ew) return wb;
	if (lp->next == list) break;  // not found...
    }
    return  NULL;
}

/*  ------------------------------------------------------------  */
butblock*	BlockOfBut (bble* list, Button* bx)
// Return button block (if any) pointing to given button bx
// ANALOG TO FINDING WINDOW FROM ITS LIST...
{
bble	*lp;
butblock  *bb;

/*  TODO:
  Create a pointer "butlist" as a global to point to a
  list of buttons.  Creation puts buttons onto this list.
  > Create a data type for the list in intgui.h, like wble
  > Also make a menu block list in the same way
  > Add to creation routine the creation of a list block,
    and push it onto the list automatically
  > Call this program in button handler; get the block and
    if it exists the data pointer, then set the data pointer
    based on the state of the data item.
  > When creating the button, initialize the data from the
    listed data value.  Pass data pointer in create call just
    as is done with the windows.
  > Do the creation similarly to window; and handle the data
    in/out similarly.  When creating, set toggle by the data
    present; when toggling, set data to toggled value.

..  */

// Actually list is "winlist" in all cases, a global...
// we could use "list = winlist" here...
// Change as that for the button block case using a global -- butlist

// Find pointer bx in a button block, return that block's pointer
    for (lp=list; lp != NULL; lp = lp->next) {
	bb = lp->blk;
	if (bb->bu == bx) return bb;
	if (lp->next == list) break;  // not found...
    }
    return  NULL;
}

/*  ------------------------------------------------------------  */

/* The problem with ToggleButton is that it is required to be
  app dependent.  It is a callback, with call sequence defined, and
  can't get a list pointer in there (unless we create a structure to
  pass, but param is defined to be a button pointer.  And we need to
  get the button block to toggle the status in the button block.
  Maybe, get the things which use the button block status to actually
  look at the button's own status?  */

/*  ------------------------------------------------------------  */
int	TextScan (wble* list)
// Scan all text windows and put current text values into
// the proper structure location...
// This routine is used to synchronize the window text values with
// the structure containing the real data.
{
wble	*le;
winblock  *wb;
int	n, k;
char	*p;
int	change=0;


    for (le=list; le != NULL; le = le->next) {
	wb = le->blk;

/* Get the data from the window text block into the data field
  specified by winblock pointer wb.  */
	switch (wb->type) {
	    case  TEXT:
		n = wb->width;
/* Get up to n characters for data field. */
		k = strlen (wb->ew->text);
		if (k > n) wb->ew->text[n] = AN;
// CHECK CHANGES
		change |= strcmp (wb->data, wb->ew->text);
		strncpy (wb->data, wb->ew->text, n);
		break;
	    case  DTEXT:
		p = *(char**)wb->data;
//		p = (char*)wb->data;
// Note: the above two are not equivalent...
// wb->data for dynamic text is a pointer to the location in which
// a pointer to the dynamic string is stored.  It is NOT a pointer to
// the data itself, but a pointer to the location in obs structure
// where the character pointer is found.

/* NOTE -- Here's the cause of some crashes.
  It is possible for p above to be NULL, which occurs if there is
  no title present at the start.
  Also, if no title has been inserted, the wb->ew->text field will
  have a zero length, which results in a NULL pointer returned when
  it is placed in a dynamic string.
  THEREFORE, we need to always test p for NULL to avoid a crash! */

		if (p != NULL) {	// debug for all this conditional
		    change |= strcmp (p, wb->ew->text);
		    free (p);
		} else change |= (strlen(wb->ew->text) > 0);
/* Above, we note a change if the text changed (for non-null text), 
  or the text is now present if it was formerly null */

		p = dynamstr (wb->ew->text);

//		wb->data = (void*)p;
// Note: the above is wrong...
		*(char**)wb->data = p;
// if (wb == &titlefld && p != NULL) printf (" ..Title: <%s>\n", p);
		break;
	    default:
		break;
	    }
	if (le->next == list) break;  // end
    }
    return  change;
}

// Globals accessed:  winlist (could be param)
//	gob-changed (could be return)

// int	UpdateTextWindow (winblock* wb);
/* The above  Function Prototype  is needed to properly link to the
  update program needed to process updatds to other windows caused
  by changed to some text windows in the application.  It will be
  called by the text window handler below.    */

int	Call_Updater (winblock* wb)
/*  Find the updating call, if any, in wb, and call it using
  wb as the argument.  Return what was found.
  The updater program is put in winblock for application independence.    */
{
int	r = 0;

/* Check null cases */
    if (wb == NULL) return  r;
    if (wb->updater == NULL) return  r;
    r = (wb->updater)(wb);
    return  r;
}

/*  ------------------------------------------------------------  */
int    HandleTextWindow (wble* list, EditWindow* w)
{
winblock	*wb;
double		val;
int		change=0;

// Get the winblock
    wb = BlockOfWin (list, w);

// Check for null case
    if (wb == NULL) return change;

/* Put the data from the data field specified by winblock pointer wb 
  into the window text block, correctly formatted.  NOTE that automatic 
  windows such as text windows do not come through this program and are 
  handled by a Text_Scan program elsewhere... */
/* This routine is usually called when we send a carriage return while
  the window is active with a green border.  It is called from
  TextBack only.  (NOT only) */
/* This is also called by ANY routine wishing to update the automatic
  window w after changing its data value! */
    switch (wb->type) {
    case  TEXT:
// Don't bother to count; probably can't exceede...
//	strcpy (w->text, wb->data);
// Should count at wb->width...
	change |= strncmp (wb->data, w->text, wb->width);
	strncpy (w->text, wb->data, wb->width);
	w->text[wb->width] = AN;
	break;
    case  DTEXT:
	change |= strcmp (*(char**)wb->data, w->text);
	strcpy (w->text, *(char**)wb->data);
// Don't bother to count, or limit to size of text field in window def...
	break;
    case  ITEXT:
	sprintf (wb->ew->text, "%*d", 
	    wb->dp, *(int*)wb->data);
	change |= 1;	// Real change?
	break;
    case  FTEXT:
//	*(double*)wb->data = wb->scale * atof (wb->ew->text);
	val = *(double*)wb->data/wb->scale;
//	if (wb == &slitpaf) val += 90.0;
	val += wb->offset;
	sprintf (wb->ew->text, "%.*f", 
	    wb->dp, val);
	change |= 1;	// Real change?
	break;
    case  DMS:
//	*(double*)wb->data = wb->scale * pdecode (wb->ew->text);
//	strcpy (wb->ew->text,  
//	    sexig( *(double*)wb->data/wb->scale, 3, wb->dp) );
	strcpy (wb->ew->text, " ");
	strcat (wb->ew->text,  
	    sexig( *(double*)wb->data/wb->scale, 3, wb->dp) );
	change |= 1;	// Real change?
	break;
    }
//    CBX_UpdateEditWindow(wb->ew);
// or reset edit window?  the reset also sets it to normal if not disabled.
//    CBX_ResetEditWindow (wb->ew);

/* Call for general updates due to this window changing.  */
//    change |= UpdateTextWindow (wb);
/* The actual program called is supplied in the winblock by the application,
  as it will be an application dependent code, but needs to be called by
  this application independent module.  We use a general calling program
  to properly do the needed NULL checks.      */
    change |= Call_Updater (wb);
/* When the same thing is done for buttons, menus, etc. we should rename
  to Window_Updater, Button_Updater, Menu_Updater, etc.  and be sure to
  provide the proper routine to the butblock, mnublock, etc.      */

// Call the general update for the window
// That will put the data into the text window,
    CBX_ResetEditWindow (w);
    return  change;
}

// globals used:
//	winlist  (could be param)
//	gob-changed (could be return)
//	&slitpaf  -- app specific window change; can be
//		generalized by adding an offset field to window block.


/*  ------------------------------------------------------------  */

#ifdef  USESETEDITWINDOW
// Intended to be undefined
void	SetEditWindow (wble* list, EditWindow* w, char* data)
/* Set the data field of w to the data value, then do a 
  HandleTextWindow operation on the window.  Note that the data
  is the text we wish to set into the window, and we de-format
  it if needed into an actual data value.   */ 
{
winblock	*wb;
double		val;
//int		change=0;

// Get the winblock
    wb = BlockOfWin (list, w);

// Check for null case
    if (wb == NULL) return;

// copy the action from handletextwindow above, with change of
// setting the data field rather than using existing data...

/* "read" the data if needed to integer, float and then put it
  into the wb->data area, and limit to wb.width if text.  */

/* The appropriately de-formatted data needs to be put in w->text,
  probably the data itself is good enough.   */

/* Put the data given into the window text block, and hope that it
  is correctly formatted.  It will be if the caller is good.  */

/* Then, depending on the data type, "read" the data into an appropriate
  variable, and put that in the wb data field if possible.  */

/* This should have the effect of HandleTextWindow without needing
  to call that program directly.  
  We do this because we need to decode the data based on type anyway.  */

/* THESE COMMENTS ARE FROM HandleTextWindow WHERE WE GOT THE CODE */
/* Put the data from the data field specified by winblock pointer wb 
  into the window text block, correctly formatted.  NOTE that automatic 
  windows such as text windows do not come through this program and are 
  handled by a Text_Scan program elsewhere... */
/* This routine is usually called when we send a carriage return while
  the window is active with a green border.  It is called from
  TextBack only.  (NOT only) */
/* This is also called by ANY routine wishing to update the automatic
  window w after changing its data value! */
/* END OF COMMENTS FROM HandleTextWindow.    */
    switch (wb->type) {
    case  TEXT:
// Don't bother to count; probably can't exceede...
////	strcpy (w->text, wb->data);
// Should count at wb->width...
//	change |= strncmp (wb->data, w->text, wb->width);
	strncpy (wb->data, data, wb->width);
	strncpy (w->text, wb->data, wb->width);
	w->text[wb->width] = AN;
	break;
    case  DTEXT:
//	change |= strcmp (*(char**)wb->data, w->text);
	strcpy (*(char**)wb->data, data);
//	strcpy (w->text, *(char**)wb->data);
	strcpy (w->text, data);
// Don't bother to count, or limit to size of text field in window def...
	break;
    case  ITEXT:
//	sprintf (wb->ew->text, "%*d", 
//	    wb->dp, *(int*)wb->data);
//	change |= 1;	// Real change?
//	strcpy (wb->ew->text, data);
	strcpy (w->text, data);
	*(int*)wb->data = atoi (data);
	break;
    case  FTEXT:
	*(double*)wb->data = atof (data);
// follows things from HandleTextWindow
//	*(double*)wb->data = wb->scale * atof (wb->ew->text);
	val = *(double*)wb->data/wb->scale;
//	if (wb == &slitpaf) val += 90.0;
	val += wb->offset;
	sprintf (wb->ew->text, "%.*f", 
	    wb->dp, val);
//	change |= 1;	// Real change?
	break;
    case  DMS:
	*(double*)wb->data = pdecode (data);
//	*(double*)wb->data = wb->scale * pdecode (wb->ew->text);
//	strcpy (wb->ew->text,  
//	    sexig( *(double*)wb->data/wb->scale, 3, wb->dp) );
	strcpy (wb->ew->text, " ");
	strcat (wb->ew->text,  
	    sexig( *(double*)wb->data/wb->scale, 3, wb->dp) );
//	change |= 1;	// Real change?
	break;
    }
//    CBX_UpdateEditWindow(wb->ew);
// or reset edit window?  the reset also sets it to normal if not disabled.
//    CBX_ResetEditWindow (wb->ew);


// Call the general update for the window
// That will put the data into the text window,
    CBX_ResetEditWindow (w);


}
#endif  // on  USESETEDITWINDOW

// the data pointer in winblock is typically set with AutoEditWin (below)
// which depends on the type, which we may use to select how to
// set the data

//int    HandleTextWindow (wble* list, EditWindow* w)

/* Could also write some specific programs, such as 
  void  SetIntWindow (wble* list, EditWindow* w, int v)
  void  SetFloatWindow (wble* list, EditWindow* w, double v)
  which would set the wb->data appropriately with v, and
  then just call HandleTextWindow to get it set.
...  */

// void	TextBack	(void* param);
/* The above  Function Prototype  is here while we call TextBack from
  this module.  It is an external defined in intgui.h and coded into
  guisubs.c which is included by intgui.c and others.  Once the specific
  code from TextBack is moved to HandleTextWindow (here, just above)
  there will no longer be a call to TextBack from here, and the
  Function Prototype will be removed.    */

void	SetFloatWindow (wble* list, winblock* wb, double v)
{
    if (wb == NULL) return;
/* wb->type should be FTEXT or DMS */
    if (wb->data != NULL) *(double*)wb->data = v;
    HandleTextWindow (list, wb->ew);
//    TextBack (wb->ew);
}

/* The HandleTextWindow call will take the value which has been put into
  the wb->data location, and write it into the window.  Then, TextBack
  will take the window's value and use it to update any other items such
  as rotator flip warnings.  It also calls HandleTextWindow.

  NOTE that this is probably the wrong way to do it -- the updating of
  those things should be in HandleTextWindow and not in TextBack 
  (which is located in guisubs.c).  Then TextBack will call HandleTextWindow
  and so will these routines.  Done this old way there is a duplication
  of effort in calling TextBack again, and values are translated from the
  text versions rather than being entered more accurately.  Fixing it
  will require moving code from TextBack to HandleTextWindow.      */

/*  NOTE that the proposed change makes this module application dependent.
  A method is needed to call the application dependent update routines which
  must be done when certain values change, and in proper order.

  TextBack takes data from window text block into winblock data field.
  It is an external defined in intgui.h and coded in guisubs.c which
  is included by intgui.c.  It is usually called when the window text
  has changed and the data is to be entered.

  HandleTextWindow takes data from winblock data field to window text block.
  It is in this module (just above) and is supposed to be application
  independent.  It does NOT call TextBack.  It is called when the data
  has changed and the window should be updated to match.

  Usually, TextBack is called when a window is edited, and HandleTextWindow
  is called when the program sets the data field.  In both cases, application
  specific changes need to be recognized.  Previously, TextBack would call
  HandleTextWindow.

  The HandleTextWindow should be primary so the data field can be set
  properly when needed.  It can call a set of application dependent
  code.  Similarly, TextBack is secondary, it should NOT be called from
  HandleTextWindow.

  The application specific change detection and globalization program
  should be in guisubs, should detect changes in either data or window text.
  The compare for change should be done after converting the primary source,
  comparing to the old secondary source.  HandleTextWindow will return the
  change status, TextBack has no return (void by definition) but can set
  the Obs->changed global as found.  The new update program called by both
  (since neither will now call the other) will return change status.  This
  update program will need to call HandleTextWindow for windows other than
  the one it was called for.
  
  A function prototype for the update program will be needed here.

  It will be important for the update program to not call itself (via the
  call to HandleTextWindow) and get in a loop.
  Also, the update program should be in guisubs.c and not in this module
  which should stay application independent as far as possible.
  The calls to TextBack in the SetxxxWindow will be removed and their
  action replaced by a call to the update program from HandleTextWindow.
  New program should be Update Text Window (without spaces).

  Calls go like this:
  HTW(w) -> UTW(w) -> HTW(others)
  TB(w) -> HTW(w) (and that calls UTW as above)
  SFW(wb,v) Set wb.data to v -> HTW(w)

  The old way was:
  TB(w) - Update, HTW(others), -> HTW(w)
  HTW(w) (no update done)

  The binary data is primary.  If it is set directly, call HTW.  If the
  window is edited, TB is called, and that sets binary data from the
  content of the window, and calls HTW to provide a good normalized and
  properly formatted window content value.  The primary path is:
  Get Binary Data -> HTW to set window -> UTW to update globals, others.
  When UTW updates data, it does it as above.

  The whole point is to get that new update program in a single place, to
  do things like compute new flip warnings based on dec or position angle,
  and to omit some warnings based on instrument (gisap/wfccd), and to
  do other HandleTextWindow calls for IFU windows and secondary wavelength
  windows which get their values changed.			*/


void	SetIntWindow (wble* list, winblock* wb, int v)
{
    if (wb == NULL) return;
/* wb->type should be ITEXT */
    if (wb->data != NULL) *(int*)wb->data = v;
    HandleTextWindow (list, wb->ew);
//    TextBack (wb->ew);
}



/*  ------------------------------------------------------------  */
// NOTE -- alternate method:  Could make this a void, and simply
// pass the list as a meta-pointer, and update it here.  That would
// cut down the code in the calls, of which there are many.
// Do the same for buttons, with a list pointer..
// Also, do similar thing for menus..

// wble*	AutoEditWin (wble* list, winblock* wb, void *data)
void	AutoEditWin (wble** wlist, winblock* wb, void *data)
{
// Like old MakeEditWin below, but makes automatic window...
int	x, y;
int	w, h;
int	t, ts;
char	*p=NULL;
char	buf[128];
int	mode;
double	val;

/* Find the proper window location and size */
//    t = strlen (wb->label);
//    x = Xloc(wb->xloc);   // was ..  + Margin1;
//    if (t > 0) x += Xsize (Xp*t + 2);  // was .. CharW * (t + 1);
//    y = Yloc(wb->yloc);

/* Find proper window size */
//    w = Xsize (Xp*(wb->width + wb->dp)+2);	// Add 1/2 char to most win's
//    h = (int)(wb->height*Vunit + 1);


/* Find the horizontal sizes of things, and location */
    t = strlen (wb->label);			// Text length of label
    ts = (t > 0) ? Xsize (Xp*t +2) : 0;		// Label size
    x = Xloc (wb->xloc);			// Aligned location
    w = Xsize (Xp*(wb->width + wb->dp)+2);	// Window width
    switch (wb->align) {
    case  Title:
    default:
	wb->laboff = 0;
	x += ts;
	break;
    case  Leftside:
	wb->laboff = -ts;
	break;
    case  Rightside:
	wb->laboff = -(ts+w);
	x -= w;
	break;
    }

/* Find the vertical size and location */
    y = Yloc(wb->yloc);
    h = (int)(wb->height*Vunit + 1);


/* Fill in the data pointer provided. */
    wb->data = data;

/* Translate data to text pointer to create window */
/* Also, come up with the mode for the auto window... */
    mode = 0;
    switch (wb->type) {
    case  TEXT:
	strncpy (buf, data, wb->width);
	buf[wb->width] = AN;
	p = buf;
/* Assure only the fixed size goes into window buffer */
	break;
    case  DTEXT:
	p = *(char**)data;  
	break;
    case  ITEXT:		// Integer value
	p = buf;
	sprintf (p, "%d", *(int*)wb->data);
	mode |= CBX_VALUE | CBX_OVERWRITE;
	w -= (int)Hunit;	// Trim un-necessary extra space
	break;
    case  FTEXT:		// Floating value
	p = buf;
	val = *(double*)wb->data/wb->scale;
//	if (wb == &slitpaf) val += 90.0;
	val += wb->offset;
	sprintf (p, "%.*f", wb->dp, val);
	mode |= CBX_VALUE | CBX_OVERWRITE;
	w -= (int)Hunit;	// Trim un-necessary extra space
	break;
    case  DMS:			// Sexigisimal value
//	strcpy (buf, sexig( *(double*)wb->data/wb->scale, 3, wb->dp));
	strcpy (buf, " ");
	strcat (buf, sexig( *(double*)wb->data/wb->scale, 3, wb->dp));
	p = buf;
	mode |= CBX_OVERWRITE;
/* Kludge width increment needed for non-VALUE fields... */
	w += Xsize (Xp/2);
/* We NEED to add one space to the window size, so that the editing
  algorithm in the CXT library will properly position the string and
  the cursor.  It does look ugly, but that's the present work-around
  for making a window.  It is only needed for fields which do not
  have the VALUE attribute, and are not text. */
// If we add Xsize(Xp)/2 we get left justified;
// If we add Xsize(Xp) we get a space to the left (when provided with
// a leading space in the text handling routine).
// At font height 18 we need w = +4 or more to get 1 character;
// and we need w += 15 to get 2 characters additional...
	break;
    }

/* Kludge width increment needed for non-VALUE fields... */
//    if (!(mode & CBX_VALUE)) w += Xsize (2) ;

/* Create the edit window */
    (void)CBX_CreateAutoEdit (wb->mwin, wb->ew,  x, y,  w, h,  
		p, mode, wb->callback );
//		p, mode, TextBack );

/* Put the window in the list */
//    return  wladd (list, wb);
// change to *wlist = wladd (*wlist, wb);
    if (wlist != NULL) *wlist = wladd (*wlist, wb);
}

// app specific things:
// global variables: 
//	slitpaf -- add the offset to the window block
//	textback -- itself app dependent.  pass as a parameter!


/*  ------------------------------------------------------------  */
void	AlignMenu (mnublock* mnu, int* x, int* t, int* w)
{
int	ts;		// Text label size


/* Find the horizontal sizes of things, and location */
    *t = strlen (mnu->label);			// Text length of label
    ts = (*t > 0) ? Xsize (*t * Xp +1) : 0;	// Label size (was +2)
    *x = Xloc (mnu->xloc);			// Aligned location
    *w = Xsize (Xp*mnu->width + 3);		// Menu width with circle
/* How much we add depends on type of menu; check type field */
    switch (mnu->align) {
    case  Title:
    default:
	mnu->laboff = 0;
	*x += ts;
	break;
    case  Leftside:
	mnu->laboff = -ts;
	break;
    case  Rightside:
	mnu->laboff = -(ts + *w);
	*x -= *w;
	break;
    }
}

// NOTE -- uses xsize, which needs to be generalized above!!

/*  ------------------------------------------------------------  */
/* The MakePMenu subroutine is obsolete.  Use MakeMenu instead, with
  the proper type field in the menu block.  */
void	MakePMenu (mble** L, mnublock* mnu)
{
int	x, y;
int	w, h;
int	t;

/* Find the proper window location and size */
    AlignMenu (mnu, &x, &t, &w);
//    t = strlen (mnu->label);
//    x = Xloc (mnu->xloc);
//    if (t > 0) x += Xsize (Xp*t + 2);  // CharW * (t + 1);
    y = Yloc(mnu->yloc);

/* Find proper window size */
//    w = Xsize (Xp*mnu->width + 3);	// Allow small circle at end
    h = (int)(mnu->height*Vunit + 1);

    mnu->type = M_POP;	// force type to comply with standard...

/* Create the menu window */
    (void)CBX_CreatePopupMenu (mnu->mwin, mnu->mu,  x, y,  w, h );
    if (L != NULL) *L = mladd (*L, mnu);
}

/*  ------------------------------------------------------------  */
/* Previously known as MakeAMenu, now the general menu maker */
void	MakeMenu (mble** L, mnublock* mnu)
{
int	x, y;
int	w, h;
int	t;

/* Find the proper window location and size */
    AlignMenu (mnu, &x, &t, &w);
    y = Yloc(mnu->yloc);

/* Find proper window size */
    h = (int)(mnu->height*Vunit + 1);

/* Create the menu window */

//    (void)CBX_CreatePopupMenu (mnu->mwin, mnu->mu,  x, y,  w, h );

// // Revise this type testing into a switch/case structure
//     if (mnu->type == M_ADROP) {
// // printf (" .- Making A-drop menu for %s.\n", mnu->label);	// debug
// 	CBX_CreateAutoDrop (mnu->mwin, mnu->mu, x, y, w, h,
// 		mnu->label, mnu->callback );
//     }

    switch (mnu->type) {
    case M_DROP:
	CBX_CreateDropMenu (mnu->mwin, mnu->mu, x, y, w, h,
		mnu->label );
	break;
    case M_ADROP:
	CBX_CreateAutoDrop (mnu->mwin, mnu->mu, x, y, w, h,
		mnu->label, mnu->callback );
	break;
    case M_POP:
	CBX_CreatePopupMenu (mnu->mwin, mnu->mu,  x, y,  w, h );
	break;
    case M_APOP:
	CBX_CreateAutoPopup (mnu->mwin, mnu->mu, x, y, w, h,
		mnu->callback );
	break;
    default:
	printf (" No menu created for %s -- bad type.\n", mnu->label);
    }

    if (L != NULL) *L = mladd (*L, mnu);

}

/*  ------------------------------------------------------------  */
void	MakeDButton (bble** blist, butblock* but, 
			int* dptr, u_long color)	//stbut
{
int	x, y;
int	w, h;
int	t, ts;

/* The label follows the button location, simplyfing things */
//    x = Xloc(but->xloc);
//    t = strlen (but->label);
//    w = Xsize (Xp*(but->width +  t) + 3 );

/* Find the horizontal sizes of things, and location */
    t = strlen (but->label);			// Text length of label
    ts = (t > 0) ? Xsize (Xp*t +2) : 0;		// Label size
    x = Xloc (but->xloc);			// Aligned location
    w = Xsize (Xp*(but->width + t) + 3);	// Window width
// NOTE - THE ABOVE IS PROBABLY WRONG, AS REGARDS +t
    switch (but->align) {
    case  Title:
    default:
	but->laboff = 0;
	x += ts;
	break;
    case  Leftside:
	but->laboff = -ts;
	break;
    case  Rightside:
	but->laboff = -(ts+w);
	x -= w;
	break;
    }

/* Fill in data pointer provided */
    but->data = dptr;

    y = Yloc(but->yloc);
    h = (int)(but->height*Vunit + 1);

// Color is displayed when selected, and is white when unselected.
//    if (dptr != NULL) but->bu->status = *dptr;
    (void)CBX_CreateAutoDButton (but->mwin, but->bu,  x, y,  w, h,
	but->label, color, but->callback);
//	but->label, app->BUT_SEL, ToggleButton);
//    butlist = bladd (butlist, but);
    if (blist != NULL) *blist = bladd (*blist, but);

    if (dptr != NULL) but->bu->status = *dptr;
// Needs to be set after create auto so it comes up looking right!
}

// global/appspecific things:
//	BUT_SEL, the button color (could be param? in but block?)
//  requires ToggleButton, which may be app specific itself
//  Pass the app-specific entry point as a parameter to this routine!
//  could make toggle button a block entry, too.
//	butlist -- make it a parameter - pointer to butlist...



/*  ------------------------------------------------------------  */
void	MakeButton (bble** blist, butblock* but)	//stbut
{
int	x, y;
int	w, h;
int	t, ts;


/* Find the horizontal sizes of things, and location */
    t = strlen (but->label);			// Text length of label
    ts = (t > 0) ? Xsize (Xp*t +2) : 0;		// Label size
    x = Xloc (but->xloc);			// Aligned location
    w = Xsize (Xp*(but->width + t) + 1);	// Window width
// NOTE - THE ABOVE IS PROBABLY WRONG, AS REGARDS +t
    switch (but->align) {
    case  Title:
    default:
	but->laboff = 0;
	x += ts;
	break;
    case  Leftside:
	but->laboff = -ts;
	break;
    case  Rightside:
	but->laboff = -(ts+w);
	x -= w;
	break;
    }


    y = Yloc(but->yloc);
    h = (int)(but->height*Vunit + 1);

// Color is displayed when selected, and is white when unselected.
    (void)CBX_CreateAutoButton (but->mwin, but->bu,  x, y,  w, h,
	but->label, but->callback);
//    butlist = bladd (butlist, but);
    if (blist != NULL) *blist = bladd (*blist, but);
}

// Pass butlist* as a parameter


/*  ------------------------------------------------------------  */
void	MakeMButton (bble** blist, butblock* but)	//stbut
/* Like MakeButton, but makes a non-automatic button */
{
int	x, y;
int	w, h;
int	t, ts;


/* Find the horizontal sizes of things, and location */
    t = strlen (but->label);			// Text length of label
    ts = (t > 0) ? Xsize (Xp*t +2) : 0;		// Label size
    x = Xloc (but->xloc);			// Aligned location
    w = Xsize (Xp*(but->width + t) + 1);	// Window width
// NOTE - THE ABOVE IS PROBABLY WRONG, AS REGARDS +t
    switch (but->align) {
    case  Title:
    default:
	but->laboff = 0;
	x += ts;
	break;
    case  Leftside:
	but->laboff = -ts;
	break;
    case  Rightside:
	but->laboff = -(ts+w);
	x -= w;
	break;
    }


    y = Yloc(but->yloc);
    h = (int)(but->height*Vunit + 1);

// Color is displayed when selected, and is white when unselected.
    (void)CBX_CreateButton (but->mwin, but->bu,  x, y,  w, h,
	but->label );
// Same as auto button but without the callback.
// The auto button callback distinguishes between left and other
// mouse buttons.  Here, the buttons must be handled by the
// main loop.  A general handling routine may be used.

//    (void)CBX_CreateAutoButton (but->mwin, but->bu,  x, y,  w, h,
//	but->label, but->callback);
    if (blist != NULL) *blist = bladd (*blist, but);
}


/*  ------------------------------------------------------------  */
void	MoveButton (butblock* but)	//stbut
/* Button is moved to new location which has been entered in the
  button block.  Used for window resize things.  */
{
int	x, y;
int	w;
//int	h;
int	t, ts;

// Copying here the same code as is used to make the button

/* Find the horizontal sizes of things, and location */
    t = strlen (but->label);			// Text length of label
    ts = (t > 0) ? Xsize (Xp*t +2) : 0;		// Label size
    x = Xloc (but->xloc);			// Aligned location
    w = Xsize (Xp*(but->width + t) + 1);	// Window width
// NOTE - THE ABOVE IS PROBABLY WRONG, AS REGARDS +t
    switch (but->align) {
    case  Title:
    default:
	but->laboff = 0;
	x += ts;
	break;
    case  Leftside:
	but->laboff = -ts;
	break;
    case  Rightside:
	but->laboff = -(ts+w);
	x -= w;
	break;
    }

    y = Yloc(but->yloc);
//    h = (int)(but->height*Vunit + 1);

//    if (but->bu->win != NULL)
// Above removed since but->bu->win is not a pointer, but the actual window!
// Change to check for correctly assigned mwin, could also check display
    if (but->bu->mwin == but->mwin && but->bu->disp == but->mwin->disp)
	XMoveWindow (but->mwin->disp, but->bu->win, x, y);

//    (void)CBX_CreateAutoButton (but->mwin, but->bu,  x, y,  w, h,
//	but->label, but->callback);
}


/* === Location of MakeFiltmenu */

/*  ------------------------------------------------------------  */
int	FillMenu (Menu* m, IMenuEntry* e, int sel)
{
/* "e" is an array of entries, terminated with one having a NULL
 character value in its text field.  This subroutine fills those
 entries into the entry array of the given menu "m", and sets its
 selection index to "sel".  Also adjusts the "m->n" value to the
 proper number of entries.  Finishes by calling the display routine. 
 Returns the selection value. */
int	n;
MenuEntry  *me;    // Defined in ccbxlib, subset of IMenuEntry stuff

/* Check for any null case exits needed... */
    if (m == NULL) return 0;
    if (e == NULL) return 0;

/* Loop to spill stuff into m */
    for (n=0,me=m->entry; *(e->text) != AN; me++,e++,n++) {
	strcpy (me->text, e->text);
	me->flag = e->flag;
    }

/* Note that a better alternate is to start by setting the
  menu count (m->n) to zero, and calling addmenuentry for
  each entry in the loop.  Run the loop on "e" only. */

    m->n = n;
    m->sel = sel;

    if (sel < n) n = sel;	// else n is upper limit of sel.
    if (n < 0) n = 0;
    return  n;
}

// The IMenuEntry is somewhat app specific.  Generalize it, add
// to the .h file and it should fly.


/*  ------------------------------------------------------------  */
// NOTE the slight similarity of DrawLabel and draw_label; the
// latter uses Xloc, Yloc on its position arguments...
// This DrawLabel routine is used in window, menu, button labeling.
void	DrawLabel (MainWindow* mwin, int x, int y, char* string)
{
// Put up the string at the given location
int	sl;
Display  *display=mwin->disp;
Window  win=mwin->win;
GC      gc= mwin->gfs.gc;

// Return any null cases
    if (string == NULL) return;

// Find length, and display the text
    sl = strlen (string);
    if (sl < 1) return;
// The x,y locations are pre-computed with Xloc, Yloc here...
// NOTE -- we may want to add TextH to y at this point?
    XDrawString (display,win,gc, x,y, string, sl);
}

/*  ------------------------------------------------------------  */
void	DrawWindow (winblock* wb)
{
int	x, y;

/* Display the window's text field in main window */
    x = Xloc (wb->xloc) + wb->laboff;
    y = Yloc(wb->yloc) + (int)(Vunit*wb->height-1);  // +1);
    DrawLabel (wb->mwin, x, y, wb->label);

/* Update the edit window here */
    CBX_UpdateEditWindow(wb->ew);
}

/*  ------------------------------------------------------------  */
// void	DrawMenu (mnublock* mb, u_long colour)
void	DrawMenu (mnublock* mb)
{
GC	lgc = mb->mwin->gfs.gc;
XGCValues	savgc;
int	x, y;

/* Display the Menu's text field in main window */
// do not do this for drop menus, the text is in the window...
    switch (mb->type) {
    case M_POP:
    case M_APOP:
//    case M_INSTR:
//    case M_FILTR:
//    case M_GRAT:
//    case M_GRIS:
	x = Xloc(mb->xloc) + mb->laboff;
//	x = Xloc(mb->xloc + mb->laboff);
	y = Yloc(mb->yloc) + (int)(Vunit*mb->height-2);  // +1);
// Plot in colour if desired...
//	if (colour != 0)
	if (mb->tcolor != 0) {	// should save the color for restore
	    XGetGCValues (mb->mwin->disp, lgc, GCForeground+GCBackground,
		&savgc);
/*  Example of how to save/restore foreground color in a GC  */
	    XSetForeground (mb->mwin->disp, lgc, mb->tcolor);
	}
// Plot in the defined color, not in any persistant color.
	DrawLabel (mb->mwin, x, y, mb->label);
	if (mb->tcolor != 0)	// need to restore the default color.
		XSetForeground (mb->mwin->disp, lgc, savgc.foreground);
// NOTE THAT WE ASSUME DEFAULT COLOR IS 0 HERE
// Since the color may persist, define that color for all calls to
// this program which are in a single loop
    }

/* Update the menu here */
// do not update automatic menus...
    switch (mb->type) {
    case M_DROP:
    case M_POP:
	CBX_UpdateMenu(mb->mu);
    }
}


/*  ------------------------------------------------------------  */
void	SetBlockData (winblock* wb, double val)
// Put val into winblock's data pointer, and write the
// data into the winblock's text field.  Then display it.
{

/* Decide what if anything to do based on block type */
    switch (wb->type) {
	case  TEXT:
	case  DTEXT:
	    return;
//	    break;
	case  ITEXT:
	    *(int*)wb->data = val;
	    sprintf (wb->ew->text, "%d", *(int*)wb->data);
	    break;
	case  FTEXT:
	    *(double*)wb->data = val;
//	    if (wb == &slitpaf) val += 90.0 * wb->scale;
	    val += wb->offset * wb->scale;
	    sprintf (wb->ew->text, "%.*f", wb->dp, val/wb->scale);
	    break;
	case  DMS:
	    *(double*)wb->data = val;
//	    strcpy (wb->ew->text,  sexig( val/wb->scale, 3, wb->dp) );
	    strcpy (wb->ew->text, " ");
	    strcat (wb->ew->text,  sexig( val/wb->scale, 3, wb->dp) );
	    break;
    }
    CBX_UpdateEditWindow(wb->ew);
}

/*  ------------------------------------------------------------  */
void	draw_label (MainWindow* mwin, int x, int y, char* string)
{
// Put up the string at the given location
int	sl;
Display  *display=mwin->disp;
Window  win=mwin->win;
GC      gc= mwin->gfs.gc;
int	sx;
int	sy;

// Return any null cases
    if (string == NULL) return;

// Find length, and display the text
    sl = strlen (string);
    if (sl < 1) return;
// NOTE -- we may want to add TextH to y at this point?
    sx = Xloc (x);
    sy = Yloc (y);
    XDrawString (display,win,gc, sx,sy, string, sl);
}


/*  ------------------------------------------------------------  */
void redraw_windows(wble* list)
{
wble	*p;
winblock  *wb;

/* Draw all windows in the window list  */
//    for (p=list; p != NULL; p = p->next) {
    C_loop (p, list) {
	wb = p->blk;
	DrawWindow (wb);
//	if (p->next == list) break;
    }
}

/*  ------------------------------------------------------------  */
/* In guisubs.c add to the redraw_menus the following:
  1) the mble* list argument
  2) a local mble *p variable as below
  3) The c-loop statement to do all the drawmenu things
..Add that to the few static titles which are there.     ...*/
void redraw_menus_x(mble* list)
{
mble	*p;

/* Draw all menus in the menu list */
    C_loop (p, list)  DrawMenu (p->blk);	// was.. , 0);
}

// Want to add same thing for buttons and menus

/*  ------------------------------------------------------------  */
