/*  guiutil.h == Extended GUI generalized support definitions  */

/*  Much of this is taken out of intgui.h, which then includes
  this file in place of what is removed.  */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                  Copyright (C) 2010 by                            **
**     Ken Clardy,  Carnegie Observatories,  Pasadena California.    **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* ---------------------------------------------------------------------- */

#ifndef	INCLUDE_GUIUTIL_H
#define	INCLUDE_GUIUTIL_H

/* INCLUDEs ------------------------------------------------------------- */

#include	"cxt.h"
#include	"maskdat.h"

/* DEFINEs -------------------------------------------------------------- */

/* Units for character fine positioning  */

#define  Xp  4
#define  Yp  6
// Units are 1/5 character width, 1/6 character height.

/* We have a font height of PXh.  The aspect ratio is about 4/7.
  A character width and height value is derived from PXh, as follows:
  Vertically, the unit of measure is PXh/Yp pixels; the text
  height is (Yp+1) units plus 1 pixel.  
  The horizontal character spacing is 4*(Yp+1)/7 times the unit
  of vertical measure in pixels; This is Xp horizontal units.
  We may assume that the actual character is (Xp-1) units wide,
  with a 1 unit space between characters.

  Our locations are specified in horizontal and vertical units
  as computed above.  Pixel locations are computed using the accurate
  floating point unit definitions and the integer units, then rounding
  to the nearest pixel.

  Windows to hold text are 1 unit larger than text height, i.e.
  (Yp+2) units plus 1 pixel.

  Inter-line spacing is (Yp+3) units.

  Actual locations -- pixels = Xloc (x) or Yloc (y)

  Horizontal spacing is given by Xsize (units), where Xp units
  is one character spacing...

.. */


/* TYPEDEFs ------------------------------------------------------------- */

enum	aligntype {
	Title,
	Leftside,
	Rightside
	};

enum	datatype {
	TEXT,		// Text value
	DTEXT,		// Dynamic text pointer
	ITEXT,		// Integer field (scaled?)
	FTEXT,		// Floating point field (scaled)
	DMS,		// dd:mm:ss.ss  Scale Degree or Hour
	EPOCH		// Special epoch
	};

/* Window block - generalized entry parameters */
typedef  struct  winblock {
	int	xloc;
	int	yloc;
	int	width;
	int	height;
	char	label[32];
	void	*data;
	int	type;
	double	scale;
	double	offset;
	int	dp;
	EditWindow  *ew;
	MainWindow  *mwin;
	int	align;
	int	laboff;
	void	(*callback)(void*);	// A pointer to a callback entry
	int	(*updater)(struct winblock*);	// Pointer to update program
	} winblock;

/* Window block list element */
typedef  struct  wble {
	struct wble *next;
	struct wble *last;
	winblock  *blk;
	} wble;

enum	mnutype {
	M_VOID,
	M_DROP,
	M_ADROP,
	M_POP,
	M_APOP
//	M_INSTR,
//	M_FILTR,
//	M_GRAT,
//	M_GRIS
	};

// Flags for menu block flag
//#define  MBred  1
//#define  MByel  2
//#define  MBgrn  4

/* Menu block - generalized entry parameters */
typedef  struct  mnublock {
	int	xloc;
	int	yloc;
	int	width;
	int	height;
	char	label[32];
	void	*data;
	int	type;
	Menu	*mu;
	MainWindow  *mwin;
	int	align;
	int	laboff;
	void	(*callback)(void*);	// A pointer to a callback entry 
//	int	flag;
	u_long	tcolor;		// Title color
	} mnublock;

/* Menu block list element */
typedef  struct  mble {
	struct mble *next;
	struct mble *last;
	mnublock  *blk;
	} mble;

/* Button block - generalized entry parameters */
typedef  struct  butblock {
	int	xloc;
	int	yloc;
	int	width;
	int	height;
	char	label[32];
	int	*data;
	Button	*bu;
	MainWindow  *mwin;
	int	align;
	int	laboff;
	void	(*callback)(void*);	// A pointer to a callback entry
	} butblock;

/* Button block list element */
typedef  struct  bble {
	struct bble *next;
	struct bble *last;
	butblock  *blk;
	} bble;

// imenuentry is application dependent, used in function below...
/* Extended IMACS name and menu entry structure... */
typedef struct imenuentry_tag {
    char name[32];	// Name in element queue
    char text[32];	// Text presented in menu
//    int  value;		// Redundant with index.
    int  order;		// Disperser order; 0 == direct
    int  type;		// 0=unknown, 1=grating, 2=grism...  See enum
    int  flag;		// Special flag, e.g. disabled
    } IMenuEntry;
// Make this general for filling a changing menu...


/* GLOBALs -------------------------------------------------------------- */

/* Acknowledging that it is bad form, nonetheless these global values
  are assumed by the generalizations in this library and get located
  here.  The variables shoule be called "extern" if this is included
  from a file which is not guiutil.c  */

#ifdef IS_GUIUTIL_C
#define EXT_GUIUTIL 
#else
#define EXT_GUIUTIL extern
#endif


/*  ====  Location Globals  ====  */

/*  -- temporarily, make these constant...
#ifdef FONTSIZE
EXT_GUIUTIL	int  PXh=FONTSIZE;
#else
EXT_GUIUTIL	int  PXh=18;
#endif
// PXh is the font size, height
.. end temporary removal.. */

// #define  PXh  18

//#define CharW	(10+(PXh-14)/2)

#define DEF_FONT        "lucidatypewriter"

// #define  Tpxh  18	// Temporary definition
// double	Vunit={Tpxh/Yp};		// Units in pixels (3)
// //double	Hunit={4*(Yp+1)*Tpxh/(7*Yp*Xp)};
// // Should work out to Hunit*Xp about 8 rather than current 12..
// double	Hunit={(4*(Yp+1)*(Tpxh/Yp)/(7*Xp))};  // (4*7/7)*(3/4) == 3
// // That looks best, but still slightly wide...


#ifdef  IS_GUIUTIL_C
double		Vunit=3.0;
double		Hunit=3.0;
#else
extern	double	Vunit;
extern	double	Hunit;
#endif

/* FUNCTION PROTOTYPEs -------------------------------------------------- */

/*  -----  Dynamic Positioning support  -----  */
int	Xsize (int c);
int	Xloc (int x);
int	Yloc (int y);

/*  --- Debug help  ---  */
// void	dpause (char* mes);

/*  -----  List support code  -----  */
wble*		wlpush		(wble*, wble*);
wble*		wlpop		(wble*);
wble*		wladd		(wble*, winblock*);
bble*		blpush		(bble* a, bble* b);
bble*		blpop		(bble* a);
bble*		bladd		(bble* a, butblock* b);
mble*		mladd		(mble* a, mnublock* b);

double		pdecode		(char*);

// int		Mselectv	(IMenuEntry*  mu, int value);

winblock*	BlockOfWin	(wble* list, EditWindow* ew);
butblock*	BlockOfBut	(bble* list, Button* bx);
int	TextScan	(wble*);
int	Call_Updater (winblock* wb);
int	HandleTextWindow (wble* list, EditWindow* w);
/* Note - handletextwindow is applicatin specific */
//void	SetEditWindow (wble* list, EditWindow* w, char* data);
void	SetFloatWindow (wble* list, winblock* wb, double v);
void	SetIntWindow (wble* list, winblock* wb, int v);
/* Also, TextBack is application specific */
// static void	TextBack	(void* param);
// wble*	AutoEditWin	(wble* list, winblock* wb, void* data);
void		AutoEditWin	(wble** list, winblock* wb, void* data);

// wble*		MakeEditWin	(wble*, winblock*, void*);
void	AlignMenu (mnublock* mnu, int* x, int* t, int* w);
void	MakePMenu	(mble**, mnublock*);
//void	MakeAMenu	(mble**, mnublock*);
void	MakeMenu	(mble**, mnublock*);
void	MakeDButton	(bble**, butblock*, int*, u_long color);

void	MakeButton	(bble** blist, butblock* but);
void	MakeMButton	(bble** blist, butblock* but);
void	MoveButton	(butblock* but);
int	FillMenu	(Menu*, IMenuEntry*, int);
void	DrawLabel	(MainWindow*, int, int, char*);
void	DrawWindow	(winblock*);
void	DrawMenu	(mnublock*);	// was.. , u_long);
void		SetBlockData	(winblock*, double);


void	draw_label	(MainWindow*, int, int, char*);

// Note, the grid line stuff may be application dependent,
// but is a valued debug ability.

void	redraw_windows	(wble*);

/* ---------------------------------------------------------------------- */
/* End of kdcutil.h package */

#endif	/* INCLUDE_GUIUTIL_H */

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
