/*  intgui.h -- Definitions for interface GUI application */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2010 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Interface GUI location definitions... */

#include	"guiutil.h"
// Needed for defines and typedefs

/* Units for character fine positioning  */

#define  Xp  4
#define  Yp  6
// Units are 1/5 character width, 1/6 character height.

/* We have a font height of PXh.  The aspect ratio is about 4/7.
  A character width and height value is derived from PXh, as follows:
  Vertically, the unit of measure is PXh/Yp pixels; the text
  height is (Yp+1) units plus 1 pixel.  
  The horizontal character spacing is 4*(Yp+1)/7 times the unit
  of vertical measure in pixels; This is Xp horizontal units.
  We may assume that the actual character is (Xp-1) units wide,
  with a 1 unit space between characters.

  Our locations are specified in horizontal and vertical units
  as computed above.  Pixel locations are computed using the accurate
  floating point unit definitions and the integer units, then rounding
  to the nearest pixel.

  Windows to hold text are 1 unit larger than text height, i.e.
  (Yp+2) units plus 1 pixel.

  Inter-line spacing is (Yp+3) units.

  Actual locations -- pixels = Xloc (x) or Yloc (y)

  Horizontal spacing is given by Xsize (units), where Xp units
  is one character spacing...

.. */

/* Vertical Locations */

// These locations are now in vertical units, not pixels...
// approximate conversion is 3 pixels per unit...

#define  TH  Yp+1
// Use TextH = TH * Vunit + 1 to call things
#define  WH  Yp+2
// Use WinH = WH * Vunit + 1 to call window creation
#define  LH  Yp+4
// TH is height of text;  WH is height of a window; 
// LH is the normal inter-line spacing.
#define  BH  Yp+5
// BH is a wider spacing between sections, candidate for line


#define  Line1	1
	// line 1 is filemenu, optmenu, runbutton
#define  Line2	(Line1+LH)
#define  Line3	(Line2+LH)
#define  Line4	(Line3+LH)

#define  Line5	(Line4+BH)	// R.A.
#define  Line6	(Line5+LH)

#define  Line7  (Line6+WH+TH)
// Line7 is start of graphic line, Line6 + 15

#define  Line8  (Line7+TH+BH)
// Line8 is Temperature or Obs. Date.; Line6 + 33 or Line7 + 18

#define  Line9  (Line8+LH)	// Inst. details
#define  Line10 (Line9+LH)	// why not LH ? (no win on previous line)
// Line10 is instrument, disperser;
#define  Line11 (Line10+LH)	// Filter wavelengths
#define  Line11a (Line11+LH)	// Special wavelengths

#define  Line12 (Line11a+LH)	// Slit specs (was +WH, then +BH)
#define  Line13 (Line12+LH)	// Width, length (was +WH)
#define  Line14 (Line13+LH)

#define  Line15 (Line14+BH)	// Align hole size

#define  Line16 (Line15+BH)
#define  Line17 (Line16+LH)
#define  Line18 (Line17+LH)

#define  Line19 (Line18+BH)

//#define  Vsize  (Line12+40)
// #define  Vsize  (Line18+WH+LH+TH)	// why so many??
#define  Vsize  (Line19+BH+TH)

/* Horizontal Locations */

/* For labels use Margin1 and CharW */
//#define  Margin1  4


#define  COL1  1
#define  COL1a COL1+5*Xp	// Warning graphic start
#define  COL2  COL1+13*Xp+2
#define  COL3  COL2+12*Xp
#define  COL4  COL3+6*Xp
#define  COL5  COL4+8*Xp
#define  COL6  COL5+10*Xp
#define  COL7  COL6+10*Xp
// then Hsize is now COL7 or COL5 + 20*Xp
#define  Hsize  (COL7 + Xp)

#define  COL2a  COL2+8*Xp

// COL2a is yellow stripe
// COL4  is blue stripe
// COL7  is blue stripe

/* Picture of actual screen... 

     g      y       g    y  g    b        g        g              b
 -| Col1           Col2    Col3 Col4     Col5     Col6           Col7
 -|_Col1________________________Field2_______________
 -| |              |       |    |        |        |  
 1| file(iq)menu   opmenu   maskgenrunbut
 2| Data File: [Filename]       Observer: [NameGoesHere]
 3| Title: [Generaltext... ]
 4| Object: [Name]          Sub-Field: [n]
  |
 5| R.A.: [ ==:==:==.----]      Equinox: {====.--}
 6| Dec.: [-==:==:==.--  ] Slit Pos.Ang. [===.--]  () Parallactic
  |
Wr|  -xx.xx |-------------------|-------------------| xx.xx
  |
 7| Date: [yyyy-mm-dd]      Time: [hh:mm:ss] {Ut/St/Ha}
b8| Temperature [-==.-] C		{No ADC [] }
  |
 9| Instrument Details:
10| Instrument: {Long/Short}    Disperser: {Gratings}
b1| Filter: {name} |wavelengths|  Center Wavelength: [====.-]
11a|  Display wav  |wavelengths|       Ext.Ord: { - - - }
  |
12| Slit Specifications:              
13| Width:  [==.=]       Lengths: [==.=] - [==.=]  Overlap[=] () Slit/hole
b4|                        Uncut: [==.=] - [==.=]  Extend Slits [ ]
  |
15| Alignment Hole size: [==.-]          Repeat Obj [==] Ref. [==]
  |
16|  Must Have: [===]   Pdecide: [===]           Ref. Select  [==]
17|  Gaps:  {on/off}
b8|
19|  {Object Files}                                         {Test}

16| Special Modes:                               Ref. Select  [==]
17| () Nod/Shuffle  () Slit auto-extend  () Temp. compensate
18| Nod/Shf. C-length: [-==.=]
  |
[==-] = Value window; {...} = Selection menu
... */



#define  BUT_UNSEL	grey
#define  BUT_SEL	red
// Those 2 used to preceed the button block definition

#define  iqmenux  COL1
#define  iqmenuy  Line1
#define  iqmenuw  (7*Xp+2)	// was.. (7*Xp + 2)

// #define  opmenux  COL2
// #define  opmenuy  Line1

// #define  mgbutx  COL3
// #define  mgbuty  Line1

#define  eqttlx  COL4	// was.. Field2
#define  eqfldx  COL5	// was.. (COL4+9*Xp)
#define  eqfldy  Line5
#define  eqfldw  (8*Xp+2)
#define  eqttly  (eqfldy+TH)


/* Things being put in windows:
Text Windows:
  Observer Name;  File Name;  Object Name;  title
Quantity windows:
  R.A.  Dec.  (1)
  Date  time (2)  
  temperature
  Center Wavelength
  Slit size, angle, overlap allowance
  Reference hole size
Choices:
  (2) (ut/sidereal/ha)
  Short/Long camera
  Grating/Grism choice (or direct?)
  Reference hole shape
Choice with alternate text entry:
  (1) Equinox
  Parallactic angle; e-w; PositionAngle
  Filter or Wavelength Limit


Controls:  File read/write;  Maskgen run;  options
  List of object position file specifications
*/

/* --- Definitions needed for rotator geometry --- */

typedef	struct	anglist {
	struct anglist	*next, *last;
	double		v;
	} anglist;

typedef	struct	bandlist {
	struct bandlist	*next, *last;
	double		a;
	double		b;
	} bandlist;

typedef	struct	pdata {
	double	dec;
	double	lat;
	double	pa;
	double	target;
	double	width;
	int	side;
	}  pdata;


/* -- Definitions needed for function prototypes... */

typedef struct	{
	Obs	*obs;
	char	filename[120];
	int	changed;
	element*  gs;
	}  Obf;


typedef  struct  filtelem {
    char	name[32];
    passband  pb;
    } filtelem;


/* ----- Function prototypes ----- */
winblock*  findwb (EditWindow* ew);
// winblock is defined in guituil.h

/*  ---  Rotator angle support  ---  */
double  gsda (double a, double d, double g);
double  sda (double d, double g);
anglist   *qflip (pdata *pd);
//bandlist  *qwarn (pdata *pd);
bandlist  *qwarn (pdata *pd, double a, double b, vect2* lim);

void	solpos (int mj);


/*  -----  Dynamic Positioning support  -----  */
// int	Xsize (int c);
// int	Xloc (int x);
// int	Yloc (int y);

// double		pdecode		(char*);

// int		Mselectv	(IMenuEntry*  mu, int value);
//IMenuEntry*	dispmpv		(int value);
// int		instmpn		(char* name);
// int		disptype	(int value);
// int		dispndx		(char* name, IMenuEntry* me);
int		Mselectn	(char* name, IMenuEntry* me);
int		instdisp	(int value);

// winblock*	BlockOfWin	(wble* list, EditWindow* ew);
// butblock*	BlockOfBut	(Button* bx);
void	ToggleButton	(void* param);
// static void	TextScan	(void);
// static void	HandleTextWindow (EditWindow* w);
//int	UpdateTextWindow (winblock* wb);
int	TextUpdate (winblock* wb);
void	TextBack	(void* param);
// wble*		AutoEditWin	(wble* list, winblock* wb, void* data);

// wble*		MakeEditWin	(wble*, winblock*, void*);
// static void	AlignMenu (mnublock* mnu, int* x, int* t, int* w);
// static void	MakePMenu	(mnublock*);
// static void	MakeDButton	(butblock* but, int* dptr);

Bool		KDC_DialogBox	(char *title, char* text, char *answer);
void	handle_button	(void* param);

// static void	MakeButton (butblock* but);
void	MakeFiltmenu	(mble**, filtelem*, mnublock*, IMenuEntry*);
// static void	FillMenu	(Menu*, IMenuEntry*, int);
// static void	DrawWindow	(winblock*);
// static void	DrawMenu	(mnublock*);
// void		handle_text	(wble*, Window, XEvent*);
// void		SetBlockData	(winblock*, double);
void	handle_iqmenu	(void*);
void	handle_eqmenu	(void);
void	handle_instmenu	(void);
void	handle_dispmenu	(void);
void	handle_filtmenu	(void);
#ifdef  GAPVO
void	handle_gapmenu (void);
void	fill_gapmenu (mnublock* gm, element* inst);
#endif
// static void	draw_label	(int, int, char*);
#ifdef  WDEBUG
void	vgridline (int x, int color);
void	hgridline (int y, int color);
void	redraw_grid(void);
#endif
// void	redraw_menus	(void);
void	redraw_menus	(mble*);
// static void	redraw_windows	(wble*);
void	redraw_text (void);
void	redraw_warnings (void);
// void		edit_win_border	(wble*, Window, u_long);

float   get_pa             (float);
float   get_cp             (float);

// static void	pgui		(Obf*);



