/*  maskcutsub.c == Temporary include file for maskcut.c subroutines  */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2010 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* This is a collection of common subroutines which will be placed in
  the front of the maskcut.c code.  These are collected to remove 
  duplication of code in the maskgen.c program set.   */

/* These are not separately compiled routines, those are in cututil.c */



	/* - - - - - - - - - - - - - - *\
	|				|
	|  Global Cutting Parameters	|
	|				|
	\* - - - - - - - - - - - - - - */

static	void	cutparams (Mindx Minst)
/* Do the setup of global cutting parameters */
{
/* All variables used here are globals, Minst is a calling parameter */
/* Add any other global setup here */

/* Set up global cutting parameters */
#ifdef  USE_CDAT
	flange   = CDhead[Minst].flange;
	aangle   = CDhead[Minst].aangle;
	aclear   = CDhead[Minst].aclear;
	LongSlit = CDhead[Minst].LongSlit;
	SlitGap  = CDhead[Minst].SlitGap;
	avtool   = CDhead[Minst].tool;
#ifdef  SIMPLECURVE
	Curve    = CDhead[Minst].curve.x;
#else
	Curve    = CDhead[Minst].curve;
#endif
#ifdef  USE_CTE
	CTE      = CDhead[Minst].cte;
	Tcut     = CDhead[Minst].tcut;
#endif
	bwid  = CDhead[Minst].beam * SF;
	quick = CDhead[Minst].slew * SF;
	slow  = CDhead[Minst].cut  * SF;
	vslow = CDhead[Minst].fine * SF;
#ifdef  USE_ACON
	Acon  = CDhead[Minst].acon * 1.e-6;
#endif
#endif	// on USE_CDAT

#ifdef  USE_GSTO
	flange = *(double*)GS_lookup (&CDhead, Minst, CD_flange, 0);
	aangle = *(double*)GS_lookup (&CDhead, Minst, CD_aangle, 0);
	aclear = *(double*)GS_lookup (&CDhead, Minst, CD_aclear, 0);
	LongSlit = *(double*)GS_lookup (&CDhead, Minst, CD_LongSlit, 0);
	SlitGap  = *(double*)GS_lookup (&CDhead, Minst, CD_SlitGap, 0);
	avtool   = *(double*)GS_lookup (&CDhead, Minst, CD_tool, 0);
#ifdef  SIMPLECURVE
	Curve    = *(vect2*)GS_lookup (&CDhead, Minst, CD_curve, 0).x;
#else
	Curve    = *(vect2*)GS_lookup (&CDhead, Minst, CD_curve, 0);
#endif
#ifdef  USE_CTE
	CTE      = *(double*)GS_lookup (&CDhead, Minst, CD_cte, 0);
	Tcut     = *(double*)GS_lookup (&CDhead, Minst, CD_tcut, 0);
#endif
	bwid  = *(double*)GS_lookup (&CDhead, Minst, CD_beam, 0) * SF;
	quick = *(double*)GS_lookup (&CDhead, Minst, CD_slew, 0) * SF;
	slow  = *(double*)GS_lookup (&CDhead, Minst, CD_cut,  0) * SF;
	vslow = *(double*)GS_lookup (&CDhead, Minst, CD_fine, 0) * SF;
#ifdef  USE_ACON
	Acon  = *(double*)GS_lookup (&CDhead, Minst, CD_Acon, 0) * 1.e-6;
#endif	
#endif	// on USE_GSTO

#ifdef  USE_BSTO
	flange   = floatvalue(getval(&CDhead,MaskCase(Minst),"Flange"));
	aangle   = floatvalue(getval(&CDhead,MaskCase(Minst),"Aangle"));
	aclear   = floatvalue(getval(&CDhead,MaskCase(Minst),"Aclear"));
	LongSlit = floatvalue(getval(&CDhead,MaskCase(Minst),"LongSlit"));
	SlitGap  = floatvalue(getval(&CDhead,MaskCase(Minst),"SlitGap"));
	avtool   = floatvalue(getval(&CDhead,MaskCase(Minst),"Rtool"));
#ifdef  SIMPLECURVE
	Curve    = floatvalue(getval(&CDhead,MaskCase(Minst),"Curve"));
#else
	Curve    = vect2value(getval(&CDhead,MaskCase(Minst),"Curve"));
#endif
#ifdef  USE_CTE
	CTE      = floatvalue(getval(&CDhead,MaskCase(Minst),"CTE"));
	Tcut     = floatvalue(getval(&CDhead,MaskCase(Minst),"Tcut"));
#endif
	bwid  = floatvalue(getval(&CDhead,MaskCase(Minst),"Beam")) * SF;
	quick = floatvalue(getval(&CDhead,MaskCase(Minst),"Slew")) * SF;
	slow  = floatvalue(getval(&CDhead,MaskCase(Minst),"Cut"))  * SF;
	vslow = floatvalue(getval(&CDhead,MaskCase(Minst),"Fine")) * SF;
#ifdef  USE_ACON
	Acon  = floatvalue(getval(&CDhead,MaskCase(Minst),"Acon")) * 1.e-6;
#endif
#endif	// on USE_BSTO

}


objq*	qstart (objq* a, Mindx Minst, opts opt)
/* Optionally reorder the queue, and find the starting location based
  on the instrument index (or could use a start point?), return the
  starting queue member.  */
{
double	d;
double	ds;	// smallest found
vect2	c;	// location of point
vect2	p;	// target point
objq	*r;
objq	*t;
int	i;
int	n;

// Based on opt, do the traveling salesman algorithm
    n = listcount (a);
    if (n < 1) return NULL;	// no points, forgetit
    i = (2 * n + 1) / 3;
    if (opt.rerd == 1) {
	a = reorder (a, i, opt.dbug);
//	opt.rerd++;
// Omit increment here to reorder each of the multi-masks if we are
// processing a multi-mask queue.
    }

// Find the point nearest to which we should start, based on Minst
    switch (Minst) {
	case  M_imacs:
	    p = make2vect ( 0.0, 312.0 );
	    break;
	case  M_lds2:
	    p = make2vect ( 0.0, 0.0 );
	    break;
	case  M_lds3:
	    p = make2vect ( 80.0, -30.0 );
	    break;
	case  M_gismo:
	    p = make2vect ( 170.0, 100.0 );
	    break;
	case  M_wfcd:
	    p = make2vect ( 0.0, 53.0 );
	    break;
	default:
	    p = make2vect ( 0.0, 0.0 );
	    break;
    }

// Run a scan loop to find the proper ending point
    ds = 9.9e6;	// a large number
    t = a;	// we need it initialized to something
    C_loop (r, a) {	// loop over all objects
	c = r->smpos;		// position of feature
	d = dist2 (c, p);	// distance to target
	if (d < ds) {		// better case here
	    ds = d;		// save distance
	    t = r->next;	// and target
	}
    }

// Return the next point as the starting point.
    return  t;
}


void	maskinit (FILE* f, Mindx Minst, Obs* obs, opts opt)
// initialize things in mask, standard start comments
// set globals
// set slow cutting feature?
// write inch/metric commands/comments
// set scale on CTE
// set work offset
// set the z offset
{

/* Set up the scale and globals */

    SF = inchout ? 1.0/Inch : 1.0;
#ifdef  USE_CTE
    if (opt.xcte) {
double	tex;
//	SF *= 1.0 + CTE * (Tcut - obs->temp);
	tex = CTE * (Tcut - obs->temp);
	if (opt.dbug > 0 && tex != 0.0)
	    printf (" Correcting thermal expansion of %9.3e\n", tex);
	SF *= 1.0 + tex;
    }
#endif
// The above should not be here, put into cutparams or a special?
/*  Setting SF to thermal correction is done only when an obs structure
  which is being processed requires it.  Not done for special masks,
  therefore should only be done with a valid obs.  Do this during the
  setup for an actual mask.
*/

    cutparams (Minst);	// or import that program here

/* Initialize the machine */

    lasermode (Autofocus_off, f);

    Astop = (obs == NULL) ? 0 : 1;
    if (Astop) lasermode (Exact_stop_on, f);

    IDent (f);
    if (inchout == 0) {		// Warning section
	fprintf (f, " ( ! DANGER!! -- Millimeter mode file! )\r\n");
	fprintf (f, " M0 ( DANGER!! Millimeter mode! )\r\n");
	fprintf (f, " ( Millimeters can cause a crash! )\r\n");
    }
    if (inchout) fprintf (f, " G20 ( Inch measurement )\r\n");
    else         fprintf (f, " G21 ( Metric measurement )\r\n");


// End with work offset, done by Minst

    switch (Minst) {
	case  M_lds2:
	    fprintf (f, " G58  ( Work offset for LDSS )\r\n");
	    break;
	default:
	    fprintf (f, " G59  ( Work offset for IMACS )\r\n");
	    break;
    }

}


void	cutstart (FILE* f, Mindx Minst, vect2 start)
/* Start cutting at the indicated start location; find Z height from
  global positioning routine, check status from globals.  */
{
double	h;

// This is from imacs cutting
    VXYout (f, start, 1);
    lasermode (Autofocus_hold, f);

// Find the proper height here...  by start and Minst
    h = maskz (start);

    Zout (f, h);
    fprintf (f, " ( Start location )\r\n");
    IDent (f);
// NOTE - this gas/shutter stuff could be a laser_mode call!!
    fprintf (f, " M10");	// select gas, 
    fprintf (f, " M08 G4 X0.2 ( Gas on )\r\n");	// turn gas on
    fprintf (f, " M14 ( Shutter open )\r\n");

//
// NOTE -- HELP!  IS H AND ZLAST IN MM OR IN?  SHOULD MULT BY SF OR NOT????
//
// Answer -- h is scaled to current measure; includes Z offset.

    lasermode (Autofocus_on, f);
    lastloc = start;

}


	/* - - - - - - - - - - *\
	|			|
	|    The Object Loop	|
	|			|
	\* - - - - - - - - - - */


static	int	object_loop (FILE* f, objq* obq, Mindx Minst, double dy, int rp)
/*  Generalized cutting loop for objects in object queue obq, return count of
  objects successfully output.  Each object cut rc times. */
/*  output is to file f,  objects are found in queue obq,
  Mask type is Minst, and for gismo we include displacement dy.
  Should probably include x and y displacements and multiples for
  full generality if we get more than that.    */
{
objq	*r;	// roving pointer
int	n = 0;
int	j;
double	tool;
vect2	c;	// location of center
vect2	start;	// location of start corner, relative to c
vect2	slth;	// length
vect2	wide;	// width
static	vect2	last = {999.0, 999.0};	// Absolute last loc for slew
vect2	next;		// for updating last...
static	double	zlast;	// For incremental z movement...
double	rs, rc;	// rotation sine, cosine
double	la, lb, w;	// corrected length, width
double	ls;	// length of slit
//objq	*r;
//objq	*aq;	// Start object

avoids	*aqp;	// Avoidance queue
double	rt;	// tool radius (tip radius for avoidances)
double	xm;	// X coordinate multiplier
int	rep = rp;	// Repeat count

// limit repeat counts...
    if (rep < 1) rep = 1;	// least repeat
    if (rep > 2) rep = 2;	// most repeat

// obtain tool from global stuff, same for other setup

#ifdef  USE_CDAT
    aqp = CDhead[Minst].av;
    rt  = CDhead[Minst].tool;
#endif
#ifdef  USE_GSTO
    aqp = *(avoids**)GS_lookup (&CDhead, Minst, CD_av, 0);
    rt  = *(double*)GS_lookup (&CDhead, Minst, CD_tool, 0);
#endif
#ifdef  USE_BSTO
    aqp = getlsq (&CDhead,MaskCase(Minst),"Avoid");
    rt = floatvalue(getval(&CDhead,MaskCase(Minst),"Rtool"));
#endif

// should also get tool above, from beamwidth...
    tool = bwid / 2.0;
    xm = (Minst == M_lds2 || Minst == M_wfcd) ? 0.0 : -1.0; // For back of mask

    C_loop (r, obq) {
/* Going to cut slit at r... */
//	if (Xcircle(r, make2vect(0.0,0.0), flange) >= 0) {

      if (Minst == M_imacs || Minst == M_lds3) {	// not needed, generalize!
	if (r->type != OBJ_AREA
	  &&  Xcircle(r, make2vect(0.0,0.0), flange) >= 0) {
	    printf ("  Object \"%s\" off mask boundry, skipped.\n",
		r->dat.name);
	    continue;
	}
      }
// Change above to generalized on mask test

//	if (Minst == M_gismo) {
/* Here, we check for any feature within the central part of GISMO, to
  avoid cutting the Gismo Long Slit which we are purposely letting be
  in the .SMF file for data reduction use.  It technically doesn't hurt to
  cut it, the middle is tossed anyway, but it looks better not to.  */
//	    if (vect2norm(r->smpos) < 51.0) continue;
// The x size is 102.5/2.0 = 51.25 mm.  Anything less is inside.
//	}

/* Skip any object which is flagged as virtual.  Ultimately, the .SMF file
  will have a method of passing the virtual status transparently, so the
  Gismo Long Slit and others may be in .SMF and be uncut.  */
	if (r->flag & OBJECT_VIRTUAL) {
	    if (dodebug) printf (" -Virtual object %s skipped.\n", r->dat.name);
	    continue;
	}

/* Find the slit position data...  */
	c = mul2vect (r->smpos, SF);	// object position, center of slit
/* c is now in current scaled measure, mm or inches  */
	c.x *= xm;	/* Correct for back of mask cutting */
	c.y += dy;	/* Displacement (used by gismo) */
/* Check for avoidance; reject any object in an avoidance */
	if (r->type == OBJ_AREA) {	// Special treatment here!
/*  Require shape to be rectangle, else we continue, also omit the
  testing for avoidance which is done in cutting subroutine...*/
	    if (r->slit.shape != RECTANGLE) continue;
// Should also continue if Minst is not imacs, the only one for which
// an area type has been formulated and tested
//	    if (Minst != M_imacs) continue;
// Also allow instrument of GISMO, which should work OK
	    if (Minst != M_imacs && Minst != M_gismo) continue;
// We can generalize area cutting to other things such as ldss3 and
// slightly relax that at a later time.
	} else if (avoidanc (r, aqp, rt) ) {
	    printf ("  Object \"%s\" within an avoidance zone, skipped.\n",
		r->dat.name);
	    continue;
	}
// NOTE THAT ONLY CHECKS CENTER OBJECT; NEED TO CHECK ALL CORNERS!!!
// ADD A ROUTINE SIMILAR TO Rmax (mgutils.c) USING in_avoid

/* Start output of plotting commands for the object. */

	fprintf (f, "( Obj: \"%s\" )\r\n", r->dat.name);
/* Insert an in-line comment if it fits and has no parenthesis in it. */
	if (r->dat.comment) {
if (dodebug > 1)  // Debug print of comments if desired.
  printf (" >Obj: %s  Comment: %s\n", r->dat.name, r->dat.comment);  // debug only
	    if (strlen (r->dat.comment) <= 68 &&
		strchr (r->dat.comment, '(') == NULL &&
		strchr (r->dat.comment, ')') == NULL )
		fprintf (f, "( %s )\r\n", r->dat.comment);
	}
/* The entire preceeding section to insert the in-line comment in the .nc
  file is actually optional.  It could be taken out without any adverse
  effect on the .nc file itself.  It merely provides extra information for
  debugging and error checking.  Added with version 2.09 in Sept. 2009.    */
	IDent (f);	// One ident for each object for now...

	rs = sin (r->slit.angle);
	rc = -cos (r->slit.angle);	/* negative for back of mask */
// w is width/2 less tool, not less than zero
// la is length a, less tool, not less than zero
// lb is like la.
// start is -la, -w rotated by rc,rs
// slth is la+lb,0 rotated from x by rc,rs
// wide is 0,2w, rotated from y by rc,rs
//	w = SF * r->slit.width / 2.0 - tool;
	w = SF * r->slit.width;
	if (w < 0.0) w = 0.0;
//	circh = 0;

	next = c;
// Find the slit length unit vector
	slth = make2vect (rc,rs);

	switch (r->slit.shape)  {
	case  RECTANGLE:
	    la = SF * r->slit.alen - tool;
	    lb = SF * r->slit.blen - tool;
/* la is movement left from center, lb is movement right */
	    ls = la + lb;
	    if ( ls <= 0.0) {
		if (la < 0.0) la = 0.0;
		if (lb < 0.0) lb = 0.0;
		ls = la + lb;
	    }
// Find the slit start position, including beam size
	    start = sub2vect (c, mul2vect (slth, la+tool));
// Get right length for the slit length vector including beam
	    slth = mul2vect (slth, ls+bwid);
/* Plot the rectangle using these parameters */
	if (r->type == OBJ_AREA) {
	    if (dodebug > 2) printf (" Process area \"%s\".\n", r->dat.name);	// debug
	    zlast = carea (start, slth, w, zlast, slow, f, aqp, rt);
	} else {
	    zlast = crect (start, slth, w, zlast, slow, rep, f);
	}
	    break;
	case  CIRCLE:
	    ls = w - bwid;
	    zlast = chole (c, ls, zlast, vslow, rep, f);
	    break;
	default:
	case  SQUARE:
	    la = lb = w/2.0 - tool;	// or la=lb= (w - bwid)/2.0;
	    ls = la + lb;
// Find the slit start position, including beam size
	    start = sub2vect (c, mul2vect (slth, la+tool));
// Get right length for the slit length vector including beam
	    slth = mul2vect (slth, ls+bwid);
/* Plot the rectangle using these parameters */
	    zlast = crect (start, slth, w, zlast, slow, rep, f);
	    break;
// ADD HERE OTHER SPECIAL SHAPES...
// E.G. THE CIRCLE AND 4 SLITS CASE...
	case  SPACED_CROSS:
	    zlast = chole (c, w-bwid, zlast, vslow, rep, f);
// wide is c to slit displacement
// slth is slit length
	    wide = mul2vect (slth, w*0.5 + SF*r->slit.blen);
	    slth = mul2vect (slth, SF*r->slit.alen);
	    for (j=0; j<4; j++) {
		start = sum2vect (c, wide);
		zlast = crect (start, slth, w, zlast, slow, rep, f);
// Rotate wide and slth left by 90 degrees...
		wide = make2vect (-wide.y, wide.x);
		slth = make2vect (-slth.y, slth.x);
	    }
	}	// end switch on slit shapes
	last = next;
	n++;	// good object
    }	// end loop on objects.
    return  n;
}	// end of object_loop program


	/* - - - - - - - - - - *\
	|			|
	|       End  Mask	|
	|			|
	\* - - - - - - - - - - */

static	void	end_mask (FILE* f, double h)
/*  Finishing commands to output f;  h is pull-back height required,
    which may vary with mask geometry.  */
{
    lasermode (Autofocus_hold, f);	// may be redundant
/* End of program */
    fprintf (f, " M09  ( Gas off )\r\n");	// turn gas off
    IDent (f);		// Debug line number
    fprintf (f, " M15 ( Shutter closed )\r\n");
    fprintf (f, "( End cutting. )\r\n");
    Feedrate (f, " G90", quick*3.0);
    Zout  (f, h);  Feol (f);		// Pull back to height h
    lasermode (Autofocus_off, f);	// may be redundant
    IDent (f);		// Debug line number
    XYout (f, 0.0*Inch*SF, 14.0*Inch*SF, 0);
    Zout (f, 3.25*Inch*SF);  Feol (f);
/* NOTE - update location for desired parking spot. */

    IDent (f);		// Debug line number
    fprintf (f, " M30  ( End of mask. )\r\n");	// End program
}



	/* - - - - - - - - - - - - - - *\
	|				|
	|   Current Date formatting	|
	|				|
	\* - - - - - - - - - - - - - - */


char*	curdate (int  fc)
/*  Find and format the current time.  Use fc as format code:
    0 = yyXdd -- year, month as exchar, day (default)
    1 = yyyy-Mon-dd -- year, month name, day
.. */
{
static	char	buf[80];
struct	tm	*ut;
time_t	now;
char    monam[12][4] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

    time (&now);
    now -= 7 * 3600;	// Offset by 7 hours for local time;
    ut = gmtime (&now);
    switch (fc) {
	case  1:
	    sprintf (buf, "%d-%s-%02d", 
		ut->tm_year + 1900, 
		monam[ut->tm_mon], 
		ut->tm_mday);
	    break;
	case  0:
	default:
	    sprintf (buf, "%02d%c%02d", 
		ut->tm_year % 100, 
		exchar(ut->tm_mon+1), 
		ut->tm_mday);
    }
    return  buf;
}


int	cutfield (Obs* a, FILE* f, opts opt)
/*  Basic field cutting utility.  Will process the objects in the obs
  structure, writing the proper commands to file f for the mask or
  sub-mask.  Do engraving based on mask type for labels.  Returns the
  count of features cut into mask.  */
/* THIS IS NOT USED AT THIS TIME */
{
double	yv = 0.0;	// y offset to be globalized
objq	*aq;
//Mindx	inst;
// inst is a global here...
int	n=0;

// Find the instrument index
    inst = MaskIndx (a);

// find start of queue, reorder; return with message if null,
// as that indicates no objects in queue.  (qstart)
    aq = qstart (a->ob, inst, opt);
    if (aq == NULL) {
	printf (" ** No mask objects, nothing written.\n");
	return  0;
    }

// Put that annoying CTE setup here; we have an obs

#ifdef  USE_CTE
    if (opt.xcte) {
double	tex;
//	SF *= 1.0 + CTE * (Tcut - a->temp);
	tex = CTE * (Tcut - a->temp);
	if (opt.dbug > 0 && tex != 0.0)
	    printf (" Correcting thermal expansion of %9.3e\n", tex);
	SF *= 1.0 + tex;
    }
#endif
// The above should not be here, put into cutparams or a special?
/*  Setting SF to thermal correction is done only when an obs structure
  which is being processed requires it.  Not done for special masks,
  therefore should only be done with a valid obs.  Do this during the
  setup for an actual mask.
*/

// do the per-field initialize, or put that into objectloop

// call to process object queue, including any start up needed
    n = object_loop (f, aq, inst, yv, opt.repc);

// switch statement to process engravings
    switch (inst) {
	case  M_imacs:
	case  M_lds2:
	case  M_lds3:
	case  M_gismo:
	    break;
	default:
	    break;
    }
// Can do it here or call subroutines by instrument...

// Add individual per-mask features (ldss3 alignment holes)

// return the feature count
    return  n;
}


int	cutmulqueue (sortq** mmhead, Mindx Minst, opts opt)
/*  Process all the queued obs files in the given queue, check that
  Minst matches for each one, and return the number of masks done.
  Can process special files if queue null, based on Minst and opt
  values.  */
/* Replaces cutmul_ldss and cutmul_gismo routines with a unified
  multiple mask cutting procedure, which could be expanded.  */
{
int	n=0;

// Check the queue for null (special) cases
// Set globals, etc. based on instrument

// Loop over items being in the queue

// check that the obs structure agrees with Minst, error if not

// construct a file name, open the file

// initialize the output file and set globals

// loop over the sub-fields up to the count allowed

// set subfield offset constants

// call cutfield to process the objects

// do any per-subfield things such as ldss3 rim cutting

// Drop the storage for the obs structure, no longer needed

// end loop on sub-fields
// do any per-mask features (gismo hole and label)

// end of mask, fill in cuttime and close file

// reset subfield offsets to zero

// report mask and feature count, and files for this Minst

    return  n;
}

