/*  maskgen.c == Mask Generation Basic Application */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2011 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Look for other example stuff in ~/source/fix.c  */

/*  Usage:
	Argument is Observation File (optional)

        Switches:

                -d      Debug output
		+d	More debug information
		-i	Minimal debug information
		-q	Quick look option
		-r	Reorder object list
		-s	Skip plotting
		-c	No halt if object check fails
		-a	Remove off-detector alignments
		-w	Omit wingchip holes
		-h	Display help


	Features:

		Includes graphic window.  Made optional later.
		Can include object entry and editing later.
.. */

#undef  OPTLISTDEBUG
//#define  OPTLISTDEBUG
/* Symbol OPTLISTDEBUG debugs the free of the optic list, which had
  some problems in the past.  Normally this is not done.  */

#define  OPTDEBUGTEST
//#undef  OPTDEBUGTEST
/* Symbol OPTDEBUGTEST is routinely used to test optic list free by
  specifying both -h and +d at the same time.  If not defined, we
  always use the simplified help stuff.  */

#undef  DEBUG

/* Customization definitions */
//#define  APPLICATION_NAME  "MaskGen"
//const	char	PROGNAME[] = "Mask Generation";
/* Changed version 2.18 for automated versiion support -- see below */
// const	char	VERSION[] = "2.15";
 
/* Standard Includes */
  
#include        <stdio.h>
#include        <string.h>
#include        <stdlib.h>
#include	<math.h>
 
/* Application includes */

// #include	"mgfeats.h"
// now in maskdat.h

#include	<unistd.h>	// For "access"

#include	"dbplot.h"
// dbplot.h   includes mgutils.h;
// mgutils.h  includes optutils.h
// optutils.h includes maskdat.h
// maskdat.h  includes ioutils.h, kdcutil.h

const	char	VERSION[] = Xstr(MASKGNV);
char	XVERSION[16] = Xstr(MASKGNV);

// Standard Version structure
const	Vers	Program = { NULL, NULL,  MASKGNV, MASKGNR, MASKGNT };
#define  APPLICATION_NAME  "Mask Generation"
const	char	PROGNAME[] = MASKGNT;
// mgutils expect PROGNAME and VERSION strings to exist as global labels

int     inst = -1;      // start with unknown
#ifdef  USE_CDAT
cdata	*CDhead = NULL;  // global, should be a parameter (of main)
#endif
#ifdef  USE_GSTO
gsto	*CDhead = NULL;  // global, should be a parameter (of main)
#endif
#ifdef  USE_BSTO
BNODE	*CDhead = NULL;  // global, should be a parameter (of main)
#endif

/*  ==  Global definitions and storage  ==  */

char	myname[16];	// Fill in our command name here
/* The command name is used to decide whether to initialize the
  debug plot with mask or detector space.  */

/*  ===  Debug Subroutines  ===  */

#include	"msdebug.c"

#if (DEBUG_GEN >= 0)
// This is here to placate lint on mgfeats.h being needed
#define  maskgen_dummy 0
#endif

/*  ===  End Debug Subs.  === */


/*  ===  Dummy Routines  ===  */

//int	UpdateTextWindow (void* wb)
/* The real one is in guisubs.c (included by intgui.c), and is
  refered to by guiutil.c which is separately compiled.  */
//{
//    return  0;
//}

/*  ===  End Dummy Subs.  ===  */



/* Count nodes not in mask space */
static	int	mask_space (Obs* ob) {
objq	*head = NULL;
objq	*p;
double	mxr;	/* maximum mask radius */
int	k=0;	/* return value */
//int	n;	/* debug method */
char	*inam;
Mindx	msk;	/* Mask index */

avoids	*aqp;	// Avoidance queue
double	rt;	// tool radius

    if (ob == NULL) return  0;
//    else  return 0;	// debug bypass.
    msk = MaskIndx (ob);	// Mask index found
    head = ob->ob;

#ifdef  GISMOAP
// Just return if we have gismo aperature special instrument.
//    if (!strncasecmp(ObInstN(ob), "IMACS_gd", 8))  return  k;
    if (FeatIndx(ob) == F_gisap) return k;
//	k = ((sfset*)ob->Frontend->data)->n;
// really should add any debug objects, but that is not actually needed.
#endif

/* Find the maximum radius for the mask space */
    inam = (ob->Instr_mt == NULL) ? NULL : ob->Instr_mt->name;
    inst = InstIndx (inam);
    mxr = (inst == I_imacs) ? D_maxradi : D_maxradl;

/* Do lookup stuff outside the loop */
#ifdef  USE_CDAT
    aqp = CDhead[inst].av;
    rt  = CDhead[inst].tool;
#endif
#ifdef  USE_GSTO
    aqp = *(avoids**)GS_lookup (&CDhead, inst, CD_av, 0);
    rt  = *(double*)GS_lookup (&CDhead, inst, CD_tool, 0);
#endif
#ifdef  USE_BSTO
    aqp = getlsq (&CDhead,InstCase(inst),"Avoid");
    rt = floatvalue(getval(&CDhead,InstCase(inst),"Rtool"));
#endif

//    for (k=0, p=head; p != NULL; p = (p->next == head) ? NULL : p->next) {
    C_loop (p, head) {

	if ( (p->flag & OBJECT_ACTIVE) &&
	    !(p->flag & OBJECT_VIRTUAL) ) {
/* We do not test virtual objects w.r.t. mask space at all; they exist
  only in detector space, logically.  */

	    if (msk == M_gismo) {	// gismo - rectangular mask
// Here, we test if the image (input) slit falls entirely within the
// sub-area in which the object is found.
bbox	bbs;	// Slit bounding box
int	bo;	// Bounding box overlap
//int	ns;	// subfields
sbfld	*sf;	// subfield array
sfset	*ss;	// data from element

// Find the front end sub area bounding box...
// If none, p has to be declared non-active here.
// element pointer is ob->Frontend
// element data points to a sfset struct
// the sfstruct has n sbfld elements in array s
// the sbfld struct s[sfx] contains bbox field.

// Check sfx pointing to a valid sub-area...
		if (ob->Frontend != NULL) {
		    ss = (sfset*) (ob->Frontend->data);
		    if (p->sfx >= 0 && p->sfx < ss->n) {  // good n
			sf = ss->s;
		    } else p->flag &= ~OBJECT_ACTIVE;
		} else {	// Very wrong, must kill object
		    p->flag &= ~OBJECT_ACTIVE;	// inactivate
		}

		if (p->flag & OBJECT_ACTIVE) {	// might have been dropped above
// Find slit bound box
		    bbs = bbslit (p->fpos, p->islit);
// compare slit to the area bounding box.
		    bo = bbover (bbs, sf[p->sfx].field);
// We want bbs entirely within subarea, so we need bo == 2 here.
		    if (bo != 2) {	// if ==, all is good, we skip out
			if (ob->dlevel > 0)
			    printf (" Object \"%s\" slit outside subarea.\n",
				p->dat.name);
			p->flag &= ~OBJECT_ACTIVE;	// inactivate
			k++;
			continue;
		    }	// End bbs check, leave silently here
		} else {	// Should not happen...
		    if (ob->dlevel > 0)
			printf (" Object \"%s\" has confused subarea!\n",
			    p->dat.name);
		    k++;
		    continue;
		}
// If we haven't done a continue here, the object passed the test,
// and we silently drop to an avoidance check section.

// could also compare p->slit to actual mask bound, but this is probably
// not necessary, as all input areas should image in bounds.

#ifdef  WFCCD
	    } else if (msk == M_wfcd) {  // WFCCD mask - rectangular
// Test object position w.r.t. a global bounding box for the
// WFCCD mask space allowed.  It is defined in the same header as
// D_maxrad values are defined, and as a const bounding_box type. (mgutils.h)
// The bbox is "WFC_mask" defined in mgutils.h.

		if (2 != bbover (bbslit (p->fpos, p->islit), WFC_mask)) {
// Copied that from some gismo valid code; hope it's correct!
if (ob->dlevel > 1) {
// ADD A STATIC COUNTER AND REPORT ONLY FIRST 10 THINGS
	printf (" Object \"%s\" outside mask boundary.\n", p->dat.name);
}	// debug section
		    p->flag &= ~OBJECT_ACTIVE;  /* Set bit off */
		    k++;
		    continue;
		}

#endif

	    } else {		// Other - circular masks
/* Check for being on mask - circular masks here */
		if (Xcircle(p, make2vect(0.0,0.0), mxr) >= 0) { // outside mask
if (ob->dlevel > 1) {
// ADD A STATIC COUNTER AND REPORT ONLY FIRST 10 THINGS
	printf (" Object \"%s\" outside mask boundary.\n",
		p->dat.name);
}	// debug section
		    p->flag &= ~OBJECT_ACTIVE;  /* Set bit off */
// The inactivated node has no conflicts yet; we do this before
// any conflicts are found.  Otherwise we would need to call
// the "set_inactive" routine specifying conflict head too.
		    k++;
		    continue;
		}	// test on active radius
	    }		// test on type of mask
// Loop to check on avoidances...
// Set up the "inst" index based on which instrument we are using.

	    if (inst != I_lds3) {	// Not needed or appropriate
// Loop over all avoids
// Find distance from this slit to the avoid
// If too small, kill the node with mesage
		if (avoidanc (p, aqp, rt) ) {
if (ob->dlevel > 0) {
	printf (" Object \"%s\" in an avoidance.\n",
		p->dat.name);
}  // end debug section
		    p->flag &= ~OBJECT_ACTIVE;  /* Set bit off */
		    k++;
		    continue;
		}	// Avoidance test
	    }	// Instrument test
	}	// active objects only test

    }	// loop over all objects
    return  k;
}


        /* - - - - - - - - *\
        |                   |
        |    Main Program   |
        |                   |
        \* - - - - - - - - */

#define  GETOPTUSE
//#undef  GETOPTUSE
/*  Symbol  GETOPTUSE  is used to turn on (or off) the use of the
  standard "getopt" utility in parsing the command line.  Use this
  local define to enable the "old" method quickly if needed.  */

/*  ====  Main Program  ====  */

#define  DATFILE  "opticdef.dat"
#define  DATtFILE  "opticdef.tst"
#define  CUTFILE  "maskcut.cfg"

int  main (int argc, char** argv) {
#ifdef  GETOPTUSE
extern char *optarg;
extern int optind,opterr,optopt;
int c;
#else
char    *pp;    /* Scan variable */
char    px;     /* Scan character */
#endif
int     i;
int     dohelp = 0;
char	infile [120];
char	*filename;
namlist	*head = NULL;
FILE	*INF;
Obs	*obs;	/* The observation structure */
objq	*ob = NULL;	/* The master object list */
//char	*a;	/* scratch */
objq	*rob = NULL;	/* Removed objects */
int	reord=0;
int	wings=0;	// Normally disabled
int	quik=0;		// Normally disabled
int	dplot=1;	// Normally enabled
int	actf=0;		// Normally disabled
int	istest = 0;
#ifdef  OBJNAM
int	pnam=0;		// Normally disabled
#endif
#ifdef  REFLIM
int	ralin=1;	// Normally enabled
#else
int	ralin=0;	// Normally disabled
#endif
#ifdef  OBJCHECK
int	cexit = 1;	// Exit if check is failed
#endif
//#ifdef  GISMO
int	tfil = 0;	// Normally disabled
int	agrf = 0;	// Default no accurate rotation
#ifdef  GISMOAP
int	gisap = 0;
#endif
//#endif
//#ifdef  GSREQUIRED
int	gsrx = 0;	// Normally required
//#endif
int	mexit = 1;	// Exit if name check fails
int	n;
int	k;
int	bad;
int	kr=0;
int	wlevel=0;	// Standard debugging level set to none
element	*gs = NULL;	/* Optic details queue pointer */
char	xnam[64];	// Part of message
char	mbuf[128];	// Message to be generated...
char	xfilen[16];	// For the compare
//extern	char	MGUVERS[];	// Needed for reference
char	*fspec = NULL;	// For finding file
namlist	*comps = NULL;
namlist *nlx;		// Scan pointer for namlist
char	*vstr;
Pvfs	*cp = NULL;
Pvfs	*fp = NULL;
int	ns = 0;		// Nod & Shuffle status
int	ucl = 0;	// Uncut length existing
char	answer[16];

/* Other things, to become obsolete... */

/* Standard input logic:  If argument given, take that argument as the
  sole specification to process.  If no argument is given, prompt for
  an argument, and repeat the prompting until a null reply.  */

    *infile = '\0';
    dodebug = 0;	/* NOTE = temporary for debugging */
    dbfile = NULL;	/* temporary (used by dbroq) */

/* To fix up the debug stuff:
  1) put dodebug into Obs structure as dlevel, drop the global
  2) Add wlevel as local variable (on -i flag) to minimize
	the debug stuff to 0, but to plot the diagram
  3) Change dodebug stuff in all subroutines and library...
..*/

/* Store our command name in a global... */
    strncpy (myname, cfilename(argv[0]), 16);
    sprintf (xnam, "Executing \"%s\".   ", myname);	// used for debug
/* NOTE that we do not do \n in the above debug line... */
    istest = strncmp (myname, "tmaskgen", 8) == 0;

#ifdef  GETOPTUSE
/* Decode all the options of the command line */
/* :: follows letter with optional argument, : follows one with required
  argument, otherwise there are no arguments to the options.  */
    while((c = getopt(argc, argv, "acd::fgimnpqrstwhS")) != -1) switch(c) {
#ifdef  REFLIM
	case 'a':	ralin = 0;	break;  // no remove alignment
#else
	case 'a':	ralin = 1;	break;  // remove alignment
#endif
#ifdef  OBJCHECK
	case 'c':	cexit = 0;	break;  // no halt if obj chk fails
#endif
	case 'd':				// debug level
	    wlevel = optarg == NULL ? 1 : atoi (optarg);	break;
	case 'f': 	actf = 1;	break;  // active file writing
#ifdef  GSREQUIRED
	case 'g': gsrx = istest ? 1 : gsrx;	break;  // special permission
	case 'S': gsrx += gsrx ? 1 : 0;	break;
// Note that the -g flag is non-pub for test and long slit type masks only
#endif
//	case 'i':	wlevel = 0;	break;  // no debug (obsolete)
	case 'm':	mexit = 0;	break;  // skip file name match
#ifdef  OBJNAM
	case 'n':	pnam = 1;	break;  // plot names
#endif
	case 'q':	quik = 1;	break;  // quick look
	case 'r':	reord = 1;	break;  // reorder
	case 's':	dplot = 0;	break;	// skip plot
//#ifdef  GISMO
	case 'p':	agrf = 1;	break;	// accurate rotation
	case 't':	tfil = 1;	break; // test file
//#endif
	case 'w':	wings = 1;	break; // wing support
	case 'h':	dohelp = 1;	break;	// parameter help
	default:
	    printf (" unused %c option\n", c);	break;
    }  /* end of switch, only statement in the getopt while loop */

    for (c = optind; c < argc; c++) {
	sprintf (mbuf, " %sArgument %s.\n", xnam, argv[c]);  /* debug (old) */
	head = pushtxt (argv[c], head);  /* Collect arguments */
/* NOTE that dynamstr must be used to put any static string into a namlist,
  usually in  get_nam, and the like.  It will cause a crash at the
  end when the name queue is dropped, if this is not done correctly.
  That is why pushtxt is used above.  */
    }

#else  // on GETOPTUSE
/* Scan for switches and arguments */
    for (i=1; i<argc; i++) {    /* argv[0] is command name */
        pp = argv[i];
        if (*pp == '-') {       /* Scan normal switch */
            while ((px = *++pp) != '\0') {
                if (px >= 'A' && px <= 'Z') px += 'a' - 'A';
                if      (px == 'd')  wlevel = 1;  /* Debug output */
/* May want -d to cause debugging later? */
                if      (px == 'i')  wlevel = 0;  /* Debug output */
		else if (px == 'r')  reord = 1;   /* Reorder */
                else if (px == 'h')  dohelp = 1;  /* show help */
		else if (px == 'q')  quik = 1;	/* Quick Look feature */
		else if (px == 's')  dplot = 0;  /* Skip plot */
#ifdef  OBJCHECK
		else if (px == 'c')  cexit = 0;
#endif
		else if (px == 'm')  mexit = 0;
#ifdef  OBJNAM
		else if (px == 'n')  pnam = 1;  /* Plot names */
#endif
//#ifdef  GISMO
		else if (px == 'p')  agrf = 1;  /* Accurate rotation */
		else if (px == 't')  tfil = 1;  /* Use test file */
//#endif
		else if (px == 'a')  ralin = 1;	/* Remove alignments */
		else if (px == 'w')  wings = 1;	/* Wing chip support */
		else if (px == 'f')  actf = 1;  /* Active file writing */
#ifdef  GSREQUIRED
		else if (px == 'g')  gsrx = 1;  /* Special permission */
// Note that the -g flag is non-pub for test and long slit type masks only
#endif
            }
        } else if (*pp == '+') {        /* Scan positive switch */
            while ((px = *++pp) != '\0') {
                if (px >= 'A' && px <= 'Z') px += 'a' - 'A';
                if      (px == 'd')  wlevel = 2;      /* Debug output */
                else if (px == 'a')  ralin = 0;  /* No remove alignments */
                else if (px == 'h')  dohelp = 1;  /* show help */
            }
        } else if (*pp == '?') {
	    dohelp = 1;
        } else {
#ifdef DEBUG
            printf ("Adding %s to list.\n", pp);  /* debug */
#endif
	    head = pushtxt (pp, head);  /* Collect arguments */
/* NOTE that dynamstr must be used to put any static string into a namlist,
  usually in  get_nam, and the like.  It will cause a crash at the
  end when the name queue is dropped, if this is not done correctly.  */
	    sprintf (mbuf, "%sArgument %s.\n", xnam, pp);  /* debug */
/*	    if (*infile == '\0') strcpy (infile, pp);  */
        }
    }    
#endif	// on GETOPTUSE

// Create the combined version dynamically for titles
    vstr = combinev (MASKGNV, MGUTILV);
    strcpy (XVERSION, vstr);
    free (vstr);

    if (wlevel > 0 || dohelp) {
//	printf (" Mask Generation program, Version %s%s\n", VERSION, MGUVERS);
//	printf (" %s, Version %s%s", PROGNAME, VERSION, MGUVERS);
//	printf (" %s, Version %s", PROGNAME, XVERSION);
	printf (" %s, Version %s", APPLICATION_NAME, XVERSION);
	if (istest) printf ("   (Test Version)");
	if (wlevel > 1) printf ("  [debug %d]", wlevel);
	printf ("\n");
    }

/* See if this version is any good, or if we continue.  */
    if (quik == 0) {	// skip check for online quick look
//    comps = pushtxt (INTGUIT, comps);
//    comps = pushtxt (SKYWINT, comps);
	comps = pushtxt (MASKGNT, comps);
	comps = pushtxt (MGUTILT, comps);
// Would add any version data for debug plot programs too
	k = check_ver (comps, &cp, &fp);
	comps = kill_name (comps);
// if feature exists, we could exit on k < -2 here.
	if (k < -1) {
	    printf (" ** Operation of Obsolete software rejected.\n");
	    kr = 1;  // or maybe kr = k; ?
	    goto main_finish;
	}
    }

    if (dohelp) {
//	printf ("\n %s -- Version %s%s", PROGNAME, VERSION, MGUVERS);
	printf ("\n  Parameters are: ");
	printf ("\n    Observation File(s) (optional)");
	printf ("\n   Switches");
#ifdef  REFLIM
#ifdef  GETOPTUSE
	printf ("\n     -a = No auto-remove alignments");
#else
	printf ("\n     +a = No auto-remove alignments");
#endif
#else
	printf ("\n     -a = Remove off-detector alignments");
#endif
#ifdef  OBJCHECK
	printf ("\n     -c = Do not halt if object check fails.");
#endif
#ifdef  GETOPTUSE
	printf ("\n    -d* = Set debug level to (*) [1]");
#else
	printf ("\n     -d = Some debug information");
	printf ("\n     +d = More debug information");
#endif
	printf ("\n     -f = Create .obx file of objects used.");
	if (istest)
	{
	printf ("\n     -g = Guide star unneeded"); //Skip guide star test.");
//	if (!istest)
	printf (" and omit .SMF output.");
	}
#ifndef  GETOPTUSE
	printf ("\n     -i = Minimal debug information [default]");
#endif
	printf ("\n     -m = Skip the file name match test.");
#ifdef  OBJNAM
	printf ("\n     -n = Plot object names");
#endif
	printf ("\n     -q = Quick Look option (no .SMF)");
	printf ("\n     -r = Reorder objects");
	printf ("\n     -s = Skip plotting");
//#ifdef  GISMO
	printf ("\n     -p = Use accurate rotation in GISMO.");
	printf ("\n     -t = Use test version of opticdef.");
//#endif
	printf ("\n     -w = Add wingchip holes");
	printf ("\n     -h = Display this help");
	printf ("\n   Output written to standard output.");
/*	printf ("\n   Output file name is prompted for."); */
	printf ("\n \n");
	kill_name (head);
	kr = 1;		// return value
#ifdef  OPTDEBUGTEST
	if (wlevel<2) goto main_finish;		// debug version
#else
	goto main_finish;
#endif
    }

/* Print leading message if level is ok */
    if (wlevel > 0) printf ("%s", mbuf);	// the \n is already there

	/* - - - - - - - - - - - - - - *\
	|				|
	|    Get optic, cutting data	|
	|				|
	\* - - - - - - - - - - - - - - */

/* Find, open the optic data file.  Issue error message and
  keep the pointer NULL if the file is not found. */
//#ifdef  GISMO
    if (tfil && access (DATtFILE, R_OK) == 0)
	gs = get_optic_data (DATtFILE, wlevel);
    else
// That's a little cheat; the next line is the else clause, but only
// in the GISMO case; otherwise it's the only thing.
//#endif
    gs = get_optic_data (DATFILE, wlevel);
    if (gs == NULL) { kr = 0; goto main_finish; }	// was return 0;
/* Note that we MUST have optic data to proceede! */
#ifdef  OPTDEBUGTEST
    if (dohelp) goto main_finish;	// debug version
/* This is here because we could have failed to go to main_finish above
  in the dohelp section if wlevel were low [default case]. */
#endif

    CDhead = get_cut_data (CUTFILE, wlevel);
    if (CDhead == NULL) { kr=0; goto main_finish; }
/* Note that we now require cut data to proceede also */

/* Check on returned value being NULL, and exit if it is, since we need
  the configuration data to correctly process the mask.  */

//    if (*infile == '\0') askmestr (infile, 100, "Input observation file:");
/* Prompt for input if no input file names have been given */
    if (head == NULL) {
	askmestr (infile, 100, "Input observation file:");
	head = pushtxt (infile, head);
    }

/* Loop on files to process */
    while (head != NULL) {
/* Get, and open if possible, the (next) file name requested */
	filename = pop_name (&head);
	if (filename == NULL) continue;
/* Change to using findfile rather than iopenx to better control
  the debug messages produced.
  And also, change to locfile to be more compact.  */
	if (fspec != NULL) {
	    free (fspec);
	    fspec = NULL;
	}
/* Now, we keep fspec for error messages below, and clear it if needed
  just before use here.  The last one might stay allocated.  Too bad.  */
//	INF = iopenx (filename, "r", ixlist);
#ifdef  LOCATE
//	fspec = locfile ("r", filename, ", .obs, .obsl", ",./");
	fspec = locfile ("r", filename, ", .obs, .obsl", NULL);
#else  // on LOCATE
	findfile (&fspec, "r", filename, PathName); // dd);  // "./");
	sprintf (mbuf, "%s.obs", filename);
	findfile (&fspec, "r", mbuf, PathName);
	sprintf (mbuf, "%s.obsl", filename);
	findfile (&fspec, "r", mbuf, PathName);
#endif  // on LOCATE
	if (fspec != NULL) {
	    INF = iopen (fspec, "r");
//	    free (fspec);
//	    fspec = NULL;
	} else INF = NULL;

	if (INF == NULL) {
	    printf (" ** Could not find a \"%s\" obs file.\n", filename);
	    free (filename);
	    filename = NULL;
	    continue;
	}

/* Process the file that is found and opened. */
	if (wlevel > 0) printf ("Processing %s.\n", filename);
//	free (filename);	/* no longer needed */
// Probably don't want to free filename, it may be pointing to
// an argument storage place...  (That's been fixed)
// Those few cases in which we get a dynamic filename, it was prompted
// for, and we can afford the memory loss.

/*  ---  Start the defined process for mask generation  ---  */

/* ** Read the observation data file, set defaults, and prompt for
  anything which will be needed */
	obs = read_obsfile (INF, gs);
/* Check the obsfile for security. */
#ifdef  FILESEC
	k = verifide (INF, dodebug-1);
#endif
/* At present, allow even the failure status to proceede.  */
	iclose (INF);

/* ** Add flag(s) from command line if needed in obs.  */
//#ifdef  GISMO
	if (agrf) obs->Flag |= OFagr;
//#endif

/* ** Check some basic things which are required in obs ... */
	if (obs->Instr_mt == NULL) {	// instrument must be present
	    printf (" ** Sorry, no instrument found in %s\n", obs->fname);
	    goto main_loop_end;
	}

// For special instrument gismo aperature mask, we need to set a special
// flag to allow skipping some checks
#ifdef  GISMOAP
//	if (strncasecmp(obs->Instr_mt->name, "IMACS_gd", 8) == 0) gisap = 1;
	if (FeatIndx(obs) == F_gisap) gisap = 1;
	else  gisap = 0;

//	if (obs->dlevel > 0) printf (" ** FeatIndex = %d; gisap = %d\n",
// dlevel has not been set yet, use wlevel.
	if (wlevel > 0) printf (" ** FeatIndex = %d; gisap = %d\n",
		FeatIndx(obs), gisap);
//#else
//	printf (" ** GISMOAP is not set.\n");
#endif

	if (obs->Disperser == NULL
#ifdef  GISMOAP
		&& gisap == 0
#endif
	   ) {	// disperser must be present
	    printf (" ** Sorry, no disperser found for %s\n", obs->fname);
	    goto main_loop_end;
	}

	strncpy (xfilen, cfilename(filename), 8);
	if (mexit && strncmp (xfilen, obs->fname, 8) ) {  // check file name
	    printf (" ** File name error.  Reading \"%s\"\n", filename);
	    printf ("   -- has internal file name: \"%-.8s\".\n", obs->fname);
	    printf ("  Names must match.  Processing denied.\n");
	    printf ("  (Use -m flag to bypass this test).\n");
	    goto main_loop_end;
	}

	k = quik ? 0 : comp_vers ((Vers*)obs->source, cp);	// was fp
	if (k < -1) {
	    printf (" ** Sorry.  The %s software version %s\n",
		((Vers*)obs->source)->tag, fvstr(((Vers*)obs->source)->v) );
	    printf ("  which produced %s", fspec);  // or filename  ?
	    printf (" is %s.\n", (k == -2) ? "Obsolete" : "Unknown" );
	    printf ("  Please fix that.  Processing denied.\n");
	    goto main_loop_end;
	} else if (k != 0) {
	    printf (" * Warning!  The %s software version %s\n",
		((Vers*)obs->source)->tag, fvstr(((Vers*)obs->source)->v) );
	    printf ("  which produced %s", fspec);  // or filename ?
	    printf (" is %s.\n", (k < 0) ? "Older" : "Beta (test)" );
	    printf ("  For best results, use Current software.\n");
	}

/* Call to check/set default temperature to cutting temp. if needed. */
	check_temp (obs, CDhead);

/* Set global debug level into obs structure */
	obs->dlevel = wlevel;

/* Put global flag(s) for options into obs */
	if (ralin) obs->Flag |= OFral;

	if (obs->dlevel > 2) dumpobs (obs);
#ifdef  OBJNAM
	if (pnam) obs->Flag |= OFpon;
#endif

/* Implement the quik (-q) option here... (kills slit extension) */
	if (quik) {
	    obs->slext = 0;
#ifdef  OBJNAM
	    obs->Flag &= ~OFpon;
#endif
	    if (obs->dlevel > 0) printf (" -- Quick Look execution. --\n");
	}

/* ** Check that the instrument and telescope are compatable... */
	if (strstr(obs->Instr_mt->name, "IMACS")) {
	    bad = strcmp(obs->Tel_scope->name, "Magellan");
	} else if (strstr(obs->Instr_mt->name, "LDSS")) {
	    bad = (strcmp(obs->Tel_scope->name, "Magellan2") ) &&
		(strcmp(obs->Tel_scope->name, "Mag2NoADC") );
	} else {
	    printf (" Unknown instrument %s.\n", obs->Instr_mt->name);
	    bad = 1;
	}
	if (bad) {
char	reply[32];
char	bell = 0x07;	// was '\7';
char	ans[] = "Yes";
	    printf (" ** Telescope incompatable with instrument -");
	    printf (" %s for %s %c!\n", 
		obs->Tel_scope->name, obs->Instr_mt->name, bell);
	    askme (reply, 30, " -- Is this intentional? ");
// User needs to answer "yes" (case insensitive) here.
	    if (strncasecmp(reply, ans, 3)) {
		printf (" Please change TELESCOPE line in .obs file.\n");
//		return 1;
		kr = 1;
//		goto main_finish;
// Within the loop, we must kill obs data and loop back.
		goto main_loop_end;
	    }
	}	// Test added KDC 05/4-18

/* ** Check for wavelength range and center wavelength resonable. */
#define  SHORTEST_WL  3399.0
#define  LONGEST_WL  12000.0
/* Those are outside published limits for IMACS and LDSS instruments. */
	if (obs->pb.blue < SHORTEST_WL || obs->pb.red > LONGEST_WL)  {
	    printf ("** Wavelength limits out of bounds.\n");
	    goto main_loop_end;
	}
	if (obs->pb2.blue < SHORTEST_WL || obs->pb2.red > LONGEST_WL)  {
	    printf ("** Detector wavelength limits out of bounds.\n");
	    goto main_loop_end;
	}
	if (obs->cw < SHORTEST_WL || obs->cw > LONGEST_WL) {
	    printf ("** Center wavelength out of bounds.\n");
	    goto main_loop_end;
	}
	if (obs->sw < SHORTEST_WL || obs->sw > LONGEST_WL) {
	    printf ("** Spectrum central wavelength out of bounds.\n");
	    goto main_loop_end;
	}	// Tests added KDC 06/4-6

/* ** Add a check on horizon-to-horizon rotator conflict (rare) */
	C_loop (nlx, obs->pcom) {
	    if (strstr(nlx->name, "Rotator conflict") == NULL) continue;
	    if (strstr(nlx->name, "Horizon to Horizon") == NULL) continue;
/* Please NOTE that these comparisons are case sensitive.  The messages are
  created in intgui.c:967 and must match case.    */
	    printf ("** Found Rotator conflict from Horizon to Horizon,\n");
	    printf ("**  OR the field is too far north to be observed.\n");
	    goto main_loop_end;  // Added KDC 10/11-17
	}
/* Also NOTE that due to the way intgui searches for rotator conflicts, the
  above test will also find and reject objects whose declination is so great
  that they never rise above the north horizon.    */

/* ** Check for guide stars properly selected */
// Leave promptly if we don't have an instrument here
	if (obs->Instr_mt == NULL) goto main_loop_end;

#ifdef  GSDATA
// Only need to check if we have IMACS instrument
	if (InstIndx(obs->Instr_mt->name) == I_imacs
#ifdef  GISMOAP
		&& gisap == 0
#endif
/*  NOTE that gisap == 0 and reflimit > 0 are redundant, and the special
  gisap test may be removed here, when reflimit is cleared by gisap.    */
		&& obs->reflimit > 0	// Not position independent
/*  ALSO NOTE that the test on reflimit could be replaced by a test
   on actual reference object count > zero.      */
/*  NOTE that at this point, the fill objects has not yet been done, so a
  count of active reference objects is not available, and obs->actref is
  in fact zero, so such a test would be wrong.    */
	   ) {
// Count the number of selected guide stars, should be 2
	    k  = (obs->gp1.ep == 0.0) ? 0 : 1;
	    k += (obs->gp2.ep == 0.0) ? 0 : 1;
// Give message if we don't have sufficient guide stars 
	    if (k < 2) {
//		printf (" ***\n");
		printf ("\n ***  --  --  --  --  --  --  --  ***\n");
		printf (" *** Please select 2 guide stars! ***\n");
//		printf ("  *  Please select 2 guide stars.\n");
		printf (" ***  --  --  --  --  --  --  --  ***\n\n");
#ifdef  GSREQUIRED
/* In case of noquikgs, we don't bypass if quik is true */
#ifdef  NOQUIKGS
		if (!(quik || gsrx)) goto main_loop_end;
#else
		if (!(gsrx)) goto main_loop_end;
#endif  // on NOQUIKGS
/* Need both gxrx (from -g flag) and istest (executing tmaskgen) 
  to allow execution to continue.  */
#endif  // on GSREQUIRED
	    }
	}
// could have else test on index for other instruments...
#endif  // on GSDATA

/* ** Check for Nod&Shuffle conflicts */
	ns = fabs(obs->nrbias) > NSrlim;
	ucl = (obs->dca != 0.0) || (obs->dcb != 0.0);
	k = 0;		// local error count
// See if LDP prism is in use, and warn if no N&S rotation bias
	if ( (InstIndx(obs->Instr_mt->name) == I_imacs) &&
#ifdef  GISMOAP
// strongly hope this prevents following line from trying
// to reference obs->Disperser which may be null if gisap is on!!
		gisap == 0 &&
#endif
	    (strcasecmp (obs->Disperser->name, "LDP_Prism") == 0) &&
	    (ns == 0) ) {
	    printf (" \n");
	    printf (" **  *  *  *  *  *  *  *  *  *  *  *  *  *  **\n");
	    printf (" **  The LDP Prism is being used, but the   **\n");
	    printf (" **  Nod & Shuffle mode is not selected.    **\n");
	    printf (" ** This is likely a mistake, please check. **\n");
	    printf (" **  *  *  *  *  *  *  *  *  *  *  *  *  *  **\n");
	    k++;
	}

// Check for N&S being set if uncut length is also being used
//	if (ucl && (ns == 0)) {
// Actually, we should only do this for IMACS as ns is not set in LDSS
	if ( (InstIndx(obs->Instr_mt->name) == I_imacs) &&
	    ucl && (ns == 0)) {
	    if (k == 0) printf (" \n");
	    printf (" **  *  *  *  *  *  *  *  *  *  *  *  *  *  **\n");
	    printf (" **  An uncut length is specified, but the  **\n");
	    printf (" **  Nod & Shuffle mode is not selected.    **\n");
	    printf (" ** This is likely a mistake, please check. **\n");
	    printf (" **  *  *  *  *  *  *  *  *  *  *  *  *  *  **\n");
	    k++;
	}

// Also check uncut length set if N&S has been set
	if (ns && (ucl == 0)) {
	    if (k == 0) printf (" \n");
	    printf (" **  *  *  *  *  *  *  *  *  *  *  *  *  *  **\n");
	    printf (" **  The Nod & Shuffle mode is selected,    **\n");
	    printf (" **  but no uncut slit length is specified. **\n");
	    printf (" ** This is likely a mistake, please check. **\n");
	    printf (" **  *  *  *  *  *  *  *  *  *  *  *  *  *  **\n");
	    k++;
	}

// If we have errors, ask to continue at this point...
	if (k > 0) {
	    askme (answer, 15, " ** Do you wish to continue [no|yes] ? ");
	    if (strncasecmp(answer,"yes",3)) goto main_loop_end;
	}

/* ** Set GISMO slit rotation flag automatic feature.  */
//#ifdef  GISMO
	if ( ucl && obs->default_slit.shape == RECTANGLE && 
	    !strncasecmp(obs->Instr_mt->name, "IMACS_l", 7) ) {
	    obs->Flag |= OFagr;
	}
/* The above sets accurate rotation flag when we have a rectangular slit
  and an uncut length and we are using long camera.  That flag is only
  used in the case of GISMO mask transforms, so we don't check for the
  presence of the GISMO frontend here.  */
//#endif

/* ** Read all the object list files in a subroutine */
	fill_objects (obs);
	ob = obs->ob;	// May be needed?
// ob is used below.  It starts as null.  Since it should point to
// the object queue, we set it here, although is may not be used yet.

// The special GISMO aperature mask is generated by obtaining a
// special object queue above in fill_objects

#ifdef  NSROT
/* ** Check for and rotate the field if a bias amount is present */
//	rotobj (obs->ob, obs->nrbias, 0);
// Above removed; see nsrot.otl for details; now rotated in object_setup
// Also, suppress wing chips if nrbias is not zero
// Suppress slit extension also (maybe others)
	if (fabs(obs->nrbias) > NSrlim) {
	    wings = 0;
	    obs->slext = 0;
	}
#endif

	dbroq ("Start", obs);

/* ** Call subroutine to process list for geometry, and construct the object
  parameters in the image plane, and detector plane. */
/* The pointers to the optics element queue are inserted in obs when
  it was read in read_obsfile above, and gs was passed to it. */
	k = listcount (obs->ob);
	if (obs->dlevel > 1) printf (" Total of %d objects read.\n", k);
#ifdef  GISMOAP
      if (gisap) n = k;   // kludge here
      else
// NOTE that this is a total kludge, affecting only the
// immediately following statement.  Should provide brackets conditionally.
#endif
	n = object_setup (obs);
	if (obs->dlevel > 1) {
	    printf (" Setup found %d objects.\n", n);
//	    if (obs->Frontend != NULL) 
//		printf (" -- %d objects were rejected for position.\n", k-n);
/* This message means that since k objects were read from the file, and only
  n exist after setup, the difference were rejected for their position,
  being not in a subfield or their spectra being off detector.  There
  should also be messages of out of field for some of these.  */
	}
	dbroq ("Did setup", obs);
	fflush (stdout);

/* ** Check if global spectrum length error exists */
	if (obs->dlevel > 1) printf (" Flag = %08X \n", obs->Flag);
	if (obs->Flag & OFexe) {	// The x40 bit here (see maskdat.h)
	    printf ("** !! Spectral range is too large.\n");
	    printf ("  All spectra are too large for the detector.\n");
	    printf ("  Reduce wavelength limits or dispersion!\n");
	    if (obs->dlevel < 2) goto main_loop_end;
// above omits the failure exit if we are debugging
	}

/* ** If the feature is on, add virtual gap objects. */
#ifdef  GAPVO
	add_gaps (obs);
//	Gap_count (1, obs->ob);		// debug only
#endif

/* ** Sort the object list on the detector position */
//    if (obs->dlevel) show_objqd (obs->ob);
// Do the sort, and then show the queue again...
	ob = sort_obj (obs->ob, sort_decide);
// printf (" --Finished sort_obj\n");  fflush (stdout);  // debug

	obs->ob = ob;	/* Put it back into structure as sorted */
/* This is also needed to keep object queue header straight for
  the processing which follows, and uses obs->ob pointer. */
	if (obs->dlevel > 2
#ifdef  GISMOAP
		&& gisap == 0
#endif
	   ) {
	    printf (" Sorted queue on Bx\n");
	    show_objqd (ob);
	}

/* Debug check of sort here. */
	k = active_objs (ob);
	n = sort_check (ob, sort_decide);
	if (n > 0) printf (" *!* Sort order failed %d times for %d!\n", n, k);
	else if (obs->dlevel > 1) 
	    printf ("  =*= Sorted %d active objects successfully.\n", k);

/* ** Call subroutine to check all slits for being on mask */
#ifdef  NSROT
/* De-rotate mask space, and rotate slits for the tests */
	rotobj (obs->ob, -(obs->nrbias), 1);
#endif
	n = mask_space (obs);
#ifdef  NSROT
/* Restore mask space rotation, and old slit angles */
	rotobj (obs->ob, (obs->nrbias), 1);
#endif

	if (obs->dlevel > 0 && n > 0)
	    printf (" Removed %d objects outside mask bound.\n", n);

/* ** Call subroutine to check for unusual object properties.  */
#ifdef  OBJCHECK
	if (object_check(ob) > 0) {
	    if (cexit) {
		printf (" If you would like to accept these errors,");
		printf (" and continue execution,\n");
		printf (" execute this program with the \"-c\" flag");
		printf (" in the command line.\n\n");
		goto main_loop_end;
	    } else {
		printf (" Executing as requested despite ERRORS.\n\n");
	    }
	}
#endif  // on OBJCHECK

/* ** Call subroutine to check overlaps or conflicts among the objects
  in the object list. */
	dbroq ("Before scan", obs);

	if (obs->dlevel > 1) {
	    printf (" Scanning %d objects for conflicts...\n",
		active_objs(obs->ob) );
	    fflush (stdout);
	}

	n = scan_conflicts (obs);
	if (obs->dlevel > 0) printf (" Found %d conflicts.\n", n);

	dbroq ("After scan", obs);

	if (obs->dlevel > 1) check_cfcount (ob, "Main:2");

/* Report total removals if they exceede count... */
	if (obs->dlevel > 0 || Nfaroff > MESCOUNT)
	    printf (" %4d objects removed as FAR off detector.\n", Nfaroff);
	if (obs->dlevel > 0 || Noutdet > MESCOUNT)
	    printf (" %4d objects flagged outside detector.\n", Noutdet);
	if (obs->dlevel > 0 || Nduppos > MESCOUNT)
	    printf (" %4d objects flagged as duplicates.\n", Nduppos);

/* Report total nodes and active nodes */
	if (obs->dlevel > 0) {
	    n = listcount (obs->ob);
	    k = active_objs (obs->ob);
	    printf (" Of %d total objects, %d are active, %d inactive.\n",
		n, k, n-k);
// debug print of gap count here...
//#ifdef  GAPVO
//	    Gap_count (2, obs->ob);		// debug only
//#endif
	}
	if (istest) gsrx &= 5;

/* ** Call subroutines to automatically deactivate conflicts */

/* First, remove all single conflict objects. */
/* Removing single conflicts is done to speed up the conflict resolution
  algorithm in deconflict2 below.  The same result should be achieved
  either way, providing a possible check on deconflict operation.  */
//#define  SKIPSINGLE
#undef   SKIPSINGLE
#ifndef  SKIPSINGLE
	n = drop_single_confs (obs);
//	if (obs->dlevel) printf (" Removed %d single conflicts.\n", n);
	if (obs->dlevel > 0) {
	    printf (" Removed %d single conflict objects,", n);
	    printf (" leaving %d active objects.\n",
		active_objs(obs->ob));
	}
#endif

	dbroq ("Before deconflict", obs);

	if (obs->dlevel > 1) {
	    printf (" Removing conflicts...\n");
	    fflush (stdout);
	}

/* Run the de-conflict code; either old or new version... */
/* deconflict will sort object list by priority during operation */
	n = deconflict (obs);  // Was deconflict2
	if (obs->dlevel > 0) printf (" Removed %d conflicting objects.\n", n);
/* deconflict leaves object list sorted into original order */

/* Report the total number of objects, or mask features */
	if (obs->dlevel >= 0) repobj (obs->ob);
// DO THE ABOVE BEFORE SEPARATING OBJECT QUEUES TO ACTIVE/INACTIVE

/* NOTE -- MAY WANT TO SEPARATELY REPORT SLITS AND ALIGNMENT HOLES? */
	dbroq ("After deconflict", obs);

/* ** Report any remaining conflicting nodes here... */
	if (obs->dlevel >= 0) {
	    n = repconf (obs);
	    if (obs->dlevel > 0 || n > 0) printf ("  %d remaining conflicts.\n", n);
	}

/* ** Check number of reference objects, report if that is small...  */
	obs->actref = RefCount(obs->ob);
	k = 0;
	if (obs->actref == 0) {
	    if (obs->reflimit > 0) {
		printf (" \n");
		printf (" ** * * * * * * * *  ERROR  * * * * * * * * **\n");
		printf (" ** NO alignment objects exist on the mask! **\n");
		printf (" **    This mask may not be observable!     **\n");
		printf (" ** * * * * * * * * * * * * * * * * * * * * **\n");
		k++;
	    } else {
		printf (" ** Special Mask with NO ALIGNMENT (Position Independent)\n");
	    }
	} else if (obs->actref < 4) {
	    if (obs->reflimit > 0) {
		printf (" \n");
		printf (" ** * * * * * * * *  ERROR  * * * * * * * * **\n");
		printf (" **  %d alignment objects exist on the mask! **\n",
			obs->actref);
		printf (" **    This mask may not be observable!     **\n");
		printf (" ** * * * * * * * * * * * * * * * * * * * * **\n");
		k++;
	    } else {
//		printf (" ** Special Mask with NO ALIGNMENT (Position Independent)\n");
		printf (" ** Found %d of %d ALIGNMENT objects.  Strange...\n",
			obs->actref, obs->reflimit);
		k++;
	    }
	}

// If we have errors, ask to continue at this point...
	if (k > 0) {
	    printf (" \n");
	    askme (answer, 15, " ** Do you wish to continue [no|yes] ? ");
	    if (strncasecmp(answer,"yes",3)) goto main_loop_end;
	}

/* ** Write a file of object queue, with use counts... */
	write_obf (obs, 0);

//#ifdef  GAPVO
//	Gap_count (4, obs->ob);		// debug only
//#endif

/* Call special subroutine to remove inactive objects from the standard
  object queue and place them in a separate queue. */
	rob = inact_obj (obs);

/* HERE WE PUT ANY ROUTINE WHICH WRITES THE INACTIVE OBJECTS
  FROM QUEUE "rob" TO A FILE, ETC.  */
// currently, queue "rob" is unused...

/* Special feature -- write a file like .obf (the .obx file) containing
  only the active objects, otherwise similar to the .obw file.  */
	if (actf) write_obf (obs, 1);

	dbroq ("Mask queue", obs);

//#ifdef  GAPVO
//	Gap_count (5, obs->ob);		// debug only
//#endif

/* Call optional subroutine to edit the object list with image data
  files, and interactive editing. */
// NOTE - this will be part of the "masktool" edition, and
//  it is not done here.

/* Here, we have only active objects in the queue.  Do any operations
  on the active set now.  */

/* Call slit extension algorithm.  The subroutine checks to see
  if the algorithm itself is enabled.  */

	extend_slits (obs);
/* Extend slits sorts objects into x order */

/* OBTAIN AND SPLICE IN A SET OF WING-CHIP SUPPORT HOLES; NO HOLES
  IF NOD/SHUFFLE OR IF INHIBITED BY A SPECIAL OBS FLAG... */
/* Default is now wings = 0, disabled.  The -w flag can be used
  to obtain wing chip holes if desired.    */
	if (wings)  add_wings (obs);

/* Re-order the object queue (may be optional) */
//	if (reord) obs->ob = reorder (obs->ob, 8, 2);
//	n = (kobjs(obs->ob) + 5) / 2;
	n = (listcount(obs->ob) + 5) / 2;
	if (reord && quik == 0) obs->ob = reorder (obs->ob, n, 2);

/* NOTE THAT THE RE-ORDER OPERATES ON USED AND UNUSED OBJECTS; NEED TO
  SPLIT THE QUEUE INTO ACTIVE AND INACTIVE QUEUES, AND OPERATE ON THE
  ACTIVE QUEUE ONLY!! */  /* See ~22 or 34 lines above */

/* ** Call subroutine to extract a SMDF descriptor structure from the
  existing image file. */
/* --this will look at those holes with active status and put them in
  a separate queue? */
// OR - we just scan the master object queue and write those
// into the smdf giving their hole information...

//#ifdef  GAPVO
//	Gap_count (8, obs->ob);		// debug only
//#endif

#ifdef  NSROT
/* ** De-rotate by bias amount, and correct slit orientations */
	rotobj (obs->ob, -(obs->nrbias), 1);
#endif

	if (quik == 0 && gsrx == 0) write_smdf (obs);
/* Note that the file name is in the obs structure... */
	if (gsrx) printf (" ** No Mask Definition file written! **\n");

/* ** Call subroutine to write the SMDF file from the descriptor structure */
/*  SMDF file will need:
	-- All the optical data to reconstruct the spectra for all the
	objects.  Names of the optic elements may suffice with the
	optic data processor.
	-- The global data in the Obs structure - will need several lines
	to put all that out in a fixed form.
	NOTE that some of these names needed in Obs will also satisfy the
	optical data requirements above.  This part of the file will look
	a lot like the observation file format.
	-- Mask positions in detail for all active objects
..*/

//#ifdef  GAPVO
//	Gap_count (9, obs->ob);		// debug only
//#endif

/* Debug section -- write some things which may help debug of
  the debug window execution.. */

//	i = strncmp (myname, "tmaskgen", 8) ? MASK : DETECTOR;
	i = istest ? DETECTOR : MASK;
/* DEBUG plot in separate interactive window -- */
//	if (obs->dlevel) pwin (obs, i);
//	if (dplot) pwin (obs, i);	/* Standard feature now */
	if (dplot
#ifdef  GISMOAP
	    && gisap == 0
#endif
	) pwin (obs, i);	/* Standard feature now */

/* Optional extra -- Do subroutine to spawn a system command which will
  plot the SMDF data as a PostScrpt file. */

/* Optional extra -- Subroutine to write an object input file consisting
  of the objects in the list not used in the SMDF output file. */

/* ** - Indicates part of minimum required set */

main_loop_end:

    free (filename);
    filename = NULL;
/* NOTE -- Above, in main loop we should put all arguments into
  a namlist structure.  Then, construct a loop to pop them out
  one at a time, and process them.  After processing, call
  kill_obs to kill the obs structure (here) and end the loop.
  Keep the exlist (extension list) around; make it before the
  loop, and remove it after the loop ends here. */

/* Done with obs storage, delete it here. */
	obs = kill_obs (obs);

/* Drop the removed object queue also */
	rob = kill_objq (rob);

    }	// End of loop on input files < while (head != NULL) >

// printf (" -- Ended loop on input.\n"); fflush(stdout); // debug

    kr = 0;
main_finish:

// printf (" --main-finish here.\n");	fflush (stdout);	// debug

/* Also, free any optic elements still stored... */
#ifndef  OPTLISTDEBUG
//    while (gs != NULL) gs = free_element (gs);
    gs = kill_elementq (gs);
/* This is the simple method, and should work; if it fails, we start with
  the debug section which follows... */
#else
//-- from here, remove debug stuff (see below)...
{	// debug test of what's happening
int	kf=0;
int	ns;

    if (gs != NULL) {
	ns = listcount (gs);
	if (wlevel>1)  printf (" -- Total of %d optical elements.\n", ns);
	if (gs->last == NULL) printf (" gs->last is wrongly NULL!\n");
	fflush (stdout);
    }

    while (gs != NULL) {
	gs = free_element (gs);
	kf++;
    }	// Loop to free elements in gs element queue.
    if (gs != NULL) printf (" -- Element queue not empty.\n");
    else if (wlevel>2) printf (" Free %d optical elements.\n", ns); // lint
}  // end debug section
#endif  // on OPTLISTDEBUG

/* Also, free the cut data stuff, too.  */
#ifdef  USE_CDAT
    CDhead = kill_cdat (CDhead);
#endif
#ifdef  USE_GSTO
    while (CDhead != NULL) CDhead = kill_gsto (CDhead);
#endif

#ifdef  USE_BSTO
    CDhead = btree_kill (CDhead, kill_kvm);
#endif

    cp = kill_pvfs (cp);
    fp = kill_pvfs (fp);

    return kr;

}

/* Revision History

11/03-17 -- Version 2.27
	-- Change fclose to iclose for file name support.
	-- Added SHFTCLK capability to plot (11/10-27)
	-- Plot window re-positioned to mid-right (11/11-14)

10/11-12 -- Version 2.26
	-- Change wing chip support default to 0, no wing chip support
	as a default.  See "wings".  The -w flag will be used to add 
	the holes if needed.
	-- Add test on "Horizon to Horizon" rotator conflict.
	--> Becomes minimum version upon release for above changes.

10/06-25 -- Version 2.25
	-- Add support for position independent masks by not testing for
	guide stars if no reference objects are being allowed.
	-- Removed unused symbol "debug".

09/01-27 -- Version 2.24
	-- Adding code for GISMO aperature mask feature
	--> Minimum version when GISMO is released
	-- Adding debug for gismoap feature

09/01-07 -- Version 2.23
	-- The -g flag is only available in test mode

08/12-16 -- Version 2.22
	-- Fix gaps, omit testing virtual object w.r.t. mask space
	-- Add automatic GISMO slit rotation feature
	--> Minimum version release 2.07 for bug fix

08/10-16 -- Version 2.21
	-- Add code for "getopt" utility scan of command line

08/07-28 -- Version 2.20
	-- Add output of active object only file (.obx file) on the
	-f flag.
	--> Minimum version for release 2.06

08/07-18 -- Version 2.19
	-- Add conditional for IMACS to error message on uncut length 
	without N&S mode selected as N&S mode only applies to IMACS and
	actual N&S operation is permitted on LDSS without rotation.

08/03-11 -- Version 2.18
	-- Change format of output to smdf file for MADE line by changing
	the PROGNAME and VERSION items
	-- Change back format of MADE line, and put Written By line before
	the MADE line for better source detection.
	-- Add Program (Vers) structure
	-- Reject operation of obsolete software.
	-- Reject processing of obsolete data.
	-- Add checks of N&S, uncut length for consistency
	--> Will be minimum version at release 2.04

08/01-14 -- Version 2.17
	-- Add message giving total object count.
	-- Add message in mgutils for gismo giving number of objects
	removed for not being in a subfield.

07/10-17 -- Version 2.16
	-- Add version support call during startup, put version
	information into mgfeats
	-- Minimum Version for release 2.00 and version control.

07/05-03 -- Version 2.15
	-- Fix filename logic and drop filename at end of loop
	-- Able to allow no guidestars if quick look feature

07/04-18 -- Version 2.14
	-- Implement guide star requirement
	-- Add output of "Test Version" to tmaskgen execution
	-- Fix bug in storing argument text in main.

07/01-17 -- Version 2.13
	-- Fix support for Nod/Shuffle rotation flagged by NSROT

06/12-13 -- Version 2.12
	-- Add support for Nod/Shuffle rotation

06/10-26 -- Version 2.11
	-- Adding more GISMO support, in dbplot.c

06/09-27 -- Version 2.10
	-- Adding GISMO support, and test version of opticdef

06/06-13 -- Version 2.09
	-- Added support for cdata table, with gsto or b-tree reading

06/05-26 -- Version 2.08
	-- Add support for b-tree storage of maskcut.cfg data

06/04-25 -- Version 2.07
	-- Added code to catch no instrument, no disperser, and file name
	field not matching file name opened.
	-- The file name test is actually optional.
	-- Modified dbplot to put smaller font on feature labels.

06/04-13 -- Version 2.06
	-- Implement plotting of object names on OBJNAM flag

06/04-06 -- Version 2.05
	-- Add tests for wavelength ranges out of bounds
	-- Add test for global spectrum range too large

06/02-23 -- Version 2.04
	-- Adding OBJCHECK feature to check object data

06/01-11 -- Version 2.03
	-- Add message if file not found.

05/12-23 -- Version 2.02
	-- Call subroutine to set ldss temperature default.

05/12-07 -- Version 2.01
	-- Remove some of the debug code on dropping the optic queue
	on gs.  Also allow debug of this.
	-- Default debug level is now 0.
	-- Removed unused extension list items.

05/12-06 -- Version 2.00
	-- Replace dodebug global with dlevel item in the Obs
	structure, with better debug level discrimination.
	-- For local variable, replace dodebug with wlevel here.

05/11-07 -- Version 1.98
	-- Call to repconf to report conflicts.
	-- Adding avoidance tests to mask_space
	-- Support of dynamic maskcut.cfg entry

05/10-06 -- Version 1.97
	-- Optutils supporting extended grism data type
	-- Also, comments on includes changed
	-- Add -a feature to remove alignments conflicting with the
	detector bounds/gaps.

05/09-08 -- Version 1.97
	-- Add -s feature to skip the debug plot of mask/field

05/06-28 -- Version 1.96
	-- Implement virtual gap feature, if feature flag is on.

05/06-14 -- Version 1.95
	-- Removal of redundant items in Obs -- telescope, instrument,
	grating, etc.  Use optical element pointers instead.
	-- Will start tmaskgen plot in detector space, otherwise mask

05/06-03 -- Version 1.94
	-- Support for -i minimal debug output flag
	-- Support for -q no slit extension flag as quick look

05/05-12 -- Version 1.93
	-- Support for ndplot instead of dbplot for testing
	-- deconflict2 changed to deconflict (5-23)

05/04-18 -- Version 1.92
	-- Error message for incompatable telescope.

05/02-02 -- Version 1.91
	-- Start of support for new LDSS

04/11-02 -- Version 1.90
	-- Support of version 1.06 mgutils -- 
	-- Removed commented code, revised loop increment logic

04/11-01 -- Version 1.90
	-- Support of version 1.05 mgutils -- use kobjs and active_objects
	instead of a local loop to count objecs in queue.
	-- Later changed to listcount general routine.

04/10-29 -- Version 1.90
	-- Support of version 1.04 mgutils -- finally put fill_objects
	call in to replace a string of inline code to read all the
	objects into the obs linked structures.
	-- Remove the fill_objects old code and some other commented
	code which is no longer used; also some variables not used.
	-- Remove all reference to check_elemqueue
	-- Change messages; add reports of counts of removed objects and
	status of scanning and removing conflicts.
	-- Only report scanned conflicts up to a limit number MESCOUNT,
	and report number found/removed when done.
	-- Improving reports of object counts and progress

04/10-27 -- Version 1.90
	-- Support of version 1.02 mgutils -- remove call to predeconflict
	as that work is done in drop_single_confs now.  Added comments,
	and change wording of a message.

04/10-14 -- Version 1.90
	-- Change version numbering, adding mgutils.c version and
	appending that to the version number here.  Version reported will be
	the concatenation of this and mgutils.
	-- Changes in utilities will show up in version reports without
	having to edit a new version of the maskgen.c source file.  When
	a new major version of utilities is made, the version number here
	must be updated however.  Next version should be 2.01 -- with the
	end digits coresponding to the utility version.
	-- Add "-w" to help information.
	-- Slit extension support both in .obs file reading and in
	maskgen exists with version 1.90.  Full support with the intgui
	button being visible is waiting for full test results.  The slit
	extension algorithm was debugged in mgutils on 9-23.

04/09-16 -- Version 1.89
	-- Call to slit extension subroutine added

04/09-14 -- Version 1.88
	-- Add bounding box subroutine support
	-- Change transform to a true matrix
	-- A fix to wing chip hole names was added.

04/08-26 -- Version 1.88
	-- Make wing chip support the default, with -w to omit.
	-- Version value was really 1.87 on this edition.

04/08-12 -- Version 1.87
	-- Avoid writing NULL title in SMDF file.
	-- Adding support for wing chips also (not default).

04/06-03 -- Version 1.86
	-- Reject with message any MagNoADC telescope files.

04/05-27 -- Version 1.85
	-- Change default from MagNoADC to Magellan (in mgutils.c).
	-- Also, change default to ex_order = 0 in mgutils.c

04/05-13 -- Version 1.84
	-- Fix bug in conflict resolution for extra orders and high
	pdecide and large spectral range.  See confprob.txt.

04/04-28 -- Version 1.82
	-- Handle default telescope name correctly (to override MagNoADC)

04/03-31 -- Version 1.80
	-- Add write of non-default slit parameters to .obf file (mgutils)

04/03-09 -- Version 1.75
	-- Add timestamp to SMF file comments
	-- Enforce no embeded blanks in file name and observer name.

04/02-18 -- Version 1.70
	-- Fix in-line comments in an object file

04/01-15 -- Version 1.60
	-- Add PROGNAME as a standard thing.

04/01-08 -- Version 1.50
	-- Add DLIMIT for wavelengths on detector limit separately.

03/12-10 -- Version 1.40
	-- Change conflict queue to double link secondary poiners
	-- Remove all reference to conflict queue header; find the
	conflict queue if needed from object queue.
	-- Add order field to object queue and sort on it after
	conflict resolution (deconflict2 only).
	-- Has debug section for element storage which faults while
	being freed and may not be circular.

03/12-08 -- Version 1.30
	-- Generalized object queue sort; pass decision function.

03/12-04 -- Version 1.25
	-- Testing new parameterized conflict resolution

03/11-05 -- Version 1.20
	-- Using get_optic_data subroutine from mgutils; now program
	will STOP if the optic data is not found.

03/10-22 -- Version 1.15
	-- Starting to put in the avoidance computations.

03/10-21 -- Version 1.12
	-- Check of sorting
	-- Fixed sort algorithm, and drop objval data for nodes which
	are removed early in the process and for way wrong positions.

03/10-17 -- Version 1.10
	-- Added timing report on conflict resolution
	-- Added predeconflict to speed up total resolution

03/10-14 -- Version 1.05
	-- Changed object storage, can have null spectra

03/10-13 -- Version 1.00
	-- Fixed conflict-while-active problems; finds all objects
	in the catalog now.

03/10-11 -- Version 0.96
	-- Adds debug for object queue problems

03/10-02 -- Version 0.95
	-- Includes correct implementation of "overlap" on detector
	-- Adds "uncut" values for global slit shortening, in
	support of nod/shuffle capabilities.
	-- Fixed some bugs in the spectrum overlap computations

03/09-10 -- Version 0.90
	-- Write object file with incremented use counts

03/09-05 -- Version 0.85
	-- Fixed poor format in SMDF file writing
	-- Added some limits on conflict reporting

03/07-28 -- Version 0.80
	-- Look at avoidance areas on mask to reject some slits
	(May not be all complete)

03/07-23 -- Version 0.75
	-- Remove inactive objects to a separate object queue.

03/07-21 -- Version 0.70
	-- Introduced new optical transformations to detector positions
	-- Default slit sizes are now kept in arcseconds
	-- Removed extra field box in mask debug plot
	-- Hid orange spectrum boundries behind a debug flag
	-- Added black detector boundry in detector debug plot
	-- Better mask boundry check

03/07-14 -- Version 0.60
	-- Make conflicts with extra orders optional
	-- Add line giving slits and alignment features totals

03/06-29 -- Version 0.55
	-- Many changes have been made to utility routines to
	accomodate new optics and obs file format
	-- Add output of number of objects

03/02-11 -- Version 0.50
	-- Integrating new optical subroutines; temporarily including
	both as a debug feature...

02/12-17 -- Version 0.40
	-- Increased sorting distance in reorder call; also based
	on number of objects to be ordered.

02/11-08 -- Version 0.35
	-- Debugging crash
	-- Changing iterative to algebraic computation of grating angle

02/08-23 -- Version 0.32
	-- Advanced ordering of output, and make it optional

02/08-22 -- Version 0.30
	-- Introduce ordering of output by distance minimizing

02/04-30 -- Version 0.25
	-- Change position computation of objects to matrix

02/03-13 -- Version 0.22
	-- Fixed bad free of filename;
	-- Changed debug level logic
	-- Fixes to utility to produce test grid of (reference) holes

02/03-13 -- Version 0.21
	-- Put debug routines into msdebug.c file

02/03-08 -- Version 0.20
	-- Add loop to process multiple input files
	-- Kill obs storage at end of processing
	-- Remove some redundant debug statements
	(earlier)
	-- Do not remove alignment hole if spectrum off detector
	-- Add spectrum overlap in pixels by augmenting sod edges

02/02-12 -- Version 0.16
	-- Added LDSS2 optics descriptions to mgutils
	-- Do not remove nodes with priority < 0.0
	-- Do not remove duplicates, off array with priority < -2.0
	-- Add dump of mode for grating/grism debug
	-- Add computation of grism center wavelength

01/09-17 -- Version 0.11
	-- Added current epoch programs, can get date
	-- Put epoch and date created into SMDF output
	-- (9-18) Removed extra push/pop for object queue
	-- Removed general.c include

01/09-14 -- Version 0.10
	-- Has graphic window debug
	-- Cleaned up data structures and calls
	-- Version number appears in mgutils.h

01/09-05 -- Version 0.09
	-- Added SMDF output

01/09-04 -- Version 0.08
	-- Improved debug dumps

01/08-30 -- Version 0.07
	-- Read object files to objq structure

01/08-28 -- Version 0.06
	-- Uppercases keywords in observation file
	-- Better use of dodebug global

01/08-15 -- Version 0.05
	-- Takes data from observation data file
	-- Uses new utility subroutines in library
	-- Starting debug of basic version of program

01/04-09 -- Version 0.04
	-- Put sexigisimal decoding in utils.

01/04-03 -- Version 0.03
	-- Moved subroutines to mgutils.c and to library

01/04-02 -- Version 0.02
	-- Testing input of object file to structure.
	-- Debug output of generated structure.
	-- Added recursive sexigisimal decoding code.

01/03-29 -- Version 0.01
	-- Testing how to add subroutines from external file,
	open the object file and read it.

01/03-07 -- Version 0.00
	-- Started by copying from fix.c for command line arguments
	and general structure.
..*/
