/*  mgfeats.h == Mask Generation Features Definitions */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2011 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **  
**                                                                   **
**      This software is proprietary, and may be used or copied      **  
**          only with the written consent of the author.             **  
**              All copies must retain this notice.                  **  
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*  Defines the program features being supported.  Coordinates these
  features between all the source files which have conditional
  compilation based on feature set.  */

#ifndef  INCLUDE_MGFEATS_H
#define  INCLUDE_MGFEATS_H

/* This is now included by maskdat.h */
/* By convention, the first def/undef line is the "normal" state. */

	/* - - - - - - - - - - - - - - *\
	|				|
	|	 Global Values		|
	|				|
	\* - - - - - - - - - - - - - - */

/*  Definitions for system-wide use go here  */

#define  PathName  "MGPATH"

	/* - - - - - - - - - - - - - - *\
	|				|
	|	 Version Control	|
	|				|
	\* - - - - - - - - - - - - - - */

/* Locations for version control use */
// #define  URLpref  "ftp://ftp.ociw.edu/pub/clardy/"
// That definition is now placed down below for a while...
#define  VERSFIL  "mgvers.dat"
// was #define  DEFILE  "versions.cfg"
#define  LOCALpfx  ".loc_"
/* LOCALpfx is the prefix to apply to a file name to create a local
  copy of the file, probably an FTP file such as VERSFIL.  */
#define  LOCALage  30.0
/* LOCALage is the acceptable age of the local file, in hours.  */
// We start age limit at 1024 seconds (17 minutes)
// It may be increased to 1/2 hour or 1 hour
// 8192 Seconds is 2.28 hours; 16384 is 4.55 hr.
// 131072 is 36.4 hours
// Change to floating hours 2010/10-29
/* The compiled value, LOCALage, is used to test local file age, not the
  value in the local file or the FTP file.  */
/* The above items are also used by the general local cache method outlined
  in localftp.otl and used to store the latest public key file which is
  used via FTP a lot.  This local caching is put on a symbol to allow it
  to be dropped if security should require it.  */

/* Version numbers for all components and utilities are placed here,
  and may be accessed from anything which includes this module.
  The title string is xxxxxxT by convention.
  Convention is that xxxxxxV is the current version, incremented
  on any functional change; with the integer part being incremented
  only on release of a substantial new functionality or organization.
  The global version is incremented for any release.
  The required version is xxxxxxR and is the minimum acceptable value
  for proper use of each component.  Execution of older versions is
  rejected, and input from older versions is warned about.   */

/*  ADD ANY NEW VERSION INFORMATION NEEDED HERE AND ALSO UPDATE
  THE REFERENCES IN THE makevfs PROGRAM.  (IN mgutils.c)  */

/* To initialize a text version string using the bare versions below, a
  two stage stringification process is needed.  For details, see
  http://gcc.gnu.org/onlinedocs/gcc-4.1.1/cpp/Stringification.html
  We define two macros here to help...
  To use them, do e.g.  char  VERSION[] = Xstr (MGUTILV); */
#define  Xstr(s) Str(s)
#define  Str(s)  #s

/* Floating point version numbers have two decimals after the point
  by convention, but may have up to 4 if needed.  They are compared
  as numeric values.  */

/* Global version numbering (increments on each release) */
#define  VERSECT  "Release"
#define  VERSECV  2.11
#define  VERSECR  2.10

/* Here are the old names; all references to these should have been
  removed starting with version 2.05 in October, 2008.    */
//#define  VERXECT  "Version"
//#define  VERXECV  2.04
//#define  VERXECR  2.03
/* Remove above here and in mgutils.c */

/* Interface GUI component */
#define  INTGUIT  "IntGui"
#define  INTGUIV  4.68
#define  INTGUIR  4.65
// 4.58 minimum

#define  SKYWINT  "SkyWin"
#define  SKYWINV  1.30
#define  SKYWINR  1.28

/* Mask Generation component */
#define  MASKGNT  "MaskGen"
#define  MASKGNV  2.27
#define  MASKGNR  2.26

#define  MGUTILT  "MGutil"
#define  MGUTILV  1.92
#define  MGUTILR  1.91
// 1.81 minimum

/* Mask Cutting component */
#define  MASKCTT  "MaskCut"
#define  MASKCTV  3.36
#define  MASKCTR  3.35

#define  CUTUTLT  "CutUtil"
#define  CUTUTLV  1.15
#define  CUTUTLR  1.15

/*  NC plotting utility */
#define  NCPLOTT  "NCplot"
#define  NCPLOTV  1.24
#define  NCPLOTR  1.22

/* Version support utility */
#define  VERSUPT  "VerSup"
#define  VERSUPV  0.83
#define  VERSUPR  0.82

/* SMF plot utility */
#define  SMFPLTT  "SMDFplt"
#define  SMFPLTV  0.83
#define  SMFPLTR  0.82

/* SMDF postscript plot */
#define  SMDFPST  "SMDFps"
#define  SMDFPSV  1.07
#define  SMDFPSR  1.06

/* Check Sum feature */
#define  VCONCVT  "MGunit"
#define  VCONCVV  1.81
#define  VCONCVR  1.19
/* Set VCONCVR to 1.19 on next release.  Was 1.17 */
/* When VCONCVR was 1.18, checksum failed on Mac-OS-X machines due
  to incompatable random generation from given seed.  Don't use 1.18 */

/* Star Catalog Server */
#define  CATXMLT  "CATxml"
#define  CATXMLV  0.50
#define  CATXMLR  0.30


/* -- Required Data Files -- */

/* Optic Definitions - opticdef.dat */
#define  OPTDEFT  "OpticD"
#define  OPTDEFV  1.50
#define  OPTDEFR  1.50

/* Cutting Definitions - maskcut.cfg */
#define  CUTDEFT  "CutDef"
#define  CUTDEFV  1.50
#define  CUTDEFR  1.50

/* Pmap data - pmap.dat */
#define  PMAPDAT  "Pmap"
#define  PMAPDAV  1.50
#define  PMAPDAR  1.50


	/* - - - - - - - - - - - - - - *\
	|				|
	|      Server Coordination	|
	|				|
	\* - - - - - - - - - - - - - - */

/* Definitions for client-server coordination, placed here so that
  the same definitions are used in compiling both client tasks and
  the server task also.
  Remember -- It is bad luck to change these without also compiling
  a new server so that both are coordinated.
  (And by "bad luck" we mean that things will not work right)    */

/*  --  For Star Catalog xml-rpc server  --  */

#define  METHOD   "PullStars"
#define  AMETHOD  "GetStars"
// PullStars gets the structure, GetStars gets text information
// (Server currently supports both methods)
#define  PortNum  0xd83b
// Above is 55355
// Use port 7080 for testing with skymap (used by hack server)
//#define  PortNum  7084

#define  TRIM_C
//#undef  TRIM_C
/*  Symbol  TRIM_C  is defined to implement circular trimming of
  sky regions about the center; opposed to the "old way" which was
  a generally rectangular trim in a,d separately.
  Should normally be defined.    */

#undef  NETGETV
//#define  NETGETV
/*  Symbol  NETGETV  is defined to get the "v" vector over the network,
  and undefined to omit that, creating v locally from alpha and delta.
  It should normally be undefined.    */

#define  MLIM  19.5
/*  Symbol  MLIM  is the magnitude limit used in both the server
  (where it was previously defined at 19.5) and in skymap.c for disk
  catalog lookup (which had no limit in v2.09 and v2.10).  It could
  be set to a large value such as 35.0 if all stars in the catalog
  were desired, or restricted to make catalog tables smaller.  It was
  used in the server to reduce the size of network data transmission,
  but the reduction is probably under 50% in most cases.  This value
  will be used by both server and local programs so that the catalog
  tables are the same for both methods in the future.    */


	/* - - - - - - - - - - - - - - *\
	|				|
	|     Application  Features	|
	|				|
	\* - - - - - - - - - - - - - - */

/*  WHEN ANY NEW FEATURE IS ADDED HERE, IT SHOULD ALSO BE ADDED TO
  THE makevfs PROGRAM TO TRACK VERSION USE.  (IN mgutils.c)
  By convention, features are added to the top of this list.    */

/*  It is recommended to use "Xfeat" to add all definitions which
  are present here to the version file.  Values are written too, if
  they are part of the definition and not "0".    */

/* By convention, the first def/undef line is the "normal" state. */

#define  SHFTCLK
//#undef  SHFTCLK
/*  Symbol  SHFTCLK  is defined to allow all interactive plotting programs
  to map a mouse button press with the "shift" key held down to a click of
  the center mouse button.  This will allow center click operations to be
  accessible for users whose physical mouse or trackpad cannot generate
  a center click input.  (This is common on MacBook trackpads).  */

#define  DYNPARSE
//#undef  DYNPARSE
/*  Symbol  DYNPARSE  is defined to use dynamic parsing in place of old
  parsing with supplied buffer.  The is more robust in case a bad line
  has an unusually long token which could overflow any defined buffer
  and cause a problem.  Used in multiple places.
  Still considered highly experimental until thoroughly tested.     */

#define  LOCATE
//#undef  LOCATE
/*  Symbol  LOCATE  is defined to add and use the feature of the
  "locfile" program replacing the older "findfile" program.
  See locate.otl for explanation of the improvement.      */
/*  When working, move locfile and support programs from mgutils.c
  to ioutils.c, and unifdef this symbol to be always true.    */

#define  CONSTEL
//#undef  CONSTEL
/*  Symbol  CONSTEL  is defined to add the feature of showing constellation
  names for the position in intgui; putting the constellation index into
  the observation structure; put name in smdf file as a comment.  */

//sunifdef < #define  NEWNAME
// Removed 2010/08-05  Release 2.09+
//#undef  NEWNAME
/*  Symbol  NEWNAME  is defined to implement the change from the old
  ociw.edu to the new obs.carnegiescience.edu domain name.  This happens
  in several places where FTP or other web access to our server is
  needed.  Will be required in 2010 about June 8.    */

/* Locations for version control use */
// Temporarily relocated here to use the above symbol definition.
//sunifdef < #ifdef  NEWNAME
#define  URLpref  "ftp://ftp.obs.carnegiescience.edu/pub/clardy/"
//sunifdef < #else
//sunifdef < #define  URLpref  "ftp://ftp.ociw.edu/pub/clardy/"
//sunifdef < #endif


#define  CACHECAT
//#undef  CACHECAT
/*  Feature  CACHECAT  is defined to implement the automatic caching
  of catalog data, and using it if possible.  When defined, the
  ALLMSTAR  feature below is required.  This should be defined when
  the  EXCAT  feature below is active, and a server is online.     */

//sunifdef < #define  CTLGKEY
// Removed 2010/08-05  Release 2.09+
//#undef  CTLGKEY
/*  Symbol  CTLGKEY  is defined to implement the Catalog Keyword feature,
  to read all non-required object catalog information using keyword=value
  syntax rather than positional.  Will also be able to read legacy files
  with positional parameters.  Also writes .obw/.obx files with the
  keyword syntax for non-default values.
  This feature is required after release 2.08    */

//sunifdef < #undef  OCKWX
// Removed 2010/08-05  Release 2.09+
//#define  OCKWX
/*  Symbol  OCKWX  was defined to allow counting use of keywords in reading
  of object data.  This should NOT be done, so keeping this symbol undefined
  is a good thing.  Later, delete all these references.   */

//#undef  GGOLDWAY
/*  Symbol  GGOLDWAY  is un-defined to implement guider testing
  in mgutils -- normalize_obs routine using a clean subroutine
  rather than the "oldway" of in-line code.  Should not be defined.
  It is present here for debug use only, not a feature.
  (Actually, does not appear anywhere else but here)...      */

//sunifdef < #define  ALLMSTAR
// Removed 2010/08-05  Release 2.09+
//#undef  ALLMSTAR
/* Symbol  ALLMSTAR  is defined to implement the trial fix to skymap
  and skywin in which all stars regardless of magnitude are selected
  for the list, and after sorting by magnitude an upper limit on star
  numbers is applied.  This is to reduce calls to the full read of 
  the star catalog data from disk or network.  */
/* Due to the liklihood of server use, the ALLMSTAR feature should
  always be defined, and will be unifdef treated to be defined.  */
#define  MAGLIM1  17.5
#define  MAGLIM2  18.5
#define  STARLIM  3072
//#define  STARLIM  2048
#define  STAROPT  1536
//#define  STAROPT  1280
//#define  STARLIM  1536
//#define  STAROPT  1024
/* Constants MAGLIM1, MAGLIM2 and STARLIM are set to coordinate the limits
  on the sky window for skymap.c and skywin.c.  We want at least MAGLIM1
  but need to limit count to STARLIM.  Things are best at STAROPT star
  count if we have deep enough magnitudes.  The number of stars plotted is
  limited, based on the magnitude at that limit.  Find ns(1) and ns(2) as
  number of stars up to MAGLIM1 and MAGLIM2.  Then ns(1) < ns(2) and
  STAROPT < STARLIM.  If ns(2) < STAROPT, use STAROPT.  If ns(2) < STARLIM
  then use ns(2).  If ns(1) < STARLIM use STARLIM.  Then, if ns(1) > STAROPT
  use max(STARLIM, 2*STAROPT).  Limit number of stars to whichever
  limit we are using.
  Set MAGLIM values based on needed guide star brightness, and set the
  count limits based on how fast the usual workstation plots a map.  */

#ifdef  CACHECAT
//sunifdef < #define  ALLMSTAR
// Removed 2010/08-05  Release 2.09+
/*  The  ALLMSTAR  feature is required by CACHECAT feature.  */
#endif

#define  EXTPRI
//#undef  EXTPRI
/* Symbol  EXTPRI  is defined to implement the extended priority window
  in intgui which supports the "EXPRI" keyword in the .obs file to use
  an extended priority for objects which increments the priority by
  use count X coefficient to allow used objects to fill-in on subsequent
  masks if possible.  */

/* See the symbol  ISOREFLIM  defined in mgutils.c for the alternate method
  of removing reference objects past reflimit.  */

#undef  NOCATS
//#define  NOCATS
/* Symbol  NOCATS  is defined only for debug purposes.  When defined,
  the catpath program only looks at environment variable CATPATH for
  a defnintion of the catalog location.  All the usual defaults, and
  parent directory things are bypassed.  This is used to test the
  internet service to verify that catalog information does not come
  from the local disk.  Always undefined in production versions.  */

#undef  HACKSERVER
//#define  HACKSERVER
/*  Symbol  HACKSERVER  is defined to use the "hack" server on fujiwara
  written in python, and undefined for normal use with the real server.
  Normally should be undefined.    Test/Debug only.    */

#define  EXCAT
//#undef  EXCAT
/* Symbol  EXCAT  is defined to implement the "hack" to skymap which
  does catalog look-up externally, i.e. from the web-based interface.
  A similar feature will be implemented to actually add the ability
  to get star catalog data either locally or over the internet.
  This should only appear in a production version when an actual
  on-line server exists to answer requests.    */

#define  IMAGM
//#undef  IMAGM
/*  Symbol  IMAGM  is defined to implement the image mode feature in
  the intgui program.  When activated by a command line flag, this mode
  changes the main window in several ways, and allows users to use the
  sky window to select guide stars, and make an observing log entry;
  use the sun/moon planning feature, but not to make a .obs file or
  to plan a spectroscopic observation.   */
/*  This symbol depends on GSDATA being defined below; if it is not,
  then IMAGM will be undef below.  This is needed to support the basic
  reason for the imaging mode, creation of an obscat entry, which
  requires GSDATA to be properly written.   */

#define  FILESEC
//#undef  FILESEC
/*  Symbol  FILESEC  is defined to implement file security; or undefined
  to temporarily bypass its use.  This writes (and reads) a security line
  at the end of each data file in the system.  */

//#define  REJOLDATA
//#undef  REJOLDATA
/* This symbol is not actually used.  Yet.  */
/* Symbol  REJOLDATA  is defined to add an optional test for the version
  in the input data, and reject the data file if the version of the
  program which made it is old or obsolete relative to the latest
  available version of that program.  A command line flag can be used
  to over-ride this test.  Will apply to maskgen and maskcut but not
  to intgui, and ncplot will only give a warning message.  */

//#define  WFCCD
#undef   WFCCD
/* Symbol  WFCCD  is defined to add features to maskcut, ncplot and intgui
  to support making of masks for the WFCCD instrument on the 100"
  NOTE -- this symbol is no longer needed or used; it may be used
  in INTGUI for support development.   */

#define  DSSGET
//#undef  DSSGET
/* Symbol  DSSGET  is defined to add feature to skywin that a control-right
  button press will request and display an image from the DSS over the net
  for the selected guide star.  This will enable users to see if a guide
  star is in fact a valid star when selecting it.  
  Use "setenv MYDISPLAY xx"; see newinstr.txt   */

//#undef  CUTREPEAT
#define  CUTREPEAT
/* Symbol  CUTREPEAT  is defined to add feature to maskcut such that a
  -x flag can cause mask features (slits, alignment boxes, wing chip
  holes) to be cut with a repeat count such as 2 times if needed to
  compensate for a tired or mis-adjusted laser machine.  This should
  not normally be needed.  */

#define  AUTOGS
//#undef  AUTOGS
/* Symbol  AUTOGS  is defined to have skywin automatically find appropriate
  guide stars from stars in its display when needed.  A global flag will
  indicate when guide stars are needed and when they have been found or
  attempted to have been found.
  Initially, this should only affect IMACS instruments.  */

#define  NSROT
//#undef  NSROT
/* Symbol  NSROT  is defined to activate the nod-and-shuffle rotation
  feature.  In IMACS short camera mode only, a 90 degree rotation of
  the grism is allowed to accomodate the nod-and-shuffle method of
  sky subtraction.  This is controlled by a check box in the intgui,
  and rotates the mask geometry by 90 degrees, affects the rotation
  warnings, the gap values, and rotates the field for mask cutting
  also.  It may be used in LDSS-3 also.   */
/* Correct operation verified for IMACS case, 2007/2-13 */
/* Feature required to support observations after 2007-05-14  */

/*  ---  Polynomial features for GISMO transforms  ---  */

#define  CompP  00	/* Complex polynomial default */
#define  CConj   1	/* Complex and conjugate polynomial */
#define  GenPol  2	/* Generalized polynomial */
/* Above define polynomial types; below sets the actual value*/
#define  POLY  CConj	/* Using conjugate case 2007/3-12 */
/* NOTE == ONLY change this when also deriving and installing the
  corresponding polynomial via the "tcpoly" utility !!  */

/*  The POLY symbol is used to define which of several possible polynomial
  standards is being used for (mostly) the GISMO transforms.  The default
  choice (was) the complex polynomial, but that does not work well, so either
  a complex polynomial with conjugate terms is used, or a general polynomial
  with separate terms for X and Y.  The latter two are purposely non-conformal
  which appears to be necessary.  The conjugate method is a little easier,
  and should need fewer points; the separate polynomials will fit a wider
  variety of problems but requires more terms.  */
/* The use of this definition is, for example
  #if  POLY == CConj     // conjugate
  #elif  POLY == GenPol // general
  #else                // default complex
  #endif
  to get separate sections for all 3 types of polynomial, one at a time */

/*  ---  End of GISMO polynomial features  ---  */

/* The GISMO instrument and features are now defined for general use
  as of 2011/02-10.  However, notwithstanding the above, the symbol
  GISMOLS will remain undefined (see below) until such time as the
  mask space location of the long slit is adequately determined to
  allow its use, which is mainly listing it in the .SMF file.    */

#define  GismoMenus
//#undef  GismoMenus
/* Symbol  GismoMenus  controls the enable of the GISMO feature in the
  intgui menu system independently of all the other GISMO features
  which are being defined generally in maskgen, maskcut, ncplot and
  general utilities to support existing GISMO mask files.  Only the
  origin of new GISMO masks is suppressed by the absence of the
  GismoMenus symbol.  It is forced on by the "make gintgui" command
  in ~/imacs/maskgen.  Defining it will provide full GISMO support
  for distributed systems.    */
/* Symbol  GismoMenus  is allowed 2011/02-10 per verbal authorization
  from Gladders (himself, in person); effectively enabling GISMO  */

/* Set the symbol  GISMENU  the same as symbol  GismoMenus  is set,
  for future use and feature list in mgutils.c;  This is the new name
  for this feature, the old  GismoMenus  is depricated.    */
#ifdef  GismoMenus
#define  GISMENU
#else
#undef  GISMENU
#endif

#define  GismoFeatures
//#undef  GismoFeatures
/* Symbol  GismoFeatures  controls the following features easily rather
  than individually selecting them.  It is used to make a test version
  with full GISMO support during commissioning of GISMO.  Should be
  defined when the instrument becomes standard.
  Actually; should be removed and features defined individually.
  NOTE that these features do NOT include GismoMenus.    */
/* The GismoFeatures may be defined to allow full GISMO support in the
  maskegn and maskcut tasks.  Only the instrument menus in intgui are
  being suppressed prior to public availibilty by the GismoMenus symbol
  being not defined.  This allows only gintgui to be different, without
  requiring a separate maskgen and maskcut task set.    */
/*  NOTE that  GismoFeatures  is ONLY USED HERE in mgfeats.h as a
  convenience.  It is no longer used in the makefile.    */

/* For coordination, the GismoFeatures is needed to be defined when
  the more inclusive GismoMenus symbol is defined.  Otherwise, it
  may be selected or not at will.  */
#ifdef  GismoMenus
#define  GismoFeatures
#endif

#ifdef  GismoFeatures
#define  GISMO
//#undef  GISMO
#else
//#define  GISMO
#undef  GISMO
#endif  // on GismoFeatures
/* Symbol  GISMO  is defined to activate all modifications related to
  the gismo instrument, a front end to IMACS f/4 spectrograph.
  Mostly, this is the menu items in intgui.  */

//#ifdef  GismoFeatures
////#undef  GISMODECKER
//#define  GISMODECKER
//#else
//#undef  GISMODECKER
////#define  GISMODECKER
//#endif  // on GismoFeatures

#undef  GISMODECKER
//#define  GISMODECKER
/* Symbol  GISMODECKER  is defined to include code to produce a decker
  mask for GISMO as an option.  It is normally not included, and not
  needed for the distribution version.  
  Turn this on only if a new decker for GISMO is needed; since that
  is an instrument part, this would ve very rare.  Affects maskcut.c  */

#ifdef  GismoFeatures
// #if  defined GismoFeatures || defined GismoMenus
#define  GISMOAP
//#undef  GISMOAP
#else
//#define  GISMOAP
#undef  GISMOAP
#endif  // on GismoFeatures
/* Symbol  GISMOAP  is defined to include the feature of a special instrument
  in the intgui which will produce a .obs file to instruct maskgen to produce
  a .SMF file for the GISMO aperature mask, consisting of rectangular areas
  through which alignment boxes can be imaged.  This is needed for regular
  use of GISMO; each observer should have his own aperature mask file.    */

#ifdef  GismoFeatures
//#undef  GISMOE
#define  GISMOE
#else
#undef  GISMOE
//#define  GISMOE
#endif  // on GismoFeatures
/* Symbol  GISMOE  is defined to temporarily allow use of MOE with
  GISMO as an experiment to see if they are actually compatable.  */

#ifdef  GismoFeatures
//#undef  GISCAM
#define  GISCAM
#else
#undef  GISCAM
//#define  GISCAM
#endif  // on GismoFeatures
/* Symbol  GISCAM  is defined to make the camera name in skywin the same
  as the frontend name, and use "chip" definitions for the subfields
  to display the special field of view.  */

#ifdef  GismoFeatures
#define  GISMOMNS
//#undef  GISMOMNS
#else
#undef  GISMOMNS
#endif  // on GismoFeatures
/*  Symbol  GISMOMNS  is defined to implement the special GISMO
  "Macro Nod and Shuffle" mode in which 1/7 of each detector in the
  f/4 N&S mode is used.  With a special instrument designation, gaps
  are simulated blocking the unused detector areas, and gap mode with
  n&s is turned on.  GAPVO is required.
  This symbol is required to read .obs file made for GISMO MNS instrument.   */

#ifdef  GismoFeatures
//#define  GISMOLS
#undef  GISMOLS
#else
#undef  GISMOLS
#endif  // on GismoFeatures
/*  Symbol  GISMOLS  is defined to implement the GISMO Long Slit feature,
  which is a special object.  The special symbol is needed to hide this
  feature until the correct parameters are found.    */

#ifdef  GismoFeatures
#define  GISMOF2
//#undef  GISMOF2
#else
#undef  GISMOF2
#endif  // on GismoFeatures
/*  Symbol  GISMOF2  is defined to implement a menu entry for GISMO on
  the IMACS f/2 camera, which points to the proper set of dispersers
  and should be installed near the f/4 or regular GISMO instrument.  */


/*  ---  Storage of data from configuration files  ---  */

#define  USE_CDAT
//#undef  USE_CDAT
/* The  USE_CDAT  symbol is defined to use the "cdata" structure in
  an array indexed by "inst", to locally store all the data read from
  the maskdat.cfg file by either BSTO or GSTO reading programs. */

#define  BSTO
//#undef  BSTO
/* The  BSTO  symbol replaces the generalize storage method of
  gsto with the b-tree method.  It drives two symbols which allow
  the use of these two competing methods.  The split is designed to
  allow introduction of a third method (or more) later, driven by
  a more complex definition.  */

/* Define other symbols based on whether USE_CDAT is active, and
  which file reading method is to be used.  */

#ifdef  USE_CDAT
/* Using cdata, not using the others at all */
#undef   USE_BSTO
#undef   USE_GSTO
#else	// on USE_CDAT
/* No cdata, need either USE_BSTO or USE_GSTO */
#ifdef  BSTO
#define  USE_BSTO
#undef   USE_GSTO
#else
#define  USE_GSTO
#undef   USE_BSTO
#endif
#endif	// on USE_CDAT

/* The following defines are independent of USE_CDAT -- these defines
  are used to drive the reading method as opposed to use. */
#ifdef  BSTO
#define  READ_BD
#undef   READ_GD
#else
#define  READ_GD
#undef   READ_BD
#endif

/*  ---  End of storage of configuration data section.  ---  */


#define  USE_ACON
//#undef  USE_ACON
/* The USE_ACON symbol implements computation of elapsed time of
  cutting within maskcut,  to be totaled and reported at end of 
  the .nc file; to be written into a trailing comment in .nc ... */

#define  OBJNAM
//#undef  OBJNAM
/* Symbol  OBJNAM  is defined to implement the plotting of names for
  each object in the debug plots.  Needs a -n flag in the maskgen
  execution to become effective.  */

#define  OBJCHECK
//#undef  OBJCHECK
/* Symbol  OBJCHECK  is defined to implement a consistency check on
  all objects in maskgen.  When present the check is done, and a
  command line flag allows processing if it fails.  */

#define  GSREQUIRED
//#undef  GSREQUIRED
/* Symbol GSREQUIRED is defined to require the use of guide star data
  for mask generation step.  When present, conditionals requiring the
  presence of guide star data in the .obs file are enforced for all
  IMACS instruments, but not for LDSS instruments.  */

#define  NOQUIKGS
//#undef  NOQUIKGS
/* Symbol  NOQUIKGS  is defined to allow a quick look execution to
  proceede without guide stars being specified.  Since no .SMF file
  is produced, this is a valid bypass.  When not defined, the strict
  guide star requirement is enforced even on quick look.  */

#define  GSDATA
//#undef  GSDATA
/* Symbol GSDATA is defined to implement the feature of designating and
  passing on the locations of up to 2 guide stars for the observing
  catalog comments written to the .SMF file.  */
/* Consider making this standard.  */

//sunifdef < #undef  GSDATA_2
// Removed 2010/08-05  Release 2.09+
/*  Symbol GSDATA_2 may be defined to clear guider data when any
  change is made to position or orientation.  Normally undefined when
  the AUTOGS feature is active; should always be undefined now. */
/*  Permanently undefined from 2010/8-5  */

//#ifndef  GSDATA
//#undef  IMAGM
//#endif
/*  Use of IMAGM requires GSDATA to also be defined.  If it is not,
  we are required to undef the IMAGM symbol here.  */
/*  This may not actually be needed.  */

#define  GUIDATE
//#undef  GUIDATE
/* Symbol GUIDATE is defined to enable a date field in place of the
  temperature window in the interface gui.  This date enables the
  display of a sun-twilight-night color bar below the graphic for
  rotator reversals.  The date information is stored as MJD in the
  .obs file, with a default value of 0 meaning no date and no
  sun display.  The windows are pull-down menus for date selection.
  When GUIDATE is defined, the maskcut program will use the temperature,
  probably the default value, in the cutting process as a default.  */

#ifdef  GUIDATE
#define  MOONLINE
//#undef  MOONLINE
#else  // on GUIDATE
#undef  MOONLINE
#endif  // on GUIDATE
/* Symbol  MOONLINE  requires GUIDATE be defined, and if defined it
  allows plotting of a moon rise/set graphic and moon % illumination
  value below the sun rise/set graphic line.  */

#ifdef  MOONLINE
#define  MOONANG
//#undef  MOONANG
#else  // on MOONLINE
#undef  MOONANG
#endif  // on MOONLINE
/* Symbol MOONANG requires moonline, and if defined will append the
  angle between field center and moon to left of moon line. */

#define  REFLIM
//#undef  REFLIM
/* Symbol REFLIM is defined to establish the new method of limiting
  the number of reference stars.  It replaces the old Ref. Select
  window in intgui with Ref. Limit window, and requires the default
  value of -a in maskgen.  Old value is still supported on reading,
  but is not in the gui.  */

#define  GAPVO
//#undef  GAPVO
/* Symbol GAPVO is defined to support the virtual object creation and
  use to model detector gaps.  If not defined, no intgui controls are
  created, and maskgen does not create virtual objects.  When defined,
  full support is compiled.  */

#ifdef  GSSMOMNS
#define  GAPVO
#endif
/* The GISMOMNS feature requires GAPVO be on.  It should be standard.  */

#undef  GAPLDSS
//#define  GAPLDSS
/* Symbol GAPLDSS is defined when (if) the LDSS instrument has a known
  gap value in the optical file; without it the menu is disabled. */

#define  DIFREF
//#undef  DIFREF
/* Symbol DIFREF is defined to support the differential refraction
  computation in mask positions.  If not defined, no change is made
  to the mask position.  Adds a Dif. Ref. check button in intgui.
  Supports Hour Angle box in intgui.    */

#ifdef  DIFREF
#define  UPARO
#else
#undef  UPARO
#endif
/* Symbol UPARO is defined to allow feature of zenith vector in the
  skywindow when testing.  After it works, rip it out and make it
  part of the difref feature... (see UPOPTION in skywin, etc.)
  Since it works, UPARO is now tied to equal DIFREF; do not change. */

#undef  ZNTHDOT
//#define  ZNTHDOT
/* Symbol ZNTHDOT is defined to allow option to plot a distinctive
  dot at the zenith in the sky view; this is to be used only for debug.
  NOTE:  This ZNTHDOT requires UPARO also be defined!  */

#define  SLXB
//#undef   SLXB
/* Symbol  SLXB is defined to support the slit extension feature,
  and mostly the button in intgui implementing it.  */

#define  DMOE
//#undef  DMOE
// DMOE is enabling flag for MOE support of extra disperser

#define  DWAV
//#undef  DWAV
/* Symbol DWAV is defined to allow entry of detector wavelength
  bounds separately from conflict wavelength bounds.  Required
  to be defined if DMOE is defined. */
#ifdef  DMOE
 #define  DWAV
#endif

#define  GTILTMOE
//#undef  GTILTMOE
/* Symbol GTILTMOE allows a grating tilt to be specified in opticdef
for the echelle specification.  */

//#define  RLSFULL
#undef  RLSFULL
/* Symbol RLSFULL allows the full computation of reordering using
  actual end points of each slit.  If not on, actual start point is
  used for both start and end of the slit. */
/* Actually used only in mgutils.c, should be coordinated between
  that use and cututils.c for reordering.  */

// #define  GCENT
// //#undef  GCENT
// // GCENT is the global centering feature 
// // Defined for skywin, and optionally used in intgui
/* The above define is now STANDARD for maskgen */

#define  NNA
//#undef  NNA
/* Symbol NNA activates the Nearest Neighbor Algorithm, which pre-sets
  each node's nearest neighbor as next to that node to jump start the
  shortest total distance part of the algorithm, hoping to speed it
  up siginficantly in the case of many nodes.  Used in the reorder
  program in mgutils; used here to define needed items in the objq
  structure in maskdat.h . */


/* --- Following items are normally defined --- */

// #define  EXORDX
// //#undef  EXORDX
/* Symbol EXORDX implements the feature to allow an extra order check
  on comparison star spectra, but not on object spectra.  This is a
  variable in the obs structure, and supported by a pull-down menu
  rather than a check box in the interface GUI.  */
/* The above define is now STANDARD for maskgen */

// #define  AXBUT
// //#undef  AXBUT
/* Symbol AXBUT implements a button which will save the current
  data, run mask generation, and return to the window.  */
/* The above define is now STANDARD for maskgen */

#define  PRIORDER
/* Symbol PRIORDER is used to drive the selection of otherwise similar
  nodes in conflict resolution activities.  When set, lower ordered
  objects will be selected over higher ordered ones; when unset the
  opposite happens, later objects override the earlier.  */



	/* - - - - - - - - - - - - - - *\
	|				|
	|        Debug  Features	|
	|				|
	\* - - - - - - - - - - - - - - */

/* Use these in a statement like:
  #if (DEBUG_XX > 1)
	printf ("bla bla bla\n");
  #endif
.. Set the level to 0 to drop the messages, and higher integers
get progressively more of the messages.  Then its easy to turn
the debug messages on/off.   */

#define  LocDebug  0
// For debugging locfile program in mgutils
/* Symbol LocDebug is the locate debug level.  0 = none,
  1 = minimal debug, 2 = basic debug, 3 = detail debug.   */
/* Actually, this will allow dodebug to decide debug level */

#define  DEBUG_GUI  0
// For controlled messages in intgui

#define  DEBUG_GEN  0
// For controlled messages in maskgen

#define  DEBUG_CUT  0
// For controlled messages in maskcut

#define  DEBUG_PLT  0
// For controlled messages in ncplot

#define  DEBUG_UTL  0
// For controlled messages in general utilities


	/* - - - - - - - - - - - - - - *\
	|				|
	|     Obsolete  Features	|
	|				|
	\* - - - - - - - - - - - - - - */

/* Below are mostly unused or obsolete */

// #define  LDSS
#undef  LDSS
// LDSS is enabling flag for LDSS-2 instrument support


/*  ====  */
#endif     /*  INCLUDE_MGFEATS_H  */
/*  ====  */

/*  ===  Revsion History  ===  

2005/03-23	Started this feature.

... */
