/* Common utility programs for .nc file reading */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                  Copyright (C) 2008 by                            **
**     Ken Clardy,  Carnegie Observatories,  Pasadena California.    **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define  IS_NCRUTIL_C
#include	"ncrutil.h"

	/* - - - - - - - - - - - - - - *\
	|				|
	|      Parsing Subroutines	|
	|				|
	\* - - - - - - - - - - - - - - */

/*  ===  Parsing Subroutines  ===  */

// Removed commented code which is actually in ioutil.c

/* Letter parse - find next letter in buffer c, return it uppercased
  at the indicated location, and return the NEXT location in buffer. */
static	char*	ltrparse (char* buf, char* ltr)
{
char	*p;
    *ltr = AN;
    if (buf == NULL) return NULL;
    for (p = buf; *p != AN; p++) {
	if (*p >= 'A' && *p <= 'Z') {
	    *ltr = *p;
	    p++;
	    return p;
	}
	if (*p >= 'a' && *p <= 'z') {
	    *ltr = *p + ('A' - 'a');
	    p++;
	    return p;
	}
    }
    return  p;
}


/* Numeric parse -- find numeric value in buffer starting at c, return
  it in the indicated location, and return the NEXT location in buffer. */
/* Scan past any trailing whitespace */
static	char*	numparse (char* buf, double* v)
{
char	*p;
char	*x;
double	z;	// normalization
int	k;	// decimal count
double	d;
int	j;	// scan count
    *v = 0.0;
    if (buf == NULL) return NULL;
// Skip over any leading whitespace (so we don't stop on it)
    p = skipover (buf, WHITESPACE);
// Scan the number
    for (j=k=0,z=1.0; *p != AN; p++) {
// Allow sign(s) as part of numeric
	if (*p == '+') { j++; continue; }
	if (*p == '-') { j++; z *= -1.0; continue; }
// Count decimal point(s) to allow normalization
	if (*p == '.') { j++;
	    k++;	// Count decimals
// If k exceeds 1, we may want to scan token and break
	    if (k > 1) {	// doubled decimal, scan out token and break
		x = strpbrk (p, WHITESPACE);
		if (x == NULL) x = strchr (p, 0);
		p = skipover (x, WHITESPACE);
		break;
	    }
	    continue;
	}
	if (*p >= '0' && *p <= '9') {
	    j++;
	    d = (double) (*p - '0');
	    *v *= 10.0;
	    *v += d;
	    if (k > 0) z /= 10.0;	// Keep normalization
	    continue;
	} else {	// Non-numeric -- terminate (including whitespace)
	    if (j > 0) p = skipover (p, WHITESPACE);
	    break;
	}
    }
    *v *= z;	// normalize, including sign
    return p;
}



static	char*	getcomment (char* c)
/* Obtain a dynamic copy of a comment */
{
char	*s;
char	*e;
char	*b;
char	*x;
int	k;
double	val;
char	form[80];
char	unit[40];

// Find the start of the actual comment
    s = strchr (c, '(');
    if (s == NULL) return s;
    for (s++; strchr(WHITESPACE, *s); s++) ;

/* New section -- verify valid comment.  Count parenthesis, and
  display an error message if the count is bad.  */
    for (b=c,k=0; *b != AN; b++) {
	if (*b == '(') k++;
	if (*b == ')') k--;
	if (k < 0) break;
    }
    if (k != 0) {
	printf (" ** Comment statement error (%d), text follows:\n", k);
	printf (" %s \n", c);
    }

// Find the end of the comment
    e = strrchr (s, ')');	// last ) if several exist...
    if (e == NULL) e = strchr(s,0);
    for (e--; strchr(WHITESPACE, *e) && e >= s; e--) ;

// Obtain storage for the comment
    k = e - s + 1;	// actual characters
    if (*s == AN) k = 0;
    if (k == 0) return NULL;
    b = malloc (k+1);

// Copy the comment characters, and append a null
    strncpy (b, s, k);
    b[k] = AN;

// Check if the comment is the time estimate
    if (strstr (b, Legend) != NULL) {
	x = &(b[strlen(Legend)]);
	sprintf (form, " %s", "%lf %s.");
	k = sscanf (x, form, &val, unit);
	if (k == 2) {
	    ComTime = val;
	    if (strncasecmp(unit, "hour", 4) == 0) ComTime *= 60.0;
//	} else {		// debug only section
//	    printf (" Comment: \"%s\".\n", b);
//	    printf (" Reduced: \"%s\".\n", x);
//	    printf (" Format: \"%s\".\n", form);
//	    printf (" k=%d, val=%.3f, unit=\"%s\".\n", k, val, unit);
	}
    }

// Return the extracted stuff...
    return  b;
}



	/* - - - - - - - - - - - - - - *\
	|				|
	|    Linked Storage Utilities	|
	|				|
	\* - - - - - - - - - - - - - - */


/*  ====  Linked Storage support utilities  ====  */


static	mlist*	newmlist (int l, int a, int i, int e, double f)
{
mlist	*p;
//    p = xalloc (mlist);
    p = (mlist*) malloc (sizeof (mlist));
    p->next = p->last = NULL;
    p->laser = l;
    p->afoc = a;
    p->interp = i;
    p->par = e;
    p->xac = 0;
    p->endl = 0;
    p->begl = 0;
    p->feed = f;
    p->v = make3vect (UNSET, UNSET, UNSET);
    p->ctr = p->v;
    p->com = NULL;
    return  p;
}

static	mlist*	apendml (mlist* p, mlist* a)
/* Append p onto the end of list at a.  Return end of list */
{
mlist	*r;
    r = a;
    if (a == NULL) r = p;
    else {
	while (a->next != NULL) a = a->next;
	a->next = p;
	if (p != NULL) {
	    p->last = a;
	    r = p;
	}
    }
    return  r;
}

static	mlist*	popml (mlist* a)
/* Return pointer to next thing available in list */
{
mlist	*r;
mlist	*p;
    if (a == NULL) return a;
    r = a->next;
    p = a->last;
// Fix up pointers
    if (r != NULL) r->last = p;
    if (p != NULL) p->next = r;
    a->next = a->last = NULL;
// Select the returned link
    if (r == NULL) r = p;
    return  r;
}

mlist*	freeml (mlist* a)
{
mlist	*p;
    if (a == NULL) return a;
    p = popml (a);
    if (a->com != NULL) free (a->com);
    free (a);
    return  p;
}


	/* - - - - - - - - - - - - - - *\
	|				|
	|      Timing for .nc file	|
	|				|
	\* - - - - - - - - - - - - - - */

//#define  DbigT		// Debug only
#undef  DbigT		// Debug only

double	find_times (mlist* head)
/* Fill in timing tables, and return total time in minutes */
{
mlist	*p;
double	tot = 0.0;
int	j;
vect3	pv;
vect3	v;
vect3	dv;
pvec3	hv; //debug
double	d;
double	t;
// #ifdef  Acon
double	fa, dr;
double	da;
int	na;
// #endif
vect2	cen;
double	rad;
vect2	rv, nv;
pvec2	pp, pn;
double  arc;
#ifdef  DbigT
int	kk = 0;
#endif

#ifdef  DbigT
    printf (" Acceleration constant Acon = %7.2e\n", Acon);	// debug only
#endif
    for (j=0; j<Ncond+1; j++) { tx[j] = dx[j] = 0.0; sx[j] = 0; }

// Loop to scan through the list and gather statistics
    pv = make3vect (UNSET, UNSET, UNSET);
//    for (p=head; p!= NULL; p = p->next) {
    C_loop (p, head) {
// Keep by laser on/off, cut/engrave, feed rate
//	if (p->interp < 0) continue;  // avoid special blocks
/* For special blocks, check for and add time of any dwell blocks */
	if (p->interp == -2) {	// Pause block, value in p->v.x
	    tx[7] += p->v.x / 60.0;
	    (sx[7])++;
	    continue;
	} else if (p->interp < 0) continue;

	v = p->v;
// See if we have a valid prevous positon, or this
//	if ( !( pv.x < VSET || pv.y < VSET || v.x < VSET || v.y < VSET)) {
	if ( !( pv.x < VSET || pv.y < VSET || v.x < VSET || v.y < VSET
		|| pv.z < VSET || v.z < VSET)) {
// Get distance from previous position
	    dv = sub3vect (v, pv);
	    if (p->interp == 1) {	// Linear
		d = vect3norm (dv);
	    } else if (p->interp == 2 || p->interp == 3) {	// circular
// Tricky.  Assume x-y (good for mask cutting).  Get center and
// position vectors in 2d.  Convert to polar, and find angles;
// then find angle x radius for distance.
		cen = get2de3v (p->ctr);
//		pv = make2vect (px, py);
		rv = get2de3v (pv);
		nv = get2de3v (p->v);
// Convert pv/nv to be relative to the center:
		rv = sub2vect (rv, cen);
		nv = sub2vect (nv, cen);
// Obtain polar versions of pv, nv:
		pp = polar2 (rv);
		pn = polar2 (nv);
		arc = pn.t - pp.t;
		rad = pn.r;
		if (p->interp == 2) {	// cw to arc
		    if (arc > 0.0) arc -= TWOPI;
		} else {		// ccw
		    if (arc < 0.0) arc += TWOPI;
		}
		if (fabs(arc) < 0.0001) arc = TWOPI;
// arc is arc length, positive is ccw here...
		if (arc < 0.0) arc = -arc;
		d = rad * arc;
	    } else continue;
// #ifndef  Acon
// 	    t = d / p->feed;
/* NOTE -- This time does not include accleration/decelleration.  If the
  exact stop mode is in effect, we should add a little time for that
  to take place.  If it takes x seconds to accelerate or decellerate,
  then a time of t=x needs to be added.  For a parameter of the assumed
  acceleration time, add min (t, acctime) to t here.  The acctime will
  be greater if feed rate is greater; therefore we should find atim
  as k * p->feed, and t += min (t, atim) here.  Units?  feed is in
  mm/minute, d is in mm, t is in minutes.  Then atim is very small,
  let's say 50 milliseconds at 500 mm/min, so k = 0.1 * feed millisec,
  and k = 0.1 * feed * .001 / 60 minutes = 1.66e-6.
.. */
// #else
	    if (p->xac) na = 2;
	    else        na = p->endl + p->begl;
	    fa = p->feed * Acon;
	    da = na * 0.5 * p->feed * fa;
	    dr = d - da;
	    if (d <= 0.0)      t = 0.0;
	    else if (dr > 0.0) t = na * fa + dr/p->feed;
	    else               t = na * sqrt (d*Acon);
// For derivation of above, see acceler.txt 

// That was previously:
//	    ax = Acon * p->feed;
//	    t += (t < ax) ? t : ax;
// #endif

/* Debug section to trap very large t values... */
#ifdef  DbigT
	if (fabs(t) > 1.0e5 && kk++ < 7) {	// more than a day!
	    printf ("--Time of movement = %10.3g minutes.\n", t);
	    printf ("-- Laser %d par %d\n", p->laser, p->par);
	    printf ("--Loc. %.3f %.3f %.3f\n", p->v.x, p->v.y, p->v.z);
//	    printf ("--d = %.2f  dr = %.2f  da = %.2f\n", d, dr, da);
//	    printf ("--na = %d  fa = %.2f\n", na, fa);
//	    printf ("--endl = %d, begl = %d\n", p->endl, p->begl);
	    printf ("--feed = %.2f  Acon = %.4e\n", p->feed, Acon);
	}
#endif

	    tot += t;
	    seg++;
	    if (p->laser) {  // cutting
		if      (p->par == 3) j = 3;		// Engraving
		else if (p->par > 3) j = 2;		// Rough cut
		else    j = (p->feed > 520.0) ? 1 : 0;  // cut speed
	    } else {
		j = (p->feed > 400.0) ? 5 : 4;
// Select case 6 as vertical movement; 3-d diff near z axis:
		hv = polar3 (dv);
		if (fabs(hv.z) > 10.0 * hv.r) j = 6;	// Vertical
	    }

	    tx[j] += t;
	    dx[j] += d;
	    (sx[j])++;
	}	// End of test for valid segment
	pv = v;		// remember previous position
//	if (p->next == head) break;
    }

    return  tot;
}


	/* - - - - - - - - - - - - - - *\
	|				|
	|      Reading of .nc file	|
	|				|
	\* - - - - - - - - - - - - - - */


/* The test file reading program... */
//mlist*	Data_Processor (char* FileName, opts opt)

mlist*	Read_ncfile (FILE* handle, int dbug, int gdbug)
{
char	buf[256];
char	token[32];
char	name[32];
char	*c;	// scan pointer
char	*p;	// alternate pointer
char	ltr;
int	k;
int	m;
double	v;
mlist	*top, *tail;
mlist	*cml;
mlist	*lastm=NULL;	// last movement pointer
double	*cp;		// Coordinate pointer
int	gc=0;	// G code count
int	gm=0;	// Message count
int	nl=0;	// Line count

// status to be entered into data structures
int	incr=0;		// incremental mode
int	mv=0;		// movement flag
int	pv=0;		// parameter flag
int	lsr=0;		// laser on/off
int	dlsr=0;		// change in laser state
int	afoc=0;		// autofocus on/off
int	intrp=0;	// interpolation mode
int	sintrp=0;	// saved interp mode
int	xact=0;		// exact stop mode
int	par=0;		// laser parameter
vect3	loc;		// Old loc in incremental mode
vect3	ctr;		// Center location
double	scale;		// scale factor mm or inch
double	x, y, z;
double	cx, cy, cz;	// centers for circular
double	feed=0.0;	// Initial value for lint
double	pause=0.0;	// Pause value in seconds
char	*com=NULL;	// current comment

    top = tail = NULL;
    x = y = z = UNSET;
    scale = 1.0;

/*  ---  Loop for Command Data file  ---  */

    while (fgets(buf, 250, handle)) {

/* Process the buffer load in this loop */
	nl++;	// keep the line number.
	if (gdbug > 1) printf (">>%s", buf);
	for (c = strchr(buf,0);	// start at end of buffer
	    c >= buf && *c < ' ';  // in buffer and a control character
	    c--)  *c = AN;	// Kill trailing control codes
	mv = pv = 0;
	k = 0;
	gc = 0;

/* Set the cx coordinates to the unset value. */
	cx = cy = cz = UNSET;
/* Any value less than -1.0e9 is considered "not set" and
  will be set to the current x,y,z when the line is entered. 
  Incremental mode values will set cx,cy,cz to the increments, and
  the absolute values will be computed when the line is entered. 
  Thus in incremental mode, unset values get set to 0.0 just before
  the absolute coordinates are resolved.  */

/* Keep the coordinates in case mode is incremental */
	loc = make3vect (x, y, z);

/* Parse the line, finding commands and codes */
	c = buf;
	while ((c = skipover (c, WHITESPACE))) {
	    if (*c == AN) {	// End of line clean-up
		if (mv > 0) {	// If we have a move
		    if (gdbug > 1)
			printf (" -> Move %7.5f %7.5f %7.5f\n", x, y, z);
// HERE WE CREATE THE MLIST ENTRY AND PUSH IT TO THE END OF THE LIST
// USE CURRENT XYZ AND STATUS VALUES
		    cml = newmlist (lsr, afoc, intrp, par, feed);
		    cml->xac = xact;
// LOAD THE LOCATION VECTORS, AND PROCESS ANY COMMENTS...
		    cml->v = make3vect (x, y, z);
//		    if (intrp == -2) cml->v.x /= 60.0;
		    if (intrp == -2) {
			cml->v = make3vect (pause, 0.0, 0.0);
			pause = 0.0;
		    }
// how we set cx and others
//      incremental  not incremental
// unset   xo + 0        x
//  set   xo + cx       cx
// if unset set to 0 for incremental, x if not
// for incremental, add former x to find center
		    if (intrp != -2) {
			if (cx < VSET) cx = incr ? 0.0 : x;
			if (cy < VSET) cy = incr ? 0.0 : y;
			if (cz < VSET) cz = incr ? 0.0 : z;
			ctr = make3vect (cx, cy, cz);
			cml->ctr = incr ? sum3vect (ctr, loc) : ctr;
		    }

// If movement has laser change, record that
		    if (dlsr && intrp != -2) {
			cml->begl = 1;
			dlsr = 0;
		    }

// Save the location of an actual move, non-comment, and non-pause
		    if (com == NULL && intrp != -2) lastm = cml;

// Old way:
// if incremental add xo to cx (can stay unset)
// if unset, set to x
//		    if (incr) { cx += x; cy += y; cz += z; }
//		    if (cx < VSET) cx = x;
//		    if (cy < VSET) cy = y;
//		    if (cz < VSET) cz = z;
//		    cml->ctr = make3vect (cx, cy, cz);

// Process a comment.  Put the current comment pointer into
// the new current motion list structure and clear it.
		    cml->com = com;
		    com = NULL;
    
// Put the constructed current motion (cml) into the queue after tail
		    if (top == NULL) top = cml;
		    tail = apendml (cml, tail);

// If sintrp not same as intrp, fix that up
		    if (sintrp != 0 || intrp < -1) intrp = sintrp;

/* Check here (within move) for duplicate G commands in line */
		    if (gc > 1) {
			if (gm++ < 10) {
printf (" ** Double G command line %5d: \"%s\".\n", nl, buf);
			}
		    }

		}	// if we have a move

// Create a special block if pause is active
		if (pause > 0.0) {
		    cml = newmlist (lsr, afoc, -2, par, feed);
		    cml->ctr = make3vect (UNSET, UNSET, UNSET);
		    cml->v = make3vect (pause, 0.0, 0.0);
		    if (top == NULL) top = cml;
		    tail = apendml (cml, tail);
		    pause = 0.0;
		}

		break;
	    }	// End of line

	    if (*c == '(') {	// Apparent comment text
/* If we have currently a comment, let it go. */
		if (com != NULL) {
//		    if (dbug > 1) printf ("  == %s ==\n", com);
// Above changed to 1 for better debug use KDC 08/10-17
		    if (dbug > 1 && gdbug < 2) printf ("  == %s ==\n", com);
// Above print is redundant, misplaced and misleading if gdbug > 1.
		    free (com);
		    com = NULL;
		}
/* Handle the comment.  Get dynamic copy of comment text. */
		com = getcomment (c);

/* Check for an instrument name field, and if found put it into
  the queue as a special block... */
		p = parse (com, token, WHITESPACE);
		if (!strcmp (token, "Instrument:")) {  // Special comment
		    p = parse (p, name, WHITESPACE);
		    cml = newmlist (lsr, afoc, -1, par, feed);
// Special block has interp -1 for instrument,
// and unset values for vectors.
		    cml->com = malloc(strlen(name)+1);
		    strcpy (cml->com, name);
		    cml->ctr = cml->v = make3vect (UNSET, UNSET, UNSET);
		    if (top == NULL) top = cml;
		    tail = apendml (cml, tail);
		}

/* Check for "By" and put comment in special block with interp -3 for
  the source comment, which may be found later. */
		if (!strcmp (token, "By")) {	// source comment
		    cml = newmlist (lsr, afoc, -3, par, feed);
// Special block has interp -3 for source comment,
// and unset values for vectors.
		    cml->com = malloc(strlen(com)+1);
		    strcpy (cml->com, com);
		    cml->ctr = cml->v = make3vect (UNSET, UNSET, UNSET);
		    if (top == NULL) top = cml;
		    tail = apendml (cml, tail);
		}

/* Check for the phrase "mask file by " and put comment in a special
  block with interp -3 as an alternate form of source comment.  */
//		if (!strcasestr (com, "mask file by ")) {
		if (strstr (com, "mask file by ") != NULL) {
		    cml = newmlist (lsr, afoc, -3, par, feed);
// Special block has interp -3 for source comment,
// and unset values for vectors.
		    cml->com = malloc(strlen(com)+1);
		    strcpy (cml->com, com);
		    cml->ctr = cml->v = make3vect (UNSET, UNSET, UNSET);
		    if (top == NULL) top = cml;
		    tail = apendml (cml, tail);
		}

/* Check for mask serial number followed by instrument name field, and
  put it into queue as a special block also... */
		p = parse (com, token, WHITESPACE);
		if (!strcmp (token, "Mask:")) {  // Special comment
		    p = parse (p, name, WHITESPACE);  // Serial number
// Parse past the serial number following
		    p = parse (p, token, WHITESPACE);  // New token

		    if (!strcmp (token, "Instrument:")) {  // Special comment
			p = parse (p, name, WHITESPACE);
			cml = newmlist (lsr, afoc, -1, par, feed);
// Special block has interp -1 for instrument,
// and unset values for vectors.
			cml->com = malloc(strlen(name)+1);
			strcpy (cml->com, name);
			cml->ctr = cml->v = make3vect (UNSET, UNSET, UNSET);
			if (top == NULL) top = cml;
			tail = apendml (cml, tail);
		    }
		}


/* Update the pointer to position past end of comment. */
		p = strchr (c, ')');
		if (p == NULL) p = strchr (c, 0);	// End
		else  p++;		// Past the ) found

		if (com != NULL && gdbug > 1) 
			printf (" Comment = \"%s\".\n", com);	// debug
		c = p;
		continue;	// back to start, in case line ended
	    }

// Parse out the next upper case letter (may be this one!)
	    c = ltrparse (c, &ltr);
	    if (ltr == AN) { continue; }

// Parse out any numeric argument following letter
	    c = numparse (c, &v);

// Now report that thing found as a debug message
/*
	    printf (" -- Found command %c %7.5f \n", ltr, v);
.. */

// Here, we have a command letter in ltr, and a numeric argument
// in v.  Good time to run a switch on ltr...
	    m = (int) v;
	    cp = NULL;	// Coordinate pointer for xyz
	    switch (ltr) {
		case 'F':
		    if (gdbug>1) printf (" -- Feed rate %5.2f\n", v);
		    feed = v * scale;
		    break;
		case 'G':
		    gc++;
		    switch (m) {
			case 61:
			if (gdbug>1) printf (" -- Exact stop mode G61\n");
			xact = 1;
			break;
			case 64:
			if (gdbug>1) printf (" -- Cutting Mode G64\n");
			xact = 0;
			break;
			case 20:
			if (gdbug>1) printf (" -- Inch measure G20\n");
			scale = 25.4;	// or Inch
// All coordinates and feed rate are multiplied by scale
// to make the data structure always metric dimensions.
			break;
			case 21:
			if (gdbug>1) printf (" -- Metric measure G21\n");
			scale = 1.0;
			printf (" !! DANGER -- Metric measure file!\n");
			break;
			case 1:
			if (gdbug>1) printf (" -- Linear mode G1\n");
			intrp = 1;
			break;
			case 2:
			if (gdbug>1) printf (" -- Circular CW mode G2\n");
			intrp = 2;
			break;
			case 3:
			if (gdbug>1) printf (" -- Circular CCW mode G3\n");
			intrp = 3;
			break;
			case 90:
			if (gdbug>1) printf (" -- Absolute coordinates G90\n");
			incr = 0;
			break;
			case 91:
			if (gdbug>1) printf (" -- Incremental mode G91\n");
			incr = 1;
			break;
			case 80:
			if (gdbug>1) printf (" -- Parameter value follows\n");
			pv = 1;   // allow p value to change tool mode
			break;
			case 4:
			if (gdbug>1) printf (" -- Dwell G4 (x used)\n");
//			if (gdbug>1) printf (" -- Dwell G4 (x unused)\n");
//			mv = -99;
			gc--;	// Does not count; non-modal
			sintrp = intrp;
			intrp = -2;
			break;
// Work Offsets
			case 52:
			if (gdbug>0) printf (" -- Local Offset used!\n");
			break;
			case 54:
			case 55:
			case 56:
			case 57:
			case 58:
			case 59:
			if (gdbug>1) printf (" -- Work Offset G%2d\n", m);
			break;
// Default - G code not yet decoded
			default:
			printf (" ?? (%d) Unknown G code %d\n", nl, m);
			break;
		    }
//		    if (gdbug>1) printf (" -- G code %d found.\n", m);
		    break;
		case 'M':
		    switch (m) {
			case 4:
			if (gdbug>1) printf (" -- Laser ON (M4).\n");
			if (lastm != NULL) lastm->endl = 1;
			pause += BeamDelay;
			dlsr = 1;
			lsr = 1;
			break;
			case 5:
			if (gdbug>1) printf (" -- Laser OFF (M5).\n");
			if (lastm != NULL) lastm->endl = 1;
			pause += BeamDelay;
			dlsr = 1;
			lsr = 0;
			break;
			case 8:
			if (gdbug>1) printf (" -- Gas ON (M8).\n");
			pause += 0.5;
			break;
			case 9:
			if (gdbug>1) printf (" -- Gas OFF (M9).\n");
			pause += 0.25;
			break;
			case 10:
			if (gdbug>1) printf (" -- Gas is AIR (M10).\n");
			break;
			case 14:
			if (gdbug>1) printf (" -- Shutter OPEN (M14).\n");
			pause += 0.25;
//			lsr = 1;
			break;
			case 15:
			if (gdbug>1) printf (" -- Shutter CLOSED (M15).\n");
			pause += 0.25;
			break;
			case 16:
			if (gdbug>1) printf (" -- Autofocus ON (M16).\n");
			afoc |= 2;
			break;
			case 17:
			if (gdbug>1) printf (" -- Autofocus OFF (M17).\n");
			afoc &= 0XFFFD;
			break;
			case 18:
			if (gdbug>1) printf (" -- Autofocus HOLD (M18).\n");
			afoc |= 1;
			break;
			case 19:
			if (gdbug>1) printf (" -- Autofocus release (M19).\n");
			afoc &= 0XFFFE;
			pause += 0.50;
			break;
			case 30:
			if (gdbug>1) printf (" -- End program (M30).\n");
			break;
			case 0:
			if (gdbug>1) printf (" -- Program Stop (M0).\n");
			break;

			default:
			printf (" ?? (%d) Unknown M code %d\n", nl, m);
			break;
		    }
//		    if (gdbug>1) printf (" -- M code %d found.\n", m);
		    break;
		case 'P':
		    if (pv) par = (int) v;
		    if (gdbug>1) printf (" - Laser Parameter %d.\n", par);
		    pv = 0;
		    break;
		case 'E':
		    printf (" -- (%d) Tool offset %d not supported!\n", nl, m);
		    break;
		case 'T':
		    printf (" -- (%d) Tool offset setting not supported!\n",
				nl);
		    break;
		case 'I':
		    cp = &cx;
		case 'J':		// case statement fallthrough OK
		    if (cp == NULL) cp = &cy;
		case 'K':		// case statement fallthrough OK
		    if (cp == NULL) cp = &cz;
		    *cp = v * scale;
//		    *cp = (intrp == -2) ? v : v * scale;
		    break;
		case 'X':
		    cp = (intrp == -2) ? &pause : &x;
		case 'Y':		// case statement fallthrough OK
		    if (cp == NULL) cp = &y;
		case 'Z':		// case statement fallthrough OK
		    if (cp == NULL) cp = &z;
// Here is common code for all 3 coordinates...
		    if (mv < -20) break;
		    if (incr) {
			if (gdbug>1) printf 
				(" - Inc. %c %7.5f = ", ltr, v*scale);
			*cp += (intrp == -2) ? v : v * scale;
			if (gdbug>1) printf ("%7.5f\n", *cp);
		    } else *cp = (intrp == -2) ? v : v * scale;
		    mv++;
		    break;
		case 'N':
		    if (gdbug > 2) printf ("Line %d.\n", m);
		    if (k > 0) printf (" ** (%d) Line N%05d not first token!\n",
			nl, m);
		    break;
		default:
		    printf (" ?? (%d) Unknown code \"%c\".\n", nl, ltr);
		    break;
	    }	// ltr decode switch ends here


// debug escape if we scan past end of buffer
	    if (++k > 40) break;
	}	// buffer scanning loop here...

// Debug exit for lines of file...
//	if (++j > 200) break;
    }	// buffer reading loop here...

    return  top;
}



	/* - - - - - - - - - - - - - - *\
	|				|
	|      Time Estimate comment	|
	|				|
	\* - - - - - - - - - - - - - - */


double	fill_cuttime (FILE* handle)
/* Enter the cutting time into the file, if it is not already there,
  and return the time in minutes.  The time in the file is in hours. */
{
mlist	*head = NULL;
double	tused;

/* The file designated is already open.  We rewind it, and read it
  (as a .nc file) to find its total time.  If the comment already
  exists, that fact is found in the global time value.  We then
  compute the time using the current global Acon, and report any
  difference if significant.  If there is no comment yet, we create
  one and append it to the file.  If there was no report of difference,
  we report the time and number of segments found also.  */

// rewind the file
    rewind (handle);

// Read the file to obtain the data structure
    ComTime = -1.0;
    head = Read_ncfile (handle, -1, -1);

// Compute the time used from the data structure
    tused = find_times (head);

// Clear the data structure to return dynamic storage
    while (head != NULL) head = freeml (head);

// See if there was already a comment
    if (ComTime > 0.0) {
// Comment already exists, check the amount
	if (fabs(ComTime - tused) < 1.0) return tused;
// Amount is OK, return silently
// Amount is inconsistant, error message and return correct value
	printf ("  Time from .nc file comment  is %.2f minutes.\n", ComTime);
	printf ("  Time from .nc file contents is %.2f minutes.\n", tused);
    }
// No comment yet, create one and append it
    fprintf (handle, "( %s %.2f hours. )\r\n", Legend, tused/60.0);

    printf ("  %s %.3f minutes ( %.2f hours) for %d segments.\n",
	Legend, tused, tused/60.0, seg );

// return the cutting time.
    return  tused;

}

