/* Utility for reading (and timing) .nc file -- Header file */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2006 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/* ---------------------------------------------------------------------- */

#ifndef	INCLUDE_NCRUTIL_H
#define	INCLUDE_NCRUTIL_H

/* INCLUDEs ------------------------------------------------------------- */

#include	<stdlib.h>
#include	<string.h>	// or should it be strings.h ??

#include	MASKDATH
//#include	IOUTILH
//#include	KDCUTILH	/* Definitions */

/* DEFINEs -------------------------------------------------------------- */

#define  AN  '\0'

/* Used to implement unset value; < VSET is unset */
#define  UNSET  -1.1e10
#define  VSET   -1.0e10

#define  Legend  "Estimated cutting time:"

#define  TWOPI  2.0*M_PI

//  #define  Acon  1.0e-6
// Acon is now a global, read from parameter file
/* Acon is constant for accelleration effect, and is a very rough
  estimate of this effect.  Use 0.0 for no acceleration; use
  1.66e-6 for 50 milliseconds at 500 mm/min.  */

#define  BeamDelay  0.005
/* The  BeamDelay  symbol is the laser on/off delay in seconds to be
  added for each laser turn on/off operation.  */


/* TYPEDEFs ------------------------------------------------------------- */

/* The motion list block.. */

typedef struct  mlist {
	struct  mlist  *next, *last;
	int	laser;
	int	afoc;
	int	interp;
	int	par;	// laser cut type
	int	xac;	// exact stop mode
	int	endl;	// end laser on/off
	int	begl;	// begin laser on/off
	double	feed;
	vect3	v;	// absolute location
	vect3	ctr;	// center
	char	*com;
	}  mlist;

/* GLOBALs -------------------------------------------------------------- */

/* Acknowledging that it is bad form, nonetheless these global values
  are assumed by the generalizations in this library and get located
  here.  The variables shoule be called "extern" if this is included
  from a file which is not ncrutil.c  */

#ifdef IS_NCRUTIL_C
#define EXT_NCRUTIL 
#else
#define EXT_NCRUTIL extern
#endif


#define  Ncond  9
EXT_NCRUTIL	double	tx[Ncond+1];
EXT_NCRUTIL	double	dx[Ncond+1];
EXT_NCRUTIL	int	sx[Ncond+1];

#ifdef  IS_NCRUTIL_C
double	ComTime = -1.0;
double	Acon = 0.0;
char	ex[Ncond+1][20] = { 
	"Slow cutting",		"Fast cutting", 
	"Rough Cutting",	"Engraving",
	"Off, slow",		"Off, fast", 
	"Vertical",		"Pause",
	"Total:",		"" };
int	seg = 0;
#else
extern	double	ComTime;
extern	double	Acon;
extern	char	ex[Ncond+1][20];
extern	int	seg;
#endif


/* FUNCTION PROTOTYPEs -------------------------------------------------- */

/* Commented entries have been declared static (local) in the ncrutil.c
  package code.  If they are needed, remove the static declaration, and
  remove the comment here... */

/*  ===  Parsing Subroutines  ===  */
// char*	ltrparse (char* buf, char* ltr);
// char*	numparse (char* buf, double* v);
// char*	getcomment (char* c);

/*  ====  Linked Storage support utilities  ====  */
// mlist*	newmlist (int l, int a, int i, int e, double f);
// mlist*	apendml (mlist* p, mlist* a);
// mlist*	popml (mlist* a);
mlist*	freeml (mlist* a);

/* --- Global timing tables derived --- */
double	find_times (mlist* head);

/* --- Reading program --- */
mlist*	Read_ncfile (FILE* handle, int dbug, int gdbug);

double	fill_cuttime (FILE* handle);


/* ---------------------------------------------------------------------- */
/* End of ncrutil.h package */

#endif	/* INCLUDE_NCRUTIL_H */

/* ---------------------------------------------------------------------- */
