/*  skymap.c  ==  Mapping functions */

/* Copied from map.c from gmap project.. */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2011 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 
#define IS_SKYMAP_C
 
/* DEFINEs -------------------------------------------------------------- */
 
/* #define DEBUG */

#undef  TIMEDEBUG
//#define  TIMEDEBUG
/*  Sumbol  TIMEDEBUG  is defined to include time debugging feature
  and report on syserr the elapsed times for service.  Normally,
  this will be undef.  Only used for test and debug.  */

#ifdef  TIMEDEBUG
#include	"timepak.h"
#define  Rflags TR_titlr|TR_elap|TR_cput|TR_count|TR_long
// The timepak is included by mgutils.c
#endif


#define MAPSIZE_STEP  256

#undef  DBredraw
//#define  DBredraw
/* Symbol DBredraw does some tracing of map_redraw execution */

#undef  DEBUGproj
//#define  DEBUGproj
/*  Symbol  DEBUGproj  is defined here to add debug code to test
  the projd program values...  Normally undefined.     */

/* Definition of CACHECAT moved to mgfeats.h */
// Move the above to mgfeats.h and if defined require ALLMSTAR to also
// be defined, as we expect to cache all the stars in the catalog.
// The ALLMSTAR symbol is used in skywin.c, intgui.c also.


/*  Symbol  LOCALSERVER  is obsolete.  To achieve what this formerly
  did, use "setenv CATSERVER localhost".  Used for testing of a new
  server or special enhancements.  There are other environment variables;
  See also CATPORT, CATPATH, CATMETHOD, CATCACHE.  (And other getenv)
  CATPORT  55355  (or 0xd83b)
  CATPATH  /data1/moon/gcats (or /dev/null) -- used for NOCATS only
  CATMETHOD  GetStars (slow way is PullStars)
  CATCACHE  /dev/null or /home/clardy/imacs/maskgen/cache or unsetenv ...
  CATSERVER  localhost or catxml.obs.carnegiescience.edu  */



/* INCLUDEs ------------------------------------------------------------- */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

// #ifdef  sun
// #include <sunmath.h>	/* nint */
// #endif
#include "cxt.h"

#include  MASKDATH
//#include  MASKDATH
#ifdef  AUTOGS
#include  MGUTILH
#endif
/* We NEED the system includes in MGUTILH to come bdfore the local includes,
  specifically the intgui.h.  From the MGUTILH include we get the library
  <openssl/evp.h> which includes <openssl/bn.h> in which some function
  prototypes use "Xp" as a variable name, but "Xp" is defined in intgui.h.  */

#ifdef  CACHECAT
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>	// for access
#include <utime.h>    // for utime
#endif

#include "skymap.h"
#include "utils.h"
#include "intgui.h"		// Define Obf structure, get_cp
// #include  "kdcutil.h"

#include  "skywin.h"

#ifdef  EXCAT
#include <xmlrpc-c/base.h>
#include <xmlrpc-c/client.h>
#endif

/* EXTERNs -------------------------------------------------------------- */
 
//#ifdef  UPOPTION
//extern double tellat, tellong;
//extern time_t  ut_now;
//#endif

// Define these in case of upoption or uparrow symbols.
#ifdef  UPARO
extern double tellat, tellong;
#endif

extern Application *app;
extern MainWindow  mwin;

extern  double	equinox;

// extern  Obs  *obs;
extern  Obf  *Gob;

extern	int	TESTMODE;

/* STATICs -------------------------------------------------------------- */
 
static GC      gc_blue,gc_green,gc_yellow,gc_red,gc_brown,gc_white,gc_black;
static GC      gc_orange,gc_grey;
static GC      map_gc;
static Bool    map_graph=False;

#ifdef  EXCAT
// static Bool    USEXMLRPC=False;  // was =True;
int	USEXMLRPC=0;
#endif

#ifdef  CACHECAT
int	crep=0;
#endif


/* function prototype(s) */
 
static void    map_sort(CATstar*,int,int);
static double  map_al(double);
static void  map_coords (int*, int*, vect3);
static void  map_vector(Display*,Window,GC,long,long,long,double);
static void  draw_circle (Display*, Window, GC, int, int, int);
#ifdef  UPARO
static double  par_angle (double ha, double dec, double lat);
#endif
//extern int	check_star (int star);
// The above is needed as we apparently do not include skywin.h here.

/* ---------------------------------------------------------------------- */

/*  ====  Some extra utilities  ====  */

#if defined DEBUGproj
static	void	runrms (double x, double* rms, int* n)
#else
void	runrms (double x, double* rms, int* n)
#endif
/*  Maintain a running rms average (*rms) of successive values of
  the given x.  Also maintain a count (*n) of averaged values.
  Usage:  Start *n off at 0, call for each x to be averaged.  The
  value of *rms is maintained at each call.    */
{
double	s = *rms;
    s  = s*s;
    s *= *n;	/* Recovers sum squares (rms = sqrt(sumsq/n) */
    s += x*x;	/*  New sum squares */
    ++(*n);	/* Update count */
    *rms = sqrt (s / *n);	/*  New rms average */
}
/*  runrms above is used in debugging the "projd" function.  Also for
  general averaging it can be placed elsewhere.  */


void	runavg (double x, double* avg, int* n)
/*  Maintain a running average (*avg) of successive values of
  the given x.  Also maintain a count (*n) of averaged values.
  Usage:  Start *n off at 0, call for each x to be averaged.  The
  value of *avg is maintained at each call.    */
{
double	s = *avg;

    s *= *n;	/* Recovers sum squares (rms = sqrt(sumsq/n) */
    s += x;	/* Add in value to average */
    ++(*n);	/* Update count */
    *avg = (s / *n);
}
/*  runavg above is used in debugging the "projd" function.  Also for
  general averaging it can be placed elsewhere.  */



/* ---------------------------------------------------------------------- */

#ifdef  EXCAT
#define XNAME "Xmlrpc-c Interface Client"
#define XVERSION "1.2"
extern  int	skywin_active;   // Really defined in skywin.c

static void die_if_fault_occurred (xmlrpc_env *env) {
    if (env->fault_occurred) {
        fprintf(stderr, "XML-RPC Fault: %s (%d)\n",
                env->fault_string, env->fault_code);
//        exit(1);
/* Too drastic; just cut off the skywin activity */
	fprintf (stderr, "  ** Internet catalog service not available. **\n");
	skywin_active = 0;
	USEXMLRPC = 1;   // Tested and found wanting
    }
}
#endif	// on EXCAT

// void map_init(Display* display,double scale,Bool graph)
void map_init(MainWindow* mw,double scale,Bool graph)
{
Display  *display;
  Window   win;
//  GfC	map_gfc;
 
  maplist       = NULL;
  map_nstars    = 0;
  mapsize       = 0;
  map_scale     = scale;
  map_sel       = -1;
  map_w = map_h = 0;
  map_pcx = map_pcy = 0.0;

 
  if (graph) {
//    win    =  DefaultRootWindow(display);
	display = mw->disp;
	win = mw->win;
    gc_blue = XCreateGC(display,win,0,0);       /* special GCs to speedup */
    XSetBackground(display,gc_blue,app->lgrey);  /* star-map redraw        */
    XSetForeground(display,gc_blue,app->blue);
    gc_green = XCreateGC(display,win,0,0);
    XSetBackground(display,gc_green,app->lgrey);
    XSetForeground(display,gc_green,app->green);
    gc_yellow = XCreateGC(display,win,0,0);
    XSetBackground(display,gc_yellow,app->lgrey);
    XSetForeground(display,gc_yellow,app->yellow);
    gc_orange = XCreateGC(display,win,0,0);
    XSetBackground(display,gc_orange,app->lgrey);
    XSetForeground(display,gc_orange,app->orange);
    gc_red = XCreateGC(display,win,0,0);
    XSetBackground(display,gc_red,app->lgrey);
    XSetForeground(display,gc_red,app->red);
    gc_brown = XCreateGC(display,win,0,0);
    XSetBackground(display,gc_brown,app->lgrey);
    XSetForeground(display,gc_brown,app->brown);
    gc_white = XCreateGC(display,win,0,0);
    XSetBackground(display,gc_white,app->lgrey);
    XSetForeground(display,gc_white,app->white);
    gc_black = XCreateGC(display,win,0,0);
    XSetBackground(display,gc_black,app->lgrey);
    XSetForeground(display,gc_black,app->black);
    gc_grey = XCreateGC(display,win,0,0);
    XSetBackground(display,gc_grey,app->black);
    XSetForeground(display,gc_grey,app->lgrey);
//    map_gc = CBX_CreateGC(app->black,app->white,"courier","medium",12);
//    map_gfc = CBX_CreateGfC(&mwin,"courier","medium",12);
//    map_gc = map_gfc.gc;
	map_gc = (CBX_CreateGfC(mw,"courier","medium",12)).gc;
    map_graph = True;
// dummy statement to avoid warning messages...
if (cat_equ[0] == cat_mag[0]) ;
  }
}
 
/* -------------------------------------------------------------- */
 
void map_free(Display* display)
/*  This is actually not called.  */
{
  if (maplist) { free((void*)maplist); maplist=NULL; }
 
  map_nstars = mapsize = 0;
 
  if (map_graph) {
    XFreeGC(display,gc_white);
    XFreeGC(display,gc_black);
    XFreeGC(display,gc_green);
    XFreeGC(display,gc_blue);
    XFreeGC(display,gc_yellow);
    XFreeGC(display,gc_brown);
    XFreeGC(display,gc_red);
    XFreeGC(display,gc_orange);
    XFreeGC(display,gc_grey);
    XFreeGC(display,map_gc);
  }
}

/* ---------------------------------------------------------------------- */

void draw_circle (Display* display, Window win, GC gc,
		int x, int y, int rs)
{
int  r;
  r = rs + 1;
  XDrawArc (display, win, gc, x-r, y-r, 2*r, 2*r, 0, 23040);
/* 23040 = 360 * 64; full circle */

}

 
/* ---------------------------------------------------------------------- */
 
//void map_circle_star(Display* display,Window win,int star,float mlim)
static	void map_circle_star(Display* display,Window win,
	u_long color,int star,float mlim)
{
//  int  r;
  int x, y;

  if (star < 0) return;                          /* star = MAPlist number */

//  XSetForeground(display,map_gc,app->white);
  XSetForeground(display,map_gc,color);

  map_coords (&x, &y, maplist[star].v);
  draw_circle (display, win, map_gc, x, y,
	imax(map_get_radius(maplist[star].mag,mlim),1) );

  XSetForeground(display,map_gc,app->black);
}
 
/* ---------------------------------------------------------------------- */

int map_get_closest_star (int x, int y)
{
/* Find closest star to screen coordinate given; use vector
  projection to obtain 2-vectors to compare. */
int    i,sel=-1;
vect2	p;
vect2	d;
double	dq, dmin=9.0e+32;

// Obtain projection vector in arc seconds (similar to xyproj, from map_ad)
//    p.x = map_pcx + (double) (x - map_w/2.0) * map_scale;
//    p.y = map_pcy - (double) (y - map_h/2.0) * map_scale;
    p = map_fpxy (x, y);

// Loop over all stars
    for (i=0; i<map_nstars; i++) {
	d = sub2vect (xyproj (maplist[i].v), p);
	dq = dot2vect (d, d);
	if (dq < dmin) {
	    dmin = dq;
	    sel = i;
	}	// Select new minimum
    }	// loop over stars
    return  sel;
}

double	map_dns (int star)
/* Find the distance to the nearest (other) star, given the
  subscript of a subject star.  */
{
int	i;
//int	sel=-1;
vect2	p;
vect2	d;
double	dq, dmin=9.0e+32;

    if (map_nstars < 1) return 0.0;

// Obtain the projected vector for the subject star.
    p = xyproj (maplist[star].v);

// Loop over all stars
    for (i=0; i<map_nstars; i++) {
	if (i == star) continue;	// only for "other" stars
	d = sub2vect (xyproj (maplist[i].v), p);
	dq = dot2vect (d, d);
	if (dq < dmin) {
	    dmin = dq;
//	    sel = i;
	}	// Select new minimum
    }	// loop over stars
    return  sqrt(dmin);
}

/* -------------------------------------------------------------- */

static	double	projd (double a0, double d0, double a1, double d1)
/*  Find the plane projected distance in arc seconds from projection
  center location a0 (hours), d0 (degrees) to star at a1 (hours),
  d1 (degrees).  Projection similar to xyproj geometry, projection
  from a virtual center at the point on the unit sphere oposite to
  the projection center a0,d0.  We do a lot of trig here.    */
{
double	da;	// diff. alpha
double	s0, c0;	// sin, cos of d0
double	s1, c1;	// sin, cos of d1
double	x, y, z;  // vector being projected
double	g, h;	// scratch

// Get the needed sine and cosine values
    da = (a1 - a0) * Hour;	// a increases to left, as does y
    d0 *= Degree;		// scale the declinations
    d1 *= Degree;		// to cut down on multiplies
    s0 = sin (d0);
    c0 = cos (d0);
    s1 = sin (d1);
    c1 = cos (d1);

// Form the position vector at equator and a0 for object 1
    g = c1 * cos (da);		// x' is forward
    y = c1 * sin (da);		// y  is left
    h = s1;			// z' is up

// rotate that "vector" to d0 center.  This is tangent plane vector.
    x = g*c0 + h*s0;	// Forward
    z = h*c0 - g*s0;	// Up

// Get radial component of tangent plane (z,y) vector in g,
// and the scaled projection height in h
    g = hypot (z, y);		// radius in tangent plane
    h = ArcSecond * 0.5 * (1.0 + x);

    return  g/h;
}

/* -------------------------------------------------------------- */

vect2	xyproj (vect3 v)
{
/* Compute plane vector in arcseconds, right and up */
/* Use the projection from back of sphere */
vect3	p;
vect2	result;
double	x, y;
double	f;
//    p = mmult3 (map_xf, v);  // Tangent plane vector
/* SHOULD use {Gob->}obs->mxpos instead of map_xf */
/* Also SHOULD put this or a similar thing into mgutils.c, just before
  the xplane routine (line 3249) and use that both here and in xplane */
    p = mmul3v (map_xf, v);  // Tangent plane vector
    f = 0.5*(1.0 + p.x);
    x = -p.y / f;
    y =  p.z / f;
    result.x = x / ArcSecond;
    result.y = y / ArcSecond;
    return result;
}

/* -------------------------------------------------------------- */

vect3	unproj (vect2 v)
{
double	p;
vect3	a;
vect3	result;
/* From a plane vector in arc seconds, right and up on view,
  obtain the spherical vector in the sky of that point.
  Inverse of xyproj function above.  */
    a.y = -v.x * ArcSecond;
    a.z =  v.y * ArcSecond;
    p = hypot (a.y, a.z) / 2.0;
    a.x = 1.0 - p*p;
//    result = norm3vect (tmult3(map_xf, a));
/* SHOULD use {Gob->}obs->mxpos instead of map_xf */
    result = norm3vect (mmul3v(mtr3(map_xf), a));
    return result;
}

/* -------------------------------------------------------------- */
void	map_pltxy (int *x, int *y, vect2 v)
{
    *x = Nint (map_w/2.0 + (v.x - map_pcx)/map_scale);
    *y = Nint (map_h/2.0 - (v.y - map_pcy)/map_scale);
}

/* -------------------------------------------------------------- */
vect2	map_fpxy (int x, int y)
{
vect2	p;	// Focal plane coordinates, from plot coordinates
    p.x = map_pcx + ((double)x - (double)map_w/2.0) * map_scale;
    p.y = map_pcy - ((double)y - (double)map_h/2.0) * map_scale;
    return  p;
}

/* -------------------------------------------------------------- */
void  map_coords (int *x, int *y, vect3 v)
{
vect2	p;	// projection vector
    p = xyproj (v);
//    *x = Nint (map_w/2.0 + p.x/map_scale);
//    *y = Nint (map_h/2.0 - p.y/map_scale);
    map_pltxy (x, y, p);
}

/* -------------------------------------------------------------- */

// void  map_ad (double* a, double* d, int x, int y)
// /* Inverse map_coords from projection */

/* ---------------------------------------------------------------------- */

static double map_al(double al)
{
  double da;

  da = (map_ca-al);
  if      (da >  12.) da = -24. + da;
  else if (da < -12.) da =  24. + da;

  return(da);
}

/* ---------------------------------------------------------------------- */
 
long map_get_x_coord(double al)
{
//  return(map_w/2 + (long)(3600.*15.*map_al(al)*cos(map_cd*Degree)/map_scale));
  return(map_w/2 + (long)((3600.*15.*map_al(al)*cos(map_cd*Degree)-map_pcx)/map_scale));
}

/* -------------------------------------------------------------- */
 
long map_get_y_coord(double de)
{
//  return(map_h/2 + (long)(3600.*(map_cd-de)/map_scale));
  return(map_h/2 + (long)((3600.*(map_cd-de)+map_pcy)/map_scale));
}

/* -------------------------------------------------------------- */
 
int map_get_radius(float v,float mlim)
{
int  r;
 
//  r = (int)((mlim-v)*2.);

// Change to make stars a little bigger, not too big.
    r = (int)(1.0 + (mlim-v)*1.8);
 
    return(r);
}
 
/* ---------------------------------------------------------------------- */

//#define  SHOSERV
/*  Symbol  SHOSERV  implements some debug programs, and a section
  in the map_getlist program which dumps details about a single
  star for decoding the data format, perhaps.  Undefine it to
  turn off all that extra compiling and such.  */

/* See old version, skymap.c18 or skymap.c.bak for the SHOSERV things,
  which were obtained from the hex dump utility.  */

/* Special debug program - hex dump of an array */


/* ---------------------------------------------------------------------- */
static  int  nsmag (float mlim, int nmap)
/* Returns index of star (zero based) in the global maplist array which
  is the last one not fainter than mlim.  The nmap is the number of
  entries in maplist (largest subscript is nmap-1).  Use a binary search
  algorithm to find the number. 
  Assumes that array is sorted by magnitude.  */
//	nmap, maplist[nmap-1].mag);
{
int	a, b;	// limit index
int	x;	// trial index
    if (nmap < 2) return 0;
    a = 0;
    if (!(maplist[a].mag <= mlim)) return a;
    b = nmap - 1;
    if (!(mlim < maplist[b].mag)) return b;
    while (a < b) {
	x = a + ( (b-a)/2 );
	if (x == a) return x;
	if (mlim < maplist[x].mag) {
	    b = x;
	} else {
	    a = x;
	}
    }
    return  a;
}


	/* - - - - - - - - - - - - - - *\
	|				|
	|    Get stars from Catalog	|
	|				|
	\* - - - - - - - - - - - - - - */


#undef  UPCACHEREAD
//#define  UPCACHEREAD
/*  Symbol  UPCACHEREAD  is defined to implement the feature of updating the
  file modification/access times for any cache file which is actually
  used.  Then the "oldest" file is actually the least recently used rather
  than the earliest created, and thus more suitable for replacement should
  that be necessary.  This is local to the get_cache and put_cache routines.
  --  This symbol should normaly be undefined  --

  When undefined, the inode time is updated using chmod, and that time is
  used, effectively implementing the least recently used algorithm, but not
  changing the creation time displayed in a directory listing.
  Using "ls -c" will sort by and show the ctime rather than mtime, and
  "ls -u" will sort and show by access time.  See the commands dx, dxa, dxc,
  dxd, dxda, dxdc, lr, lra, lrc; and others in ~dxcom.      */

/* Define of ExtraField was moved to mgfeats.h for coordination.  */
/* -- It's back! -- */
#define  ExtraField  0.15
// was 0.1
/*  Symbol  ExtraField  is the amount by which the radius of the area
  of the star catalog read is enhanced over the diagonal of the
  display window.  */

#define  HeaderForm   "%d %10.7f %10.6f %8.3f %5d\n"
#define  HeaderFormI  "%d  %lf    %lf    %lf   %d"
#define  CL_header  32
// HeaderForm contains cat, alpha, delta, frad, stars
#define  CacheForm  "%10lu %10.7f %10.6f %5.2f %.8s\n"
#define  CacheFormI "%lu   %lf    %lf    %lf   %s\n"
#define  CL_star  44
// The header and star lengths are just estimates now
// CacheForm contains id, alpha, delta, mag, spc
#define  CacheBUFL 127

/*  Here we introduce on CACHECAT symbol two programs:
int  get_cache (a,d,r);  and void put_cache (a,d,r,n).
Both use the global cache control structure.  The maplist may
be enhanced to hold the double magnitude integer data if needed;
or the CATstar structure could be extended if needed.
... */

#ifdef  CACHECAT

/* Status bits returned by get_cachendx */
#define  GCX_nofile  (1 <<  0)	// file could not be opened
#define  GCX_badhdr  (1 <<  1)	// file has bad header format
#define  GCX_nochdr  (1 <<  2)	// cache directory not readable/initialized
#define  GCX_badfil  (1 <<  3)	// file not readable
#define  GCX_badptr  (1 <<  4)	// cache pointer not valid
#define  GCX_badidx  (1 <<  5)	// index out of range

static	int	get_cachendx (int i, cachecon* C)
/*  Try to read index file i header into cache *C, returning a status value
  indicating any errors found.  Zero indicates no problems, 
  error bits are defined above.
  The cache must have been previously initialized.
  When 0 is returned, the header information is placed in *C, and any
  previous data there is grecefully removed.    */
{
cacheindx	X;	// Local index entry
FILE	*F=NULL;
char	*filename;
char	fnbuf[24];
int	r = 0;
int	j;
struct stat  sbuf;
int	stars=0;
char	headbuf[CacheBUFL+1];

/*  Go through the trivial checks for basic things.  */
    if (C == NULL) return GCX_badptr;	// 32;
    if (access (C->directory, R_OK)) return GCX_nochdr; // May want to set C->status ?
    if (i < 0 || i >= Ncache) return GCX_badidx;  // 64;

//    sprintf (fnbuf, FNform, i);
    sprintf (fnbuf, FNform, exchar(i));
    filename = fncat (C->directory, fnbuf);   // free this later
    if (access (filename, R_OK)) r = GCX_badfil;  //16;
    if (r == 0) F = xopen (filename, "r");
    else        F = NULL;
    if (F == NULL) r |= GCX_nofile;  // 1;
    if (r) goto errorexit;

    X.status = CI_exist;
    fgets (headbuf, CacheBUFL, F);   // get header
    j = sscanf (headbuf, HeaderFormI, &(X.cat),
	&(X.alpha), &(X.delta), &(X.frad), &(X.stars) );

/* Below was older way to try to get data (not so good) */
//    j = fscanf (F, HeaderFormI, &(X.cat),
//	&(X.alpha), &(X.delta), &(X.frad), &(X.stars) );

    if (j >= 4) X.status |= CI_good;
    else  {
	r |= GCX_badhdr;  // 2;  // bad header format
	if (dodebug > 0) printf (" *= Header of %s has only %d values.\n",
			filename, j);
    }
    if (j >= 5) stars = X.stars;  // some early data had no star count
    iclose (F);   F = NULL;
    if (r == 0) X.filename = filename;
    else { free(X.filename); X.filename = NULL; }

    j = stat (filename, &sbuf);
#ifdef  UPCACHEREAD
    X.date = sbuf.st_mtime;   // modification time
#else
    X.date = sbuf.st_ctime;   // inode change time
#endif
//    X.date = sbuf.st_atime;   // access time (do not use this)
    if (stars == 0) {  // Guess star count if not in header
	stars = (sbuf.st_size - CL_header) / CL_star;
	X.stars = stars;
    }
    if (dodebug > 1) printf (" *= Cache %s has about %d stars.\n",
			filename, stars);
    if (r) goto errorexit;
    free (C->ndx[i].filename);  // Prevent memory leak here
    C->ndx[i] = X;	// Store derived Cache index entry
    return  r;

errorexit:
    free (filename);
    if (F) iclose (F);
    return  r;
}

static	void	multline (int* a, int* b, char* m1, char* m2, int w)
/*  Multi line support program -- could be generalized.
  Write either "%d: m1" or "%d - %d: m2", depending on *a, *b, w
  and update *a and *b.  w is 1 to write, 0 to accumulate.
  For the w == 0 case:
    If b < 0, set a = b, return.
    If a < 0, set a = b, return.
    Otherwise, return.
  For the w == 1 case:
    If b < 0, set a = b, return.
    If a == b, write m1 and set b=a=-1, return
    If a < b, write m2, set b=a=-1, return
    Otherwise write m1 (b), set b=a=-1, return  
  */
/*  Usage:
  Start with b = -1 and a = -1.
  When adding a line to be written, set b = x and call with w = 0.
  Before writing a different line, call with w = 1.
  At end, call with w = 1.
 */
{
    if (*b < 0) goto out2;
    if (w == 0) {	// Save, but don't write yet.
	if (*a < 0) goto out2;
	return;
    } else {
	if (*a < *b && *a >= 0) {
	    printf ("%2d - %d: %s\n", *a, *b, m2);
	} else {
	    printf ("%2d: %s\n", *b, m1);
	}
    }
// out1:
    *b = -1;
out2:
    *a = *b;
}

static	int	get_cache (double alpha, double delta, double rad)
/*  Try to find a cache locally which provides data for the current
  catalog at the given alpha delta position of radius rad.  If found,
  use the cache data to fill the maplist array and return the number
  of stars found.  If not found, return zero.  Implement the usual rules
  for forming the cache disk name and check for disk files.  Put these
  values into the local cache index, and use the cache index if it exists.
  When getting individual stars, generate v vector from alpha, delta
  values.  Also mag/spc from double magnitude data if so formatted.
  ..    */
{
double	dist;
char	*dptr;		// Directory string pointer
char	bufr[CacheBUFL+1];
unsigned long id;
double	ai, di, mi;
char	sp[10];
int	i;		// scratch index
int	j;		// scratch
int	n=0;		// number of stars returned
int	k=-1;		// cache index active
int	ni=0;		// debug number in radius
FILE	*F=NULL;
int	old=-1;
time_t	oldage;
char	*fnam = NULL;
#ifndef  UPCACHEREAD
struct  stat	sbuf;
#endif

    if (dodebug > 0) printf (" *= get_cache initial status is %d.\n", Cache.status);

/* First, initialize the cache index if it has not already been done.  */
    if ((Cache.status & CS_init) == 0) {	// Need to initialize

/* Get the directory, verify that it is proper.  If not, set skip status
  and return without loading index.  Initialize directory name.      */
	dptr = getenv (CacheDir);	// defined as "CATCACHE"
	Cache.directory = strdup (dptr ? dptr : "./");
	if (dodebug > 2) printf (" *= Cache directory \"%s\".\n",
				Cache.directory);
/* Check the directory.  If read/write it should be good.  Otherwise, it
  is bad and we skip caching.  The "/dev/null" is special and indicates
  no caching is to be attempted.  Even if it is attempted, it would not
  work, but the index would get inproperly filled with data which was
  not written, wasting effort.    */
	if (access (Cache.directory, R_OK + W_OK) ) {	// problem exists
	    if (crep || dodebug > 0) 
		printf (" *= Cache directory \"%s\" NOT read/write.\n",
				Cache.directory);
	    Cache.status |= (CS_init | CS_skip);   // Directory disallowed
	    return  0;
	} else {	// Cache directory permits read and write
	    if (dodebug > 0) 
		printf (" *= Cache directory \"%s\" is read/write.\n",
			Cache.directory);
/* Seems that /dev/null checks out OK, so we need to test for it directly. */
	    if (!strcmp (Cache.directory, "/dev/null")) {
		if (dodebug >= 0)
		    printf (" *= Cache directory \"%s\" inhibits caching.\n",
			Cache.directory);
		Cache.status |= (CS_init | CS_skip);
		return  0;
	    }
	}
	if (crep || dodebug > 1) printf (" *= Using Cache directory \"%s\".\n",
			Cache.directory);

/* Run a loop 0 to Ncache, and within the loop look for properly named
  cache files within the cache directory.  Set the status for each one
  found.  Open such files and read the header; fill in the data.  If the
  header decodes correctly, set status.  */

	for (i=0; i<Ncache; i++) {	// File scan loop
	    j = get_cachendx (i, &Cache);
	    if (j && dodebug>1) printf (" *= Cache %d status %d.\n", i, j);

// find the oldest for debug report
	    if (Cache.ndx[i].status & CI_good &&
		(old < 0 || (old >= 0 && Cache.ndx[i].date < oldage)) ) {
		old = i;
		oldage = Cache.ndx[i].date;
		if (dodebug > 1) printf (" *= Cache %d date %ld\n", old, oldage);
	    }
	}   // loop over possible index values

	Cache.status |= CS_init;	// Initialized
	if (dodebug > 0) printf (" *= Oldest cache is %d.\n", old);

    } else {  // already initialized
	if (dodebug > 0) printf (" *= Initialized cache directory is \"%s\".\n",
			Cache.directory);
	if (Cache.status & CS_skip) {
	    if (dodebug >= 0) printf (" *= No caching input.\n");
	    return 0;
	}
    }  // Initialization of Cache section

    if (Cache.status & CS_skip) {   // see if we skip the cache
	if (dodebug >= 0) printf (" *= Skip cache input.\n");
	return  0;
    }
/* At this point, the Cache structure has been filled with cache index
  data, and it is ready to go.    */

/*  Special section, reporting on cache content  */
    if (crep && alpha <= -1.0) {
char	abuf[32];
char	dbuf[32];
int	m = -1;
int	n = -1;
char	s[] = "has no entry.";
char	q[] = "no entry.";
	for (i=0; i<Ncache; i++) {
//	    printf (" %d:", i);
	    if (Cache.ndx[i].status & CI_exist) {  // Entry exists
		multline (&m, &n, s, q, 1);
		printf ("%2d:", i);
		if (Cache.ndx[i].filename)  printf (" %s", Cache.ndx[i].filename);
		else  printf (" (no filename)");
		if (Cache.ndx[i].status & CI_good) {
		    printf (" (good stat)");
		    printf (" Cat: %s\n", cat_name(Cache.ndx[i].cat) );
		} else  printf (" (bad stat)\n");
		sexigw (abuf, Cache.ndx[i].alpha,   3, 3);
		sexigw (dbuf, Cache.ndx[i].delta, -3, 2);  // force sign
		printf ("   : R.A. Dec. Rad: %s %s  %.2f", 
			abuf, dbuf, Cache.ndx[i].frad);
		printf (", %d stars.\n", Cache.ndx[i].stars);
// Here's where we add date print if possible
	    } else {  // No entry
		if (Cache.ndx[i].filename) {
		    multline (&m, &n, s, q, 1);
		    printf ("%2d: %s does not exist.\n", 
			i, Cache.ndx[i].filename);
		} else {
		    n = i;
		    multline (&m, &n, s, q, 0);
//		    printf (" has no entry.");
		}
//		printf ("\n");
	    }  // conditional on entry existing
	}  // for loop (i) on index array
	multline (&m, &n, s, q, 1);
	crep = 0;
	return -1;   // Indicating report has been done
    }  // conditional on cache report

/*  Loop to look through cache index to find a match.  If not found, return
  a 0 count.  If found, we will read it.  */

    for (i=0; i<Ncache; i++) {
	if ((Cache.ndx[i].status & CI_exist) == 0) continue;
	if ((Cache.ndx[i].status & CI_good) == 0) continue;
	if (catalog != Cache.ndx[i].cat) continue;
	dist = projd (alpha, delta, Cache.ndx[i].alpha, Cache.ndx[i].delta)
		+ rad;
/* Now called with r1, and cache has r2 radius (r1 < r2) will allow the
  cache data to be used.    */
//		+ rad / (1.0 + ExtraField);
/* dist is the needed cached field radius from cached field center */
	if (dist > Cache.ndx[i].frad) continue;
/* If we get past all those rejection possibilities, we have a good center */
	k = i;
	break;
    }

    if (k >= 0 && dodebug > 1) printf (" *= Cache file %d being used.\n", k);
// may want dodebug >= 0 above when testing.

/* Get the file name, open the file, read header data, and then read all
  the star data into the maplist array.  Be sure to extend maplist when
  needed.  Return number of stars filled into maplist.  */

    if (k < 0) {	// Did not find a cache file
	if (dodebug > 0) printf (" *= No valid cache file found.\n");
	return  0;
    }

/*  ====  Load data from cache k  ====  */

/* Try to anticipate maplist size from number of stars in header which
  has been cached already.  */
    if (mapsize <= Cache.ndx[k].stars) {
	mapsize = MAPSIZE_STEP * (1 + (Cache.ndx[k].stars / MAPSIZE_STEP) );
	maplist = (CATstar*)realloc(maplist, mapsize*sizeof(CATstar));
    }

/* Open file based on cached file name  */
    fnam = Cache.ndx[k].filename;
//    printf (" Using %s\n", fnam);		// Display cache used
    if (TESTMODE) printf (" Star data from %s\n", fnam);
    else     printf (" Star data obtained from local cache.\n");
    F = xopen (fnam, "r");
    fgets (bufr, CacheBUFL, F);   // get past header
    while (fgets (bufr, CacheBUFL, F) != NULL) {
	trimend (bufr, NEWLINES);
	j = sscanf (bufr, CacheFormI, &id, &ai, &di, &mi, sp);
	if (j < 4) continue;
	maplist[n].id = id;
	maplist[n].alpha = ai;
	maplist[n].delta = di;
/* If we change the format of the cache, or make it dependent on catalog,
  these lines following might need to be changed too.    */
	maplist[n].mag = mi;
	if (j >= 5) strncpy (maplist[n].spc, sp, 8);
//	maplist[n].v = sph2vec (ai*Hour, di*Degree);
	maplist[n].v = had2vec (ai, di);

	if (projd(alpha,delta, ai,di) <= rad) ni++;	// count stars in rad

        n++;
        if (n == mapsize) {           /* increase maplist */
          mapsize += MAPSIZE_STEP;  // should be a global
          maplist = (CATstar*)realloc(maplist, mapsize*sizeof(CATstar));
	}  // End of maplist expand section


    }  // End of read loop

/* Possible message if stars in header don't match stars in file. */
    if (dodebug >= 0 && n != Cache.ndx[k].stars) {
	printf (" *= Header gives %d stars, cache file has %d stars.\n",
		Cache.ndx[k].stars, n);
    }

/* Update the time (touch) the file used */
#ifdef  UPCACHEREAD
    utime (fnam, NULL);  // set st_mtime to current time
#else
// Change inode; write mode to what it already was; updates st_ctime
    stat  (fnam, &sbuf);
    chmod (fnam, sbuf.st_mode);  // Sets st_ctime to now
#endif
/* We note that since all cache files have their header read by the indexing
  section, all the st_atime values (access time) will be set to the current
  time whenever we look at the cache.  Thus atime is not a valid method of
  judging which cache file is "oldest" or least recently used.      */

/* Clean up the open file */
    iclose (F);

/* Debug report of stars within radius. */
    if (dodebug > 0) {
	printf (" *= get_cache found %d of %d stars (%.4f) \n", 
			ni, n, (float)ni/(float)n);
	printf (" *-  within %.2f radius of %.2f cached.\n",
			rad, Cache.ndx[k].frad );
    }

/* All done now.  */
    return  n;
}  // end of get_cache


static	void	put_cache (double alpha, double delta, double rad, int n)
/*  Presuming that existing maplist data was not obtained from cache, we
  want to write the data from the maplist to a cache file.  Check the
  index (which exists as get_cache was called earlier) to find an appropriate
  slot in which to write the data.  Delete any existing data in that slot, and
  update the index and write the current data.
  If existing data is a super-set of an existing file, replace it.
  Otherwise put in first unused slot, otherwise the oldest (least
  recently used).        */
{
int	i;	// index
int	j;	// star index
int	fb=-1;	// first blank
int	old=-1;
time_t	oldage;
int	sub=-1;	// subset
double	dist;
char	fnbuf[24];
FILE	*F;

    if (dodebug > 0)
	printf (" *= Put_Cache called (status %d).\n", Cache.status);

    if ( (Cache.status & CS_init) == 0) {   // Not initialized (impossible?)
	printf (" *= Caching not initialized!  (Should not happen)\n");
	return;
    }
    if (n < 1) {  // number of stars found
	printf (" *= No stars to write to cache. (Should not happen)\n");
	return;
    }
    if (Cache.status & CS_skip) {   // No caching
	if (dodebug >= 0) printf (" *= No Caching output.\n");
	return;
    }

/*  NOTE --
  Here we optionally insert a conditional to omit caching of data not
  from the USNO-AO catalog, the one used as a gmap standard and the one
  provided for guide star data.  It is also the one served over the
  network and used as the default standard.  If the format of other
  catalog cache data would have to be different, this should be done;
  also it may be done to save a little disk space.    */

//    if (catalog != CAT_USNO_AO)  return;   // cache only default catalog

/*  -- See where to write the cache data...
  Run a loop over the index.  Retain the oldest file found.  Retain the
  first entry which does not exist.  Retain the first entry which is a
  subset of the current data set based on location and radius.  */
    for (i=0; i<Ncache; i++) {   // index scanning loop
	if ((Cache.ndx[i].status & CI_exist) == 0) {  // non-existing file
	    if (fb < 0) fb = i;  // remember first one
	    continue;  // or even break...
	}
// after here, entry i exists.
	if ((Cache.ndx[i].status & CI_good) == 0) continue;
// after here, entry i is also good

// find the oldest
	if (old < 0 || (old >= 0 && Cache.ndx[i].date < oldage)) {
	    old = i;
	    oldage = Cache.ndx[i].date;
	}

// find first subset (within same catalog)
	dist = projd (alpha, delta, Cache.ndx[i].alpha, Cache.ndx[i].delta)
		+ Cache.ndx[i].frad;
	if (sub < 0 && dist < rad && Cache.ndx[i].cat == catalog) sub = i;
    }

/* The file to use is first subset, and if no such, the first blank, and if
  no such, the oldest.  Choose that entry index as active.  */
    i = -1;
// By setting i above, all 3 selection statements are the same form;
// these can be interchanged to optimize the desired effects now.
    if (i < 0) i = sub;
    if (i < 0) i = fb;
    if (i < 0) i = old;
    if (i < 0) {  // Should not happen, can't find output location
	printf (" *= Cannot find cache output.  Should not happen.\n");
	return;
    }

/* Make up file name if needed (it may not exist yet, for a new cache) */
    if (Cache.ndx[i].filename == NULL) {
//	sprintf (fnbuf, FNform, i);
	sprintf (fnbuf, FNform, exchar(i));
	Cache.ndx[i].filename = fncat (Cache.directory, fnbuf);
    }
/* NOTE -- Don't think it is possible for Cache.ndx[i].filename to both
  exist and be wrong.  However it may well not exist yet.  */

/* Delete the indicated file if it exists.
  However, we actually do not delete an existing file.  Instead, we
  overwrite (or attempt to) the file.  If that succeeds, we will update
  the index, otherwise it remains.  */

/*  Write the indicated file and update the index.  */
    F = xopen (Cache.ndx[i].filename, "w");
    if (F == NULL) {
	printf (" *= Could not open cache output to %s\n", 
		Cache.ndx[i].filename);
	return;
    }

/*  ==  Write the cache file;  header and all data  ==  */

    fprintf (F, HeaderForm, catalog, alpha, delta, rad, n);
    for (j = 0; j < n; j++) {  /* Loop over all stars to cache file */
	fprintf (F, CacheForm, maplist[j].id,
		maplist[j].alpha, maplist[j].delta,
		maplist[j].mag, maplist[j].spc );
    }
    iclose (F);   // Close the cache file

/*  Using subroutine, obtain the proper index value for this index file;
  verifying that it is readable.  If it cannot be written, consider turning
  on the skip flag for the whole cache status.
  ..Includes getting date from a stat on the file.    */

/* Properly update the Cache index, set all values including date */
    j = get_cachendx (i, &Cache);

// Write debug message if cache was written correctly, or an error message
//  if it was not.
//    if (j==0 && dodebug >= 0) printf (" *= Cache wrote %d stars to %s.\n", 
    if (j==0 && dodebug >= 0) printf (" Saved %d stars to %s.\n", 
			n, Cache.ndx[i].filename);
    if (j && dodebug >= 0) printf (" *= Cache index status %d for %s.\n",
			j, Cache.ndx[i].filename);

/*  Clean up dynamic storage.  If any.  */
}	// End of put_cache
#endif  // on CACHECAT



int map_getlist(double a0,double d0,double rx,double ry)
/*  Return number of stars found (nmap) and stored in the global
  array "maplist" of type CATstar (defined in catalog.h as a stuct).
  Parameters:  a0 = R.A. in hours, d0 = Dec. in degrees, rx and ry
  are field size in arc seconds.  mlim is a magnitude limit, now unused.  */
{
  int     i,j,ia1,ia2,id1,id2,n,n1,n2,ok=0;  /* a0,d0 is assumed to be */
  int     nmap=0;                            /* equinox=cat_equ[catalog] */
  long    id=0;
#ifndef  TRIM_C
  double  dista;
#endif
  double  cosd15,epoch;
  double  r0;
  double  r1, r2;
#ifdef  EXCAT
  int     l;                            /* equinox=cat_equ[catalog] */
#ifdef  NETGETV
  double  vx,vy,vz;
#endif
  double  tmag;
char	*server = NULL;
#endif  // on EXCAT
int	ns1, ns2;
static  double  sava0=-25.0, savd0=-2500.0;
static  double  savr0=0.0;
static	int	savcat= -1;

  CATstar tmp;

#ifdef  EXCAT
    xmlrpc_env env;
    xmlrpc_value *result;
    xmlrpc_value *value;
//    xmlrpc_value  xvalue;
    unsigned int result_size;
//    char * const serverUrl = "http://localhost:7080/";
//    char * const serverUrl = "http://mifune.ociw.edu:7080/";
//    char * const serverUrl = "http://fujiwara.ociw.edu:7080/";

// timing of catalog data here
// need timers for getlist, getcache, putcache, network, decode, diskread
#ifdef  TIMEDEBUG
// NOTE that it is required that these be NULL if unused !!
T_timer  *gotlist = NULL;
T_timer  *gotcache = NULL;
T_timer  *putcache = NULL;
T_timer  *networkt = NULL;
T_timer  *tdecode = NULL;
T_timer  *diskread = NULL;
#endif

// For each timer, need a T_new and a T_xit call.

/*  All this server stuff is really not needed or desired.  The prefered
  way to set the server for any testing is to use the
     environment variable  CATSERVER
  The DefaultServer value should only be defined for the distributed
  system, and then only to the "official" server location.  The proper
  location for that is indentified by the comment below.

  Similarly for the DefaultMethod below, the distributed system uses
  the "offical" default; for testing use the
     environment variable  CATMETHOD
  to set the server method name.            */

#ifdef  HACKSERVER
// hack server uses port 7080 on fujiwara.ociw.edu and PullStars method.
    char* const DefaultServer = "fujiwara.obs.carnegiescience.edu";
//    char * const methodName = "GetStars";
//    char * const methodName = "GetUSNO_AO";
    char * const methodName = "PullStars";
#else  // for HACKSERVER; regular server stuff follows
// *** Official server name defaulted here ***
//    char* const  DefaultServer = "draco.obs.carnegiescience.edu";
    char* const  DefaultServer = "catxml.obs.carnegiescience.edu";
#endif  // on HACKSERVER
char	*method;	// needs this larger scope
// *** Official server method defaulted here ***
char * const  DefaultMethod = AMETHOD;	// Was "GetStars";
    char  SURL[256];
#endif  // on EXCAT

//printf (" #- Starting map_getlist, d = %.3f\n",d0); fflush(stdout); //debug

//printf (" #- Starting map_getlist, catalog %s\n", cat_name(catalog) ); //debug

#ifdef  CACHECAT
/*  Special section -- See if request for cache report  */
    if (a0 <= -1.0 && rx <= 0.0) {
// Might need to make values a little more specific in the future?
	crep = 1;
	crep = get_cache (a0, d0, rx);
	if (crep == 0) {
	    printf (" -- Catalog caching is not in effect.\n");
	} else  crep = 0;
	return  0;
    }
#endif

  if (map_sel > -1) id = maplist[map_sel].id; /* save selected star */
 
  cosd15 = 15.*cos(d0*Degree);           /* local 'cos(delta)' */
/* Eliminate possible loop problem if near pole */
  if (fabs(cosd15) < 1.e-8) d0 *= (1.0 - 1.3e-8);   // > (4/pi) * 1.e-8
// in ibound, cosd = 15 * 3600 * cos(d)
// 1/cosd must be over 480 at smallest cosd;
// so 15cos(d) * 3600 * 480 < 1 is required;
// cos(d) < 3.858e-8, 15cos(d) < 5.78e-7
// test 15cosd vs 1.e-7 = 10.0e-8
// at 1.e-7, cosd=(1/15)*1.e-7, d = (pi/2) + (1/15)*1.e-7
// need to subtract (2/15)*1.e-7
// relative to pi/2 that is (2/pi)*(2/15)*1.e-7 = 8.5e-9.
// cannot subtract more than:  cosd > 5.78e-7,
// which is (2/pi)*5.78e-7 = 3.68e-7.
  epoch  = get_epoch(0);

/*  Transform rx (input parameter) to field of view radius, and use it
  as that in the rest of the process...  */

//  if (map_pa != 0.) { rx *= 1.42; ry *= 1.42; }
/* For varying position angles, and other reasons such as non-square
  aspect ratios, it is helpful to increase rx, ry to a little more
  than the diagonal of the window; at all times. */
  r0 = hypot (rx, ry);
//  dista = 1.1 * hypot (rx, ry);
//  dista = ( 1.0 + ExtraField ) * r0;
//  rx = ry = dista;
  r1 = ( 1.0 + ExtraField ) * r0;
  r2 = ( 1.0 + ExtraField ) * r1;
  rx = ry = r2;

/* Rationalized radius values:
  r0 is the required radius
  r1 is enhanced radius used for requiring cache size
  r2 is full data radius used to trim disk data and to obtain data
  from a network server.  All data is trimmed to r2.
  rx and ry in call to cat_ibound are set then to r2.      */

/* If we are at the location and with the radius saved, we can return the
  saved nmap and the current array without going through the effort of
  actually reading the catalog.  This can happen only with the  ALLMSTAR
  feature active (not requiring magnitude limit) or with MAPBUFR feature
  active to reduce the reading.  */

    if (savr0 > 5.0 && map_savn > 0 && savcat == catalog) {
double  dab;
	dab = fabs(r0 - savr0) / savr0;
#ifndef  TRIM_C
	dista = fabs(a0 - sava0);
	if (dista > 12.0) dista = 24.0 - dista;
	dab += dista*cosd15*3600.;
	dab += fabs(d0 - savd0) * 3600.;
#else
	dab += projd (sava0,savd0, a0,d0);
#endif
/* dab is now the absolute sum of radius difference, position differences
  in ra and dec., all in arc seconds.  If small enough, that is within the
  0.1 extra radius we established in finding dista above, we will allow
  the existing data to persist.  */
	if (dab < savr0 * 0.9 * ExtraField) {
	    map_nstars = nmap = map_savn;
//printf (" -- Keeping old maplist within map_getlist --\n");  // debug
	    return  nmap;	// Omit reading things!
	}
    }

/*  Change the above as follows:
  1) Find radius difference  dab  as  projd (sava0, savd0, a0, d0)
  2) Compare dab with rx * ExtraField
  // CHECK WHERE AND HOW WE GET savr0 sava0 savd0  (before return below)
  // Do we check catalog and epoch?  We should.
... */

/*  NOTE
  What we have to do here, both in the ALLMSTAR case above, and in
  the CACHECAT case below is to see if the current array is acceptable.
  To be acceptable, it must satisfy all these things:
  1) Not have moved too far from center as to omit needed stars
  2) Not have expanded or zoomed in too far as to require stars of
     a different magnitude range or to need extra stars
  3) Have the same catalog requested
  4) Have the same epoch if applicable (or be very close)
  -- If any of these are not satisfied, we must get the data again, and
  that may mean re-reading the cache if applicable, or getting the data
  from disk again.
... */

// start timing for getlist here ==gotlist
#ifdef  TIMEDEBUG
    if (Gob->obs->dlevel > 0)
	gotlist = T_new (" == Catalog lookup:", TF_auto);
#endif

  cat_ibound(a0,d0,rx,ry,&id1,&id2,&ia1,&ia2);
#ifdef DEBUG
  fprintf(stderr,"a=%f,d=%f, id1=%d,id2=%d, ia1=%d,ia2=%d\n",
          a0,d0,id1,id2,ia1,ia2);
  fflush(stderr);
#endif

/* Debug print if debug selected */
  if (Gob->obs->dlevel > 0)
    printf (" .getlist for %.5f, %.5f;  %.3f, %.3f\n",
	a0, d0,  rx, ry );

/*  At or near this point, we must call get_cache to attempt to get the
  data from a cache and return it in the maplist global array with the
  value of nmap set appropriately.
  Also, if the maplist is set non-null, we must goto the appropriate
  point below; just after a possible put_cache call, at which the
  magnitude limit trimming might take place.  call it "map_gotlist".
  ...  */


  if (maplist == NULL) {             /* no semaphore necessary */
    mapsize = MAPSIZE_STEP;          /* single threaded process */
//    maplist = (CATstar*)malloc(mapsize*sizeof(CATstar));
    maplist = (CATstar*)calloc(mapsize, sizeof(CATstar));
// SHOULD WE NOT BE SETTING nmap TO ZERO HERE ??
    nmap = 0;	// Added 2009/8-14 KDC  may be redundant?
  }
/* NOTE:
  And what if maplist is not NULL here?  What then?  Why would we have
  some storage already allocated?  Shouldn't we clear it?   */

#ifdef  CACHECAT

/*  Call the get_cache program to get any cached catalog data which might
  be able to satisfy the current center.  If we get data, bypass the catalog
  reading by going to map_gotlist below.  Otherwise, fall through to doing
  the standard catalog operations.      */

// start timing of get cache here/ and end too ==gotcache
#ifdef  TIMEDEBUG
    if (Gob->obs->dlevel > 0)
	gotcache = T_new (" == Cache reading:", TF_auto);
#endif
//    nmap = get_cache (a0, d0, rx);
    nmap = get_cache (a0, d0, r1);
#ifdef  TIMEDEBUG
    if (Gob->obs->dlevel > 0)  T_xit (&gotcache, Rflags);
#endif
    if (dodebug > 1 || (dodebug > 0 && nmap > 0) )
	printf (" *= Got %d stars from cache.\n", nmap);
    if (nmap > 0) goto map_gotlist;

#endif  // on CACHECAT


/* -- Get stars from the chosen catalog (or network) -- */

#ifdef  CACHECAT
//    dista = rx;
//    rx *= (1.0 + ExtraField);
//    ry = rx;
#endif

#ifdef  EXCAT
  if (USEXMLRPC & 2) {   // Using network catalog
//char	*server = NULL;
char	*portname = NULL;
//char	*method = NULL;  // currently needs the larger scope
// The method could be defined here if the above test were combined
// with the similar test just following the completion of the above section...
int	port = PortNum;
// start timing of network catalog lookup here ==networkt
#ifdef  TIMEDEBUG
    if (Gob->obs->dlevel > 0)
	networkt = T_new (" == Network lookup:", TF_auto);
#endif
    /* Initialize our error-handling environment. */
    xmlrpc_env_init(&env);

//    catalog = CAT_USNO_AO;	// That's the only one we serve...
// was debug only; set catalog in redraw_skywin

    xmlrpc_client_init2(&env, XMLRPC_CLIENT_NO_FLAGS, XNAME, XVERSION, NULL, 0);
    die_if_fault_occurred(&env);
//    if (!skywin_active) return 0;	// leave if no longer useful
    if (!skywin_active) goto inactive;	// leave if no longer useful
// also cancel timing if  active -- networkt, gotlist

/* Try lookup of environment variable  CATSERVER and if it comes up null,
  substitute DefalutServer.  Then put it into SURL using the sprintf below. */
    server = getenv ("CATSERVER");	// This might be optional?
    if (server == NULL) server = DefaultServer;
    portname = getenv ("CATPORT");
    if (portname != NULL) {
	j = sscanf (portname, "%d", &port);
	if (j != 1) port = PortNum;
    }
#ifdef  HACKSERVER
    sprintf (SURL, "http://%s:%d", server, 7080);  // A terminal / is not needed
// We use the literal port number for the hack server since we don't
// change the port on the hack server.
#else
    sprintf (SURL, "http://%s:%d/RPC2", server, port);
//    sprintf (SURL, "http://%s:%d/CAT", server, port);
/* REALLY need that /RPC2 following the server and port; without it we get
  a 404 error and nothing comes out!
  Also, the server apparently can't change the default easily...      */
#endif
/* A terminal / on the URL is not needed, yet it does not hurt either,
  and even a // on the end appears to work.
  The URL is like:  http://fujiwara.ociw.edu:7080
  Or, for the real server:  http://draco.ociw.edu:7084/RPC2  */

/* Find the method to be used.  Look at CATMETHOD env.vbl. and check for
  valid methods; then set the one used.  */
    method = getenv ("CATMETHOD");
    if (method == NULL) method = DefaultMethod;
#ifdef  HACKSERVER
    method = methodName;	// Only one in the hack
#endif
// Test if method is one of the acceptable methods, otherwise use default
    if ( strcmp(method, "GetStars") && strcmp(method, "PullStars") )
	method = DefaultMethod;

    if (Gob->obs->dlevel > 0)
	printf("Making XMLRPC call to server url '%s' method '%s' \n",
           SURL, method);
  }  // end of first network use test
// NOTE above and below lines refer to the logically identical test ---
  if (USEXMLRPC & 2) {    // using internet to get catalog data
/* -------------------------------------------------------------- */
int	smeth;	// server method type
int	kfv=0;
int	kml=0;	// debug counts of fov and mag limited in decode

//    r0 = rx * 1.5;	// Temporary fix to limiting cache

    xmlrpc_limit_set(XMLRPC_XML_SIZE_LIMIT_ID,320*1024*1024);  // was 288
//    xmlrpc_limit_set(XMLRPC_XML_SIZE_LIMIT_ID,64*1024*1024);  // was 16, 48, 64
#ifdef  HACKSERVER
    result = xmlrpc_client_call(&env, SURL, methodName, "(ddddi)",
                (xmlrpc_double) a0, (xmlrpc_double) d0,
                (xmlrpc_double) rx, (xmlrpc_double) ry, (xmlrpc_int) catalog);
#else
    result = xmlrpc_client_call(&env, SURL, method, "(dddi)",
                (xmlrpc_double) a0, (xmlrpc_double) d0,
                (xmlrpc_double) rx, (xmlrpc_int) catalog);
#endif
// NOTE rx, ry above are set to r2
// adding catalog indicator for new server
            die_if_fault_occurred(&env);
//	    if (!skywin_active) return 0;	// leave if no longer useful
	    if (!skywin_active) goto inactive;	// leave if no longer useful

// start timing of result decode here  ==tdecode
#ifdef  TIMEDEBUG
    if (Gob->obs->dlevel > 0)
	tdecode = T_new (" == Decode netdata:", TF_auto);
#endif
    result_size = xmlrpc_array_size(&env, result);
    if (Gob->obs->dlevel > 0) printf(" Got %d stars from server: %s\n",
		result_size, server);

/* Extraction of data from the "result" array (an xmlrpc array) and placing
  the data values in the maplist array of CATstar structure data types.  */

#ifdef  HACKSERVER
    smeth = 0;
#else
    smeth = strcmp (method, "GetStars") == 0;
#endif

    for(l=0;l<result_size;l++) {    // loop over internet results
char*	tspc;		// for string decode...
char  const	*line;	// data line found
/* Using above obviates warning of incompatable types from compile for the
  xmlrpc_read_string below, but produces a warning about free below,
  unless a cast is added to its use as an argument to free.  Stupid.  */
unsigned long  idv;
double	aiv, div, miv;
char	spv[12];

	xmlrpc_array_read_item(&env,result,l,&value);
      if (smeth) {   // GetStars -- string value in array
	xmlrpc_read_string (&env, value, &line );
/* In the downloaded source, it says that the last argument above is a
  pointer to a pointer, which gets set to newly malloc'ed storage which
  the user must then free.  Doing so works, although the compiler
  still complains about the 3rd argument above being incompatable pointer type;
  which is why the definition includes the "const".  Stupid.   */
	j = sscanf (line, CacheFormI, &idv, &aiv, &div, &miv, spv);
	free ( (char*) line);  // cast needed to eliminate stupid warning
	if (j < 4) continue;  // just forget any bad line...
	tmp.id = idv;
	tmp.alpha = aiv;
	tmp.delta = div;
/* If we change the format of the cache, or make it dependent on catalog,
  these lines following might need to be changed too.    */
	tmp.mag = miv;
	if (j >= 5) strncpy (tmp.spc, spv, 8);
	tmp.v = had2vec (aiv, div);
      } else {  // smeth false - PullStars case, structure in array
	xmlrpc_decompose_value(&env,value,
#ifdef  NETGETV
		 "{s:i,s:d,s:d,s:d,s:s,s:d,s:d,s:d,*}",
#else
		 "{s:i,s:d,s:d,s:d,s:s,*}",
#endif
		 "id",&tmp.id,
		 "alpha",&tmp.alpha,
		 "delta",&tmp.delta,
		 "mag",&tmag,

//		 "spc",&tmp.spc,	// THIS IS WRONG.
		 "spc",&tspc	// Need address of character pointer
/* What happens here is that the address of the character data in the array
  is placed into the pointer at &tspc, i.e. the tspc pointer itself.
  This differs from int and double things which values are placed in the
  location given by their pointer.  Instead of &(double), we need
  here to give &(char*), and copy the actual data later.    */

#ifdef  NETGETV
		,"vx",&vx ,"vy",&vy ,"vz",&vz
#endif
		);

	strncpy (tmp.spc, tspc, 8);  // copy the temporary location
/* To get the character data, we copy the character string now pointed to
  by txpc into the proper character location in the tmp structure.  */

	tmp.mag = (float) tmag;

#ifdef  NETGETV
/* The "v" vector is completely redundant with alpha and delta.  For
  speed and network traffic reduction we can eliminate sending the
  3 components of v and derive the v vector here from the alpha and
  delta values we need anyway.  This is controled by NETGETV symbol.  */
	tmp.v.x = vx;
	tmp.v.y = vy;
	tmp.v.z = vz;
#else
	tmp.v = had2vec (tmp.alpha, tmp.delta);
#endif  // on NETGETV

/* NOTE -- when using &tmp.spc (that is &(char[]) ), we find that instead
  of the character data, a pointer to the character data is placed there.
  So the fix is to pass a pointer and copy out the characters later rather
  than to try to convince the server to place the characters.  The routine
  xmlrpc_decompose_value is the one at fault for treating character values
  differently from arithemetic values here.  Its documentation does not
  explicitly state this difference, but an example therein indicates that
  the string decodes to a pointer and not a character array.   */

/* We may have to get blue and red magnitudes from the server, and use the
  logic seen in catalog.c:510 to obtain both mag and the spc which will in
  that case be a color index value.
  If this is done, expect the server to pass an integer (32 bit) which
  will be decoded into the red and blue magnitudes as in catalog.c.  This
  then replaces the floating point mag and the character spc values.
  The decoding would take place here, into the tmp structure.
  It would be expected that the CATstar would be extended to include the
  special integer to be written to cache too.      */

      }  // End of conditional on smeth - string method vs. structure method

/*  Here we are sub-selecting the current object in tmp (a CATstar struct)
  obtained from the server for location, field of view and perhaps for
  magnitude limit.  Continue the loop for non-selected objects, and put
  selected ones into maplist array.  If that array fills up, use realloc
  to obtain a larger array.
  This is exactly what is done for stars obtained from disk below.    */

#ifdef  NotDefinedTrim
#ifndef  TRIM_C
	if (fabs(d0-tmp.delta)*3600. > ry) {kfv++; continue;} /* distd */
        dista = fabs(a0-tmp.alpha);                  /* dista */
        if (dista > 12.) dista = 24.-dista;
        if (dista*cosd15*3600. > rx) {kfv++; continue;}  /* out of FOV */
#else
	if (projd(a0,d0, tmp.alpha,tmp.delta) > r0) {kfv++; continue;}  // out of FOV
/* Using r0 instead of rx above temporarily to allow most of fov from
  network to appear in the cache.    */
#endif  // on TRIM_C
#endif  // on NotDefinedTrim
// Dropping all trim activity after server call, it is not needed

/*  NOTE -
  Here we could put a selection on a fixed magnitude limit, say 19.5 to
  test if it substantially reduces the array size.  Also do it in the
  code getting stars from disk below.      */

	maplist[nmap] = tmp;	// Much better, cleaner...
        nmap++;
        if (nmap == mapsize) {           /* increase maplist */
          mapsize += MAPSIZE_STEP;
          maplist = (CATstar*)realloc(maplist, mapsize*sizeof(CATstar));
        }
     }  // loop on l over internet results

/* Debug report on decoding activity.  Only those stars decoded will
  be cached, so we need to accept most into FOV at this point.  */
    if (Gob->obs->dlevel > 1) 
	printf (" *= Decoded %d of %d stars: %d to FOV, %d to mlim.\n", 
		nmap, result_size,  kfv, kml);

//    ok = nmap > 0;		// fake needed if not reading catalog
// Code below reordered and fixed, this is no longer needed.

/* The clean-up after section.  Here we clean up the storage obtained
  for the xmlrpc process which got the internet data.  */
    die_if_fault_occurred(&env);
//    if (!skywin_active) return 0;	// leave if no longer useful
    if (!skywin_active) goto inactive;	// leave if no longer useful

    /* Dispose of our result value. */
    xmlrpc_DECREF(result);

    /* Clean up our error-handling environment. */
    xmlrpc_env_clean(&env);
    xmlrpc_client_cleanup();

#ifdef  TIMEDEBUG
    if (Gob->obs->dlevel > 0) {
	T_xit (&tdecode, Rflags);
	T_xit (&networkt, Rflags);
    }
#endif

    if (result_size > 2) {
	if (TESTMODE) printf (" Star data obtained from internet server: %s\n",
		server);
	else          printf (" Star data obtained from internet server.\n");
    }

// Here ends the conditional on USEXMLRPC & 2
// and the local section from way far up above...
  } else {	// not using internet

#endif	// on EXCAT

/*  ===  Get catalog data from local disk file.  ===  */

#ifdef  TIMEDEBUG
    if (Gob->obs->dlevel > 0)
	diskread = T_new (" == Disk reading:", TF_auto);
#endif

/*  Above was for possible use of network server.  Here is the code
  for getting data from the disk based catalogs.      */

  for (j=id1; j<=id2; j++) {         /* delta-bands */
    for (i=ia1; i<=ia2; i++) {       /* hour-bands */
      cat_indices(i,j,&n1,&n2);
      for (n=n1; n<=n2; n++) {       /* loop over indices */
        switch (catalog) {
        case CAT_GSCP:
	  ok = GetGSCplus(j,n,&tmp); break;	/* guessing at j */
        case CAT_GSC:
          ok = GetGSC(n,&tmp); break;
        case CAT_PPM:
          ok = GetPPM(n,&tmp); break;
        case CAT_SAO:
          ok = GetSAO(n,&tmp,epoch); break;
        case CAT_USNO_SA:
          ok = GetUSNO_SA(j,n,&tmp); break;
        case CAT_USNO_AO:
          ok = GetUSNO_AO(j,n,&tmp); break;
        default:
          ok = 0; break;
        }
        if (!ok) { 
          if ((catalog==CAT_USNO_SA) || (catalog==CAT_USNO_AO)) continue; 
          else break;
        }
/* Select here for location and possibly magnitude; unwanted stars
  go to continue the loop, others are inserted in maplist array.    */
#ifndef TRIM_C
        if (fabs(d0-tmp.delta)*3600. > ry) continue; /* distd */
        dista = fabs(a0-tmp.alpha);                  /* dista */
        if (dista > 12.) dista = 24.-dista;
        if (dista*cosd15*3600. > rx) continue;  /* out of FOV */
#else
	if (projd(a0,d0, tmp.alpha,tmp.delta) > rx) continue;  // out of FOV
#endif  // on TRIM_C
// Note that rx, ry used above are set to r2

/*  NOTE -
  Here we could put a selection on a fixed magnitude limit, say 19.5 to
  test if it substantially reduces the array size.  Also do it in the
  code getting stars from network above.      */

	if (tmp.mag > MLIM) continue;           /* star too faint */
/* Above magnitude limit added 2011/02-17 (after version 2.10) to make the
  network and disk lookup results the same, network server uses the same
  limit (set in mgfeats.h) as we use here.  However, the server may be
  compiled at a different time, so full equality is not guaranteed.    */

	maplist[nmap] = tmp;	// Much better, cleaner...
        nmap++;
        if (nmap == mapsize) {           /* increase maplist */
          mapsize += MAPSIZE_STEP;
          maplist = (CATstar*)realloc(maplist, mapsize*sizeof(CATstar));
        }
      }  /* endfor: 'n' */
      if (!ok) { 
        if ((catalog==CAT_USNO_SA) || (catalog==CAT_USNO_AO)) continue; 
        else break;
      }
    }  /* endfor: 'i' */
    if (!ok) { 
      if ((catalog==CAT_USNO_SA) || (catalog==CAT_USNO_AO)) continue; 
      else break; 
    }
  }  /* endfor 'j' */

  if (!ok) { 
    if ((catalog!=CAT_USNO_SA) && (catalog!=CAT_USNO_AO)) {
#ifdef  TIMEDEBUG
	if (Gob->obs->dlevel > 0)  T_xit (&diskread, Rflags);
#endif
	return(-1);
    }  // catalog USNO test
  }  // the !ok test above ends here
/* Having the above test outside the internet conditional cause all catalogs
  other than USNO to fail to display when data was from internet, as the
  "ok" status is not set unless we actually read datalog data.   */

#ifdef  TIMEDEBUG
    if (Gob->obs->dlevel > 0)  T_xit (&diskread, Rflags);
#endif

    if (nmap > 2) {
	if (TESTMODE)
		printf (" Star data obtained from local disk catalog: %s\n",
			catpath() );
	else    printf (" Star data obtained from local disk catalog.\n");
    }

#ifdef  EXCAT
  }  // End of not using internet conditional
// if conditional on USEXMLRPC & 2 to decide internet or disk ends here.
// (Actually, this is the end of the else clause of that conditional,
//  so it ends the disk use code.)
// This ifdef conditional section is just for that closing curly brace...
#endif  // on EXCAT


#ifdef  CACHECAT
/* Having gotten star data from actual catalog or from network, we make a
  local cache file here before continuing.  */

    if (nmap > 0) {
#ifdef  TIMEDEBUG
	if (Gob->obs->dlevel > 0)
	    putcache = T_new (" == Cache writing:", TF_auto);
#endif
	put_cache (a0, d0, rx, nmap);  // rx is r2 here
#ifdef  TIMEDEBUG
	if (Gob->obs->dlevel > 0)  T_xit (&putcache, Rflags);
#endif
    }
//    if (nmap > 0) put_cache (a0, d0, rx, nmap);

//    rx = dista;    // Removing the enhancement by ExtraField
//    ry = rx;

/* If we got stars from cache, we transfer to the next line bypassing all the
  above catalog reading stuff.  */
map_gotlist:

#endif  // on CACHECAT


#ifdef DEBUG
//  IS THIS EVEN USED ANYMORE ??
  fprintf(stderr,"nmap=%d (opt=%d, max=%d)\n", nmap, STAROPT, STARLIM);
#endif  // on DEBUG
 
  if ((nmap < mapsize/2) && (nmap > MAPSIZE_STEP)) { /* decrease maplist */
    mapsize /= 2;
    maplist = (CATstar*)realloc(maplist, mapsize*sizeof(CATstar));
  }

/*   NOTE:
  Probably, we want to put the above section which can reduce the
  storage allocation below the following section which can reduce
  the nmap value by limiting the magnitude range when ALLMSTAR is
  defined.  Also, the above should be in a for loop, or a better method
  of choosing the reduced mapsize should be chosen in case the reduction
  is larger.  Probably look at nmap/MAPSIZE_STEP and mapsize/MAPSIZE_STEP
  to find the appropriate mapsize.    */
 
  if (nmap > 1) map_sort(maplist,0,nmap-1);  /* faint over bright stars */

  if (Gob->obs->dlevel > 0)  printf (" .. %d stars to %.2f", 
	nmap, nmap > 0 ? maplist[nmap-1].mag : 19.98);
//  if (nmap > maxlist) { nmap = maxlist;  // limit the map size
/* Limit the number/magnitude, complex here */
    if (nmap > STAROPT) {		// See if we limit at all
	ns1 = nsmag (MAGLIM1, nmap);
	ns2 = nsmag (MAGLIM2, nmap);
	if (Gob->obs->dlevel > 0) printf (" (%d/%.1f, %d/%.1f)",
		ns1, MAGLIM1, ns2, MAGLIM2);
	if (ns2 < STAROPT) {	// limit to staropt
	    if (nmap > STAROPT) nmap = STAROPT;
	} else if (ns2 < STARLIM) {  // limit to ns2
	    if (nmap > ns2) nmap = ns2;
	} else if (ns1 < STARLIM) {  // limit to STARLIM
	    if (nmap > STARLIM) nmap = STARLIM;
	} else {  // limit to maximum of starlim or 2x staropt
	    ns1 = STARLIM;
	    if (2*STAROPT > ns1) ns1 = 2*STAROPT;
	    if (nmap > ns1) nmap = ns1;
	}
//    } else {	// Use all nmap stars
    }

    if (Gob->obs->dlevel > 0) printf (" (at %d, %.2f)\n", 
	nmap, nmap > 0 ? maplist[nmap-1].mag : 0.0);

//  if (Gob->obs->dlevel > 0) printf ("\n");

  map_sel = -1;
  if (id > 0) {
    for (i=0; i<nmap; i++) {
      if (maplist[i].id == id) {
        map_sel = i;
        break;
      }
    }
  }
  map_nstars = nmap;

/* Save the position for later possible use */
    map_savn = nmap;
    sava0 = a0;
    savd0 = d0;
    savr0 = r0;
    savcat = catalog;

#ifdef  TIMEDEBUG
    if (Gob->obs->dlevel > 0)  T_xit (&gotlist, Rflags);
#endif

/*  NOTE
  We also need to save the catalog value, the equinox and epoch,
  (maybe not equinox, determined by catalog).  Redundant probably
  to save both map_nstars and map_savn here, they are the same.
  ALSO check that there are no other returns which could leave
  the saved items un-specified.        */

  if (nmap <= 0) skywin_active = 0;	// no stars, no window
  return nmap;

/*  Special section to return 0 if window is inactive - check timing */
inactive:
#ifdef  TIMEDEBUG
// networkt, gotlist, tdecode might be active, so close them
    if (Gob->obs->dlevel > 0) {
//	if (tdecode == NULL) fprintf (stderr, "  ** tdecode NULL\n");
//	fflush (stderr);
	T_xit (&tdecode, Rflags);
//	fflush (stderr);
	T_xit (&networkt, Rflags);
	T_xit (&gotlist, Rflags);
    }
#endif
// Add any other things which might need to be shut down...
// Since skywin_active being zero sent us here, this should be redundant...
    skywin_active = 0;	// no stars, no window (redundant here)
    return  0;
}	// End of map_getlist
 
/* ---------------------------------------------------------------------- */

void  map_draw_x (Display* display, Window win, GC gc, int x, int y, int s)
{
// Centered, arm of s*root(2)
    XDrawLine(display,win,gc, x-s,y+s, x+s,y-s);
    XDrawLine(display,win,gc, x-s,y-s, x+s,y+s);
}

/* ---------------------------------------------------------------------- */

void  map_draw_sq (Display* display, Window win, GC gc, int x, int y, int s)
{
// Centered, side of 2s
    XDrawLine(display,win,gc, x-s,y+s, x+s,y+s);
    XDrawLine(display,win,gc, x+s,y+s, x+s,y-s);
    XDrawLine(display,win,gc, x+s,y-s, x-s,y-s);
    XDrawLine(display,win,gc, x-s,y-s, x-s,y+s);
}

/* ---------------------------------------------------------------------- */

void  map_draw_tr (Display* display, Window win, GC gc, int x, int y, int s)
{
// Centered, side of 4s
    XDrawLine(display,win,gc, x-s*2,y+s, x+s*2,y+s);
    XDrawLine(display,win,gc, x+s,y+s, x,y-s*2);
    XDrawLine(display,win,gc, x,y-s*2, x-s*2,y+s);
}

/* ---------------------------------------------------------------------- */
/* Count real objects, minding use count... */
int	krobjs (Obs* obs)
/* Return count of nodes in object queue */
{
int	k;
objq	*ob;
objq	*p;
    if (obs == NULL) return 0;
    ob = obs->ob;
    if (ob == NULL) return 0;
    for (k=0,p=ob; p!=NULL; p = p->next) {
	if ( (p->type == OBJ_OBJECT) &&
	     (p->use <= obs->reuseob) )  k++;
	if (p->next == ob) break;
    }
    return  k;
}

 
/* ---------------------------------------------------------------------- */
// move this up above all calls to it...
// #define  DEBUG_GS
#ifdef  GSDATA
int	check_guidestar (gpos* gs)
//static	int	check_guidestar (gpos* gs)
//static	int	check_guidestar (gpos* gs, int g)
/* Returns index of star matching gs for guider window g */
{
//int	ms;
int	star = -1;
int	i;
int	sel = -1;
//int	gf;
vect2	p;
vect2	d;
//vect2	x;
vect3	v;
double	dq, dmin = 1.e+32;

// catch nulls
#ifdef  DEBUG_GS
    if (gs == NULL) {
	printf (" Null gpos for guider %d\n", g+1);
	return  star;
    }
#else
    if (gs == NULL) return star;
#endif  // on DEBUG_GS

// return -1 if gs has no guidestar
#ifdef  DEBUG_GS
    if (gs->ep == 0.0) {		// debug section only
	printf (" Epoch 0 in guider %d\n", g+1);
	return  star;
    }
#else
    if (gs->ep == 0.0) return  star;
#endif  // on DEBUG_GS

// using code similar to get_closest_star obtain star close to
// the position g as a star index; return -1 if not found
//    v = prec (map_eq, sph2vec(gs->ra * Hour, gs->dec * Degree), gs->ep);
    v = prec (map_eq, had2vec(gs->ra, gs->dec), gs->ep);
    p = xyproj (v);

    for (i=0,dmin=99.0; i<map_nstars; i++) {
	d = sub2vect (xyproj (maplist[i].v), p);
	dq = dot2vect (d, d);
	if (dq < dmin) {
	    dmin = dq;
	    sel = i;
	    if (dmin == 0.0) break;
	}
    }

// if the dmin value is too large, return -1
#ifdef  DEBUG_GS
    if (sqrt(dmin) > 0.4) {
	printf (" Guider %d out by %.1f arcsec.\n", g+1,
		sqrt(dmin) );
	return  star;
    }
#else
    if (sqrt(dmin) > 0.4) return  star;
//    if (sqrt(dmin)/ArcSecond > 0.4) return  star;
#endif  // on DEBUG_GS

/* Sorry, the skymap program here knows nothing of the guider
  windows or magnitude limits.  Thus, we can't' check these... */

// // use find_guider to see if star found is in guider window,
// // substituting for map_sel properly.
// // If not in the proper window, silently clear the gs
//     ms = map_sel;
//     map_sel = sel;
//     gf = find_guider (&x, &d);
//     map_sel = ms;
// 
// /* Check that we have a guide star in the proper window */
//     if (gf != g+1) {	// Found bad guide star
// 	gs->ra = gs->dec = gs->ep = 0.0;
// 	gs->cn = -1;
// 	return  star;
//     }
// 
// // Check the magnitude also, and clear if bad
//     if (maplist[sel].mag > gdr[g].mag) {
// 	gs->ra = gs->dec = gs->ep = 0.0;
// 	gs->cn = -1;
// 	return  star;
//     }

    star = sel;

// if found in window, and bright, return star index.
    return  star;

}
#endif  // on GSDATA

#ifdef  GSDATA
int	guider_index (gpos g)
/* Return the guider index plus one value for the guider position
  found in structure g.  Return 0 if not in a proper guider, and
  the guider number if found.  If g.ep is zero, return zero.  */
{
int	r=0;	// return value
vect3	vp;	// Local converted position

    if (g.ep == 0.0) return r;
//    vp = sph2vec (g.ra * Hour,  g.dec * Degree);
    vp = had2vec (g.ra,  g.dec);
    r = find_gpos (vp);
    return  r;
}
#endif  // on GSDATA

/* NOTE!  This following program should be moved to skymap.c as it
  is probably now only used there; and set to static (local) there.
  When done, correct skywin.h also.  */
#ifdef  GSDATA
static	void	validate_guiders (Obs* ob)
/* Check both guiders for current validity and clear any which are
  found to be not valid.  Set agscan flag too.  Used for centering
  and recentering operations.  */
{
gpos	gz = {0.0, 0.0, 0.0, -1};
int	gs;
//int	st;
//int	f;

/* NOTE -- We need to have a valid transform matrix computed to use
  guider_index properly, otherwise it returns 0.  So, we exit here
  if map_nstars has not been properly set yet.  */
    if (map_nstars < 10) return;

//    st = check_guidestar (&(ob->gp1));
//    f = check_star (st);
//    if (f < 1) {
/* NOTE -- It may be that guide star in ob is from a different catalog
  than the one currently displayed.  That should not be an error, so
  we only clear the guider if the star exists AND is not in guider now.  */
//    if (st >= 0 && f < 1) {
    gs = guider_index (ob->gp1);
    if (gs != 1) {
	ob->gp1 = gz;
#ifdef  AUTOGS
	ob->agscan = 0;	// Will need rescan of autogs
#endif  // on AUTOGS
    }
//    st = check_guidestar (&(ob->gp2));
//    f = check_star (st);
//    if (st >= 0 && f < 1) {
    gs = guider_index (ob->gp2);
    if (gs != 2) {
//    if (guider_index(ob->gp2) != 2) {
	ob->gp2 = gz;
#ifdef  AUTOGS
	ob->agscan = 0;	// Will need rescan of autogs
#endif  // on AUTOGS
    }
}
#endif  // on GSDATA

/* ---------------------------------------------------------------------- */
 
void map_redraw(Display* display,Window win,float mlim)
{
  int   i,r,p=0;
  int	d;
  int	x, y;
  int	xc, yc;
int	n, cs;
  float c;
  GC    lgc;
#ifdef  UPARO
  double  lst;
//  double  catza, catzd;
#endif
double	catza = 0.0;
double	catzd = 0.0;
objq	*obq, *q;
vect3	v;
vect3	pm[3];	// Precession matrix
Obs	*obs;
#ifdef  DEBUGproj
double	rd, rms;
double	ac, dc;
int	np=0;
#endif

  static double sclst[] = {
    30.,60.,120.,180.,300.,600.,1200.,1800.,3600.,7200.,18000.,0.};
  static char *sctxt[] = {
    "30\"","1'","2'","3'","5'","10'","20'","30'","1dg","2dg","5dg",""};
 
  XClearWindow(display,win);

#ifdef  DBredraw
  printf (" ! Start map_redraw.  For %d stars.\n", map_nstars); fflush (stdout);  // debug
#endif

#ifdef  DEBUGproj
    vec2had (&ac, &dc, unproj (make2vect(0.0, 0.0)));  // kludge, yes
// like vec2sph
#endif
  for (i=0; i<map_nstars; i++) {               /* loop over stars in list */
    map_coords(&x, &y, maplist[i].v);
#ifdef  DEBUGproj
    rd = vect2norm(xyproj(maplist[i].v)) - 
	projd(ac,dc, maplist[i].alpha, maplist[i].delta);
    runrms (rd, &rms, &np);
#endif
    r = map_get_radius(maplist[i].mag,mlim);
    switch(maplist[i].spc[0]) {
    case 'O':
    case 'B':
      lgc = gc_blue;
      break;
    case 'A':
    case 'F':
      lgc = gc_green;
      break;
    case 'G':
      lgc = gc_yellow;
      break;
    case 'K':
      lgc = gc_red;
      break;
    case 'M':
    case 'R':
    case 'N':
      lgc = gc_brown;
      break;
    case 'C':
      sscanf(maplist[i].spc,"C%f",&c);  
      if      (c > 3.00) lgc = gc_brown;   /* M */
      else if (c > 1.50) lgc = gc_red;     /* K */
      else if (c > 1.00) lgc = gc_yellow;  /* G */
      else if (c > 0.00) lgc = gc_green;   /* AF */
      else               lgc = gc_blue;    /* OB */
      break;
    default:
      lgc = gc_black;
      break;
    }
/* NOTE -- at this point, a global could turn off star coloring by
  setting "lgc = gc_black" - to drop all color */
    if (map_sc == 0) lgc = gc_black;	/* Kill star colors */
    switch (r) {
    case 0:
      XDrawPoint(display,win,lgc,x,y); break;
    case 1:
      XFillRectangle(display,win,lgc,x-1,y-1,2,2); break;
    default:
/*      XFillArc(display,win,lgc,x-r,y-r,2*r,2*r,0,23040); break; */
      XFillArc(display,win,lgc,x-r,y-r,2*r,2*r,0,23040); break;
    } 
  }		/* End loop over stars */
#ifdef  DEBUGproj
  printf (" map_redraw rms %.2f for %d stars.\n", rms, np);
#endif

#ifdef  DBredraw
  printf (" ! map_redraw location 1. %d stars.\n", i); fflush (stdout);  // debug
#endif

/* Plot the objects, if selected by the feature... */

// Find the object queue start, skipping if null
// CHANGE TO ALSO GET USE COUNTS AND REFSEL VALUE
    obs = (Gob == NULL) ? NULL : Gob->obs;
    obq = (obs == NULL) ? NULL : obs->ob;
//    if (obs != NULL) obq = obs->ob;
//    else  obq = NULL;
/* Set obq NULL if neither plot feature is currently active. */
    if (!(map_ap | map_op)) obq = NULL;

/* Set the size for the object plotting */
// CHANGE TO USE OBS AND TAKE USE COUNT
    n = krobjs (obs);
    if      (n <   10) cs = 5;
    else if (n <   50) cs = 4;
    else if (n <  200) cs = 3;
    else if (n < 1000) cs = 2;
    else               cs = 1;

    pmatrix (pm, map_eq, equinox);	// Get precession matrix to map_eq

    i = 0;		// reference counter
    d = obs->refsel;
    if (d < 1) d = 1;
    for (q=obq; q != NULL; q=q->next) {		// Loop over object queue
// CHANGE TO ADD SELECTION BASED ON USE COUNT
// REJECTED OBJECTS DO NOT CONTINUE, USE goto TO THE BREAK TEST
// OR PUT THE BREAK TEST WITH THE REJECTION.
// bypass any object types with exceeded counts...
	if ( (q->type == OBJ_OBJECT) && (q->use > obs->reuseob) ) {
	    if (q->next == obq) break;
	    continue;
	}
// For reference, we count and apply colors too.

	v = sph2vec (q->dat.ra, q->dat.dec);	// Find x,y location

// NOTE -- We must precess the object position from "equinox" to the
// catalog epoch, "map_eq" or cat_equ[catalog].

/* Here is how to do precession:
1) From utils.c (C. Birk's utilities) is:
void precess(double ain,double din,double ein, double *aout, double *dout,
             double eout)
(The a,d are in hours and degrees here).

2) From kdcutil.c:
vect3	prec (double e, vect3 pf, double ef)
  ( Precess vector position pf at epoch ef to epoch e )

void	pprec (double *a, double *d, double e,
	double a1, double d1, double e1)  {
  ( Precess from position a1, d1 at e1 to *a, *d at e.  Positions are
  in radians, and epoch in Julian years  )

The vector precession would probably be the best thing here.
  .. */

//	vc = prec (map_eq, v, equinox);   // Precess to map_eq

//	vc = mmult3 (pm, v);		// Precess using matrix
//	map_coords(&x, &y, vc);

	map_coords (&x, &y, mmult3 (pm, v) );

//	map_coords(&x, &y, v);

// Decide color based on object type
//	lgc = gc_orange;	// for now, same
//	lgc = gc_red;	// for now, same
//	lgc = gc_grey;	// for now, same
// Plot a x or box at the location
	if (q->type > 1)  {		// 2 = reference
//	    if (q->use > 0) lgc = gc_yellow;
	    if (map_ap) {
//		lgc = gc_yellow;
// Find color white = unused, yellow = used
		if (q->use > obs->reuserf) {
		    lgc = gc_white;
		} else {
		    if ((i % d) == 0)  lgc = gc_yellow;
// d set to obs->refsel above outside loop and limited to 1-n
		    else   lgc = gc_white;
		    i++;
		}
		map_draw_sq (display,win,lgc, x, y, 4);
	    }
	}  else  {			// 0 = unknown, 1 = object
	    if (map_op) {
//		lgc = gc_grey;
		lgc = gc_white;
		map_draw_x (display,win,lgc, x, y, cs);
	    }
//	    ;  // do nothing for now
	}
// End of object plotting loop...
	if (q->next == obq) break;
    }

#ifdef  DBredraw
  printf (" ! map_redraw location 3.  %d objs\n", i); fflush (stdout);  // debug
#endif

/* Find plot center and telescope field center points */
  x = map_w/2;
  y = map_h/2;
  map_pltxy (&xc, &yc, make2vect(0.0, 0.0));

/* If we are panning, we draw 2 center crosses, one in blue at the center
  of the display, and one in black at the center of the field.  The blue
  one is drawn second so it is displayed in front... */

  if (map_pn) {
    XDrawLine(display,win,map_gc,xc-5,yc,xc+5,yc);  /* field center cross  */
    XDrawLine(display,win,map_gc,xc,yc-5,xc,yc+5);
    XDrawLine(display,win,gc_blue,x-5,y,x+5,y);     /* plot center cross   */
    XDrawLine(display,win,gc_blue,x,y-5,x,y+5);
  } else {
    XDrawLine(display,win,map_gc,x-5,y,x+5,y);      /*  center cross    */
    XDrawLine(display,win,map_gc,x,y-5,x,y+5);
  }

/* North vector, and east vector too. */
  if (map_nv) {
    map_vector(display,win,gc_green,xc,yc,(map_w+map_h)/10,0.0);
    map_vector(display,win,gc_green,xc,yc,(map_w+map_h)/25,90.0*Degree);
  }

//#ifdef  UPOPTION
///* Precession will be required for up or zenith */
//  if (map_uv || map_zd) {
//    lst=get_siderial(ut_now,tellong)/3600.;
//    precess (lst, tellat, cur_epoch, &catza, &catzd, map_eq );
//  }
//#else
  if (catza != 0.0) catza = 0.0;  // To avoid unused message
  if (catzd != 0.0) catzd = 0.0;  // To avoid unused message
//#endif

#ifdef  UPARO
/* Alternate method when hour angle specified... */
  if ( (map_uv || map_zd) && obs->ha > -24.0) {
    lst = obs->ra/Hour + obs->ha;	// NOTE that ra is in radians...
    precess (lst, tellat, cur_epoch, &catza, &catzd, map_eq );
  }
#endif

#ifdef  UPARO
/* Now, the zenith pointing vector in yellow... */
//  if (map_uv) {
  if (map_uv && obs->ha > -24.0) {
    map_vector(display,win,gc_yellow,xc,yc,(map_w+map_h)/9.0,
	Degree * par_angle (catza - map_ca, map_cd, catzd));
/*	par_angle (get_ha(ut_now,tellong,map_ca), map_cd, tellat)); */
  }
#endif

#ifdef  ZNTHDOT
/* Draw the zenith as a special point if requested. */
//  if (map_zd) {
  if (map_zd && obs->ha > -24.0) {
//    map_coords (&x, &y, sph2vec (catza*Hour, catzd*Degree));
    map_coords (&x, &y, had2vec (catza, catzd));
/*    XFillRectangle(display,win,gc_white,x-1,y-1,2,2); */
/* Plot yellow circle, with two blue filled arcs inside */
/* Plot red circle, with two white filled arcs inside */
    XFillArc(display,win,gc_red,x-6,y-6,12,12,0,23040);
    for (i=0; i<4; i += 2) XFillArc(display,win,gc_white,
	x-5,y-5,10,10, i*(23040/4), 23040/4 );
/* Start and length of arc are last 2 */

  }
#endif

#ifdef  DBredraw
// printf (" ! map_redraw location 7.\n"); fflush (stdout);  // debug
#endif
  for (i=0; sclst[i]>0.; i++) {
    p = (int)floor(sclst[i]/map_scale+0.5);
    if (p > map_w/8) break;
    if (p > 50) break;
  }
  if (sclst[i] == 0.) i--;
  XDrawLine(display,win,map_gc,10,30,10,36);      /* draw scale           */
  XDrawLine(display,win,map_gc,10+p,30,10+p,36);
  XDrawLine(display,win,map_gc,10,33,10+p,33);
  XDrawString(display,win,map_gc,(10+p)/2-(int)(strlen(sctxt[i])*3.0),47,
    sctxt[i],strlen(sctxt[i]));
 
#ifdef  DBredraw
//  printf (" ! map_redraw location 9.\n"); fflush (stdout);  // debug
#endif

//  map_circle_star(display,win,map_sel,mlim);      /* circle selected star */
  map_circle_star(display,win,
	app->white,map_sel,mlim);      /* circle selected star */

// use map_get_closest_star to find guide stars, and plot them
#ifdef  GSDATA
// At this point, catch out any guide stars which may have been
// placed outside of their fields by cenering or recentering movements,
// and get this done before doing automatic guide star selection
    validate_guiders (obs);

// Now would be a good time to automatically select guide stars;
// Do this by calling get_autogs here
#ifdef  AUTOGS
    i = get_autogs (obs);
// And, we don't really care what the status is (yet)
#endif

// For each guide star find its subscript if any in the display;
// Check that the guide star is valid, in magnitude and guider location,
// using check_guider.  If not checked, delete the gs.
/* The gp1 or gp2 position is cleared in map_guideset (skywin.c:2314) if
  the guide star is not in field or not in brightness range.  This is
  detected here by i being out of range when set through check_guidestar.
  This is because check_guidestar cannot check position and magnitude
  limits directly itself.  */

    i = check_guidestar (&(obs->gp1));
//    i = check_guidestar (&(obs->gp1), 0);
    map_circle_star (display, win,
	app->red, i, mlim);
//    i = check_guidestar (&(obs->gp2), 1);
    i = check_guidestar (&(obs->gp2));
    map_circle_star (display, win,
	app->red, i, mlim);

// plot the circle around that star
#endif  // on GSDATA

#ifdef  DBredraw
  printf (" !  End  map_redraw.\n"); fflush (stdout);  // debug
#endif
}

/* ---------------------------------------------------------------------- */
//#define  GetAutoDB	// Debug prints for get auto guide star
#undef  GetAutoDB	// Debug off
#ifdef  AUTOGS
#ifdef  GSDATA
// NOTE that automatic guide stars depends both on AUTOGS feature
// and on the GSDATA feature being on!
int	get_autogs (Obs* ob)
/* Look through the star information (if any) to find any guide stars
  (if none already present) for either guide star field.  Return a status
  depending on what action is performed;
   -1 if no guide star information is present or other error
    0 if guide stars are present and nothing is done, or not found
    1 if guide stars are found and entered.
  The guide stars found are entered in the Obs structure provided.   */
{
int	gs = 0;
int	gn = 0;  // guider data needed flags
int	i;
int	f;
int	st[2];
double	sm[2];
double	t;

#ifdef  GetAutoDB
    printf ("..Start get_autogs %d/%d.\n", ob->autogs, ob->agscan);
#endif
// -- Operational checks
    if (ob == NULL) return -1;		// Need a valid structure
// check autogs flag on
    if (ob->autogs == 0) return 0;
// check proper instrument
// Needs to be imacs currently; leave early if not.
    if (InstIndx(ObInstN(ob)) != I_imacs) return 0;
// check guide stars exist - which ones are needed
    gn = 0;
    if (ob->gp1.ep == 0.0) gn |= 1;
    if (ob->gp2.ep == 0.0) gn |= 2;
// try changing != to == on above 2 tests - works better
#ifdef  GetAutoDB
    printf (" ..get_autogs guider flag = %d, stars = %d.\n", gn, map_nstars);
#endif
    if (gn == 0) return 0;
// check already searched
    if (ob->agscan) return 0;
#ifdef  GetAutoDB
//    printf (" ..get_autogs called for gn=%d.\n", gn);
#endif

// -- Resource checks
// check catalog data present
    if (map_nstars < 1) return -1;
// check if proper catalog being used
	// HOW??
    if (catalog != CAT_USNO_AO) return 0;
// it is an external int defined in catalog.h and
// set when the catalog is read.  This is optional...
#ifdef  GetAutoDB
    printf (" ..get_autogs has nstars %d and catalog.\n", map_nstars);
#endif

// -- Scan available stars in catalog
// Clear saving arrays
    for (i=0; i<2; i++) {
	st[i] = -1;
	sm[i] = 99.0;	// the small is better metric
    }
    for (i=0; i<map_nstars; i++) {
// see if this star is a valid guide star 
	f = check_star (i);
	if (f < 1) continue;
	if (f > 2) continue;
// check that this field is needed
	if ((f & gn) == 0) continue;
// have a guide star candidate, compute star metric (subroutine)
	t = guide_metric (i);
#ifdef  GetAutoDB
    printf (" ..Test star %d for guider %d metric %.3f\n", i, f, t);
#endif
// remember this star if better than existing
	if (t < sm[f-1]) {
	    st[f-1] = i;
	    sm[f-1] = t;
	}
    }	// Star scanning loop

// -- Fill in any guide stars found, set status
    for (i=0; i<2; i++) {
	f = i+1;
	if ((f & gn) == 0) continue;	// not needed (redundant?)
//	if (st[i] < 0) continue;	// no star, redundant
	if (st[i] < 0) {	// Star needed, not found message
	    printf (" *-* No guidestar could be found for area %d.\n", f);
	}
	map_guideset (st[i], 0);	// set the guidestar
    }

// -- Set scan flag and Return the proper status
    ob->agscan |= 1;
    return  gs;
}
#endif  // on GSDATA
#endif  // on AUTOGS

/* ---------------------------------------------------------------------- */

/* ---------------------------------------------------------------------- */

// #ifdef  GUIDEWIN
// #endif  /* on GUIDEWIN */

/* ---------------------------------------------------------------------- */

// double  par_angle (double ha, double dec, double lat)
// /* Computes parallactic angle; i.e. the position angle of the
//   zenith from the given hour angle and declination. */

#ifdef  UPARO
// compute the paralactic angle, degrees
double  par_angle (double h, double d, double g)
{
// double  par_angle (double ha, double dec, double lat)
/* Computes parallactic angle; i.e. the position angle of the
  zenith from the given hour angle and declination. */

double  st, ct;
double  sp, cp;
double  sd, cd;
double  cz;
double  x;

// look up the trig functions
    sincos (h*Hour, &st, &ct);
    sincos (d*Degree, &sd, &cd);
    sincos (g*Degree, &sp, &cp);

    cz = ct*cp*cd + sp*sd;
/*  x = atan2 (st*cp, (sp - cz*sd)/cd); */
    x = atan2 (st*cp*cd, sp - cz*sd );
            
/* Alternate is omit cz, and find
    x = atan2 (cp*st, (sp*cd - cp*sd*ct));
.. */       
            
  return  x/Degree;
}
#endif  

/* ---------------------------------------------------------------------- */

static void map_vector(Display* display, Window win, GC lgc,
	long x, long y, long length, double vpa)
{
/* Try to put a vector on the map... */
long	ex, ey;
long	px, py;
double  theta;
double	phi;
double  vv;

/*  theta = 270.0*Degree - map_pa - vpa; */	/* Added arbitrary 180 deg. */
  theta = 90.0*Degree + map_pa + vpa;	/* Added arbitrary 180 deg. */
/* Probably, y increases downwards, use -y and normal signs? */
  vv = length;
  ex = x + (long) (vv * cos (theta));
  ey = y - (long) (vv * sin (theta));

  XDrawLine(display,win,lgc,x,y,ex,ey);

/* Draw an arrow head at the end of the vector */
  vv *= 0.15;

  phi = theta + 150.0*Degree;
  px = ex + (long) (vv * cos (phi));
  py = ey - (long) (vv * sin (phi));
  XDrawLine(display,win,lgc,ex,ey,px,py);

  phi = theta - 150.0*Degree;
  px = ex + (long) (vv * cos (phi));
  py = ey - (long) (vv * sin (phi));
  XDrawLine(display,win,lgc,ex,ey,px,py);

}

/* ---------------------------------------------------------------------- */

static void map_sort(CATstar* array,int l,int r)
{
  int     i,j;
  float   m;
  CATstar temp;
 
  i = l;
  j = r;
 
  if (j>i) {
    m = array[(i+j)/2].mag;
    do {
      while (array[i].mag < m) i++;
      while (array[j].mag > m) j--;
      if (i<=j) {
        memcpy(&temp,&array[i],sizeof(CATstar));
        memcpy(&array[i++],&array[j],sizeof(CATstar));
        memcpy(&array[j--],&temp,sizeof(CATstar));
      }
    } while (!(i>j));
    map_sort(array,l,j);
    map_sort(array,i,r);
  }
}

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */

