/*  skywin.c  ==  Sky window for interface GUI  */

/* Modified from telmap.c of the gmap project.. */
/* Also see the file  skymap.c  */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2011 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define IS_SKYWIN_C

/* DEFINEs -------------------------------------------------------------- */

//  #define DEBUG

//#define  DBGsd
#undef  DBGsd
// These debugs give sky redraw indication messages
//#define  DBGsd2
#undef  DBGsd2

#undef  DBrsky
//#define  DBrsky
// Above to give redraw skywin messages

#undef  DBzep
//#define  DBzep
// Above for toggle off guide star message

#undef  DBconf
//#define  DBconf
// Above for configuration debug extra messages

/* The GCENT feature is moved to mgfeats.h */

#define  MODULE_NAME  "SkyWindow"
const	char	SKWNAME[] = "Sky Window";
//const	char	SKHIVD[] = "1";
//const	char	SKWVERS[] = ".18";
// versions moved down ~42 lines

#define MAX_SCALE_SAO  1200.
#define MAX_SCALE_GSC  200.
#define MAX_SCALE_USS  40.
#define MAX_SCALE_USA  15.

#define  BrightLim  11.0

#define STAR_INFO      1
#define STAR_OFFSET    2
#define	STAR_DATA      3
/* STAR_DATA is guider info without moving */

#define  MovingColor  brown
// Tried, poor: orange pink ivory white green lgrey
// Look good:   red  brown
// Not tried:    

/* INCLUDEs ------------------------------------------------------------- */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>                          /* getcwd() */
#include <math.h>                            /* log(),cos() */
#ifdef  sun
#include <sunmath.h>			/* nint, sincos */
#else
extern void  sincos (double, double*, double*);
#endif
#include	<ctype.h>		// isalpha, tolower

//#include "ccbxlib.h"                         /* Xlib stuff */
#include  "cxt.h"
#include "utils.h"                           /* generic utilities */

//#include "mgutils.h"
// #include "kdcutil.h"		/* Geometry package */
#include	MGUTILH

//const	char	SKHIVD[] = "1";
const	char	SKWVERS[] = Xstr(SKYWINV);
// was ".18";

#include "datutil.h"			/* default data utilities */

//#include	"mgfeats.h"
// Now included by maskdat.h

#include	"../intgui.h"

// To include guiutil.h when it is made too.

#include "skymap.h"
#include "skywin.h"
#include "catalog.h"

/* External function prototypes */
// void  command_interface (int i, int dim, double val);
/* Above is coded in gmap.c */

/* EXTERNs -------------------------------------------------------------- */

// some of these become non-extern globals here, removed from pmap

Bool       recalc_on_map=False;
double     alpha,delta,equinox;
       float      pixscale,choffx[8],choffy[8];
double	hangl;	// for DIFREF

/* These things are done in main, once only */
extern gdata      *GDAT;		/* Global Data structure */
extern int        nsects[3];
extern section    **csections;
extern int        dimx,dimy,nchips;
extern char       c_name[64],g_name[64],t_name[64];
// guiderdata  gdr[4];
// enum  dmenus  { MNU_TEL, MNU_GDR, MNU_CAM };	// copied from main;
// the above should be in the header file for skywin?
// extern int	  Gchange;

// extern  Obs	*obs;		// Maybe needed???

/* The IFU offsets, from intgui.c global */
extern	double	IFUdx[2];
extern	double	IFUdy[2];

/*  The following externs are actually defined in and used
  for the main task, we access them here to display the data
  wich is being edited by the main task.   */

/* This first group may not be needed here */
extern EditWindow alwin,dewin;
// extern EditWindow pawin;
// extern EditWindow gdr1r, gdr1a;
// extern EditWindow gdr2r, gdr2a;

extern	MainWindow  mwin;
extern	Application *app;

/* Managing the skywin option in the main option menu; this stuff is
  copied from intgui.c ... */
extern	Menu	optmenu;
enum opt_menus { 
// add more here -- must agree with add menu options when creating menu
		OPT_SKYWIN,
		OPT_DUMMY,
		OPT_EXIT
};

extern	wble*	  winlist;	// GCENT use only
extern  winblock  RAfield;
extern  winblock  DECfield;
extern  winblock  slitpaf;
extern  winblock  hafield;	// for DIFREF

extern	double	RAsave;
extern	double	DECsave;
extern	int	recent;

extern  float	pa;
extern  float	cp;

extern  Obf*	Gob;

/* Some external function prototypes...
  these should be defined in another header.  Actually defined
  in the main task itself.   */
extern  float   get_cp             (float);

extern	int	TESTMODE;

#ifdef  EXCAT
//extern Bool    USEXMLRPC;
extern int    USEXMLRPC;
#endif

/* GLOBALs -------------------------------------------------------------- */

int	nguiders;
guiderdata	gdr[4];
double	sx = 1960.0; // 0.0;	/* Scale maximum size, for create_skywin */
// 1960.0 = 980 * 2 -- to include imacs guiders as well.

u_long	skycolor = 0;
int	  skydraw;

int	skywin_active = 0;
// int	default_cat=CAT_GSC;
int	default_cat=CAT_USNO_AO;	// Per Dressler's request 04/01

/* Button blocks for button widgets */
/* butblock:  xloc,yloc, width,height, label,
	*data,  *button, aligntype, laboff  */

/* Put this button block stuff below the statics.
  Needs inclusion of guisubs.h or intgui.h stuff... */
// butblock  PCbutn = {
// 	FXw/2 + 20, FXy, 30, FXh,  "Set Center",
// 	NULL, &pcbut, Leftside);
// // Dynamically add xwa.width/2 to the x value

// butblock  GSbutn = {
// 	FXw/2 + 20, FXy, 30, FXh,  "P-guide",
// 	NULL, &gsbut, Rightside);
// // Dynamically subtract x from xwa.width/2

// A button list is needed, too.
// bble*	swbuts = NULL;

// Also need hanglerl programs; see dunbut in xmap.c for example

/* STATICs -------------------------------------------------------------- */

static Bool  auto_maglim=True;
static float man_maglim=16.,maglim;
static GC    skywin_gc;

static float g_offaxis_x,g_offaxis_y;
static float g_field_w,g_field_h;
static float g_unit_x,g_unit_y;
static float g_maglimit;
static float g_brightlim;
static int   g_center_x,g_center_y;
static int   mapsel_type;

/* These buttons are globals, and go in skywin.h; like camenu */
//static Button  pcbut;		// pan center set to telescope
static	Button	rcbut;		// The recenter button
// static Button  gsbut;		// guide star selection button

/* function prototype(s) */

#ifdef  DBconf
static void  config_skywin (int x, int y);
#else
static void  config_skywin (void);
#endif

static	int	astrcmp (char* a, char* b);
static void    start_skywin (void);

static void  redraw_skymenus  (void);
static void  redraw_skytext    (Display*);
static void  redraw_buttons (void);

static void  recalc_map     (Display*);
static float get_mlim       (int,double);
static void  adjust_scale   (void);
static Bool  is_maxscale    (void);

// static void	handle_rcbut (Display* display, int button);
static void	handle_rcbut (void);

// static void	Telescope_name  (char*);
// note the above is not yet external; move to skywin.h if needed outside

/* ---------------------------------------------------------------------- */
/*   Utility used by skywin  */
void	init_gcenter (void) {
/* Uses only global values */

    if (RAfield.data == NULL || DECfield.data == NULL) return;
/* The above exists for the REFRESH feature where this is called from
  write_obs without the window having been started.  Also helps if for
  any reason we get here without the window being initialized.   */

    RAsave = *(double*)RAfield.data / RAfield.scale;
    DECsave = *(double*)DECfield.data / DECfield.scale;
    if (recent) skydraw |= 1;	// Up for drawing
    recent = 0;
}

/* ---------------------------------------------------------------------- */
/*             Here's the code for the sky window                         */
/* ---------------------------------------------------------------------- */

/* ---------------------------------------------------------------------- */

//int	skywin_handle (XEvent* event)
//	    if (CBX_AutoEvent(&skywin,&subevent,&sleeptime)) {
//void	skywin_handle (void)
int	skywin_handle (int* sleeptime)
/* Handle the event loop of the skywin window.  The "sleeptime"
  parameter is a pointer to the current sleep time, which gets
  incremented when there are few events, and drastically cut
  when there are many events.  Returns true (non-zero) if it handles
  events, and zero if no events are handled.  Probably should return
  a count of the events handled.  */
{
/* Handles the skywin window; independently of events in main window.
  Senses the active flag and starts skywin if needed, or closes it
  if needed.  Handles skywin events.  Checks for changes to position
  and updates the display properly.  */

XEvent	event;
//int	sleeptime = 10;
double	c;
// int	replot = 0;

int	events = 0;

static	int	rob=0;
static	int	ref=0;
static	int	rsl=0;

static int  ermask=0;				// debug
static int  erlimit=0;				// debug
static int	mescount = 0;			// debug
char	buffer[256];				// debug

static char	instname[32] = "";	// to track instrument changes

// if (!catalog) {			// debug section
//   printf (" >- Start skywin-handle catalog=%d, set %d.\n", 
// 		catalog, default_cat);
//   catalog = default_cat;
// }

    if (map_kd) skywin_active = map_kd = 0;	// Kill feature
    if (!skywin_active) map_savn = 0;

    if (skywin.disp == NULL) {		// No display
	if (!skywin_active) return events;	// and don't want one
	start_skywin ();
    } else {				// Have display
	if (!skywin_active) {
	    CBX_CloseMainWindow (&skywin);
// should also uncheck the option in the option menu here...
	    CBX_CheckMenu (&optmenu, OPT_SKYWIN, skywin_active);
	    cat_close();
	    return events;
	}
    }

	/* - - - - - - - - - - *\
	|			|
	|    Loop   Program	|
	|			|
	\* - - - - - - - - - - */

/* Start of skywin loop program */

// Look for any changes in ra, dec, epoch (maybe others)
// Then, check for any event, and run switch if found.
// Set a redraw flag if changes found, and in switch if
// an event requires it.
// Finally, if redraw is set, do the redraw once only.

// CHANGE ALL PROGRAMS to set the skywin_redraw flag instead of actually
// calling the redraw_skywin program itself.  This to eliminate multiple
// redraws when things change.
// The flag needs different values to decide if recalc is true.
// A 1 bit indicates redraw; a 2 bit indicates recalc and redraw.

// Look for R.A. change (RAfield)
    c = *(double*)RAfield.data / RAfield.scale;
    if (c != alpha) {
#ifdef  DBGsd2
	printf (" Dra alpha=%.4g, c=%.4g", alpha, c);	//debug
#endif
	alpha = c;
	recent = 1;	// Enable recentering

// 	printf (" --Redraw for alpha (hs) = %.5f\n", c);
// 	printf ("  RAfield data pointer = %8.8lx\n", RAfield.data); //debug
//	redraw_skywin(skywin.disp,True);
//	replot = True;

	skydraw |= 3;
#ifdef  DBGsd2
	printf (" change ra, skydraw=3\n");	// debug
#endif
    }

// Look for Dec. change (DECfield)
    c = *(double*)DECfield.data / DECfield.scale;
    if (c != delta) {
#ifdef  DBGsd2
	printf (" Ddec delta=%.4g, c=%.4g", delta, c);	//debug
#endif
	delta = c;
	recent = 1;	// Enable recentering

// 	printf (" --Redraw for delta (hs) = %.5f\n", c);
//	redraw_skywin(skywin.disp,True);
// NOTE - IF THIS OR PA CHANGES, WE NEED TO RECHECK THE ROTATOR,
// AS IS DONE IN INTGUI!!
//	replot = True;

	skydraw |= 3;
#ifdef  DBGsd2
	printf (" change dec, skydraw=3\n");	//debug
#endif
    }
// Need recalc True above to get the new map!


// Look for PA change (slitpaf)
    c = *(double*)slitpaf.data / slitpaf.scale;


//    if (c != pa) {
// Can get roundoff errors with scale, and -30 = 330 in angle too.
//    if (fabsdif(c, pa, 360.0) > 3.0e-6) {
/* This compare needs to be to about float accuracy rather than double,
  for unknown reasons in order to work properly.  */
    if (fabsdif(c, pa, 360.0) > 3.0e-5) {
	pa = c;

// if (!catalog) {			// debug section
//   printf (" >- skywin-handle 2.3 catalog=%d, set %d.\n", 
// 	catalog, default_cat);
//   catalog = default_cat;
// }

	cp = get_cp (c);

// if (!catalog) {			// debug section
//   printf (" >- skywin-handle 2.4 catalog=%d, set %d.\n", 
// 	catalog, default_cat);
//   catalog = default_cat;
// }

	map_pa = -pa * Degree;
//	redraw_skywin(skywin.disp,True);
//	replot = True;
	skydraw |= 3;
#ifdef  DBGsd
	printf (" change pa, skydraw=3\n");	//debug
#endif
    }

#ifdef  DIFREF
/* This section just detects changes to the hour angle box, the
  hangl variable is used only to detect change, it is not otherwise
  used as an hour angle value.  */
    c = *(double*)hafield.data / hafield.scale;
    if (c != hangl) {
	hangl = c;
	skydraw |= 3;
#ifdef  DBGsd
	printf (" change hangl, skydraw=3\n");	// debug
#endif
    }
	hangl = c;
#endif


// Add check of equinox too.
// If changes, reset equinox but not alpha,delta; then redraw.
// Should be handled by handle_eqmenu ?
    if (equinox != Gob->obs->equinox) {
	equinox = Gob->obs->equinox;
// printf (" Skywin set equinox to %.2f\n", equinox);
//	redraw_skywin(skywin.disp,True);
//	replot = True;
	skydraw |= 3;
#ifdef  DBGsd
	printf (" set equinox, skydraw=3\n");	// debug
#endif
    }

// if (!catalog) {			// debug section
//   printf (" >- skywin-handle 4 catalog=%d, set %d.\n", 
// 	catalog, default_cat);
//   catalog = default_cat;
// }

// CHANGE TO ADD TRACKING OF SELECTION COUNT IN OBS, AND IF ANY
// DO CHANGE, FLAG FOR DRAWING THE MAP AGAIN.

/* Track any changes in the use counts or refsel; we will need to
  redraw the window if any of these change and option is on... */
    if (rob != Gob->obs->reuseob) {
	if (map_op) skydraw |= 1;
	rob = Gob->obs->reuseob;
    }

    if (ref != Gob->obs->reuserf) {
	if (map_ap) skydraw |= 1;
	ref = Gob->obs->reuserf;
    }

    if (rsl != Gob->obs->refsel) {
	if (map_ap) skydraw |= 1;
	rsl = Gob->obs->refsel;
    }


/* Check for change of insrument name, and change camera field if so */
    if (Gob != NULL && Gob->obs != NULL && Gob->obs->Instr_mt != NULL) {
//	if (strcmp(Gob->obs->Instr_mt->name, instname)) {  // change in inst.
#ifdef  GISCAM
	if (strcmp(ObFrontN(Gob->obs), instname)) {  // change in inst.
#else
	if (strcmp(ObInstN(Gob->obs), instname)) {  // change in inst.
#endif
	section	*cam, *sp;	// local loop variables
    
	    if (*instname) {	// Not just initializing
		skydraw |= 1;
#ifdef  DBGsd
		printf (" change inst. skydraw=1\n");	// debug
#endif
	    }
//	    strcpy (instname, Gob->obs->Instr_mt->name);
#ifdef  GISCAM
	    strcpy (instname, ObFrontN(Gob->obs));
#else
	    strcpy (instname, ObInstN(Gob->obs));
#endif
    // printf (" >Looking for camera %s here.\n", instname);	// debug
	    for (sp=cam=GDAT->camera; sp != NULL; sp = sp->next) {
    // See if we have name in sp equal to camera name in obs
		if (!astrcmp(sp->name, instname)) {
    // NOTE - made astrcmp non-case-sensitive to find camera names,
    // matching e.g. IMACS_sc to IMACS-S
    // printf (" >Camera %s found as %s.\n", instname, sp->name);  // debug
#if (DEBUG_GUI > 0)
		printf (" >Cam %s found = %s.\n", instname, sp->name);
#endif
		    GDAT->camera = sp;	// The actual set of camera...
    // NOTE - don't know if that is really necessary.
    //		    setup_defaults ();
		    get_camera_defaults ();  // Needed to change plot
		    break;
		}
    //printf ("Not camera %s; loop again...\n", sp->name);
		if (sp->next == cam) break;	// circular stopper
	    }	// Loop over all camera sections

/* Also check for guider here... */
	    for (sp=cam=GDAT->guider; sp != NULL; sp = sp->next) {
    // See if we have name in sp equal to guider name in obs
		if (!astrcmp(sp->name, instname)) {
    // NOTE - made astrcmp non-case-sensitive to find camera names,
    // matching e.g. IMACS_sc to IMACS-S
    // printf (" >Camera %s found as %s.\n", instname, sp->name);  // debug
#if (DEBUG_GUI > 0)
		printf (" >Gdr %s found = %s.\n", instname, sp->name);
#endif
		    GDAT->guider = sp;	// The actual set of guider...
    // NOTE - don't know if that is really necessary.
    //		    setup_defaults ();
		    get_camera_defaults ();  // Needed to change plot
		    break;
		}
		if (sp->next == cam) break;	// circular stopper
	    }	// Loop over all guider sections

	}	// Test change of instrument name
    }

	/* - - - - - - - - - - *\
	|			|
	|      Event  Loop  	|
	|			|
	\* - - - - - - - - - - */

    if (skydraw) events++;
//    sleeptime = 10;	// from 10;
//    if (CBX_AutoEvent(&skywin, &event, sleeptime) ) {
/* Change to keep looping on any events detected which need to be handled
  by this program, and do so without pausing.  */
    while (CBX_AutoEvent(&skywin, &event, sleeptime) ) {
//    if (CBX_AutoEvent(&skywin, &event, sleeptime) ) {
	events++;

	if (++mescount < 0) {		// debug messages
    Window	pw;
	    printf (" --Sky-, event.type=%d, ser=%ld, win=%8.8lx",
		    event.type, event.xany.serial, event.xany.window);
    //	printf (" --Loop, event.type = %d", event.type);
	    if (event.type == Expose) {
		printf (", Expose ");
		pw = event.xexpose.window;
		if (pw == skywin.win) printf ("skywin");
//		else if (pw == win) printf ("main");
		else printf ("other");
		printf (" count = %d", event.xexpose.count);
	    } else if (event.type == MapNotify) {
		printf (", MapNotify ");
		pw = event.xmap.window;
		if (pw == skywin.win) printf ("skywin");
//		else if (pw == win) printf ("main");
		else printf ("other");
		printf (" (no-icon).");
	    } else if (event.type == LeaveNotify) {
		printf (", LeaveNotify ");
		pw = event.xcrossing.window;
		if (pw == skywin.win) printf ("skywin");
//		else if (pw == win) printf ("main");
		else printf ("other");
	    } else if (event.type == ConfigureNotify) {
		printf (", ConfigureNotify ");
		pw = event.xconfigure.window;
		if (pw == skywin.win) printf ("skywin");
		else printf ("other");
	    } else if (event.type == ButtonPress) {
		printf (", ButtonPress ");
		pw = event.xbutton.window;
		if (pw == skywin.win) printf ("skywin");
//		else if (pw == win) printf ("main");
		else printf ("other");
	    }
	    printf ("\n");
	}					// end debug message section


	switch (event.type) {			/// loop switch
    
	case KeyPress:                        /* key pressed event   *2*   */
	    break;
    /* key release is *3* */
    
	case ButtonPress:                    /* button pressed event *4*  */
    /* -- ButtonPress is a mouse button activation -- */
#ifdef  SHFTCLK  // was just: __APPLE__
/* For Apple computers, shift-click should be center click */
	if (event.xbutton.state & ShiftMask) event.xbutton.button = 2;
#endif
	    if (event.xbutton.window == zibut.win)
//		handle_zibut(skywin.disp,event.xbutton.button);
		handle_zibut(event.xbutton.button);
	    else if (event.xbutton.window == zobut.win)
//		handle_zobut(skywin.disp,event.xbutton.button);
		handle_zobut(event.xbutton.button);
	    else if (event.xbutton.window == rcbut.win)
//		handle_rcbut (skywin.disp, event.xbutton.button);
		handle_rcbut ();
	    else if (event.xbutton.window == skywin.win)
		handle_mapsel ( event.xbutton.x, event.xbutton.y,
			event.xbutton.button, event.xbutton.state);
	    else if (event.xbutton.window == camenu.win)
		handle_camenu ();	// (skywin.disp);
	    else if (event.xbutton.window == fxmenu.win) 
		handle_fxmenu ();	// (skywin.disp);
    //	else if (event.xbutton.button != 1) break;	// Not needed??
	    else  break;
    
//	    return (1);
	    break;
    
	case ButtonRelease:                 /* mouse button released *5*   */
#ifdef DEBUG
	    sprintf(buffer,"unrelated button release");
	    adebug(P_TITLE,buffer);
	    fprintf(stderr,"%s\n",buffer);
#endif
	    break;
    
// #ifdef DEBUG
	case MotionNotify:			/* pointer motion   *6*    */
		    /* ignore */
	    break;
// #endif
    
	case EnterNotify:                      /* window enter event *7*   */
	    break;
	case LeaveNotify:                      /* window leave event *8*   */
	    break;
    
	case Expose:                           /* expose events    *12*    */
//#ifdef  DBGsd
//	if (event.xexpose.count > 0) printf (" Expose, count=%d\n",
//		event.xexpose.count);
//#endif
	    if (event.xexpose.count > 0) break;     /* ignore */
	    if (event.xexpose.window == skywin.win) {
//		redraw_skywin(skywin.disp,False);
//		redraw_skywin(skywin.disp,replot);
//		replot = 0;
		skydraw |= 1;
#ifdef  DBGsd
		printf (" expose, skydraw=1\n");	// debug
#endif
//		return (1);
	    }
	    break;
    
	case MappingNotify:                       /* keyboard mapping event   */
    //	  XRefreshKeyboardMapping((XMappingEvent*)&event);
    //	  tdebug(P_TITLE,"XRefreshKeyboardMapping");
    //#ifdef DEBUG
    //	  fprintf(stderr,"XRefreshKeyboardMapping\n"); fflush(stderr);
    //#endif
	      break;
    
	case DestroyNotify:				/* 17 */
	    if (event.xdestroywindow.window == skywin.win) {  /* skymap window */
		CBX_CloseMainWindow (&skywin);
		cat_close();
    printf (" *Sky window closed.  Event = %d.\n", event.type);	// debug
//		return (1);
	    }
	    break;
    
	case KillNotify:			/* Also for kill... *52*  */
// This is the special CXT "event" detection
//	CBX_CloseMainWindow (&skywin);
// We can't do that if we get killed, best hope is to close catalogs
// and null out our display to show main that we're gone.
	    cat_close();
	    skywin.disp = NULL;
// We have not turned the main window's option menu flag off (yet)
// Also, we should turn skywin_active off, too.
	    CBX_CloseMainWindow (&skywin);	// kinda redundant...
	    skywin_active = map_kd = map_savn = 0;
	    CBX_CheckMenu (&optmenu, OPT_SKYWIN, skywin_active);
	    break;
    
	case UnmapNotify:				/* 18 */
    //	if (event.xmap.window == win) iconized = True;
	    break;
    
	case MapNotify:				/* 19 */
    //	if (event.xmap.window == win) iconized = False;
	    break;
    
	case ReparentNotify:			/* 21 */
		    /* Ignored */
	    break;
    
	case ConfigureNotify:                    /* configure event   *22*    */
//    printf (" ..ConfigNotif (22) sky found.\n");  fflush (stdout);  //debug
	    if (event.xconfigure.window == skywin.win && skydraw == 0) {
#ifdef  DBconf
		config_skywin (event.xconfigure.x, event.xconfigure.y);
#else
		config_skywin ();
#endif
//		skydraw |= 3;
	    }
// Like, what other window could it be?
//    printf (" ..ConfigNotif (22) sky handled.\n");  fflush (stdout);  //debug
	    break;
    
    /*  Some others...
	    ResizeRequest	25
	    ColormapNotify	32
	    MappingNotify	34
    .. */
    
	default:
    
//#ifdef DEBUG
	    if (event.type == ermask) {
		if (++erlimit > 4) break;
	    } else {
		ermask = event.type;
		erlimit = 1;	// This is error number 1.
	    }
	    sprintf(buffer,"unknown sky event '%d'",event.type);
//	    adebug(P_TITLE,buffer);
	    fprintf(stderr,"%s\n",buffer);
//#endif
    /* If get here, see /usr/openwin/share/include/X11/X.h  */
//	    return (1);
	    break;
	}		// End of switch for event

    }	// End of event found test.  NOW the event found loop.

// if (!catalog) {			// debug section
//   printf (" >- After event switch catalog=%d, set %d.\n", 
// 	catalog, default_cat);
//   catalog = default_cat;
// }

// Check for the redraw flag.  If on, redraw and turn off;
// if off, do nothing.
// NOTE that skydraw needs to be a global value here.
    if (skydraw & 1) {
// printf ("  -- Doing redraw skywin, flag = %d.\n", skydraw);	// debug
#ifdef  DBGsd
	printf (" draw skywin from loop, skydraw=%d\n", skydraw); //debug only
#endif
	redraw_skywin (skywin.disp, skydraw & 2);
	skydraw = 0;
    }

    return  events;
//    return  0;		/* Not handled here */
//  ***  event->xx or event.xx *** global change catcher...
}



/* ---------------------------------------------------------------------- */

static void redraw_skymenus(void)
{
//  CBX_UpdatePMenu(&camenu);
//  CBX_UpdateDMenu(&fxmenu);
//#ifdef  EXCAT
//#if defined (EXCAT) && defined (HACKSERVER)
#ifdef  HACKSERVER
//  if (!USEXMLRPC)
  if ((USEXMLRPC & 2) == 0)  // only updating camenu if no network operation
#endif
  CBX_UpdateMenu (&camenu);
  CBX_UpdateMenu (&fxmenu);
}

/* ---------------------------------------------------------------------- */
int	find_gpos (vect3 vcp)
/* Given a celestial position vector v, find whether the position is
  in a guider field, returning 0 if not, and guider subscript+1 if
  it is.  This is similar to find_guider (below) but does not construct
  the target and projection vectors, and does not take an implicit
  catalog entry as input.  Thus it is better suited to testing any
  position with respect to guider field for non-catalog entries.  */
{
int	g=0;
vect2	pv;

    pv = xyproj (vcp);

    if (nguiders < 1) {	// Old way
	g = 1;
	if (-pv.x > g_offaxis_x+g_field_w/2.) g = 0;
	if (-pv.x < g_offaxis_x-g_field_w/2.) g = 0;
	if ( pv.y > g_offaxis_y+g_field_h/2.) g = 0;
	if ( pv.y < g_offaxis_y-g_field_h/2.) g = 0;
    } else {	// nguiders > 0, IMACS way

    double	r, a;
    int	i;
    // int	found;  // maybe make this g?
    double	p, ax;
    double	gx, gy;
    
    // get r,a if we need them for polar guiders
	vec2pol (&r, &a, pv);
	a /= Degree;
	gx = gy = 0.0;
    
    /* We look only in first two guiders for a guide star, not in
     any other subsidary guiders, e.g. center-field; using 2, not nguiders */
	for (i=g=0; i<2; i++) {
	    if (gdr[i].mode & GDR_POLAR) {	/* gdr[i].mode is **Polar** */
		if (gdr[i].mode & GDR_SH) {
		    p = 0.0;
		    ax = 0.0;
		} else {
		    p = gdr[i].px * gdr[i].psize / 2.0;
		    ax = atan (p / gdr[i].position_u.polr_data.rp)/Degree;
		}
		if (r < gdr[i].position_u.polr_data.rlo - p) continue;
		if (r > gdr[i].position_u.polr_data.rhi + p) continue;
		if (a < gdr[i].position_u.polr_data.alo - ax) continue;
		if (a > gdr[i].position_u.polr_data.ahi + ax) continue;
		g = i + 1;	// Found guider number g
	    } else {	/* gdr[i].mode is **Rectangular** */
	/* This section has not been tested */
		if (pv.x < gdr[i].position_u.rect_data.xlo) continue;
		if (pv.x > gdr[i].position_u.rect_data.xhi) continue;
		if (pv.y < gdr[i].position_u.rect_data.ylo) continue;
		if (pv.y > gdr[i].position_u.rect_data.yhi) continue;
		g = i + 1;
	    }	/* end guider mode selection */
	    if (g > 0) break;	// Found the guider  ?
	}	/* for loop over guiders */
    }	// end nguiders conditional
    return  g;
}


static	int	find_guider (vect2 *v, vect2 *d)
{
/* Find whether star number map_sel is in the guider field, returning
  zero if not (or no map_sel) and guider subscript+1 if it is.
  The "v" vector returns the target location; x,y for rectangular
  guiders, and either r,a or x,y for imacs.
  The "d" vector returns extra information, the projected vector
  for rectangular, and the gx,gy vector for imacs.
  Global values map_sel and nguiders are used to drive the logic
  and which selected star is used.
  The v and p vectors are used with the global guider field definitions
  to find if we are in a guider field; they are computed for this
  and returned to avoid having to be computed again.
  (see also find_gpos)
*/
/*  Revised to call find_gpos to obtain the guider subscript; and
  also to set the v and d vectors as expected.  This is and should
  be used only in the skywin.c source file, hence it is static.
  Although we do have to doubly compute the v and d that are found,
  we simplify the finding of the guider field value by having all
  that happen in find_gpos.
*/
/*  When commented code is cleaned here and in find_gpos above, all the
  guider finding logic is in find_gpos, with the computation of the target
  location and extra position values remaining here in find_guider; these
  are used only in some specific skywin.c programs below.
*/
int	g=0;
vect3	msv;
vect2	pv;

    if (map_sel < 0) return g;
//    pv = xyproj (maplist[map_sel].v);
    msv = maplist[map_sel].v;
    pv = xyproj (msv);
    g = find_gpos (msv);

    if (nguiders < 1) {	// Old way
// pv.x is -dx2 in the old way
	v->x = g_center_x - pv.x * g_unit_x;
	v->y = g_center_y + pv.y * g_unit_y;
	*d = pv;
    } else {	// nguiders > 0, IMACS way

    double	r, a;
    double	rs, as;
    int	i;
    // int	found;  // maybe make this g?
    double	p, ax;
    double	gx = 0.0;
    double	gy = 0.0;

/* Apply the offsets from gdr[n].xc and .yc to the pv position to correct
  for the offset of IMACS guiders.  */
/* pv is the projected location of the guide star; displace it by an
  offset to pick up the desired guider location which will result in the
  IMACS stars matching the actual mask locations.  */
/* The sense of xc,yc should be the [observed mask center] - [observed
  center star] projected to standard x,y as seen on the sky plot.  */
	i = g - 1;	// g is known here
	pv.x += gdr[i].xc;
	pv.y += gdr[i].yc;
/* The values of xc, yc for guiders 0,1 should be the same here. */

    // get r,a if we need them for polar guiders
	vec2pol (&r, &a, pv);
	a /= Degree;
	gx = gy = 0.0;

    /* We look only in first two guiders for a guide star, not in
     any other subsidary guiders, e.g. center-field; using 2, not nguiders */
//	for (i=g=0; i<2; i++) {
	    if (g < 1) {	/*  There is no guider; make up v.  i is -1 */
//		*v = pv;	/* With no g, this is a default choice. */
		v->x = r;	// Better default for imacs
		v->y = a;	// guiders is polar values

/* The d vector is made up from gx, gy which are already cleared. */
	    } else if (gdr[i].mode & GDR_POLAR) {	/* gdr[i].mode is **Polar** */

		if (gdr[i].mode & GDR_SH) {
		    p = 0.0;
		    ax = 0.0;
		} else {
		    p = gdr[i].px * gdr[i].psize / 2.0;
		    ax = atan (p / gdr[i].position_u.polr_data.rp)/Degree;
		}	
		rs = r;  as = a;	/* the actual star positions */
		if (r < gdr[i].position_u.polr_data.rlo) 
			r = gdr[i].position_u.polr_data.rlo;
		if (r > gdr[i].position_u.polr_data.rhi) 
			r = gdr[i].position_u.polr_data.rhi;
	/* above 2 appeared reversed rlo/rhi ??? */
		if (a < gdr[i].position_u.polr_data.alo) 
			a = gdr[i].position_u.polr_data.alo;
		if (a > gdr[i].position_u.polr_data.ahi) 
			a = gdr[i].position_u.polr_data.ahi;
	
	/* Our first target is the r,a just found */
		v->x = r;
		v->y = a;
	
	/* gx, gy are coordinates in arcseconds in guider window aligned with
	 the polar coordinate system... */
	// THESE ARE PROBABLY WRONG DURING MOTION, MAYBE LATER TOO
		gx = rs * cos ((a - as)*Degree) - r;
		gy = rs * sin ((a - as)*Degree);
	
	    } else {	/* gdr[i].mode is **Rectangular** */
	/* This section has not been tested */
		*v = pv;	// just guessing that this is correct...
	
		gx = pv.x - 
			(gdr[i].position_u.rect_data.xlo +
			 gdr[i].position_u.rect_data.xhi)/2.0;
		gy = pv.y - 
			(gdr[i].position_u.rect_data.ylo +
			 gdr[i].position_u.rect_data.yhi)/2.0;
	/* THAT PROBABLY HAS TO BE ROTATED BY map_pa ?? or 90 degrees? */
	    }	/* end guider mode selection */

    /* The gx,gy found is the secondary vector */
	d->x = gx;
	d->y = gy;

    }	// end nguiders conditional
    return  g;
}

/* ---------------------------------------------------------------------- */

static void redraw_skytext(Display* disp)
{
//static char buf[128];
/* DEBUG change to provide large enough buffer; real problem is that
  things put into buffer are wrong and too large.  */
static char buf[512];
  char   name[32];
  char   nfld[80];
  char   pfld[80];
  char   xfld[80];
  double ra, dc;
  char	rastr[20], dcstr[20];

// Make sure colors are correct.  They don't get set wrong,
// but background has sometimes become blask for no reason.
    XSetForeground(disp,skywin_gc,app->black);
    XSetBackground(disp,skywin_gc,app->white);

/* #define OLD_TEXT */
/* First, select the old or new guider methods, then find guide star. */
  if (nguiders < 1) {		// The old method

  int    inf;
  vect2  v, pp;

  if (map_sel < 0) return;
/* Definitions for buttons:
  Left (1) -- STAR_OFFSET = Select star, show x,y or out of field
  Center (2) -- STAR_INFO = Obtain catalog information on star
  Right (3) -- STAR_DATA  = Show all star position data
.. */
/* For STAR_OFFSET just show star x,y or out of field; for
  STAR_INFO show name,spect,mag, ra, dec.;
  for STAR_DATA show that plus actual x,y  */
/* Added 2003/8 -- for STAR_OFFSET show full data, plus out of field */

    if (dodebug) sprintf(nfld,"%s,%.8s,V=%.1f%c",
      cat_create_name(name,catalog,maplist[map_sel].id),
      maplist[map_sel].spc,  maplist[map_sel].mag, '\0');
    else  sprintf(nfld,"%s V=%.1f%c",
      cat_create_name(name,catalog,maplist[map_sel].id),
      maplist[map_sel].mag, '\0');

    precess(maplist[map_sel].alpha,maplist[map_sel].delta,cat_equ[catalog],
              &ra,&dc,equinox);
    alpha_str (rastr, ra);
    delta_str (dcstr, dc);
      sprintf(pfld," %s %s%c",rastr,dcstr, '\0');

// New way...
    inf = find_guider (&v, &pp);
    if (dodebug) sprintf(xfld," x=%.2f, y=%.2f", v.x, v.y);
    else  *xfld = '\0';

// Add faint warning for selection; location depends on type...
    if (mapsel_type == STAR_INFO) {	// xy or out of field
//	if (maplist[map_sel].mag > gdr[0].mag) strcat (pfld, " Faint!");
	if (maplist[map_sel].mag > g_maglimit) strcat (pfld, " Faint!");
	if (maplist[map_sel].mag < g_brightlim) strcat (pfld, " Bright!");
    } else {
	if (maplist[map_sel].mag > g_maglimit) strcat (xfld, " Faint!");
	if (maplist[map_sel].mag < g_brightlim) strcat (xfld, " Bright!");
    }

    if (mapsel_type == STAR_OFFSET) {	// xy or out of field
// 	if (inf > -1) strcpy (buf, xfld);
//	if (inf < 1) strcat (xfld, " OUT OF FIELD");
//	sprintf (buf, "%s%s%s" nfld, pfld, xfld);
//	if (inf > 0) strcpy (buf, xfld);
//	else  strcpy (buf, "OUT OF FIELD");
	if (inf < 1) strcat (xfld, " OUT OF FIELD");
	sprintf (buf, "%s%s", nfld, xfld);
    } else if (mapsel_type == STAR_INFO) {	// catalog information
	sprintf (buf, "%s%s", nfld, pfld);
    } else {			// all information
	sprintf (buf, "%s%s%s", nfld, pfld, xfld);
    }

// #endif
  } else {			// The new method
/* This is the IMACS code version...  */

#ifdef  NOCATS
  char  xbuf[64];
#endif

  int	i;
  int	found=0;
//  double  rs, as;
//  double  gx, gy;
vect2  pp;   // position
vect2  gg;   // star in guider location


if (map_sel < 0) {
  if (mapsel_type == STAR_OFFSET && found != 0) 
    XDrawImageString(disp,skywin.win,skywin_gc,3,map_h-3,buf,strlen(buf));
/*    XDrawImageString(skywin.disp,skywin.win,skywin_gc,3,map_h-3,buf,strlen(buf)); */
  return;
}
/* Note:  map_sel is subscript of selected closest star */

/* change mapsel stuff here.
  Get star position information.  If starinfo, report catalog info; if
  stardata, report that plus xy (here there is no xy) and if staroffset,
  report positions or out of field.
.. */

/* Get nfld -- the star number, name, spectrum and magnitude */
#ifndef  NOCATS
    if (dodebug) sprintf(nfld,"%d:%s,%s,V=%.2f", map_sel,
      cat_create_name(name,catalog,maplist[map_sel].id),
      maplist[map_sel].spc,  maplist[map_sel].mag);
#else
    memset (xbuf, 0, 64);
    hex4bin (xbuf, maplist[map_sel].spc, 8);
    if (dodebug > 1) sprintf(nfld,"%d:%s,%.8s(%.16s),V=%.2f", map_sel,
      cat_create_name(name,catalog,maplist[map_sel].id),
      maplist[map_sel].spc, xbuf,  maplist[map_sel].mag);
    else if (dodebug) sprintf(nfld,"%d:%s,%s,V=%.2f", map_sel,
      cat_create_name(name,catalog,maplist[map_sel].id),
      maplist[map_sel].spc,     maplist[map_sel].mag);
#endif
/* Debug test -- guessing that the internet server does not supply the
  spectrum and is giving garbage there.. */
//    if (dodebug) sprintf(nfld,"%d:%s,V=%.1f", map_sel,
//      cat_create_name(name,catalog,maplist[map_sel].id),
//      maplist[map_sel].mag);
#ifndef  NOCATS
// temporarily change nocats sense for debug
    else  sprintf(nfld,"%d:%s %s V=%.2f", map_sel,
      cat_create_name(name,catalog,maplist[map_sel].id),
      maplist[map_sel].spc,  maplist[map_sel].mag);
#else
// When debugging with catalog server, state the spectrum also...
//    else  sprintf(nfld,"%d:%s (%8s = %.16s) V=%.1f", map_sel,
    else  sprintf(nfld,"%d:%s %s V=%.2f", map_sel,
      cat_create_name(name,catalog,maplist[map_sel].id),
      maplist[map_sel].spc,
//      xbuf,
      maplist[map_sel].mag);
#endif

/* Get pfld -- The standard ra/dec for the star */
    precess(maplist[map_sel].alpha,maplist[map_sel].delta,cat_equ[catalog],
              &ra,&dc,equinox);
    alpha_str (rastr, ra);
    delta_str (dcstr, dc);
    sprintf(pfld," %s %s%c",rastr,dcstr, '\0');

/* Find xfld -- whether star is in guider, and if so, where... */
    found = find_guider (&pp, &gg);

  if (found) {
    i = found - 1;
    if (gdr[i].mode & GDR_POLAR) {
//	sprintf (xfld, " G=%d, r=%.1f, a=%.1f", i+1, pp.x, pp.y);
	sprintf (xfld, " G=%d r=%.1f a=%.1f  gx=%.1f gy=%.1f",
		i+1, pp.x, pp.y,  gg.x, gg.y);
    } else {	/* gdr[i].mode is **Rectangular** */
	sprintf (xfld, " G=%d dx=%.1f dy=%.1f", i+1, pp.x, pp.y);
    }	/* end guider mode selection */
//  if (maplist[map_sel].mag > gdr[i].mag) strcat (xfld, " Too Faint!");
/* If dodebug is not on, we drop all that xfld guider position stuff
  since the user is not possibly intereseted in it, and the xfld label
  is too long to print anyway!  */
    if (dodebug == 0) *xfld = '\0';
    if (maplist[map_sel].mag > gdr[i].mag) strcat (xfld, " Faint!");
    if (maplist[map_sel].mag < gdr[i].bmag) strcat (xfld, " Bright!");
  } else {	// i.e. not found...
    strcpy(xfld," OUT OF FIELD");
  }

/* Make up the actual buffer to show based on mapsel_type... */
    if (mapsel_type == STAR_OFFSET) {	// xy or out of field
//	sprintf (buf, "%s%s", nfld, xfld);
	sprintf (buf, "%s%s%s", nfld, pfld, xfld);
    } else if (mapsel_type == STAR_INFO) {	// catalog information
	sprintf (buf, "%s%s", nfld, pfld);
    } else {			// all information
	sprintf (buf, "%s%s%s", nfld, pfld, xfld);
    }


  }	// End of nguiders conditional

/* Finally, display the correct string in the window */
  if (Gob->obs->dlevel > 1) {	// debug to see why buffer is in error
	printf (" ..Buffer length = %d\n", (int)strlen(buf));
	printf (" nfld: %s\n", nfld);
	printf (" pfld: %s\n", pfld);
	printf (" xfld: %s\n", xfld);
	printf (" Buf: %s\n", buf);
	fflush (stdout);
  }
  XDrawImageString(disp,skywin.win,skywin_gc,3,map_h-3,buf,strlen(buf));
}

/* ---------------------------------------------------------------------- */

static void redraw_buttons(void)
{
  CBX_UpdateButton(&zibut);
  CBX_UpdateButton(&zobut);
// was.. UpdateSButton on the above

    if (recent) {	// enable and update
	rcbut.bg = app->yellow;		// Brighter enabled button
	CBX_EnableButton (&rcbut);
    } else {		// disable, don't update
	CBX_DisableButton (&rcbut, CBX_DISABLED);
    }

}

/* ---------------------------------------------------------------------- */


/* ---------------------------------------------------------------------- */

/* A non-global entry to set default position for guiders at half
the field limits; used only locally at present, and here to avoid
lengthy code duplication in 2 locations where this is used. */

/* This is where the guider accessable field sizes are finally
  determined from the data file values...  Needs to be called
  when we read and initialize camera data  */

/* was: void  set_default_guiders(Display* display) */
void  set_default_guiders (void)
{
int  i;

// NOTE that this is only effective for new style, where nguiders > 0
    for (i=0; i<nguiders; i++) {
	if (gdr[i].mode & GDR_POLAR) {	/* Polar data */
	    gdr[i].position_u.polr_data.rp = 0.5 * (
			gdr[i].position_u.polr_data.rlo +
			gdr[i].position_u.polr_data.rhi );
	    gdr[i].position_u.polr_data.ap = 0.5 * (
			gdr[i].position_u.polr_data.alo +
			gdr[i].position_u.polr_data.ahi );
	} else {		/* Rectangular data */
	    gdr[i].position_u.rect_data.xp = 0.5 * (
			 gdr[i].position_u.rect_data.xlo +
			 gdr[i].position_u.rect_data.xhi );
	    gdr[i].position_u.rect_data.yp = 0.5 * (
			 gdr[i].position_u.rect_data.ylo +
			 gdr[i].position_u.rect_data.yhi );
	}
    }
//    update_guider_windows ();
/* Actually, we don't use the display variable, yet. */
}

/* ---------------------------------------------------------------------- */
#ifdef  DBconf
static void  config_skywin (int x, int y)
#else
static void  config_skywin (void)
#endif
{
XWindowAttributes xwa;
#ifdef  DBconf
static  int  xx=0;
static  int  yy=0;
#endif

#ifdef  DBconf
// Generate a message if we have moved.
    if (x != xx || y != yy) {
	printf (" skywin move from %d, %d to %d, %d\n", xx, yy, x, y);
	xx = x;
	yy = y;
	return;
    }
#endif
//    skydraw |= 1;	// going to need a redraw for this event.
    XGetWindowAttributes (skywin.disp, skywin.win, &xwa);
    if ((xwa.width==map_w) && (xwa.height==map_h)) return;  // break;
// Catalog menu is stuck on left side, does not move here

// Select guidestar button

// Features menu
    XMoveWindow (skywin.disp, fxmenu.win, (xwa.width-FXw)/2, FXy);
    XMoveWindow (skywin.disp, rcbut.win, xwa.width/2+RCx, RCy);

// Center set button

// Zoom in/out buttons
    XMoveWindow (skywin.disp,zobut.win, xwa.width-Zw-5,     2);
    XMoveWindow (skywin.disp,zibut.win, xwa.width-Zw-5-Zw-5,2);
// Also need to move features menu and other such buttons!
    skydraw |= 2;	// was.. 2;
// Will need 3 above if expose does not follow a configure...



}

/* ---------------------------------------------------------------------- */
// This is local to start_skywin, for now.  May add prototype later...
static	int	astrcmp (char* a, char* b)
{
/* Compare alphabetic characters in a and b.  Return -1 if a less,
  0 if equal, and +1 if a more than b.  Stop at end of string, and
  return equal if no difference has been found. */

    for (; *a != AN && *b != AN; a++,b++) {
//	if (*a == *b) continue;			// pass equal
	while (!isalpha(*a) && *a != AN) a++;	// skip non-alpha
	while (!isalpha(*b) && *b != AN) b++;
//	if (*a == *b) continue;			// pass equal
	if (tolower(*a) == tolower(*b)) continue; // pass equal non case sens.
	if (*a == AN || *b == AN) break;	// leave if ended
// Here, we have an inequality...
	if (*a < *b) return -1;
	else  return  1;
    }
    return  0;
}

/* ---------------------------------------------------------------------- */
static void  start_skywin (void)
{

  int        x,y;
  int        w,h;
  int        sw, sh;
  int        i;
  XSizeHints hint;

// printf (" #- Starting start_skywin.\n");  fflush(stdout);	// debug

/*      CBX_GetGeometry(win,&x,&y,&w,&h); */
/* Above would get geometry of root window, to put this one on relative
 to the root.  However, we wish to anchor the window to the bottom of the
 screen, so we get the position more directly. */
/*      hint.width  = 400;
    hint.height = 400;  */
/* Making the window a little larger... */
    w = 800;   // 750;
    h = 650;
    hint.width  = w;
    hint.height = h;
//      CBX_GetXY (CBX_BOTTOMRIGHT, &x, &y, w, h);			//<==
// CHANGE to proper new thing for getting x,y location...
//	cbx_GetXY (CBX_BOTTOMRIGHT, &x, &y, w, h, 0);
    x = 200;
    y = 250;
/* Doesn't matter, x,y get set by position word in setup anyway... */

/*      hint.x      = imax(x-hint.width-15,0); */
    hint.x = x;
/* Set to left, 15 is 3 border widths, this win + 2 sides of new win */
/*      hint.y      = y-hint.height-55+h; */
    hint.y = y;
/* Bottom at same height; y+h is bottom of this window; */
/* 55 is 2x25 top borders + 5 bottom border */
    if (hint.y < 0) hint.y = y-25;
    hint.min_width  = 400;   hint.min_height = 300;
//      hint.max_width  = 815;   hint.max_height = 750;
    sw = DisplayWidth(mwin.disp,DefaultScreen(mwin.disp));
    sh = DisplayHeight(mwin.disp,DefaultScreen(mwin.disp));
    hint.max_width  = sw;   hint.max_height = sh - 55;
    hint.flags = PPosition | PSize | PMinSize | PMaxSize;

/* We change the creation of skywin here, all the window operation is
  now done in the create subroutine, and "skywin" refers to a mainwindow
  data structure.  */

    alpha = RAsave;
    delta = DECsave;

// printf (" #- Before create skywin\n"); fflush(stdout); // debug

    create_skywin (hint);
    if (!skywin_active) return;   // got error starting, abandon.
    map_pa = -pa*Degree;

// printf (" #- start_skywin loc. 7.\n");  fflush(stdout);	// debug
//      if (nguiders > 0) GMflag = 7;  /* Initialize guider locations */
/*      redraw_skywin(display,True); */  /* (delay this) */
/* Create two guider windows, too... */
    for (i=0; i<4; i++) {
	gdr[i].sel = -1;
	gdr[i].used = 0;
	gdr[i].size = 0;
	gdr[i].tbl = NULL;
    }

/* We need objects, they may have changed or not been read since we
  have been absent, so just unconditionally read all objects here.  */
/* NOTE -- it is possible to track whether a non-null object queue
  pointer has been possibly changed by setting a change flag on it in
  the button handler (handle_button currently) where it would call
  to fill objects if skywin is active.  Then only fill if the object
  queue is empty (new data) or the flag is on.  */
    if (Gob != NULL) {
	fill_objects (Gob->obs);
    }

/* Need to change camera set-up, since obs->Camera no longer exists.
  Basically, the camera name can be IMACS-L, IMACS-S, LDSS3 only.
  What we will do is to get the instrument name from the obs structure,
  and using its name, find which pmap name is to be used to match the
  data, and then get that one.  */
/* Since the compare is case insensitive and compares only to the
  length of the shortest string, the instrument name itself is
  probably sufficient for the purpose... */

    if (Gob != NULL) {
    char	*cnam = NULL;
    Obs	*obs;
    section	*cam, *sp;
	obs = Gob->obs;
	    if (obs != NULL) {
		if (obs->Instr_mt != NULL) cnam = obs->Instr_mt->name;
//		printf ("Looking for camera %s in skywin.\n", cnam);
// if frontend is not null, make cnam the front end name...
#ifdef  GISCAM
		if (obs->Frontend != NULL) cnam = obs->Frontend->name;
#endif
// Could replace with ObInstN (obs) or ObFrontN (obs) function.
	    }
	    if (cnam != NULL) {
		cam = GDAT->camera;
		for (sp=cam; sp != NULL; sp = sp->next) {
	// See if we have name in sp equal to camera name in obs
		    if (!astrcmp(sp->name, cnam)) {
	if (obs->dlevel > 0)
	  printf ("Camera %s found as %s.\n", cnam, sp->name);
			GDAT->camera = sp;	// The actual set of camera...
	// find index in csections[i] of sp;
			for (i=0; i<nsects[MNU_CAM]; i++) {
			    if (sp == csections[i]) break;
			}
//			if (i<nsects[MNU_CAM]) cammenu.sel = i;
// setting cammenu only applies in gmap and its clones, that is
// a menu in the gmap main window, not used here.
	// NOTE - don't know if that is really necessary.
	//		setup_defaults ();
			get_camera_defaults ();
			break;
		    }
	//printf ("Not camera %s; loop again...\n", sp->name);
		    if (sp->next == cam) break;	// circular stopper
		}
	    }
    }


// printf (" #- start_skywin loc. 9.\n");  fflush(stdout);	// debug

/* Now, draw all the windows (delayed from above) */
// FAULT IN FOLLOWING CALL!!
#ifdef  DBGsd
//    printf (" Call redraw_skywin from startup\n");
#endif
//    redraw_skywin(skywin.disp,True);
// The data is not ready yet to draw sky win, it will be done
// in the loop instead.
    skydraw |= 3;	// Do it in the loop.

if (!catalog) printf (" >- Start skywin, catalog=%d\n", catalog); //debug
// printf (" #- Ending start_skywin.\n");  fflush(stdout);	// debug

}


/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
u_long	get_named_color (Display* disp, char* name, u_long dflt_color)
/* Return pixval for the named color */
{
Colormap	cmap = app->cmap;
Status		s;
u_long		pval;
XColor		xcol, ycol;

    s = XAllocNamedColor (disp, cmap, name, &xcol, &ycol);
    if (app->vdepth > 8) {
	float f=0.975; 
	xcol.red  =(int)(xcol.red  *f);
	xcol.green=(int)(xcol.green*f);
	xcol.blue =(int)(xcol.blue *f);
	s = XAllocColor (disp, cmap, &xcol);
    }
    if (!s) pval = dflt_color;
    else    pval = xcol.pixel;

    return  pval;
}

/* ---------------------------------------------------------------------- */
void	setmwbackground (MainWindow mw, u_long color)
{
    mw.bg = color;
    XSetWindowBackground (mw.disp, mw.win, mw.bg);
}

/* ---------------------------------------------------------------------- */


// void create_skywin(Display* display,GC gc)
void	create_skywin (XSizeHints  hint)
{
// For the create_skywin subroutine...
GC	gc;
  XWindowAttributes xwa;
//  double  sx=0.0;   /* Scale maximum size */

// fprintf (stderr, " ## create_skywin start.\n"); fflush(stderr);	// debug

/* Here is where we really do it, moved from start_skywin */

//       skywin = XCreateSimpleWindow(mwin.disp,DefaultRootWindow(mwin.disp),
//         hint.x,hint.y,hint.width,hint.height,2,app->black,app->SKY_COLOR);

    CBX_OpenMainWindow (
	&skywin,
//	CBX_BOTTOMRIGHT,
//	CBX_BOTTOMLEFT,
	CBX_MIDLEFT,
	hint.width, hint.height,
	&hint,
	"XMap-Sky",  "Sky",
	False);		// Makes this a dependent window...

// need to set colors
// may also need to set button input? ...
//       XSelectInput(mwin.disp,skywin,StructureNotifyMask | ExposureMask 
//                    | ButtonPressMask);

    if (skycolor == 0)
	skycolor = get_named_color (skywin.disp, "lightskyblue", app->beige);
    setmwbackground (skywin, skycolor);

// THIS COLOR SETTING IS OBSERVED TO WORK...

/* Correct the mask setting normally given a main window to allow
  this window to receive mouse clicks on stars and locations... */
    CBX_SelectInput (&skywin, StructureNotifyMask |
		ExposureMask  |  ButtonPressMask );

    gc = skywin.gfs.gc;		// get a copy of that...

// printf (" ## create_skywin Loc. 1.\n"); fflush(stderr);	// debug

// fprintf (stderr, " ** create_skywin.  Display = %8.8lx\n",
// 	skywin.disp);  fflush (stderr);  //debug

    XGetWindowAttributes(skywin.disp, skywin.win, &xwa);


//mnublock	catfield = {
//	CAx, CAy,  CAw, CAh, NULL,
//	NULL, M_VOID, &camenu, Leftside };

//  (void)CBX_CreateDropMenu(&swin, &camenu, CAx,CAy,CAw,CAh,NULL);
//  CBX_CreateAutoDrop_Ext(&mwin, &camenu, skywin, mwin.gfs,

// fprintf (stderr, " ## create_skywin Loc. 2.\n"); fflush(stderr);	// debug

/*  Creation of Catalog Menu was formerly here; now it is now below,
  just before we set scale and turn on the map.  We need to select the
  catalog and try to open it before creating the menu to enable the
  automatic default to network catalog retrieval.    */

/*  Features Menu  */

  CBX_CreateDropMenu (&skywin, &fxmenu, 
		(xwa.width-FXw)/2,FXy,FXw,FXh, "Features");
  (void)CBX_AddMenuEntry(&fxmenu,"N-vector", CBX_CHECKED); //FX_NV, CBX_NORMAL);
    map_nv = 1;
// Defalut changed per Dressler, 2004/01 - KDC
#ifdef  UPARO
  (void)CBX_AddMenuEntry(&fxmenu,"Up-vector", CBX_NORMAL); //FX_UV, CBX_NORMAL);
#endif
    map_uv = 0;
#ifdef  ZNTHDOT
  (void)CBX_AddMenuEntry(&fxmenu,"Z-dot", CBX_NORMAL); //FX_ZD, CBX_NORMAL);
#endif
    map_zd = 0;
//  (void)CBX_AddMenuEntry(&fxmenu,"Random offset", FX_RO, CBX_DISABLED);
  (void)CBX_AddMenuEntry(&fxmenu,"Pan center", CBX_NORMAL); //FX_PN, CBX_NORMAL);
    map_pn = 0;
  (void)CBX_AddMenuEntry(&fxmenu,"Star Colors", CBX_NORMAL); //FX_SC, CBX_CHECKED);
    map_sc = 0;
// Default changed 2004/01 to match USNO catalog - KDC
  (void)CBX_AddMenuEntry(&fxmenu,"Plot Alignment", CBX_CHECKED); //FX_AP, CBX_CHECKED);
    map_ap = 1;
  (void)CBX_AddMenuEntry(&fxmenu,"Plot Objects", CBX_NORMAL); //FX_OP, CBX_NORMAL);
    map_op = 0;
  (void)CBX_AddMenuEntry(&fxmenu,"Prin. Guider", CBX_CHECKED); //FX_GP, CBX_CHECKED);
    map_pg = 1;
  (void)CBX_AddMenuEntry(&fxmenu,"Sec. Guider", CBX_CHECKED); //FX_GS, CBX_CHECKED);
    map_sg = 1;
//  (void)CBX_AddMenuEntry(&fxmenu,"Ctr. Guider", CBX_NORMAL); //FX_GC, CBX_NORMAL);
//    map_cg = 0;
  (void)CBX_AddMenuEntry(&fxmenu,"Camera FOV", CBX_CHECKED); //FX_CF, CBX_CHECKED);
    map_cf = 1;
/* We drop the close window option for now.  The handler is not clearing
  the option flag bit in the main window option menu, so that menu is
  still displayed with the window active when it is not.  A better method
  of flagging the window activity should be implemented. */
/* Re-introduce the quit window option.  Now properly handles the menu
  option and the closing of the window. */
  (void)CBX_AddMenuEntry(&fxmenu,"Quit Window", CBX_NORMAL); //FX_KD, CBX_NORMAL);
//    map_kd = 0;


//  CBX_CreateSButton(&zobut,skywin,gc,xwa.width-Zw-5,     2,Zw,Zh,CBX_UP_TRI);
//  CBX_CreateSButton(&zibut,skywin,gc,xwa.width-Zw-5-Zw-5,2,Zw,Zh,CBX_DN_TRI);

// printf (" ## create_skywin Loc. 4.\n"); fflush(stdout);	// debug

/* Zoom out/in buttons */

    CBX_CreateSButton (&skywin,&zobut, xwa.width-Zw-5,     2,Zw,Zh,CBX_UP_TRI);
    CBX_CreateSButton (&skywin,&zibut, xwa.width-Zw-5-Zw-5,2,Zw,Zh,CBX_DN_TRI);

/* Centering feature recenter button */
    CBX_CreateButton (&skywin,&rcbut, xwa.width/2+RCx,RCy, 
			RCw,RCh, "ReCenter");
// add rcx,rcy,rcw,rch to skywin.h first.

/*  Pan select center button */
//    MakeButton (&PCbutn);
// See analogue for DUNbutn in xmap.c

/* Guide star select button */
//    MakeButton (&GSbutn);


/* Set the proper default catalog for the attempt to open the catalog
  which follows.  */
    catalog = camenu.sel = default_cat;  // needed for proper catalog test.
/* Alternately, we could use default_cat instead of catmenu.sel in the
  cat_open call(s) below.  Then use the selected catalog when handling
  the selection from the pop-up later.  The pop-up will not be created
  if we don't have the default catalog present on disk.  */

/* The catalog variable is set above because the -i test flag sets
  USEXMLRPC and bypasses the open test which otherwise is used to set
  the catalog variable.  Thus, it is not always redundant.  The -i flag
  is only active in the "tintgui" edition of the process.  */

//#ifdef  EXCAT
//  if (!USEXMLRPC) {
//#endif
#ifdef  EXCAT
/* If the USEXMLRPC has not been tested, we try to open catalogs and if we
  succeede, set the flag to tested and off; if no success, set it to
  tested and on, and skip trying to open.  If it has been tested, if on
  we skip opening, and if off, we do the normal open which follows.
  Bit 1 == tested;  Bit 2 == XMLRPC is on.      */
    if ((USEXMLRPC & 1) == 0) {  // non-tested, test it

	if (!cat_open(camenu.sel)) {               /* try open catalog */
char buffer[1024];
	    sprintf(buffer," local cat_open (%s) failed", catpath() );
	    fprintf(stderr,"%s\n",buffer); fflush(stderr);
	    fprintf (stderr, " If you want to use a local disk catalog,\n");
// The CATPATH method is now primary.  KDC 2011/11-14
//#ifdef  NOCATS
	    fprintf (stderr, " Make sure \"CATPATH\" is defined correctly.\n");
//#else
//	    fprintf (stderr, " Make sure \"GMAPCATS\" is defined correctly, or");
//	    fprintf (stderr, " that \"../CATS\" is linked to your catalog data.\n");
//#endif
#ifdef DEBUG
	    adebug(__FILE__, buffer);
#endif
	    USEXMLRPC = 3;	// Set to use network catalog
	    fprintf (stderr, " Using network catalog feature.\n");
//	    catalog = CAT_USNO_AO;	// redundant (1)
	} else {  // open success, set tested and off
// catalog was set upon successful call to cat_open above
	    USEXMLRPC = 1;
// Bit 1 == tested;  Bit 2 == Using network catalog
	}

    } else {
	if (Gob->obs->dlevel > 0)
	  fprintf (stderr, " Network catalog status: %d\n", USEXMLRPC);
    }  // it has now been tested and set, and open is done if possible


#else  // the non-EXCAT case (old way - just try to open or fail.)
    if (!cat_open(camenu.sel)) {               /* open catalog              */
	char buffer[1024];
	sprintf(buffer," cat_open (%s) failed", catpath() );
// was..		genv3 ("GMAPCATS", "CATPATH", "../CATS") );
	fprintf(stderr,"%s\n",buffer); fflush(stderr);
// The CATPATH method is now primary.  KDC 2011/11-14
//#ifdef  NOCATS
	fprintf (stderr, " Make sure \"CATPATH\" is defined correctly.\n");
//#else
//	fprintf (stderr, " Make sure \"GMAPCATS\" is defined correctly, or");
//	fprintf (stderr, " that \"../CATS\" is linked to your catalog data.\n");
//#endif
#ifdef DEBUG
	adebug(__FILE__, buffer);
#endif
//	done=True; // skywin=0;
	skywin_active = map_savn = 0;	// Drop the window
	CBX_CloseMainWindow (&skywin);
	    CBX_CheckMenu (&optmenu, OPT_SKYWIN, skywin_active);
	return;
    }
//#ifdef  EXCAT
//  }
//#endif
#endif  // on EXCAT
    skywin_gc = gc;
//  printf (" ## create_skywin Loc. 5.\n"); fflush(stdout);	// debug

// THESE SECTIONS, ABOVE AND BELOW, NEED TO BE PLACED IN A SUBROUTINE
// TO BE CALLED HERE, AND FROM THE MENU HANDLER FOR THE TELESCOPE,
// CAMERA AND GUIDER MENUS WHEN DEFAULTS CHANGE.
  get_camera_defaults ();
//    set_default_guiders();
// The above sets the radius for the principal guider,
// which has no radial motion.
// was..    set_default_guiders(display);


/* Catalog sub-menu relocated here so USEXMLRPC flag if present will have
  been set when the menu (if any) is created.  */

/*  Catalog Menu */
//#ifdef  EXCAT
#if defined (EXCAT) && defined (HACKSERVER)
//  if (!USEXMLRPC) {
//  if ((USEXMLRPC & 2) == 0) {
  if (USEXMLRPC == 1) {  // tested and not network feature, can have menu
#endif
  CBX_CreatePopupMenu(&skywin, &camenu, CAx,CAy,CAw,CAh);
  (void)CBX_AddMenuEntry(&camenu,"SAO",    CBX_NORMAL); //CAT_SAO, CBX_NORMAL);
  (void)CBX_AddMenuEntry(&camenu,"PPM",    CBX_NORMAL); //CAT_PPM, CBX_NORMAL);
  (void)CBX_AddMenuEntry(&camenu,"GSC",    CBX_NORMAL); //CAT_GSC, CBX_NORMAL);
  (void)CBX_AddMenuEntry(&camenu,"GSC+",   CBX_NORMAL); //CAT_GSCP, CBX_NORMAL);
// The USNO_SA entry in the catalog is disabled, to keep the enum
// count the same in the catalog program and header file.  We could
// remove it if the menu count value were still a parameter...
  (void)CBX_AddMenuEntry(&camenu,"USNO-SA",CBX_DISABLED); //CAT_USNO_SA,CBX_NORMAL);
  (void)CBX_AddMenuEntry(&camenu,"USNO-A", CBX_NORMAL); //CAT_USNO_AO,CBX_NORMAL);
    camenu.sel = default_cat;
//    camenu.sel = catalog = default_cat;

//    if (TESTMODE & 1) camenu.sel = CAT_SAO;	// Safe catalog


// printf (" ## create_skywin Loc. 3.\n"); fflush(stdout);	// debug
//#ifdef  EXCAT
#if defined (EXCAT) && defined (HACKSERVER)
  } else {  // can only have the default catalog, no menu created
    catalog = CAT_USNO_AO;	// The only catalog on the server
// above is redundant (2)
  }
#endif

/* The catalog pop-up menu above was relocated to here to enable the
  network catalog feature to be enabled by default if no catalog was
  found on disk.  In that case, no pop-up menu would be created.  */

/* Set default scale and start up the map... */
  {
    int  i;
    double c,g,h;

    c = 0.5 * sx / (float)xwa.height;   // Global default scale
// Factor of 0.5 above to allow g to determine scale below if the
// instrument is specified.

/* Find scale dependent on the instrument */
//    i = InstIndx (Gob->obs->Instr_mt->name);
    i = InstIndx (ObInstN(Gob->obs));
    if (i == I_imacs) {
	for (g=0.0,i=0; i<2; i++) g += gdr[i].position_u.polr_data.rhi;
    } else if (i == I_lds3) {
	h = fabs (gdr[0].position_u.rect_data.xhi);
	h = dmax (h, fabs (gdr[0].position_u.rect_data.xlo));
	g = fabs (gdr[0].position_u.rect_data.yhi);
	g = dmax (g, fabs (gdr[0].position_u.rect_data.ylo));
	g = hypot (h, g);
    } else {
	g = sx;	// default scale
    }
    g = 1.2 * g / (float)xwa.height;
//    printf (" >> Scale computation c = %.2f, g = %.2f\n", c, g);	// debug
// c=1.58; g=3.52 IMACS, 1.72 LDSS3;
    map_init (&skywin, dmax(c,g), True);
/* Second argument is initial scale, unknown units - maybe arcsec/pixel */
  }
}

/* ---------------------------------------------------------------------- */

void get_camera_defaults (void)
{
/* Set up the camera and guider defaults in global section */

/* Get guider file information */
  {
    int  i, m;
    double  rx, ry;	/* size value */
    char buffer[256];
//    char inifile[256];
//    sprintf(inifile,"%s/guider.cfg",genv2("PWD","."));
    section  *sec;
    sec = GDAT->guider;
    (void)get_gstring(sec,"name",g_name,"Guider?");

/* New guider information reading */
//    nguiders = (int)get_glong (sec, "nguiders", 1);

    nguiders = (int)get_glong (sec, "nguiders", 0);
/* Change things for guider compatability:
  if nguiders < 1, we have "old style" data; read it in as was
  done previously.  Otherwise, we have new style, and we will
  be displaying IMACS style guiders.  */

    if (nguiders < 1) {
	g_offaxis_x = (float)get_gdouble(sec,"offaxis-x",0.);
	g_offaxis_y = (float)get_gdouble(sec,"offaxis-y",0.);
	g_field_w   = (float)get_gdouble(sec,"field-w",600.);
	g_field_h   = (float)get_gdouble(sec,"field-h",600.);
	g_unit_x    = (float)get_gdouble(sec,"unit-x",1.);
	g_unit_y    = (float)get_gdouble(sec,"unit-y",1.);
	g_center_x  = (int)get_glong(sec,"center-x",0);
	g_center_y  = (int)get_glong(sec,"center-y",0);
	g_maglimit  = (float)get_gdouble(sec,"maglim",17.0);
//printf (" Guider %s mag.limit = %.1f\n", g_name, g_maglimit);  //debug
	g_brightlim = (float)get_gdouble(sec,"brightlim",11.0);

    } else {	// new style (IMACS) guiders
	for (i=0; i<nguiders; i++) {
	    sprintf (buffer, "mode%d", i+1);
	    m = (int)get_glong(sec, buffer, 0);
	    gdr[i].mode = m;
	    if (m & GDR_POLAR) {	/* Polar data */
		sprintf (buffer, "rlo%d", i+1);
		gdr[i].position_u.polr_data.rlo = (float)get_gdouble (sec, buffer, 900.0);
		sprintf (buffer, "rhi%d", i+1);
		gdr[i].position_u.polr_data.rhi = (float)get_gdouble (sec, buffer, 900.0);
		sprintf (buffer, "alo%d", i+1);
		gdr[i].position_u.polr_data.alo = (float)get_gdouble (sec, buffer, 0.0);
		sprintf (buffer, "ahi%d", i+1);
		gdr[i].position_u.polr_data.ahi = (float)get_gdouble (sec, buffer, 40.0);
		rx = gdr[i].position_u.polr_data.rhi;
/* Initialize the position to middle of allowed range */
/* done below in set_default_guiders */
//		gdr[i].position_u.polr_data.rp = 0.5 * (
//		    gdr[i].position_u.polr_data.rlo +
//		    gdr[i].position_u.polr_data.rhi );
//		gdr[i].position_u.polr_data.ap = 0.5 * (
//		    gdr[i].position_u.polr_data.alo +
//		    gdr[i].position_u.polr_data.ahi );
	    } else {		/* Rectangular data */
		sprintf (buffer, "xlo%d", i+1);
		gdr[i].position_u.rect_data.xlo = (float)get_gdouble (sec, buffer, 0.0);
		rx = fabs(gdr[i].position_u.rect_data.xlo);
		sprintf (buffer, "xhi%d", i+1);
		gdr[i].position_u.rect_data.xhi = (float)get_gdouble (sec, buffer, 300.0);
		rx = dmax (rx, fabs(gdr[i].position_u.rect_data.xhi));
		sprintf (buffer, "ylo%d", i+1);
		gdr[i].position_u.rect_data.ylo = (float)get_gdouble (sec, buffer, 0.0);
		ry = fabs(gdr[i].position_u.rect_data.ylo);
		sprintf (buffer, "yhi%d", i+1);
		gdr[i].position_u.rect_data.yhi = (float)get_gdouble (sec, buffer, 300.0);
		ry = dmax (ry, fabs(gdr[i].position_u.rect_data.yhi));
		rx = dmax (rx, ry);
	    }
	    sprintf (buffer, "rot%d", i+1);
	    gdr[i].grot = (float)get_gdouble (sec, buffer, 0.0);
/* Magnitude limit for guiders - stars numerically greater would be
  flagged as "too faint" if selected for that guider.  Default is
  just a guess.  Shack-Hartmann guiders will have a brighter limit... */
	    sprintf (buffer, "mag%d", i+1);
	    gdr[i].mag = (float)get_gdouble (sec, buffer, 17.0);
/* Also, do the bright magnitude limit... */
	    sprintf (buffer, "bmag%d", i+1);
	    gdr[i].bmag = (float)get_gdouble (sec, buffer, 11.0);

    /* Some values are not (yet) subscripted in the configuration file */
	    gdr[i].px = (float)get_gdouble (sec, "px", 512.0);
	    gdr[i].psize = (float)get_gdouble (sec, "psize", 0.2);
	    gdr[i].scale = (float)get_gdouble (sec, "scale", 1.0);
	    gdr[i].xc = (float)get_gdouble (sec, "offaxis-x", 0.0);
	    gdr[i].yc = (float)get_gdouble (sec, "offaxis-y", 0.0);
    /* NOTE - xc, yc values and others are implicitly in arcseconds; we should
      adjust them for scale? */
	    sx = dmax (sx, gdr[i].px * gdr[i].psize * 0.5 + rx);
	    gdr[i].moving = 0;  /* Initialize motion */
	}	/* Loop over guiders */

    }  // End conditional on old/new style guiders
    set_default_guiders ();
//    update_guider_windows ();

  }


  {
    int  i;
//    char inifile[256];
    char buffer[1024];
    section  *sec;
    sec = GDAT->camera;
//    sprintf(inifile,"%s/camera.ini",genv2("PWD","."));
    (void)get_gstring(sec,"name",c_name,"Camera?");
    nchips = (int)get_glong(sec,"nchips",1);
    for (i=0; i<nchips; i++) {
      sprintf(buffer,"choffx%d",i+1);
      choffx[i] = (float)get_gdouble(sec,buffer,0.);
      sprintf(buffer,"choffy%d",i+1);
      choffy[i] = (float)get_gdouble(sec,buffer,0.);
    }
    dimx     = (int)get_glong(sec,"dimx",256);
    dimy     = (int)get_glong(sec,"dimy",256);
    pixscale = (float)get_gdouble(sec,"pixscale",1.);
  }

}

/* ---------------------------------------------------------------------- */

/* Subroutine to draw an arc at given position, radius, angles */
void DrawArc (Display* disp, Window win, GC gc,
	int x, int y,
	float ra, float a1, float a2)
{
int	r;
    r = (int)floor(ra + 0.5);
    XDrawArc (disp, win, gc,
	x-r, y-r, 2*r, 2*r, 
	(int)(a1*64.0), (int)((a2-a1)*64.0) );
/* Angles are start, length */
}

/* ---------------------------------------------------------------------- */

void PlotGuider (Display* disp, Window win, GC gc, 
	guiderdata* gx)
{
/* Plot location of guider accessable field, and guider locations */
int	x, y;
  double  a1,a2,ax;
  double  r1, r2;	// was... , rx;
  double  p;
//  int  i;
//  int  s, c;
//  int  x0, y0;
  int  x1, y1, x2, y2;
//  int  x3, y3, x4, y4;

    XSetForeground(disp,gc,app->blue);
    XSetLineAttributes(disp,gc,1,LineOnOffDash,CapRound,JoinRound);
/* When we add active guider locations in yellow or yellow/blue, we put
  those things into the loop below... */

//	x = map_w/2 - gx->xc / map_scale;
//	y = map_h/2 - gx->yc / map_scale;
// NOTE that the gx->xc is implicitly negated in above code...
    map_pltxy (&x, &y, make2vect (-(double)gx->xc, (double)gx->yc));
    p = gx->px * gx->psize / 2.0;		/* Half size of box */

    if (gx->mode & GDR_POLAR) {
/* Find location and angle values for the arcs and lines... */
//	x = map_w/2 - gx->xc/map_scale;
//	y = map_h/2 - gx->yc/map_scale;
//	p = gx->px * gx->psize / 2.0;

	if (gx->mode & GDR_SH) {
	    a1 = gx->position_u.polr_data.alo;
	    a2 = gx->position_u.polr_data.ahi;
	    r1 = gx->position_u.polr_data.rlo;
	    r2 = gx->position_u.polr_data.rhi;
	} else {
/*	    ax = RADS * atan (p / gx->position_u.polr_data.rp);  */
	    ax = atan (p / gx->position_u.polr_data.rp) / Degree;
	    a1 = gx->position_u.polr_data.alo - ax;
	    a2 = gx->position_u.polr_data.ahi + ax;
	    r1 = gx->position_u.polr_data.rlo - p;
	    r2 = gx->position_u.polr_data.rhi + p;
	}
	DrawArc (disp,win,gc,x,y, r1/map_scale, a1, a2);
	DrawArc (disp,win,gc,x,y, r2/map_scale, a1, a2);
/* Add the radial lines here after finding end points... */
	x1 = x + (Nint)(cos(a1*Degree) * r1/map_scale);
	y1 = y - (Nint)(sin(a1*Degree) * r1/map_scale);
	x2 = x + (Nint)(cos(a1*Degree) * r2/map_scale);
	y2 = y - (Nint)(sin(a1*Degree) * r2/map_scale);
/* NOTE - We may want to change rounding here */
	XDrawLine (disp,win,gc, x1,y1, x2,y2);
	x1 = x + (Nint)(cos(a2*Degree) * r1/map_scale);
	y1 = y - (Nint)(sin(a2*Degree) * r1/map_scale);
	x2 = x + (Nint)(cos(a2*Degree) * r2/map_scale);
	y2 = y - (Nint)(sin(a2*Degree) * r2/map_scale);
	XDrawLine (disp,win,gc, x1,y1, x2,y2);

/* Plot the active position of the window, depending on mode */

    } else {		/* Polar/rectangular */
/* Find and plot the x-y coordinates of a rectangle */
//	x = map_w/2 - gx->xc / map_scale;
//	y = map_h/2 - gx->yc / map_scale;
// Moved to outside the if/else statement...

/* Find corners of box range of motion */
	x1 = x + (Nint)(gx->position_u.rect_data.xlo/map_scale);
	y1 = y - (Nint)(gx->position_u.rect_data.ylo/map_scale);
	x2 = x + (Nint)(gx->position_u.rect_data.xhi/map_scale);
	y2 = y - (Nint)(gx->position_u.rect_data.yhi/map_scale);
	if (x1 != x2 || y1 != y2) {
	    XDrawLine (disp,win,gc, x1,y1, x1,y2);
	    XDrawLine (disp,win,gc, x1,y2, x2,y2);
	    XDrawLine (disp,win,gc, x2,y2, x2,y1);
	    XDrawLine (disp,win,gc, x2,y1, x1,y1);
	}
/* Draw the rectangle itself, active position. */
    }	/* End of polar/rectangular test */
    XSetLineAttributes(disp,gc,1,LineSolid,CapRound,JoinRound);
    XSetForeground(disp,gc,app->black);
}

/* ---------------------------------------------------------------------- */

void  my_rectangle (Display* disp, Window win, GC gc, 
	int x, int y, int w, int h )
/* Replicate XDrawRectangle as much as possible */
{
// typedef struct {
//      short x, y;
// } XPoint;

XPoint  points[5];

	points[0].x = x;
	points[0].y = y;
	points[1].x = x+w;
	points[1].y = y;
	points[2].x = x+w;
	points[2].y = y+h;
	points[3].x = x;
	points[3].y = y+h;
	points[4] = points[0];

	XDrawLines (disp,win,gc, points, 5, CoordModeOrigin);

}

/* It seems that the XDrawRectangle for some unknown reason disregards
  the LineDoubleDash attribute and draws all rectangles as LineSolid.
  We put in this replacement to draw the lines with the proper
  line attribute as in the GC.  */

/* ---------------------------------------------------------------------- */

// Many commented lines are being removed from redraw_skywin; to
// see the old things, look at skywin.c03 and earlier...
void redraw_skywin(Display* disp,Bool recalc)
{
  int    i,x,y,w,h;
  float  r;
  static  int  initial = 1;
vect2	cofs = { 0.0, 0.0 };  // center offset
vect2	chof = { 0.0, 0.0 };  // chip offset

#ifdef  DBrsky
  printf (" -# redraw_skywin start.  At %.3f %.3f %.2f\n", alpha, delta, pa); fflush(stdout);	// debug
#endif

//  if (skywin == 0) return;
//    if (skywin.disp == NULL) return;
    if (disp == NULL) return;
  if (!CBX_IsMapped(skywin.win)) { 
    if (recalc) recalc_on_map=True;
    return;
  }

  redraw_skymenus();
  redraw_buttons();

  if (recalc) {                             /* recalculate map */
//#ifdef  EXCAT
//    if (USEXMLRPC) catalog = CAT_USNO_AO;  // Special debug test
//#endif
#ifdef  DBrsky
    printf (" Calling recalc_map from redraw_skywin.\n");
#endif
    recalc_map(disp); recalc_on_map=False;
/*    map_sel = -1; */	/* Experiment to clear on change */
/* The above was tried to eliminate selection of guide star being
 remembered during a movement.  An explicit reset when any movement
 is done is now implemented.  A better way may be to clear it after
 moving the guider? */
  }

// experiment with no redraw map on recalc?  -- only first times
    if (initial || (recalc == 0) ) {
#ifdef  DBrsky
  printf (" -# Calling map_redraw from redraw_skywin\n");		// debug
#endif
  map_redraw(disp,skywin.win,maglim);
  initial = 1;
    }		// experiment
/*  NOTE - on the above.
  When skywin starts up, it is called from the loop two times with the
  recalc flag on.  Both of these are just setting things up, and later
  we are called with recalc off.  At that time, we need to do the
  map_redraw to actually draw the map things and get running.  The
  initial trys have a wrong location and some missing data.
  .. */

  redraw_skytext(disp);                     /* on top of 'skywin' */

  if (!Gob->obs->IFUmode)  {
// Omit camera field when in IFU mode...
    w = (int)floor(dimx*pixscale/map_scale+0.5); /* redraw camera-FOV */
    h = (int)floor(dimy*pixscale/map_scale+0.5);
    if ((w>1) && (h>1) && (h<map_h) && (w<map_w) &&(map_cf != 0)) {
// CORE DUMP IN NEXT STATEMENT...
    if (skywin_gc == NULL) {
	printf (" #!! skywin_gc NULL!!\n");
	fflush(stdout);
    }
    if (app == NULL) {printf (" #!! app NULL!!\n"); fflush(stdout);}
    XSetForeground(disp,skywin_gc,app->red);
    XSetLineAttributes(disp,skywin_gc,1,LineOnOffDash,CapRound,JoinRound);
    cofs = (nguiders < 1) ? make2vect (g_offaxis_x, g_offaxis_y) :
		make2vect (gdr[0].xc, gdr[0].yc);
/* NOTE that it may be proper above to make cofs zero in the first case
  of nguiders < 1.  However, it is probable that we need to displace the
  camera in all cases in which the guider field is also displaced, which
  will be done by using offaxis values in both cases.
  -- Actually, in telmap.c we have cofs = zero in the first case.  */

/*  Draw the camera field of view.  */
//    if (InstIndx(Gob->obs->Instr_mt->name) == I_lds3) {
    if (InstIndx(ObInstN(Gob->obs)) == I_lds3) {
/* For LDSS, draw a circular field.  Diameter of field is given as
  8.25 arc minutes; convert to arc seconds radius..  */
	r = (8.25 * 60.0 / map_scale) / 2.0;		// radius, arcsec
//	map_pltxy (&x, &y, make2vect (0.0, 0.0));	// center
	map_pltxy (&x, &y, cofs);	// center
	DrawArc (disp, skywin.win, skywin_gc, x, y, r, 0.0, 360.0);
    } else {
/* Otherwise, draw rectangular chips */
      for (i=0; i<nchips; i++) {
        chof = make2vect (-(double)choffx[i], (double)choffy[i]);
        map_pltxy (&x, &y, sum2vect (cofs, chof) );  // Add center, chip offsets
//        map_pltxy (&x, &y, make2vect (-(double)choffx[i], (double)choffy[i]));
    
/* For debug, replace the following with 4 lines using draw line calls
        XDrawRectangle(disp,skywin,skywin_gc,x-w/2,y-h/2,w,h);
.. */
/* That produces a dashed box where XDrawRectangle does not! */
        my_rectangle (disp,skywin.win,skywin_gc,x-w/2,y-h/2,w,h);
/* By either method of plotting, a dashed line which extends off the
display will not be plotted... */
    
      }
    }	// Test for instrument.
    XSetForeground(disp,skywin_gc,app->black);
    XSetLineAttributes(disp,skywin_gc,1,LineSolid,CapRound,JoinRound);
    }
  }  // Test for IFU mode

/* (Re-)Drawing of guider-field */

/* Plot of guider field depends on nguiders; if < 1 we do the 
  "old style" guider plot, and if 1 or more, do the new (IMACS) style */

  if (nguiders < 1) {		// old style (LDSS, others)
      w = (int)floor(g_field_w/map_scale+0.5);   // redraw guider-field
      h = (int)floor(g_field_h/map_scale+0.5);
      if (map_pg && (w>1) && (h>1) && (h<2*map_h) && (w<2*map_w)) {
	XSetForeground(disp,skywin_gc,app->blue);
	XSetLineAttributes(disp,skywin_gc,1,LineOnOffDash,CapRound,JoinRound);
// NOTE - NOTE - using old method, we do not use map_pcx and map_pcy
// plot center data for pan mode.  If needed, revise this code!!!
	x = map_get_x_coord(map_ca+g_offaxis_x/(3600.*15.*cos(map_cd*Degree)));
	y = map_get_y_coord(map_cd+g_offaxis_y/3600.);
//	XDrawRectangle(disp,skywin,skywin_gc,x-w/2,y-h/2,w,h);
	my_rectangle (disp,skywin.win,skywin_gc,x-w/2,y-h/2,w,h);
	XSetForeground(disp,skywin_gc,app->black);
	XSetLineAttributes(disp,skywin_gc,1,LineSolid,CapRound,JoinRound);
      }
  } else  {			// new style (IMACS)

    /* Draw the (new) guider field positions */
      w = (int)floor(900.0/map_scale+0.5);	/* Rough scale of plot */
      if ((w>1) && (w < (5*map_w) )) {	/* Guess it can be seen... */
// above was w < 2*map_w
// Alternate is to put guider2 stuff below outside this if, so the center
// guider is always tested and plotted.
// i.e. restore 2* above, remove comments and put PlotGuider for 2
// outside this if conditional block below.
    /*    for (i=0; i<nguiders; i++) {
	    PlotGuider (disp, skywin.win, skywin_gc, &(gdr[i]));
	}  */		/* Loop over guider fields */
    /* Plot guiders based on feature flags */
	if (map_pg != 0) PlotGuider (disp, skywin.win, skywin_gc, &(gdr[0]));
	if (map_sg != 0) PlotGuider (disp, skywin.win, skywin_gc, &(gdr[1]));
	if (map_cg != 0) PlotGuider (disp, skywin.win, skywin_gc, &(gdr[2]));
    /* Restore "standard" color and line attributes (now in plotguider) */
//	XSetForeground(disp,skywin_gc,black);
//	XSetLineAttributes(disp,skywin_gc,1,LineSolid,CapRound,JoinRound);
      }
  }

/* If IFU mode, plot the IFU windows at proper scale */
    if (Gob->obs->IFUmode) {
	w = (int)floor(5.0/map_scale+0.5);   // Set window dimensions
	h = (int)floor(5.0/map_scale+0.5);
	XSetLineAttributes (disp, skywin_gc, 1,	// Set dashed line ?
			LineOnOffDash, CapRound, JoinRound);
	for (i=0; i<2; i++) {	// Draw each window
	    XSetForeground (disp, skywin_gc,
		(i+1 == Gob->obs->IFUmode) ? app->orange : app->yellow);
	    map_pltxy (&x, &y, make2vect(IFUdx[i], IFUdy[i]));
	    my_rectangle (disp, skywin.win, skywin_gc, x-w/2, y-h/2, w, h);
	}
// Restore line stuff
	XSetForeground(disp,skywin_gc,app->black);
	XSetLineAttributes(disp,skywin_gc,1,LineSolid,CapRound,JoinRound);
    }

#ifdef  DBrsky
  printf (" -# redraw_skywin End.\n"); fflush(stdout);	// debug
#endif
}

/* ---------------------------------------------------------------------- */

static void recalc_map(Display* display)
{
  double            rx,ry;
  double	apc, dpc;
  vect3    pv;
  vect3    nv;
  double   sp, cp;
  XWindowAttributes xwa;
int	nstars;

/*  precess(alpha,delta,equinox,&map_ca,&map_cd,cat_equ[catalog]); */ /* precess */
  map_eq = cat_equ[catalog];
  cur_epoch = get_epoch (0);
  precess(alpha,delta,equinox,&map_ca,&map_cd,map_eq); /* precess */
//  printf (" - recalc: equinox = %.2f  map_eq = %.2f\n", equinox, map_eq);
  
/* Generate the global transform matrix, map_xf */
/* This is similar to smatrix, but we add map_pa rotation, too.
  The sense of x,y,z is here consistant with tang2vec and similar.
  x is along center vector; y is east and z is north in sky. */
/* Although here, y is left and z up on the view surface */
//    pv = sph2vec (map_ca*Hour, map_cd*Degree);
    pv = had2vec (map_ca, map_cd);
    /* map_xf[0] */ map_xf.x = norm3vect (pv);	/* Along reference */
    sincos (map_pa, &sp, &cp);
    nv = make3vect (0.0, -sp, cp);
//    nv = rotate (nv, -map_ca*Hour, 3);
//    nv = rotate (nv,  map_cd*Degree, 2);
    nv = rotate (nv, -map_cd*Degree, 2);
    nv = rotate (nv,  map_ca*Hour, 3);
    /* map_xf[2] */ map_xf.z = norm3vect (nv);
    /* map_xf[1] = norm3vect (cross3vect(map_xf[2], map_xf[0])); */
    map_xf.y = norm3vect (cross3vect(map_xf.z, map_xf.x));
/* SHOULD use {Gob->}obs->mxpos instead of map_xf  -- except that this
  matrix is in map_eq coordinates, and the map_pa angle may differ.  
  Otherwise we could call a common routine to get_mpos for this.  */
    if (Gob != NULL) Gob->obs->mxpos = map_xf;
// Above may be wrong if equinox of map is not equinox of catalog! 
/* See get_mpos in mgutils.c for how that is done.  */

  XGetWindowAttributes(display,skywin.win,&xwa);
  map_w  = xwa.width;
  map_h  = xwa.height;
  rx     = (double)map_w/2. * map_scale;
  ry     = (double)map_h/2. * map_scale;
  maglim = get_mlim(catalog,rx*ry*4.);

  CBX_BusyCursor(skywin.win);
//  CBX_BusyCursor_Ext(display,skywin.win);
//  CBX_BusyCursor_Ext(mwin.disp,skywin.win);
//  (void)map_getlist(map_ca,map_cd,rx,ry,maglim);
// Adjust maglim to get the right number of stars
// NOTE - we need to get list at plot center, not map center!
// map_ca is hours, map_cd is degrees, and at equinox of catalog.

    vec2sph (&apc, &dpc, unproj(make2vect(map_pcx, map_pcy)));
// Maybe should precess from display equinox to catalog equinox?  No.
// We want apc and dpc in catalog epoch, as the transform gives them.
// Could do a precess with that 3-vector from unproj...

// Change apc, dpc from radians to hours, degrees
    apc /= Hour;
    dpc /= Degree;

// First, try a simple integer limit on total stars retrieved...
	nstars = map_getlist (apc, dpc, rx, ry);
	maglim = maplist[nstars-1].mag;  // needed elsewhere
    if (Gob->obs->dlevel > 0) printf ("  .map_getlist got %d stars.\n", nstars);

  CBX_NormalCursor(skywin.win);
//  CBX_NormalCursor_Ext(display,skywin.win);
//  CBX_NormalCursor_Ext(mwin.disp,skywin.win);
}	// end of recalc_map

/* ---------------------------------------------------------------------- */

void handle_camenu (void)	// (Display* display)
{
  int        new;

/* Handle catalog menu.  Open catalog and redraw map and window */
/* NOTE: this is not handled or called when the network catalog window is 
  active, there is no catalog menu in that window.  */
/* NOTE:  when network catalog is active, this will get called if we do not
  also have the HACKSERVER defined.  So we need a call to getlist or equivalent
  to get the new window data up.    */

  new = CBX_HandleMenu(&camenu);	// was HandlePMenu
  if ((new < 0) || (new == catalog)) return;

// printf (" >> Handle catalog %d replacing catalog %d.\n", 
//	new, catalog); //debug

#ifdef  EXCAT
  if (USEXMLRPC == 3) {
/* If we are using external catalog (server) and we have tested for and are
  using an external server, then we need to change catalog, set skydraw,
  and return.    */
	catalog = new;   // since we don't do cat_open
	camenu.sel = new;	// update that menu
//	default_cat = new;	// Need this?  no.
	map_sel = -1;
	adjust_scale();
	skydraw |= 3;
	return;
  }
#endif

  cat_close();
  if (!cat_open(new)) {
    char buffer[1024];
//  sprintf(buffer,"cat_open(%s) failed",genv3("GMAPCATS","CATPATH","../CATS"));
    sprintf(buffer,"cat_open(%s) failed",catpath());
    fprintf(stderr,"%s\n",buffer); fflush(stderr);
#ifdef DEBUG
    adebug(__FILE__,buffer);
#endif
    cat_open(catalog);	// re-open old on error
  } else {  // Appears to open, catalog was set in cat_open
    default_cat = camenu.sel = new;
//    if (catalog != new) printf (" ** Catalog not set ERROR.\n");  // debug
// Use above line if catalog menu changes appear not to work correctly...
    map_sel = -1;
    adjust_scale();
  }
//  redraw_skywin (display,True);
//  skydraw |= 2;
// Add redraw bit (1) for Mac-OS-X redisplay bug
  skydraw |= 3;
#ifdef  DBGsd
  printf (" handled camenu, skydraw=3\n");
#endif
//  redraw_skywin (skywin.disp,True);  // (display,True);
}	// end of handle_camenu

/* ---------------------------------------------------------------------- */

void handle_fxmenu (void)	// (Display* display)
{
  int        new;

/* Handles the features menu, set/clear flags as needed. */

  new = CBX_HandleMenu(&fxmenu);	// was HandleDMenu

  switch (new) {
  case FX_NV:
    if (fxmenu.entry[FX_NV].flag == CBX_NORMAL) {
      fxmenu.entry[FX_NV].flag = CBX_CHECKED;
      map_nv = 1;
    } else {
      fxmenu.entry[FX_NV].flag = CBX_NORMAL;
      map_nv = 0;
    }
    break;
#ifdef  UPARO
  case FX_UV:
    if (fxmenu.entry[FX_UV].flag == CBX_NORMAL) {
      fxmenu.entry[FX_UV].flag = CBX_CHECKED;
      map_uv = 1;
    } else {
      fxmenu.entry[FX_UV].flag = CBX_NORMAL;
      map_uv = 0;
    }
    break;
#endif
#ifdef  ZNTHDOT
  case FX_ZD:
    if (fxmenu.entry[FX_ZD].flag == CBX_NORMAL) {
      fxmenu.entry[FX_ZD].flag = CBX_CHECKED;
      map_zd = 1;
    } else {
      fxmenu.entry[FX_ZD].flag = CBX_NORMAL;
      map_zd = 0;
    }
    break;
#endif
//  case FX_RO:
//    move_offset (
//	uranx (32.0/(3600.0*15.0*cos(delta*Degree))),
//	uranx (32.0/(3600.0)),
//	uranx (2.0) );
//    break;
  case FX_PN:
    if (fxmenu.entry[FX_PN].flag == CBX_NORMAL) {
      fxmenu.entry[FX_PN].flag = CBX_CHECKED;
      map_pn = 1;
    } else {
      fxmenu.entry[FX_PN].flag = CBX_NORMAL;
      map_pn = 0;
// ALSO, we re-zero the display here
      map_pcx = map_pcy = 0.0;
    }
    break;
  case FX_SC:
    if (fxmenu.entry[FX_SC].flag == CBX_NORMAL) {
      fxmenu.entry[FX_SC].flag = CBX_CHECKED;
      map_sc = 1;
    } else {
      fxmenu.entry[FX_SC].flag = CBX_NORMAL;
      map_sc = 0;
    }
    break;
  case FX_AP:
    if (fxmenu.entry[FX_AP].flag & CBX_CHECKED) {	// the new method
      fxmenu.entry[FX_AP].flag = CBX_NORMAL;
      map_ap = 0;
    } else {
      fxmenu.entry[FX_AP].flag = CBX_CHECKED;
      map_ap = 1;
    }
    break;
  case FX_OP:
    if (fxmenu.entry[FX_OP].flag & CBX_CHECKED) {	// the new method
      fxmenu.entry[FX_OP].flag = CBX_NORMAL;
      map_op = 0;
    } else {
      fxmenu.entry[FX_OP].flag = CBX_CHECKED;
      map_op = 1;
    }
    break;
  case FX_GP:
    if (fxmenu.entry[FX_GP].flag == CBX_NORMAL) {
      fxmenu.entry[FX_GP].flag = CBX_CHECKED;
      map_pg = 1;
    } else {
      fxmenu.entry[FX_GP].flag = CBX_NORMAL;
      map_pg = 0;
    }
    break;
  case FX_GS:
    if (fxmenu.entry[FX_GS].flag == CBX_NORMAL) {
      fxmenu.entry[FX_GS].flag = CBX_CHECKED;
      map_sg = 1;
    } else {
      fxmenu.entry[FX_GS].flag = CBX_NORMAL;
      map_sg = 0;
    }
    break;
//  case FX_GC:
//    if (fxmenu.entry[FX_GC].flag == CBX_NORMAL) {
//      fxmenu.entry[FX_GC].flag = CBX_CHECKED;
//      map_cg = 1;
//    } else {
//      fxmenu.entry[FX_GC].flag = CBX_NORMAL;
//      map_cg = 0;
//    }
//    break;
  case FX_CF:
    if (fxmenu.entry[FX_CF].flag == CBX_NORMAL) {
      fxmenu.entry[FX_CF].flag = CBX_CHECKED;
      map_cf = 1;
    } else {
      fxmenu.entry[FX_CF].flag = CBX_NORMAL;
      map_cf = 0;
    }
    break;
  case FX_KD:
    if (fxmenu.entry[FX_KD].flag & CBX_CHECKED) {	// the new method
      fxmenu.entry[FX_KD].flag = CBX_NORMAL;	// Will never do this
      map_kd = 0;
    } else {
      fxmenu.entry[FX_KD].flag = CBX_CHECKED;	// This will kill window
      map_kd = 1;
//	skywin_active = 0;
    }
    break;
  }
//  redraw_skywin (skywin.disp,True);  //(display,True);
//  redraw_skywin (display,True);
//  skydraw |= 2;
// Add redraw bit (1) for Mac-OS-X display refresh bug
  skydraw |= 3;
#ifdef  DBGsd
  printf (" handled fx menu skydraw=3\n");
#endif
}

/* ---------------------------------------------------------------------- */


// void	handle_rcbut (Display* display, int button) {
void	handle_rcbut (void)  {
  if (CBX_HandleButton(&rcbut) == CBX_PRESSED) {

// Put saved values back into current position
    *(double*)RAfield.data = RAsave * RAfield.scale;
    alpha = RAsave;
    *(double*)DECfield.data = DECsave * DECfield.scale;
    delta = DECsave;
    recent = 0;
/* NOTE!  Should now check guide stars for being correct, and clear
  any guide star which is not in a guider field; if autogs is in
  effect, also clear agscan when doing this.  
  To check, use same procedure as done when reading;
  */
//    validate_guiders (Gob->obs);
// This is now moved to redraw_map in skymap

// Also need to update text window values
    HandleTextWindow (winlist, RAfield.ew);
    HandleTextWindow (winlist, DECfield.ew);

    skydraw |= 3;
#ifdef  DBGsd
    printf (" handled rcbut skydraw=3\n");	// debug
#endif
    CBX_ReleaseButton(&rcbut);
  }
  CBX_UpdateButton(&rcbut);
}

/* ---------------------------------------------------------------------- */

//void handle_zibut(Display* display,int button)
void handle_zibut(int button)
{
  if (CBX_HandleButton3(&zibut,button) == CBX_PRESSED) {
    switch (button) {
    case 3:
      map_scale /= 2.00; break;
    case 2:
      map_scale /= 1.50; break;
    case 1:
    default:
      map_scale /= 1.25; break;
    }
    adjust_scale();
//    redraw_skywin (skywin.disp,True);  //(display,True);
//    redraw_skywin (display,True);
    skydraw |= 3;
#ifdef  DBGsd
    printf (" handled zibut skydraw=3\n");
#endif
    CBX_ReleaseButton(&zibut);
  }
  CBX_UpdateButton(&zibut);	// Was UpdateSButton
}

/* ---------------------------------------------------------------------- */
 
//void handle_zobut(Display* display,int button)
void handle_zobut(int button)
{
  if (CBX_HandleButton3(&zobut,button) == CBX_PRESSED) {
    if (!is_maxscale()) {
      switch (button) {
      case 3:
        map_scale *= 2.00; break;
      case 2:
        map_scale *= 1.50; break;
      case 1:
      default:
        map_scale *= 1.25; break;
      }
      adjust_scale();
//      redraw_skywin (skywin.disp,True);  //(display,True); 
//      redraw_skywin (display,True); 
	skydraw |= 3;
#ifdef  DBGsd
	printf (" handled zobut skydraw=3\n");
#endif
    }
    CBX_ReleaseButton(&zobut);
  }
  CBX_UpdateButton(&zobut);	// Was UpdateSButton
}
 
/* ---------------------------------------------------------------------- */

 

/* ---------------------------------------------------------------------- */

#define SAO_CUT   2.00
#define GSC_CUT   0.50
#define USS_CUT   0.15
// #define USA_CUT   0.01
#define USA_CUT   0.014
 
static float get_mlim(int cat,double fov)   /* limiting magnitude         */
{
  float m;
 
  if (!auto_maglim) return man_maglim;
 
  fov /= (3600.*3600.);                     /* convert to square-degrees  */
 
  switch (cat) {                            /* calculate lim.-magnitude   */
  case CAT_SAO:
  case CAT_PPM:
    if (fov > SAO_CUT) m = (float)dmax(5.0,cat_mag[cat]-0.8*log(fov/SAO_CUT));
    else               m = cat_mag[cat];
    break;
  case CAT_GSC:
  case CAT_GSCP:
    if (fov > GSC_CUT) m = (float)dmax(6.5,cat_mag[cat]-1.3*log(fov/GSC_CUT));
    else               m = cat_mag[cat];
    break;
  case CAT_USNO_SA:
    if (fov > USS_CUT) m = (float)dmax(14.,cat_mag[cat]-1.0*log(fov/USS_CUT));
    else               m = cat_mag[cat];
    break;
  case CAT_USNO_AO: 
//    if (fov > USA_CUT) m = (float)dmax(14.,cat_mag[cat]-1.5*log(fov/USA_CUT));
    if (fov > USA_CUT) m = (float)dmax(14.,cat_mag[cat]-1.3*log(fov/USA_CUT));
    else               m = cat_mag[cat];
#ifdef DEBUG
    fprintf(stderr,"fov=%f, m=%f\n",fov,m);
#endif
    break;
  default:
    m = 30.;
    break;
  }
  return m;
}
 
#undef SAO_CUT
#undef GSC_CUT
#undef USS_CUT
#undef USA_CUT


/* ---------------------------------------------------------------------- */

int	check_star (int star)
/* Check whether the given star is a valid guide star; if so return a
  subscript 1 greater than the guider field index (i.e. 1 or possibly 2),
  and if not return zero.  This program checks the field position using
  the "find_guider" subroutine, and also checks the magnitude limits.
  It is called by the map_guideset program, as well as the automatic guide
  star program in skymap.c.						*/
{
//int	ms;
int	gf=0;
int	gs;
//vect2	v;
//vect2	dx;

    if (star < 0) return 0;
// For star, see if it may be a guide star
// use find_guider to see if it is in guider window,
// set and restore map_sel to do this, probably redundant.
//    ms = map_sel;
//    map_sel = star;
//    gf = find_guider (&v, &dx);
//    map_sel = ms;

// Instead of above use find_gpos to find based on position vector
    gf = find_gpos (maplist[star].v);
    if (gf <= 0) return gf;	// no guider

/* See if star is within proper magnitude limits.  Use gdr[gf] when
  number of guiders is over 1, and g_maglimit or g_brightlim if the
  nguider is 0 or so. */
    gs = gf - 1;
    if (nguiders < 1) {
	if (maplist[star].mag > g_maglimit) return 0;	// faint.
	if (maplist[star].mag < g_brightlim) return 0;	// bright.
    } else {
	if (maplist[star].mag > gdr[gs].mag) return 0;	// faint.
	if (maplist[star].mag < gdr[gs].bmag) return 0;	// bright.
    }
    return  gf;
}

#ifdef  AUTOGS	// temporary

double	range_scale (double v, double r1, double r2)
/*  Compute the normalized distance from the center of the range
  r1..r2 at which v lies; but limit this distance to the range
  0.0 to 1.0.  Then return the square of this normalized distance.  */
{
double	r;	// the range
double	c;	// the center
double	d;	// the distance
double	x;	// our result

// Compute range, center and distance from center
    r = r2 - r1;
    c = (r1 + r2) * 0.5;
    d = v - c;

// Find normalized distance from center, allowing for null range
    if (r == 0.0) {
	x = (d == 0.0) ? 0.0 : 1.0;
    } else {
	x = d / r;
    }

// Compute squared distance and limit range to 0.0 .. 1.0
    x = x * x;
    if (x > 1.0) x = 1.0;
    return  x;
}

double	guide_metric (int star)
/*  Compute a selection metric for the given star depending on its
  position within the field (better if near center) and magnitude
  (better if middle of range).  When a range of values is tested, 
  the measure is computed by the range_scale function above.  For
  computing the isolation quantity, the nearest star distance is
  found by the near_star subroutine in skymap.c, and scaled.
  Our result is then scaled to produce a normalized value which
  is 1.0 for an ideal result, and 0.0 for a poor result.   */
{

double	sm = 0.0;	// The answer
int	g=0;		// the guider, one greater than subscript
vect2	pv;		// The position
double	dr, da;		// Normalized metric values for location
double	dm;		// Normalized metric for magnitude
double	dn;		// Distance to nearest
double	fs;		// Field size for normalize

// Find the guider field in which we are (etc. etc.)
// This partly duplicates find_guider subroutine
//    if (map_sel < 0) return sm;
    if (star < 0) return sm;
//    pv = xyproj (maplist[map_sel].v);
    pv = xyproj (maplist[star].v);
    if (nguiders < 1) {	// Old way
	g = 1;
	dr = range_scale (-pv.x, g_offaxis_x+g_field_w/2.0,
			g_offaxis_x-g_field_w/2.0);
	da = range_scale (pv.y, g_offaxis_y+g_field_h/2.0,
			g_offaxis_y-g_field_h/2.0);
	fs = (g_field_w + g_field_h) / 2.0;
    } else {	// nguiders > 0, IMACS way (from find_guider)
    double	r, a;
//    double	rs, as;
    int	i;
    double	p, ax;
    double	gx, gy;
    double	sr[2], sa[2], fw[2];
    
    // get r,a if we need them for polar guiders
	vec2pol (&r, &a, pv);
	a /= Degree;
	gx = gy = 0.0;
    
    /* We look only in first two guiders for a guide star, not in
     any other subsidary guiders, e.g. center-field; using 2, not nguiders */
	for (i=g=0; i<2; i++) {
	    if (gdr[i].mode & GDR_POLAR) {	/* gdr[i].mode is **Polar** */
		if (gdr[i].mode & GDR_SH) {
		    p = 0.0;
		    ax = 0.0;
		} else {
		    p = gdr[i].px * gdr[i].psize / 2.0;
		    ax = atan (p / gdr[i].position_u.polr_data.rp)/Degree;
		}
//		if (r < gdr[i].position_u.polr_data.rlo - p) continue;
//		if (r > gdr[i].position_u.polr_data.rhi + p) continue;
//		if (a < gdr[i].position_u.polr_data.alo - ax) continue;
//		if (a > gdr[i].position_u.polr_data.ahi + ax) continue;
//		g = i + 1;	// Found guider number g
		sr[i] = range_scale (r, gdr[i].position_u.polr_data.rlo - p,
				gdr[i].position_u.polr_data.rhi + p);
		sa[i] = range_scale (a, gdr[i].position_u.polr_data.alo - ax,
				gdr[i].position_u.polr_data.ahi + ax);

//		rs = r;  as = a;	/* the actual star positions */
//		if (r < gdr[i].position_u.polr_data.rlo) 
//			r = gdr[i].position_u.polr_data.rlo;
//		if (r > gdr[i].position_u.polr_data.rhi) 
//			r = gdr[i].position_u.polr_data.rhi;
	/* above 2 appeared reversed rlo/rhi ??? */
//		if (a < gdr[i].position_u.polr_data.alo) 
//			a = gdr[i].position_u.polr_data.alo;
//		if (a > gdr[i].position_u.polr_data.ahi) 
//			a = gdr[i].position_u.polr_data.ahi;
	
	/* Our first target is the r,a just found */
//		v->x = r;
//		v->y = a;
	
	/* gx, gy are coordinates in arcseconds in guider window aligned with
	 the polar coordinate system... */
	// THESE ARE PROBABLY WRONG DURING MOTION, MAYBE LATER TOO
//		gx = rs * cos ((a - as)*Degree) - r;
//		gy = rs * sin ((a - as)*Degree);
	
	    } else {	/* gdr[i].mode is **Rectangular** */
	/* This section has not been tested */
//		if (pv.x < gdr[i].position_u.rect_data.xlo) continue;
//		if (pv.x > gdr[i].position_u.rect_data.xhi) continue;
//		if (pv.y < gdr[i].position_u.rect_data.ylo) continue;
//		if (pv.y > gdr[i].position_u.rect_data.yhi) continue;
//		g = i + 1;
		sr[i] = range_scale (pv.x, gdr[i].position_u.rect_data.xlo,
				gdr[i].position_u.rect_data.xhi);
		sa[i] = range_scale (pv.y, gdr[i].position_u.rect_data.ylo,
				gdr[i].position_u.rect_data.yhi);
	
//		*v = pv;	// just guessing that this is correct...
	
//		gx = pv.x - 
//			(gdr[i].position_u.rect_data.xlo +
//			 gdr[i].position_u.rect_data.xhi)/2.0;
//		gy = pv.y - 
//			(gdr[i].position_u.rect_data.ylo +
//			 gdr[i].position_u.rect_data.yhi)/2.0;
	/* THAT PROBABLY HAS TO BE ROTATED BY map_pa ?? or 90 degrees? */
	    }	/* end guider mode selection */
	    fw[i] = gdr[i].px * gdr[i].psize;

//	    break;   /* split out of if statements */
	}	/* for loop over guiders */
	g = ((sr[0] + sa[0]) < (sr[1] + sa[1])) ? 0 : 1;
	dr = sr[g];
	da = sa[g];
	fs = fw[g];
	g++;
    
    /* The gx,gy found is the secondary vector */
//	d->x = gx;
//	d->y = gy;

    }	// end nguiders conditional


//-----

//int	find_guider (vect2 *v, vect2 *d)
//{
/* Find whether star number map_sel is in the guider field, returning
  zero if not (or no map_sel) and guider subscript+1 if it is.
  The "v" vector returns the target location; x,y for rectangular
  guiders, and either r,a or x,y for imacs.
  The "d" vector returns extra information, the projected vector
  for rectangular, and the gx,gy vector for imacs.
  Global values map_sel and nguiders are used to drive the logic
  and which selected star is used.
  The v and p vectors are used with the global guider field definitions
  to find if we are in a guider field; they are computed for this
  and returned to avoid having to be computed again.
*/


//    return  g;
//}

//-----


// Find distance from angular range  (see find_guider)

// Find distance from radial range

// Find distance from magnitude range
    if (nguiders < 1) {
	dm = range_scale (maplist[star].mag, g_maglimit, g_brightlim);
    } else {
    int	gs;
	gs = (g > 0) ? g-1 : 0;		// protection
	dm = range_scale (maplist[star].mag, gdr[gs].mag, gdr[gs].bmag);
    }

// Combine dr, da and dm

    sm += (dr + da) / 2.0;
    sm = sqrt (0.70 * sm + 0.30 * dm);

// Scale and normalize all range distances  (small is better)
// sm is normalized to 0.0 to 1.0, we hope.

// Find distance to nearest neighbor  (big is better) in arcseconds.
// call a subroutine.
//    dn = 20.0;		// temporary kludge.
    dn = map_dns (star);
    if (dn <= 0.0) dn = 20.0;	// Catch bad data
// (normalize to guider field size?)
    dn = dn / (0.5 * fs);	// Normalize to 1.0 for 1/2 field size
    if (dn > 1.0) dn = 1.0;	// Limit biggest normalized distance to 1.0

// Combine small and big targets -- small is better.
    return 0.6 * sm + 0.4 * (1.0 - dn);
}
#endif	// on AUTOGS - temporary

/* ---------------------------------------------------------------------- */
#ifdef  GSDATA
void	map_guideset (int  star, int tog)
/* Set the named star as a guide star.  If tog is non-zero, allow
  the call to toggle the setting of the guide star, and if zero
  only force the setting.  */
{
//int	ms;
int	gf;
int	ss;
//vect2	v;
//vect2	dx;
gpos	g;
gpos	og;
Obs	*obs;
double	d;

// check null cases
    if (star < 0) return;
    if (Gob == NULL) return;
    obs = Gob->obs;
    if (obs == NULL) return;

// See if star is a valid guide star based on catalog info for
// position and for brightness...
    gf = check_star (star);
    if (gf < 1) return;
// Now, if gf is > 0, we have a guider for star;
    gf--;

// obtain the ra/dec/epoch of the star and store in a gs structure
    g.ra = maplist[star].alpha;	// is in hours
    g.dec = maplist[star].delta;	// is in degrees
    g.ep = cat_equ[catalog];	// is in years
//    g.cn = star;
    g.cn = maplist[star].id;

    og = (gf == 1) ? obs->gp2 : obs->gp1;

// See if we already have it, if so clear otherwise set
    if (og.ep == 0.0) ss = 1;
    else {
	d  = 2.7e+4 * fabs (g.ra - og.ra);	// half arcsecond
	d += 1800.0 * fabs (g.dec - og.dec);	// half arcsecond
	d += 1.0e+3 * fabs (g.ep - og.ep);	// .001 year
	ss = d > 1.0;
    }
    if (!ss && tog) {	// toggle off is only allowed if flag is on
#ifdef  DBzep
	printf (" Toggle off guide star %d.\n", gf);
#endif
	g.ra = g.dec = g.ep = 0.0;
	g.cn = -1;
    }

    if (gf == 1) obs->gp2 = g;
    else         obs->gp1 = g;
// all done, the guider position has been set if it is valid.
    Gob->changed |= 1;

}
#endif  // on GSDATA


/* ---------------------------------------------------------------------- */

//void handle_mapsel(Display* display,int x,int y,int b)
void handle_mapsel(int x,int y,int b,int s)
{

//  if      (b == 1) mapsel_type = STAR_INFO;
//  else             mapsel_type = STAR_OFFSET;

/* Definitions for buttons:
  Left (1) -- STAR_OFFSET = Select star, show x,y or out of field
  Center (2) -- STAR_INFO = Obtain catalog information on star
  Right (3) -- STAR_DATA  = Show all star position data
.. */
/* State s is bit mapped, contains such bits as ShiftMask,
  LockMask, ControlMask */

/* "b" here is button pressed left=1, middle=2, right=3 */
  if      (b == 2) mapsel_type = STAR_INFO;
  else if (b == 1) mapsel_type = STAR_OFFSET;
  else if (b == 3) mapsel_type = STAR_DATA;
  else return;

  if (b == 2 && map_pn) {
vect2	p;
	p = map_fpxy (x, y);
	map_pcx = p.x;
	map_pcy = p.y;
/* Redraw, since we changed the center... */
//	redraw_skywin (display, True);
	skydraw |= 3;
#ifdef  DBGsd
	printf (" in mapsel skydraw=3\n");	// debug
#endif
  } else if (b == 2) {		// move center operation
vect2	p;
double	apc, dpc;
//#ifdef  GSDATA
//const  gpos  gz = { 0.0, 0.0, 0.0, 0 };
//#endif
//  printf (" - mapsel: equinox = %.2f  map_eq = %.2f\n", equinox, map_eq);
    p = map_fpxy (x, y);
// p is focal plane coordinates in arc seconds
// use unproj...
//    vec2sph (&apc, &dpc, unproj(p));
/* If the observation epoch is not the same as the catalog epoch, we
  need to precess the position found from catalog epoch as used by the
  star projection matrix to observation epoch so we can set the field
  center to that position correctly.  The following will precess the
  unprojected vector at map_eq to the actual equinox.  */
    vec2sph (&apc, &dpc, prec (equinox, unproj(p), map_eq));

// apc, dpc are new current center...  in radians;
// put into RAfield, DECfield
//    alpha = apc/Hour;
//    delta = dpc/Degree;
//    *(double*)RAfield.data = alpha * RAfield.scale;
//    *(double*)DECfield.data = delta * DECfield.scale;
// NOTE -- don't set alpha/delta here; they need to be detected
// as different and set in the skywin loop to draw it properly.
    *(double*)RAfield.data = apc; // Was.. RAfield.scale * apc / Hour;
    *(double*)DECfield.data = dpc; // Was..DECfield.scale * dpc / Degree;
// Also need to update text window values
    HandleTextWindow (winlist, RAfield.ew);
    HandleTextWindow (winlist, DECfield.ew);
    Gob->changed |= 1;	// mark as a change
/* Also clear any guide stars which need to be cleared */
#ifdef  GSDATA
//    Gob->obs->gp1 = Gob->obs->gp2 = gz;
//    Gob->obs->agscan = 0;
//    validate_guiders (Gob->obs);  // moved to redraw_map
#endif
//
#ifdef  DSSGET
  } else if (b == 3 && ( s & ControlMask )) {	// Get a DSS image (ctrl-right)
int	mapfound;
double	ra, dc;
char	cgi[80];
char	url[80];
char	defcom[40];
char	line[256];
char	*comm=NULL;
      mapfound = map_get_closest_star (x, y);
#if (DEBUG_GUI > 0)
    printf (" Control-right on object %d", mapfound);  // debug only
#endif
/* Construct DSS request for catalog object position.  */
      sprintf (url, "http://archive.stsci.edu/cgi-bin/dss_search");
      precess(maplist[mapfound].alpha,maplist[mapfound].delta,cat_equ[catalog],
              &ra, &dc, 2000.00);	// may be null case
      sprintf (cgi, "v=3&r=%.5f&d=%.5f&e=J2000&h=1.5&w=1.5&f=gif", ra*15.0, dc);

/* NOTE -- we could derive h and w data from the guider px and psize values. */

#if (DEBUG_GUI > 0)
    printf (" - %s\n", cgi);			// debug only
#endif
/* Set the default command to use the generated URL */
/*  On mac-os, __APPLE_CC__ and __APPLE__ are defined, while Linux has
  __unix__ and __i386__ defined.  */
#ifdef  __APPLE__
      sprintf (defcom, "open");
#else
      sprintf (defcom, "firefox");	// was "htmlview"
#endif
/* Above, we need the command which will reliably handle a URL properly;
  usually that is a browser.  For Apple, the "open" command is intelegent
  enough to recognize a URL and assign it to the current browser; for Linux
  there appears not to be such a facility, so we go with the most common
  browser.  If a more general method comes online, change to that.  */

/* Obtain the command to use.  It should be a command capable of taking
  a general URL as an argument, and then viewing the image returned as a
  result, or displaying an error message.  It is defined by the environment
  variable "MYDISPLAY" or the defaults chosen above based on O.S.
  This allows the user to configure a different command if desired.   */
      comm = getenv ("MYDISPLAY");
      if (comm == NULL) comm = defcom;

/* Pass the command and URL on to the system to get the picture. */
      sprintf (line, "%s \"%s?%s\" &", comm, url, cgi);
      system (line);	// do the deed
#endif	// on DSSGET -- cntrl+right button case
  } else {		// All other button press actions
      map_sel = map_get_closest_star (x, y);
#ifdef  GSDATA
// Call program to set/unset global guide star location...
      if (b == 1) map_guideset (map_sel, 1);
#endif  // on GSDATA

#ifdef DEBUG
      fprintf(stderr,"handle_mapsel(%d,%d) -> %d (%d)\n",x,y,map_sel,mapsel_type);
#endif

      if (map_sel > -1) {
//      redraw_skywin (display,False);
	skydraw |= 1;
#ifdef  DBGsd
	printf (" in mapsel skydraw=1\n");	// debug
#endif
      }

  }	// button and modifier sections

}

/* ---------------------------------------------------------------------- */

static void adjust_scale(void)
{
  switch (catalog) {
  case CAT_USNO_SA:
    if (map_scale > MAX_SCALE_USS) map_scale = MAX_SCALE_USS;
    break;
  case CAT_USNO_AO:
    if (map_scale > MAX_SCALE_USA) map_scale = MAX_SCALE_USA;
    break;
  case CAT_GSC:
  case CAT_GSCP:
    if (map_scale > MAX_SCALE_GSC) map_scale = MAX_SCALE_GSC;
    break;
  case CAT_SAO:
  default:
    if (map_scale > MAX_SCALE_SAO) map_scale = MAX_SCALE_SAO;
    break;

  }
  if (map_scale < 0.2) map_scale = 0.2;
}

/* ---------------------------------------------------------------------- */

static Bool is_maxscale(void)
{
  Bool r=False;

  switch (catalog) {
  case CAT_USNO_SA:
    if (map_scale >= MAX_SCALE_USS) r=True;
    break;
  case CAT_USNO_AO:
    if (map_scale >= MAX_SCALE_USA) r=True;
    break;
  case CAT_GSC:
  case CAT_GSCP:
    if (map_scale >= MAX_SCALE_GSC) r=True;
    break;
  case CAT_SAO:
  default:
    if (map_scale >= MAX_SCALE_SAO) r=True;
    break;
  }
  return(r);
}

/* ---------------------------------------------------------------------- */


/* ---------------------------------------------------------------------- */


/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */

/*   ===   Revision History   ===  


2011/10-26 -- Version 1.30
	-- On SHFTCLK, the feature that a center click of the
	mouse may be achieved by a shift-click is added.  This maps all
	clicks with shift held to center click.
	-- Window position changed from CBS_BOTTOMLEFT to CBX_MIDLEFT

2009/11-24 -- Version 1.29
	-- Includes client/server method for star catalog, implementing
	the EXCAT feature.  Tested with production installed server.
	Most of the implementation is in skymap.c.
	-- Add messages for source of star catalog; detailed in TESTMODE

2009/09-15 -- Version 1.28
	-- Includes CACHECAT feature and implements caching of catalog
	data (optionally).  The feature is required when EXCAT is used
	in a released version.
	-- Includes use of GetStars method, which reads star data from
	server as text.  See AMETHOD symbol in mgfeats.h.

2009/07-29 -- Version 1.27
	-- Trying out test of catalog on disk and defaulting to catalog
	via net if EXCAT feature is defined.
	--> Will be minimum version when catalog server is released.

2009/03-03 -- Version 1.26
	-- Cleaned up find_gpos and find_guider to have actual guider
	subscript found in former, and only extra vectors things set
	in the latter which is now a local subroutine.  See comments.

2008/09-17 -- Version 1.25
	-- Change skywin_handle to take pointer to sleep time, and to
	return the number of events handled by the window.  Also change
	to handle all events possible in a loop rapidly before giving
	any chance to wait for more.

2008/09-10 -- Version 1.24
	-- Add support for reading catalog data from Internet, and for
	partial caching of sky window data.
	--> To become Minimum Version about release 2.06+

2008/03-28 -- Version 1.23
	-- Add displacement of coordinate in find_guider to match that in
	the gmap program.  This is needed to correctly apply the guider
	boundaries; they are already plotted correctly.
	--> Minimum Version for release 2.05+, added in release 2.04

2008/01-29 -- Version 1.22
	-- Precession done on position clicked on to center the display,
	and not on position of selected star.
	--> Minimum Version for release 2.04+

2007/11-13 -- Version 1.21
	-- Support for GISCAM feature, which will probably not be used
	since gismo areas are not same sized as chips are.

2007/11-05 -- Version 1.20
	-- Mouse clicks in sky window extended to include recognition of
	shift or control keypress with the mouse click.  To implement
	the automatic retrieval of DSS images of guide stars.
	-- Add MYDISPLAY environment variable to define display command.

2007/10-29 -- Version 1.19
	-- Fix compare to establish new position angle, the error amount was
	too small causing continious mis-compares.
	-- Minimum Version for release 2.02+

2007/08-21 -- Version 1.18
	-- Adding support for automatic guide star; added "check_star"
	-- Minimum Version for release 2.0 and version control.

2007/07-06 -- Version 1.17
	-- Using ObInstN to get instrument name to avoid crash

2007/04-11 -- Version 1.16
	-- Reduce star position display line if not debug

2006/09-28 -- Version 1.15
	-- Add test mode and select SAO catalog in tintgui case.

2006/06-30 -- Version 1.14
	-- Make symbol GCENT standard and comment out use.

2006/01-26 -- Version 1.13
	-- Add bright limits to pmap and use them from there

2006/01-24 -- Version 1.12
	-- Added BrightLim as 11.0 for bright magnitude limit

2006/01-13 -- Version 1.11
	-- Added g_maglimit for single guider mangitude limit feature,
	also added to pmap.dat the data line.

2006/01-10 -- Version 1.10
	-- Add support for GSDATA guide star selection

2005/11-17 -- Version 1.06
	-- Revised for Mac-OS refresh bug

2005/08-29 -- Version 1.05
	-- Better scale computation for LDSS, and others too.

2004/11-15 -- Version 1.04
	-- Pass version number increment on.

2004/11-10 -- Version 1.03
	-- Pass version number increment on.

2004/10-26 -- Version 1.02
	-- Pass version number increment on.

2004/10-22 -- Version 1.01
	-- Introduced the SKWVERS string, basically global to be used
	by routines here and including program (intgui).  Allows increment
	of sky window things without disturbing the base version.
	Similar to maskgen/mgutils version extension done earlier.

... */
