/*   skywin.h  ==  Definitions for skywin programs  */

/* Modified from telmap.h from gmap project... */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2008 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef INCLUDE_SKYWIN_H
#define INCLUDE_SKYWIN_H

/* ---------------------------------------------------------------------- */

#ifdef IS_SKYWIN_C
#define EXT_SKYWIN
#else
#define EXT_SKYWIN extern
#endif

#undef  UPOPTION
/* Not used at all */
/* Allows zenith dot and up vector options -- useful for debug */
/* See also, FX_UV and FX_ZD enum, and map_uv and map_zd flags */

/* DEFINEs -------------------------------------------------------------- */

#define CAx         2                         /* catalog menu             */
#define CAy         2
#define CAw         70  // was 55	// was 80
#define CAh         18

#define STx         (CAx+CAw+10)
#define STy         CAy
#define STw         36
#define STh         CAh

#define FXy         CAy
#define FXw         104   // was 80
#define FXh         CAh

// #define P1x         (STx+STw+20)
// #define P1y         STy
// #define P1w         STw
// #define P1h         STh

// #define P2x         (P1x+P1w+10)
// #define P2y         P1y
// #define P2w         P1w
// #define P2h         P1h

// #define TLx         (P2x+P2w+10)
// #define TLy         P2y
// #define TLw         P2w
// #define TLh         P2h

// #define GTx         (TLx+TLw+20)
// #define GTy         TLy
// #define GTw         TLw
// #define GTh         TLh

#define CLx         (STx+STw+10)
#define CLy         CAy
#define CLw         54
#define CLh         CAh
 
#define Zw          18                        /* zoom-in/zoom-out buttons */
#define Zh          CAh                       /* star-map                 */

// For GCENT things
#define  RCx	(FXw/2+16)
#define  RCy	FXy
#define  RCw	92
#define  RCh	STh

// #define CTx         (FXw/2+15)
// #define CTy         STy
// #define CTw         STw
// #define CTh         STh

// #define PKx         (CTx+CTw+10)
// #define PKy         STy
// #define PKw         STw
// #define PKh         STh

enum  fx_menus {		/* Features menu */
		FX_NV,		/* North vector */
//#ifdef  UPOPTION
#ifdef  UPARO
		FX_UV,		/* Up vector */
#endif
#ifdef  ZNTHDOT
		FX_ZD,		/* Zenith marker */
#endif
// 		FX_RO,		/* Random offset */
		FX_PN,		/* Panning of plot center */
		FX_SC,		/* Star Colors(*) */
		FX_AP,		/* Plot alignment(*) */
		FX_OP,		/* Plot objects */
		FX_GP,		/* Guider 1 display(*) */
		FX_GS,		/* Guider 2 display(*) */
//		FX_GC,		/* Center Guider display */
		FX_CF,		/* Camera FOV display(*) */
		FX_KD		/* KILL the sky window */
	};

/* INCLUDEs ------------------------------------------------------------- */

#include <X11/X.h>

/* TYPEDEFs ------------------------------------------------------------- */

enum  dmenus  { MNU_TEL, MNU_GDR, MNU_CAM };  // From gmap

/* GLOBALs -------------------------------------------------------------- */

EXT_SKYWIN MainWindow skywin;
EXT_SKYWIN Menu   camenu;
EXT_SKYWIN Menu   fxmenu;
EXT_SKYWIN Button zibut,zobut;
// EXT_SKYWIN Button stbut;
// EXT_SKYWIN Button ctbut;
// EXT_SKYWIN Button pkbut;
// EXT_SKYWIN Button p1but;
// EXT_SKYWIN Button p2but;
// EXT_SKYWIN Button tlbut;
// EXT_SKYWIN Button gtbut;

// EXT_SKYWIN Window gdr1win;
// EXT_SKYWIN Window gdr2win;

/* function prototype(s) */

int	find_gpos (vect3 vcp);
//int	find_guider (vect2 *v, vect2 *d);	// non-local ?

//#ifdef  GCENT
void	init_gcenter (void);
//#endif

//void	skywin_handle (void);
int	skywin_handle (int* sleeptime);

u_long	get_named_color (Display* disp, char* name, u_long dflt_color);
// void create_skywin (Display*,GC);
void create_skywin (XSizeHints);
void get_camera_defaults (void);
void redraw_skywin (Display*,Bool);

void	update_guider_windows (void);

void handle_camenu (void);	// (Display*);
void handle_fxmenu (void);	// (Display*);
//#ifdef  GSDATA
//void	validate_guiders (Obs* ob);
//#endif
void handle_zibut  (int);	// (Display*,int);
void handle_zobut  (int);	// (Display*,int);
// void handle_stbut  (Display*,int);
// void clear_ctsel (void);
// void handle_ctbut  (Display*,int);
// void handle_p1but  (Display*);
// void handle_p2but  (Display*);
// void handle_tlbut  (Display*);
// void handle_gtbut  (Display*);
// void handle_clbut  (Display*);
// void handle_pkbut  (Display*);
#ifdef  AUTOGS
double	guide_metric (int star);
#endif
void handle_mapsel (int,int,int,int);	// (Display*,int,int,int);
//EXT_SKYWIN int	check_star (int star);
int	check_star (int star);

/* ---------------------------------------------------------------------- */

#endif  /* INCLUDE_SKYWIN_H */

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
