/* Utility programs for performance timing - Header file */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                  Copyright (C) 2009 by                            **
**     Ken Clardy,  Carnegie Observatories,  Pasadena California.    **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* ---------------------------------------------------------------------- */

#ifndef INCLUDE_TIMEPAK_H
#define INCLUDE_TIMEPAK_H

/* INCLUDEs ------------------------------------------------------------- */

#include	<time.h>
#include	<sys/time.h>
#include	<string.h>

/* DEFINEs -------------------------------------------------------------- */

//#ifdef  IS_MGU_C
//#define  EXT_MGU
//#else
//#define  EXT_MGU extern
//#endif

/* -- Bits in flag word -- */
#define  TF_auto  (1 << 0)
#define  TF_rest  (1 << 1)

#define  T_ReStart  (1 << 2)
/* Flag bits operate as follows:
  auto  = automatic start and stop timer
  rest  = allow timer restart if begun while running
  */

/* -- Report definition bits -- */
#define  TR_titlr (1 << 0)
#define  TR_titlf (1 << 1)
#define  TR_elap  (1 << 2)
#define  TR_cput  (1 << 3)
#define  TR_count (1 << 4)
#define  TR_errp  (1 << 5)
#define  TR_long  (1 << 6)
/* Report formatting flags do this:
  titlr = report the title
  titlf = title on single line followed with line feed
  elap  = elapsed time reported
  cput  = cpu time reported
  count = report counts and time/count also
  errp  = report errors if any
  long  = allow lines up to 80 char, otherwise 60.
  */

/* TYPEDEFs ------------------------------------------------------------- */

struct	TS_timer {
	char	*title;		// Report identifier
	double	et;		// Total seconds found CPU
	double	ct;		// Total seconds clock time
	double	abst;		// Latest start absolute
	clock_t	start;		// Latest start time
	int	n;		// Interval total
	int	state;		// state = start count
	int	flag;		// flag bits
	int	er;		// count of errors
	};
typedef	struct	TS_timer	T_timer;

/* GLOBALs -------------------------------------------------------------- */

/* Be careful with external and initialization here. */
#ifdef  IS_MGU_C
int	N_TX = 0;
T_timer	*T_TX = NULL;
#else
extern  int	N_TX;
extern  T_timer	*T_TX;
#endif
 
/* FUNCTION PROTOTYPEs -------------------------------------------------- */

T_timer*  T_new (char* title, int flag);
T_timer*  T_kil (T_timer* b);

void	T_clr (T_timer* b);
void	T_beg (T_timer* b);
void	T_end (T_timer* b);
char*	T_rep (T_timer* *b, int r);
void	T_xit (T_timer* *b, int r);

void	TX_con (int n);
void	TX_beg (int n);
void	TX_end (int n);
void	TX_nam (int n, char* title);
void	TX_enm (int n, char* title);
void	TX_rep (int n);


/* ---------------------------------------------------------------------- */
/* End of timepak.h package */
 
#endif  /* INCLUDE_TIMEPAK_H */
 
/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
