/* ------------------------------------------------------------------------
 *
 * utils.c
 * 
 * Project: GMAP  (OCIW, Pasadena, CA)
 *
 * general utility functions ...  
 *
 * v0.01   1998-06-01  Christoph C. Birk, birk@ociw.edu
 * v0.10   1999-07-01  (CCB)
 * v0.20   2000-03-30  gmap (CCB)
 *
 * ---------------------------------------------------------------------- */

#define IS_UTILS_C

/* DEFINEs -------------------------------------------------------------- */

/* #define DEBUG */

#ifdef SOLARIS
#ifndef SVR4
#define SVR4
#endif
#endif

#ifndef  sun
#define  SVR4
#endif
// Above is to compile correctly on Linux

/* INCLUDES ------------------------------------------------------------- */

#include <stdlib.h>
#include <stdio.h>                        /* fprintf() */
#include <string.h>                       /* strcpy() */
#include <math.h>                         /* fabs(),floor() */
#include <ctype.h>                        /* toupper() */
#include <unistd.h>                       /* fork(),_exit(),open() */
#include <fcntl.h>                        /* O_RDWR */
#ifdef SVR4
#include <time.h>                         /* nanosleep() */
#endif

#include "utils.h"


/* GLOBALs -------------------------------------------------------------- */

/* none */


/* EXTERNs -------------------------------------------------------------- */

/* none */


/* ---------------------------------------------------------------------- */

char* extend(char *s, char c, int l)
{
  int i;
 
  i = strlen(s);
  while (i < l) {
    s[i++] = c;
  }
  s[i] = '\0';                                          /* s[l] = '\0';   */

  return(s);
}
 
/* ---------------------------------------------------------------------- */

char* uppercase(char *string)
{
  char c,*s;

  s = string;
  while (*s != '\0') { 
    c = toupper(*s);
    *s++ = c;
  }
  return(string);
}

/* ---------------------------------------------------------------------- */

char* lowercase(char *string)
{
  char c,*s;
 
  s = string;
  while (*s != '\0') {
    c = tolower(*s);
    *s++ = c;
  }
  return(string);
}
 
/* ---------------------------------------------------------------------- */

char* cut_spaces(char* line)
{
  char *p1,*p2;
 
  p1 = line;
  p2 = line + strlen(line) -1;
 
  if (p2 < p1) return(line);

  while (isspace(*p1) && p1<p2) p1++;
  while (isspace(*p2) && p2>p1) *p2-- = '\0';
 
  if (p1 > line) {
    char *p0;
    p0 = line;
    while (*p1) {
      *p0 = *p1; p0++; p1++;
    }
    *p0 = '\0';
  }
    
  return(line);
}

/* ---------------------------------------------------------------------- */

char* inc_filename(char *name)
{
  int  i;
 
  if (*name == '\0') return(name);
  name[strlen(name)-1] += 1;                  /* increment last character */ 
 
  for (i=strlen(name)-1; i>0; i--) {          /* 'overflow' characters    */ 
    if (name[i] == (char)('z'+1)) {           /* 'z+1' -> 'a'             */ 
      name[i] = 'a';
      name[i-1] += 1;
    } else if (name[i] == (char)('Z'+1)) {    /* 'Z+1' -> 'A'             */ 
      name[i] = 'A';
      name[i-1] += 1;
    } else if (name[i] == (char)('9'+1)) {    /* '9+1' -> '0'             */ 
      name[i] = '0';
      name[i-1] += 1;
    }
  }
  return(name);
}

/* ---------------------------------------------------------------------- */
 
char* dec_filename(char *name)
{
  int  i;
 
  if (*name == '\0') return(name);
  name[strlen(name)-1] -= 1;                  /* decrement last character */
 
  for (i=strlen(name)-1; i>0; i--) {          /* 'overflow' characters    */
    if (name[i] == (char)('a'-1)) {           /* 'a-1' -> 'z'             */
      name[i] = 'z';
      name[i-1] -= 1;
    } else if (name[i] == (char)('A'-1)) {    /* 'A-1' -> 'Z'             */
      name[i] = 'Z';
      name[i-1] -= 1;
    } else if (name[i] == (char)('0'-1)) {    /* '0-1' -> '9'             */
      name[i] = '9';
      name[i-1] -= 1;
    }
  }
  return(name);
}

/* ---------------------------------------------------------------------- */

char* extract_filename(char* filename,const char* path)
{
  char *h;
 
  h = strrchr(path,'/');                   /* WARNING: U**X specific path */
  if (h == NULL) strcpy(filename,path);
  else           strcpy(filename,h+1);
  h = strrchr(filename,'.'); if (h) *h = '\0';
 
  return filename;
}

/* ---------------------------------------------------------------------- */

char* extract_pathname(char* pathname,const char* path)
{
  char *h;
 
  strcpy(pathname,path);
  h = strrchr(pathname,'/');               /* WARNING: U**X specific path */
  if (h) *h = '\0';
  else   strcpy(pathname,"");
 
  return pathname;
}
 
/* ---------------------------------------------------------------------- */

char* alpha_str(char *string,double alpha)   /* alpha in hours -> string  */
{
  double hr,mi,se;
 
  hr = floor(alpha);                                    /* alpha > 0.     */
  mi = floor((alpha-hr)*60.);
  se = (alpha - hr - mi/60.)*3600.;
/*  if (se >= 59.95) { se -= 59.95; mi += 1.; } */
  if (se >= 59.995) { se -= 59.995; mi += 1.; }
  if (mi >= 60.00) { mi -= 60.00; hr += 1.; }
  if (hr >= 24.00) { hr -= 24.00; }
/*  (void)sprintf(string," %02.0f:%02.0f:%04.1f",hr,mi,se); */
  (void)sprintf(string," %02.0f:%02.0f:%04.2f",hr,mi,se);

  return(string);
}
 
/* ---------------------------------------------------------------------- */
 
char* delta_str(char *string,double delta)
{
  double dg,pd,mi,se;
  char   sign=' ';
 
  pd = fabs(delta);
  dg = floor(pd);                        /* delta in degrees -> string    */
  mi = floor((pd - dg)*60.);
  se = (pd - dg - mi/60.)*3600.;
  if (delta < 0.) sign = '-';
  if (se >= 59.95) { se -= 59.95; mi += 1.; }
  if (mi >= 60.00) { mi -= 60.00; dg += 1.; }
  (void)sprintf(string,"%c%02.0f:%02.0f:%04.1f",sign,dg,mi,se);

  return(string);
}

/* ---------------------------------------------------------------------- */
 
char* time_str(char *string,double st)
{
  int hr,mi,se;
 
  hr  = (int)floor(st/3600.);
  st -= hr*3600.;
  mi  = (int)floor(st/60.);
  st -= mi*60.;
  se  = (int)floor(st);
  (void)sprintf(string,"%2d:%02d:%02d",hr,mi,se);

  return(string);
}

/* ---------------------------------------------------------------------- */

double d_alpha(const char* text)
{
  double alh=0.,alm=0.,als=0.,a;
 
  if (!strchr(text,(int)':'))
    sscanf(text,"%lf %lf %lf",&alh,&alm,&als);
  else
    sscanf(text,"%lf:%lf:%lf",&alh,&alm,&als);
 
  a = alh + alm/60. + als/3600.;
  while (a >= 24.) a -= 24.;
  while (a < 0.)  a += 24.;
 
  return (a);
}

/* ---------------------------------------------------------------------- */

double d_delta(const char* text)
{
  double deg=0.,dem=0.,des=0.,d;
 
  if (!strchr(text,(int)':'))
    sscanf(text,"%lf %lf %lf",&deg,&dem,&des);
  else
    sscanf(text,"%lf:%lf:%lf",&deg,&dem,&des);
 
  d = fabs(deg)+fabs(dem)/60.+fabs(des)/3600.;
  while (d > 90.) d -= 90;
 
  if ((deg<0.) || (dem<0.) || (des<0.) || strchr(text,(int)'-'))
    d = -d;
 
  return d;
}

/* ---------------------------------------------------------------------- */

void q_sort(int *zahl, int l, int r)
{
  int  x,i,j,h;

  i = l;
  j = r;

  if (j>i) {
    x = zahl[(i+j)/2];
    do {
      while (zahl[i] < x) i++;
      while (zahl[j] > x) j--;
      if (i<=j) {
        h         = zahl[i];
        zahl[i++] = zahl[j];
        zahl[j--] = h;
      }
    } while (!(i>j));
    q_sort(zahl,l,j);
    q_sort(zahl,i,r);
  }
}

/* ---------------------------------------------------------------------- */

void d_sort(double *zahl,int l,int r)
{
  int    i,j;
  double x,h;
 
  i = l;
  j = r;
 
  if (j>i) {
    x = zahl[(i+j)/2];
    do {
      while (zahl[i] < x) i++;
      while (zahl[j] > x) j--;
      if (i<=j) {
        h         = zahl[i];
        zahl[i++] = zahl[j];
        zahl[j--] = h;
      }
    } while (!(i>j));
    d_sort(zahl,l,j);
    d_sort(zahl,i,r);
  }
}
 
/* ---------------------------------------------------------------------- */

char *dtime(void)
{                                        /* create data/time-string       */ 
  time_t t1;

  (void)time(&t1); 
 
  return(ctime(&t1)); 
} 

/* ---------------------------------------------------------------------- */
 
void adebug(const char *who,const char *text)
{
  char dfile[256];
  FILE *fp;
 
  sprintf(dfile,"%s/debug.log",genv2("CAMTMP",getenv("HOME")));

  if ((fp=fopen(dfile,"a")) != NULL) {
    if (who == NULL) who = "?";
    fprintf(fp,"\t(%s) %s\n",who,text);
    (void)fclose(fp);
  }
}
 
/* ---------------------------------------------------------------------- */

void tdebug(const char *who,const char *text)
{
  char dfile[256];
  FILE *fp;
 
  sprintf(dfile,"%s/debug.log",genv2("CAMTMP",getenv("HOME")));

  if ((fp=fopen(dfile,"a")) != NULL) {
    if (who == NULL) who = "?";
    fprintf(fp,"%s\t(%s) %s\n",dtime(),who,text);
    (void)fclose(fp);
  }
}

/* ---------------------------------------------------------------------- */

void append_logfile(const char* file,const char* who,const char* text)
{
  FILE   *fp;
  char   tstr[32];
  time_t ct;

  fp = fopen(file,"a");
  if (fp == NULL) return;                      /* cannot open file */

  ct = cor_time(0);
#ifdef  sparc
  cftime(tstr,"%H:%M:%S",&ct);
#else
// Substitute equivalent strftime (no cftime in Linux?)
  strftime (tstr, 12, "%H:%M:%S", gmtime(&ct));
#endif
  fprintf(fp,"%s - %s - %s\n",tstr,who,text);
  fclose(fp);
}
  
/* ---------------------------------------------------------------------- */
 
time_t cor_time(int timeoff)
{
  return(time((time_t*)NULL) - (time_t)timeoff);
}

/* ---------------------------------------------------------------------- */
 
/* this is a SunOS 4.x.x <-> SunOS 5.x.x compatible version of
   usleep(), nanosleep()
*/
 
void msleep(int msec)
{
#ifdef SVR4
  struct timespec delay;
  if (msec > 0) {
    delay.tv_sec  = msec/1000;
    delay.tv_nsec = (msec % 1000) * 1000000;
    (void)nanosleep(&delay,NULL);
  }
#else
  while (msec >= 1000) {
    sleep(1);
    msec -= 1000;
  }
  if (msec > 0) usleep((u_int)msec*1000);     /* avoid to call usleep(0)  */
#endif
}
 
/* ---------------------------------------------------------------------- */

#define STG0    24059.               /* Greenwich Siderial Time (seconds) */
                                     /* on 1970.01.01.00:00:00 GMT        */
double get_siderial(time_t ut,double tl)
{
  double st;

  st  = (double)((int)(STG0+366.2422/365.2422*ut)%(int)86400)*24./86400.;
  st += tl/15.;
  if (st < 0.)  st += 24.;
  if (st > 24.) st -= 24.;

  return(st*3600.);                     /* seconds                        */
}

/* ---------------------------------------------------------------------- */

double get_epoch(time_t ut)
{
  int    yr;
  double doy,hr, /* mi,sc, */ dpy,ep;
  char   buffer[128];
  struct tm *gtime;

  if (ut == 0) ut = time(NULL);                   /* now */
  gtime = gmtime(&ut);                            /* get tm-structure (UT)*/

  strftime(buffer,80,"%Y",gtime);                 /* calculate epoch      */
  yr  = atoi(buffer);                             /* year                 */
#ifdef SVR4                                       /* get days per year    */
  if ((yr % 4) == 0) {
    if ((yr % 100) == 0) {
      if ((yr % 400) == 0)
        dpy = 366.;
      else
        dpy = 365.;
    } else {
      dpy = 366.;
    }
  } else {
    dpy = 365.;
  }
#else
  dpy = (double)dysize(yr);
#endif
  strftime(buffer,80,"%j",gtime);                 /* day                  */
  doy = atof(buffer) - 1.;
  strftime(buffer,80,"%H",gtime);                 /* hour                 */
  hr = atof(buffer); 
#if 0
  strftime(buffer,80,"%M",gtime);                 /* minute               */
  mi = atof(buffer);
  strftime(buffer,80,"%S",gtime);                 /* second               */
  sc = atof(buffer);
#endif
  ep = (double)yr + doy/dpy + hr/(24.*dpy);
       /* + mi/(1440.*dpy) + sc/(86400.*dpy) */

  return(ep);
}

/* ---------------------------------------------------------------------- */

int get_uts(time_t ut)
{
  struct  tm *gtime;
 
  gtime = gmtime(&ut);                        /* GMT */
 
  return((int)gtime->tm_hour*3600+(int)gtime->tm_min*60+(int)gtime->tm_sec);
}

/* ---------------------------------------------------------------------- */

char* get_night(char* night,time_t when)
{
  struct tm* ltime;
  time_t     ut;
 
  if (when == 0) ut = cor_time(0);
  else           ut = when;
  ut   -= 12*3600;
  ltime = localtime(&ut);                     /* get local time */
 
  strftime(night,16,"%d%b%Y",ltime);
 
  return night;
}
 
/* ---------------------------------------------------------------------- */

// #define RADS    57.29577951                  /* Degrees per radian        */
#define  Degree     (M_PI/180.0)
#define SCALE   750.                         /* Atmospheric scale height  */

double get_airmass(double lat, double dec, double ha)
{
  double cos_zd,x,airmass;

  cos_zd = sin (lat*Degree) * sin (dec*Degree) +
	   cos (lat*Degree) * cos (dec*Degree) * cos (15.*ha*Degree);
  if ((cos_zd > 1.) || (cos_zd < 0.)) return (9.999);

  x  = SCALE * cos_zd;
  airmass = sqrt (x*x + 2.*SCALE + 1.) - x;
  if (airmass > 9.999) airmass = 9.999;
 
  return (airmass);
}

/* ---------------------------------------------------------------------- */

double get_ha(time_t ut,double tellong,double alpha)
{
  double ha;
 
  ha = get_siderial(ut,tellong)/3600. - alpha;
 
  return ha;
}

/* ---------------------------------------------------------------------- */

void precess(double ain,double din,double ein, double *aout, double *dout,
             double eout)
{
  double c1=2.062648e5;
  double pi=3.141592653;
  double al,del,tau0,tau;
  double zeta,theta,z,ac,bc,cc,zx,a1,d1;
 
  if (ein == eout) {                          /* input = output equinox ? */
    *aout = ain;
    *dout = din;
    return;
  }
  al  = ain*(pi/12.);                         /*  coordinates in radians  */
  del = din*(pi/180.);
 
  tau0   = (ein-1900.)/100.;                  /* Newcomb:                 */
  tau    = (eout-1900.)/100. - tau0;
  zeta   = (2304.25+1.396*tau0)*tau+.302*tau*tau+0.018*tau*tau*tau;
  z      = zeta+0.791*tau*tau+0.001*tau*tau*tau;
  theta  = (2004.682-0.853*tau0)*tau-0.426*tau*tau-0.042*tau*tau*tau;
  zeta  /= c1;
  z     /= c1;
  theta /= c1;
 
  ac = cos(del)*sin(al+zeta);
  bc = cos(theta)*cos(del)*cos(al+zeta)-sin(theta)*sin(del);
  cc = sin(theta)*cos(del)*cos(al+zeta)+cos(theta)*sin(del);
  d1 = asin(cc);
  zx = sqrt(1.-cc*cc);
  a1 = acos(bc/zx);
  if (ac/zx < 0.) a1=2.*pi-a1;
  a1 += z;

  *aout = a1*12./pi;                         /* -> hours                  */
  if (*aout <  0.)  *aout += 24.;
  if (*aout >= 24.) *aout -= 24.;
  *dout  = d1*180./pi;                       /* -> degrees                */
}

/* ---------------------------------------------------------------------- */

char *genv2(char *first,char *def_pointer)
{
  if (getenv(first) == NULL) return(def_pointer);
  return(getenv(first));
}
 
/* ---------------------------------------------------------------------- */

char *genv3(char *first,char *second,char *def_pointer)
{
  if (getenv(first) == NULL) {
    if (getenv(second) == NULL) return(def_pointer);
    else                        return(getenv(second));
  }
  return(getenv(first));
}

/* ---------------------------------------------------------------------- */

int imin(int a,int b)
{
  if (a < b) return(a);
  return(b);
}

/* ---------------------------------------------------------------------- */

int imax(int a,int b)
{
  if (a > b) return(a);
  return(b);
}

/* ---------------------------------------------------------------------- */
 
double dmin(double a,double b)
{
  if (a < b) return(a);
  return(b);
}
 
/* ---------------------------------------------------------------------- */
 
double dmax(double a,double b)
{
  if (a > b) return(a);
  return(b);
}

/* ---------------------------------------------------------------------- */

char* itos(char* string,char* format,int value)
{
  (void)sprintf(string,format,value);
  return(string);
}

/* ---------------------------------------------------------------------- */
 
char* ltos(char* string,char* format,long value)
{
  (void)sprintf(string,format,value);
  return(string);
}
 
/* ---------------------------------------------------------------------- */
 
char* ftos(char* string,char* format,double value)
{
  (void)sprintf(string,format,value);
  return(string);
}

/* ---------------------------------------------------------------------- */

int get_list_index(char* string,char** list)
{
  int i;
  
  for (i=0; *list[i] != '\0'; i++) {
    if (!strcasecmp(string,list[i])) break;
  }
  if (*list[i] == '\0') return(-1);
  return(i);
}

/* ---------------------------------------------------------------------- */

int str2int(char* string,char** list)
{
  int i;
 
  for (i=0; *list[i] != '\0'; i++) {
    if (!strcasecmp(string,list[i])) break;
  }
  if (*list[i] == '\0') return(-1);
  return(i);
}
 
/* ---------------------------------------------------------------------- */

char* get_local_datestr(char* string,time_t ut)
{
  struct tm *now;
 
  if (ut <= 0) ut = time(NULL);
  now = localtime(&ut);
  strftime(string,32,"%b %d %H:%M:%S",now);     /* min.length of 'string' */
  return(string);
}
 
/* ---------------------------------------------------------------------- */
 
char* get_ut_timestr(char* string,time_t ut)
{
  struct tm *now;
 
  if (ut <= 0) ut = time(NULL);
  now = gmtime(&ut);
  strftime(string,32,"%H:%M:%S",now);           /* min.length of 'string' */
  return(string);
}

/* ---------------------------------------------------------------------- */

char* get_string(char* file,char* key,char* value,char* defstring)
{
  char   buffer[1024],val[128],*h,*r=NULL;
  FILE   *fp;

  if (defstring) { strcpy(value,defstring); r=value; }

  fp = fopen(file,"r");
  if (fp != NULL) {                               /* file exists */
    while (fgets(buffer,sizeof(buffer),fp)) {     /* search for 'key' */
      cut_spaces(buffer);                         /* remove <CR> */
      h = strchr(buffer,';');                     /* comment */
      if (h) *h='\0';
      h = strchr(buffer,'=');                     /* delimiter */
      if (h==NULL) continue;
      strcpy(val,h+1);
      *h = '\0';
      cut_spaces(buffer);                         /* leading and trailing */
      if (!strcasecmp(buffer,key)) {              /* key found */
        cut_spaces(val);
        strcpy(value,val); r=value;
        break;
      }
    }
    fclose(fp);
  }

  return(r);
}

/* ---------------------------------------------------------------------- */

void put_string(char* file,char* key,char* string)
{
  int  i,n=0,found=0;
  char buffer[1024],*lines[1024],lkey[128],*h;
  FILE *fp;

  lines[0] = (char*)malloc(128*1024*sizeof(char)/8);
  sprintf(lkey,"%s = ",key); extend(lkey,' ',20);

  fp = fopen(file,"r");
  if (fp != NULL) {                               /* search for 'key' */
    while (fgets(buffer,sizeof(buffer),fp)) {
      cut_spaces(buffer); if (*buffer == '\0') continue;
      strcpy(lines[n],buffer);                    /* copy orig. line */
      h = strchr(buffer,';');                     /* comment */
      if (h) *h='\0';
      h = strchr(buffer,'=');                     /* delimiter */
      if (h==NULL) continue;
      *h = '\0';
      cut_spaces(buffer);                         /* leading and trailing */
      if (!strcasecmp(buffer,key)) {
        sprintf(lines[n],"%s %s",lkey,string);    /* replace line */
        found = 1;
      }
      lines[n+1] = lines[n] + strlen(lines[n])+1;
      n++;
    }
    fclose(fp);
  } /* endif(fp!=NULL) */
 
  if (found==0) {                                 /* append new line */
    sprintf(lines[n],"%s %s",lkey,string);  n++;
  }

  fp = fopen(file,"w");
  if (fp != NULL) {
    for (i=0; i<n; i++) fprintf(fp,"%s\n",lines[i]);
    fclose(fp);
  } else {
    fprintf(stderr,"utils.c/put_string(): cannot open '%s' for writing",file);
    fflush(stderr);
  }

  free((void*)lines[0]);
}

/* ---------------------------------------------------------------------- */

long get_long(char* file,char* key,long defvalue)
{
  char   buf[128],buf2[128];
  long   value,l;

  if (get_string(file,key,buf,NULL) == NULL) return(defvalue);
  if (!strncmp(buf,"0x",2)) {                      /* hex value */
    sscanf(buf,"%lx",&value); 
    return(value);
  }
  if ((l=strlen(buf)) > 1) {
    if ((buf[l-1] == 'h') || (buf[l-1] == 'H')) {  /* hex value */
      buf[l-1] = '\0'; 
      sprintf(buf2,"0x%s",buf);
      sscanf(buf2,"%lx",&value);
      return(value);
    }
  }
  return(atol(buf));                               /* decimal value */
}

/* ---------------------------------------------------------------------- */
 
double get_double(char* file,char* key,double defvalue)
{
  char   buf[128];
 
  if (get_string(file,key,buf,NULL) == NULL) return(defvalue);
  return(atof(buf));                               /* decimal value */
}

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */

