/* Balanced Tree algorithms...   See Knuth, V3, P455 */

/* ----- structure definition ----- */

typedef struct  bnode  {
    struct bnode	*llink;
    struct bnode	*rlink;
    int		rank;
    int		balance;
    void	*data;		/* General object pointer */
    } BNODE;

/* In BNODE, llink and rlink point to the next nodes in the tree;
  rank is 1 plus the number of nodes in the left sub-tree, and is
  used to "index" the n-th element, and to count total elements;
  balance is the difference in height of right and left sub-trees,
  used to keep tree in balance, must be -1 <= balance <= +1.  */

/* The compare sub, given two object pointers returns -, 0 or + to
  indicate the comparison of those values */

/*  ----  Function prototypes for btree.c routines  ----  */

void *btree_add (BNODE **head, void *data, int (*subr)() );

int  btree_length (BNODE **head);

void *btree_index (BNODE **head, int n);

void  btree_visit (BNODE *node, void (*subr)(), void *args);

void  btree_visit_rest (BNODE *node, int(*range)(), void(*subr)(), void *args);

void *btree_find (BNODE **root, void *key, int (*subr)() );

BNODE *btree_kill (BNODE *node, int (*subr)() );

/*  =*=  */

