/*  ioutils.h -- Input/Output Utility Subroutines */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2011 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef  INCLUDE_IOUTILS_H
#define  INCLUDE_IOUTILS_H

#include	<unistd.h>	// For access, stat
//#include	<stdio.h>	// redundant, done in kdcutil.h
#include	<string.h>	// Move to kdcutil.h if needed there
#include	"kdcutil.h"	// For vect2 type
// Putting vect2 into ioutils.c
// Works if kdcutil.h is included before this...
// But NOT in some other includes of ioutils without kdcutil

/*  ====  Definitions  ====  */
#define  AN  '\0'
//#define  WHITESPACE " \t"
// Define as in isspace() -- see man isspace
//#define  WHITESPACE " \f\n\r\t\v"
// Add characters inserted by word processors
#define  WHITESPACE " \f\n\r\t\v\xa0\xc2"
#define  BLANKSPACE " \t"

#ifdef IS_IOU_C
#define EXT_IOU
#else
#define EXT_IOU extern
#endif

#define  C_loop(q,a)  for(q=(a); q!=NULL; q=(q->next!=(a))?q->next:NULL)
/* C_loop (indexpointer, headpointer)
  Forms a "for" loop over all elements of the (possibly circularly)
  linked list, using the "next" element pointer.  Loop terminates with
  the index pointer NULL if a NULL pointer is encountered, or the
  head pointer is encountered in a "next" link.  Use of "continue"
  or "break" will work normally.  The head pointer may be modified,
  so long as the "next" pointer referencing it is also modified.      */


/*  ====  Typedefs  ====  */
typedef  struct  namlist {
        struct namlist  *next, *last;
        char    *name;
        } namlist;

typedef struct  {
	char	*buf;
	size_t	size;
	}  Dbuf;
// above was formerly "Dchr"

/*  ----  Data in support of version control  ----  */

typedef struct	Vers {	/* Version structure */
	struct	Vers	*next, *last;
	float	v;
	float	m;
	char	*tag;	// was nam
	}  Vers;
// above was formerly "Vers"

typedef struct  Pvfs {	/* Program version feature structure */
	Vers	*ver;	// was pri
	namlist	*feat;	// was feat
	char	*date;	// was date
	char	*name;	// was name
	}  Pvfs;
// above was formerly "Phead"

// typedef struct  vector2 { double x, y; } vect2;
// Above would conflict with definition in kdcutil.h

/*  ====  Globals  ====  */
EXT_IOU	char	LastXopen[256];

// EXT_IOU	char*	FNT[32]={[0 ... 31]NULL};
// FNT is static global local to ioutils.c

/*  ====  Function Prototypes  ====  */

/*  ====  String Manipulation  ====  */
void  trimend (char* a, char* space);
void  trimbeg (char* a, char* space);
void  trims (char* a, char* space);
char*  dyncat (char* pfx, char* nam);
char*  fncat  (char* pfx, char* nam);
char  *isotime (char *tbuf);
char*	upcase (char* s);
char*	lowcase (char* s);
char*	upcasen (char* s, int n);
char*	lowcasen (char* s, int n);

/*  ====  Queue Manipulation  ==== */
char	*dynamstr (char *a);
char    *pop_name (namlist **a);
namlist	*get_nam (char *a);
namlist *push_name (namlist *b, namlist *a);
namlist *pushtxt (char *text, namlist *a);
namlist	*kill_name (namlist *a);

namlist	*new_nam (char *a);
//namlist *push_nl (char* a, namlist* p);

/*  ====  Prompting  ====  */
void  askme (char* buf, int blen, char* prompt);
void  askmestr (char *buf, int blen, char *message);

/*  ====  I/O Routines  ====  */
/*  ====  File Name Support  ====  */
char*	fnx (int fd);
char*	fnt (int fd);
char*	fhn (FILE* f);
char*	fhname (FILE* f);
void	setfnt (int  fd,  char*  file);
void	clearfnt (void);

/*  ====  Open / Close Support  ====  */
FILE*	iopen (char* fs, char* tp);
FILE*   xopen (char* fs, char* tp);
FILE*	iopenx (char *fs, char *tp, namlist *exts);
// char*	findrfile (char* fname, namlist* nl);
void	findfile (char** spec, char* mode, char* name, char* dir);
void	iclose (FILE* f);

void	evtran (char* *nam);
/* -- "locfile" goes here -- */

/*  ====  Parsing Routines  ====  */

char*   skipover (char* line, char* delim);
char*    parse (char* line, char*  token, char* delim);

char*   dparse (char* line, char* *token, char* delim);
char*	getok (char** line, char* delim);
int	itok (char* *line, char* delim);
double	ftok (char* *line, char* delim);
double	postok (char* *line, char* delim);
long	ltok (char* *line, char* delim);

void	xcat (char** a, char* b);
char*	extok (char** line, char* fredel, char* specdel);
namlist*  cdlparse (char* line, char* fredel, char* specdel);
int	kval (char* tok, double* v);

char*	cfilename (char* fspec);

size_t	szx (char* x, int k);
char*	cpyan (char* dest, char* src);


/*  ====  Floating Point Version Support  ====  */

char*	fvstr (float a);
char*	vstr (float a);
char*	combinev (float a, float b);

/*  ====  Parameter File Reading  ====  */

void*  get_data (char* filespec, void* (*userprog)(char*, void*),
	char* delim,	char* leadcom,
	char* anycom,	char* dxcom,
	char* xdcom,	char* dxdcom );

/*  ====  Sexigisimal decode/encode  ====  */

double	posvalue (char*  field);
int	intvalue (char*  field);
long	longval  (char*  field);
double	floatvalue (char*  field);
vect2	vect2value (char* field);
vect2	complxvalue (char* field);
vect3	vect3value (char* field);
char*	decomp (int i, char s, int n, int r);
char*	dsexig (double val, int np, int nd, char* *buf);
char*	sexig (double val, int np, int nd);
void	sexigw (char* buf, double val, int np, int nd);

/*  ====  Date decode and encode  ====  */

int	nmonth (char* name);
long int mjdate (int y, int m, int d);
long int mjd4cdate (char* date);
void    date4mjd (long int mjd, int* year, int* month, int* day);
long int mjd4tm (time_t* now);
int	weekofyear (long int mj);
//char*	dayandtime (void);
char*	dayntim (char* date, char* tfld);

/*  ====  Binary Data Debug Utilities  ====  */
void	hdstr (char* title, unsigned char* dat, int n);
void	cdstr (char* title, char* dat);

/*  ====  Binary Encoding Utilities  ====  */
int	hex2bin (char* hx, unsigned char* bn, int n);
int	hex4bin (char* hx, unsigned char* bn, int n);
char*	hextex (char* xx, int n);
char	sixbc (int c);
int	bcsix (char c);
int	six2bin (char* sx, unsigned char* bn, int n);
int	six4bin (char* sx, unsigned char* bn, int n);

/*  ====  Dynamic Buffer support  ====  */

void	Dsadd (Dbuf* d, char* str);

/*  ====  Support for version control  ====  */

Vers*	kill_vers (Vers* a);
Pvfs*	clean_pvfs (Pvfs* a);
Pvfs*	kill_pvfs (Pvfs* a);

int	fvcmp (float a, float b);
int	vvcmp (Vers* a, Vers* b);

int	hasfeat (char* name, namlist* list);
Vers*	findver (char* name, Pvfs* head);
int	comp_vers (Vers* v, Pvfs* p);

Vers*	addvers (char* tag, float v, float m, Vers* head);
namlist*  addfeat (char* name, namlist* head);
namlist*  addvfeat (char* name, int v, namlist* head);
namlist*  addxfeat (char* name, char* v, namlist* head);

char*	putvfs (Pvfs* head);
Pvfs*	getvfs (char* bufr, char* source);

char*	getafile (char* filename);
void	putafile (char* buf, char* filename);

void	displist (namlist* list);

/*  ====  Support for libcurl actions  ====  */

char*	getFTP (char*  url);


/*  ====  */
#endif     /*  INCLUDE_IOUTILS_H  */
/*  ====  */
