/*  intgui.c == Astronomer interface GUI task */

/* Input is an obs file, or if none, all defaults to start.  Output
  is edited obs file, possibly tested through mask generation. */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2011 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Optional command line argument is a .obs file; 
  output is edited .obs file, via the GUI option "save" or "save as".  */

/*  ====  STANDARD INCLUDES  ====  */
#include	<stdio.h>
#include	<stdlib.h>
#include        <string.h>
#include	<unistd.h>	/* for access */
#include	<ctype.h>	/* For character search */
#include	<math.h>

/*  ====  GLOBAL DEFINITIONS  ====  */

const	char	PROGNAME[] = "Observer Interface GUI";

// version stuff moved, see below.
//const	char	VERSION[]  =  "4.43";
//extern	char	SKWVERS[];	// Extended version for sky window
//char	VERSION[16] = "4.42";	// Should use INTGUIV in quotes
//char	XVERSION[16] = Xstr(INTGUIV);
// above moved to after include of mgfeats
// see it below for automatic initialization of version stuff...

//int     dodebug = 0;	/* Global debug flag */
// dodebug defined in optutils, included in mgutils.h stuff...
unsigned  int	TESTMODE = 0;	// Can be set true when executed as tintgui

// #include	"mgfeats.h"
// now features come with maskdat.h included by mgutils.h below

#undef  WDEBUG
//#define  WDEBUG
// WDEBUG is debug flag for window appearance...
// Allows output of alignment lines, and appearance of the test button

#define  ANGLEDBUG
//#undef  ANGLEDBUG
/*  Symbol  ANGLEDBUG  is defined to use the "Test" button to provide a
  report of rotator angle computation.  When not defined, the old use
  of "Test" to test dialog boxes will remain.  These are effective only
  when the WDEBUG flag above is defined, and when "tintgui" is used as
  a test mode program; otherwise none of this is done.    */

#define  REFRESH
//#undef  REFRESH
/*  Symbol  REFRESH  enables the feature of the -r flag causing an
  automatic immediate save and exit (if data has been read).  This
  is to provide proper checksum, signature and formatting for an
  .obs file perhaps generated by a script.    */

// ----- Below are defines which are mostly fixed -----

#undef  OBSOUT
// OBSOUT enables adding obsolete items to the .obs file output

#define  IFUM
// #undef  IFUM
// IFUM is enabling flag for IFU mode support

#define  SKYOPT
// #undef  SKYOPT
// SKYOPT is enabling flag for sky window features

#define  INTGUI
// INTGUI is used to omit bad calls in simulation cases...
// (Used in guisubs.c only)

//#define  DISPXPTR
/*  Symbol  DISPXPTR  is defined to implement the separate disperser menu
  method, replacing the indexing into a single array of menu entries which
  required a clumsy synchronization between an enum and a data array.  See
  the guimenu.otl document for more details.  */

/*  Code now assumes the  DISPXPTR  symbol is defined, and conditional
  statements are being removed.  */

#define  OBJSHAPE
//#undef  OBJSHAPE
/*  Symbol  OBJSHAPE  is defined to enable a pop-up menu window to select
  the shape of the default slit.  */

#ifdef  OBJSHAPE
//#define  OBJAUTO
#undef  OBJAUTO
/*  Symbol  OBJAUTO  is defined to implement the shape pop-up as auto, and
  undefined to implement it as non-automatic.  One should be decided
  upon and only that one used.  If  OBJSHAPE is not defined, this symbol
  should also be undefined.  */
/*  It turns out that non-automatic is the easier and more clear method
  of handling this menu; leave it undefined and eventually remove all
  the auto items.  Keep for a while as documentation.  */
#endif

// Global storage for rotator angle computation, warning
double	hrf = 30.0;	// Rotator flip hour angle for warning
double	lat = -29.01418;	// Latitude of IMACS
int	warnflag = 0;	// Global warning flag
// See below for rotaq and rotbq globals.

int	NoADC = 0;	// Special telescope name
// ADC installed 2004/5, now default is 0 rather than 1
// The NoADC now specifies ldss operation 2005/2


typedef struct  {
	int	dbug;
	int	size;
	int	skwa;
#ifdef  REFRESH
	int	refr;
#endif
//#ifdef	CACHECAT
/* Symbol  CACHECAT  is not defined yet, that happens in MGUTILH
  just below.  However, there's no harm in putting this item into
  the structure even when it is not used.    */
	int	crep;
//#endif
	}  opts;

int	FontSize = 18;

/*  ====  Utility programs, library programs, debug code.  ====  */

#include	MGUTILH
// #include	"mgutils.h"
/* Above is all that's needed -- it includes everything else... */
/* mgutils.h includes:  ioutils, kdcutils, maskdat header files */

/* All mgfeats.h things are now defined here... */

// This is the new way to initialize version strings, after mgfeats included
// compatable with old version, but needs to be placed later.
const	char	VERSION[] = Xstr(INTGUIV);
char	XVERSION[16] = Xstr(INTGUIV);
// above needs include of mgfeats.
// XVERSION is char and not const char because we replace it later
// 

// Standard Version structure
const	Vers	Program = { NULL, NULL,  INTGUIV, INTGUIR, INTGUIT };

#include	"cxt.h"		/* Included from library directory */
/* The cxt.h replaces the older ccbxlib.c which is now unused. */

#include	"utils.h"	/* No longer included by ccb lib */

#include "datutil.h"		/* New data utilities (from gmap) */

#include	"intgui.h"	/* Definition of interface */

#include	"msdebug.c"	/* Debug subroutines */

#ifdef  SKYOPT
#include "skywin.h"
#include "skymap.h"
#endif

/* The graphic global which needs to be defined real soon, and
  established like right away on startup... */
// Gob

Obf	*Gob=NULL;		/* Graphics section obf pointer global */
// Above is a globalization kludge...


/* Test only switch for using web interface for catalog lookup */
#ifdef  EXCAT
//extern Bool    USEXMLRPC;
extern int    USEXMLRPC;
#endif

#ifdef  IMAGM
//extern int    IMAGUI;
int	IMAGUI = 0;
#endif

/* Globals needing defines from intgui.h */
anglist  *rotaq = NULL;	// Queue of rotator flip angles
bandlist *rotbq = NULL;	// Queue of occultation bands
bandlist *rotcq = NULL; // Combined warning queue
bandlist *rotdq = NULL; // Queue of warning bands
//bandlist *roteq = NULL;
pdata	pdx;		// Needed for computing rotator angles

// Globals for GUIDATE feature:
double	rats = -25.0;	// R.A. of True Sun, hours
double	dets = 0.0;	// Dec. of True Sun, degrees
#ifdef  GUIDATE
double	rafc;		// R.A. field center in hours
double	defc;		// Dec. field center in degrees
double	eqfc;		// Equinox of fielc center
vect3	vfc;		// Vector position of field center
double	lmm;		// Local mean midnight in days from 2000.0
double	stlm;		// Mean Sidereal time in hours of local mean midnight
double	stm;		// Sidereal time at true solar midnight
double	twit;		// Twilight time difference from stm, sidereal hr.
double	srt;		// Sunrise time, sidereal hours from stm
double	sst;		// Sunset time, sidereal hours from stm
#ifdef  MOONLINE
double	ltt;		// Lunar transit time diff from lmm in sidereal hours
double	mrt;		// Moonrise time, in sidereal hours from ltt
double	mst;		// Moonset time, in sidereal hours from ltt
int	pci;		// Percent illuminated
int	adm;		// Angle distance to moon
#endif  // on MOONLINE
#endif  // on GUIDATE
double	cepoch = 2000.0;	// Current epoch
double	wepoch = -1.5;	// Window's epoch
int	cmjd = 0;	// Current mjd, set when initialized
int	mjd = 0;
int	yr = 0;
int	mo = 0;
int	dy = 0;
int	baseyear = 2002;


char	myname[32];	// Fill in our command name here
/* The command name is used to decide whether to run maskgen or
  tmaskgen for the active button.  Allows test debugging.  */

/*  Extern definitions for sky window stuff  ----------  */

#ifdef  SKYOPT
extern	int	nguiders;
extern  int	skywin_active;
#endif

/* Unlike all other other guiders at Magellan,
 * the IMACS guider is not mounted square with the rotator.
 * When IMACS is in its "nominal position",
 * the rotator encoder reads the following value.
 * Unfortunately, right now, this value seems to change
 * everytime they unhook and rehook up IMACS to the rotator.
 * They're going to try to fix this.
 */
//  #define CHUECO (-44.936)	// Early set-up
//  #define CHUECO (-47.340)
//#define CHUECO (-46.150)
//  #define CHUECO (-45.0)	// This will be intended standard
//#define LDSSROA (-60.3)

// Those offset angles are now in mgutils.h for general
// distribution to more programs.
// Current CHUECO is (-46.150)

double	IMACSrot = CHUECO;
char	Chuecn[] = "IMACSCHUECO";	/* Environment name */
double	RotOff = CHUECO;

/*  Add rotator limits.  Basic rotator limit is ROTLIM, and
  we will warn between 1.0 - ROTLIM and -180.0 degrees, and
  also between ROTLIM - 1.0 and +180.0 degrees.  This gives
  an additional 1 degree buffer past the hard limit ROTLIM.
  Environment name ROTLIMIT may be used to change this.
  This limit applies to both IMACS and LDSS-3.      */

#define  ROTLIM  (178.0)
//char	Rotlimn[] = "ROTLIMIT"

/*  Add individual rotator limits.  These default to the 
  standard values defined by ROTLIM, but may be changed by 
  the environment names ROTATORLOW and ROTATORHIGH when
  needed.  These limits apply to both IMACS and LDSS-3.    */

/* Temporarily as of December 2010, the low limit is set to
  -151 degrees.  This is implemented as a hardware limit for
  the rotator motion, so that the IMACS wrap limit is not reached
  at all, even during a slew motion.  In the event that this
  limitation can be removed, change this data.  This limit may
  also be dynamically changed by using "ROTATORLOW=XXX" in the
  mgvers.dat file in the FTP area.    */

//#define  DROT1  (1.0 - ROTLIM)   // 1.0 - 178.0 = -177.0
#define  DROT2  (ROTLIM - 1.0)   // 178.0 - 1.0 = +177.0
#define  DROT1  (-151.0)   // "Temporary" restriction

double	RotLim1 = DROT1;
char	RotLimLow[] = "ROTATORLOW";

double	RotLim2 = DROT2;
char	RotLimHigh[] = "ROTATORHIGH";

/*  Add IMACS angle limits.  IMACS is limited in ccw motion
  (positive IMACS angle) at about 105 degrees, and this 
  occurs only near the zenith.  The global IMACSlimit is
  set to the negative of the lowest possible rotator angle 
  observed when the telescope is at zenith.  The environment
  name IMACSLIMIT may be used to change this.  This limit
  applies only to IMACS.  It is assumed that a small safety
  value of about one degree is built in to this value.  A
  rotator warning is given when the IMACS angle falls in the
  range IMACSlimit + CHUECO to 180.0 + CHUECO.       */

double	IMACSlimit = 151.0;  // Standard is 151.0;
char	IMACSlimnam[] = "IMACSLIMIT";


/*  Globals for sky window stuff  ----------  */

// These globals are for the datutil stuff to get gmap data
gdata      *GDAT;		/* Global Data structure */
int        nsects[3];
section    **tsections;
section    **csections;
section    **gsections;
double     tellong,tellat;              /* telescope geography */
int        dimx,dimy,nchips;
char       t_name[64],g_name[64],c_name[64];
// enum  dmenus  { MNU_TEL, MNU_GDR, MNU_CAM };

double	RAsave, DECsave;
int	recent;

/*  ====  Global data for IFU mode support  ====  */

/* Offsets are in arc seconds from nominal center, - is east and south */
double	IFUdx[2] = { -13.1, 13.9 };
double	IFUdy[2] = { -45.7, 16.8 };
// double	IFUdx[2] = { 14.3, -14.3 };
// double	IFUdy[2] = { 31.1, -31.1 };
// double	IFUdx[2] = { -29.55, +29.55 };
// double	IFUdy[2] = { -12.0, -12.0 };

/* Data for global Nod & Shuffle rotation and rotator bias.  */
int	nrbiask = 0;	// Integer flag field

/*  ====  Function Prototypes for global support  ====  */

#ifdef  GUIDATE
static void   setup_dates (void);
#endif

/*  ====  Data structures for interface GUI  ====  */

// Moved to intgui.h

#define  FORCEOBS
//#undef  FORCEOBS
/*  Symbol  FORCEOBS  is defined to force input files to have the .obs
  extension, to change input requests to use that extension, and thus
  to only allow such files to be opened.  Also, if a file is not found,
  the name to create it is forced to the .obs extension.
  If undefined, other extensions are allowed on obs files.
  Actually, when the file is written, it is done with the .obs extension;
  what we can do with this symbol is to allow reading of the given
  extension file rather than forcing the input to .obs immediately.
  As originally written, assumed to be defined.    */

/* All obs files are written with .obs extension, and since FORCEOBS is
  normally defined, all reading forces a .obs extension to be used even
  if the specified file exists and has a proper checksum etc.  */

/*  Filling in the default obs structure, from a file, or with simple
  defaults to initialize  */

/* Subroutine to free an obs structure */
/* (See kill_obs in mgutils.c) */

static	double	drange (double v, double modulus, double limit)
/* Move v by modulus so that it is within +/- modulus, and not
  greater than limit which should not exceed modulus.  */
{
    while (v < 0.0) v += modulus;
    while (v > modulus) v -= modulus;
    if (v > limit) v -= modulus;
    return  v;
}

/* Subroutine like findfile to append the .obs to obs file names */
/* Follows the method of findsmdf in itest.c here... */
void	findobs (char** spec, char* mode, char* name, char* dir)
{
static  char  ext[]=".obs";
//#ifdef  LOCATE
#if defined (LOCATE) && ! defined (FORCEOBS)
char	*s;
#endif
char	*n;
char	*c;

/* Look for a current result, return if so. */
    if (*spec != NULL) return;

/* Without a name, we can't do a thing. */
    if (name == NULL) return;

/* See if name has ".obs" in it, and just try that. */
    if (strstr (name, ext)) {
#ifdef  LOCATE
	*spec = locfile (mode, name, NULL, dir);
#else
	findfile (spec, mode, name, dir);
#endif
    } else {
/* Otherwise, strip any trailing .xxx, append the extension and try that. */
//	n = strdup (name);	// changed to dynamic buffer
	n = calloc (1, strlen(name)+strlen(ext)+2);  // to allow strcat
	strcpy (n, name);
	c = strrchr (n, '.');
#ifdef  LOCATE
#ifndef  FORCEOBS
// This section is not normally used
// Try two names; as given, and stripped of an extension if one exists.
	s = calloc (1, strlen(ext)+2);
	sprintf (s, ",%s", ext);
/* Need to explicitly add null extension to what is searched for */
	*spec = locfile (mode, n, s, dir);
	if (c != NULL && *spec == NULL) {
	    *c = AN;
	    *spec = locfile (mode, n, s, dir);
	}
	free (s);
#else  // on FORCEOBS true
// This is the section normally used
	if (c != NULL) *c = AN;
	*spec = locfile (mode, n, ext, dir);
#endif  // on FORCEOBS
#else  // on LOCATE -- undef section follows
#ifndef  FORCEOBS
// This is not normally done
	findfile (spec, mode, n, dir);
#endif  // on FORCEOBS
// This section is the "old way" before locfile
	if (c != NULL) *c = AN;
	strcat (n, ext);
	findfile (spec, mode, n, dir);
#endif  // on LOCATE
	free (n);
    }

}


	/* - - - - - - - - - - *\
	|			|
	|  File name utilities	|
	|			|
	\* - - - - - - - - - - */

/*  ----  Character handling routines (candidates for ioutils)  ---- */

char*	limcpy (char* dest, char* src, int n)
/* Copy up to n characters from src to dest; put a null following
  what is copied, for a total of at most n+1 characters in dest.
  Return the pointer to dest. */
{
int	s;
char	*c;
    s = strlen (src);
    if (s > n) {		// Artifical truncation
	c = strncpy (dest, src, n);
	dest[n] = AN;
    } else {			// Copy s+1 characters.
	c = strncpy (dest, src, s+1);
    }
    return  c;
}

static	bool	isvalidch (char c)
/* Return true if character c is valid in a file name */
{
/* Check a few obvious things, like delimiters */
    if (c == '/') return False;
    if (c == '.') return False;

/* Look out for those control characters */
    if (c <= ' ') return False;

/* Check for some basic allowable things */
//	if (isalnum(c)) return True;
    if (c == '-') return True;
    if (c == '+') return True;
    if (c == '_') return True;

/* Now, we really need a default thing */
/* Should return false if not printable character */
//	return  True;
/* An alternate is to specifically include those things which
  we like, and finally return isalnum(c) for all else. */
    return  isalnum(c);
}

char*	filenamloc (char* fspec)
/* Return pointer to location in fspec of the start of the filename
  portion.  Finds end of filename (last dot or end) and backtracks
  to the location of the start of the name */
{
char	*c, *e;

/* Take care of a few null cases */
    if (fspec == NULL) return fspec;

/* Find the end of the file name (or end of string) */
    e = strrchr (fspec, '.');
    if (e == NULL) e = fspec + strlen(fspec);
    while ( e > fspec && ! isvalidch(*e) ) e--;
    if (e == fspec) return e;		/* No filename */
/* Now, e is the last character in the filename */

/* Find beginning of file name string... */
    c = e;
    while ( c > fspec && isvalidch(*c) ) c--;
    if (!isvalidch(*c)) c++;  /* Went too far, back up */
    return  c;
}

char*	getfilenam (char* fspec)
/* Return a local pointer to a copy of a filename, extracted from
  the fspec text. */
{
static	char	buf[16];
char	*c, *e;
int	n;

/* Take care of a few null cases */
    *buf = AN;
    if (fspec == NULL) return buf;

/* Find the end of the file name (or end of string) */
    e = strrchr (fspec, '.');
    if (e == NULL) e = fspec + strlen(fspec);
    while ( e > fspec && ! isvalidch(*e) ) e--;
    if ( e == fspec && ! isvalidch(*e) ) return  buf;  /* No filename */
/* Now, e is the last character in the filename */

/* Find beginning of file name string... */
    c = e;
    while ( c > fspec && isvalidch(*c) ) c--;
    if (!isvalidch(*c)) c++;  /* Went too far, back up */
/* Now, c is first character in the filename */
    n = 1 + e - c;	/* Characters in filename */
/* Should probably use filenamloc and a scan to find c and e ? */

/* Copy up to 16 characters to buf... */
    if (n > 15) n = 15;
    strncpy (buf, c, n);
    buf[n] = AN;	/* Supply the terminating null */
    return  buf;
}



	/* - - - - - - - - - - - - - - *\
	|				|
	|   Read  Observation  file	|
	|				|
	\* - - - - - - - - - - - - - - */


/* Subroutine to obtain obs structure from file, or fill in defaults */
/* Fills in filename with file name actually read from */
Obs*	newobs (char* filename, element* gs)
{
Obs	*os = NULL;
FILE	*INF = NULL;
char	*inspec = NULL;   /* Input file spec. */
char	*cp;
char	fname[10];	/* Holds fname for obs structure */
// char	Tname[16];	/* Default telescope */
#ifdef  FILESEC
int	k;	// status of file security
#endif

/* If the file name is blank, or null, skip the finding of inspec... */

/* Verify file location and readable.  Else, return null */

    if (filename != NULL) {
	if (strlen(filename) > 0) {
// NOTE -- need to add LOCATE here and fileloc call
#ifdef  LOCATE
	    inspec = locfile ("r", filename, ", .obs, .obsl", " ");
/* locfile will automatically append an expansion of PathName since
  the directory field is not a zero length string.
  Were we to specify PathName for directory, a duplicate path would
  be appended and searched.    */
#else
	    findobs (&inspec, "r", filename, "");
//	    findobs (&inspec, "r", filename, "SMDATA");
	    findobs (&inspec, "r", filename, PathName);
#endif
// NOTE THAT inspec is a dynamic string, and must be freed later...
	}
    }

/* Add a section here, such that if filename is NULL, or possibly 
a null string, we instead call "defobs(gs)" to supply a clean
obs structure for initial things... */
    if (inspec == NULL) {
	if (filename != NULL && strlen(filename) > 0) {
	    printf
	    (" ** \"%s\" not found.  Use blank data.\n", filename); //debug
	    printf 
	    ("  (You might want to set %s properly.)\n", PathName);
//	    ("  (You may use \"setenv SMDATA xx\" to set a directory.)\n");
	}
	os = defobs (gs);	// A new strucrure created...
//	strcpy (filename, "");	// And kill the filename
//	limcpy (fname, os->fname, 8);	// copy the default here
/* Try assigning fname as requested instead... */
	if (strlen(filename) > 0) {
	    limcpy (fname, filenamloc(filename), 8);	// supplied
// null out any . in fname
	    cp = strchr (fname, '.');
	    if (cp != NULL) *cp = AN;	
	} else {
	    limcpy (fname, os->fname, 8);	// null input
	}

//	strcpy (Tname, NoADC ? "MagNoADC" : "Magellan");
//	strcpy (Tname, NoADC ? "Mag2NoADC" : "Magellan");
// Below Tname is actaully unused
//	strcpy (Tname, NoADC ? "Magellan2" : "Magellan");
//	os->Telescope = get_optic (Tname);
//	os->Telescope = NULL;
	normalize_obs (os);	// Set things up correctly
    } else {

// Below is how it is done in maskgen itself...
/* Open the file, get file pointer, "INF" */
//	printf (" == Reading \"%s\".\n", inspec);  //debug
	strcpy (filename, inspec);
	INF = iopen (inspec, "r");
	if (INF == NULL) {
	    strcpy (filename, "");
	    return NULL;
	}

/* Obtain new obs structure from read_obsfile */
	os = read_obsfile (INF, gs);
/* That does the full read of observation file, but does not read
  any of the objects to form a queue (thus saving time). */

#ifdef  FILESEC
	k = verifide (INF, dodebug-1);
/* Currently, let even the fail status pass by */
#endif

/* See if we have the NoADC case... */
		NoADC = (strcmp (os->Tel_scope->name, "MagNoADC") == 0) ||
		(strcmp (os->Tel_scope->name, "Magellan2") == 0) ||
#ifdef  WFCCD
		(strcmp (os->Tel_scope->name, "DuPont") == 0) ||
#endif
		(strcmp (os->Tel_scope->name, "Mag2NoADC") == 0);
// ALSO NEED TO CHECK FOR Mag2NoADC ??
// There is no longer an imacs no adc mode; use for ldss only!
// IMACS have ADC (0) but LDSS does not (1).
//	if (strncasecmp(os->instrument,"LDSS",4) == 0)
	if (os->Instr_mt != NULL) {
//int	ii = InstIndx(os->Instr_mt->name);
int	ii = InstIndx(ObInstN(os));
//	    if (strncasecmp(os->Instr_mt->name,"LDSS",4) == 0)
//		  RotOff = LDSSROA;
//	    else  RotOff = CHUECO;
// SHOULD USE InstIndx IN THE ABOVE FOR FINDING OFFSET
	    if      (ii == I_lds2) RotOff = LDSSROA;
	    else if (ii == I_lds3) RotOff = LDSSROA;
#ifdef  WFCCD
	    else if (ii == I_wfcd) RotOff = 270;	// NOTE - change?
#endif
	    else                   RotOff = IMACSrot;  // was CHUECO;
	} else  RotOff = CHUECO;	// default

/* Obtain the fname from the file specification */
	strncpy (fname, getfilenam(inspec), 8);
/* Using the local getfilenam rather than cfilename to eliminate any
  non-valid characters from the filespec in the fname field. */
	fname[8] = AN;

/* Close the file */
	iclose (INF);
	free (inspec);
    }
    strncpy (os->fname, fname, 8);	/* Get the filename */
    if (dodebug > 1) dumpobs (os);

/* Set-up for features */
#ifdef  GUIDATE
    if (os->date > 0) {
	mjd = os->date;
	setup_dates ();
// We skip setup_dates here as the needed app pointer and colors
// have not yet been defined if this is the initial display.
// Also, the cmjd is not yet defined if this is initial.
// (Those things have now been added as conditional in setup_dates.)
//	solpos (mjd);
/* Note that solpos will be done again when window epoch has been
  set, later in the initialization.  So it is not always needed here. */
    }
#endif

/* Return result */
    return  os;
}


	/* - - - - - - - - - - - - - - *\
	|				|
	|   Write  Observation  file	|
	|				|
	\* - - - - - - - - - - - - - - */


/* Subroutine to write an obs structure to an output file, as a
  text entity.  Ultimately used to save data structure... */
bool	writeobs (char* filename, Obs* ob)
{
FILE	*f;
namlist	*namp;
// double	scale;
char	fn[12];
char	buf[32];
char	message[256];
int	yr, mo, dy;
// int	k=0;	// scratch
char	*cp;
anglist	*al;
bandlist *bl;
//#ifdef  GISMOAP
int	gisap = 0;	// always defined (easier)
//#endif

/* Check for any null case stuff, and message if so. */
    if (filename == NULL) {
	printf (" ** Can't save to NULL file name.\n");
	return False;
    }
    if (*filename == AN) {
	printf (" ** Can't save to blank file name.\n");
	return False;
    }
    if (ob == NULL) {
	printf (" ** Can't save NULL obs structure.\n");
	return False;
    }

/* Drop those Unnamed.obs and .obs file names we might get... */
//    strncpy (fn, getfilenam(filename), 8);
    strncpy (fn, cfilename(filename), 8);
    if (!strcmp(fn, "Unnamed")) {
	printf (" ** We don't write \"Unnamed\" as a file.\n");
	return False;
    }
    if (strlen(fn) < 1) {
	printf (" ** Can't save to a null named file.\n");
	return  False;
    }

/* Set up a special flag for GISMOAP case, indicating whether a special
  instrument name exists.  If special, some tests are skipped, and some
  output is skipped.  */
#ifdef  GISMOAP
//    if (strncasecmp(ObInstN(ob), "IMACS_gd", 8) == 0) gisap = 1;
    if (FeatIndx(ob) == F_gisap) { gisap = 1; ob->reflimit = 0; }
#endif

// Check for rotator angle flip, give warning box, and only
// proceed if it is OK
//  if (hrf < 24.0 && gisap == 0 && ob->reflimit > 0) {
    if (fabs(hrf) < gsda (30.0, defc, pdx.lat) && 
	gisap == 0 && ob->reflimit > 0) {
//	sprintf (message, " WARNING!  Rotator flips at %.2f hours!", hrf);
	sprintf (message, " WARNING!  Rotator limit at %.2f hours!", hrf);
	if (!CBX_AlertBox("WARNING", message)) return False;
    }

// Consistency tests
// in writeobs

// Check for a valid Instrument and Disperser selected
#ifdef  IFUM
    if (ob->Instr_mt == NULL && ob->IFUmode == 0) {
#else
    if (ob->Instr_mt == NULL) {
#endif
	sprintf (message, " WARNING!  No Instrument selected!");
	if (!CBX_AlertBox("WARNING", message)) return False;
#ifdef  IFUM
    } else if (ob->Disperser == NULL && ob->IFUmode == 0 && gisap == 0) {
#else
    } else if (ob->Disperser == NULL && gisap == 0 && ob->reflimit > 0) {
#endif
	sprintf (message, " WARNING!  No Disperser selected!");
	if (!CBX_AlertBox("WARNING", message)) return False;
    }

// if extra orders is selected, check that spectra do not self-overlap
// with wavelength limits and adjacent orders.  Check order x high wavl
// does not exceed (order+1) x low wavel.
// ONLY check if extra orders of objects (vs. reference) is on
#ifdef  IFUM
    if ((ob->ex_order & 2) && !ob->IFUmode) {
#else
    if (ob->ex_order & 2) {
#endif
int	ord = ob->order;
	if ( (ord * ob->pb.red) > ( (ord+1) * ob->pb.blue) ) {
	    sprintf (message, 
		" WARNING!  Order %d red overlaps order %d blue!",
		ord, ord+1 );
	    if (!CBX_AlertBox("WARNING", message)) return False;
	}
    }

#ifdef  GSREQUIRED
/* Count the guide stars selected, and if IMACS require 2 and warn
  if fewer have been selected... */
//    if (InstIndx(ob->Instr_mt->name) == I_imacs) {
//    if (InstIndx(ObInstN(ob)) == I_imacs && gisap == 0) {
    if (InstIndx(ObInstN(ob)) == I_imacs && ob != NULL && ob->reflimit > 0) {
int	k;
// Count the number of selected guide stars, should be 2
	k  = (ob->gp1.ep == 0.0) ? 0 : 1;
	k += (ob->gp2.ep == 0.0) ? 0 : 1;
	if (k < 2) {
	    sprintf (message,
		" WARNING!  Fewer than 2 guide stars selected!");
	    if (!CBX_AlertBox("WARNING", message)) return False;
	}
    }
#endif  // on GSREQUIRED

/* Check here for any other possible errors, omissions, or inconsistancies
  which might give rise to a bad obs file.  Issue alert box messages
  for each, and make user press the continue button. */


// Open the named file for writing, or error message if not
    f = iopen (filename, "w+");  // Always use "w+"
    if (f == NULL) {
	printf (" ** No save possible to \"%s\".\n", filename);
	return False;
    }

/* Normalize the Obs structure; check for new element names, and fill
  in old style optics data based on those.  */
    normalize_obs (ob);
/* The normalize function will also compute the proper transform matrix
  in the Obs structure; that is needed because normalize will also
  do a small check on the guide star locations, and can clear any
  guide stars found to be bad.  */

/* Write all the things we have in the obs structure, in a nice
  order, into the open output file... */

    fprintf (f, "# Obs file (%s)\n", filename);

/* New official program line for reading by read_obsfile  */
    fprintf (f, "# Written By:  %s %s \n",
	Program.tag, fvstr (Program.v) );

/* Come up with date (pacific time date) here -- from mgutils */
    jd2date (jdnow()-(8.0/24.0), &yr, &mo, &dy);
    sprintf (buf, "%d-%02d-%02d%c", yr, mo, dy, AN);
/* A better way:
    time (&now);
    tm = localtime (&now);
    strftime (buf, "%Y-%m-%d", tm);
  see "man strftime" for details.    .. */

//    fprintf (f, "! Edited %s By %s version %s%s\n", 
//		buf, PROGNAME, VERSION, SKWVERS);
//    fprintf (f, "! Edited %s By %s version %d.%02d.%02d\n", 
//		buf, PROGNAME, INTGUIP, INTGUIV, SKYWINV);

//    cp = combinev (INTGUIV, SKYWINV);
//    fprintf (f, "! Edited %s By %s version %s\n", 
//		buf, PROGNAME, cp);
//    free (cp);	// cp is a scratch pointer here.

    fprintf (f, "! Edited %s By %s version %s\n", 
		buf, PROGNAME, XVERSION);

/* Remove any embeded blanks in observer and file name by replacing them
  with underscores before writing... */
    sprintf (buf, "%.16s", ob->oname);
    while ((cp=strchr(buf,' '))) *cp = '_';
    fprintf (f, "OBSERVER  %s\n", buf);

    sprintf (buf, "%.8s", ob->fname);
    while ((cp=strchr(buf,' '))) *cp = '_';
    fprintf (f, "FILENAME  %s\n", buf);

    if (ob->comments != NULL) {
	fprintf (f, "#\n");
	fprintf (f, "# -- Global comment list follows. --\n");
	C_loop (namp, ob->comments) {
//	    fprintf (f, "!->%s\n", namp->name);
// No longer add that '>' character each time we write...
	    fprintf (f, "!-%s\n", namp->name);
	}
	fprintf (f, "# -- End of comment list. --\n");
	fprintf (f, "#\n");
    }

    if (ob->title != NULL) fprintf (f, "TITLE   %s\n", ob->title);
    else                   fprintf (f, "#  NULL title \n");

/* Omit some things if Position Independent (test reflimit?) */
  if (gisap == 0) {
/* This requires some explanation.
  In general, if we have a Position Independent mask, the center position
  should not matter.  In the case of the GISMO aperture feature this is
  exactly true, and the mask fetures are defined in terms of input field
  locations.  However, on other masks such as a long slit, the mask features
  are being defined by reference to a "fake" sky position, which should be
  maintained here.  We can let the user specify a zero center if needed, or
  some other fake center when convenient.  This is why we cannot also skip
  the center position output when we just have reflimit == 0 case.
  However, we can, and do, skip other things such as disperser requirements,
  guide star output, wavelength limits, and the like for Position Independent
  mask cases.  This may be revised as needed.

  It is hereby acknowledged that ob->reflimit is set to zero whenever gisap
  is non-zero, so testing both is actually redundant, and a test on the
  ob->reflimit > 0 would be sufficient here; however it is done so that the
  test on ob->reflimit > 0 may be removed easily when needed, and only the
  gisap test will be done, since ob->reflimit MAY be cleared independently
  from the gisap case, and sometimes tested independently too.        */
//  if (gisap == 0 && ob->reflimit > 0) {
/* Make up the observation center... */
    strcpy (buf, sexig( ob->ra/Hour, 3, 3));
    fprintf (f, "CENTER   %s ", buf);
    fflush (f);		// As we are re-using buf here...
    strcpy (buf, sexig( ob->dec/Degree, 3, 2));
    fprintf (f, "%s\n", buf);

    fprintf (f, "EQUINOX  %.5f\n", ob->equinox);

/* Add some other things which we can find in obs... */
    fprintf (f, "POSITION %.5f\n", ob->angle/Degree);

/* Add hour angle for differential refraction */
#ifdef  DIFREF
    if (ob->dref) {	// differential refraction active
//    if (!(ob->ha < -24.0)) fprintf (f, "HANGLE %.3f\n", ob->ha);
	fprintf (f, "DREF  1\n");
	if (ob->ha == 0.0) fprintf (f, "HANGLE %4.4G\n", ob->ha);
	else  fprintf (f, "HANGLE %+4.4G\n", (ob->ha < -24.0) ? -25.0 : ob->ha);
    } else {		// No differential refraction active
	if (ob->ha != 0.0) 
	  fprintf (f, "HANGLE %+4.4G\n", (ob->ha < -24.0) ? -25.0 : ob->ha);
    }
#endif  // on DIFREF

  } else {  // on gisap true (or reflimit <= zero)
// clear position windows:  alwin, dewin, spawin, eqmenu
// set (ob->) ra, dec, angle  to zero
    ob->ra = ob->dec = 0.0;
//    HandleTextWindow (winlist, RAfield.ew);  // That's &alwin there...
//    HandleTextWindow (winlist, DECfield.ew);  // That's &dewin there...
    ob->angle = 0.0;
//    HandleTextWindow (winlist, slitpaf.ew);  // That's &spawin there...
  }  // on gisap

/* Some explanation:
  1) winlist is defined below, as a global in the file.  It could be defined
    above and used here, but that is not necessary.
  2) When a value such as ob->ra is changed at the binary value as above,
    the value presented in the window gets changed only when a call such as
    HandleTextWindow is done to output format the value for the window.
    This is the recommended thing for updating text window values from the
    program.  When updated by editing the text, the window handler takes 
    the text as input and updates the binary value.
  3) Elsewhere, when ra, dec, angle are changed or cleared to zero, the
    HandleTextWindow program is called to update the windows.
  4) It is not necessary to set ra, dec and angle here, and in fact is
    rather too late; they should be set already.        */

/* Initialize the saved position for global centering feature */
    init_gcenter ();

/* Write pass-through comments from rotaq queue, without reference
  to the ob->pcom which was read in.  Those originate here */
//  if (gisap == 0) {
/* Here, we also test for Position Independent mask case, since this is
  the real reason for it, the ability to skip guide star output.    */
  if (gisap == 0 && ob->reflimit > 0) {
// Debug print for flip queue
/* The rotaq queue should be empty, and only is used for those masks
  made for instruments which are not supported with rotator soft limits,
  but still require rotator warning support.  Temporarily, put these
  messages to debug output only.  For long term, put these in the file
  when the rotcq queue is empty and rotaq is not empty in place of the
  no rotator warnings message.    */
   if (ob->dlevel > 1) {
    if (rotaq == NULL) {
	fprintf (stderr, "#! No rotator flip above horizon.\n");
    } else  C_loop (al, rotaq) {
	fprintf (stderr, "#! !> Rotator flip at Hour Angle %s.\n", 
		sexig( al->v, 2, 0) );
    }
   } // test on dlevel
/* Write pass-through comments for rotator limits non-standard */
/* NOTE -- Only do this if instrument is I_imacs */
//    if (RotLim1 != RotLow && InstNdx(ob) == I_imacs)
    if (RotLim1 != DROT1)
	fprintf (f, "#! !> Rotator Low limit set to %.1f from %.1f\n",
		RotLim1, DROT1 );
//    if (RotLim2 != RotHigh && InstNdx(ob) == I_imacs)
    if (RotLim2 != DROT2)
	fprintf (f, "#! !> Rotator High limit set to %.1f from %.1f\n",
		RotLim2, DROT2 );
/* Same for actual warning queue */
    if (rotcq == NULL) {
      if (rotaq == NULL)
	fprintf (f, "#! No rotator warnings above horizon.\n");
/* For compatability, if we have no rotcq but we DO have rotaq, we output
  the old rotator flip messages at this point.  */
      else  C_loop (al, rotaq) {
	fprintf (f, "#! !> Rotator flip at Hour Angle %s.\n", 
		sexig( al->v, 2, 0) );
      }
    } else  C_loop (bl, rotcq) {
/* When we have an actual rotcq list of conflicts, write the pass-through
  comments for all the regions.  Test vs. semi-diurnal-arc to see if one
  end is "Horizon", but allow 12h as a non-horizon value.    */
char	*hbf=NULL, *hbt=NULL;
double	sdx;
	sdx = sda (ob->dec/Degree, pdx.lat) - 1.0/60.0;
	if (fabs(bl->a) < sdx || sdx > 11.9) hbf = dsexig (bl->a, 2, 0, NULL);
	else  hbf = strdup ("Horizon");
	if (fabs(bl->b) < sdx || sdx > 11.9) hbt = dsexig (bl->b, 2, 0, NULL);
	else  hbt = strdup ("Horizon");
	fprintf (f, "#! !> Rotator conflict at Hour Angles:  %s to %s.\n", 
		hbf, hbt);
//	    dsexig ( bl->a, 2, 0, &hbf), dsexig (bl->b, 2, 0, &hbt) );
	free (hbf);	free (hbt);
    }
/* Add similar for start of rotator warning bands.  */

//#ifdef  GSDATA
    if (ob->gp1.ep != 0.0) {
//	fprintf (f, "GS1 %9.6f %9.5f  %8.3f \n",
//		ob->gp1.ra, ob->gp1.dec, ob->gp1.ep);
	fprintf (f, "GS1 %9.6f %9.5f  %8.3f  %5lu\n",
		ob->gp1.ra, ob->gp1.dec, ob->gp1.ep, ob->gp1.cn);
    }
    if (ob->gp2.ep != 0.0) {
//	fprintf (f, "GS2 %9.6f %9.5f  %8.3f \n",
//		ob->gp2.ra, ob->gp2.dec, ob->gp2.ep);
	fprintf (f, "GS2 %9.6f %9.5f  %8.3f  %5lu\n",
		ob->gp2.ra, ob->gp2.dec, ob->gp2.ep, ob->gp2.cn);
    }
//#endif
  }  // on gisap or Position Independent

/* For IFU support, show the object position and IFU mode */
#ifdef  IFUM
    if (ob->IFUmode) {
	fprintf (f, "IFU  %d\n", ob->IFUmode);
	strcpy (buf, sexig( ob->IFUra/Hour, 3, 3));
	fprintf (f, "OFFCENTER %s ", buf);
	fflush (f);
	strcpy (buf, sexig( ob->IFUdec/Degree, 3, 2));
	fprintf (f, "%s\n", buf);
    }
#endif

/* The following items are not output in IFU mode. */
#ifdef  IFUM
    if (!ob->IFUmode) {
#endif
/* Only write temperature entry if it is not the default; this may
  include it being set to cutting temp for ldss case.   */
    if (ob->temp != DefTemp)
	fprintf (f, "TEMPERATURE  %.2f\n", ob->temp);
  if (gisap == 0 && ob->reflimit > 0) {
    fprintf (f, "WLIMIT  %.1f %.1f\n", ob->pb.blue, ob->pb.red);
/* Check if pb2 has been set and set pb2act if so... */
#ifdef  DWAV
//    if ( (ob->pb2.blue != ob->pb.blue) ||
//	 (ob->pb2.red  != ob->pb.red ) )  ob->pb2act = 1;
    ob->pb2act = (ob->pb2.blue != ob->pb.blue) ||
		 (ob->pb2.red  != ob->pb.red );
#endif
    if (ob->pb2act)
	fprintf (f, "DLIMIT %.1f %.1f\n", ob->pb2.blue, ob->pb2.red);
    fprintf (f, "Wavelength  %.2f\n", ob->cw);
  }  // on gisap

/* Parameterized decision priorities if not default */
    if (ob->Pmusthave != -2.0) fprintf (f, "MUSTHAVE %.2f\n", ob->Pmusthave);
    if (ob->Pdecide != 0.0)    fprintf (f, "PDECIDE  %.2f\n", ob->Pdecide);
#ifdef  IFUM
    }	// End non-IFU case test
#endif

/* Obtain optic elements names, if possible... */
    if (ob->Tel_scope == NULL) fprintf (f, "# NULL telescope\n");
    else	fprintf (f, "TELESCOPE  %s\n", ob->Tel_scope->name);
// Restoring the telescope name as read to output...

/* Put in any non-null front end element name */
    if (ob->Frontend != NULL) fprintf (f, "FRONTEND %s \n", ob->Frontend->name);
// Does not depend on GISMO defined

/* The instrument details are not present in IFU mode. */
#ifdef  IFUM
    if (!ob->IFUmode) {
#endif
    if (ob->Instr_mt == NULL) fprintf (f, "# NULL instrument\n");
    else fprintf (f, "INSTRUMENT %s\n", ob->Instr_mt->name);

/* The DBOUND and GAPS values */
    if (ob->Dstat) {
//int	k;
int	db;
int	ix;
int	n;
	n = (ob->Dstat & DS_ns) ? 2 : 1;
//	if (ob->Dstat & DS_gaps)   fprintf (f, "GAPS  %d\n", n+1);
//	if (ob->Dstat & DS_orient) fprintf (f, "DBOUND  %d\n", n+1);
//	if (ob->Dstat & (DS_ns | DS_gaps | DS_orient))
/* Omit testing the DS_ns bit which indicates only the orientation and
  not a request for gap support now.    */
	db = ob->Dstat & (DS_gaps | DS_orient);
	ix = InstIndx(ObInstN(ob));
#ifdef  NSROT
// Include the NS bit unless instrument is IMACS short camera, or ldss-3
// which have on/off only; see fill_gapmenu at ~4153
	if (ob->dlevel > 1) {
	    printf (" Inst = %d (IMACS = %d); name = %s\n",
	        ix, I_imacs, ObInstN(ob));
	}			// debug only section
//	if (InstIndx(ObInstN(ob)) != I_imacs || 
//	  strncasecmp(ObInstN(ob), "IMACS_s", 7) )  // Redundant!
	if (ix == I_lds3 || !strncasecmp(ObInstN(ob), "IMACS_s", 7) )
		db = ob->Dstat & DS_gaps;	// IMACS short camera case
	else
		db = ob->Dstat & (DS_ns | DS_gaps | DS_orient);
// NOTE that we must assure that detector orientation is defaulted to
// the proper one for all cameras, not just IMACS
// Also NOTE that ldss3 has Gaps menu disabled, so the test above is moot.
#else
	if (ix == I_lds3)
		db = ob->Dstat & DS_gaps;	// on/off only case
	else
		db = ob->Dstat & (DS_ns | DS_gaps | DS_orient);
#endif
	if (db) fprintf (f, "GAPS  %d\n", n);
	if (ob->dlevel > 1) printf (" Dstat = %d, db = %d, GAPS %d\n",
		ob->Dstat, db, db ? n : 0);		// Debug only
/* A value of DS_ns with no other bits set is not normally read in,
  unless the MOE disperser is present with a GAPS value and has the
  gaps bit cleared in normalize.  In any case, we write only non-zero
  values for the gaps and orient bits, with the N&S orientation
  bit value added.  If the value would be zero, it is not written.  */
    }

  if (gisap == 0 && ob->reflimit > 0) {
    if (ob->Disperser == NULL) fprintf (f, "# NULL disperser\n");
    else fprintf (f, "DISPERSER  %s\n", ob->Disperser->name);

/* Write order if it exceeds the normal default */
    if (ob->order > 1) fprintf (f, "ORDER %d\n", ob->order);

/* Nod & Shuffle rotation bias value if present and allowable  */
#ifdef  NSROT
    if (fabs(ob->nrbias) > NSrlim &&		// has bias value
/* The N&S bias feature applies only to IMACS short camera observations,
  for which the grism is rotated by 90 degrees to accomodate N&S
  detector orientation requirements without rotating the detector.  */
//	InstIndx(ob->Instr_mt->name) == I_imacs &&	// is IMACS
	InstIndx(ObInstN(ob)) == I_imacs &&	// is IMACS
		ob->Disperser->kw == GRIsm ) {		// uses grism
//	(a->Disperser->kw == GRIsm || a->Disperser->kw == EGRIsm ) ) {
// The test for EGRIsm might be omitted?
// May also want to add uncut lengths not zero present?
	fprintf (f, "NSROT  %5.1f\n", ob->nrbias / Degree);
    }
#endif
  }  // on gisap

/* Slit and hole sizes as available */
//    fprintf (f, "SLITSIZE  %.3f %.3f %.3f %.3f\n",
//	ob->default_slit.width,
//	ob->default_slit.alen,
//	ob->default_slit.blen,
//	ob->default_slit.angle / Degree  );

  if (gisap == 0) {
    fprintf (f, "SLITSIZE  %.3f",
	ob->default_slit.width);

    if (ob->default_slit.shape != RECTANGLE)
	fprintf (f, " %d", ob->default_slit.shape);

/* Also set default alen, blen if needed for default as for reference. */
    if (ob->default_slit.shape == SQUARE ||
	ob->default_slit.shape == CIRCLE )
	ob->default_slit.alen = ob->default_slit.blen =
	ob->default_slit.width / 2.0;

    fprintf (f, " %.3f %.3f %.3f\n",
	ob->default_slit.alen,
	ob->default_slit.blen,
	ob->default_slit.angle / Degree  );

/* If shape is square or circle, set alen and blen to width/2; this will
  set the proper spectrum size... */
    if (ob->reference_slit.shape == SQUARE ||
	ob->reference_slit.shape == CIRCLE )
	ob->reference_slit.alen = ob->reference_slit.blen =
	ob->reference_slit.width / 2.0;
/* The above is copied from the place in mgutils.c where this data will
  be read.  Doing it here makes the file more consistent for those users
  who might otherwise find fault with the alen and blen values.  */
  }  // on gisap

    fprintf (f, "REFHOLE   %.3f %d %.3f %.3f %.3f\n",
	ob->reference_slit.width,
	ob->reference_slit.shape,
	ob->reference_slit.alen,
	ob->reference_slit.blen,
	ob->reference_slit.angle / Degree  );

    if (ob->slext) fprintf (f, "SLEXTEND  1\n");
/* Could do as ex_order is done below instead... */

/* Overlap and do-not-cut values */
    if (ob->minsep != D_minsep)	// that is as in mgutils...
	fprintf (f, "OVERLAP    %.2f\n", -(ob->minsep) );
    if (ob->dca != 0.0)
	fprintf (f, "UNCUTLEFT  %.3f\n", ob->dca);
    if (ob->dcb != 0.0)
	fprintf (f, "UNCUTRIGHT %.3f\n", ob->dcb);

  if (gisap == 0 && ob->reflimit > 0) {
/* Output the ex_order flag */
    fprintf (f, "EXORDER  %d\n", ob->ex_order );

/* Output the repeat and select parameters */
    if (ob->reuseob < 99)
	fprintf (f, "REPOBJ  %d\n", ob->reuseob);
    if (ob->reuserf < 99)
	fprintf (f, "REPREF  %d\n", ob->reuserf);
  }  // on gisap

    if (ob->refsel > 1)
	fprintf (f, "REFSEL  %d\n", ob->refsel);
    if (ob->reflimit < 99)
	fprintf (f, "REFLIMIT %d\n", ob->reflimit);

/* Output extended priority line if non-zero */
    if (ob->apri != 0.0)
	fprintf (f, "EXPRI  %.3f\n", ob->apri);

/* Output date value if present */
    if (mjd > 0) fprintf (f, "DATE %5d\n", mjd);

/* Output all the object files */
    if (ob->objfiles != NULL)
	fprintf (f, "#  Object file list\n");
    C_loop (namp, ob->objfiles) {
	fprintf (f, "OBJFILE  %s\n", namp->name);
    }

/* Output some obsolete things for debug */
#ifdef  OBSOUT
    fprintf (f, "#  Following items are obsolete use\n");
// // (see above)	fprintf (f, "INSTNAME  %s\n", ob->instrument);
    if (ob->instrument != NULL) 
	 fprintf (f, "INSTNAME %s\n", ob->instrument);
    else fprintf (f, "#  NO instrument name given.\n");
  if (gisap == 0 && ob->reflimit > 0) {
    if (ob->grating != NULL) fprintf (f, "# GRATING %s\n",
	ob->grating->name);
    if (ob->filter != NULL) fprintf (f, "FILTER %s\n",
	ob->filter->name);
  }  // on gisap
#endif

#ifdef  IFUM
    }	// End non-IFU case test
#endif

#ifdef  FILESEC
/* Add checksum and signature line to file */
    signfile (f, '!', dodebug-1);
#endif

/* End of process; close the file. */
    iclose (f);
    return True;
}

	/* - - - - - - - - - - *\
	|			|
	|   Rotator Geometry	|
	|			|
	\* - - - - - - - - - - */

/*  === Rotator geometry routines needed  ===  */

// #define  ROFF  (-47.0)
// Rotator offset (a.k.a. CHUECO)
// Replace ROFF with "RotOff" -- a global variable

// Compute zd and parallactic at one step
void	pazd (double* pa, double* zd, double h, double d, double g)
{
double  st, ct;
double  sp, cp;
double  sd, cd;
double  x, y, z;

// look up the trig functions
    sincos (h*Hour, &st, &ct);		// Hour Angle
    sincos (d*Degree, &sd, &cd);	// Declination
    sincos (g*Degree, &sp, &cp);	// Latitude

// Find vector position of zenith relative to star
    x = cp*st;			// East
    y = sp*cd - cp*sd*ct;	// Toward North Pole
    z = sp*sd + cp*cd*ct;	// Toward star
    *zd = atan2(sqrt(x*x + y*y), z) / Degree;
    *pa = atan2(x, y) / Degree;
}

double  gsda (double a, double d, double g)
// Find, return semi-dirurnal arc in hours
{
double  sp, cp;
double  sd, cd;
double	sal;
double	csd;

// look up the trig functions
    sincos (d*Degree, &sd, &cd);	// Declination
    sincos (g*Degree, &sp, &cp);	// Latitude
    sal = sin (a*Degree);		// sine of altitude...
    if (cd < 0.001) cd = 0.001;
    csd = (sal - sd*sp)/(cd*cp);
    if (csd < -1.0) csd = -1.0;
    if (csd >  1.0) csd =  1.0;
    return  acos(csd)/Hour;
}

double  sda (double d, double g)
// Find, return semi-dirurnal arc in hours
{
// Use the general program above.
    return  gsda (0.0, d, g);
}

// For GUIDATE support:
#ifdef  GUIDATE

// SHOULD MODIFY THIS TO COMPUTE RA,DEC IN HOURS AND DEGREES,
// FROM THE VECTOR POSITION RATHER THAN RETURNING VECTOR.

vect3	smpos (double t, int b)
/* Return equitorial vector for body b (sun=0, moon=1) at the
  time t, in days from standard 2000.0 epoch.  */
{
vect3	r={0.0, 0.0, 0.0};
vect3	v;
double	sml;	// mean longitude of sun
double	ob;
double	g;
double	stl, sdl;
double	mlon;	// not using , mlst;
double	mlat;
double	tt;

/* Position computation from ~/apps/position/sst.c where they are
  taken from an approximation to low accuracy in the AENA.  */

    sml = fmod (280.460 + 0.9856474*t, 360.0) * Degree;	// mean solar longitude
    if (sml < 0.0) sml += 360.0;

    ob = (23.439 - 4.0e-7*t)*Degree;

//    if (strcmp (body, "sun") == 0)  {                   /*  FOR SUN */

    if (b == 0) {	// sun

/* Compute solar true longitude */
        g = (357.528 + 0.9856003*t)*Degree;
        sdl = (1.915*sin(g) + 0.020*sin(2.0*g))*Degree;
        stl = sml + sdl;		// True solar longitude radians

/* Compute solar position vector */
	v = sph2vec (stl, 0.0);
//        vmake (v, cos(stl/dir), sin(stl/dir), 0.0);
//        ob = (23.439 - 4.0e-7*t)*Degree;
//        vrot (v, ob/dir, 1);
	r = rotate (v, ob, 1);

// /* Rotate the solar position into horizon coordinates */
//         vrot (v, (-lst)/dir, 3);
//         sphere (&a, &z, v);
//         vrot (v, (latitude-90.0)/dir, 2);       /* Geodetic */

// /* Extract the current altitude and azimuth */
//         sphere (&a, &z, v);

// /* semidiameter is 961.18 arcsec/distance.  */
//         *s = dosds ? 0.267 * (1.0 + 0.0167*cos((g+sdl)/dir)) : 0.0;


//     } else if (strcmp (body, "moon") == 0)  {           /* FOR MOON */
    } else if (b == 1) {	// moon

/* Find moon position, if specified */
        tt = t / 36525.0;
        mlon = 218.32 + 481267.883*tt
                + 6.29*sin((134.9+477198.85*tt)*Degree)
                - 1.27*sin((259.2-413335.38*tt)*Degree)
                + 0.66*sin((235.7+890534.23*tt)*Degree)
                + 0.21*sin((269.9+954397.70*tt)*Degree)
                - 0.19*sin((357.5+ 35999.05*tt)*Degree)
                - 0.11*sin((186.6+966404.05*tt)*Degree);
        mlat =
                  5.13*sin(( 93.3+483202.03*tt)*Degree)
                + 0.28*sin((228.2+960400.87*tt)*Degree)
                - 0.28*sin((318.3+  6003.18*tt)*Degree)
                - 0.17*sin((217.6-407332.20*tt)*Degree);
//        mpi = 0.9508
//                + 0.0518*cos((134.9+477198.85*tt)*Degree)
//                + 0.0095*cos((259.2-413335.38*tt)*Degree)
//                + 0.0078*cos((235.7+890534.23*tt)*Degree)
//                + 0.0028*cos((269.9+954397.70*tt)*Degree);
//        mr = 1.0 / sin(mpi*Degree);
 
/* Make moon position vector in equitorial coordinates */
//        vmake (v, mr*cos(mlat/dir), 0.0, mr*sin(mlat/dir));
	v = sph2vec (mlon*Degree, mlat*Degree);
//        vrot (v, mlon/dir, 3);
//        ob = 23.439 - 4.0e-7*t;
//        vrot (v, ob/dir, 1);
	r = rotate (v, ob, 1);
    }

    return  r;
}
#endif  // on GUIDATE

//#define  DBUGMOON
#undef  DBUGMOON

#ifdef  GUIDATE
void	solpos (int mj)
/* Find true solar position at mj, setting ra and dec values in
  the global values section.  Also uses epoch of coordinates. */
// Call this when changing mjd or epoch
{
double	sml;
vect3	vsun;
#ifdef  MOONLINE
double	t;
vect3	vmoon;
double	ram, dem;
int	i;
#endif

// Set up the field position, by precession to date;
// initially set equal to given position as a default.

    if (Gob != NULL) {
	rafc = Gob->obs->ra / Hour;
	defc = Gob->obs->dec / Degree;
	eqfc = Gob->obs->equinox;
    }

//    rafc = *(double*)RAfield.data / Hour;
//    defc = *(double*)DECfield.data / Degree;


// If the mj is not valid current date, i.e. past 1 Jan 2003,
// we reject it, and establish null position.

// printf (" - solpos mjd = %d\n", mj);	// debug

// mjd for 2003-1-1 is 52640
// JD at lco at 2003-1-2 at 4:40:24 Z = 2452641.69472
// Really, that should end in 69673 for proper longitude.
    if (mj < 52640) {
	rats = -25.0;
	dets = 0.0;
	return;
    }


// Establish the local mean midnight reference time
    lmm = (double)(mj - 51545) + 1.69673;	// Time in days from 2000.0
// The 1.6942 is for midnight at end of day at LCO,
// The 51545 is standard JD epoch, 2000-Jan-0, 12h. (J.D. 2451545.0)
// NOTE that the above includes longitude of LCO as .19673 days west.

// Find solar position for lmm

// These formulae are from ~/apps/position/sst.c
    sml = (280.460 + 0.9856474*lmm)*Degree;	// Mean solar longitude
// We actually will need that sml below...

/* -- comment out this code
    g = (357.528 + 0.9856003*lmm)*Degree;	// Mean anomaly
    sdl = (1.915*sin(g) + 0.020*sin(2.0*g))*Degree;// Differential longitude
    stl = sml + sdl;				// True longitude
    ob = (23.439 - 4.0e-7*lmm)*Degree;		// Obliquity of date
// Above angles are in radians

// Find the declination and ra from the longitude and obliquity...
    dets = asin ( sin(stl) * sin(ob) ) / Degree;	// Declination
// dets is in degrees

// Find the ra of the true sun at lmm
    rats = atan2 (cos(ob)*sin(stl), cos(stl)) / Hour;	// R.A. of sun
    if (rats < 0.0) rats += 24.0;
//printf (" Old sun = %.5f, %.4f\n", rats, dets);	// debug
-- end commented code here.  */

// Note, we haven't done precession to window epoch yet
/* Do not do precession above.  Instead, precess the field center from
  the window epoch to date to get sidereal time of transit of center
  in terms of local mean sidereal time. */

/* Precess the field position from field epoch to epoch of given date */

//    pprec (&rafc, &defc, 2000.0 + (lmm / 365.25),
//	*(double*)RAfield.data, *(double*)DECfield.data, wepoch);

    if (Gob != NULL) {
//	pprec (&rafc, &defc, 2000.0 + (lmm / 365.25),
//	    Gob->obs->ra, Gob->obs->dec, Gob->obs->equinox);
	vfc = prec (2000.0 + (lmm / 365.25),
	    sph2vec(Gob->obs->ra, Gob->obs->dec),
	    Gob->obs->equinox );
	vec2sph (&rafc, &defc, vfc);
	rafc /= Hour;
	defc /= Degree;
	eqfc = Gob->obs->equinox;  // may be redundant
    }


// Now, the real position (compare to old)
    vsun = smpos (lmm, 0);
    vec2had (&rats, &dets, vsun);
//    rats /= Hour;
//    dets /= Degree;
//printf (" New sun = %.5f, %.4f\n", rats, dets);	// debug

/* Find mean sidereal time at local midnight, from sml */
    stlm = fmod (12.0 + (sml/Hour), 24.0);

/* Using rats, find hour angle of anti-sun at lmm, and convert that
  solar interval to sidereal interval to find stm itself... */
    stm = fdifmod (rats+12.0, stlm, 24.0) * Sidereal + stlm;
#ifdef  DBUGMOON
printf (" STlm = %7.4f, dstm = %7.4f", stlm, stm-stlm);	// debug
//printf (" rats = %8.5f, dets = %8.4f\n Lat = %8.4f", rats, dets, pdx.lat);
#endif

/* Using complement sda, find sidereal interval to twilight */
    twit = (12.0 - gsda (-18.0, dets, pdx.lat)) * Sidereal;
#ifdef  DBUGMOON
printf (", dTwi = %7.4f", twit);		// debug
#endif

/* Find the srt and sst times using complement sda also */
    srt = (12.0 - gsda ( 0.0, dets, pdx.lat)) * Sidereal;
    sst = (gsda (-1.0, dets, pdx.lat) - 12.0) * Sidereal;
#ifdef  DBUGMOON
printf (" dSs = %7.4f, dSr = %6.4f", sst, srt);	// debug
printf ("\n");			// debug
#endif

#ifdef  MOONLINE
// #define  Lunar  1.036601
/* Lunar is the mean lunar to sidereal rate.  The sidereal month is
  about 27.321661 days, use 1 + 1/27.3 here... */

/* Find nearest lunar transit by approximating lunar ra at lmm and
  finding the hour angle of moon at lmm.  We may choose one day later
  when moon is near sun, but that's not a problem... */
/* Iterate this approximation about 3 times... */
    for (i=0,ltt=0.0; i<3; i++) {
	t = lmm + (ltt/Sidereal)/24.0;
	vmoon = smpos (t, 1);
	vec2had (&ram, &dem, vmoon);
// We find the position of the moon at ltt hours sidereal after stlm;
// Then that moon transits ltt hours sidereal after stlm.
	ltt = fdifmod (ram, stlm, 24.0);
    }
#ifdef  DBUGMOON
printf (" Ltt = %7.4f", ltt);		// debug
#endif

/* Given ltt, we find pci by vector positions of moon and sun at
  that instant.  We already have vmoon then... */
    pci = 50.0 * (1.0 - dot3vect (smpos(t,0), vmoon));
#ifdef  DBUGMOON
printf (" Pci = %02d", pci);		// debug
#endif
/* Also find angular distance to moon. */
//    adm = acosd (dot3vect (vfc, vmoon));
/* Could use acos(--)/Degree if acosd is not welcome in linux */
    adm = (int) (acos (dot3vect (vfc, vmoon)) / Degree);

/* Knowing the ltt and hence the mean time of transit, find the
  moonrise and moonset times by obtaining the sda of the lunar position
  at a time near rise/set and iterating on that too... */
    for (i=0,mrt=-6.0; i<3; i++) {
	vmoon = smpos (lmm + ((ltt+mrt)/Sidereal)/24.0, 1);
	vec2had (&ram, &dem, vmoon);
// We find position of moon at stm+ltt+mrt, and at that time of
// moonrise, ram - sda should equal the sidereal time...
//	mrt = fdifmod (ram, ltt+stlm, 24.0) - gsda (0.0, dem, pdx.lat);
	mrt = fdifmod (ram, ltt+stlm, 24.0) - gsda (1.0, dem, pdx.lat);
    }
#ifdef  DBUGMOON
printf (" dLr = %7.4f", mrt);		// debug
#endif

/* Moonrise and moonset times are taken for an altitude of 1 degree
  above the actual horizon.  The moon positions are geocentric, and 
  the average geocentric parallax is about one degree at rise and set. */

    for (i=0,mst=6.0; i<3; i++) {
	vmoon = smpos (lmm + ((ltt+mst)/Sidereal)/24.0, 1);
	vec2had (&ram, &dem, vmoon);
// We find position of moon at stm+ltt+mst, and at that time of
// moonset, ram + sda should equal the sidereal time...
//	mst = fdifmod (ram, ltt+stlm, 24.0) + gsda (-1.0, dem, pdx.lat);
	mst = fdifmod (ram, ltt+stlm, 24.0) + gsda (0.0, dem, pdx.lat);
    }
#ifdef  DBUGMOON
printf (" dLs = %7.4f", mst);		// debug

/* Note that we add 1 degree to the moon altitude at rise and set to
  compensate for the approximate geocentric parallax at rise/set */
printf ("\n");			// debug
#endif

#endif  // on MOONLINE

}
#endif  // on GUIDATE

double	rotang (double* Imx, double h, double d, double g, double pa, int s)
// Find and return the rotation angle, and IMACS angle
{
double	sp;		// side parameter
double	z;		// zeinth distance
double	p;		// parallactic angle
//double	zvp;		// zenith vector position
//double	nvp;		// north vector position
//double	imx;		// imax angle position
double	rrv;		// result

    sp = ( s == 0 ) ? 1.0 : -1.0;
// For imacs, s=1 and sp = -1.0
    pazd (&p, &z, h, d, g);
//    zvp = sp * (z - 90.0);		// "up" left of "vertical"
//    nvp = zvp + p;			// North left of vertical
//    imx = nvp - pa;			// Imacs angle
//    rrv = RotOff - imx + z*sp;		// Rotation angle
// Substitute imx = nvp - pa and get:
//    rrv = RotOff + pa - nvp + z*sp;
// Substitute nvp = zvp + p and get:
//    rrv = Rotoff + pa - p - zvp + z*sp;
// Substitute zvp = sp*z - sp*90 and get:
//    rrv = Rotoff + pa - p + 90*sp - z*sp + z*sp;
//    while (rrv < 0.0) rrv += 360.0;
//    while (rrv > 360.0) rrv -= 360.0;
// The above may be simplified algebraically to this:
    rrv = 90.0*sp - p + RotOff + pa;
/* Please NOTE that since rotation angle is measured in telescope space,
  it does not depend on the zenith distance.  However, the IMACS angle
  is measured relative to the platform, so it DOES depend on z.    */
    rrv = drange (rrv, 360.0, 180.0);	// Actual rotation angle
    *Imx = -(rrv - RotOff - sp*z);	// Actual IMACS angle, positive is CCW
/* When on side 1 (west) and telescope at zenith, imacs angle is -46 with
  RotOff = -46 and rotator at 0.  With a rotator range of -180 to +180, the
  imacs angle has a range of -46 -z +/-180 = -46 -0 +180 = +134 to
  -46 -90 -180 = -316.  With rotator range -151 to +176, the imacs angle
  range is more like -46 - -151 = +105 to -46 -90 -176 = -312.  If we stop
  a little above the horizon, the -312 is a little more positive, like -300.
  Since IMACS range has been reported as about -310 to +120, we actually
  are running into the limit near zenith, reducing that +120 to +105; the
  other limit is not actually limiting the rotation.  Thus the rotation
  limit for IMACS angle is +105 to +132 which it would have reached if
  it could move with the rotator to the rotator limit of -178.    */
    return  rrv;
}

#define  TDEL  0.0005
//#define  TDEL  0.05
/* TDEL is in hours, and 0.05 is 3 minutes, 0.0005 is 1.8 seconds. */

static	double	rotrate (double* x, double h, double d, double g, double pa, int s)
// Call rotang twice to find rates for rotangle and imacs angle
{
double	a1, a2;
double	da;
double	r;
double	x1, x2;
    a1 = rotang (&x1, h - TDEL, d, g, pa, s);
    a2 = rotang (&x2, h + TDEL, d, g, pa, s);
    da = fdifmod (a2, a1, 360.0);
    r = 0.5 * da / TDEL;
//    da = fdifmod (x2, x1, 360.0);
    da = x2 - x1;
    *x = 0.5 * da / TDEL;
    return  r;
}

static	vect2	dfunc (double h, pdata* p)
/* Return a vector giving x=descriminator function and y = rate, for
  the location described in p */
{
vect2	res;
double	v, r;
double	x, y;
//double	q;
//double	w, z;
    v = rotang  (&x, h, p->dec, p->lat, p->pa, p->side & 1);
    r = rotrate (&y, h, p->dec, p->lat, p->pa, p->side & 1);
// x is the imacs angle, and y its rate
// Alternately, we could compute imacs angle and its rate.
    if (p->side & 2) {	// use imacs angle (x, y)
//	x = fdifmod (x, p->target, 360.0);
	x = x - p->target;
/* If fabs(y*TDEL) > 90.0 degrees, we are at a rotator flip and
  x is either small (under 180) or large, and if small it is right,
  but if large it needs to be replaced by a large value of the oposite sign
  to the small equivalent value, indicating a root.  Maybe set y to zero.  */
/* Even if y is not large, x could have changed a lot and have fabs(x) > 180,
  in which case we need to find the proper sign of x.  Like this... 
	if (fabs(x) > 270.0) {  // flip detected
	    if (x > 0.0) {
		q = x - 360.0;
		if (q > 0.0) x = q - 360.0;
		else  x = q;
	    } else if (x < 0.0) {
		q = x + 360.0;
		if (q < 0.0) x = q + 360.0;
		else x = q;
	    }
	}
... */
// Actually, the above should be done in search which calls this...
// end of added section
	if (fabs(y*TDEL) > 90.0) y = 0.0;
	r = y;
    } else {		// use rotator angle (v, r)
	x = fdifmod (v, p->target, 360.0);
	y = r;
    }
//    x = fdifmod (v, p->target, 360.0);
// For a single point, return the x value...
    if (p->width == 0.0) {
	res.x = x;
	res.y = y;	// was = r;
	return res;
    }
// For band, we need to find when within width...
// x = v - target; if x > 0, then v > target and d = width - x
// otherwise, with v < target, x < 0, d = width + x
// so d = width - abs (v-target), and d > 0 when in the interval.
    if (x > 0.0) {
	x = -x;
	y = -y;		// was r = -r;
    }
    res.x = x + p->width;
    res.y = y;		// was = r;
    return  res;
}


/*  =====  Set-up defines for binary search system  =====  */

#define  BINSEARCH
//#undef  BINSEARCH
/*  Symbol  BINSEARCH  is defined to implement the new binary search
  for the next root in an interval.  This replaces the older search
  program (below) with a more general method.  See "rootfind.otl" for
  a description of how it is done.    */

#undef  DBSEARCH
//#define  DBSEARCH

#undef  DBQWARN
//#define  DBQWARN

/* Use the  DBSEARCH  symbol for detailed debug on the search algorithm
  and use the  DBQWARN  symbol to give debug prints on the qwarn calls
  and the queues of flips and warning segments.  All those queues are
  assumed to be in order when imacs angles are searched.    */

/* Using DBQWARN gives a nice display of the warning bands, and also
  shows the imacs angle warnings in orange.  */

#undef  ALLOWFAKESEARCH
//#define  ALLOWFAKESEARCH
/*  Use symbol  ALLOWFAKESEARCH  to allow use of a fake search program
  so the calling programs can use the old method of doing search rather
  than actually using the binary search directly.  This is intended for
  debug operations only, and should not be used in production.    */

//#define  USEXQWARN
#undef  USEXQWARN
/*  Symbol  USEXQWARN  causes the xqwarn subroutine to be compiled and
  used for finding warnings of rotator limits.  It may not be needed, or
  it may actually cause wrong results.  Appears to be not needed.    */


#define  Close  0.01
/*  Symbol  Close  is the number of hours of hour angle within which
  two warning elements will be regarded as being the same.    */

//#undef  LIMITPREC
#define  LIMITPREC  0.5
/*  Use symbol  LIMITPREC  to add code to limit precision value based
  on the width and derivatives.  This might not be needed.  The
  precision is set at least as small as the x distance implied by a
  function value of LIMITPREC and the largest slope at an interval end.
  A value no larger than 0.5 degrees should be used, and lower values
  might be useful in some cases.  Smaller can cause slow execution.    */

#define  Precision  0.01
/*  Symbol  Precision  is used in calls to gbsearch for the precision
  value and hence determines the execution time needed...
  Experimentally on some problem locations shows that higher values
  can cause roots to be missed, and lower values will cause much longer
  execution times.  With LIMITPREC above working, most any value can be
  used for precision.  Both it and required split may be defined in the
  program or together in the future.
  There is some logic in making Precision similar to Close      */

/*  ====  Binary search should be set up now.  ====  */


#ifdef  BINSEARCH

pdata	gpd;  // global structure for easy reference in subroutines

static	double	signof (double x)
/* Returns 1.0 if x > 0, -1.0 if x < 0 and 0.0 if x == 0.  */
{
    if      (x > 0.0) return  1.0;
    else if (x < 0.0) return -1.0;
    else              return  0.0;
}

static	double	xrecip (double x)
/* Returns corrected reciprocal.  If |x| > 1.0 this is 1.0/x,
  and otherwise is signof (x).  If x is 0.0, return 2.0.  */
{
    if (fabs(x) > 1.0) return  1.0/x;
    if (x == 0.0) return  2.0;
    return  signof (x);
}

static	vect3	bsfunc (double x)
/*  Binary Search interface function.  Finds function value y and
  derivative z at point x, and packages them in a 3 vector.  */
{
vect3	r;
vect2	a;
    a = dfunc (x, &gpd);
    r.x = x;
    r.y = a.x;
    r.z = a.y;
    return  r;
}

// Use dmax instead (it is in utils.o which is linked)
//static	double	amax (double a, double b) {
//    if (a > b) return  a;
//    return  b;
//}

static	vect3  gbsearch (vect3 left, vect3 right, double prec, double req)
/*  General, binary, recursive search for the first root in the interval
  from left to right.  The return value of a root will be within prec of the 
  true root, and to the right of it, and left of the right limit.  When
  no root exists, the right value is returned.  This program is recursive,
  and may call itself multiple times.  Any interval longer than req must be
  split and shorter intervals examined.  The right value may be greater
  (usually) or less than the left value.  Uses bsfunc to find values for
  the function inside the interval.    */
{
vect3	mid;
vect3	res;
vect3	a, b;
double	p, r;
double	d, dir, dp;
double	ad;
double	cr, cl, cm;
#ifdef  LIMITPREC
double	cs;
#endif
int	ri;

// Some housekeeping and trivial stops
    d = right.x - left.x;
    if (d == 0.0) return  right;	// Stop infinite loop
    dir = signof (d);
    ad = fabs (d);
    r = fabs (req);
    if (r <= 0.0) r =  ad/20.0;
    p = fabs (prec);
    if (p <= 0.0) p = ad * 1.0e-4;
    dp = p * dir;

#ifdef  LIMITPREC
/* Enforce an upper limit on p when required by the largest of the
  derivatives, so as not to miss a root.    */
    cs = dmax (fabs(left.z), fabs(right.z));	// highest slope
//    cs = LIMITPREC * fabs( dmax(gpd.width, 1.0) * xrecip (cs));
    cs = LIMITPREC * fabs(xrecip (cs));
// Note that width is zero when looking for point, not range.
    if (cs < p) { p = cs;  dp = p*dir; }
#endif  // on LIMITPREC

/* See if left side is an exact root, and if so, return a
  point p/2 past it.  */
    if (left.y == 0.0) return bsfunc (left.x + 0.5*dp);

/* See if a root is in the interval; it is in if the signs of the
  function values at the ends differ, and the derivatives have same signs  */
    ri = signof(left.y) != signof(right.y) && signof(left.z) == signof(right.z);

// Quickly leave if width is under p and no root
    if (ad < p && !ri) return  right;

/* In case y difference between left and right is over 180 degrees, correct
  the largest by 360 degrees (if not IMACS angle) */
    if ((gpd.side & 2) == 0) {  // not IMACS angle solution
	if (fabs(right.y - left.y) > 180.0) {
	    cl = signof (right.y - left.y);
	    if (fabs(left.y) > fabs(right.y)) {
		left.y += cl*360.0;
	    } else {
		right.y -= cl*360.0;
	    }
	}
    }

/* At least one root is in the interval if the function values change sign,
  but not if the sign change is due to modulo arithmetic.  A root may be
  in the interval, and should be searched for, if the function values and
  derivative indicate that a root is right of the left end and left of the 
  right end.  Use this criterion here.  At a root, we have r, 0, z; while
  at a near point we have x, y, z.  Then z = dy/dx = (y-0)/(x-r).  To have
  a root at least as close as d from x, substitute d = r-x, and obtain the
  critical slope zc = -y/d.  Slope more than zc in absolute value and a
  sign of y opposite to that of the slope will result in intercept at d
  or closer to x, the desired condition.  If y is negative, take
  z - (-y/d) and if positive there is a root; if y is positive, take
  z - (-y/d) and if negative there is a root, so if signof (y) is same
  as - signof (z - (-y/d)), or -signof(z+y/d), there may be a root.
  For the right side, use -d for the equivalent operation.
  Actually, use y/2d to allow more intervals to be tested...
  Here, set ri to a root expected in interval:    */
//    ri = ((signof( -left.y) == signof( left.z+ 0.5*left.y/d) ) ||  // was &&
//	  (signof(-right.y) == signof(right.z+0.5*right.y/d) ))  ||
//	  (left.y == 0.0)  ||  (right.y == 0.0) ;
/* It appears that the above computation does not find whether a root is
  inside the interval with any reliability, in fact it will miss many.  */
/* We note that when a flip occurs, the z value is often forced to zero. */

// Find mid point
    mid = bsfunc (0.5*(left.x + right.x));
/* Do not (yet) correct mid.y for modulo 360 as done with left, right
  above, as it may not be necessary.  */

#ifdef  LIMITPREC
/* Enforce an upper limit on p when required by the largest of the
  derivatives, so as not to miss a root.    */
    cs = LIMITPREC * fabs(xrecip (mid.z));
// Note that width is zero when looking for point, not range.
    if (cs < p) { p = cs;  dp = p*dir; }
#endif  // on LIMITPREC

/* See if we happen to have an exact root, and if so, recurse to the left
  of this point, and if nothing found there, return a point a little to the
  right of this found point.  NOTE that there will be a problem if derivative
  is also zero.    */
    if (mid.y == 0.0) {  // Stumbled into a root
	a = bsfunc (mid.x - 0.5*dir*dmin(p, 0.5*ad));
	b = gbsearch (left, a, p, r);
	if (b.x == a.x) {
	    res = bsfunc (mid.x+0.5*dp);
// DEBUG only
//printf ("x-x gbsearch returning mid root %.2f %.2f %.2f\n", res.x, res.y, res.z);
	    return  res;
	}
    }

//#define  XTRABINSRCH
#undef  XTRABINSRCH
/* Using symbol  XTRABINSRCH  to implement a quicker binary search
  when the interval is under p to find a better root.  This extra
  precision is not necessary, only finding the root is necessary.  */
/* NOTE that at dec -28, spa 328 the XTRABINSRCH true case produces
  3 bogus roots for flip angles.  For this reason, and the fact that
  it is probably a needless complication, it is negated.    */

// Special treatments for large and small intervals
    if (ad > r)  goto final;
    else if (ad < p) {  // if not ri, we returned above
#ifdef  XTRABINSRCH
int	n;
double	x, rx;
// if right is a root, return 0.25*dp past it 
// DEBUG only
//if (right.y == 0.0) printf ("x-x gbsearch returns right+ %.2f\n", right.x);
	if (right.y == 0.0) return bsfunc (right.x + 0.25*dp);
// Drop any condition in which the left-right function values change by
// more than the average derivative would allow to avoid mod 360 problems
//	if (fabs(right.y-left.y) > (90.0 + ad*fabs(right.z+left.z)) ) {
//DEBUG MESSAGE ONLY
//		printf ("x-x gbsearch bump 1 at %.2f <%.2f> %.2f %.2f (%.2f)\n", 
//			right.x, d, left.y, right.y, right.z);
//		return  right;
//	}
// save right location
	rx = right.x;
// do up to 10 divides to improve precision
// if an exact mid point is found, return its right, or 1/2 between
// it and saved right if its right is the saved one
	for (n=0; n<12; n++) {  // Up to 12 iterations on this
	    if (signof(left.y) == signof(mid.y)) left = mid;
	    else if (signof(right.y) == signof(mid.y)) {
// DEBUG
//if (n > 8 && fabs(mid.y) > 0.5) printf ("x-x gbsearch error 3 %.2f %.2f %.2f\n",
//	mid.x, mid.y, mid.z);
		if (n > 8) return  mid;
		right = mid;
	    }
	    else {  // should not happen, give best result
// DEBUG ONLY
//printf ("x-x gbsearch bad error 2 near %.2f - (%.2f, %.2f, %.2f)\n", 
//	mid.x, left.y, mid.y, right.y);
//		return  gbsearch (left, mid, p, r);
		return  bsfunc (right.x);  // the basic right
//		return  right;
	    }
// Update p, dp in case we need them
	    dp *= 0.5;
	    p = fabs (dp);
// find a new mid location to compare
	    mid = bsfunc (0.5*(left.x + right.x));
// If we found a root, return its right or incremented location
	    if (mid.y == 0.0) {
// DEBUG ONLY
//printf ("x-x gbsearch returning mid root at %.2f\n", mid.x);
		if (right.x == rx) return bsfunc (mid.x + 0.25*dp);
		else  return  right;
	    }
	}
	x = right.y / (right.y - left.y);  // right fraction of interval to root
//DEBUG ONLY
//if (fabs(right.y) > 0.5) printf ("x-x gbsearch error 5 %.2f %.2f %.2f\n",
//	right.x, right.y, right.z);
	if (right.x == rx) return bsfunc (right.x - 0.5*x*(left.x-right.x));
/* The above puts the return 1/2 way between the actual root, which is
  at right.x - x*(left.s+right.x) and right.x; this satisfies the requirement
  that the return be on the right of the root, and left of the end of the
  segment, and distinct from both.    */
/*  also note:  b = l*x + (1-x)*r = r - r*x + l*x = r - x*(l-r)  */
	else  return  right;  // easy way if right has moved

#else  // on XTRABINSRCH
	if (signof(left.y) != signof(mid.y)) return  mid;
/* Within p, we are completing a binary search for the root.  Supposedly,
  before we get this far, if the root were bogus (different or incompatable
  signs of derivatives), the interval containing this would have been returned
  as no root.  If we find a root in the left section, we return mid as the
  return value.  Otherwise, we need to split the right section and find
  either a root, or we need to return a slightly incremented right value
  if the right value is an exact root.    */
	if (right.y == 0.0) return bsfunc (right.x + 0.5*dp);
	return  gbsearch (mid, right, p, r);
#endif  // on XTRABINSRCH
    }

/* See if the midpoint and its slope adds the possibility of a root
  within the entire interval.  Here, the slope might intercept
  either way, so we look at absolute value only.  */
//    ri = ri || ( fabs(mid.z) > fabs(mid.y/d) );

/* Now, if ri is not on, we have an interval with insufficient slopes to
  make an intercept from the known function values.  So long as the required
  splitting has been met, we declare this interval to be devoid of any
  root, and return the right end.  */
//    if (!ri) return  right;
// Temporarily ri fails, so we do not use it.

/* Now, consider whether this interval may not have a root at all.  It this
  appears to be the case, return the right end.
  To qualify, we need the derivatives at left, mid, right points to be
  substantially the same, and the characteristic length given by y/z to
  be longer than the interval.  Also the signs of y at all 3 points
  must be the same.  Compute the needed values and decide.
  The distance to right of point x is -y/z; so if -left.y/z > d and
  right.y/z > d and abs(mid.y/z)>d/2 we may be clear.  Actually, insist
  on twice that.  Compare dir*y/z to ad.    */
/* Comment out this stuff... */
    cl = -dir*left.y*xrecip(left.z);
// cl is distance from left to predicted root in direction dir
    cr =  dir*right.y*xrecip(right.z);
// and similarly for cr (direction -dr) and cm (either direction).
    cm =  fabs(mid.y*xrecip(mid.z));
    if (cl > 2.0*ad && cr > 2.0*ad && cm > ad &&
	signof(left.y) == signof (right.y) && 
	signof(left.y) == signof(mid.y)  )  return  right;
/*  ... End of commenting  */
/* We could and perhaps should, also consider computing slope from total
  dy/dx of the interval, comparing that to slope at left, right to see if
  it is between them, and that the mid slope is between the end point
  slopes to pass this test.  The test may need to be more restrictive
  to catch all roots.    */

#ifdef  LIMITPREC2  // fake out this one
/* At this point, if the minimum of cl, cr, cm is not larger than 2p, we
  need to decrease p to such a minimum; otherwise we risk missing the
  root by allowing too large a final interval.  */

// Try this one, it does not decrease when nearing root and y decreases...
    cs = amax (fabs(left.z), fabs(right.z));
    cs = amax (fabs(mid.z), cs);
    cs = 0.25 * fabs( amax(gpd.width, 1.0) * xrecip (cs));
/* The mistake is that width can validly be zero, in which case we don't do
  this in the same way.  */
    if (cs < p) p = cs;

// some debug stuff here -- bail out on bad value
    if (p < 1.0e-7) {
	printf ("   Left: %7.3f %7.3f %8.1f\n",  left.x,  left.y,  left.z);
	printf ("  Right: %7.3f %7.3f %8.1f\n", right.x, right.y, right.z);
	printf ("    Mid: %7.3f %7.3f %8.1f\n",   mid.x,   mid.y,   mid.z);
	printf ("  Width: %7.3f  cs %11.3e  p %11.3e\n", gpd.width, cs, p);
	exit (0);
    }
#endif  // on LIMITPREC

/* Finally, since we have not been able to exclude the possibility of a root,
  recursively examine each of the two sections.    */
final:
    res = gbsearch (left, mid, p, r);
    if (res.x == mid.x) res = gbsearch (mid, right, p, r);
    return  res;
}

#ifdef  ALLOWFAKESEARCH  // normally not defined
static	double	search (double start, double fin, pdata* p)
/* Interface for testing the regular calls; calling routine will not
 have the exact root this way, and it may mask errors in implementing
 the full search properly in the user subroutines.  */
{
vect3	left, right, res;
    gpd = *p;
    left = bsfunc (start);
    right = bsfunc (fin);
    res = gbsearch (left, right, Precision, 1.0);  // was  0.001, 1.0);
// Precision is needed looking for imacs angle, even though width of
// region is 14 degrees and max slope is about 751 deg/hr.
    return  res.x;
}
#endif  // on ALLOWFAKESEARCH

static	vect3	gsearch (double start, double fin, pdata* p)
/*  Similar to search, this entry point uses gbsearch with the proper
  arguments, and returns the full result.  Allows easier interface
  to programs, and can make precision and requirement global
  or arguments.  Allows more general gbsearch, and perhaps quicker.  */
{
vect3	left, right, res;
    gpd = *p;
    left = bsfunc (start);
    right = bsfunc (fin);
    res = gbsearch (left, right, Precision, 1.0);  // was  0.001, 1.0);
// Precision is needed looking for imacs angle, even though width of
// region is 14 degrees and max slope is about 751 deg/hr.
    return  res;
}

#else  // on BINSEARCH

static	double	search (double start, double fin, pdata* p)
/* Search from start to fin for zero of the dfunc function;
  return the value found.  If none, return a value at or past fin. */
{
double	dir;
double	beg;
double	nex;
vect2	val;
double	x;
double	step;
double	bstep;
double	rem;
vect2	ev;
vect2	cv;
int	bad;
int	fwd;

/* Find the direction; we may want to search either way */
    dir = fin - start;
    bstep = dir/10.0;		// basic step
    if (dir > 0.0) dir = 1.0;
    if (dir < 0.0) dir = -1.0;
#ifdef  IMACSANGLE
    if (p->side & 2) {
// Artificially limit IMACS angle step to about 2 degrees
	if (fabs(bstep) > 0.13) bstep = 0.13 * dir;
    }
#endif

    nex = beg = start;
    rem = fin - beg;
    ev = val = dfunc (beg, p);
#ifdef  DBSEARCH
printf (" Search from %.2f to %.2f (%.1f) %.2f rem.  Angl: %d\n", 
	start, fin, dir, rem, p->side & 2);
#endif

#ifdef  NOTDEFINED
/* Start by moving away from beg by enough to make slope and value
  the same in sign.  This might be a very small ways... */
    for (step=bstep, bad=1; bad; nex += bstep*0.2) {
	ev = dfunc (nex, p);
	bad = ev.x * ev.y <= 0.0;
// bad = true has us moving toward a root.  We don't want to return the
// root we are supposed to be leaving.  It is not bad if the root is
// a non-trivial distance away, so that ev.x > 0.1 or -ev.x/ev.y > 0.02.
// it is still bad if those ae not met, ev.x < 0.1 and -ev.x/ev.y < 0.02.
	bad = bad && ev.x < 0.1;
	if (ev.y != 0.0) bad = bad && -ev.x/ev.y < 0.02;
// This allows us to step away from a root gracefully, but is hard on
// finding two very close roots.
	bad = bad && ev.y*val.y > 0.0 && ev.x*val.x > 0.0;
// Also, do not step away further when value or slope changes sign.
    }
    beg = nex;
    rem = fin - beg;
    ev = val = dfunc (beg, p);
// above actually moves us away from some solutiions, so it is bad.
#endif  // above section removed...

    while ( dir*rem > 0.0 ) {
/* Always return if we have found it... */
	if (val.x == 0.0) return beg;
/* Compute a step size from our current beg, val, slope */
	if (val.y != 0.0) {
//	if (val.y != 0.0 && (p->side & 2 == 0)) {  // omit slope for A 2
/* We have a slope, see where it leads... */
/* Operating from:
    y = dv/dx = (v-0) / (beg-x)
    beg-x = v/y
    x = beg - v/y			*/
	    x = beg - val.x / val.y;
	    if (dir*(x-beg) < 0.0) {	// before beg, getting bigger
		step = bstep;
		fwd = 0;
	    } else {	// Step just a little past (x-beg)
//		step = bstep/4.0 + x - beg;
		step = x-beg + fabs(val.x/val.y)/5.0;
/* Limit the step to under 2*bstep when doing this */
		if (step*dir > 2.0*bstep*dir) step = 2.0*bstep;
		fwd = 1;
	    }
	} else {
	    step = bstep;
	    fwd = 1;
	}
/* Having a step size, find a trial end to the interval, nex; 
  check if that is within the bound (fin);
  and at that point find the function. */
#ifdef  DBSEARCH
printf (" Step start = %.2f", step);
#endif
	if (step*dir > dir*rem ) step = rem;
	for (bad=1; bad; step /= 2.0) {
	    nex = beg + step;
/* It is possible to reduce step below a useful increment value,
  and thus enter an infinite loop here.  To prevent this, we will
  break this loop if step becomes so small as to not change the
  argument; and restore the previous step.  KDC 2004/5-12 */
	    if (nex == beg) {  // safety escape.
//	printf (" -- Debug exit from loop.  Step size is %.7g\n", step);
//	printf (" -- val is %.5g / %.5g at %.5g\n", val.x, val.y, beg);
//	printf (" --  ev is %.5g / %.5g at %.5g\n", ev.x, ev.y, nex);
		step *= 2.0;	// Increase to 4.0 if problems...
		break;	// loop exit
	    }  // end safety section
	    ev = dfunc (nex, p);
 /* Can't have movement over 30 deg. -- see if so... */
	    bad = fabsdif (val.x, ev.x, 360.0) > 60.0;
// use val.x - ev.x if p.side & 2 flag is on			<---
// Also, detect if ev inticates a flip in this case             <---
/* Also bad if we had fwd, and now don't */
/* Bad if slope changed sign, while value didn't */
	    bad = bad || (fwd && val.x*ev.x > 0.0 && val.y*ev.y < 0.0 );
/* This method of loop breaking is perhaps safer, but can result in
  the break happening when it otherwise would not, and would otherwise
  not result in a problem.  It may be needed later, however. */
//	    if (fabs(step/beg) < 2.e-15) {  // safety escape.
//		break;	// debug loop exit
//	    }  // end debug section
	}


#ifdef  DBSEARCH
printf (" refined to %.2f\n", step);
#endif

/* Step size should be smaller at beginning and end of interval.
  A relatively fine search between is good, since targets with small
  width are frequent.  We need to find roots near the peak of the
  target function and within the current slope regeme, so slope
  changing sign from its value at beg is bad, indicating we are
  probably past a root, unless acompanied by a uniformly large
  value amount.  */

/* See if we have found the root, or can straddle it... */
/* NOTE -- If we are passing around the back, and x values are near
  +/- 180, we don't have a root, but we do change signs.  Requiring
  a small change in x allows us to find the nearest root, and then
  requiring the small difference be about 0 finds only true roots. */
	if ( (val.x*ev.x <= 0.0) && (fabs(val.x-ev.x) < 90.0) ) break;
// The next line detects a flip in the IMACS angle search case,
// which should be treated like a root, so we do a break here.
#ifdef  IMACSANGLE
	if ( (p->side & 2) && (fabs(val.x-ev.x) > 180.0) ) break;
#endif
/* Havn't found a root within beg-nex; keep trying */
#ifdef  DBSEARCH
printf (" Beg %.2f at %.2f : Fin %.2f at %.2f skipped.\n",
   val.x, beg,  ev.x, nex);
#endif
	beg = nex;
	val = ev;
	rem = fin - beg;
    }	// End of interval search

/* Now we have one of the following cases:
  1) beg = fin, and we have found nothing
  2) The beg,nex interval straddles a root  */
    if (beg == fin) {
//printf (" Return end of interval %.2f\n", fin + 0.001*dir);
	return fin + 0.001*dir;
    }

#ifdef  DBSEARCH
printf ("  Finding root between %.2f at %.2f and %.2f at %.2f\n",
   val.x, beg,  ev.x, nex);
#endif

/* We enter a binary search for the root, using a .001 interval
  as our resolution limit */
    while (fabs(nex-beg) > 0.001) {
	x = (beg+nex) / 2.0;
	cv = dfunc (x, p);
// if p->side & 2 flag is on, check for possible flip      <--
	if (cv.x == 0.0) return x;
	if (cv.x*val.x > 0.0) {	// same side as beg
	    beg = x;
	    val = cv;
	} else {		// same side as nex
	    nex = x;
	    ev = cv;
	}
    }

/* Do a final sub-division, and return that as the result. */
#ifdef  DBSEARCH
printf (" Search returning solution at %.2f\n", (beg+nex)/2.0);
#endif
    return ( (beg+nex) / 2.0 );
}

#endif  // on BINSEARCH


/* Find a queue of rotator flip angles */
anglist   *qflip (pdata *pd)
{
anglist	*p;
anglist *q = NULL;
#ifdef  BINSEARCH
vect3	next;
//vect3	left, right;
#else
double	h;
#endif
double	x;
double	sdh;

/* Set up for the scan */
    sdh = sda (pd->dec, pd->lat);
    pd->width = 0.0;
    pd->target = 180.0;	// May parameterize?
#ifdef  BINSEARCH
//    gpd = *pd;
//    left = bsfunc (-sdh);
//    right = bsfunc (sdh);
    x = -sdh;
/* Find all possible instances of target... */
    for (;;) {
//	left = gbsearch (left, right, Precision, 1.0);
	next = gsearch (x, sdh, pd);
#ifdef  DBQWARN
	printf (" Searching %.3f to %.3f for 180 flip.\n", x, sdh);
#endif
	x = next.x;
	if (x < sdh) {
/* Found a root, put it in the queue */
#ifdef  DBQWARN
	printf (" -found flip at %.3f root %.2f dy/dx %.2f\n",
		next.x, next.y, next.z );
//	if (fabs(next.y) > 0.5) continue;  // omit the bad ones...
#endif
	    p = xalloc (anglist);
//	    p->v = next.x - next.y/next.z;
	    p->v = x;
	    if (next.z != 0.0) p->v -= next.y/next.z;
	    q = G_push_l (q, p);
	} else if (x == sdh && sdh == 0.0) {
// Special case -- invisible object (never rises) shown as rotator conflict.
	    p = xalloc (anglist);
	    p->v = 0.0;
	    q = G_push_l (q, p);
	    break;
	} else  break;
    }
#else
//    x = -sdh;
    while (x < sdh) {
	h = search (x, sdh, pd);
	if (h < sdh) {
/* Found a root at h.  Push it onto a queue */
	    p = xalloc (anglist);  // defined in optutils.h
	    p->v = h;
	    q = G_push_l (q, p);
	}
	x = h + 0.001;
    }
#endif
    return  q;
}

bandlist  *qkill (bandlist *q)
/*  Kill dynamic bandlist queue at *q, returning NULL.  */
{
bandlist  *b;

    while (q != NULL) {	// Kill existing queue
//	b = q;
	q = G_pop_l (b=q);
//	q = q->next;
	free (b);
    }
    return  q;
}


/* Find a queue of occultation bands */
// bandlist  *qwarn (double sdh, pdata *pd)
bandlist  *qwarn (pdata *pd, double a, double b, vect2* lim)
/* For band from rotation value a to b, find warning bands.
  Vector lim is search limits lim->x to lim->y in ha hours;
  if NULL, search all the semi-diurnal arc.    */
{
bandlist *p;
bandlist *q = NULL;
#ifdef  BINSEARCH
//vect3	left, next, right;
vect3	left, right;
#else  // on BINSEARCH
// put unused variables here
double	x;
double	h;
double	f;
vect2	v;
int	nl=0;	// debug
#endif  // on BINSEARCH
double	sdh;
double	sdl;

/* Set up for the scan */
    if (lim == NULL) {  // default to full arc if not specified
	sdh = sda (pd->dec, pd->lat);
	sdl = -sdh;
    } else {		// use any limits actually specified
	sdh = lim->y;
	sdl = lim->x;
    }
//    pd->width = 45.0;	// May parameterize?
//    pd->target = RotOff;	// May parameterize?
    pd->target = (a + b)/2.0;
    pd->width = fabs (b - a)/2.0;

#ifdef  DBQWARN
	printf (" Qwarn from %.3f to %.3f for %.2f +- %.2f  A %d\n",
		sdl, sdh, pd->target, pd->width, pd->side & 2 );
#endif

/* This is more complex than single target.
  We do:
Find if we are within the band at x.
If so, set start as our first root;
If not, find the next root.

The loop:
	If no root, break the loop.
	Test is if current x is past sdh => no root.
	Check that our start root is going in. (call dfunc, look at y)
	Find the next root, it will be going out (check).
	Put the two roots into the queue.	
	Find the next root and end the loop.
*/

#ifdef  BINSEARCH

// initialize for the loop
// Find the left band start, either at x or the next root
    gpd = *pd;
    left = bsfunc (sdl);
//    if (left.y < 0.0) left = gbsearch (left, right, Precision, 1.0);
    if (left.y < 0.0 || (left.y == 0.0 && left.z < 0.0)) {
/* If the starting function value is negative, we are not in a
  desired region, and if it is zero but derivative is negative,
  we are ending, not starting a region.  Unless the function value
  is actually positive, or zero with positive derivative, we have
  to proceede to the next root, which should start a region.    */
	left = gsearch (left.x, sdh, pd);
// should check that right.z is positive?  (We do that soon enough)
    }

// left is now the start of a band, or if advanced to sdh, there is none
    while (left.x < sdh) {
// Find the right root of the band (or end), and call it h
	right = gsearch (left.x, sdh, pd);

// If the derivatives are wrong sign, we need to bitch and moan here,
// and we also need to drop out of the loop as it is lost...
// Carefully check slopes only when not at sdl, sdh limits!
	if ( (left.x > sdl && left.z  < 0.0) || 
	    (right.x < sdh && right.z > 0.0) ) {
	    printf (" *Scan Error 3: Looking %.3f to %.3f, found:\n", sdl, sdh);
	    printf ("   %.3f %.2f (%.1f) -- %.3f %.2f (%.1f) FAIL\n",
		left.x, left.y, left.z,  right.x, right.y, right.z );
/* If we get here, the problem is that point right, which is supposed to be
  the final point of the band has a positive derivative when it should be
  negative; or point left, which is supposed to be the start of the 
  band has a negative derivative when it should be positive.  In any case,
  we appear to have missed a root, so we fail.  */
/* One possible recovery strategy is to set right to left, and then look
  for left again between the previous right and the current right.  To do
  that properly, save the right.x somewhere, start it at sdl.  */
	    break;
	}

// Create a queue entry for left to right and insert it
	p = xalloc (bandlist);
	p->a = left.x;
	p->b = right.x;
// Carefully, testing for actual roots (not just search limits),
// use .y and .z values to find the exact roots;
// The gsearch returns the starting point for the next search,
// which is usually slightly behond the true root.
	if (left.x  > sdl &&  left.z != 0.0) p->a -= left.y/left.z;
	if (right.x < sdh && right.z != 0.0) p->b -= right.y/right.z;
	q = G_push_l (q, p);
	if (right.x >= sdh) break;  // reached the end there

// To continue from here, we will need to find a new left band start by
// searching from right, and verify that it is an "in" node (done at top)
	left = gsearch (right.x, sdh, pd);
	if (left.x >= sdh) break;	// there is no next band
    }	// end of search loop

#else  // on BINSEARCH

    h = f = x = sdl;

/* Find this or next x which is in band or going in */
    v = dfunc (x, pd);
#ifdef  DBSEARCH
    printf ("  Initial x %.2f, val %.2f, rate %.2f\n", x, v.x, v.y);
#endif
/* Initially, we are in an interval if v.x >= 0.0 and do not search,
  or we are out, and a search is needed.  */
//    if (v.x < 0.0 || (v.x == 0.0 && v.y < 0.0)) {	// out or leaving
    if (v.x < 0.0) {   // Not in an interval, so we search for one
	x = search (x, sdh, pd);  // was x+0.001
    }
// x is thus the start of a band, which may be in process at start.

// Search for out (v.x < 0.0) or end of range
    while (x < sdh) {  // was <= which is wrong
	if (nl++ > 4) break;	// debug loop killer
	v = dfunc (x, pd);
/* At this point, we expect to be in or starting an interval, with
  v.x at least small or positive, never too negative.  And unless we are
  just starting v.y needs to be positive.  If these are not true, we have
  found an end while expecting a begining.  That is a scan error, and
  we need to set h as the end, and x falls back to f (unless x == f).  */
//	if (v.y < 0.0 && x > (0.01-sdh)) {
// That test is for low limit, not actually for sdh
	if (v.y < 0.0 && x > (sdh-0.01)) break;	// found the end
	if (v.y < 0.0 && x > sdl) {
	    printf ("   Search fail 1. (Expect in) %.2f d %.2f at %.2f  A %d\n", 
	    	v.x, v.y, x, pd->side & 2);
/* What does this mean?  x is either the start (sdl) (in which case
  the v.y can be either + or -) or it is a band start, where v.y > 0 is
  probably required.  At start the x>sdl is false, and at band start
  the y<0 will be false.  If either of these fails we are not starting
  a band and the search is logically failed.  It is possible that the
  search found the out edge rather than an in edge.  Restore x before
  the last search.  */
/* At this point we should have v.x > 0 or at least v.x very small; and
  if x is away from sdl, v.y should be positive.  This would indicate
  the start of a band.  If x >= sdh we are at the end and should break.  */
/* Since v.y is negative, we have an out edge expecting an in edge.  Two
  choices - could be we did not get away from the last out edge in the
  scan, or we might have missed the in edge.  We keep the last x in f,
  and the last out edge is in h, so should be able to tell.  If we have
  not advanced from f, we do another search from further out; if we did
  advance, we need to search better, and have a real error.
  Assuming f < h, if x nearer f need to search from h, if past h,
  need to search better.  */
	    if (x <= h) {  // did not advance
		x = search (h+0.01, sdh, pd);
	    } else {  // Skipped over entry - true error
		printf ("  Skipped band, trying again.\n");
		x = search (0.5*(h+x), sdh, pd);
	    }
	    continue;
	}
/* Should now have x at start of a band. */
/* Find the next root, the going out side */
	h = search (x+0.0001, sdh, pd);  // Was 0.01
	v = dfunc (h, pd);
	if (h >= sdh) break;  // found the end
	if (v.y > 0.0 && h < (sdh-0.0001)) {	// was 0.01
	    printf ("  Search fail 2. (Expect out) %.2f d %.2f at %.2f  A %d\n", 
			v.x, v.y, h, pd->side & 2);
/* This is similar to the above, h was looking for the outgoing side, and
  an in-going side has been found.  Just continue, keeping x as in-going.
  There are two possibilities, either h did not advance, in which case
  h will be near x, or it went too far  */
#ifdef  DBSEARCH
//	    printf ("  Searching from %.2f got %.2f, val %.2f, rate %.2f; to %.2f\n",
//		    x, h, v.x, v.y, sdh);
#endif
	    if (h <= x) {
		h = search (x+0.0001, sdh, pd);  // was 0.01
	    } else {  // skipped, true error.
		printf (" Skipped band, trying again.\n");
//		h = search (x - fabs(v.x/v.y) + 0.01, sdh, pd);
		h = search (0.5*(x+h), sdh, pd);
	    }
//	    x = f =  h;	// Save in-going side found
	    continue;
	}
	if (x < h) {
//	    p = (bandlist*) malloc (sizeof(bandlist));
	    p = xalloc (bandlist);
	    p->a = x;
	    p->b = h;
//	    q = G_push_l (p, q);	// wrong order!
	    q = G_push_l (q, p);
// printf (" Found occultation from %.2f to %.2f\n", x, h);
	}
	if (h >= sdh) break;
/* NOTE -- We need to explicitly break on h here, or else the next
  search for x will search backward from a value past sdh to sdh itself! */
	f = x;
	x = search (h+0.0001, sdh, pd);  // was 0.01
// printf (" Loop on occultation test at x = %.2f\n", x);
    }

#endif  // on BINSEARCH

#ifdef  DBQWARN
    if (listcount(q) > 0) {  // later change to listcount(q) > 1
	printf (" -returns:");
	C_loop (p, q) {
	    printf (" (%.3f,%.3f)", p->a, p->b);
	}
	printf ("\n");
    }
#endif
    return  q;
}


static	bandlist*  merge_band (bandlist* aq, bandlist* bq)
/* Merge two queues of bandlist items into a single queue, with proper
  overlap and sequencing.  If the ends of items are within the
  "Close" parameter, the items are considered to overlap, and
  are combined into one.    */
{
bandlist  *al, *bl;
bandlist  *rq=NULL;
bandlist  *p;

    if (bq == NULL) return  aq;
    if (aq == NULL) return  bq;

#ifdef  DBQWARN
    printf (" Merge lists:\n");
    printf ("  A:");
    C_loop (p, aq) printf (" (%.3f,%.3f)", p->a, p->b);
    printf ("\n  B:");
    C_loop (p, bq) printf (" (%.3f,%.3f)", p->a, p->b);
    printf ("\n");
#endif

    for (al=aq, bl=bq; al != NULL && bl != NULL; ) {  // special logic
// If a comes first, pop a and put it after rq, and continue
// If b comes first, pop b and put it after rq, and continue
	if (al->b + Close < bl->a) {
	    al = G_pop_l (p = al);
	    rq = G_push_l (rq, p);
	    continue;
	}
	if (bl->b + Close < al->a) {
	    bl = G_pop_l (p = bl);
	    rq = G_push_l (rq, p);
	    continue;
	}
// Solve the overlap.
// Pop a, it is the next item for result
// First edge of a is min of a, b first edge;
// last edge is max of a, b last edge.
// push the a item as corrected onto result queue
// Pop b, it will be killed.
// free the b item which was popped.
	al = G_pop_l (p = al);		// Take a list item
	p->a = p->a < bl->a ? p->a : bl->a;	// Find new left edge
	p->b = p->b > bl->b ? p->b : bl->b;	// Find new right edge
	rq = G_push_l (rq, p);		// Put item on result
	bl = G_pop_l (p = bl);		// Pop the b list item
	free (p);		// Free the redundant item
    }

// If a tail of a or b remains, put it at the end of result.
    rq = G_push_l (rq, al);
    rq = G_push_l (rq, bl);
// return the result
#ifdef  DBQWARN
    printf ("  Q:");
    C_loop (p, rq) printf (" (%.3f,%.3f)", p->a, p->b);
    printf ("\n");
#endif
    return  rq;
}

#ifdef  USEXQWARN

static	bandlist*  xqwarn (pdata *pd, anglist* qa, double a, double b)
/* For band from rotation value a to b, find warning bands, searching the
  limits from horizon to horizon, breaking up the search at the flip
  angles in list qa (if any) and merging the resulting queues.  Use
  qwarn to search each segment.    */
{
anglist  *ap;
bandlist  *dq = NULL;
bandlist  *eq = NULL;
vect2	lim;
double	sdh;


#ifdef  DBQWARN
bandlist  *p;
	printf (" Xqwarn %.3f to %.3f   A %d\n",
		 a, b, /* pd->target, pd->width, */ pd->side & 2 );
#endif
// Set up the horizon limit and initial limit
    sdh = sda (pd->dec, pd->lat);
    lim.x = -sdh;
// Loop on end points at each known flip, find queues inside
    C_loop (ap, qa) {
	lim.y = ap->v;
	eq = qwarn (pd, a, b, &lim);
	dq = merge_band (dq, eq);
	eq = NULL;
	lim.x = ap->v;
    }
// Find the last interval
    lim.y = sdh;
    eq = qwarn (pd, a, b, &lim);
    dq = merge_band (dq, eq);
    eq = NULL;

// the return is the summed queue of all results found
#ifdef  DBQWARN
    if (listcount(dq) > 0) {  // later change to listcount(q) > 1
	printf (" =Xqwarn Returns:");
	C_loop (p, dq) {
	    printf (" (%.3f,%.3f)", p->a, p->b);
	}
	printf ("\n");
    }
#endif
    return  dq;
}

#endif  // on USEXQWARN

#define  SIMPLEBLEND
//#undef  SIMPLEBLEND
/*  Using symbol  SIMPLEBLEND  to change the code section which combines
  the flip locations and rotator warning bands into a blended queue of
  warning bands.  The old way seems needlessly complicated, and the new
  way just finds all the warnings and merges them with the merge program.  */

static	void	find_rotator (int  doq, int df)
/* On action flag  doq,  and globals such as queues rotaq - rotcq,
  rotator limits, the pdx global structure defined above,
  and the appropriate subroutines, find  all the rotator limits and
  warning values.  Called by initialization below and by TextUpdate
  in guisubs, this combined code will result in queues of lone flips
  (which should not exist), single sided warnings and combined warnings.
  The debug flag df allows for some informative output.  */
{
anglist   *al;
bandlist  *bl;
bandlist  *cl;
//bandlist *rotdq = NULL;
bandlist *roteq = NULL;
/* Above are now local here. */

#ifndef  SIMPLEBLEND
anglist	*xl;
double	a;
double	b, c;
int	k;
#endif

/* To make new queues, first kill the old ones to prevent memory leak. */
    while (rotaq != NULL) {  // kill the rotaq queue
	rotaq = G_pop_l (al = rotaq);
	free (al);
    }
    rotcq = qkill (rotcq);
//    rotdq = qkill (rotdq);
//    roteq = qkill (roteq);
/* NOTE -- It would be a good thing for rotdq and roteq to be local
  to this program, and rotcq is the only one which needs to be global.  */

/* Find queues of rotator flips and rotator soft limit warning zones. */
    if (doq) {
	rotaq = qflip (&pdx);

#ifdef  DBQWARN
    printf ("\n Dec. %.2f,   PA %.2f\n", pdx.dec, pdx.pa);
    if (listcount (rotaq) > 0) {
	printf ("  Flip Angles found:");
	C_loop (al, rotaq) printf (" %.3f", al->v);
	printf ("\n");
    }
#endif
// Add computation of rotdq, roteq here also
// Use standard limits for rotator on both instruments
/* NOTE -- Add here the ldss rotator limits for doq == 2 case.  */
//      if (doq == 1) {	// These are IMACS only items
//	rotdq = qwarn (&pdx, -180.0, -176.0, NULL);	// or -151.0
//	roteq = qwarn (&pdx,  176.0,  180.0, NULL);
//	if (df) printf (" Warn rotator 176.0 -176.0\n");
//      } else if (doq == 2) {  // For LDSS, for example
/* NOTE - please substitute actual rotator soft limits for LDSS when
  these values are known.  For now, guess at some limits near 180.    */
/* These are now rotator limits for everyone...  */
#ifdef  USEXQWARN
	rotdq = xqwarn (&pdx, rotaq, -180.0, RotLim1);
	roteq = xqwarn (&pdx, rotaq, RotLim2,  180.0);
#else  // on USEXQWARN
	rotdq = qwarn (&pdx, -180.0, RotLim1, NULL);
	roteq = qwarn (&pdx, RotLim2,  180.0, NULL);
#endif  // on USEXQWARN
	if (df) printf (" Warn rotator %.2f %.2f\n", RotLim2, RotLim1);
//      }
/* If the instrument, slightly encoded in doq, is not supported here,
  the conflict zones found by qwarn and rotator limits are empty, and
  only the flip points found above by qflip are non-NULL.  This allows
  compatable mode flip warnings.    */
    }
// Find hrf for backward compatability
    hrf = 30.0;
    C_loop (al, rotaq) {
	if (fabs(al->v) < fabs(hrf)) hrf = al->v;
    }
// hrf is now the smallest (highest) flip location
/* Consider using the same method to find hrf as the smallest rotator
  warning band start location.  In fact, hrf is improved below, using
  the rotcq result (if any).  Still use hrf global as a warning in the
  intgui window, and when needed in the pop-up warning for write obs file.  */

/* Combine flip queues and soft limit warning queues where they meet or
  overlap.  Drop queue elements thusly combined.  */

#ifdef  SIMPLEBLEND
/* Make up a band queue for the flip locations in rotcq. */
    C_loop (al, rotaq) {
	cl = xalloc (bandlist);
	cl->a = cl->b = al->v;
	rotcq = G_push_l (rotcq, cl);
    }
// Having done this, drop rotaq entirely
    while (rotaq != NULL) {  // kill the rotaq queue
	rotaq = G_pop_l (al = rotaq);
	free (al);
    }

#else  // on SIMPLEBLEND
/* For all flip locations found, there should be at least one and usually
  two warning queues starting/ending at that flip.  Gather them all.  */
    C_loop (al, rotaq) {
	cl = NULL;
	a = b = c = al->v;  // save current flip point
	C_loop (bl, rotdq) {
	    k = fabs(bl->a - a) < Close ? 1 : 0;
	    k = fabs(bl->b - a) < Close ? 2 : k;
	    if (k)  {
		if (k == 1) b = bl->b; else b = bl->a;
		rotdq = G_head(G_pop_l (cl = bl));
		al->v = -99.0;  // mark, to not loose queue
		break;  // popped current member, can't continue
	    }
	}
	C_loop (bl, roteq) {
	    k = fabs(bl->a - a) < Close ? 1 : 0;
	    k = fabs(bl->b - a) < Close ? 2 : k;
	    if (k)  {
		if (k == 1) c = bl->b; else c = bl->a;
		if (cl == NULL) {
		    roteq = G_head(G_pop_l (cl = bl));
		} else {
		    roteq = G_head(G_pop_l (bl));
		    free (bl);
		}
		al->v = -99.0;  // mark, to not loose queue
		break;  // popped current member, can't continue
	    }
	}
// cl should point to the rotdq element matching al or at least the
// roteq element matching.  If it is null, then qwarn has failed to
// find the flip regions, which is bad.
	if (cl == NULL)  {  // continue;  // should not happen...
// MIGHT WANT TO PUT AN ERROR MESSAGE HERE AS WE SHOULD FIND SOMETHING
/* At this point, b and c are the "other" sides of the warning band(s)
  which are being mated.  Put them in the proper order.  */
	    printf (" *=* Flip combine ERROR, cl NULL.\n");
	    continue;
	}
	if (c > b) {
	    cl->a = b;
	    cl->b = c;
	} else {
	    cl->a = c;
	    cl->b = b;
	}
	if (a < cl->a) cl->a = a;  //| seldom used but needed for
	if (a > cl->b) cl->b = a;  //| proper logic here.
	rotcq = G_push_l (rotcq, cl);
    }

/* Need to drop any marked items in the rotaq queue. */
    for (al=rotaq; al != NULL; ) {  // special queue logic here
	xl = al->next;
	if (al->v < -98.0) {
	    rotaq = G_head (G_pop_l (al));
	    free (al);
	}
	al = xl;
    }
#endif  // on SIMPLEBLEND

#ifdef  DBQWARN
//	printf (" -rotcq:");
//	C_loop (bl, rotcq) {
//	    printf (" (%.3f,%.3f)", bl->a, bl->b);
//	}
//	printf ("\n");
//	printf (" -rotdq:");
//	C_loop (bl, rotdq) {
//	    printf (" (%.3f,%.3f)", bl->a, bl->b);
//	}
//	printf ("\n");
//	printf (" -roteq:");
//	C_loop (bl, roteq) {
//	    printf (" (%.3f,%.3f)", bl->a, bl->b);
//	}
//	printf ("\n");
#endif

/* Merge the queues, two at a time with overlap combining.  */
    rotdq = merge_band (rotdq, roteq);
    roteq = NULL;
    rotcq = merge_band (rotcq, rotdq);
    rotdq = NULL;

/* For IMACS only, compute rotdq as limit of IMACS angles, and
  combine that with the combined queue when debugged.  */
    if (doq == 1 && (-IMACSlimit > RotLim1)) {	// IMACS only
double	sdh;
vect2	lim;
bandlist  *bp;
double	ixlo = IMACSlimit + IMACSrot;
double	ixhi = 180.0 + IMACSrot;

	pdx.side |= 2;
	sdh = sda (pdx.dec, pdx.lat);
/* In a loop, call qwarn with limits set to the areas between the sda
  limits and the known bands currently in rotcq.  For each returned
  set of limits, merge that with the rotdq.  When done, merge rotdq
  with rotcq.  We need the sda and limits and a local vector here.  */
	if (df && (-IMACSlimit > RotLim1)) 
		printf (" Warn IMACS angle > %.2f\n", ixlo);
	lim.x = -sdh;
#ifdef  DBQWARN
	printf (" Flips:");
	C_loop (bp, rotcq) {
	    printf (" (%.3f,%.3f)", bp->a, bp->b);
	}
	printf ("\n");
#endif
	C_loop (bp, rotcq) {
	    lim.y = bp->a;  // was  + 0.001;
	    roteq = qwarn (&pdx, ixlo, ixhi, &lim);
	    rotdq = merge_band (rotdq, roteq);
	    roteq = NULL;
	    lim.x = bp->b;  // was  - 0.001;
/* The use of 0.001 into the flip region is stopped.  We should not go
  any distance into the flip, as those regions need to be avoided.  And,
  gaps between regions smaller than "Close" (currently 0.01) are removed
  by the merge_band process we will do anyway.    */
	}
	lim.y = sdh;
	roteq = qwarn (&pdx, ixlo, ixhi, &lim);
	rotdq = merge_band (rotdq, roteq);
	roteq = NULL;
// Above is really 151 + RotOff, 180 + RotOff, current RotOff = -46.15
	pdx.side &= 1;  /* should use (bitwise not)2 */
//  Add these two lines after debug with orange band for rotdq (in guisubs)
// ALSO change rotdq from global to local define!
#ifndef  DBQWARN
	rotcq = merge_band (rotcq, rotdq);
	rotdq = NULL;
#endif
/* NOTE-NOTE the definition of "rotdq" should be changed from global
  to local, although this is not strictly necessary.  And, when that
  is done, we should comment out the plotting of the orange band
  in guisubs.c, although setting rotdq to NULL effectively skips any
  plotting.  One might add a feature symbol for orange band on debug
  if feeling really ambitious...  */
    }	// End of IMACS angle search

    if (dodebug > 1 && listcount(rotcq) > 0) {
	printf (" Rotator:");
	C_loop (bl, rotcq) {
	    printf (" (%.3f,%.3f)", bl->a, bl->b);
	}
	printf ("\n");
    }

/* Compute a better hrf using rotcq intervals... */
    C_loop (bl, rotcq) {
	if (bl->a < 0.0 && bl->b > 0.0) hrf = 0.0;
	if (fabs(bl->a) < fabs(hrf)) hrf = bl->a;
	if (fabs(bl->b) < fabs(hrf)) hrf = bl->b;
    }

}	// end of find_rotator


/* Obsolete gflip, ggflip programs have been removed from here. */


	/* - - - - - - - - - - - *\
	|                         |
	|  IFU support programs   |
	|                         |
	\* - - - - - - - - - - - */

#ifdef  IFUM

void	IFUfindcenter (Obs* ob)
{
/* From data in ob structure on IFUmode and object position,
  find the correct center position as ra/dec in structure.  */

double	ab, db;		// object
double	da, dd;		// coordinate differences
double	dm;		// Mid declination
double	cdel;		// Cosine mid-delta for scale
vect2	base;		// base offset of object from center
vect2	p;		// offset at o (pa augmented by da)
int	m;		// Mode index
int	i;		// Loop counter

// First, return any null or trivial cases
// Check mode; we only compute for mode 1 or 2
    m = ob->IFUmode - 1;
    if (m<0 || m>1) return;

// Establish the offset vector in radians, in normal xy, from
// the as yet unestablished center.
// This vector would not change, it is c->o at c.
    base = mul2vect (make2vect (IFUdx[m], IFUdy[m]), ArcSecond);
    p = rot2 (base, ob->angle);

// Set up values for finding center
    ab = ob->IFUra;
    db = ob->IFUdec;

// Initialize the differences based on the offset
    da = -p.x;
    dd =  p.y;

// Do following 2 or 3 times...
// Loop about 3 times to converge on offset
    for (i=0; i<3; i++) {
	dm = db - 0.5*dd;
	cdel = cos(dm);
	p = rot2 (base, ob->angle + da * sin(dm) );	// think sign ok
	da = -p.x / cdel;
	dd =  p.y;
    }

// Place result into the structure...
    ob->ra = ab - da;
    ob->dec = db - dd;
}

#endif
//  End IFUM conditional


	/* - - - - - - - - - - - *\
	|                         |
	|  X-window GUI programs  |
	|                         |
	\* - - - - - - - - - - - */


/*  ====  X-Window program for the GUI  ====  */

/* These programs are "borrowed" from "dbplot.c" stuff,
  and also the main GUI of GMAP itself.. (gmap.c)... */


/*  Function prototypes of named handlers  */

// static void	handle_iqmenu (void*);

#ifdef  SKYOPT
static void	handle_optmenu	(void*);

// For reading the map data, as gmap does it...
static void    setup_defaults     (void);
static section**  get_sections       (int);

#endif

static void	handle_exmenu (void);

void  handle_active (void* param);

#ifdef  GUIDATE
//void	handle_datemnu (void* param);
static void   handle_datemnu (void* param);
#endif

#ifdef  OBJSHAPE
#ifdef  OBJAUTO
static void   handle_shape (void* param);
#else
static void   handle_shape (void);
#endif
#endif

/*  ==== Graphics Globals ==== */

// Obf	*Gob;		/* Graphics section obf pointer global */
// Above is a globalization kludge...
// That's been moved up, near the start...

Bool	done=False;
int	writ=0;

// done is also in dbplot.o in the library... ? should not be linked there.

EditWindow  dfwin;		/* Data file name */
EditWindow  obwin;		/* Observer name */
EditWindow  ttlwin;		/* Title field */
EditWindow  alwin, dewin;	/* Coordinates */
EditWindow  temwin;		/* Temperature */
EditWindow  hawin;		/* Hour Angle */
EditWindow  cwwin;		/* Center Wavelength */
EditWindow  spawin;		/* Slit Position Angle */
EditWindow  swidwin;		/* Slit Width */
EditWindow  slawin, slbwin;	/* Slit width windows */
EditWindow  dcawin, dcbwin;	/* Do-not-cut values */
EditWindow  sltwin;		/* Slit tilt window */
EditWindow  rwidwin;		/* Ref. hole size */
EditWindow  blupb, redpb;	/* PassBand windows */
EditWindow  bludb, reddb;	/* Detector PassBand windows */
EditWindow  ordwin;		/* Disperser order */
// EditWindow  exordw;	/* Extra order flag */
EditWindow  repob, reprf;	/* Re-use count windows */
EditWindow  refsv;		/* Reference selection */
EditWindow  ovlwin;		/* Overlap Specifier */
EditWindow  mhwin, pdwin;	/* Must have and decide priorities */
EditWindow  xpwin;		/* Extend priority for expri/apri; on EXTPRI */

// With skywindow, the following need to be global, not static
  Application *app;
  MainWindow  mwin;

static Window	win;

//static Button	slitype;	/* Slit/hole button */	//stbut
static Button	objlist;	/* Object list button */
static Button	noadc;		/* For no adc telescope */
static Button	slext;		/* Slit extension */
static Button	drefbut;	/* Differential Refraction */
static Button	testbut;
static Button	actbut;		/* Trial maskgen control */
static Button	kilbut;		/* Test utility kill */
static Button	nsrot;		/* Nod/Shufl rotation feature */


static Menu    iqmenu;		/* init/quit (drop-down) */
//  static Menu    optmenu;		/* option drop-down */

static Menu	eqmenu;		/* Equinox pop-up  */
// change iqmenu to filemenu

static Menu	xordmu;		/* Extra order drop-down */

#ifdef  SKYOPT
Menu    optmenu;		/* option drop-down */
// NOT static, as skywin needs to know it.
#endif

//#ifdef  SKYOPT
float	cp;		/* casspos = -(180-pa) */
float	pa = 0.0;	/* an actual position angle */
//#endif


static Menu	instmu;		/* Instrument pop-up */
static Menu	dispmu;		/* Disperser pop-up */

static Menu	filtmu;		/* Filter menu */

#ifdef  GUIDATE
static Menu	yearmnu;
static Menu	monthmnu;
static Menu	daymnu;
#endif

#ifdef  OBJSHAPE
static Menu	shapemnu;
#endif

#ifdef  GAPVO
static Menu	gapmu;		/* Virtual gap pop-up */
#endif

static	Bool	iconized=False;


/*  --  IFU window things...  */
#ifdef  IFUM
int	IFU_active;
MainWindow	IFUwin;
EditWindow	aobwin;
EditWindow	dobwin;
Menu		sidemu;
// Also need selector for left/right
#endif


// static  int	wx, wy;   // These are unused

enum file_menus { 
// add more here for iqmenu entries
		    FILE_SAVE,
		    FILE_SAVEAS,
		    FILE_DUMMY,
		    FILE_DONE,
		    FILE_KILL,
		    FILE_EXIT
		};

enum equnioxes { EQU_1950,EQU_2000,EQU_TODAY,EQU_SET};

#ifdef  SKYOPT
enum opt_menus { 
// add more here -- must agree with add menu options when creating menu
		OPT_SKYWIN,
		OPT_IFU,
		OPT_DUMMY,
		OPT_EXIT
		};
#endif


/* Window blocks for all widgets */
/* winblock:  xloc,yloc, width,height, label,
	*data, type, scale, decimals, *editwindow, aligntype, laboff  */
/* winblock:  xloc, yloc, width, height, label, *data, type,
	scale, offset, decimals, *edwin, *mwin, align, laboff, callbak, updater*/
winblock  datafile = {
	COL1, Line2,  8, TH, "Observ. File:", NULL, TEXT,  
	1.0, 0.0, 1, &dfwin, &mwin, Title, 0, TextBack, TextUpdate };
// Try width 8 with 1 decimal...  that seems to work...
// The old kludge was to specify a width of 9

winblock  observer = {
	COL5, Line2, 16, TH, "Observer:", NULL, TEXT,
	1.0, 0.0, 1, &obwin, &mwin, Leftside, 0, TextBack, TextUpdate };
// Add decimal to text window if window sometimes comes out too small...
// Was COL4 and Title location

winblock  titlefld = {
	COL1, Line3,  48, TH, "Title:", NULL, DTEXT,
	1.0, 0.0, 0, &ttlwin, &mwin, Title, 0, TextBack, TextUpdate };

winblock  RAfield = {
	COL1, Line5,  9, TH,  "R.A.", NULL, DMS,
	Hour, 0.0, 3, &alwin, &mwin, Title, 0, TextBack, TextUpdate };

winblock  DECfield = {
	COL1, Line6,  10, TH,  "Dec.", NULL, DMS,
	Degree, 0.0, 2, &dewin, &mwin, Title, 0, TextBack, TextUpdate };

#ifdef  IFUM
#define  Uline1  5
winblock  URAfield = {
	COL1, Uline1,  9, TH,  "R.A.", NULL, DMS,
	Hour, 0.0, 3, &aobwin, &IFUwin, Title, 0, TextBack, TextUpdate };

winblock  UDECfield = {
	COL1, Uline1+LH,  10, TH,  "Dec.", NULL, DMS,
	Degree, 0.0, 2, &dobwin, &IFUwin, Title, 0, TextBack, TextUpdate };

wble*	IFUwinlist=NULL;	// A global pointer
#endif  // on IFUM

winblock  slitpaf = {
//	COL4-6*Xp
	COL5, Line6, 4, TH, "Slit Pos. Ang.", NULL, FTEXT, 
	Degree, 90.0, 1, &spawin, &mwin, Leftside, 0, TextBack, TextUpdate };

winblock  tmpfield = {
	COL1, Line8,  3, TH, "Temperature", NULL, FTEXT, 
	1.0, 0.0, 1, &temwin, &mwin, Title, 0, TextBack, TextUpdate };

winblock  hafield = {
	COL5, Line8,  5, TH, "H.A.", NULL, FTEXT, 
	1.0, 0.0, 1, &hawin, &mwin, Leftside, 0, TextBack, TextUpdate };

winblock  ordfield = {
	COL5-2*Xp, Line9, 1, TH, "Order:", NULL, ITEXT, 
	1.0, 0.0, 1, &ordwin, &mwin, Rightside, 0, TextBack, TextUpdate };
// order was at COL5 Rightside; then at COL3 Title; and COL4+2*Xp Leftside

winblock  cwfield = {
	COL4, Line11, 5, TH, "Center Wavelength:", NULL, FTEXT, 
	1.0, 0.0, 1, &cwwin, &mwin, Title, 0, TextBack, TextUpdate };

//	COL6, Line11, 5, TH, "Center Wavelength:",
//	NULL, FTEXT, 1.0, 1, &cwwin, Leftside };

winblock  swidfield = {
	COL1, Line13, 2, TH, "Width:", NULL, FTEXT, 
	1.0, 0.0, 2, &swidwin, &mwin, Title, 0, TextBack, TextUpdate };

winblock  slafield = {
	COL3, Line13, 2, TH, "Lengths:", NULL, FTEXT, 
	1.0, 0.0, 1, &slawin, &mwin, Rightside, 0, TextBack, TextUpdate };
//	COL3, Line13, 2, TH, "Lengths:",	// was COL2+5*Xp
//	NULL, FTEXT, 1.0, 1, &slawin, Rightside };	// Was Title

winblock  slbfield = {
	COL3+2, Line13, 2, TH, "--", NULL, FTEXT, 
	1.0, 0.0, 1, &slbwin, &mwin, Title, 0, TextBack, TextUpdate };
// was COL4-1*Xp

winblock  dcafield = {
	COL3, Line14, 2, TH, "Uncut:", NULL, FTEXT, 
	1.0, 0.0, 1, &dcawin, &mwin, Rightside, 0, TextBack, TextUpdate };	
// was COL2+5*Xp    // Was Title

winblock  dcbfield = {
	COL3+2, Line14, 2, TH, "--", NULL, FTEXT, 
	1.0, 0.0, 1, &dcbwin, &mwin, Title, 0, TextBack, TextUpdate };
// was COL4-1*Xp

winblock  ovlfield = {
	COL6, Line13, 3, TH, "Overlap pixels:", NULL, FTEXT, 
	-1.0, 0.0, 2, &ovlwin, &mwin, Leftside, 0, TextBack, TextUpdate };

winblock  sltfield = {
	COL5, Line13, 2, TH, "Tilt:", NULL, FTEXT, 
	Degree, 0.0, 2, &sltwin, &mwin, Leftside, 0, TextBack, TextUpdate };

// exordfld orphaned, use name for xordbbk
//winblock  exordfld = {
//	COL6, Line13, 2, TH, "ExOrd",
//	NULL, ITEXT, 1.0, 1, &exordw, Leftside };

winblock  rwidfield = {
	COL1, Line15, 3, TH, "Align Hole Size:", NULL, FTEXT, 
	1.0, 0.0, 1, &rwidwin, &mwin, Title, 0, TextBack, TextUpdate };

winblock  robfield = {
	COL5, Line15, 2, TH, "Repeat Obj:", NULL, ITEXT, 
	1.0, 0.0, 1, &repob, &mwin, Leftside, 0, TextBack, TextUpdate };

winblock  rrefield = {
	COL6, Line15, 2, TH, "Ref:", NULL, ITEXT, 
	1.0, 0.0, 1, &reprf, &mwin, Leftside, 0, TextBack, TextUpdate };

// NOTE -- WE WILL CHANGE REFSEL TO REFLIMIT IN THIS WINDOW
#ifdef  REFLIM
winblock  rsefield = {
	COL6, Line16, 1, TH, "Ref. Limit:", NULL, ITEXT, 
	1.0, 0.0, 1, &refsv, &mwin, Leftside, 0, TextBack, TextUpdate };
#else
winblock  rsefield = {
	COL6, Line16, 1, TH, "Ref. Select:", NULL, ITEXT, 
	1.0, 0.0, 1, &refsv, &mwin, Leftside, 0, TextBack, TextUpdate };
#endif
// Previous (alternate) position for repeat and refsel was
// on lines 14 and 15.  Also can fit on linew 7, 8.
// may move them back, and put date time info on line 7,8
// and warning about rotator angle?

#ifdef  EXTPRI
winblock  xpfield = {
	COL3, Line17, 2, TH, "Prio. Sup:", NULL, FTEXT, 
	1.0, 0.0, 2, &xpwin, &mwin, Leftside, 0, TextBack, TextUpdate };
#endif

/* winblock:  xloc, yloc, width, height, label, *data, type,
	scale, offset, decimals, *edwin, *mwin, align, laboff, callbak*/

winblock  mhpfield = {
	COL1, Line16, 3, TH, "MustHave:", NULL, FTEXT,
	1.0, 0.0, 1, &mhwin, &mwin, Title, 0, TextBack, TextUpdate };

winblock  pdfield = {
	COL3, Line16, 3, TH, "Pdecide:", NULL, FTEXT,
	1.0, 0.0, 1, &pdwin, &mwin, Leftside, 0, TextBack, TextUpdate };

/* And, a list of windows... */
wble*	winlist = NULL;

/* The winlist needs to be a global because it is referenced in the
  callback routine for an automatic window, and also in the handler
  routines for automatic windows.  It was previously local to the
  main loop program... */

/* Button blocks for button widgets */
/* butblock:  xloc,yloc, width,height, label, *data,
	*button, *mwin, aligntype, laboff, callback  */

//int	stflag = 0;
//butblock  stbfield = {					//stbut
//	COL5, Line14, 2, WH, "Slit/Hole", &stflag, 
//	&slitype, &mwin, Leftside, 0, NULL };
//// Was Line13
//// Above is orphan, make a new handler for it if used.

butblock  slxfld = {
	COL5, Line14, 2, WH, "Extend Slits", NULL, 
	&slext, &mwin, Leftside, 0, ToggleButton };	// slit extension

butblock  drefld = {
	COL5, Line17, 2, WH, "Dif. Ref. ", NULL, 
	&drefbut, &mwin, Leftside, 0, ToggleButton };	// dif refract 

butblock   adcfld = {
	COL5, Line8, 2, WH, "No ADC  ", NULL, 
	&noadc, &mwin, Leftside, 0, ToggleButton };  // ToggleButton

static int	nobj = 0;
butblock  objlfld = {
	COL1, Line19, 1, WH, "Object Files", &nobj, 
	&objlist, &mwin, Leftside, 0, handle_button };

butblock  tbfield = {
	COL7, Line19, 0, WH, "Test", NULL, 
	&testbut, &mwin, Rightside, 0, handle_button };   // Test

butblock  actfield = {
	COL3, Line1, 0, WH, "MaskGen", NULL,
	&actbut, &mwin, Leftside, 0, handle_active };
// Either col3 or col4 works

butblock  killfld = {
	COL7, Line1, 1, WH, "Exit", NULL,
	&kilbut, &mwin, Rightside, 0, handle_button };	// kill

#ifdef  NSROT
// button block for N&S rotation feature
// line 17 col 4+ or so
// check box similar to extend slits
butblock  nsrfld = {
	COL5, Line9, 2, WH, "N & S mode", NULL, 	// Was Line17
	&nsrot, &mwin, Leftside, 0, ToggleButton };	// N&S rot. bias
// Previous label was "N&S rotation"
#endif

/* And, a list of buttons... */
bble*	butlist = NULL;

/* Menu blocks for all menu widgets (instruments, dispersers, filters) */

// enum instruments { INST_sel, INST_Ilc, INST_Isc,
enum instruments { INST_sel, 

// imacs f/2 things
		INST_Isc,
// To include f/2 gismo, put the below thing above and match instme data below
//#ifdef  GISMOF2
#if  defined GISMOF2 && defined GismoMenus
		INST_Gss,	// GISMO with f/2
//#ifdef  GISMOAP
//		INST_Gsas,	// like the Gsla long camera case
//#endif
#endif

// imacs f/4 things
		INST_Ilc,
//#ifdef  GISMO
#ifdef  GismoMenus
		INST_Gsl,	// GISMO with f/4
#ifdef  GISMOMNS
		INST_Gslx,	// GISMO at f/4 with macro n&s
#endif
#ifdef  GISMOAP
		INST_Gsal,	// For Gismo aperature long
#endif
#endif  // on GISMO -- GismoMenus

// other things
		INST_Lc, 
#ifdef  WFCCD
		INST_wfcd,
#endif
		INST_dummy, INST_Ild, INST_Isd,
		INST_Ld, 

//  SYNC_HERE
		INST_end, INST_tot };
// the above enum MUST sync to the instme data below


enum  dispersers  { DISP_sel,		// Inst-1, long cam (0)
		DISP_r150, DISP_r300, DISP_r600, DISP_r1200,
#ifdef  DMOE
		DISP_moe,
#else
		DISP_nomoe,
#endif
		DISP_rz, DISP_rx,
		DISP_sel1,		// Inst-2, short cam (5)
		DISP_p150, DISP_p200, DISP_p300, DISP_p600, DISP_p300r,
		DISP_pz, DISP_px,
		DISP_sel2,		// Inst-3, ldss (11)
// CHANGE THESE NAMES...
		DISP_s150, DISP_s300b, DISP_s300r, DISP_s600,
		DISP_vphr, DISP_vphb,  DISP_vpha,
		DISP_sz, DISP_sx,
#ifdef	WFCCD
		DISP_sel3,	// Inst-4, wfccd
		DISP_wf1,	// Add wfccd dispersers here
#endif
//#ifdef  GISMO
#ifdef  GismoMenus
		DISP_selG,
		DISP_r150g, DISP_r300g, DISP_r600g, DISP_r1200g,
// Note the omission of any MOE disperser at this point for gismo.
#ifdef  GISMOE
		DISP_gmoe,
#else
		DISP_nogmoe,
#endif
// All these names are really place-keepers to keep count right.
		DISP_rzg, DISP_rxg,
		DISP_selGs,		// Inst-2, short cam (5)
		DISP_p150g, DISP_p200g, DISP_p300g, DISP_p600g, DISP_p300gr,
		DISP_pzg, DISP_pxg,
#endif  // on GISMO -- GismoMenus
		DISP_end, DISP_tot };		// last, 11 or 20.
// the above enum MUST sync to the dispme data below
// No longer synced; thete is no dispme any longeer...
/* The DISP_sel, DISP_sel1, DISP_sel2 and DISP_end index values are
  REQUIRED to match the dispme subscripts; others are placeholders */

enum  disptypes { DTYPE_unknown, DTYPE_grating, DTYPE_grism };

// Instruments:
/* IMenuEntry:  name, text, order, type, flag  */
/*  
  .. name is item name in the opticdef.dat file
  .. text is in pull-down
  .. order is the order to use on that instrument
  .. type is from enum (unknown, grating, grism...)
  .. flag is to disable the selection in the pulldown.
  */
// #ifdef  LDSS_NEW
// #define  Ninsts 10
// #else
// #define  Ninsts 8
// #endif
#define  Ninsts INST_tot
IMenuEntry	instme[Ninsts] = {
	{" ",        "-Select-",    0, DTYPE_unknown, CBX_DISABLED},

// f/2 items
	{"IMACS_sc", "IMACS_f/2",   1, DTYPE_grism,   0},
// If desired as a gismo feature, put the below section above next to the
// f/4 version of gismo, and change instme section above to match
//#ifdef  GISMOF2
#if  defined GISMOF2 && defined GismoMenus
	{"IMACS_sc", "GISMO_f/2",   1, DTYPE_grism, 0},	// or "GISMO_f/2"
//#ifdef	GISMOAP
//	{"IMACS_gds", "GISMO f/2AP",   0, DTYPE_grating, 0},  
//#endif
#endif

// f/4 items
	{"IMACS_lc", "IMACS_f/4",   1, DTYPE_grating, 0},
//#ifdef  GISMO
#ifdef  GismoMenus
//	{"IMACS_lc", " GISMO",   1, DTYPE_grating, 0},	// or "GISMO_f/4"
	{"IMACS_lc", "GISMO_f/4",   1, DTYPE_grating, 0},	// or "GISMO_f/4"
#ifdef  GISMOMNS
	{"IMACS_xlc", "GISMO MNS",   1, DTYPE_grating, 0},  // or "GISMO_f/4"
#endif
#ifdef	GISMOAP
	{"IMACS_gdl", "GISMO AP",   0, DTYPE_grating, 0},  // or "GISMO_f/4"
#endif
#endif  // on GISMO -- GismoMenus

// Others
//	{"LDSS:NEW",  " LDSS",       1, DTYPE_grism,   0},
	{"LDSS3",    " LDSS",       1, DTYPE_grism,   0},
#ifdef	WFCCD
	{"WFCCD",   "WFCCD",      1, DTYPE_grism,  0},
#endif

	{"",         "",            0,  0,       0},  /* Drop direct inst's */
/* Note that entries below this are effectively dropped by the above null entry */
	{"IMACS_ld", "f/4-direct",  0, DTYPE_grating, 0},
	{"IMACS_sd", "f/2-direct",  0, DTYPE_grism,   0},
	{"LDSS3_D",  "LDSS-direct", 0, DTYPE_grism,   0},

//  SYNC_HERE
	{"",         "",            0,  0,            0}  };
#define  maxinstxt  11


// See the definition of dispindx below the menu data



// Dispersers, Gratings, Grisms:
/* IMenuEntry:  name, text, order, type, flag  */
// #ifndef  LDSS_NEW
// #define  Ndisps  16
// #else
// #define  Ndisps  26
// #endif

/*  The new way disperser menu set  */

/* The disperser menu for IMACS long camera gratings, and maybe MOE */
IMenuEntry	disp_If4[8] = {
	{" ",                 "-Select-",    0, DTYPE_unknown, CBX_DISABLED},
	{"IMACS_150A",        "Grating 150", 1, DTYPE_grating,  0},
	{"IMACS_300A",        "Grating 300", 1, DTYPE_grating,  0},
	{"IMACS_600A",        "Grating 600", 1, DTYPE_grating,  0},
	{"IMACS_1200A",	      "Grat'g 1200", 1, DTYPE_grating,  0},
#ifdef  DMOE
	{"MOE",	              " M.O.E.    ", 1, DTYPE_grating,  0},
#else
	{" ",	              "           ", 0, DTYPE_unknown,  0},
#endif
	{"IMACS_direct",      "Dir. Mirror", 0, DTYPE_grating,  0},
	{"",                   "",           0,  0,             0}  };

/* The disperser menu for IMACS short camera grisms  */
IMenuEntry	disp_If2[8] = {
	{" ",                  "-Select-",   0, DTYPE_unknown, CBX_DISABLED},
	{"IMACS_grism_150",    "Grism 150",  1, DTYPE_grism,    0},
	{"IMACS_grism_200",    "Grism 200",  1, DTYPE_grism,    0},
	{"IMACS_grism_300",    "Grism 300",  1, DTYPE_grism,    0},
	{"IMACS_grism_600",    "Grism 600",  1, DTYPE_grism,    0},
	{"IMACS_grism_300R",   "Grism 300R", 1, DTYPE_grism,    0},
	{"IMACS_direct_grism", "No Grism",   0, DTYPE_grism,    0},
	{"",                   "",           0,  0,             0}  };

/* The disperser menu for the LDSS instrument grisms  */
IMenuEntry	disp_Lds[10] = {
	{" ",                  "-Select-",   0, DTYPE_unknown, CBX_DISABLED},
	{"LDSS3_Low",          "Grism low",  1, DTYPE_grism,    0},
	{"LDSS3_Med_blue",     "Grism blu",  1, DTYPE_grism,    0},
	{"LDSS3_Med_red",      "Grism red",  1, DTYPE_grism,    0},
	{"LDSS3_High_blue",    "Grism high", 1, DTYPE_grism,    0},
	{"LDSS3_VPHred",       "VPH Red",    1, DTYPE_grism,    0},
	{"LDSS3_VPHblue",      "VPH Blue",   1, DTYPE_grism,    0},
	{"LDSS3_VPHall",       "VPH All",    1, DTYPE_grism,    0},
	{"LDSS_direct_grism",  "No Grism",   0, DTYPE_grism,    0},
	{"",                   "",           0,  0,             0}  };

/* Disperser menu for WFCCD dispersers  */
IMenuEntry	disp_WFC[3] = {
	{" ",                  "-Select-",   0, DTYPE_unknown, CBX_DISABLED},
	{"WFCCD_G1",           "Grism",      1, DTYPE_grism,    0},
	{"",                   "",           0,  0,             0}  };

/* A special disperser menu for Gismo selected for instrument MOE */
IMenuEntry	disp_Gis[8] = {
	{" ",                  "-Select-",   0, DTYPE_unknown, CBX_DISABLED},
	{"IMACS_150A",        "Grating 150", 1, DTYPE_grating,  0},
	{"IMACS_300A",        "Grating 300", 1, DTYPE_grating,  0},
	{"IMACS_600A",        "Grating 600", 1, DTYPE_grating,  0},
	{"IMACS_1200A",	      "Grat'g 1200", 1, DTYPE_grating,  0},
#ifdef  GISMOE
	{"MOE",	              " M.O.E.    ", 1, DTYPE_grating,  0},
#else
	{" ",	              "           ", 0, DTYPE_unknown,  0},
#endif
	{"IMACS_direct",      "Dir. Mirror", 0, DTYPE_grating,  0},
	{"",                   "",           0,  0,             0}  };

/* A special disperser menu entry with no dispersers */
IMenuEntry	disp_Zro[3] = {
	{" ",                  "-Select-",   0, DTYPE_unknown, CBX_DISABLED},
	{"IMACS_direct",      "Dir. Mirror", 0, DTYPE_grating,  0},
	{"",                   "",           0,  0,             0}  };

/* A NULL disperser menu when no disperser is known */
IMenuEntry	disp_Nul[1] = {
	{"",                   "",           0,  0,             0}  };

#define  maxdsptxt  11


// This definition had to be put below the menus referenced
// herein, so things are slightly out of order here...

// This array must synchronize with the instrument array, but has pointers
// to separate disperser arrays initialized separately.
IMenuEntry*	dispindx[Ninsts+1] = {
	NULL,		// unknown

// f/2 things
	disp_If2,	// IMACS short camera
//#ifdef  GISMOF2
#if  defined GISMOF2 && defined GismoMenus
	disp_If2,	// imacs with gismo f/2
// may need to change for gismo with imacs and f/2 ??
#endif

// f/4 things
	disp_If4,	// Imacs long cameera
//#ifdef  GISMO
#ifdef  GismoMenus
	disp_Gis,	// IMACS with gismo
#ifdef  GISMOMNS
	disp_Gis,	// Second gismo entry for MNS
#endif
#ifdef  GISMOAP
	disp_Zro,	// Third gismo entry for AP window
// If we use disp_Nul above, we get "?" in disperser menu window; using
// disp_Gis gives a "-Select-" there which looks more consistent.
// Special disp_Zro is only a select entry and no dispersers
#endif
#endif	// on GISMO -- GismoMenus

// other things
	disp_Lds,
#ifdef	WFCCD
	disp_WFC,
#endif
	disp_Nul,
///
	disp_If4, disp_If2,
	disp_Lds,
//#ifdef  GISMO
//	disp_Gis,	// IMACS with gismo
//#endif
//  SYNC_HERE
	disp_Nul,
	NULL };

/* Global to track which disperser menu is up */
// static	int	Ityp = DTYPE_unknown;

/* Should revise the gratme and grisme to a single array, and
  index the entry into that array by the instrument index using
  this table... */

/* Menu Block:  xloc,yloc, width,height, label, *data,
	type, *menu, *mwin, align, laboff, callback, tcolor */

mnublock  iqfield = {
	COL1, Line1, 10, WH, "File", NULL,
	M_ADROP, &iqmenu, &mwin, Leftside, 0, handle_iqmenu, 0 };

#ifdef  SKYOPT
mnublock  OPTfield = {
	COL2a, Line1, 8, WH, "Options", NULL,
	M_ADROP, &optmenu, &mwin, Rightside, 0, handle_optmenu, 0 };
//	COL2, Line1, 8, WH, "Options", NULL,
//	M_ADROP, &optmenu, &mwin, Leftside, 0, handle_optmenu, 0 };
#endif

mnublock  eqfield = {
	COL5, Line5, 8, WH, "Equinox:", NULL,
	M_POP, &eqmenu, &mwin, Leftside, 0, NULL, 0 };  // no handler yet
//was..
//	COL4, Line5, 8, WH, "Equinox:", NULL,
//	M_POP, &eqmenu, &mwin, Title, 0, NULL, 0 };  // no handler yet
//was..
//	COL5, Line5, 8, WH, "Equinox:", NULL,
//	M_POP, &eqmenu, &mwin, Leftside, 0, NULL, 0 };  // no handler yet
// COL5 and Leftside aligns slightly left of COL4 and Title.

mnublock  exordfld = {
	COL6+3*Xp, Line11a, 9, TH, "Ex.Ord:", NULL,
	M_POP, &xordmu, &mwin, Rightside, 0, NULL, 0 };  // handle_exmenu };

mnublock  instfield = {
	COL1, Line10,  maxinstxt, TH, "Instrument:", NULL, 
	M_POP, &instmu, &mwin, Title, 0, NULL, 0 };


mnublock  dispfield = {
	COL5, Line10,  maxdsptxt+1, TH, "Disperser:", NULL, 
	M_POP, &dispmu, &mwin, Leftside, 0, NULL, 0 };
//	COL4, Line10,  maxdsptxt, TH, "Disperser:",
//	NULL, M_GRAT, &dispmu, Title };

#ifdef  GAPVO
mnublock  gapfield = {
	COL1, Line17,  6, TH, "Gaps:", NULL, 
	M_POP, &gapmu, &mwin, Title, 0, NULL, 0 };
#endif

/* Things for filter and passband menu items... */
mnublock  filtfield = {
	COL1, Line11,  0, TH, "Filter:", NULL, 
	M_POP, &filtmu, &mwin, Title, 0, NULL, 0 };

/* Filter Names and Passbands; data here is to be updated! */

filtelem  filters [10] = {
	{"Clear",	{ 3500.0, 9700.0 } },
	{"Blu",		{ 3700.0, 4900.0 } },
	{"Vis",		{ 4900.0, 6050.0 } },
	{"Red",		{ 5600.0, 7600.0 } },
	{"I-Red",	{ 7150.0, 8850.0 } },
	{"W-Blu",	{ 4800.0, 7900.0 } },
	{"W-Red",	{ 6330.0, 9650.0 } },
	{"N.B.",	{ 8100.0, 8260.0 } },
	{"",		{ 0.0, 0.0 } }
	};

/* was...
	{"clear",	{ 3500.0, 9500.0 } },
	{"visual",	{ 4100.0, 7900.0 } },
	{"IR",		{ 8000.0, 11000.0} },
	{"UV",		{ 3500.0, 4100.0 } },
	{"",		{ 0.0, 0.0 } }
	};
.. */

IMenuEntry	filtme[10];	// Dynamically formed in MakeFiltmenu

#ifdef  GUIDATE
/* Menus for selection of date */

//#define  GDbase  COL2a-2
#define  GDbase  COL2a
#define  GDgap   2*Xp
// Above used to bind these 3 menus together across line.

mnublock  yearfield = {
	GDbase-GDgap, Line8,  5, TH, "Obs. Date:", &yr, 
	M_APOP, &yearmnu, &mwin, Rightside, 0, handle_datemnu, 0 };

mnublock  monthfield = {
	GDbase, Line8,  5, TH, " ", &mo, 
	M_APOP, &monthmnu, &mwin, Leftside, 0, handle_datemnu, 0 };

mnublock  dayfield = {
	GDbase+4*GDgap-1, Line8,  3, TH, " ", &dy, 
	M_APOP, &daymnu, &mwin, Leftside, 0, handle_datemnu, 0 };
// ADD CALLBACK TO HANDLE THESE MENUS AS DATES...
// ADD DATA POINTER TO YMD FOR THESE INSTEAD OF NULL
// may need those back to null for handling in handler??
/* Menu Block:  xloc,yloc, width,height, label, *data,
	type, *menu, *mwin, align, laboff, callback */

#endif  // on GUIDATE

#ifdef  OBJSHAPE
/* Menu block for object slit shape */

mnublock  ssfield = {
	COL1, Line14,  5, TH, "Shape:", NULL, 
#ifdef  OBJAUTO
	M_APOP, &shapemnu, &mwin, Title, 0, handle_shape, 0 };
#else
	M_POP, &shapemnu, &mwin, Title, 0, NULL, 0 };
#endif

#endif  // on OBJSHAPE

/* And... a list of menus */
mble*	mnulist = NULL;

/* The subsidiary windows, with passband information */
// NOTE - adjust x position (first arg) for these!
winblock  blupbfld = {
	COL2a, Line11,  5, TH, " ", NULL, FTEXT, 
	1.0, 0.0, 0, &blupb, &mwin, Rightside, 0, TextBack, TextUpdate };
//	COL2+2*Xp, Line11,  5, TH, " ",
//	NULL, FTEXT, 1.0, 0, &blupb, Title };

winblock  redpbfld = {
	COL2a+2, Line11,  5, TH, "--", NULL, FTEXT, 
	1.0, 0.0, 0, &redpb, &mwin, Title, 0, TextBack, TextUpdate };
//	COL3-4*Xp, Line11,  5, TH, "--",
//	NULL, FTEXT, 1.0, 0, &redpb, Title };

/* The windows for wavelength range used for detector test */
winblock  bludbfld = {
	COL2a, Line11a,  5, TH, "Detector range:", NULL, FTEXT, 
	1.0, 0.0, 0, &bludb, &mwin, Rightside, 0, TextBack, TextUpdate };

winblock  reddbfld = {
	COL2a+2, Line11a,  5, TH, "--", NULL, FTEXT, 
	1.0, 0.0, 0, &reddb, &mwin, Title, 0, TextBack, TextUpdate };


/*  NOTE  - - - - - -
  To simplify some of the menu, window, etc. entries here, we could
  introduce an interface routne when calling create-xxx which would
  supply the proper scale, offset, main window, laboff and callback
  when those fields are NULL in the data.
  - - - - - - - - - - - -  */

winblock*  findwb (EditWindow* ew)
/*  Using the global list(s), find the winblock which points to
  a given edit window, and return the pointer to that winblock.  */
{
wble	*lp;	// Search pointer
winblock  *wb;	// Result pointer

/*  Try looking at standard global list first.  */
    C_loop (lp, winlist) {
	wb = lp->blk;
	if (wb->ew == ew) return  wb;
    }

/* If needed, and defined, look similarly in secondary list(s) */
#ifdef  IFUM
    C_loop (lp, IFUwinlist) {
	wb = lp->blk;
	if (wb->ew == ew) return  wb;
    }
#endif  // on IFUM

/* If not found, we must return NULL */
    return  NULL;
}

/*  --- NOTE --- The guisubs.c file is created to allow xmap to be
  integrated into intgui.  Its content is used in parallel by the
  xmap main program and intgui itself.  The code will return here
  when xmap is integrated into intgui and ceases to be a stand-alone
  program using guisubs.c as an include.   */

/* = = = = = = = = = = = = = */
#include	"guisubs.c"
/* = = = = = = = = = = = = = */

/* following things were removed from guisubs.c, and their places
  noted for future return when reassembling things. */


/*  ------------------------------------------------------------  */

#ifdef  SKYOPT
static void   handle_optmenu(void* param)
{
// int	new;
Menu	*mnu=(Menu*)param;

// (now auto)    new = CBX_HandleMenu(&optmenu);
// Going to auto menu, no need to do handlemenu.

//    printf (" ?Handle option %d?\n", new);
//    switch (new) {
//    switch (optmenu.sel) {
    switch (mnu->sel) {		// *mnu is optmenu, for lint
	case  OPT_SKYWIN :
//	    printf (" Option - toggle the sky window - ");
	    if (optmenu.entry[OPT_SKYWIN].flag & CBX_CHECKED) {
		skywin_active = 0;
//		printf ("off");
	    } else {
		skywin_active = 1;
//		printf ("on");
	    }
	    CBX_CheckMenu (&optmenu, OPT_SKYWIN, skywin_active);
//	    printf (".\n");
	    break;
#ifdef  IFUM
	case  OPT_IFU :
	    if (optmenu.entry[OPT_IFU].flag & CBX_CHECKED) {
		IFU_active = 0;
//		printf ("off");
	    } else {
		IFU_active = 1;
//		printf ("on");
	    }
	    CBX_CheckMenu (&optmenu, OPT_IFU, IFU_active);
	    Gob->changed |= 1;
	    break;
#endif
	case  OPT_EXIT :
//	    printf (" Option - end it all.\n");
	    done = True;
	    break;
	case  -1:
//	    printf (" Option - non-select.\n");
	    break;
	default :
	    printf (" ?? How did I get here? (Option %d)\n", optmenu.sel);
	    break;
    }
}
#endif

// the val:  optmenu.entry[OPT_SKYWIN].flag & CBX_CHECKED
// should equal skywin_active; simply do:
//  CBX_CheckMenu (&optmenu, OPT_SKYWIN, skywin_active);
// When skywin could possibly have changed!

/*  ------------------------------------------------------------  */
#ifdef  GUIDATE
static void   setup_dates (void)
{
int	days[13] = { 31,
	31, 28, 31,  30, 31, 30,  31, 31, 30,  31, 30, 31 };
/* Based on global values for yr, mo, dy and mjd, and the number
  of days in month, set up the menu selectors for the date menus
  and set n in the daymenu to days in month plus one. */

/* First, key off the mjd value which is supposed to be correct */
// no -- we only set mjd based on valid ymd in menu handler
    if (mjd != 0) {
	date4mjd (mjd, &yr, &mo, &dy);
	yearmnu.sel = (yr > baseyear && yr < baseyear + yearmnu.n) ?
		yr - baseyear : 0;
	monthmnu.sel = mo;
	daymnu.sel = dy;
    }
    daymnu.n = days[mo] + 1;
    if (yr%4 == 0 && mo == 2) daymnu.n = 30;

//  printf (" - setup_dates mjd = %d, cmjd = %d\n", mjd, cmjd);  // debug
//  fflush (stdout);		// debug

/*  Set the color of yearfld based on date  */

/* Omit this section if app is not (yet) defined */
if (app != NULL && cmjd > 0) {
    if      (mjd <= 0)      yearfield.tcolor = app->black;
    else if (mjd < cmjd+21) yearfield.tcolor = app->red;
    else if (mjd < cmjd+42) yearfield.tcolor = app->yellow;
    else                    yearfield.tcolor = app->green;
}

// NOTE THAT WE ASSUME 0 (or black) TO BE THE DEFAULT COLOR HERE!

// printf (" - setup_dates mjd = %d\n", mjd);  // debug

}

// NOTE that this is within an  #ifdef GUIDATE
static void   handle_datemnu (void* param)
{
// int	new;
Menu	*mnu=(Menu*)param;
int	xmjd;

/* BORROWED from optmenu above.  We do these things here:
  Find which menu year-month-day we have, and set the
  proper global value to the menu select value.
  Note if this is a change in the value.
  Check the date.  If we have selected a 0, the date must become 0.
  Otherwise, see if we have all good values, then compute mjd.
  When mjd is computed or set call for ra/dec computation too (solpos).
  If we had a valid mjd and we selected 0, other memory selectors
  need to become 0 also.  If we selected a valid value, set that
  for the proper integer.  Be sure to add base year in that case.
  When the month changes, select an appropriate ending value for
  the date menu.  Month 0 has 31 days; month 2 has days which depend
  on the year value mod 4.  If current day value is beyond the month,
  set it right before proceeding.  */

    xmjd = mjd;
    if (mnu == &yearmnu) {
	if (mnu->sel == 0) {
	    yr = 0;
//	    if (mjd != 0) mo = dy = 0;
	} else {
	    yr = mnu->sel + baseyear;
	}
    } else if (mnu == &monthmnu) {
	if (mnu->sel == 0) {
	    mo = 0;
//	    if (mjd != 0) yr = dy = 0;
	} else {
	    mo = mnu->sel;
	}
    } else if (mnu == &daymnu) {
	if (mnu->sel == 0) {
	    dy = 0;
//	    if (mjd != 0) yr = mo = 0;
	} else {
	    dy = mnu->sel;
	}
    } else {
	printf (" ??*? Unknown date menu!\n");
	return;
    }
    if (yr > 0 && mo > 0 && dy > 0) mjd = mjdate (yr, mo, dy);
    else  mjd = 0;
    if (xmjd != mjd) {
	Gob->changed |= 1;
	warnflag |= 1;
    }
//    solpos (mjd);
// The position routine is now called when warnflag is raised.
    setup_dates();
    CBX_UpdateMenu(mnu);	/* Update needed for MacOSX */
}
#endif  // on GUIDATE

/*  ------------------------------------------------------------  */
static void   handle_exmenu(void)
{
int	new;
int	mask;
// (now auto)    new = CBX_HandleMenu(&optmenu);
// Going to auto menu, no need to do handlemenu.

    new = CBX_HandleMenu(&xordmu);
    if (new < 0) return;
    if (new == xordmu.sel) return;	// Null case
    if (new > 2) return;	// or xordmu.n perhaps
//    printf (" Handle exorder menu %d\n", new);	// debug only
    Gob->changed |= (xordmu.sel != new);
    xordmu.sel = new;
/* Set up the actual mask; menu 0,1,2 == mask 0,1,3 */
    mask = new;
    if (mask == 2) mask = 3;
    Gob->obs->ex_order = mask;
    CBX_UpdateMenu(&xordmu);	/* Update needed for MacOSX */
}


#ifdef  OBJSHAPE
/*  ------------------------------------------------------------  */
// change to shape, to handle shape pull-down
// Need to set the shape flag in obs
// if not slit, disable the slit length windows
// if changed to slit, enable the slit length windows
#ifdef  OBJAUTO
static void   handle_shape(void* param)
#else
static void   handle_shape(void)
#endif
{
#ifdef  OBJAUTO
Menu	*mnu=(Menu*)param;
#endif
int	new;
int	old;
int	sc;
// (now auto)    new = CBX_HandleMenu(&optmenu);
// Going to auto menu, no need to do handlemenu.

#ifdef  OBJAUTO
    switch (Gob->obs->default_slit.shape) {
	case RECTANGLE:	old = 0;	break;
	case CIRCLE:	old = 1;	break;
	case SQUARE:	old = 2;	break;
    }
    new = mnu->sel;
#else
    old = shapemnu.sel;
    new = CBX_HandleMenu(&shapemnu);
#endif
//printf (" -*shape old = %d, new = %d\n", old, new);  // debug only
    if (new < 0) return;
    if (new == old) return;	// Null case
    if (new > 2) return;	// or xordmu.n perhaps
//    printf (" Handle shape menu %d\n", new);	// debug only
    Gob->changed |= (old != new);
#ifndef  OBJAUTO
    shapemnu.sel = new;
#endif
/* Set up the actual mask; menu 0,1,2 == RECTANGLE CIRCLE SQUARE */
// The slitshape is CIRCLE - SQUARE - RECTANGLE
// slit struct has shape, width, alen, blen, angle
// When changing to/from RECTANGLE, adjust alen/blen, angle appropriately;
// for circle/square alen=blen=width/2; keep angle
// to rectangle, set alen, blen to default value of 6.0 (or 5.0?)
/* If we change from rectangle, disable windows and set alen/blen */
// from rectangle is old == 0
    if (old == 0) {
// disable alen/blen windows
// set alen/blen as width/2
	Gob->obs->default_slit.alen = Gob->obs->default_slit.width / 2.0;
	Gob->obs->default_slit.blen = Gob->obs->default_slit.alen;
	HandleTextWindow (winlist, &slawin);
	CBX_DisableEditWindow (&slawin, CBX_DISABLED);
	HandleTextWindow (winlist, &slbwin);
	CBX_DisableEditWindow (&slbwin, CBX_DISABLED);
	HandleTextWindow (winlist, &dcawin);
	CBX_DisableEditWindow (&dcawin, CBX_DISABLED);
	HandleTextWindow (winlist, &dcbwin);
	CBX_DisableEditWindow (&dcbwin, CBX_DISABLED);
/* Also clear and disable any slit extension if going to hole/box (slext) */
#ifdef  SLXB
	if (Gob->obs->slext != 0) Gob->obs->slext = 0;
	if (slext.flag == CBX_NORMAL) CBX_DisableButton (&slext, CBX_DISABLED);
#endif

/* If we change to rectangle, enable windows and set default alen/blen */
// to rectangle is new == 0
    } else if (new == 0) {
// enable alen/blen windows
// set alen/blen as default 5.0
	Gob->obs->default_slit.alen = 5.0;
	Gob->obs->default_slit.blen = 5.0;
	HandleTextWindow (winlist, &slawin);
	CBX_EnableEditWindow (&slawin);
	HandleTextWindow (winlist, &slbwin);
	CBX_EnableEditWindow (&slbwin);
	HandleTextWindow (winlist, &dcawin);
	CBX_EnableEditWindow (&dcawin);
	HandleTextWindow (winlist, &dcbwin);
	CBX_EnableEditWindow (&dcbwin);
/* Otherwise, to/from circle and square do not need alen/blen adjustment */
/* Enable slit extension for rectangle (slext) */
#ifdef  SLXB
//      if (Gob->obs->default_slit.shape == RECTANGLE) {
	if (slext.flag == CBX_DISABLED) CBX_EnableButton (&slext);
//      }
#endif
    }

/*  Set the actual shape code indicated by menu option */
    switch  (new) {
    	case 0:		sc = RECTANGLE;		break;
    	case 1:		sc = CIRCLE;		break;
    	case 2:		sc = SQUARE;		break;
    }
    Gob->obs->default_slit.shape = sc;

    CBX_UpdateMenu(&shapemnu);	/* Update needed for MacOSX */
}
#endif  // on OBJSHAPE

/*  ------------------------------------------------------------  */
int	Mselectn (char* name, IMenuEntry* me)
{
IMenuEntry	*p;
int		i;

/* Check first for exact name... */
    for (i=0,p=me; *p->name != AN; p++,i++) {
	if (p->flag) continue;
	if (!strcmp(name, p->name)) return i;
    }

/* Now, check without case... */
    for (i=0,p=me; *p->name != AN; p++,i++) {
	if (p->flag) continue;
	if (!strcasecmp(name, p->name)) return i;
    }

/* Nothing found, return the selector default */
    return  0;
}

/*  ------------------------------------------------------------  */
int	instdisp (int value)
// Return disperser type of given instrument
// does not exist with new LDSS support
{
int	type;
    switch (value) {
	case  INST_Ilc:
	case  INST_Ild:
//#ifdef  GISMO
#ifdef  GismoMenus  // The INST_Gxx things are only defined if this is true
	case  INST_Gsl:
#ifdef  GISMOMNS
	case  INST_Gslx:
#endif
#ifdef  GISMOAP
	case  INST_Gsal:
#endif
#endif  // on GISMO
	    type = DTYPE_grating;
	    break;
	case  INST_Isc:
//#ifdef  GISMOF2
#if  defined GISMOF2 && defined GismoMenus
	case  INST_Gss:  // should not be needed
#endif
	case  INST_Isd:
	case  INST_Lc:
	case  INST_Ld:
	    type = DTYPE_grism;
	    break;
	default:
	    type = DTYPE_unknown;
	    break;
    }
    return  type;
}
/* NOTE  Should replace the above function by a reference to
  the instme array, and the type member; index the array
  by value.  e.g. use  "instme[value].type" in place of function.  */
/* Also NOTE that this function is never used now. */


/*  ------------------------------------------------------------  */


	/* - - - - - - - - - - - *\
	|                         |
	|  IFU  window  programs  |
	|                         |
	\* - - - - - - - - - - - */

/*  ------------------------------------------------------------  */

#ifdef  IFUM

/* Save mode in a global value within the session */
static	int	oldmode = 0;	// Saves mode of current session

void	redraw_IFUwin (void)
{

    if (IFUwin.disp == NULL) return;
    if (!CBX_IsMapped(IFUwin.win)) return;

// draw the windows
// draw the menu -- CBX_UpdateMenu (&xxmenu)
// draw any text

    redraw_windows (IFUwinlist);
//    CBX_UpdateEditWindow (&alwin);	// does not work to update
//    CBX_UpdateEditWindow (&dewin);

    HandleTextWindow (winlist, RAfield.ew);  // That's &alwin there...
    HandleTextWindow (winlist, DECfield.ew);  // That's &dewin there...
// The above updates when changing left/right; not when new ra/dec
// Might instead put these after IFUfindcenter in handle_sidemenu below


// Draw the menu; should draw any text around it if needed...
    CBX_UpdateMenu(&sidemu);

}

/*  ------------------------------------------------------------  */

void	handle_sidemu (void)
{
/* Semi-standard menu handler... */
int	new;
int	mode;
int	changed;

    new = CBX_HandleMenu(&sidemu);
    if (new < 0) return;
    if (new > 1) return;

    mode = new + 1;
    changed = (mode != Gob->obs->IFUmode);
    Gob->obs->IFUmode = mode;
    Gob->changed |= changed;
    oldmode = mode;

    if (changed)  IFUfindcenter (Gob->obs);

    sidemu.sel = new;
    CBX_UpdateMenu(&sidemu);	/* Update needed for MacOSX */
}

/*  ------------------------------------------------------------  */

static void	start_IFUwin (void)
{
// Patterened on start_skywin
// ...

int        x,y;
int        w,h;
//int        sw, sh;
//int        i;
XSizeHints hint;
int	mode;
char	fontname[80];

//Obs	*ob;

double	ac, dc;		// Center location
double	da, dd;		// coordinate differences
double	ab, db;		// Object location
double	dm;		// Mid declination
double	cdel;		// Cosine mid-delta for scale
vect2	p;		// offset vector

 printf (" #- Starting start_IFUwin.\n");  fflush(stdout);	// debug


// Actually, we want to locate just above win, with defined width, height.
// Call that, using "mwin.win" as the window argument.
// Set our origin to same x + 0.5 * (width - our.width)
// and to y - our.height.

/* Set up our window size and location... */
/* Our size */
    hint.width = 400;
    hint.height = 90;
    hint.min_width = hint.width;	hint.max_width = 600;
    hint.min_height = hint.height;	hint.max_height = 300;

/* Our location */
    CBX_GetGeometry (mwin.win, &x, &y, &w, &h);
    hint.x = x + (w - hint.width)/2;
    hint.y = y - hint.height - 55;

    hint.flags = PPosition | PSize | PMinSize | PMaxSize;

/* Simulate the same things as create_skywin here... */
//    create_skywin (hint);

    CBX_OpenMainWindow (
	&IFUwin,
	hint.x + 32000 * hint.y,
	hint.width, hint.height,
	&hint,
	"IFU Mode Locator", "IFU",
	False);

    CBX_SelectInput (&IFUwin, StructureNotifyMask |
		ExposureMask  |  ButtonPressMask );

    strcpy(fontname,genv2("IMACSFONT",DEF_FONT)); /* default font */
    (void)CBX_CreateGfC (&IFUwin, fontname, "bold", FontSize);

// Set the mode, choosing one if none is current.
    mode = Gob->obs->IFUmode;
    if (mode == 0) mode = oldmode;
    if (mode < 1 || mode > 2) {
	mode = 1;
	Gob->obs->IFUmode = mode;
    }
    oldmode = mode;	// And save the mode locally

/* Using the center coordinates and current position angle, we
  compute the offset coordinates of the currently specified IFU
  window, and enter that in the obs data structure.  This computation
  is similar to that in IFUfindcenter, except that it is direct. */

    ac = Gob->obs->ra;
    dc = Gob->obs->dec;
    if (dc < 70.0 * Degree) {	// i.e. non-default position
	p = make2vect (IFUdx[mode-1], IFUdy[mode-1]);
	p = rot2 (mul2vect (p, ArcSecond), Gob->obs->angle);
	dd =  p.y;
	dm = dc + 0.5*dd;
	cdel = cos(dm);
	if (cdel < 0.001) cdel = 0.001;
	da = -p.x / cdel;
	ab = fmodsum (ac, da, TwoPi);
	db = dc + dd;

/* Set the computed position as the object position, unless the
  current object position is within 0.1" in that coordiante. */

	if ((cdel*fabs(Gob->obs->IFUra - ab)) > (0.1 * ArcSecond) )
	    Gob->obs->IFUra  = ab;
	if (fabs(Gob->obs->IFUdec - db) > (0.1 * ArcSecond) )
	    Gob->obs->IFUdec = db;
    }

/* The above computations solve the problem of the initial object
  coordinates in the IFU finder window being a default location.  It
  is skipped if the center is a default location.  Also, if the center
  has been changed in non-IFU mode, we pick up that center when going
  into IFU mode again.  */

/* R.A. and Dec. window creation... */
    AutoEditWin (&IFUwinlist, &URAfield, &(Gob->obs->IFUra) );
    AutoEditWin (&IFUwinlist, &UDECfield, &(Gob->obs->IFUdec) );

// Start the popup menu for left/right selection
// Reference places:  handle_instmenu  sidemu  handle_eqmenu (in guisubs)
// and creation:  Popup  and eqmenu

/* Create the popup menu for left/right side selection */

    (void)CBX_CreatePopupMenu(&IFUwin, &sidemu,
		Xloc(1+22*Xp),		// was 24 - position
		Yloc(Uline1) + Yloc(LH)/2,	// They NEEED to be separate
		Xsize(10*Xp+2),		// was 6 - menu size
		(int)(Vunit*TH+1) );
    CBX_AddMenuEntry (&sidemu, "Up/right",   0);
    CBX_AddMenuEntry (&sidemu, "Down/left",  0);
//    CBX_AddMenuEntry (&sidemu, "Left",   0);
//    CBX_AddMenuEntry (&sidemu, "Right",  0);

// Set the selection based on the current mode.
    sidemu.sel = mode - 1;

/* Call the draw routine for the window... */
    redraw_IFUwin();

/* Disable the RA/Dec input windows for center when IFU window is up */
    CBX_DisableEditWindow (&alwin, CBX_DISABLED);
    CBX_DisableEditWindow (&dewin, CBX_DISABLED);

}

/*  ------------------------------------------------------------  */

void	kill_IFUwin (void)
{

//  Close the main window
    CBX_CloseMainWindow (&IFUwin);

//  Turn off the active flag in global and menu
    IFU_active = 0;
    CBX_CheckMenu (&optmenu, OPT_IFU, IFU_active);

// Make sure the display is set inactive.
    IFUwin.disp = NULL;

// Clean all entries in the IFUwinlist queue...
    while (IFUwinlist != NULL) IFUwinlist = wlpop (IFUwinlist);

// Also, we should re-enable the ra/dec windows
    CBX_EnableEditWindow (&alwin);
    CBX_EnableEditWindow (&dewin);

}


/*  ------------------------------------------------------------  */

void	IFU_handle (int* sleeptime)
{
/* Handles the IFU window; independently of events in main window.
  Senses the active flag and starts IFUwin if needed, or closes it
  if needed.  Handles IFUwin events.  Checks for changes to position
  and updates the display properly.  */
/* Copied from skywin in large part... */


XEvent	event;

static int  ermask=0;				// debug
static int  erlimit=0;				// debug
// static int	mescount = 0;			// debug
char	buffer[256];				// debug

    if (IFUwin.disp == NULL) {		// No display
	if (!IFU_active) return;	// and don't want one
	start_IFUwin ();
    } else {				// Have display
	if (!IFU_active) {		// And don't want it
	    kill_IFUwin ();
	    return;
	}
    }

// This has the loop for the ifu window

/* Passing the sleep time pointer on to the event test */
    if (CBX_AutoEvent(&IFUwin, &event, sleeptime) ) {

	switch (event.type) {			/// loop switch
    
	case KeyPress:                        /* key pressed event   *2*   */
	    break;
    /* key release is *3* */

	case ButtonPress:                    /* button pressed event *4*  */
// Don't care which button it was here...
    /* -- ButtonPress is a mouse button activation -- */
	    if      (event.xbutton.window == sidemu.win) handle_sidemu();

	    break;

	case ButtonRelease:                 /* mouse button released *5*   */
#ifdef DEBUG
	    sprintf(buffer,"unrelated button release");
	    adebug(P_TITLE,buffer);
	    fprintf(stderr,"%s\n",buffer);
#endif
	    break;
    
// #ifdef DEBUG
	case MotionNotify:			/* pointer motion   *6*    */
		    /* ignore */
	    break;
// #endif
    
	case EnterNotify:                      /* window enter event *7*   */
	    break;
	case LeaveNotify:                      /* window leave event *8*   */
	    break;
    
	case Expose:                           /* expose events    *12*    */
	    if (event.xexpose.count > 0) break;     /* ignore */
	    if (event.xexpose.window == IFUwin.win) {
// call to redraw the window
//		redraw_IFUwin();
	    }
	    break;


	case MappingNotify:                       /* keyboard mapping event   */
	    break;
    
	case DestroyNotify:				/* 17 */
	    if (event.xdestroywindow.window == IFUwin.win) {  /* IFU window */
		kill_IFUwin ();
    printf (" *IFU window closed.  Event = %d.\n", event.type);	// debug
	    }
	    break;
    
	case KillNotify:			/* Also for kill... *52*  */
// This is the special CXT "event" detection
	    kill_IFUwin ();
    printf (" *IFU window killed.  Event = %d.\n", event.type);  //debug
	    break;
    
	case UnmapNotify:				/* 18 */
	    break;
    
	case MapNotify:				/* 19 */
	    break;

	case ReparentNotify:			/* 21 */
		    /* Ignored */
	    break;
    
	case ConfigureNotify:                    /* configure event   *22*    */
//    printf (" ..ConfigNotif (22) IFU found.\n");  fflush (stdout);  //debug
//?	    if (event.xconfigure.window == IFUwin.win) config_IFUwin ();
// Like, what other window could it be?
//    printf (" ..ConfigNotif (22) IFU handled.\n");  fflush (stdout);  //debug
	    break;

    /*  Some others...
	    ResizeRequest	25
	    ColormapNotify	32
	    MappingNotify	34
    .. */

	default:
    
//#ifdef DEBUG
	    if (event.type == ermask) {
		if (++erlimit > 4) break;
	    } else {
		ermask = event.type;
		erlimit = 1;	// This is error number 1.
	    }
	    sprintf(buffer,"unknown IFU event '%d'",event.type);
//	    adebug(P_TITLE,buffer);
	    fprintf(stderr,"%s\n",buffer);
//#endif
    /* If get here, see /usr/openwin/share/include/X11/X.h  */
//	    return (1);
	    break;


	}		// End of switch for event

	redraw_IFUwin ();	// Always if any event
    }		// End of event test
}	// End of IFU_handle
#endif	// end IFU conditional


/*  ------------------------------------------------------------  */

extern int   cbx_max  (int,int);

Bool KDC_DialogBox(char *title, char* text, char *answer)
{
MainWindow box;
int	done=0;
int	w, w2;
int	PXh,PXw,XXh;
// int  PAy=PXh/2,OKy=PAy+XXh+PXh,OKw=4*PXw,CNw=6*PXw+PXw/2;
// int  h=OKy+XXh+PXh/2;
int	TXy, PAy, OKy, OKw, CNw;
int	h;
XSizeHints hint;
XEvent     event;
EditWindow pawin;
Button     okbut,cabut;

// dynamic intiialize of font...
    PXh = Vunit * Yp;
    PXw = Hunit * Xp;
    XXh = PXh + Yp;

//    PXw = 10 + (PXh-14)/2;
//    XXh = PXh + 4;
    w = 16*PXw;
    TXy = PXh + PXh/2;	// Height of text
//    PAy=TXy + XXh + PXh/2; 	// Height of edit window
    PAy=TXy + PXh; 	// Height of edit window
    OKy=PAy+XXh+PXh;	// Height of OK buttons
    OKw=4*PXw; 
    CNw=6*PXw+PXw/2;
    h=OKy+XXh+PXh/2;

  w = cbx_max(w,PXw*(strlen(title)+3));
  w = cbx_max(w,PXw*(strlen(answer)+15));
  w = cbx_max(w,PXw*(strlen(text)+1));

  (void)CBX_OpenMainWindow(&box,CBX_CENTER,w,h,&hint,title,"ParBox",True);
  w = cbx_max(w,CBX_TextWidth(&box,title)+3*PXw);
  w = cbx_max(w,CBX_TextWidth(&box,answer)+15*PXw);
  w = cbx_max(w,w2=CBX_TextWidth(&box,text)+PXw);

  CBX_MoveResizeMainWindow(&box,CBX_CENTER,w,h);
 
  (void)CBX_CreateEditWindow(&box,&pawin,PXw,PAy, w-2*PXw,XXh, 2,answer,0);
  (void)CBX_CreateButton(&box,&okbut,w/4-OKw/2,OKy, OKw,XXh, "OK");
  (void)CBX_CreateButton(&box,&cabut,(2*w)/3-CNw/2,OKy,CNw,XXh,"cancel");
 
  while (done == 0) {                           /* event loop */
    (void)CBX_WaitEvent(&box,&event,100);
    switch(event.type) {                        /* select event response */
    case Expose:                                /* expose event */
      CBX_DrawString(&box, 5+(w-w2)/2, TXy, text);
      CBX_UpdateEditWindow(&pawin);
      CBX_UpdateButton(&okbut);
      CBX_UpdateButton(&cabut);
      break;
    case EnterNotify:                           /* window crossing */
    case LeaveNotify:
      if (event.xcrossing.window == pawin.win) {
        CBX_CrossEditWindow(&pawin,event.type);
      }  
      break;
    case KeyPress:                              /* key pressed in window */
      if (event.xkey.window ==  pawin.win) {
        if (CBX_HandleEditWindow(&pawin,(XKeyEvent*)&event)) done = 1;
      }  
      break;
    case ButtonPress:                          /* button-press event     */
#ifdef  SHFTCLK  // was just: __APPLE__
/* For Apple computers, shift-click should be center click */
	if (event.xbutton.state & ShiftMask) event.xbutton.button = 2;
#endif
      if (event.xbutton.button != 1) break;
      if (event.xbutton.window == okbut.win) { /* OK-button              */
        done =   CBX_HandleButton(&okbut);
      } else
      if (event.xbutton.window == cabut.win) { /* cancel-button          */
        done =  -CBX_HandleButton(&cabut);
      }  
    }  
  } /* while (done == 0) */

  CBX_CloseMainWindow(&box);

  if (done < 0) return(False);
  (void)strcpy(answer,pawin.text);
  return(True);
}

/*  ------------------------------------------------------------  */
/*  Special subroutine to find highest leading line number in file  */
int	hsn (char* filename)
/* Read the file and track the highest number in the first blank
  delimited token of each line.  Return that value.  Zero if no
  file exists or no line is read.  */
{
int	r = 0;	// result
int	j, k;	// status and value read
char	line[256];	// line buffer
char	tokn[16];	// token buffer
char	*c;
FILE	*f;	// the file pointer

// Open the file; return for error
    if (filename == NULL) return  r;
//    f = fopen (filename, "r");
    f = xopen (filename, "r");
    if (f == NULL) return r;

// Loop to read the file
    while ( fgets (line, 256, f) != NULL) {
	c = parse (line, tokn, " ");
	j = sscanf (tokn, "%d", &k);
	if (j == 1) {	// got one
	    if (k > r) r = k;
	}
    }

// Close the file
//    fclose (f);
    iclose (f);

// Return result
    return  r;
}

/*  ------------------------------------------------------------  */
#ifdef  IMAGM
//static	void	handle_obscat (Obs* obs, char* filename)
static	void	handle_obscat (Obs* obs)
/* Append a new line to the current obscat file containing the
  current information in the obs structure.  */
{
int	k=0;	// line number;
FILE	*f;
double	roff, angl;
char	abuf[32];
char	dbuf[32];
int	wl;

//    wl = (obs->dlevel > 1);	// or debug level when replaced
    wl = 1;	// temporary always debug

// Get the highest line number in the desired file, and increment it
//    k = hsn (filename);
    k = hsn (obs->title);
    k++;

// Open the file for append access  (a)
//    f = iopen (filename, "a");
    f = iopen (obs->title, "a");
    if (f == NULL) return;

// Write the new line to the file
//    roff = (InstIndx(ObInstN(obs)) == I_lds3) ? LDSSROA : CHUECO;
//    roff = CHUECO;	// Defined in mgutils.h
// SHOULD THIS BE RotOff here???
    roff = RotOff;	// As set by default or environemnt value
    angl = fmod (obs->angle/Degree + 90.0 + roff, 360.0) - 180.0;
    sexigw (abuf, obs->ra /Hour,   3, 3);
    sexigw (dbuf, obs->dec/Degree, -3, 2);	// Special feature!
    fprintf (f, "%03d %.8s %s %s %6.1f 0.0 0.0 %7.2f OFF ",
	k, obs->fname, abuf, dbuf, obs->equinox, angl);
    if (wl) fprintf (stderr, "%03d %.8s %s %s %6.1f 0.0 0.0 %7.2f OFF ",
	k, obs->fname, abuf, dbuf, obs->equinox, angl);
// Add guider positions also.
#ifdef  GSDATA
    gpout (obs->gp1, f, stderr, wl);
    gpout (obs->gp2, f, stderr, wl);
// see mgutils.c for gpout definition and use
#endif
    fprintf (f, " \n");
    if (wl) fprintf (stderr, " \n");

// Close the file, and done.
    iclose (f);
}
#endif  // on IMAGM

/*  ------------------------------------------------------------  */
void  handle_active (void* param)
/* Do a save, call maskgen with file name using "system" */
{
Button	*but=(Button*)param;
char	buffer[128];
char	inspec[128];
char	message[256];
char	fname[10];
char	progname[16];
char	*c;
//char	lc;

// Copy of the save code from handle iqmenu -- look to make a
// common copy of this stuff later?

// Synchronize the obs structure...
    Gob->changed |= TextScan (winlist);
// Also want to check a scan of IFUwinlist...
#ifdef  IFUM
    Gob->changed |= TextScan (IFUwinlist);
#endif

/* If we happen to have image mode, and it happens to be active, we do
  the image mode thing rather than running mask generation.  Call a
  special program, and then we are done.  */

#ifdef  IMAGM
    if (IMAGUI) {
//	handle_obscat (Gob->obs, "observing.cat");
	handle_obscat (Gob->obs);
	goto act_done;
    }
#endif

/* Find the file name to use... */
    strncpy (fname, Gob->obs->fname, 8);
    fname[8] = AN;			/* Our file name */
    strcpy (inspec, Gob->filename);	/* Current full specification */
    c = filenamloc (inspec);
    *c = AN;				/* Now spec. without filename */
    sprintf (buffer, "%s%s.obs%c", inspec, fname, AN);
/* That puts the candidate file name into the buffer */
#if (DEBUG_GUI > 0)
fprintf (stderr, " !Mangled fname of %.8s to %s.\n", Gob->obs->fname, fname);
fprintf (stderr, " !Mangled filename of %s to %s.\n", Gob->filename, buffer);
#endif

/* If we lack a fname, give a message to enter it and then exit */
    if (!(strcmp (fname, "") && strcmp(fname,"Unnamed")) ) {
	sprintf (message, "Please enter a file name.");
	CBX_MessageBox ("No file name found.", message);
	goto  act_done;
    }

/* Check for consistency; give message and exit if these things are
  lacking in the current data.  We may add some...  */

// Consistency tests
// in handle_active

    if (instmu.sel < 1) {
	sprintf (message, "Please enter an instrument.");
	CBX_MessageBox ("No instrument found.", message);
	goto  act_done;
    }

    if (dispmu.sel < 1) {
	sprintf (message, "Please enter a disperser.");
	CBX_MessageBox ("No disperser found.", message);
	goto  act_done;
    }

    if (Gob->obs->objfiles == NULL) {
	sprintf (message, "Please enter an object file.");
	CBX_MessageBox ("No object files found.", message);
	goto  act_done;
    }

    if (Gob->obs->dec > 61.0*Degree) {
	sprintf (message, "Please enter a valid position.");
	CBX_MessageBox ("Invalid position found.", message);
	goto  act_done;
    }



/* Having assured a correct fname, write to it if data has changed. */

    if (Gob->changed) {
	printf (" == Writing to \"%s\"\n", buffer);  // debug
// Write output to current file name if possible
	strcpy (Gob->filename, buffer);
	if (!writeobs (Gob->filename, Gob->obs)) {
// change into alert box!  (change to message box).
	    sprintf (message, " !! File not written to \"%s\"",
		    Gob->filename);
	    CBX_MessageBox ("Error Alert", message);
	    goto  act_done;
	} else { Gob->changed = 0;  writ++; }
    }

/* Having a correct data file, do the system command next, using
  the fname as the argument to maskgen (or tmaskgen) */

#define  MaskGenPro  "maskgen"

//    lc = strncmp(myname, "tintgui", 7) ? ' ' : 't';
// Above uses tmaskgen if we are executing tintgui -- debug feature

    sprintf (progname, "%s", MaskGenPro);
    if (!strncmp(myname, "tintgui", 7)) sprintf (progname, "%s", "tmaskgen");
// The following line is not really needed anymore...
    if (!strncmp(myname, "gintgui", 7)) sprintf (progname, "%s", "./maskgen");

#ifdef  OBJCHECK
//    sprintf (buffer, "%c%s -i -q -c %s", lc, MaskGenPro, fname);
//    sprintf (buffer, "%c%s -q -c %s", lc, MaskGenPro, fname);
    sprintf (buffer, "%s -q -c %s", progname, fname);
#else
//    sprintf (buffer, "%c%s -i -q %s", lc, MaskGenPro, fname);
//    sprintf (buffer, "%c%s -q %s", lc, MaskGenPro, fname);
    sprintf (buffer, "%s -q %s", progname, fname);
#endif  // on OBJCHECK
/* The -i flag above was to specify minimum debugging; it has been
  removed as the default is expected to be -d0.  If that changes,
  the appropriate flag would be -d0 here.  */
    printf ("-Executing: \"%s\" \n", buffer);
    system (buffer);

/* All done, we return; but release the button first */
act_done:
    CBX_ReleaseButton (but);
}

/*  ------------------------------------------------------------  */
/* A generalized button handler -- this program handles the Test button,
  and also the Object Files button, the optional Exit button, and any
  other simple-to-handle button which might need it.  */
void  handle_button(void* param)
{
int	n;
int	k;
int	i;
// char	qs[]=" ";
int	*taboo;
char	**legend, **title;
Button	*but=(Button*)param;
namlist	*p;

#define  Xtra  4
#define  Xtrap  Xtra+1
#define  WinSiz  80

// Things for testing..
#ifndef  ANGLEDBUG
char	testbuf[64];
char	*items[4]={"Button 1", "Button 2", "Button 3", ""};
int	flags[4]={1,0,1,0};
#endif  // on ANGLEDBUG

/* This "if" stack selects the actual button being handled... */
    if (but == &objlist) {
	printf (" .. Handle object list\n");

// For real - for object list:
// count number of items in list, n+4 to be used
	n = 0;
	C_loop (p, Gob->obs->objfiles) n++;

// dynamically allocate taboo, legend, title arrays
// taboo is int, legend and title are arrays of char*
	taboo = (int*) malloc (sizeof(int) * (n+Xtrap) );
	legend = (char**) malloc (sizeof(char*) * (n+Xtrap) );
	title = (char**) malloc (sizeof(char*) * (n+Xtrap) );

// fill in title arrays with known strings and blanks
// make the blanks about 32 or 48 characters long
// all legends are single blanks
// all taboos are zero
	for (i=0; i<n+Xtrap; i++) {
	    taboo[i] = 0;
//	    legend[i] = &(" ");
	    legend[i] = " ";	// All legends point to same blank literal
	    title[i] = (char*) malloc (WinSiz);
	    strcpy (title[i], legend[i]);
	}
/* At this point, put about 16 blanks in the last title[] to make
  the displayed default window a resonable width. */
	memset (title[n+Xtra-1], '\0', 20);
	memset (title[n+Xtra-1], ' ', 16);
//	for (i=0,p=Gob->obs->objfiles; p != NULL; p=p->next,i++) {
	i = 0;
	C_loop (p, Gob->obs->objfiles) { strcpy (title[i], p->name); i++; }

// Follow with a legend that is null string.
	legend[n+Xtra] = "";
	strcpy (title[n+Xtra], legend[n+Xtra]);
    
// Wants PXh to be the font height as a static variable?
	if (CBX_MultiParBox ("Object Files", legend, title, taboo)) {
// True if we pushed "ok" whether or not a change happened.
// Could look here to see if title[] array matches the objfiles
// array from which it was created, and if they match, skip the rest.
int	j;
int	tag;
	    for (j=tag=0,p=Gob->obs->objfiles; 
		p != NULL && tag == 0; p=p->next,j++) {
		tag = strcmp (title[j], p->name);
		if (p->next == Gob->obs->objfiles) break;
// BAD CONSTRUCT BREAK
	    }
// Also look for non-null entries in formerly enpty lines
	    for (j=n; j < n+Xtra-1; j++) 
		tag |= strcmp (title[j], " ");
// Now, if tag is not zero, we have a real change and we actually
// make the changes here...
/* We don't test the last line which had the extra blanks placed
  in it for spacing.  For a better, more accurate test, we should
  use trims(title[i],WHITESPACE) on all and test for non-null.  This
  method above can be fooled by putting a new file in the last line.  */
	    if (tag) {		// Test for a real change    
// Remove the old queue of file names
		p = Gob->obs->objfiles;
		p = kill_name (p);	// leaves p as NULL
// Put each non-null from title into a new queue
		for (i=0; i<n+Xtrap; i++) {
// See if name at title[i] is/not null
		    trims (title[i], WHITESPACE);
		    if (*title[i] == AN) continue;
// If not null, put it in new queue p
		    p = pushtxt (title[i], p);
		}
// Finally, put p into structure...
		Gob->obs->objfiles = p;
		Gob->changed |= 1;	// Mark as a change
#ifdef  SKYOPT
		if (skywin.disp && skywin_active) fill_objects (Gob->obs);
#endif
	    }	// end if (tag)
	}	// if on multipar box ("ok" pressed)

// Clean up all the dynamic array stuff... 
// Whether or not the edited titles were actually used.
	for (i=0; i<n+Xtrap; i++) {
	    free ((void*)title[i]);
	}
	free ((void*)taboo);
	free ((void*)legend);
	free ((void*)title);

    } else if (but == &testbut) {	// The test button - alerts, etc.
// The testbut appears only on symbol WDEBUG defined at the top.

#ifdef  ANGLEDBUG
/* Here, convert the Test button to an angle report.  See the #else below
  for the traditional use in testing alert box things.  */

// Do a report at a set of hour angles, computing the rotator angle as it
// is usually done, perhaps with intermediate results.

// local variables
double	v;
double	pa, zd,  zv, nv;
double	sp,  imx;
double	rx, rv;
double	rrx;

// define the set of declinations reported (from rotrate.c)
#define  Nh  19
double	Hs[Nh] = { -6.0, -5.0, -4.0, -3.0, -2.0, -1.0,
		-0.5, -0.3, -0.1, 0.0, 0.1, 0.3, 0.5,
		1.0, 2.0, 3.0, 4.0, 5.0, 6.0 };

// report the declination and position angle
    printf ("-- Rotator Angles for Dec. %.1f, Pa %.1f, RotOff %.1f\n",
	pdx.dec, pdx.pa, RotOff);
    printf ("  H.A.  Rotator  pa    zd    zv     nv    imx     rx     rv    imx\n");

// In a loop for hour angles,
// find paralactic angle, north direction and rotator angle,
// then report the ha, and intermediate angles to stdout.
    sp = (pdx.side == 0) ? 1.0 : -1.0;
    for (k=0; k < Nh; k++) {
	v = rotang (&imx, Hs[k], pdx.dec, pdx.lat, pdx.pa, pdx.side);
	if (v > 180.0) v -= 360.0;

	pazd (&pa, &zd, Hs[k], pdx.dec, pdx.lat);
	zv = sp * (zd - 90.0);  // "up" left of "vertical"
	nv = zv + pa;		// "north" left of vertical
	imx = nv - pdx.pa;	// imacs angle (not yet), and it's negative.
/*	rx = zd*sp - imx;	// rotation angle */
//	rv = rx + RotOff;	// Add offset
//	rv = - (nv + zd);	// should be rotator? == rx

	rx =  90.0*sp - pa + pdx.pa;

//	rv = RotOff - 90.0 - pa + pdx.pa;
	rv = RotOff + rx;
	rv = drange(rv, 360.0, 180.0);  // rotator with offset

//	rrx = RotOff+(pa + zd);	// rotrate report
	rrx = -(rv - RotOff - sp*zd);	// Imacs angle actual
/* This looks real good.  Now, we need to get rv computation into rotang,
  and put in address of imacs angle too.  */

	printf ("  %4.1f %6.1f  %6.1f %5.1f %5.1f %6.1f %6.1f %6.1f %6.1f %6.1f\n",
		  Hs[k],   v,    pa,   zd,   zv,   nv,   -imx,  rx,  rv, rrx);
    }


#else  // on ANGLEDBUG
/* The alert box presents a message and allows OK or cancel choices. */
	k = CBX_AlertBox ("Alert Box", "Test dialog boxes?");
	if (!k) goto skip_tests;


/* Test dialog box
  The dialog box presents text and a set of labeled buttons;
  pressing a button returns the index of the button pressed.
  Kinda like an extended alert box...		*/
	k = CBX_DialogBox ("Dialog Box", "Dialog Text", 
		"Button 1|Button 2|Button 3");
	sprintf (testbuf, "You pressed button %d.", k);
/* The MessageBox sends a message, allows it to be read and cleared. */
	CBX_MessageBox ("Message Box", testbuf);
	if (!CBX_AlertBox("Alert Box", "More Testing?")) goto skip_tests;



/* Test parameter box
  The parameter box presents an edit window with the string in it,
  and allows it to be edited.  Select OK or cancel, and return result  */
	strcpy (testbuf, "Edit me");
	k = CBX_ParameterBox ("Parameter Title", testbuf);
	if (k) {
	    CBX_MessageBox ("Message Box", testbuf);
	    if (!CBX_AlertBox("Alert Box", "Continue?")) goto skip_tests;
	}

/* We need a combination of dialog box (presenting text) and parameter box
  (having edit window) for the save-as dialog.  So we write our own... */

//Bool KDC_DialogBox(char *title, char* text, char *answer)

	strcpy (testbuf, "Edit Me");
	k = KDC_DialogBox ("Parameter Title", "Supply the answer", testbuf);
	if (k) {
	    CBX_MessageBox ("Message Box", testbuf);
	    if (!CBX_AlertBox("Alert Box", "Continue?")) goto skip_tests;
	}

/* Test multi-select box
  A multi-select will take array of character items, create a
  set of D-buttons with active color, and allow each to be selected
  initialize and return the flags array;
  can select OK, all or cancel and modified stuff is returned.  */
	k = CBX_MultiSelBox ("Multi-Select", items, flags, app->BUT_SEL);
	if (k) {
	    sprintf (testbuf, " Set: %d %d %d\n", 
		flags[0], flags[1], flags[2]);
	    CBX_MessageBox ("Message Box", testbuf);
	    if (!CBX_AlertBox("Alert Box", "More?")) goto skip_tests;

	}



/* Test select box.
  The select takes an array of character items, ending with a null one,
  and returns subscript of the button which was selected.  Similar effect
  to the dialog above, but with longer text items.
  Also has a "cancel" option, and buttons are arranged vertically.  */
	k = CBX_SelectBox ("Select Box", items);
	sprintf (testbuf, "You pressed button %d.", k+1);
	CBX_MessageBox ("Message Box", testbuf);


/*  End of tests...
  When no more tests are selected, this goto target will exit the test
  section.  Can't use continue or break in this "if" stack... */
skip_tests:  ;
#endif  // on ANGLEDBUG
    } else if (but == &kilbut) {	// The kill button 
	done = True;	// Real simple, but effective!
    } else {	// Some other button...  (catch-all)
	printf (" ..Handle that button!\n");
    }


    CBX_ReleaseButton (but);
}
/* End of test button -- used only in test mode */


/*  ------------------------------------------------------------  */
void	MakeFiltmenu (mble** L, filtelem* f, mnublock* mnu, IMenuEntry* fme)
{
int	x, y;
int	w, h;
int	t;
//char	title[] = "Filter:";
int	i;
char	*c;

/* Get the filter windows created... */

/* Should dynamically find largest text label... */
/* Dynamically create filter menu entries, label width */
//    for (w=i=0; *(filters[i].name) != AN; i++) {
    for (w=i=0; *(f[i].name) != AN; i++) {
//	c = filters[i].name;
	c = f[i].name;
	strcpy (fme[i].name, c);
	strcpy (fme[i].text, c);
	t = strlen (c);
	if (t > w) w = t;
//	fme[i].value = i;
	fme[i].order = 0;
	fme[i].type = 0;
	fme[i].flag = 0;
    }
    mnu->width = w;

    AlignMenu (mnu, &x, &t, &w);

    y = Yloc(mnu->yloc);
    h = (int)(mnu->height*Vunit + 1);

/* Create the menu window */
    (void)CBX_CreatePopupMenu (&mwin, mnu->mu,  x, y,  w, h );

// Fill in passband data if not already in the structure
    if (Gob->obs->pb.blue == 0.0) Gob->obs->pb.blue = f[0].pb.blue;
    if (Gob->obs->pb.red  == 0.0) Gob->obs->pb.red  = f[0].pb.red;

/* Create the passband text windows also */
// find position and size of passband windows
// NOTE - there will be 2 titles printed, "Filter:" and "-",
// these are to be defined above as globals, and used in title stuff
// both here and in the draw filter menu subroutine
// initialize passband with starting pb values...

/* NOTE -- we also need:
 -- The filtmu = a pmenu data item above
 -- subroutine to handle filter menu - put up the name, and
	initialize the subsidiary passband windows with values
 -- subroutine to draw the menu:  draw the titles, and
	draw the subsidiary windows also
 -- Maybe we make the passband windows into standard windows and 
	provide titles (second one is "-" title), and use the standard
	draw and handle routines for them.
.. */
    if (L != NULL) *L = mladd (*L, mnu);

}

/*  ------------------------------------------------------------  */
#ifdef  GUIDATE
void	MakeNumMenu (mble** L, int na, int nb, mnublock* mue)
{
// use general menu maker to make the menu
Menu	*m;
int	w;
int	j;
char	buf[32];


    MakeMenu (L, mue);
// in a loop, fill entries na to nb into the menu using add menu entry.
// we may need to find the largest field width?
// see other instances of AddMenuEntry for how to do this.
// also check the fill menu things used for filters, dispersers, etc.
// See those in guiutil.c, guisubs.c

//    CBX_AddMenuEntry (&optmenu, "Skywin",  CBX_NORMAL);

    m = mue->mu;
    w = mue->width;
    if (m == &yearmnu) baseyear = na - 1;

/* Make the zero-th entry a blank thing */
    sprintf (buf, "- - - - -");
    buf[w] = AN;
    CBX_AddMenuEntry (m, buf, CBX_NORMAL);

/* Make entries for all the specified numbers */
    for (j=na; j<nb+1; j++) {
	sprintf (buf, "%-*d", w, j);
	CBX_AddMenuEntry (m, buf, CBX_NORMAL);
    }
    m->sel = 0;
}

// look at other menu makers for hints...
/*  ------------------------------------------------------------  */

//void	MakeMonthMenu (mble** L, IMenuEntry* mfld)
void	MakeMonthMenu (mble** L,  mnublock* mue)
{
// use general menu maker
// loop over the standard month name stuff to fill menu
Menu	*m;
int	j;
char	monam[12][4] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

    MakeMenu (L, mue);
    m = mue->mu;
    CBX_AddMenuEntry (m, "- -", CBX_NORMAL);
    for (j=0; j<12; j++) {
	CBX_AddMenuEntry (m, monam[j], CBX_NORMAL);
    }
    m->sel = 0;
}

//    MakeMonthMenu (&mnulist, &monthfield);
//    MakeNumMenu (&mnulist, 1, 31, &dayfield);

//void	MakeFiltmenu (mble** L, filtelem* f, mnublock* mnu, IMenuEntry* fme)


#endif  // on GUIDATE

/*  ------------------------------------------------------------  */
void handle_iqmenu(void* param)
// param is menu pointer, ignored.
{
int	new;
int	gf;	// got filename
char	buffer[128];
char	inspec[128];
char	message[256];
char	fname[10];
char	*c;
winblock  *wb = &datafile;
bool	good;
Menu	*mnu=(Menu*)param;
//int        x,y,w,h;
//int        i;
//XSizeHints hint;

//    new = CBX_HandleMenu(&iqmenu);
// automatic doesn't need additional handling...
//    new = iqmenu.sel;
    new = mnu->sel;	// *mnu is iqmenu, for lint

/*  SYNCHRONIZE -- Get ready for save or save-as:
  copy all text to get obs structure filled with latest things
  get the inspec file pre-pend stuff and current filename
  Put the likely full spec into the buffer. */

    Gob->changed |= TextScan (winlist);
// Also want to check a scan of IFUwinlist...
#ifdef  IFUM
    Gob->changed |= TextScan (IFUwinlist);
#endif
    strncpy (fname, Gob->obs->fname, 8);
    fname[8] = AN;			/* Our file name */
    strcpy (inspec, Gob->filename);	/* Current full specification */
    c = filenamloc (inspec);
    *c = AN;				/* Now spec. without filename */
    sprintf (buffer, "%s%s.obs%c", inspec, fname, AN);
/* That puts the candidate file name into the buffer */
#if (DEBUG_GUI > 0)
fprintf (stderr, " !Mangled fname of %.8s to %s.\n", Gob->obs->fname, fname);
fprintf (stderr, " !Mangled filename of %s to %s.\n", Gob->filename, buffer);
#endif

#ifdef  IMAGM
/* In image mode, if it is on, we can only exit.  */
    if (IMAGUI) {
	done = True;
	return;
    }
#endif

    switch (new) {
    case FILE_SAVE:
// Skip if not  Gob->changed
// NOTE that the changed flag is not always maintained!!!

saveme:	;	// For auto-save

// we can write file if fname is non-null, otherwise
// we need to fall to save-as.


// Write to  Gob->filename, unless that is null -- then fall to saveas
// Use standard filename if debug version.
//	if (strcmp (Gob->filename, "")) {	// file not null
	if (strcmp (fname, "")) {	// We have non-null fname
/* Would test if Gob->changed, and return if not changed. */
	    printf (" == Writing to \"%s\"\n", buffer);  // debug
// Write output to current file name if possible
	    strcpy (Gob->filename, buffer);
	    if (!writeobs (Gob->filename, Gob->obs)) {
// change into alert box!  (change to message box).
		sprintf (message, " !! File not written to \"%s\"",
			Gob->filename);
		CBX_MessageBox ("Error Alert", message);
	    } else { Gob->changed = 0;  writ++; }
	    break;
	}  // Fall through to saveas if fname was null.
    case FILE_SAVEAS:
// Use kdc dialog to get name from user, then write on that file
// Write to the file name if given, and store filename in Gob->filename
//	    strcpy (buffer, Gob->filename);
	    gf = KDC_DialogBox (
		"Save As",		// Title
		"Filespec to save data into:",	// Request text
		buffer);
/* After asking nicely, if gf is false, we got a "cancel" in the
  dialog, and don't save.  Write to the file as specified by the
  buffer, and if it succeeds, change the filename to buffer data.
.. */
// NOTE - we may want to add ".obs" to end of buffer if no extension
// is already present, for consistency of use
	    if (gf) {
/* We got a suggestion from the user of what to save as.  So, we extract
  the filename from it, and re-make the string with the pre-pend part,
  with the filename and add ".obs" to the end.  Then write that, and
  put it into the current filename location. */

		strcpy (inspec, buffer);	/* What was requested */
		c = filenamloc (inspec);	/* Location for filename */
		strncpy (fname, getfilenam(buffer), 8); /* User's filename */
		fname[8] = AN;
fprintf (stderr, " !From request %s at %s got filename %s\n", buffer, c, fname);
		if (strlen(fname) < 1) {
		    sprintf (message, 
			"No Save -- NULL Data File not allowed.");
		    fprintf (stderr, "%s\n", message);
		    CBX_MessageBox ("No Save Done", message);
		    break;
		}
		sprintf (c, "%s.obs%c", fname, AN);
// THAT LOOKS BAD?
fprintf (stderr, " !File %s part of spec: %s\n", c, inspec);
/* Check if it exists, and skip the copy if not overwrite */
		good = access (inspec, F_OK);
		if (!good) {
		    sprintf (message,
			"File \"%s\" exists!  Overwrite?", fname);
		    good = CBX_AlertBox ("Overwrite File Alert",
			message);
		}
/* If we have good set here, we can write the thing. */

		if (good)  {
/* OK to write.  Put the actual fname into structure and window. */
		    strncpy (Gob->obs->fname, fname, 8);
/* Also, put fname into the datafile window text buffer:
  defined wb above as a pointer there, so we copy the data... 
  An inverse if what is done in TextScan */
// like..	    strncpy (wb->ew->text, wb->data, wb->width);
		    strncpy (wb->ew->text, fname, 8);
/* And, actually write the structure to the file. */
		    if (writeobs (inspec, Gob->obs)) {
			strcpy (Gob->filename, inspec);
			Gob->changed = 0;
			writ++;
		    }
		}	// Test for overwrite
	    }
	break;
    case FILE_DONE:		// save and exit
	done = True;
// possibly add "if (Gob->changed)" here?
// NO - not all edits (e.g. change name of object file) will change
// the global change flag, so we always do a save for this selection.
	goto saveme;
    case FILE_KILL:
/* The immediate exit, quit with no save */
	done = True;
	return;
    case FILE_EXIT:
	done = True;
	if (Gob->changed) {	// Auto save; ask first
	    sprintf (message, "Data changed.  Save?");
	    if (CBX_AlertBox ("Auto Save", message))  {
		fprintf (stderr, " !!Automatic file save.\n");
		goto saveme;
	    }
	}
	break;
    }	// switch on menu entry
}

/* NOTES
  Gob->obs->fname is not synchronized with filename;
  filename is not present in the GUI in a window.
.. */


/*  ------------------------------------------------------------  */
static	void	check_moe (int inst, int disp)
{
/* Check if instrument/disperser is moe; set or unset slit length and
  uncut length windows depending on the moe state.  Leave early if
  the windows check out already set properly. */
/* Center wavelength is also greyed out if moe active */
/* Also implement restriction on extra orders feature, and slit extension. */

// See if it is moe currently
#ifdef  GISMOE
    if ((inst == INST_Ilc && disp == DISP_moe) 
#ifdef  GismoMenus  // for definition of INST_Gxx existing
	|| (inst == INST_Gsl && disp == DISP_gmoe - DISP_selG)
// NOTE that the -DISP_selG may be wrong, and we may need to do
// this test differently.  Perhaps test the name of the disp item in the menu?
#ifdef  GISMOMNS
	|| (inst == INST_Gslx && disp == DISP_gmoe - DISP_selG) 
#endif
#endif  // on GismoMenus
	) {  // not DISM_gmoe
#else
    if (inst == INST_Ilc && disp == DISP_moe) {
#endif
// For moe, we need to set slit widths to 2.5 and disable windows
// First, set the values if needed
#ifdef  OBJSHAPE
	if (Gob->obs->default_slit.shape != RECTANGLE) {
	    Gob->obs->default_slit.shape = RECTANGLE;
	    shapemnu.sel = 0;
	    CBX_UpdateMenu (&shapemnu);
	}
#else
// Slit shape is needed in any case, probably is already RECTANGLE
	Gob->obs->default_slit.shape = RECTANGLE;
#endif
	if (Gob->obs->default_slit.alen != 2.5) {
	    Gob->obs->default_slit.alen = 2.5;
	    HandleTextWindow (winlist, &slawin);
	}
// Could use e.g. SetFloatWindow (winlist, &slafield, 2.5)
	if (Gob->obs->default_slit.blen != 2.5) {
	    Gob->obs->default_slit.blen = 2.5;
	    HandleTextWindow (winlist, &slbwin);
	}
	if (Gob->obs->dca != 0.0) {
	    Gob->obs->dca = 0.0;
	    HandleTextWindow (winlist, &dcawin);
	}
	if (Gob->obs->dcb != 0.0) {
	    Gob->obs->dcb = 0.0;
	    HandleTextWindow (winlist, &dcbwin);
	}

// Second, disable the windows if they are not already disabled
	if (slawin.flag == CBX_NORMAL) 
	    CBX_DisableEditWindow (&slawin, CBX_DISABLED);
	if (slbwin.flag == CBX_NORMAL) 
	    CBX_DisableEditWindow (&slbwin, CBX_DISABLED);
	if (dcawin.flag == CBX_NORMAL) 
	    CBX_DisableEditWindow (&dcawin, CBX_DISABLED);
	if (dcbwin.flag == CBX_NORMAL) 
	    CBX_DisableEditWindow (&dcbwin, CBX_DISABLED);

	if (cwwin.flag == CBX_NORMAL) 
	    CBX_DisableEditWindow (&cwwin, CBX_DISABLED);
	if (ordwin.flag == CBX_NORMAL) 
	    CBX_DisableEditWindow (&ordwin, CBX_DISABLED);

// Disable extra orders too

// Disable slit extension feature if present
#ifdef  SLXB
	if (Gob->obs->slext != 0) Gob->obs->slext = 0;
	if (slext.flag == CBX_NORMAL) CBX_DisableButton (&slext, CBX_DISABLED);
#endif

// Check on extra order menu also
	if (Gob->obs->ex_order != 0) {
	    Gob->obs->ex_order = 0;
	    xordmu.sel = 0;
	    CBX_UpdateMenu(&xordmu);
	}
	if (xordmu.flag == CBX_NORMAL) 
	    CBX_DisableMenu (&xordmu, CBX_DISABLED);

/* If slit shape menu is supported, it needs to be disabled for MOE */

#ifdef  OBJSHAPE
	if (shapemnu.flag == CBX_NORMAL)
	    CBX_DisableMenu (&shapemnu, CBX_DISABLED);
#endif

    } else {						// not moe

// Enable slit shape menu if it exists and was disabled.
#ifdef  OBJSHAPE
	if (shapemnu.flag == CBX_DISABLED
//#ifdef  GISMOAP
#if  defined GISMOAP && defined GismoMenus
	    && (inst != INST_Gsal)
#endif
	   ) CBX_EnableMenu (&shapemnu);
#endif

// Not moe, check that windows are enabled
/*  NOTE -- Should only enable windows if slit shape is RECTANGLE; check
  that before doing these enable items!  */
      if (Gob->obs->default_slit.shape == RECTANGLE
//#ifdef  GISMOAP
#if  defined GISMOAP && defined GismoMenus
	    && (inst != INST_Gsal)
#endif
         ) {
	if (slawin.flag == CBX_DISABLED) CBX_EnableEditWindow (&slawin);
	if (slbwin.flag == CBX_DISABLED) CBX_EnableEditWindow (&slbwin);
	if (dcawin.flag == CBX_DISABLED) CBX_EnableEditWindow (&dcawin);
	if (dcbwin.flag == CBX_DISABLED) CBX_EnableEditWindow (&dcbwin);
      }

	if (ordwin.flag == CBX_DISABLED
//#ifdef  GISMOAP
#if  defined GISMOAP && defined GismoMenus
	    && (inst != INST_Gsal)
#endif
	   ) CBX_EnableEditWindow (&ordwin);

	if (cwwin.flag == CBX_DISABLED && instme[inst].type == DTYPE_grating
//#ifdef  GISMOAP
#if  defined GISMOAP && defined GismoMenus
	    && (inst != INST_Gsal)
#endif
	   )
	    CBX_EnableEditWindow (&cwwin);
	if (xordmu.flag == CBX_DISABLED
//#ifdef  GISMOAP
#if  defined GISMOAP && defined GismoMenus
	    && (inst != INST_Gsal)
#endif
	   ) CBX_EnableMenu (&xordmu);

// Enable slit extension if present, and slit shape is RECTANGLE
#ifdef  SLXB
      if (Gob->obs->default_slit.shape == RECTANGLE) {
	if (slext.flag == CBX_DISABLED
//#ifdef  GISMOAP
#if  defined GISMOAP && defined GismoMenus
	    && (inst != INST_Gsal)
#endif
	   ) CBX_EnableButton (&slext);
      }
#endif
    }
}

/*  ------------------------------------------------------------  */
#ifdef  GISMOAP
void  disable4ap (int flag)
/*  Disable the set of features if flag is on, enable if off.  */
{
    if (flag) {
	CBX_DisableEditWindow (&cwwin, CBX_DISABLED);
	CBX_DisableEditWindow (&swidwin, CBX_DISABLED);
	CBX_DisableEditWindow (&slawin, CBX_DISABLED);
	CBX_DisableEditWindow (&slbwin, CBX_DISABLED);
	CBX_DisableEditWindow (&dcawin, CBX_DISABLED);
	CBX_DisableEditWindow (&dcbwin, CBX_DISABLED);
	CBX_DisableEditWindow (&blupb, CBX_DISABLED);
	CBX_DisableEditWindow (&redpb, CBX_DISABLED);
	CBX_DisableEditWindow (&bludb, CBX_DISABLED);
	CBX_DisableEditWindow (&reddb, CBX_DISABLED);
	CBX_DisableEditWindow (&ordwin, CBX_DISABLED);
	CBX_DisableEditWindow (&repob, CBX_DISABLED);
	CBX_DisableEditWindow (&reprf, CBX_DISABLED);
	CBX_DisableEditWindow (&refsv, CBX_DISABLED);
	CBX_DisableEditWindow (&ovlwin, CBX_DISABLED);
	CBX_DisableEditWindow (&mhwin, CBX_DISABLED);
	CBX_DisableEditWindow (&pdwin, CBX_DISABLED);
	CBX_DisableEditWindow (&xpwin, CBX_DISABLED);

	CBX_DisableMenu (&xordmu, CBX_DISABLED);
	CBX_DisableMenu (&shapemnu, CBX_DISABLED);
	CBX_DisableMenu (&filtmu, CBX_DISABLED);
	CBX_DisableMenu (&gapmu, CBX_DISABLED);

	CBX_DisableButton (&objlist, CBX_DISABLED);
	CBX_DisableButton (&slext, CBX_DISABLED);

/* Clear the RA/DEC positions and the angle, and disable the windows  */
//	if (RAfield.data != NULL) *(RAfield.data) = 0.0;
//	HandleTextWindow (winlist, RAfield.ew);  // That's &alwin there...
	SetFloatWindow (winlist, &RAfield,  0.0*Hour);  // &alwin, 0.0);
	SetFloatWindow (winlist, &DECfield, 0.0*Degree);  // &dewin, 0.0);
	SetFloatWindow (winlist, &slitpaf,  0.0*Degree);  // &spawin, 0.0);
/* The SFW subroutine sets the wb.data value, and then calls HTW to
  properly normalize the window text.  See guiutil.c    */

// &alwin is the address of the EditWindow
// RAfield is the winblock  (wb.ew points to EditWindow)
// handletextwin gets address of wble and address of edit window;
// and finds winblock from the list as it needs winblock.
// We should call using the address of winblock directly?
// or with address of the window?

/* There are general configuration issues which resulted in calling with
  the winlist and window pointer instead of winblock pointer; but yes,
  using the global list head winlist it is possible to write a general
  winblock finder given the window pointer alone.  Main issue is that
  callback routines which we need to write have as a parameter just the
  window pointer which is active, and we often need the winblock.      */

	CBX_DisableEditWindow (&alwin, CBX_DISABLED);
	CBX_DisableEditWindow (&dewin, CBX_DISABLED);
// position windows:  alwin, dewin, spawin, eqmenu
	CBX_DisableEditWindow (&spawin, CBX_DISABLED);
	CBX_DisableMenu (&eqmenu, CBX_DISABLED);
/* For equinox menu, only disable the menu, do not clear.  */
// see clear position

#ifdef  DIFREF
	CBX_DisableEditWindow (&hawin, CBX_DISABLED);
	CBX_DisableButton (&drefbut, CBX_DISABLED);
#endif  // on DIFREF

#ifdef  GUIDATE
/* Clear the date things, clear the mjd (global) and disable menus */
	mjd = 0;
	yr = mo = dy = 0;
	setup_dates();
	yearmnu.sel = 0;
	CBX_DisableMenu (&yearmnu, CBX_DISABLED);
	monthmnu.sel = 0;
	CBX_DisableMenu (&monthmnu, CBX_DISABLED);
	daymnu.sel = 0;
	CBX_DisableMenu (&daymnu, CBX_DISABLED);
#endif  // on GUIDATE

    } else {
//	CBX_EnableEditWindow (&cwwin);
	CBX_EnableEditWindow (&swidwin);
	CBX_EnableEditWindow (&slawin);
	CBX_EnableEditWindow (&slbwin);
	CBX_EnableEditWindow (&dcawin);
	CBX_EnableEditWindow (&dcbwin);
	CBX_EnableEditWindow (&blupb);
	CBX_EnableEditWindow (&redpb);
	CBX_EnableEditWindow (&bludb);
	CBX_EnableEditWindow (&reddb);
	CBX_EnableEditWindow (&ordwin);
	CBX_EnableEditWindow (&repob);
	CBX_EnableEditWindow (&reprf);
	CBX_EnableEditWindow (&refsv);
	CBX_EnableEditWindow (&ovlwin);
	CBX_EnableEditWindow (&mhwin);
	CBX_EnableEditWindow (&pdwin);
	CBX_EnableEditWindow (&xpwin);

	CBX_EnableMenu (&xordmu);
	CBX_EnableMenu (&shapemnu);
	CBX_EnableMenu (&filtmu);
	CBX_EnableMenu (&gapmu);

	CBX_EnableButton (&objlist);
	CBX_EnableButton (&slext);

#ifdef  DIFREF
	CBX_EnableButton (&drefbut);
// if dref is actually on, then...
	if (Gob->obs->dref) CBX_EnableEditWindow (&hawin);
#endif

// position windows:  alwin, dewin, spawin, eqmenu
	CBX_EnableEditWindow (&alwin);
	CBX_EnableEditWindow (&dewin);
	CBX_EnableEditWindow (&spawin);
	CBX_EnableMenu (&eqmenu);

#ifdef  GUIDATE
	CBX_EnableMenu (&yearmnu);
	CBX_EnableMenu (&monthmnu);
	CBX_EnableMenu (&daymnu);
#endif  // on GUIDATE
    }
}
#endif  // on GISMOAP
/*  ------------------------------------------------------------  */
void handle_instmenu(void)
{
int    new;
IMenuEntry	*ip;
IMenuEntry	*mp;
int	newval;
int	dval = 0;
int	old;
int	dif;
//int	ix;
bool	changemenu;
int	ord;

    new = CBX_HandleMenu(&instmu);
    if (new < 0) return;

/* Get the proper instrument and proper disperser menu stuff */
    newval = new;	// value same as index here
    old = instmu.sel;
    dif = old != new;
    ip = &(instme[new]);
    Gob->obs->Instr_mt = find_element (Gob->gs, ip->name);
//    ix = InstIndex (ip->name);

    mp = dispindx[new];
//#ifdef  GISMO
#ifdef  GismoMenus  // Needed for INST defines
// Do not use GismoMenus as new gets set only if menu item exists above
    if (new == INST_Gsl || new == INST_Gss
#ifdef  GISMOMNS
	|| new == INST_Gslx
#endif
#ifdef  GISMOAP
	|| new == INST_Gsal	// and later INST_Gsas
#endif
       ) {
	Gob->obs->Frontend = find_element (Gob->gs, "GISMO");
    } else if (old == INST_Gsl || old == INST_Gss
#ifdef  GISMOMNS
	|| old == INST_Gslx
#endif
#ifdef  GISMOAP
	|| old == INST_Gsal
#endif
       ) {
	Gob->obs->Frontend = NULL;
    }
#endif  // on GISMO
/* In these things (as above) ALWAYS check new == <specific instruments> and
  then check old == <instruments> with an else.  This prevents a problem if
  the instrument is re-selected within the menu.    */

#if (DEBUG_GUI > 1)
{  // show old, new inst and disp types here //debug
int	oldit, newit;
int	olddt, newdt;
	oldit = instme[old].type;	// was oldtyp;
	newit = instme[new].type;
	olddt = oldit;		// was oldtyp;
	newdt = newit;
printf (" ..Change inst,dispt From %d,%d to %d,%d\n", 
			old,olddt, new,newdt);	//debug
fflush (stdout);
}  // end debug section
#endif

/*  Add section for GISMOAP to disable a list of windows if new == INST_Gsal,
  and to enable the same list if old == INST_Gsal.  Or Gsls similarly. */
// #ifdef  GISMOAP
#if  defined GISMOAP && defined GismoMenus
    if (new == INST_Gsal) {	// disable subwindow set
	Gob->obs->reflimit = 0;
// clear position windows:  alwin, dewin, spawin, eqmenu
	disable4ap (1);
    } else if (old == INST_Gsal) {	// enable subwindows
	Gob->obs->reflimit = 99;	// Just a guess...
	disable4ap (0);
    }
/* In the above, check first for the new instrument.  If we are already at
  the Gsal instrument and re-select it, the disable(0) should not be done.  */
#endif


/* Fill in the disperser menu, its selection, and enable it. */

    if (new != old || new == 0) {
//	DISP_sel = 0;	// Force selection.
// Redundant; DISP_sel is an enum value defined as zero...
	dval = FillMenu (&dispmu, mp, DISP_sel);
	dispmu.sel = dval;
	warnflag = 1;	// Due to rotator offset differences
	Gob->obs->Disperser = NULL;
//	Gob->obs->order = 0;	// Put order into current obs
//	HandleTextWindow (winlist, ordfield.ew); // That's &ordwin there...
//	SetEditWindow (winlist, &ordwin, "0");
	SetIntWindow (winlist, &ordfield, 0);
// or we could use dispmu.sel = 0; -- all the same!

// Set the menu to disabled if the instrument is 0, enabled otherwise
	if (instme[new].type == DTYPE_unknown) {
	    CBX_DisableMenu (&dispmu, mp->flag);
//#ifdef  GISMOAP
#if  defined GISMOAP && defined GismoMenus
	} else if (new == INST_Gsal) {
	    CBX_DisableMenu (&dispmu, CBX_DISABLED);
	    dispmu.sel = 0;  // not 1;
#endif
	} else {
	    CBX_EnableMenu (&dispmu);
	}
	check_moe(new, dispmu.sel);

// Disable wavelength window if not a grating...
	if (instme[new].type != DTYPE_grating
//#ifdef  GISMOAP
#if  defined GISMOAP && defined GismoMenus
	    || new == INST_Gsal
#endif
	   ) {
	    CBX_DisableEditWindow (&cwwin, CBX_DISABLED);
//#ifdef  GISMOE
#if  defined GISMOE && defined GismoMenus
	} else if ( ! ( (new == INST_Ilc && dval == DISP_moe) ||
		(new == INST_Gsl && dval == DISP_gmoe) 
#ifdef  GISMOMNS
		|| (new == INST_Gslx && dval == DISP_gmoe) 
#endif
		) ) {
#else  // on GISMOE + GismoMenus
	} else if (new != INST_Ilc || dval != DISP_moe) {
#endif  // on GISMOE + GismoMenus
	    CBX_EnableEditWindow (&cwwin);
	}

/* Use the new method of finding instrument and telescope */
//	if (new == 3 || new == 7) {	// The LDSS instrument
	if (new == INST_Lc || new == INST_Ld) {	// The LDSS instrument
	    NoADC = 1;
	    RotOff = LDSSROA;
	    Gob->obs->Tel_scope = find_element (Gob->gs, "Magellan2");
#ifdef  WFCCD
	} else if (new = INST_wfcd) {  // The WFCCD instrument
	    NoADC = 1;
	    RotOff = 270;	// NOTE: Fix this!
	    Gob->obs->Tel_scope = find_element (Gob->gs, "DuPont");
#endif
	} else {			// default to IMACS
	    NoADC = 0;
	    RotOff = IMACSrot;  // was CHUECO;
	    Gob->obs->Tel_scope = find_element (Gob->gs, "Magellan");
	}
/*  NOTE that the above ignores the possible change by environment variable
  to the basic offset angle.  May need to be fixed.  */
    }

/* Force selection of direct dispersers if a direct instrument is selected. */
    if (newval == INST_Ild) dval = DISP_rz;
    if (newval == INST_Isd) dval = DISP_pz;
// also as Ld when ldss; selects disp-sz.
    changemenu = dval != DISP_sel;
    Gob->changed |= changemenu;		// MAY GIVE FALSE CHANGES
    if (changemenu) {
// NOTE need to find which menu dval is in grat/gris; then
// need to find index within that menu of dval; then
// may need to refill the menu...
	dispmu.sel = dval;
	CBX_UpdateMenu (&dispmu);

	mp = dispindx[new] + dispmu.sel;

	ord = mp->order;

// NOTE CHECK that dispmu.sel is what is desired; also dval may be wrong!
	Gob->obs->Disperser = find_element (Gob->gs, mp->name);
//	Gob->obs->order = ord;	// Put order into current obs
//	HandleTextWindow (winlist, ordfield.ew);  // That's &ordwin there...
	SetIntWindow (winlist, &ordfield, ord);
// would need to write ord to text to do...
//	SetEditWindow (winlist, &ordwin, text)
#if (DEBUG_GUI > 0)
printf (" .. Selected %d, (%d) \"%s\".\n", 
	new, dispmu.sel, mp->name);  fflush (stdout); //debug
#endif
    }

/*  Fill the gap menu based on instrument */
#ifdef  GAPVO
// call to fill, give gapfield, Gob->obs->Instr_mt
    if (dif) Gob->obs->Dstat = 0;
    fill_gapmenu (&gapfield, Gob->obs->Instr_mt);
#endif

//    Gob->changed |= (instmu.sel != new);
    Gob->changed |= dif;
    instmu.sel = new;
#ifdef  NSROT
    if (instmu.sel == INST_Isc) CBX_EnableButton (&nsrot);
    else {
	CBX_DisableButton (&nsrot, CBX_DISABLED);
// Disable nsrot button, and set bias to zero everywhere.
	nrbiask = 0;
	nsrot.status = 0;
	Gob->obs->nrbias = 0.0;
	slitpaf.offset = 90.0;
	HandleTextWindow (winlist, slitpaf.ew);
    }
#endif

//#ifdef  GISMOMNS
/* Special logic for GISMO MNS case.  If new is INST_Gslx we need to
  set the gap menu to select N&S and disable it.  Otherwise, if not
  INST_Gslx enable the gap menu.  */
//    if (new == INST_Gslx) {	// GISMO MNS needs gaps
//	gapmu.sel = 1;
//	Gob->obs->Dstat = DS_ns | DS_gaps | DS_orient;
//	CBX_DisableMenu (&gapmu, CBX_DISABLED);
//    } else {		// No GISMO MNS, enable gap selection
//	CBX_EnableMenu (&gapmu);
//    }
//#endif

/* The above is not necessary.  The gap menu was correctly set, or
  disabled in the fill_gapmenu call which happened above.  */

    CBX_UpdateMenu(&instmu);	/* Update needed for MacOSX */
}

/*  ------------------------------------------------------------  */
void handle_dispmenu(void)
{
int    new;
IMenuEntry	*dp;
//int	ord;
int	old;
int	inst;

    new = CBX_HandleMenu(&dispmu);
    if (new < 0) return;
    old = dispmu.sel;
    if (new == old && new != 0) return;	// Null case there

/* We skip out if a the instrument currently selected is "direct", so
  as not to choose a disperser for it... */
    if (instmu.sel == INST_Ild) return;
    if (instmu.sel == INST_Isd) return;
    if (instmu.sel == INST_Ld)  return;

/* Get the proper disperser into the structure. */

/* NOTE what we really need here is which menu (grat/gris) is up! (Ityp) */

/* Obtain the optic name of selected entry, and get its element */

/* What we probably need is to change the grisme/gratme to a single
  array, and index the starting part for the instrument, and also
  index the order.  Another array will tell whether we have grating
  or grism for this instrument.
By instrument:  (imacs f/2, imacs f/4, ldss, other) have:
  > starting index of disperser menu array
  > grating or grism value
  > 
By disperser:
  > Usual disperser menu array
  > Order for each disperser

.. */

// Replace above by selection of dp and ord, depending on instrument
    inst = instmu.sel;		// Instrument index

    dp = dispindx[inst] + new;

//    ord = dp->order;		// Order from table

    Gob->obs->Disperser = find_element (Gob->gs, dp->name);
// The gratordr and grisordr are initialized above in the section
// where the grisme dna gratme structures are initialized.
// These set the default order for this GUI; there is no special
// order defined in the optics data file for dispersers...

/* Set the proper default order here, too.  And, update the order
  window to reflect it! */
////    Gob->obs->order = ord;	// Put order into current obs
//    Gob->obs->order = dp->order;	// Put order into current obs

//    HandleTextWindow (winlist, ordfield.ew);	// That's &ordwin there...
    SetIntWindow (winlist, &ordfield, dp->order);

/* If a grism, put the center wavelength in Gob->obs->cw and
  handle that window to write the wavelength... */
    if (instme[inst].type == DTYPE_grism) {
element	*gg;			// local variable as abbreviation
	gg = Gob->obs->Disperser;
	if (gg->kw == EGRIsm) {
//	    *GRangle = ((sdgrism*)gg->data)->grism.angle * Degree;
// Defined as grating angle, we should probably use this...
	    Gob->obs->D_angle = ((egrism*)gg->data)->gangle * Degree;
//	} else {		// otherwise assume a regular grism...
	} else if (gg->kw == GRIsm) {	// otherwise assume a regular grism...
//	    *GRangle = ((stgrism*)gg->data)->angle * Degree;
	    Gob->obs->D_angle = ((stgrism*)gg->data)->angle * Degree;
	} else {
	    Gob->obs->D_angle = 0.0;	// complex disperser
	}
//	*GRorder = (*GRangle < 0.0) ? -1 : 1;
/* Above uses of GRangle and GRorder should set the proper values in
  the obs structure directly as they point to the obs values. */
/* Alternately, we replace with Gob->obs->D_angle... */
// Those have to be set for grwavc to work correctly!
	Gob->obs->cw = grwavc (gg, Gob->obs->temp);
// printf (" -- c.w. set to %.3f\n", Gob->obs->cw);  // debug only
	HandleTextWindow (winlist, cwfield.ew);
//	SetFloatWindow (winlist, &cwfield, grwavc (gg, Gob->obs->temp) );
    }
    check_moe(inst, new);

#if (DEBUG_GUI > 0)
printf (" .. Selected %d, (= %d) \"%s\".\n", 
	new, new, dp->name);  fflush (stdout); //debug
#endif
    Gob->changed |= (dispmu.sel != new);
    dispmu.sel = new;
    CBX_UpdateMenu(&dispmu);	/* Update needed for MacOSX */
}

/*  ------------------------------------------------------------  */
void  handle_filtmenu (void)
{
int    new;

/* Check for any problem or null case */
    new = CBX_HandleMenu(&filtmu);
    if (new < 0) return;
    if (new == filtmu.sel) return;	// Null case
// try checking new vs. the count available in filtmu structure...
    if (new > filtmu.n) return;   // or is that >= that we need ???

/* Appears to be a new case, initialize the windows */
// grab the data from filtelem(new), and force into obs->pb values
// call to update these windows with new data --
// write new values into text fields of the two menus, 
// using the pointers, etc in blupbfld and redpbfld winblocks

// What we do is create a "SetBlockData" routine called with
// args of winblock and d.p. data which will write the data field
// into the window's text and set it into the data pointer;
// use same thing in the handle fblock possibly...
// And, call the block display, (DrawWindow) too.
    SetBlockData (&blupbfld, filters[new].pb.blue);
    SetBlockData (&redpbfld, filters[new].pb.red);

#ifdef  DWAV
    SetBlockData (&bludbfld, filters[new].pb.blue);
    SetBlockData (&reddbfld, filters[new].pb.red);
    Gob->obs->pb2act = 0;
#endif

    Gob->changed |= (filtmu.sel != new);
    filtmu.sel = new;
    CBX_UpdateMenu(&filtmu);	/* Update needed for MacOSX */
}


/*  ------------------------------------------------------------  */
#ifdef  GAPVO
/* Handle the gap menu itself here */
void	handle_gapmenu (void)
{
int    new;
int	ds;
int	is;

/* Find the new menu value */
    new = CBX_HandleMenu(&gapmu);
    if (new < 0) return;

/* Check for any problem or null case */
    if (Gob == NULL) return;
    if (Gob->obs == NULL) return;
    if (Gob->obs->Instr_mt == NULL) return;

/* Set the status word based on values */
    ds = 0;
    if (new > 0) ds = DS_orient | DS_gaps;
//    is = InstIndx (Gob->obs->Instr_mt->name);
    is = InstIndx (ObInstN(Gob->obs));
#ifdef  NSROT
// Set the ns bit based on ns rotation value if short camera
    if (is == I_imacs) {
	if (!strncasecmp(Gob->obs->Instr_mt->name, "IMACS_s", 7)) {
	    if (fabs(Gob->obs->nrbias) > NSrlim) ds |= DS_ns;
	} else {
	    if (new == 1) ds |= DS_ns;
	}
    }
#else
    if (is == I_imacs && new == 1) ds |= DS_ns;
#endif  // on NSROT

/* Put the status word into the global obs data */
    Gob->changed |= (Gob->obs->Dstat != ds);
    Gob->obs->Dstat = ds;
    if (Gob->obs->dlevel > 1) printf (" Dstat set to %d\n", ds);  // debug only
    gapmu.sel = new;
    CBX_UpdateMenu(&gapmu);	/* Update needed for MacOSX */
}
#endif  // on GAPVO

/*  ------------------------------------------------------------  */
#ifdef  GAPVO
void	fill_gapmenu (mnublock* gm, element* inst)  {
Menu	*m;
Indx	is;
int	ds;

// Leave early if pointers are null
    if (gm == NULL) return;
//    if (inst == NULL) return;

// To start, empty the menu of its content.
// see also FillMenu in guiutil.c
    m = gm->mu;
    m->n = 0;	// clear the menu

// Find the instrument from its name;
    if (inst == NULL) {
	is = I_other;
    } else {
	is = InstIndx (inst->name);
    }

    ds = Gob->obs->Dstat;
// If imacs fill the menu with imacs things, and enable
    if (is == I_imacs) {
	CBX_AddMenuEntry (m, "-off-", 0);
/*  Check long/short camera.  If long we can have 3 values; but if short
  camera, the gaps is only off/on as the n&s vs norm orientation is done
  by the nsrotation bias when the NSROT feature is active.  */
#ifdef  NSROT
/* Added section for short camera in the case NSROT is active */
      if (!strncasecmp(inst->name, "IMACS_s", 7)) {	// Short camera
	CBX_AddMenuEntry (m, "-on-", 0);
	CBX_EnableMenu (m);
	if (ds & DS_gaps) m->sel = 1;
	else  m->sel = 0;
      } else {   // not short camera section follows
#endif	// on NSROT
/* This section is for long camera, and any camera without NSROT */
	CBX_AddMenuEntry (m, "N & S", 0);
	CBX_AddMenuEntry (m, "Norm.", 0);
	CBX_EnableMenu (m);
	if (ds & DS_gaps) {
	    m->sel = 2;
	    if (ds & DS_ns) m->sel = 1;
	} else  m->sel = 0;
/* This section is for short camera with NSROT where n&s is selected by
  the value of the rotation bias */
#ifdef  NSROT
      }   // end short camera conditional
#endif  // on NSROT
#ifdef  GISMOAP
// For any GISMO aperature mask, disable this feature
      if (!strncasecmp(inst->name, "IMACS_gd", 8)) {
// don't need to introduce FeatIndx here, just compare name.
	m->sel = 1;
	CBX_DisableMenu (m, CBX_DISABLED);
      }
#endif
    } else if (is == I_lds3) {
// if ldss, fill the menu, and enable
	CBX_AddMenuEntry (m, "-off-", 0);
#ifdef  GAPLDSS
	CBX_AddMenuEntry (m, "-on-", 0);
// use calls to CBX_AddMenuEntry to fill the menu.
	CBX_EnableMenu (m);
	if (ds & DS_gaps) m->sel = 1;
	else              m->sel = 0;
#else
	CBX_DisableMenu (m, CBX_DISABLED);
	m->sel = 0;
#endif  // on GAPLDSS
    } else {
	CBX_AddMenuEntry (m, "-off-", 0);
// otherwise, disable the menu itself.
	CBX_DisableMenu (m, CBX_DISABLED);
	m->sel = 0;
    }
#ifdef  GISMOMNS
/* Special -- Set select to 2 and disable menu for special instrument */
    if (is == I_imacs) {
// Also verifies that inst->name exists if we got is above correctly
	if (!strncasecmp(inst->name, "IMACS_xl", 8)) {
	    ds = Gob->obs->Dstat = DS_ns | DS_gaps | DS_orient;
	    m->sel = 1;  // so long as we are not short camera...
	    CBX_DisableMenu (m, CBX_DISABLED);
	}
    }
#endif
}
#endif  // on GAPVO


/*  ------------------------------------------------------------  */
/*  ------------------------------------------------------------  */
/*    End of things shuffled from guisubs.c  */
/*  ------------------------------------------------------------  */
/*  ------------------------------------------------------------  */
/*  ------------------------------------------------------------  */

// The pa and cp things are used a little in skymap; they should
// be replaced by more rational things, however...
// ONLY get_cp is referenced in skywin.c
/* ---------------------------------------------------------------------- */

float get_pa(float cp)
{
  return((float)(180.-(double)cp));
}

/* ---------------------------------------------------------------------- */

float get_cp(float pa)
{
    return((float)(180.-(double)pa));  // real method
}

/* ---------------------------------------------------------------------- */


/*  ------------------------------------------------------------  */

/*  ===  Main GUI sub-task  ===  */

void	pgui (Obf*  ob, int  h)
/* Plot the GUI itself */
{

/* Local variables */
int	dflag = 1;
XEvent		event;
XSizeHints	hint;
XWindowAttributes	xwa;
char		xserver[80];
char		title[80];
char		I_title[32];
char		fontname[80];
char	dbuffer[64];	// debug only
int		sleeptime;
element		*eptr;
IMenuEntry	*mp;
int		ival, dval;
double		xmul, ymul;
double		cjd;
int		y,m,d;
GfC		mgfc;	// debug
int		events=0;
int		totevents=0;

#ifdef  SKYOPT
char	*filespec = NULL;
char	datfile[]="pmap.dat";
#endif

/* Make up title string */
// Used as window title and icon title...
    sprintf (title, "SlitMask Interface -- Ver. %s", XVERSION);
    sprintf (I_title, "SM-GUI");

/* Test for proper argument */
    if (ob == NULL) {
	printf (" ** NULL observation function structure!\n");
	return;
    }

#ifdef  SKYOPT

    data_file (&filespec, datfile, "DEFAULTS", 1);
//    if (filespec == NULL) return;
    if (filespec == NULL) {
	fprintf (stderr, " ** pmap.dat file not available.\n");
	fprintf (stderr, " Set \"DEFAULTS\" or obtain pmap.dat,\n");
	fprintf (stderr, " Or set %s to include possible directories.\n", PathName);
	return;
    }

    if (ob->obs->dlevel > 0) {  /* Show the file used. */
	printf (" -pmap data from: %s \n", filespec);
    }

    GDAT = (gdata*) get_data (
	filespec, 	// Filespec of the data file
	*gdatstore,	// Program to store data in structure
	WHITESPACE,  	// Delimiter list
	"!#",  		// Entire line comment introducers (first col.)
	"!",		// Comment introducers appearing anywhere
	"#;",  		// Comment if preceeded by a delimiter
	"",  		// Comment if followed by delimiter
	"" );		// Comment if preceeded and followed by delimiter

    free (filespec);

// Call to setup the tel,cam,gdr arrays of section pointers from GDAT;
// do that exactly once here.
	tsections = get_sections (MNU_TEL);
	gsections = get_sections (MNU_GDR);
	csections = get_sections (MNU_CAM);

    setup_defaults ();	// Do here, and if default sections change
//    map_savn = 0;	// Start with clean sky window

#endif  // SKYOPT

//#if (DEBUG_GUI > 1)
    if (ob->obs->dlevel > 0) printf (" ++ pgui starting...\n");	// debug
//#endif

/* Set up globalization kludge */
    Gob = ob;

/* Get a display */
    strcpy (xserver, genv2("DISPLAY","") );
    if ((app=CBX_InitXlib(xserver)) == NULL) {      /* connect to 'xserver' */
	fprintf(stderr,"%s: CBX_InitXlib(%s) failed\n", "Interface",xserver);
	exit(1);
    }


/* Start up a window */

/* Make sure subsequent windows start */
    done = False;

// EDIT TO USE SIZES, ETC FROM intgui.h HERE!

/* Set dynamic font size stuff... */

/* The call parameter "h" is the requested font size.  If 0, use
  the default.  If too large, reduce it to the maximum possible, computed
  from the display size and computed window sizes.  */

    if (h > 0) {
	if (h < 9) h = 9;		// Enforce minimum size!
	Vunit = (double) h / (double) Yp;
	Hunit = Vunit * (4.0*((double)Yp +0.5)) / (double)(7*Xp);
// Testing a smaller Hunit as some spaces seem to large...
/* Check for too large... */
	xmul = (double)Xloc(Hsize) / 1270.0;
	ymul = (double)Yloc(Vsize) / 1024.0;
	if (ymul > xmul) xmul = ymul;
	if (xmul > 1.0) {
	    Vunit /= xmul;
	    Hunit /= xmul;
	    h = (double)h / xmul;
// NOTE -- WE HAVE TO REDUCE H ENOUGH THAT AT LEAST XMUL REDUCTION
// IS ACHIEVED IN VUNIT,HUNIT VALUES AND RECOMPUTE VUNIT, HUNIT.
	}
    }
//    FontSize = 18;	// Store as a global
// SHOULD THAT BE FontSize = h ????
    FontSize = h;

//#if (DEBUG_GUI > 0)
if (ob->obs->dlevel > 0)
printf (" Font %d: Hunit=%.3f, Vunit=%.3f, Hsize=%d, Vsize=%d\n",
	FontSize, Hunit, Vunit, Hsize, Vsize);
if (ob->obs->dlevel > 0)
printf (" Window creation at %d by %d\n",
	Xloc(Hsize), Yloc(Vsize) );
//#endif

/* The definitions of Vunit and Hunit (which should be equal) define the
  font size and window sizes. */
/* Converting old way to new way, 
	PXh = Vunit * Yp;
	PXw = Hunit * Xp;
	XXh = PXh + Xp;  .. */

    win = CBX_OpenMainWindow (
	&mwin, 
//	CBX_BOTTOMRIGHT, 
	CBX_MIDRIGHT, 
	Xloc(Hsize), Yloc(Vsize),		/* Window size */
	&hint, 
	title, I_title, 
	True);

dpause ("Opened window.");

    strcpy(fontname,genv2("IMACSFONT",DEF_FONT)); /* default font */

if (ob->obs->dlevel > 0) printf (" Font = %s at size %d\n", fontname, FontSize);

//    (void)CBX_CreateGfC (&mwin, fontname, "bold", FontSize);
    mgfc = CBX_CreateGfC (&mwin, fontname, "bold", FontSize);
if (ob->obs->dlevel > 0) printf (" Create font f_s=%p for FontSize %d\n",
  (void*)mgfc.f_struct, FontSize );	// debug

/* Here, we test to see if the font was properly created using the
  FontSize parameter.  If mgfc.f_struct is null, the CreateGfC was not
  able to find the right font.  To prevent a segmentation fault when
  that bad font is referenced, we need to exit now with an error
  message.  Some font sizes are just bad, while others work.
  It would be possible to try other FontSize parameters near the current
  one and perhaps find one which works; but two ugly error messages have
  already been produced for each one which fails.
  In linux, these work:  10, 11, 12, 14, 17, 18, 19, 20, 24
  and these fail:  8, 9, 13, 15, 16, 21, 22, 23 
  The default size is 18.  */

    if (mgfc.f_struct == NULL) {
	printf (" Bad font size.  Try another using -f<nnn> parameter.\n");
	exit (1);
    }

/* -- The file menu -- */

dpause ("Start to create menu");
    MakeMenu (&mnulist, &iqfield);

#ifdef  IMAGM
    if (IMAGUI == 0) {
// Omit all these things when in image mode...  See handle_iqmenu
#endif
    (void)CBX_AddMenuEntry(&iqmenu,"  Save",    0);
    (void)CBX_AddMenuEntry(&iqmenu," Save As", 0);
    (void)CBX_AddMenuEntry(&iqmenu,"",        CBX_SEPARATOR);
    (void)CBX_AddMenuEntry(&iqmenu," SavExit", 0);
    (void)CBX_AddMenuEntry(&iqmenu,"Quit nosav", 0);
#ifdef  IMAGM
    }
#endif
    (void)CBX_AddMenuEntry(&iqmenu,"  Exit",    0);

/* -- The options menu -- */

#ifdef  SKYOPT
    MakeMenu (&mnulist, &OPTfield);
// later we add pointer to menu block queue here, just like button, window.
    CBX_AddMenuEntry (&optmenu, "Skywin",  CBX_NORMAL);
//    skywin_active = 0;
    CBX_CheckMenu (&optmenu, OPT_SKYWIN, skywin_active);
/* The option for skywin gets set in the main program at startup, and a
  command line flag can set it to be on at start.  This will make the
  menu synchronized with the selection.  It is external global.  */
#ifdef  IFUM
    CBX_AddMenuEntry (&optmenu, " IFU ",  CBX_NORMAL);
    IFU_active = (ob->obs->IFUmode != 0);
    CBX_CheckMenu (&optmenu, OPT_IFU, IFU_active);
// We check the menu entry based on the activity flag
#endif
#endif

// add other buttons and menus; see gmap.c and dbplot.c for examples

/* -- The active MaskGen button -- */

#ifdef  IMAGM
    if (IMAGUI) strcpy (actfield.label, "Obscat");
#endif
    MakeButton (&butlist, &actfield);

/* Kill button for test mode only  -- on tintgui test mode (kilbut) */
    if (!strncmp(myname, "tintgui", 7)) MakeButton (&butlist, &killfld);

/* -- Sub windows for Identification -- */

//dpause ("Start to create subwindows");
/* Data file name */
#ifdef  IMAGM
    if (IMAGUI) {
// Change label of "observing file" to "object name"
	strcpy (datafile.label, "Object Name:");
    }
#endif
    AutoEditWin (&winlist, &datafile, Gob->obs->fname );

#ifdef  IMAGM
    if (IMAGUI == 0) {
// Omit all these things when in image mode...  See handle_iqmenu
#endif
//dpause ("Did datafile, next observer");
/* Observer name */
    AutoEditWin (&winlist, &observer, Gob->obs->oname );
#ifdef  IMAGM
    }
#endif

#ifdef  IMAGM
    if (IMAGUI) {
	strcpy (titlefld.label, "Obscat file:");
	titlefld.width = 20;
	if (Gob->obs->title == NULL) Gob->obs->title = calloc (1,32);
	strcpy (Gob->obs->title, "observing.cat");
    }
// Omit all these things when in image mode...  See handle_iqmenu
#endif
//dpause ("Next is title");
/* Title */
    AutoEditWin (&winlist, &titlefld, &(Gob->obs->title) );
//#ifdef  IMAGM
//    }
//#endif
/* -- Sub windows for Position -- */

#ifdef  IMAGM
// Establish new defaults for position in image mode
    if (IMAGUI) {
	Gob->obs->ra  = 6.0 * Hour;
	Gob->obs->dec = 0.0;
    }
#endif
//dpause ("Next is ra/dec positions");
/* R.A. and Dec. window creation... */
    AutoEditWin (&winlist, &RAfield, &(Gob->obs->ra) );
    AutoEditWin (&winlist, &DECfield, &(Gob->obs->dec) );

    init_gcenter();	// Initialize with  current position

/* Equinox window */
//    (void)CBX_CreatePopupMenu(&mwin, &eqmenu,
//		Xloc(eqfldx),Yloc(eqfldy), Xsize(eqfldw),thp);

    MakeMenu (&mnulist, &eqfield);

    CBX_AddMenuEntry(&eqmenu,"1950",  0);
    CBX_AddMenuEntry(&eqmenu,"2000",  0);
    CBX_AddMenuEntry(&eqmenu,"today", 0);
    CBX_AddMenuEntry(&eqmenu,"set",   0);
    if        (Gob->obs->equinox == 1950.0) {
	eqmenu.sel = EQU_1950;
    } else if (Gob->obs->equinox == 2000.0) {
	eqmenu.sel = EQU_2000;
    } else {
	eqmenu.sel = EQU_SET;
	sprintf(eqmenu.entry[EQU_SET].text,"%.2f",Gob->obs->equinox);
    }
    wepoch = Gob->obs->equinox;
//    solpos (mjd);	// Needed since epoch was set

/* Slit position angle -- instrument angle */
//#ifdef  IMAGM
#ifdef  NoSuchThing
    if (IMAGUI) {
// Special label/offset for image mode
	strcpy (slitpaf.label, "Pos. Angle");
	slitpaf.offset = 0.0;
    }
#endif

    AutoEditWin (&winlist, &slitpaf, &(Gob->obs->angle) );

/* Tempereature */
// NOTE -- this section should probably be dropped entirely?
// i.e. no temperature window support.
#ifndef  GUIDATE
    AutoEditWin (&winlist, &tmpfield, &(Gob->obs->temp) );
#endif

/* Hour Angle */
#ifdef  DIFREF
    AutoEditWin (&winlist, &hafield, &(Gob->obs->ha) );
#endif

/* -- Instrument and Disperser menus -- */

/* Instrument menu */
#ifdef  IMAGM
/* If image mode selected, omit making instrument, disperser windows at all;
  i.e. do the following only if image mode is false. */
  if (IMAGUI == 0) {
#endif
//printf (" ..Starting instrument menu\n"); fflush(stdout);  //debug
//    MakePMenu (&mnulist, &instfield);
    MakeMenu (&mnulist, &instfield);
//    Ityp = FillMenu (&instmu, instme, INST_sel);
    FillMenu (&instmu, instme, INST_sel);	// ignore return
// Find possible instrument from the obs structure
    eptr = Gob->obs->Instr_mt;
    ival = eptr == NULL ? INST_sel : Mselectn (eptr->name, instme);
//#ifdef  GISMO
#if  defined GismoMenus
    eptr = Gob->obs->Frontend;
//    if (eptr != NULL) ival = Mselectn (eptr->name, instme);
// gismo is not in instme except as display name.
    if (eptr != NULL) {
	if (strncasecmp(eptr->name, "gismo", 5) == 0) {  // is gismo
	    if      (ival == INST_Ilc)  ival = INST_Gsl;
/* Do we need to set ival to INST_Gslx if gismo with macro n/s is
  selected?  Probably put a test for gismomns text before this section
  to solve for that case, and key on GISMOMNS symbol... */
	    else if (ival == INST_Isc)  ival = INST_Gss;
/* An alternate would be to add the opticdef name of the frontend as a
  separate entry in the IMenuEntry structure and test on both the instrument
  and front end names to find the proper entry.  */
//	    else if (ival == INST_Gsal)  ival = INST_Gsal
#ifdef  GISMOAP
	    else if (!strcasecmp(eptr->name, "IMACS_gdl")) ival = INST_Gsal;
//	    else if (!strcasecmp(eptr->name, "IMACS_gds")) ival = INST_Gsas;
#endif
	}
    }
#endif  // on GISMO -- GismoMenus
// Fix for new LDSS support method -- get mp from a table, not this way
/*
  What we are trying to do is:
    1) Find the instrument index of the given instrument name
    2) Get the proper pointer for the disperser table
    3) Set up the instrument menu and selection
    4) Derive mp pointer for filling the disperser menu
.. */
/* We will need to replace instmpn and dispndx programs with a
  unified subroutine to look up a name in a table...  Mselectn
  Ultimately similar to dispndx.
.. */

    mp = dispindx[ival];

    instmu.sel = ival;	// inst. menu is straight

/* Disperser menu */
//printf (" ..Starting disperser menu\n"); fflush(stdout);  //debug
    MakeMenu (&mnulist, &dispfield);
    dval = FillMenu (&dispmu, mp, DISP_sel);  /* Note - mp selected above */
    eptr = Gob->obs->Disperser;
    dval = eptr == NULL ? DISP_sel : Mselectn (eptr->name, mp);
//	printf ("  dval = %d for %s\n", dval, eptr->name);  // debug for oxxxo
//    dispmu.sel = Mselectv (mp, dval);
// Replacing Mselectv with dispndx here...
    dispmu.sel = dval;
//NOTE fix Mselectv to return index in mp table of dval  
// Disable menu if no instrument is selected yet
/* 
  What we try to do here is:
    1) Knowing the instrument (above) find the disperser
	within the disperser table
    2) Set up the disperser menu and selection
.. */
    if (ival == INST_sel
//#ifdef  GISMOAP
#if  defined GISMOAP && defined GismoMenus
// These are needed for gismo aperature feature
	|| ival == INST_Gsal  /* || ival == INST_Gsas */
#endif
       ) CBX_DisableMenu (&dispmu, CBX_DISABLED);
#ifdef  IMAGM
  }	// End of conditionally applied if above.
#endif

#ifdef  IMAGM
/* If image mode selected, we need to do this:
  1) Select IMACS f/2 as instrument in instmu.sel
  2) Select the dispmu.sel to dval
  3) Disable dispmu
  4) Disable instmu
  .. */
  if (IMAGUI) {
    instmu.sel = ival = INST_Isc;
    dispmu.sel = dval = DISP_sel;
    CBX_DisableMenu (&dispmu, CBX_DISABLED);
    CBX_DisableMenu (&instmu, CBX_DISABLED);
//    if (Gob->obs->Instr_mt == NULL) 
	Gob->obs->Instr_mt = find_element (Gob->gs, "IMACS_short");
//    else  Gob->obs->Instr_mt = ??
/* If we need instrument selection, pay more attention to the setting of
  the default instrument as short, and the menu handler needs to have
  conditional code about IMAGM to allow only IMACS and no dispersers.  */
  }
#endif

#ifdef  IMAGM
/* If image mode selected, omit making filter, passband windows at all;
  i.e. do the following only if image mode is false. */
  if (IMAGUI == 0) {
#endif
/* Filter menu and passband windows set-up */
    MakeFiltmenu (&mnulist, filters, &filtfield, filtme);
    FillMenu (&filtmu, filtme, 0);
// Add creation of the passband windows too...
    AutoEditWin (&winlist, &blupbfld, &(Gob->obs->pb.blue) );
    AutoEditWin (&winlist, &redpbfld, &(Gob->obs->pb.red) );
    AutoEditWin (&winlist, &cwfield,  &(Gob->obs->cw) );
    if (instme[ival].type != DTYPE_grating) 
	CBX_DisableEditWindow (&cwwin, CBX_DISABLED);
#ifdef  IMAGM
  }	// End of conditionally applied if above.
#endif

#ifdef  GUIDATE
/* Find current time -- jdnow()  then subtract 2400000.0 and
  another 0.666666 days, and truncate to current mjd.  From that,
  use date4mjd to find current y,m,d and set menu to y-2, y+3 range;
  and store current mjd to warn about date less than 30 days away. */
	cjd = jdnow();
	cepoch = Jyear (cjd);	// current epoch for precession
	cjd -= 2400000.0;
	cjd -= 0.6666666666667;
	cmjd = (int)cjd;	// current mjd for date test
	date4mjd (cmjd, &y, &m, &d);

/* Date menus - year month and day.. */
//    MakeNumMenu (&mnulist, 2003, 2008, &yearfield);
    MakeNumMenu (&mnulist, y-2, y+3, &yearfield);
/* An improvement would be to find current year, and select values
  for year-1 to year+3 or so. */
    MakeMonthMenu (&mnulist, &monthfield);
    MakeNumMenu (&mnulist, 1, 31, &dayfield);
	setup_dates();
//	yearmnu.sel = yr - baseyear;
//	monthmnu.sel = mo;
//	daymnu.sel = dy;
#endif  // on GUIDATE

#ifdef  IMAGM
    if (IMAGUI == 0) {  // Do not start these windows in image mode.
#endif
/* Extra windows for detector pass band definition */
#ifdef DWAV
    AutoEditWin (&winlist, &bludbfld, &(Gob->obs->pb2.blue) );
    AutoEditWin (&winlist, &reddbfld, &(Gob->obs->pb2.red) );
#endif

/* Order */
    AutoEditWin (&winlist, &ordfield, &(Gob->obs->order));

/* -- Slit parameter sub windows -- */

/* Slit Width */
//printf (" ..Location 2.\n");  fflush (stdout);  //debug

//    FocalLength = foclen (Gob->obs);

/* With focal length found above, we set actual scale to be used;
  The astronomer's display has arcseconds, yet the Obs structure
  stores mm in the focal plane for slit size specifications.  */
//    swidfield.scale = ArcSecond * FocalLength;

/* Obs structure now storing default slit spec's as arcseconds */
    swidfield.scale = 1.0;

//printf (" ..Location 3.\n");  fflush (stdout);  //debug

    AutoEditWin (&winlist, &swidfield, &(Gob->obs->default_slit.width) );
// Apply the same scale factor to all other slit width things.

//    slafield.scale = ArcSecond * FocalLength;
    slafield.scale = 1.0;
    AutoEditWin (&winlist, &slafield, &(Gob->obs->default_slit.alen) );
//    slbfield.scale = ArcSecond * FocalLength;
    slbfield.scale = 1.0;
    AutoEditWin (&winlist, &slbfield, &(Gob->obs->default_slit.blen) );
// Drop the slit tilt window
//    winlist = AutoEditWin (winlist, &sltfield, &(Gob->obs->default_slit.angle));

    AutoEditWin (&winlist, &dcafield, &(Gob->obs->dca) );
    AutoEditWin (&winlist, &dcbfield, &(Gob->obs->dcb) );

    AutoEditWin (&winlist, &ovlfield, &(Gob->obs->minsep) );

#ifdef  SLXB

/* -- Slit extension feature button -- */

//    MakeDButton (&butlist, &slxfld, &(Gob->obs->slext), app->BUT_UNSEL);
    MakeDButton (&butlist, &slxfld, &(Gob->obs->slext), app->BUT_SEL);
    if ( *(double*)dcafield.data != 0.0 ||
	 *(double*)dcbfield.data != 0.0 ) {   // disable button
	CBX_DisableButton (&slext, CBX_DISABLED);
    }
// We do it this way since the dcafield and dcbfield have already been
// set up above.  If not, we would test Gob->obs->dca, etc.
#endif

#ifdef  DIFREF
    MakeDButton (&butlist, &drefld, &(Gob->obs->dref), app->BUT_SEL);
// if dref is on/off, enable/disable the ha window
    if (Gob->obs->dref) CBX_EnableEditWindow (&hawin);
    else  CBX_DisableEditWindow (&hawin, CBX_DISABLED);
#endif

/*  The section to set up slit shape menu, and to check on and disable
  the length flags if hole/box requires a modification to check_moe since
  hole shape is fixed for MOE use.  With the proper modification, both
  will work; slit shape can disable slit length windows, and selection
  of MOE can disable slit shape menu too.  */

#ifdef  OBJSHAPE
/* Set up the  -- slit shape menu --   */
    MakeMenu (&mnulist, &ssfield);
    CBX_AddMenuEntry (&shapemnu, "Slit", CBX_NORMAL);
    CBX_AddMenuEntry (&shapemnu, "Hole", CBX_NORMAL);
    CBX_AddMenuEntry (&shapemnu, "Box",  CBX_NORMAL);
    switch (Gob->obs->default_slit.shape) {
	case	RECTANGLE:	shapemnu.sel = 0;	break;
	case	CIRCLE:		shapemnu.sel = 1;	break;
	case	SQUARE:		shapemnu.sel = 2;	break;
    }
    if (Gob->obs->default_slit.shape != RECTANGLE) {	// Disable lengths if not slit
//printf (" -*disable slit length windows*-\n");  // debug only
	CBX_DisableEditWindow (&slawin, CBX_DISABLED);
	CBX_DisableEditWindow (&slbwin, CBX_DISABLED);
	CBX_DisableEditWindow (&dcawin, CBX_DISABLED);
	CBX_DisableEditWindow (&dcbwin, CBX_DISABLED);
/* Also disable slit extension and clear the extension flag (slext) */
#ifdef  SLXB
	if (Gob->obs->slext != 0) Gob->obs->slext = 0;
	if (slext.flag == CBX_NORMAL) CBX_DisableButton (&slext, CBX_DISABLED);
#endif

    }
#endif  // on OBJSHAPE

    MakeMenu (&mnulist, &exordfld);
    CBX_AddMenuEntry (&xordmu, " - Off -",  CBX_NORMAL);
    CBX_AddMenuEntry (&xordmu, "Reference",  CBX_NORMAL);
//    CBX_AddMenuEntry (&xordmu, "  - - -  ",  CBX_SEPARATOR);
    CBX_AddMenuEntry (&xordmu, " - All -",  CBX_NORMAL);
    {	// local variable section
    int	mm;
	mm = Gob->obs->ex_order;
	if (mm == 3) mm = 2;
	xordmu.sel = mm;
    }

//    MakeDButton (&butlist, &adcfld, &NoADC, app->BUT_SEL);
// The ADC option is removed 2004/6 with ADC permanently installed

// creation of slit extend moved up before slit shape from here

#ifdef  IMAGM
    }  // end of conditional above
#endif
    check_moe (instmu.sel, dispmu.sel);	// Set correctly if just read in...


//    MakeDButton (&stbfield, NULL);  //slitype   	//stbut

/*  Change as follows:
   the butlist will be global, otherwise like the autoeditwin
   put pointer to gob-obs-value into call
  We need such a call for the extra orders feature toggle.
   For slit type, we can retain the global data value and sync
   that to the alignment slit type at the "sync" call. (normalize_obs)
   Or a call done like/with "TextScan" to update botton stuff...
   Possibly add a button section to TextScan sync routine.
  (see that in mgutils.c; the value toggled by the button, and the value
   of the value of alignment hole shape needs to be synced. )
  Also, the toggled value of extraorder is in obs->exorder.
  NEEDS POINTER TO FLAG FOR SLIT TYPE...
-- */

/* -- Button to bring up list of object files -- */
#ifdef  IMAGM
  if (IMAGUI == 0) {
#endif
    MakeButton (&butlist, &objlfld);	// Object list
#ifdef  IMAGM
  }
#endif


/* -- Optional test button -- */

#ifdef  WDEBUG
    if (!strncmp(myname, "tintgui", 7))
	MakeButton (&butlist, &tbfield);	// Test button
#endif

#ifdef  IMAGM
  if (IMAGUI == 0) {
#endif
//    rwidfield.scale = ArcSecond * FocalLength;
    rwidfield.scale = 1.0;
    AutoEditWin (&winlist, &rwidfield, &(Gob->obs->reference_slit.width) );

/* -- Selection and priority sub-windows -- */

// Windows for entry of repeat count and select criteria
    AutoEditWin (&winlist, &robfield, &(Gob->obs->reuseob) );
    AutoEditWin (&winlist, &rrefield, &(Gob->obs->reuserf) );
#ifdef  REFLIM
    AutoEditWin (&winlist, &rsefield, &(Gob->obs->reflimit) );
#else
    AutoEditWin (&winlist, &rsefield, &(Gob->obs->refsel) );
#endif

#ifdef  EXTPRI
    AutoEditWin (&winlist, &xpfield, &(Gob->obs->apri) );
#endif

// #ifdef  WDEBUG
// Priority level selection
    AutoEditWin (&winlist, &mhpfield, &(Gob->obs->Pmusthave) );
    AutoEditWin (&winlist, &pdfield,  &(Gob->obs->Pdecide) );
// #endif

#ifdef  IMAGM
  }  // end of conditional
#endif

/* -- Rotator flip bar feature set-up -- */

/* Some other fields need to be computed... */
    {
    double	dec, ipa;
//    anglist	*al;
    int		doqflip = 1;
    int		Indx;
	dec = *(double*)DECfield.data / Degree;
	ipa = *(double*)slitpaf.data / Degree;
//	hrf = gflip (dec, lat, ipa, 1);	// Side is 1
	pdx.dec = dec;
	pdx.lat = lat;
	pdx.pa  = ipa;
	pdx.side = 1;	// IMACS side is 1

/* Set doqflip based on the instrument in use.  */
	Indx = InstNdx (Gob->obs);
	if (Indx == I_imacs) doqflip = 1;
	if (Indx == I_lds2 || Indx == I_lds3) doqflip = 2;
	if (Indx == I_cfg) doqflip = 0;
	if (Indx == I_wfcd) doqflip = 0;
#ifdef  WFCCD
/* Only find rotaq if Magellan instrument; omit this on others such as wfccd...
  We assume rotaq is NULL on entry here.  If that is not the case,
  drop all rotaq entries and only compute if right instrument.  See guisubs.c
  for how to drop entries in "rotaq".   */
/* Use qkill to clean the queue.  Use G_pop_l to pop, G_push_l to insert.  */

//	if (InstNdx (Gob->obs) != I_wfcd)
//	if (InstNdx (Gob->obs) == I_wfcd) doqflip = 0;
#endif
#ifdef  GISAP
	if (FeatIndx(Gob->obs) == F_gisap) doqflip = 0;
#endif

//	find_rotator (doqflip, dodebug > 1);
// The above is redundant, as find_rotator is called in text_update which
// is done here anyway.  So we should just do the messages here ---
	if (dodebug > 0) {  // Some rotator messages
	    printf (" Warn rotator %.2f %.2f\n", RotLim2, RotLim1);
	    if (-IMACSlimit > RotLim1)
	    printf (" Warn IMACS angle > %.2f\n", IMACSlimit + IMACSrot);
	}
// NOTE  -- now we can remove the second argument to find_rotator
// as it is actually redundant.

// In IMAGM case, put all this on a conditional...
#ifdef  IMAGM
    if (IMAGUI == 0) {
// Setting offset is not done in the image mode case.  At all.
#endif
// Initialize the rotation offset
#ifdef  NSROT
	if (fabs(Gob->obs->nrbias) > NSrlim) {
	    nrbiask = 1;
	    slitpaf.offset = 90.0 - Gob->obs->nrbias/slitpaf.scale;
	} else {
	    nrbiask = 0;
	    slitpaf.offset = 90.0;
	}
// Also initialize detector status for gap orientation
	if (nrbiask) Gob->obs->Dstat |= DS_ns;
	HandleTextWindow (winlist, slitpaf.ew);
#endif
#ifdef  IMAGM
    }
#endif

    }

#ifdef  IMAGM
  if (IMAGUI == 0) {
#endif

#ifdef  NSROT
    MakeDButton (&butlist, &nsrfld, &nrbiask, app->BUT_UNSEL);
// Enable the button for imacs f/2 instrument only
    if (instmu.sel == INST_Isc) CBX_EnableButton (&nsrot);
    else         CBX_DisableButton (&nsrot, CBX_DISABLED);
// also enable/disable in menu handlers there based on imacs/grism
// The togglebutton handler in guisubs.c sets bias based on nrbiask
// need the nrbiask (an int value) to be globally defined!
#endif


/*  -- Menu for gap feature... -- */
#ifdef  GAPVO
    MakeMenu (&mnulist, &gapfield);
// call to fill in based on current instrument...
// parameters to fill:  gapfield, Gob->obs->Instr_mt
//void	fill_gapmenu (mnublock* gm, element* inst)  {
    fill_gapmenu (&gapfield, Gob->obs->Instr_mt);

#endif

#ifdef  IMAGM
  }
#endif

    solpos (mjd);	// Finish initialization with setup here.

#ifdef  GISMOAP
    if (Gob != NULL) {
	if (FeatIndx(Gob->obs) == F_gisap) {
	    if (Gob->obs != NULL) {
		Gob->obs->reflimit = 0;
// clear position windows:  alwin, dewin, spawin, eqmenu
// set (ob->) ra, dec, angle  to zero
// Set ra, dec by setting rafc, defc -- not needed
	    }
	    disable4ap(1);
	}
/* disable this...
//char	*inam;
//	inam = ObInstN(Gob->obs);
//// We have to be so <blankety> careful here...
//	if (inam != NULL) {
//	    if (strncasecmp(inam, "IMACS_gd", 8) == 0) disable4ap(1);
//	}
.. to this */
    }
    // not handle_instmenu();
#endif

#if (DEBUG_GUI > 0)
printf (" Initialized, start window loop.\n"); fflush (stdout);  //debug
#endif

    ob->changed = 0;
dpause ("Done width fields.  Next starts loop.");

// Also, will need to add a large loop handling status here!
/* ----------------- The Loop ------------------ */

	/* - - - - - - - - - - *\
	|			|
	|      The  Loop  	|
	|			|
	\* - - - - - - - - - - */

/* Loop for mouse and keyboard inputs */
    sleeptime = 10;	// initial value
    while (!done) {
//	sleeptime = 10;  // was 30;

	while (!CBX_AutoEvent(&mwin,&event,&sleeptime)) {
#ifdef  SKYOPT
//	    (void) skywin_handle();
	    events += skywin_handle (&sleeptime);
#endif
#ifdef  IFUM
	    (void) IFU_handle(&sleeptime);
// At later time, this will return an event count too, used to
// increment the events counter as for skywin.
/* Pass the sleep time on to the inner handler; it will get reset if
  that inner handler detects an event.  This will get us out of this
  loop quickly.  */
#endif
	    if (done) break;			/* break event loop */
//	    if (dflag || warnflag) break;   /* Get out if work to be done */
	    if (warnflag) break;   /* Get out if work to be done */
// MAKE SURE dflag and warnflag ARE CLEARED IF FOUND SET!
/* The above break conditionals are not needed and should be removed;
  if needed to exit due to an external event, simply pass the pointer
  to sleeptime to the inner loop handler as done for IFU_handle; it
  will get reduced if an event is detected... */
/*  The warning flag is set in textback and the button callback which
  are done automatically by AutoEvent.  They don't have a handle to
  the sleeptime, so they need another way to get out.  However, dflag
  is local to this program, and is probably not needed in this loop.  */
	    if (events > 0) {
		totevents += events;
		events = 0;
		continue;	// i.e. don't sleep at all if found some.
	    }
dpause ("In auto event loop.");
//	    CBX_Sleep(&sleeptime,20,300);
	    CBX_Sleep(&sleeptime,15,200);
	}
	if (done) break; 

// printf (" ..Loop entered, for event %d.\n", event.type);  fflush (stdout);  //debug

//dpause ("Top of loop.");
sprintf (dbuffer, "Top of loop, type %d.", event.type);
dpause (dbuffer);
	switch (event.type) {

	case KeyPress:		/* 2 */
// printf (" ..KeyPress (2) found.\n");  fflush (stdout);  //debug
	    break;


	case ButtonPress:	/* 4 */
// printf (" ..ButtonPress (4) found.\n");  fflush (stdout);  //debug
#ifdef  SHFTCLK  // was just: __APPLE__
/* For Apple computers, shift-click should be center click */
	if (event.xbutton.state & ShiftMask) event.xbutton.button = 2;
#endif
//	    if      (event.xbutton.window == iqmenu.win) handle_iqmenu();
	    if      (event.xbutton.window == instmu.win) handle_instmenu();
	    else if (event.xbutton.window == eqmenu.win) handle_eqmenu();
	    else if (event.xbutton.window == dispmu.win) handle_dispmenu();
	    else if (event.xbutton.window == filtmu.win) handle_filtmenu();
#ifdef  GAPVO
	    else if (event.xbutton.window ==  gapmu.win) handle_gapmenu();
#endif
	    else if (event.xbutton.window == xordmu.win) handle_exmenu();
#ifdef  OBJSHAPE
#ifndef  OBJAUTO
	    else if (event.xbutton.window == shapemnu.win) handle_shape();
#endif
#endif

//	    if (event.xbutton.window == dunbut.win) done = True;
// printf (" ..ButtonPress (4) handled.\n");  fflush (stdout);  //debug
	    break;

	case ButtonRelease:	/* 5 */
	    if ((event.xbutton.window == win)) {
		break;	// We don't do it locally
	    }
	    else  printf (" Button Release detected.\n");
	    break;

	case MotionNotify:	/* 6 */
//	    printf (" Motion notification.\n");
/* This is an annoyance; there is no need to worry about motion in this
  application, we don't move anything or drag anything... */
	    break;

	case EnterNotify:	/* 7 */
#if (DEBUG_GUI > 0)
printf (" ..EnterNotify (7) found.\n");  fflush (stdout);  //debug
#endif
//	    edit_win_border (winlist, event.xcrossing.window, app->green);
	    break;

	case LeaveNotify:	/* 8 */
#if (DEBUG_GUI > 0)
printf (" ..LeaveNotify (8) found.\n");  fflush (stdout);  //debug
#endif
//	    edit_win_border (winlist, event.xcrossing.window, app->black);
	    break;

	case Expose:		/* 12 */
//printf (" ..Expose (12) found.\n");  fflush (stdout);  //debug
	    if (event.xexpose.count > 0) break;  /* ignore */
	    if (event.xexpose.window == win) {	/* that's us */
		dflag = 1;
	    }
//printf (" ..Expose (12) handled.\n");  fflush (stdout);  //debug
	    break;

	case CreateNotify:	/* 16 */
/* Ignore?  Generated by propertychangemask or substructurenotifymask */
	    break;

	case DestroyNotify:	/* 17 */
	    if (event.xdestroywindow.window == win) done = True;
/* OK, but that doesn't break the loop when quit selected */
	    break;

	case UnmapNotify:	/* 18 */
	    if (event.xmap.window == win) iconized = True;
	    break;

	case MapNotify:		/* 19 */
	    if (event.xmap.window == win) iconized = False;
	    break;

	case ReparentNotify:	/* 21 */
		/* Ignored */
	    break;

	case ConfigureNotify:	/* 22 */
printf (" ..ConfigNotif (22) found.\n");  fflush (stdout);  //debug
	    if (event.xconfigure.window == win) {	/* main window */
/* Find new config - attributes */
		XGetWindowAttributes (mwin.disp, win, &xwa);
//		set_sizes (&xwa);
//		XMoveWindow (display, dunbut.win, 
//			xwa.width - bw -10, 10);
/* The "done" button needs to move to track right side; the others don't,
  should be protected by min width value */

/* Redraw the window */
//		dflag = 1;
// Temporarily remove to prevent extra refresh at initiallize

	    } else {
//		printf (" -- Configure other window?\n");  /* debug */
		;	/* Null action */
	    }
printf (" ..ConfigNotif (22) handled.\n");  fflush (stdout);  //debug
	    break;

/* Events needed/observed:
... See /usr/include/X11/X.h for these definitions ...
..*/

	case PropertyNotify:	/* 28 */
/* Ignore?  Generated by propertychangemask or substructurenotifymask */
	    break;

	case KillNotify:	/* 52 */
/* The window has been detected to have been closed... (def. in cxt.h) */
	    done = True;
	    break;

	default:
	  if (event.type > (LASTEvent + 32) ) {
/* Added debug section for greatly out of range event numbers, which
  may actually be pointers from some program bug.  */
	    printf (" Event out of range: '%d' (%08X)\n",
		event.type, event.type);
	  } else {
	    printf (" Other event '%2d'\n", event.type);
	  }
/* See /usr/include/X11/X.h for a definition of the event. */
	    fflush (stdout);
	    break;

	} /* End event switch */

/* Do global things which the switch set up (still in loop) */

//printf (" ..End event switch found.\n");  fflush (stdout);  //debug
dpause ("Near end of loop.");

	if (iconized) dflag &= ~2;  // For lint, iconized must be used

/* Check if display needed */
	if (dflag || warnflag) {
// if warnflag call solpos
	    if (warnflag) solpos (mjd);
	    warnflag = 0;	// In case it gets set in window drawing
	    XClearWindow (mwin.disp,win);	// For message text
//	    draw_window (obs, dflag);
	    redraw_menus (mnulist);	// in guisubs.c
#ifdef  IMAGM
/* A kludge to put some window text on a conditional */
  if (IMAGUI == 0)
#endif
	    redraw_text ();		// App. specific text
	    redraw_windows (winlist);
#ifdef  WDEBUG
	    if (!strncmp(myname, "tintgui", 7))
		redraw_grid();  //debug
#endif

#ifdef  GISMOAP
//    if (FeatIndx(ob) == F_gisap) { gisap = 1; ob->reflimit = 0; }
//?	    if (FeatIndx(Gob->obs) != F_gisap)
// try making warnings conditional on not being gismo aperture file
// that drops whole bar; try SetFloatWindow for dec to +90?
#endif

	    redraw_warnings();	// ROTATOR flip, e.g.
	    dflag = 0;
	}

    }  /* main loop */

/* Free up resources and clean up generally */

#ifdef  SKYOPT
    if (skywin.disp) {                   /* close skywin               */
	CBX_CloseMainWindow (&skywin);
	cat_close();
    }
#endif

    CBX_CloseMainWindow (&mwin);

/* Clean up lists */
    while (winlist != NULL) winlist = wlpop (winlist);
    while (butlist != NULL) butlist = blpop (butlist);

/* All done */

#if (DEBUG_GUI > 0)
    printf (" -- pgui exiting...\n");	// debug
#endif

#ifdef  SKYOPT
// dummy statement to avoid warning messages...
if (cat_equ[0] == cat_mag[0]) ;
#endif

}  // End of pgui subroutine


/*  ====  End of X-windows Subroutines  ====  */


/* ---------------------------------------------------------------------- */
#ifdef  SKYOPT
static void    setup_defaults     (void)
{
/* Establish defaults from current default structure. */
/* All values used here are globals */


  /* get camera parmeters */

  {
//    char inifile[256];
//    sprintf(inifile,"%s/camera.ini",genv2("PWD","."));
    (void)get_gstring(GDAT->camera,"name",c_name,"Camera?");
  }

  /* get telescope parameters */

  {
//    char inifile[256];
//    sprintf(inifile,"%s/telescope.ini",genv2("PWD",".")); 
    (void)get_gstring(GDAT->telescope,"name",t_name,"Telescope?");
    tellong = get_gdouble(GDAT->telescope,"longitude",0.);
    tellat  = get_gdouble(GDAT->telescope,"latitude",0.);
  }

  /* get guider parameters */

  {
//    char inifile[256];
//    sprintf(inifile,"%s/guider.ini",genv2("PWD","."));
    (void)get_gstring(GDAT->guider,"name",g_name,"Guider?");
  }
}
#endif

/* ---------------------------------------------------------------------- */
#ifdef  SKYOPT
section**  get_sections (int msel)
{
section	**a;
section	**sx;
section	*p;
section	*d;
int	i, j;
/* What we do:
  Find the actual section pointer from msel using GDAT structure;
  Find the number of unique sections in that queue, it is probably
  a circular list.
  Allocate an array of that number of pointers.
  Populate the new array with the actual pointers to the members
  of the queue.
  Return the pointer to the array.
.. */
    switch (msel) {
	case MNU_TEL:
	    sx = &(GDAT->telescope);
	    break;
	case MNU_GDR:
	    sx = &(GDAT->guider);
	    break;
	case MNU_CAM:
	    sx = &(GDAT->camera);
	    break;
	default:
	    sx = NULL;
    }

/* Count the pointers, and find default */
//    for (i=0, d=p=*sx; p != NULL; p = (p->next == *sx) ? NULL : p->next) {
    i = 0;
    d = *sx;
    C_loop (p, *sx) {
	i++;
	if (p->flag & 1) d = p;		// See datutil.c under "flag"
//	if (p->next == *sx) break;
// BAD CONSTRUCT BREAK
    }
    nsects[msel] = i;

/* If no section pointers return a NULL array */
    if (i < 1) return NULL;

/* Obtain the storage */
    a = (section**) malloc (i * sizeof(section*));

/* Populate the array */
    for (p=*sx,j=0; j<i; j++) {
	a[j] = p;
	p = p->next;
    }

/* Make sure default is the one used... */
    *sx = d;

    return  a;
}
#endif

/* ---------------------------------------------------------------------- */


/* ---------------------------------------------------------------------- */

#ifdef  REFRESH

//#define  BOX3
#undef  BOX3
/*  Symbol  BOX3  uses 3x values rather than 2x when formatting
  the display box.  */

//#define  DebugDisplay
#undef  DebugDisplay

static	void	ErrorHeader (int n)
/*  Write 3 blanks, **, n fields of " * ", and ** at end  --
  or 3 blands, "**", n fields of " *", and " **" at end if 2x form.  */
{
int	i;
    if (n < 1) return;
    fprintf (stderr, "   **");
#ifdef  BOX3
    for (i=0; i<n; i++) fprintf (stderr, " * ");
    fprintf (stderr, "**\n");
#else
    for (i=0; i<n; i++) fprintf (stderr, " *");
    fprintf (stderr, " **\n");
#endif
}

static	void	MessageDisplay (char* msg, int n)
/*  Write msg centered within n x 3 characters, with 3 blanks and
  a ** leading, and ** trailing.  Pad odd blank on right if needed.
  Call with a " " message for a blank line.    */
{
int	m;		// message length
int	x, a, b;	// padding computed
int	i;		// index

    if (msg == NULL) return;
    m = strlen (msg);
    if (m < 1) return;

// compute left padding size and right padding size.  Can be zero.
// have 3n spaces to fill; need (3n - len)/2 spaces before and after

#ifdef  BOX3
    x = 3*n - m;	// eXtra spaces
#else
    x = 2*n + 1 - m;	// eXtra spaces
#endif
    if (x < 0) x = 0;
#ifdef  BOX3
    a = x/2;		// on left
#else
    a = (x+1)/2;	// on left
#endif
    if (a < 0) a = 0;
#ifdef  BOX3
    b = x - a;		// on right
#else
    b = x - 1 - a;
#endif
    if (b < 0) b = 0;

// Do the display of the message
    fprintf (stderr, "   **");   // Start of line
    for (i=0; i<a; i++) fprintf (stderr, " ");
    fprintf (stderr, "%s", msg);
    for (i=0; i<b; i++) fprintf (stderr, " ");
#ifdef  BOX3
    fprintf (stderr, "**\n");
#else
    fprintf (stderr, " **\n");
#endif
}

static	void	ErrorDisplay (char* title, char* mesg)
/*  Write on stderr the title and message, surrounded by some
  extra ** characters.  Add newline before and after.  Center
  the strings within a box.  */
{
int	lt, lm, lx;	// lengths
int	n;		// box size

// Find lengths of messages, and the longest.
    if (title == NULL)	lt = 0;
    else		lt = strlen (title);

    if (mesg == NULL)	lm = 0;
    else		lm = strlen (mesg);
    lx = (lt > lm) ? lt : lm;

// if longest is too short, abandon the effort.
    if (lx < 1) return;

#ifdef  BOX3
// Find multiple of 3 larger than longest
    n = 1 + (lx+2)/3;
#else
// Find next multiple of 2 beyond longest
    n = 1 + lx/2;
//    n = (lx+1)/2;
#endif

// Write blank line and leading decoration
    fprintf (stderr, "\n");
    ErrorHeader (n);

// Write messages (if any)
    MessageDisplay (title, n);
    MessageDisplay (mesg,  n);

// Write trailing decoration and blank line
    ErrorHeader (n);
    fprintf (stderr, "\n");
}
#endif  // on REFREESH

	/* - - - - - - - - - - - *\
	|                         |
	|   Controlling Process   |
	|                         |
	\* - - - - - - - - - - - */

/* Basic process for the GUI */

int	sm_process (opts opt, char* infile, element* gs)
/*  Fill in an obs structure, from infile or from
  basic defaults.  Then start the GUI.  When the GUI
  completes, we simply exit. */
{
static	Obf	ob;
char	*cv;
double	val;
int	i;
#ifdef	REFRESH
int	good = 0;
#endif

// if infile is NULL or empty, we create a new obs structure;
// if not, read the named file into an obs structure.
// Then start the GUI on that file...

    writ = 0;	// no file has been written yet; clear global flag

// Debug/error message is gs is null...
    if (gs == NULL) {
	printf (" ** Optic data is NULL!!\n");
	printf ("  Please be sure \"OPTICDEF\" is properly defined,\n");
	printf ("  or that a link to opticdef.dat is present in the\n");
	printf ("  current default directory.\n");
	return  0;
    }

//    CBX_MessageBox ("Null Optics Alert", 
//	"No Optical Data found!!");

#ifdef  CACHECAT
/*  Special feature.  If we have opt.crep, we will call the skymap cache
  with a special request to report the cache content.  Then, if infile is
  also NULL, we will return without processing.  Once done, we fall back
  into what passes for normal; if a file is present, we operate on it.    */
    if (opt.crep) {   // catalog cache report requested.
	if (opt.dbug >= 0) printf ("  -- Cache content report request\n");
	opt.crep = map_getlist (-1.0, 0.0, 0.0, 0.0);
	opt.crep = -1;  // done
/* When doing -c, we will exit if no actual file name is present.
  However, if one exists, continue on with it.   */
	if (infile == NULL) return 0;
	if (strlen(infile) < 1) return 0;
    }
#endif

// Change this all around.  We have some infile, maybe null or
// maybe blank.  Try to open it anyway.  If it does not open,
// we get a blank structure to create a new file.

    if (infile == NULL) {
	strcpy (ob.filename, "");
#ifdef  IMAGM
      if (IMAGUI)
	printf (" Executing in IMAGE mode.\n");
      else
// slight kludge here to change message if image mode active
#endif
	printf (" Process a new observation file.\n");
    } else {
	strcpy (ob.filename, infile);
	printf (" Process file %s.\n", infile);
    }
    ob.obs = newobs (ob.filename, gs);
    ob.changed = 0;

    if (opt.dbug > 0)
    printf (" File name is %s\n", ob.filename);  // debug only
/* The file name may have changed since the request which was displayed
  a little before as Process file xxxx
  This is not shown in debug.    */

    ob.gs = gs;    /* fake kludge way to globalize element queue! */

/* Get IMACS encoder zero point from environment, if present */
    cv = getenv (Chuecn);
    if (cv != NULL) {
        val = 0.0;
        i = sscanf (cv, "%lf", &val);
        if (fabs(val - IMACSrot) > 10.0 || i < 1) {
            fprintf (stderr, " ** Value of %12.4g for %s is ", val, Chuecn);
            fprintf (stderr, "out of range, using %.3f\n", IMACSrot);
        } else {
            IMACSrot = val;
            if (opt.dbug > 1) printf (" *%s (IMACSrot) set to %.2f\n", 
        		Chuecn, val);
        }
// Only set RotOff if we have IMACS instrument from newobs
	if (InstIndx(ObInstN(ob.obs)) == I_imacs)
		RotOff = IMACSrot;
    }

/* Also, get any specified low and high rotator limits.
  Low limit is nominally -151; High limit is +176.  Allow new limit to
  be set within 45 degrees of 180, -135 for low, +135 for high.   */
    cv = getenv (RotLimLow);
    if (cv != NULL) {
	val = 0.0;
	i = sscanf (cv, "%lf", &val);
	if (val < -180.0 || val > -135.0 || i < 1) {
	    fprintf (stderr, " ** Value of %12.4g for %s is ", val, RotLimLow);
	    fprintf (stderr, "out of range, using %.3f\n", RotLim1);
	} else {
	    RotLim1 = val;
            if (opt.dbug > 1) printf (" *%s (RotLim1) set to %.2f\n", 
        		RotLimLow, val);
	}
    }

    cv = getenv (RotLimHigh);
    if (cv != NULL) {
	val = 0.0;
	i = sscanf (cv, "%lf", &val);
	if (val > 180.0 || val < 135.0 || i < 1) {
	    fprintf (stderr, " ** Value of %12.4g for %s is ", val, RotLimHigh);
	    fprintf (stderr, "out of range, using %.3f\n", RotLim2);
	} else {
	    RotLim2 = val;
            if (opt.dbug > 1) printf (" *%s (RotLim2) set to %.2f\n", 
        		RotLimHigh, val);
	}
    }

    cv = getenv (IMACSlimnam);
    if (cv != NULL) {
	val = 0.0;
	i = sscanf (cv, "%lf", &val);
	if (val > ROTLIM || val < 135.0 || i < 1) {
	    fprintf (stderr, " ** Value of %12.4g for %s is ", val, IMACSlimnam);
	    fprintf (stderr, "out of range, using %.3f\n", IMACSlimit);
	} else {
	    IMACSlimit = val;
            if (opt.dbug > 1) printf (" *%s (IMACSlimit) set to %.2f\n", 
        		IMACSlimnam, val);
	}
    }


/* Some needed initialization */
    rafc = ob.obs->ra / Hour;
    defc = ob.obs->dec / Degree;
    eqfc = ob.obs->equinox;
    wepoch = ob.obs->epoch;
    pdx.lat = -29.01418;
    ob.obs->dlevel = opt.dbug;

/* Set sky window active based on command line flag */
    skywin_active = opt.skwa;

/* Call to start the GUI window, given the obs structure and
  its file name, etc. */
/* Pass &ob as the argument to the interactive window. */

#ifdef  REFRESH
/* The -r refresh feature -- if requested, write the Obs structure data
  to the same file as it was read from; but only if there are no obvious
  errors or omissions in the data as found.  If this is not done, or
  if an error was found, the window is started.  If the refresh is done
  properly, we will skip starting the window.  This is decided based on
  whether the "writ" flag has been successfully incremented.  */
    if (opt.refr && infile != NULL) {
char	fname[10];
char	inspec[128];
char	buffer[128];
char	message[256];
char	*c;
int	gisap = 0;

/* Refresh flag for a real file; write output and exit here */
// Like in handle_active and handle_iqmenu
// We do some basic checks for inst, disp, objlist, declination

// checks for consistency similar to those in handle_active; if not
// passed, we don't allow auto refresh here...
	good = 7;	// innocent until proven
	strncpy (fname, ob.obs->fname, 8);
	fname[8] = AN;		// our current filename
	if (strlen(fname) < 1) good = 0;
	strcpy (inspec, ob.filename);
	c = filenamloc (inspec);
	*c = AN;		// Now, spec without filename
	sprintf (buffer, "%s%s.obs", inspec, fname);	// candidate filename
	if (opt.dbug > 1) {
fprintf (stderr, " !Mangled fname of %.8s to %s.\n", ob.obs->fname, fname);
fprintf (stderr, " !Mangled filename of %s to %s.\n", ob.filename, buffer);
	}

/* Checking for consistency similarly to handle_active; we may add other
  tests if needed.  Only do refresh if resonably good obs file here.  */

	if (!strcmp(fname, "Unnamed") || good == 0) {
//	    sprintf (message, "Please enter a file name.");
//	    CBX_MessageBox ("No file name found.", message);
	    ErrorDisplay ("No file name present.",
			  "Please provide a file name.");
	    good = 0;
	}


#ifdef  IFUM
	if (ob.obs->Instr_mt == NULL && ob.obs->IFUmode == 0) {
#else
	if (ob.obs->Instr_mt == NULL) {
#endif
	    ErrorDisplay ("No instrument found.",
			  "Please provide an instrument.");
	    good = 0;

#ifdef  IFUM
	} else if (ob.obs->Disperser == NULL && ob.obs->IFUmode == 0 && gisap == 0) {
#else
	} else if (ob.obs->Disperser == NULL && gisap == 0) {
#endif
	    ErrorDisplay ("No disperser found.",
			  "Please provide a disperser.");
	    good = 0;
	}

#ifdef  IFUM
	if ((ob.obs->ex_order & 2) && !ob.obs->IFUmode) {
#else
	if (ob.obs->ex_order & 2) {
#endif
int	ord = ob.obs->order;
	    if ( (ord * ob.obs->pb.red) > ( (ord+1) * ob.obs->pb.blue) ) {
		sprintf (message, 
		    "Warning..  Order %d red overlaps order %d blue.",
		    ord, ord+1 );
//		if (!CBX_AlertBox("WARNING", message)) return False;
		ErrorDisplay (message, "Please reduce spectral range.");
		good = 0;
	    }
	}

// Consistency tests
// in sm_process
/* Testing hrf for rotator flip hour angle is not done here, as the hrf is
  computed only in the window loop.  */

	if (ob.obs->objfiles == NULL) {
	    ErrorDisplay ("No object files present.",
			  "Please enter an object file.");
	    good = 0;
	}

	if (ob.obs->dec > 61.0*Degree) {
	    ErrorDisplay ("Invalid position.",
			  "Please provide a valid position.");
	    good = 0;
	}

/* Copy the test for guide stars from writing program.  May need to
  copy others to avoid calling writeobs and having tests fail which
  would cause an AlertBox call to crash the system...  */

#ifdef  GSREQUIRED
/* Count the guide stars selected, and if IMACS require 2 and warn
  if fewer have been selected... */

#ifdef  GISMOAP
	if (FeatIndx(ob.obs) == F_gisap) { gisap = 1; ob.obs->reflimit = 0; }
#endif

//	if (InstIndx(ObInstN(ob.obs)) == I_imacs && gisap == 0) {
	if (InstIndx(ObInstN(ob.obs)) == I_imacs && ob.obs->reflimit > 0) {
int	k;
// Count the number of selected guide stars, should be 2
	    k  = (ob.obs->gp1.ep == 0.0) ? 0 : 1;
	    k += (ob.obs->gp2.ep == 0.0) ? 0 : 1;
	    if (k < 2) {
//		sprintf (message,
//		" WARNING!  Fewer than 2 guide stars selected!");
//		if (!CBX_AlertBox("WARNING", message)) return False;
		ErrorDisplay ("Fewer than 2 guide stars selected!",
			"Please select guide stars in skywin.");
		good = 0;
	    }
	}
#endif  // on GSREQUIRED

/* At this point we should have tried all the tests normally done on
  output structures, and "good" should reflect the actual state of
  the Obs structure data.  This will give a very strong chance of
  actually being able to write the file here.   */

      if (good) {  // No input errors found in the above tests
	printf (" == Writing to \"%s\"\n", infile);	// debug
	if (!writeobs (ob.filename, ob.obs)) {
	    sprintf (message, " !! File not written to \"%s\"",
		ob.filename);
	    ErrorDisplay (NULL,
			  message);
	    good = 0;
	} else writ++;   // Actually written by refresh option

// ADD ELSE CLAUSE AND NON-WRITING MESSAGE HERE
      } else {  // the not good case; input errors precluded writing
	sprintf (message, "-- File not written to \"%s\" --", ob.filename);
	ErrorDisplay ("Due to input errors,", message);

// Here is where we debug those display box line lengths
#ifdef  DebugDisplay
	ErrorDisplay ("Even", "Odd");	// debug message  (4 and 3)
	ErrorDisplay ("Even", "xOddx");  // debug message (4 and 5)
#ifdef  BOX3
	ErrorDisplay ("xEvenx", "xOddx");	// 5 and 6
#endif

#endif  // on DebugDisplay
      }    // End of conditional on detected errors

    }	// End of the refresh feature conditional
    if (writ == 0)  /* Do the window if *not* written by refresh. */
#endif  // on REFRESH
    pgui (&ob, opt.size);	// Interactive process window done here.

/* When the window has exited, control comes back here.  The "writ" flag
  may have been set if the window operation did an actual write of the
  Obs structure data (ob.obs).  It could also have been set by writing
  in the refresh option section above, which would have caused the window
  operation to be skipped.  */

/* Return 0 if file processing fails; 1 is count of success... */
//    return  1;
    i = (writ == 0) ? 0 : 1;
    return  i;
}




	/* - - - - - - - - *\
	|                   |
	|    Main Program   |
	|                   |
	\* - - - - - - - - */

#define  GETOPTUSE
//#undef  GETOPTUSE
/*  Symbol  GETOPTUSE  is used to turn on (or off) the use of the
  standard "getopt" utility in parsing the command line.  Use this
  local define to enable the "old" method quickly if needed.  */

/*  ====  MAIN PROGRAM  ====  */

#define  PROMPTLOOP
#define  DATFILE  "opticdef.dat"
#define  DATtFILE  "opticdef.tst"

#define  F_size  18

int  main (int argc, char** argv) {
#ifdef  GETOPTUSE
extern char *optarg;
extern int optind,opterr,optopt;
int c;
#else
char    *pp;    /* Scan variable */
char    px;     /* Scan character */
int     i;
#endif
int     dohelp = 0;
char	promst [80];	/* Dynamic prompt... */
char	infile [120];	/* For optional argument */
opts	opt;
namlist	*head = NULL;
char	*filename;
int	k, m, n;	/* Process counts */
int	wlevel;		// Debug level
element	*gs = NULL;	/* Optic details queue pointer */
char	xnam[64];	// Part of message
char	mbuf[128];	// Message to be generated...
//#ifdef  GISMO
#ifdef  GismoMenus
int	tfil = 0;	// Normally disabled
#endif
char	*vstr;		// pointer for dynamic version string
namlist	*comps=NULL;		// component list for check_ver

// ADD THE PROMPTING LOGIC FOR INPUT FILE!!

/* Get parameter as name of SMDF file to read */
/* Usual flags, debug and help */

/* Standard input logic:  If argument given, take that argument as the
  sole specification to process.  If no argument is given, prompt for
  an argument, and repeat the prompting until a null reply.  */

    opt.dbug = 0;	/* Or - put data in define above */
    opt.size = F_size;
#ifdef  REFRESH
    opt.refr = 0;
#endif
//#ifdef  CACHECAT
    opt.crep = 0;
//#endif
    dodebug = 0;	/* NOTE = temporary for debugging */
    wlevel = 0;
    *infile = '\0';
//    h = 18;		// replace with defined default...

#ifdef  EXCAT
//    USEXMLRPC = 0;   // was False;
#endif

#ifdef  IMAGM
    IMAGUI = 0;
#endif

#ifdef  CACHECAT
    Cache.status = 0;	// Redundant
#endif

    opt.skwa = 0;	// Default is no sky window at start
//    skywin_active = 0;		// Default is no sky window at start

/* Store our command name in a global... */
    strcpy (myname, cfilename(argv[0]));
    sprintf (xnam, "Executing \"%s\".   ", myname);	// used for debug
/* NOTE that we do not do \n in the above debug line... */
    sprintf (mbuf, " %s\n", xnam);  /* debug */

/* Also create a test mode global. */
    TESTMODE = (strncmp (myname, "tintgui", 7) == 0) ? 1 : 0;

#ifdef  GETOPTUSE
/* Decode all the options of the command line */
/* :: follows letter with optional argument, : follows one with required
  argument, otherwise there are no arguments to the options.  */
    while((c = getopt(argc, argv, "cd::f:ikmorstwh")) != -1) switch(c) {
//#ifdef  CACHECAT
	case 'c':	opt.crep = 1;		break;  // Cache content
//#endif
	case 'd':				// debug level
	    wlevel = optarg == NULL ? 1 : atoi (optarg);	break;
	case 'f': 
	    opt.size = optarg == NULL ? F_size : atoi (optarg);	break;
#ifdef  EXCAT
	case 'i':  if (TESTMODE) USEXMLRPC = 3;  break;  // test remote catalog
#endif
	case 'k':	opt.skwa = 1;		break;	// skywin start
	case 'm':	TESTMODE <<= 1;		break;  // modify test (obsolete)
	case 'o':	TESTMODE = 0;		break;	// test off
#ifdef  REFRESH
	case 'r':	opt.refr = 1;		break;	// refresh mode
#endif
//#ifdef  GISMO
#ifdef  GismoMenus
	case 's':	tfil = 1;		break;	// test file
#endif
#ifdef  IMAGM
	case 't':	opt.skwa = IMAGUI = 1;	break; // image mode
#endif
	case 'h':	dohelp = 1;		break;	// parameter help
	default:
/* When c is '?' an invalid option message has already been written. */
	    if (c != '?') printf (" unused -%c option\n", c);	break;
    }  /* end of switch, only statement in the getopt while loop */

    for (c = optind; c < argc; c++) {
	sprintf (mbuf, " %sArgument %s.\n", xnam, argv[c]);  /* debug (old) */
	head = pushtxt (argv[c], head);  /* Collect arguments */
    }

#else  // on GETOPTUSE
/* Scan for switches and arguments */
    for (i=1; i<argc; i++) {    /* argv[0] is command name */
        pp = argv[i];
        if (*pp == '-') {       /* Scan normal switch */
            while ((px = *++pp) != '\0') {
                if (px >= 'A' && px <= 'Z') px += 'a' - 'A';
                if      (px == 'd')  wlevel = 1;  /* Debug output */
/* May want -d to cause debugging later? */
		else if (px == 'f') {
		    k = sscanf (++pp, "%d", &opt.size);
		    if (k < 1) opt.size = F_size;
		    break;
		}
		else if (px == 'h')  dohelp = 1;  /* show help */
//#ifdef  GISMO
#ifdef  GismoMenus
		else if (px == 's')  tfil = 1;  /* Use test file */
#endif
#ifdef  EXCAT
		else if (px == 'i' && TESTMODE)  USEXMLRPC = 3;  /* Test only */
#endif
#ifdef  IMAGM
		else if (px == 't')  opt.skwa = IMAGUI = 1;   /* Image mode flag */
#endif
		else if (px == 'k')  opt.skwa = 1;  /* start skywin */
		else if (px == 'm')  TESTMODE <<= 1;	/* Modify test */
		else if (px == 'o')  TESTMODE = 0;  /* kill test mode */
            }
        } else if (*pp == '+') {        /* Scan positive switch */
            while ((px = *++pp) != '\0') {
                if (px >= 'A' && px <= 'Z') px += 'a' - 'A';
                if      (px == 'd')  wlevel = 2;      /* Debug output */
                else if (px == 'h')  dohelp = 1;  /* show help */
            }
        } else if (*pp == '?') {
                dohelp = 1;
        } else {
	    sprintf (mbuf, "%sArgument %s.\n", xnam, pp);  /* debug */
//	    if (*infile == '\0') strcpy (infile, pp);
//	    else printf (" -- Multiple argument not used.\n");
#ifdef debug
	    printf (mbuf, "Adding %s to list.\n", pp);  /* debug */
#endif
	    head = pushtxt (pp, head);	/* Collect arguments */
        }
    }
#endif	// on GETOPTUSE

//    vstr = fvstr (INTGUIV);
    if (wlevel > 0)
      printf ("Version %s, initial %s\n", VERSION, XVERSION); //debug
//    strcpy (VERSION, vstr);
//    free (vstr);
    vstr = combinev (INTGUIV, SKYWINV);
    strcpy (XVERSION, vstr);	// Put combined version in global
    free (vstr);

    if (wlevel > 0) printf ("%s", mbuf);

//    if (wlevel > 0 || dohelp || TESTMODE)
//	printf ("\n %s -- Version %s%s\n", PROGNAME, VERSION, SKWVERS);
    if (wlevel > 0 || dohelp || TESTMODE) {
	printf ("\n %s -- Version %s", PROGNAME, XVERSION);
	if (wlevel > 0) printf ("  Debug: %d", wlevel);
	printf ("\n");
//			INTGUIP, INTGUIV, SKYWINV);
    }
    dodebug = wlevel;	// set it early

/* See if this version is any good, or if we continue.  */
    comps = pushtxt (INTGUIT, comps);
    comps = pushtxt (SKYWINT, comps);
    comps = pushtxt (MASKGNT, comps);
    comps = pushtxt (MGUTILT, comps);
    k = check_ver (comps, NULL, NULL);
// Use NULL queues for now, we don't test for obsolete on the
// file being processed here, so we can update it.
    comps = kill_name (comps);
// if feature exists, we could exit on k < -2 here.
    if (k < -1) {
	printf (" ** Operation of Obsolete software rejected.\n");
//	kr = 1;  // or maybe kr = k; ?
//	goto main_finish;
	return  1;
    }

    if (dohelp) {
//	printf ("\n %s -- Version %s%s", PROGNAME, VERSION, SKWVERS);
	printf ("\n  Parameters are: ");
	printf ("\n     Observation file (optional) ");
#ifdef  PROMPTLOOP
	printf ("\n    Inputs are prompted for if not given.");
#else
//	printf ("\n    Use \"%s\" if not given.", "/dev/null");
	printf ("\n    Make new file if none given.");
#endif
	printf ("\n   Switches");
#ifdef  CACHECAT
	if (TESTMODE) 
	printf ("\n     -c  = Report cache contents");
#endif
#ifdef  GETOPTUSE
	printf ("\n     -d* = Set debug level to (*) [1]");
#else
	printf ("\n     -d  = Some debug information");
	printf ("\n     +d  = More debug information");
#endif
#ifdef  GETOPTUSE
	printf ("\n     -f* = Set font height to (*) [%d]", F_size);
#else
	printf ("\n     -fn = Set font height [%d]", F_size);
#endif
	printf ("\n     -h  = Display this help");
//#ifdef  GISMO
#ifdef  GismoMenus
// special tfil option
	printf ("\n     -s  = Use opticdef.tst instead of .dat");
#endif
#ifdef  EXCAT
	if (TESTMODE)
	printf ("\n     -i  = Force web catalog lookup");
#endif
	if (TESTMODE)
	printf ("\n     -o  = Turn off Test Mode features");
#ifdef	REFRESH
	printf ("\n     -r  = Refresh; save output and exit immediately");
#endif
	printf ("\n     -k  = Start skywin immediately");
#ifdef  IMAGM
	printf ("\n     -t  = Image test mode");
#endif
	if (TESTMODE)
//	printf ("\n     -m  = Modified test features (standard catalog)");
	printf ("\n     -m  = Modified test features (unused)");
//	printf (" (standard catalog)");
// The above is actually unused, and should not be displayed
//	printf ("\n     ");
//	printf ("\n   Dump(s) written to standard output.");
	printf ("\n \n");
	return (1);
    }

/* Find and open the optic data file, and all that stuff... */
//#ifdef  GISMO
#ifdef  GismoMenus

    if (tfil && access (DATtFILE, R_OK) == 0)
       gs = get_optic_data (DATtFILE, wlevel);
    else
// That's a little cheat; the next line is the else clause, but only
// in the GISMO case; otherwise it's the only thing.
// Test optic data stuff copied from maskgen.c here.
#endif
    gs = get_optic_data (DATFILE, wlevel);

    opt.dbug = wlevel;
    dodebug = wlevel;

//    Gob->obs = NULL;
//    Gob->gs = gs;
//    Gob->changed = 0;
//    Gob->filename[0] = `\0`;

/*  -- Loop to process arguments, or prompt if no arguments -- */

/* Insert a default file name here, so as to not need to prompt the
  user if no file name is given on command line. */
//    if (head == NULL) head = pushtxt ("/dev/null", head);


    for (k=n=m=0; k >= 0; ) {	// Forever loop
#ifdef  IMAGM
	if (IMAGUI) {
/* In the case of an active image mode, the file argument should
  be ignored, and a new obs structure used, and the loop proceede 
  only once.  Set this up here.  */
	    filename = NULL;
	    if (m > 0) break;
	    m++;	// So we will stop NEXT time.
	} else if (head == NULL) {
#else
	if (head == NULL) {	// No arguments, or looped out
#endif  // on IMAGM
#ifdef  PROMPTLOOP
	    if (m > 0) break;	// Had real arguments, looped out
#else
	    if (n > 0) break;	// No loop after any counted argument
#endif

#ifdef  CACHECAT
	  if (opt.crep == 0) {
#endif
	    if (n == 0)
		sprintf (promst, "Observation file [return for new]:");
	    else
		sprintf (promst, "Observation file [return if none]:");
// THAT WORKS FOR GIVEN FILE(S) AND FOR ONE NEW FILE, OR FOR
// SEVERAL PROMPTED FILES.  AFTER PROMPTED FILE, CANNOT ENTER
// A NEW FILE.  THIS MAY BE ACCEPTABLE.  A BARE RETURN CAN EITHER
// TERMINATE THE LIST, OR ASK FOR A NEW FILE, BUT NOT BOTH!
	    askmestr (infile, 110, promst);
#ifdef  CACHECAT
	  }
#endif

	    if (*infile == AN) {
		filename = NULL;
		if (n > 0) break;	// already gave arg, exit
		m++; 
	    } // break; 
// Exit on any null reply -- changed to null file name
	    else {
		filename = dynamstr (infile);
	    }	// (test for null reply to prompt)
	} else {		// We have a command line argument given
	    filename = pop_name (&head);
	    if ( filename == NULL) continue;
	    if (*filename == AN) continue;
	    m++;	// Count number of command line arguments given
	}	// (test for arguments given)
	n++;		// Count number of arguments requested
// The main subroutine call goes here
//	printf (" Processing observation file:  %s\n", filename);
//	k += Command_Processor (filename, opt);
/* Process the observation (mask generation) data file */
	k += sm_process (opt, filename, gs);
	free (filename);	// MUST be dynamic
#ifdef  CACHECAT
//	if (opt.crep < 0) break;  // Do not process the file.  // not needed
#endif
    }

/* All done, write a message about how it came out... */
#ifdef  IMAGM
    if (IMAGUI == 0)  // Skip this when image mode is active.
#endif
    printf ("\n Processed %d of %d files.\n", k, n);

    return  0;
}


/* Revision History

    2011/11-23 -- Version 4.68 (.30)
	-- Addition of better support for DIFREF feature, a button to turn
	on/off the feature and enable/disable the HA window, and compatability
	with older data without the feature.

    2011/10-26 -- Version 4.67 (.30)
	-- On SHFTCLK, add the feature that shift-click gets
	translated to center click, since some or many apples do not
	have a readily accessible center button, but often will have
	a right click capability.  This removes the shift modifier from
	other uses in the future.  Also added in sky window.
	-- Window position changed from CBS_BOTTOMRIGHT to CBX_MIDRIGHT
	-> Is NOT a minimum version, although it should be considered
	as such on apple machines; however it is not needed for most.

    2011/02-17 -- Version 4.66 (.29)
	-- Add magnitude limiting to skymap.c, with same MLIM as in the
	server catxml.c so disk and net catalog results are the same.  The
	MLIM is currently 1 magnitude fainter than the faintest usable
	guide stars defined in pmap.dat.
	-- (03-17) Change fclose to iclose for file name support.

    2010/06-25 -- Version 4.65 (.28)
	-- Change test on gismoap to set reflimit to 0 with gismoap, and then
	when reflimit is zero, bypass the test for guide stars.  Allows the
	selection of reflimit to zero for position independent mask.
	-- (8-13) Add use of locfile, and introduce FORCEOBS to indicate that
	only .obs files are read and written (as it has ever been).
	-- (8-25) Change to test for new instrument first, then for old when
	setting enable/disable windows for gismoap case.  Also, set disperser
	selection to 0 rather than 1 if gismoap instrument is selected.
	-- (8-27) Add symbol GismoMenus for special selection of GISMO menu
	features being enabled.  Otherwise support existing GISMO features.
	The makefile supports changing GismoMenus to compile gintgui.
	-- (9-3) Add TextUpdate subroutine to all winblock definitions to
	do updates whenever window data changes.
	-- (9-21) Add extended message for event out of range.
	-- (9-30) Add tests on ob->reflimit > 0 to skip some things when
	Position Independent masks are created.
	-- (10-5) Change help listing, match to test mode.
	-- (11-15) Add range of rotator limits, not just flip at 180
	--> Will be minimum version at release due to rotator limits.
	-- (12-6) Introduce binary search on symbol  BINSEARCH
	-- (12-9) Report on debug the rotator/IMACS limits, and do not search
	IMACS angle limit if already rotator limited;  Report modifications
	due to Environment Variables on debug level 2.  ioutuls.c sets EV
	based on special tokens in mgvers.dat file now.

    2009/11-10 -- Version 4.64 (.28)
	-- Add CONSTEL feature to give constellation for position.  The work
	is actually done in guisubs.c:534 and kdcutil.c
	-- Add -o flag to cancel TESTMODE

    2009/08-13 -- Version 4.63 (.25)
	-- The skymap contains new "projd" function to find projected distance
	between two a,d pairs, used to implement a single field radius value
	instead of x,y field limits.  Changing call sequences to do this.
	Mostly affects map_getlist catalog interface.
	-- Also modifications for cache report (-c).
	--> Will be minimum version with compatable server release.

    2009/07-23 -- Version 4.62 (.25)
	-- Add code for "refresh" option to immediately write an updated
	.obs file without interactive window, if data is good.
	-- Symbol  EXCAT  now enables network catalog lookup via test
	server, with -i or automatically.
	--> Will be minimum version when catalog server is released.

    2009/07-10 -- Version 4.61 (.25)
	-- Fixed writing of reference hole size to make alen & blen 1/2 of
	width for shapes circle and square.

    2009/05-21 -- Version 4.60 (.25)
	-- Removing conditionals for DISPXPTR which is now always defined.
	-- Adding GISMOF2 as a conditional feature.
	-- Change menu order and labels: IMACS_f/2 IMACS_f/4 GISMO_f/2 GISMO_f/4
	to look similar and to be in same order.
	--> Minimum version for GISMO

    2009/01-27 -- Version 4.59 (.25)
	-- Includes code for GISMOAP support.
	--> Minimum version when GISMO is released

    2009/01-15 -- Version 4.58 (.25)
	-- The Exit and Quit-nosav entries in file menu were reversed
	due to the enum being out of order, now fixed.
	-- Fixed GISMO test of instrument name when it could be null.
	-- Process count is 1 only if a file was written.
	--> New minimum version due to bug fix (start with 2.08)

    2008/12-15 -- Version 4.57 (.25)
	-- Add code under OBJSHAPE defined locally to implement the shape
	pop-up menu for global object default slit shape definition.
	-- Add code for GISMO Macro N/S mode as new instrument, and
	ready to provide f/4 and f/2 GISMO instrument entries.  These
	will require the DISPXPTR method.
	-- New menu entry to quit with no save added.

    2008/10-14 -- Version 4.56 (.25)
	-- Add code to implement "getopt" parsing loop.

    2008/09-16 -- Version 4.55 (.25)
	-- Adding code for IMAGM feature support behind that symbol.
	--> Will be minimum version at release 2.07

    2008/09-16 -- Version 4.54 (.25)
	-- Add support for EXCAT feature to obtain standard catalog
	from internet server.
	-- Drop the -t feature from tintgui, use standard catalog
	-- Fixed loop logic and sleep time for skywindow
	--> Will be minimum version for 2.06 due to bug fix

    2008/09-04 -- Version 4.53 (.24)
	-- Add support of "expri" keyword for obs file
	-- Add conditional code for extended priority window  EXTPRI
	-- Introduce new disperser menu scheme on symbol  DISPXPTR

    2008/08-28 -- Version 4.52 (.23)
	-- Add -k to start skywin immediately
	--> Minimum versiion for relase 2.05

    2008/08-21 -- Version 4.51 (.23)
	-- Fix bug where N&S orientation would write GAPS data
	--> Minimum Version for release 2.04 due to bug fix

    2008/04-28 -- Version 4.50 (.22)
	-- Add file checksum and signature (secutil.c)
	-- Add error message if pmap.dat not found.

    2008/03-06 -- Version 4.49 (.22)
	-- Change output .obs file format to include written by line
	-- Add Program (Vers) structure
	-- Reject operation of obsolete software

    2008/03-06 -- Version 4.48 (.22)
	-- N&S button moved above the disperser menu
	-- Slight movement of Order: and Observer: windows
	-- Detect bad font size before crash
	-- Changed debug messages to find the bad font size fix
	-- Set debug to 0, -d gives 1, +d gives 2

    2008/01-29 -- Version 4.47 (.22)
	-- Sky Window now does precession on input centering click;
	-- Normalize Obs before writing has been fixed in mgutils.
	-- Minimum Version for release 2.03 and following

    2007/11-26 -- Version 4.46 (.21)
	-- Add support for IMACS 150 lpmm grating in menus.
	-- Change gap menu to on/off when using nsrot feature.

    2007/10-25 -- Version 4.45 (.18)
	-- Fix GISMO mode special flag problem.

    2007/10-17 -- Version 4.44 (.18)
	-- Adding check_ver checking
	-- Minimum Version for release 2.00 and version control.

    2007/10-11 -- Version 4.43 (.18)
	-- Changed version support to report decimal versions.
	-- Version number itself is now in mgfeats.h

    2007/09-21 -- Version 4.42 (.18)
	-- Extended TESTMODE for -t option to use standard catalog.
	-- Also, automatic guide star supported in skywin.
	-- Skywin initialization changed for auto guide star.
	-- Guide star clearing on change now depends on actually testing
	the given guide star for validity.
	-- Display version number when in test mode.

    2007/07-06 -- Version 4.41 (.04)
	-- Using ObInstN to find name to avoid null instrument crash

    2007/06-07 -- Version 4.40 (.04)
	-- Add symbol GISMOE support to allow GISMO to have MOE
	disperser, as an experiment.

    2007/04-03 -- Version 4.39 (.04)
	-- Minor change to allow use of complex (instrument) disperser
	composed of multiple grisms.  Such an instrument is not added
	yet, it will be the LDP_prism when made a menu element.
	-- Also fixed bug causing crash if input file had a date present.

    2007/02-15 -- Version 4.38 (.04)
	-- Change mnublock to have text color rather than flag, and to
	set color green/yellow/red for over 6, over 3 or less weeks.
	-- Add output of non-rectangle default slit shape if it exists.

    2007/01-24 -- Version 4.37 (.04)
	-- Adding NSROT support

    2006/10-12 -- Version 4.36 (.04)
    	-- Adding GISMO support
    
    2006/06-29 -- Version 4.35 (.04)
	-- The AXBUT feature is made standard, commented out.
	-- The EXORDX feature is made standard, commented out.
	-- The GCENT feature is made standard, commented out.

    2006/04-27 -- Version 4.34 (.04)
	-- The LeaveNotify message (and EnterNotify) were put behind
	compiler conditionals to reduce screen clutter.

    2006/04-05 -- Version 4.33 (.04)
	-- The MOE disperser blanks out center wavelength, extra orders,
	order, slit extension, slit lengths.
	-- Support for new LDSS grism in menu.

    2006/02-22 -- Version 4.32 (.04)
	-- Support the OBJCHECK feature

    2006/02-10 -- Version 4.31 (.04)
	-- Better vertical spacing, and graphic bar varies in width
	and position when sun and moon bars are added.
	-- Cue the WDEBUG things to the tintgui name feature.

    2006/02-08 -- Version 4.30 (.04)
	-- Includes moon line, and debug version exit button.

    2006/01-25 -- Version 4.29 (.04)
	-- More fixes for MacOS-X menu updating.

    2006/01-24 -- Version 4.28 (.04)
	-- Add 2.0 airmass color feature (in guisubs.c).

    2006/01-19 -- Version 4.27 (.04)
	-- Supporting guide star selection; omitting the writing of
	integer part of guide star data as it is not used.

    2005/12-17 -- Version 4.26 (.04)
	-- Adding support for date and sun bar; added plot of sun bar
	in guisubs, and computation of sun coordinates.

    2005/12-12 -- Version 4.25 (.04)
	-- Add Ref. Limit window under control of REFLIM symbol.

    2005/12-06 -- Version 4.24 (.04)
	-- Use local wlevel to set dodebug, which will become obsolete.
	Also, default level is now 0.

    2005/11-17 -- Version 4.22 (.04)
	-- Add some real filter names

    2005/10-11 -- Version 4.21 (.04)
	-- Add extra update of menu for MacOSX which does not send redraw
	event when pull-down (or popup) menu closes...

    2005/09-30 -- Version 4.20 (.04)
	-- Add support for new 150 line IMACS grism.

    2005/08-24 -- Version 4.15 (.04)
	-- If pb2 values differ from pb, set pb2act to find the
	changes in the secondary data windows.
	-- Also, do not add '>' in the global comment writing...

    2005/07-15 -- Version 4.14 (.04)
	-- If named file is not found, use the name as a default for
	the new blank obs structure created.

    2005/07-06 -- Version 4.13 (.04)
	-- Introducing gap menu, behind the GAPVO symbol
	-- Put circular field of view into LDSS sky view

    2005/06-29 -- Version 4.12 (.04)
	-- Supports existence of gap and orientation bits in detector
	status word of the Obs structure when writing .obs file.
	Needed for full support of gap feature, to come later.

    2005/06-14 -- Version 4.11 (.04)
	-- Support removal of instrument and grating elements from
	the Obs structure with equivalent data use.

    2005/05-27 -- Version 4.10 (.04)
	-- Add active button to activate mask generation, hidden with
	the AXBUT feature symbol.
	-- Modify to recognize tintgui and call tmaskgen then...
	-- Put "Magellan2" as preferred to "Mag2NoADC"

    2005/05-27 -- Version 4.09 (.04)
	-- Generalized the menu creation
	-- New redraw_text subroutine for app. specific text
	-- File and equinox menus now part of menu list
	-- Menu labels displayed by general redraw
	-- Some debug from menu handlers suppressed
	-- Menu handler messages for file name suppressed

    2005/05-26 -- Version 4.08 (.04)
	-- Support of EXORDX feature for pull-down menu on extra
	order checking support.

    2005/05-11 -- Version 4.07 (.04)
	-- Fixed reporting of grism center wavelength

    2005/04-19 -- Version 4.06 (.04)
	-- Write telescope name as it was read; 
	-- Match chuei rotation angle to instrument field

    2005/03-23 -- Version 4.05 (.04)
	-- Implementing global feature method with mgfeats.h

    2005/03-02 -- Version 4.04 (.04)
	-- Add support for Differential Refraction (Hour Angle window)

    2005/03-01 -- Version 4.03 (.04)
	-- Add rotator offset angle set based on imacs/ldss instrument.

    2005/02-25 -- Version 4.02 (.04)
	-- Change ldss2 == NoADC and use Mag2NoADC instrument; to get
	a better focal length on the masks.

    2005/02-18 -- Version 4.01 (.04)
	-- Add new grisms VPH-red, VPH-blue descriptions

    2005/02-03 -- Version 4.00 (.04)
	-- Add support for new LDSS instrument
	-- Remove support for old LDSS instrument

    2004/11-10 -- Version 3.98 (.04)
	-- Add wavelength windows for detector passband.

    2004/11-10 -- Version 3.98 (.03)
	-- Adding symbol DMOE for support of MOE disperser.

    2004/10-22 -- Version 3.98 (.02)
	-- Writing rotator flip commentary into .obs file generated.
	-- Comments currently #! but !# will be acceptable too if needed.

    2004/10-22 -- Version 3.98
	-- Appending sky window version .01 to end and keeping sky window
	version separately for updates.

    2004/10-21 -- Version 3.989
	-- Add GCENT - centering feature code (With GCENT symbol)

    2004/10-19 -- Version 3.988
	-- Same version; removing lots of commented code no longer needed

    2004/10-15 -- Version 3.988
	-- Testing addition of button to run slit extend feature (SLXB)
	-- Added SavExit menu item to save and exit in one step.

    2004/09-23 -- Version 3.987
	-- Writing 3 instead of 1 for ex_orders to be compatable with
	planned extension of the flag's meaning.

    2004/09-20 -- Version 3.986
	-- Fixed bug writing minsep (overlap) with wrong default.
	-- Also, added default coordination section to mgutils.h

    2004/09-10 -- Version 3.985
	-- Allow slit extension flag in observation file

    2004/09-08 -- Version 3.984
	-- Add 1200 lpmm grating to menu

    2004/06-03 -- Version 3.983
	-- The No ADC button is not present in the GUI

    2004/05-25 -- Version 3.982
	-- Revised position of IFU aperatures.
	-- NoADC now defaults to 0 (ADC support).

    2004/05-12 -- Version 3.98
	-- Bug fix to catch an infinite loop possibility in search for
	rotator flip and other problems.

    2004/03-09 -- Version 3.97
	-- Enforcing no embeded blanks in file and observer name fields.

    2004/02-12 -- Version 3.95
	-- New (measured) offsets for IFU and new names for menu put
	in place.  Probably correct for distribution.

    2004/02-05 -- Version 3.90
	-- Improve IFU support, plot in skywin, initial values + defaults
	-- Add support for IMACSCHUECO environment variable

    2004/01-30 -- Version 3.80
	-- Add support for IFU mode, hidden behind define "IFUM"

    2004/01-15 -- Version 3.75
	Changes the way instrument and disperser menus are handled, and
	the data tables which drive them.
	-- Add version stamp to the .obs file output.

    2004/01-12 -- Version 3.70
	Introducing LDSS support hidden behind define variable "LDSS"

    2004/01-09 -- Version 3.65
	Added detection of changes to object file list.
	Default catalog is USNO with no star colors; also default
	of the North arrow.

    2004/01-08 -- Version 3.60
	Support the addition of DLIMIT to the .obs file (writing a
	proper output file, not allowing editing yet)

    2003/12-19 -- Version 3.50
	Fix catalog search to look in local ./gcats in addition to a
	few others.  Improves chances of finding catalogs.
	Also, we go live with the musthave and pdecide things.

    2003/12-18 -- Version 3.45
	Adding musthave and pdecide as debug things prior to release.

    2003/12-16 -- Version 3.40
	Adds a graphic warning for rotator flip and secondary guider
	occultation (if in NoADC mode).

    2003/12-04 -- Version 3.30
	Supports writing of musthave and pdecide parameters in the
	output .obs file.  Does not yet have windows for these.

    2003/12-02 -- Version 3.25
	Can kill sky window directly

    2003/11-28 -- Version 3.20
	Fixed sky window update on instrument change

    2003/11-26 -- Version 3.15
	Correctly plots objects in the sky window.
	Needs to track instrument in plotting camera field.
	Fixed bug which would not give rotator warning when the
	window initially came up with new data.

    2003/11-25 -- Version 3.10
	Introduce the sky window as part of new option menu; at first
	hidden behind a #define constant.

    2003/11-24 -- Version 3.00
	Introduced guiutil.c subroutines, and changed the structure
	of several window, button and menu blocks.

    2003/10-13 -- Version 2.80
	Changed rotator angle computation to use correct values being
	kept in the current Obs structure rather than quirky globals.

    2003/10-02 -- Version 2.75
	Correctly implements the "OVERLAP" parameter, which is read as
	a negative and written back that way; due to the fix of that
	parameter in mgutils.c.

    2003/10-01 -- Version 2.70
	Add overlap specification window
	Add do-not-cut slit specifications for nod/shuffle

    2003/9-23 -- Version 2.65
	Default order active for dispersers
	Warning if no instrument/disperser selected

    2003/9-17 -- Version 2.60
	Add order for disperser

    2003/9-16 -- Version 2.55
	Add warning of rotator flip

    2003/9-12 -- Version 2.50
	Add repeat count windows for object and reference
	Add reference select window
	Dump obs data only with dodebug > 1

    2003/8-13 -- Version 2.40
	Remove some debugging stuff
	Using "Tname" from mgutils to default telescope name

    2003/7-22 -- Version 2.35
	Drop the "Slit/hole" checkbox

    2003/7-18 -- Version 2.30
	Scaling in structure for default slit dimensions changed
	to arc seconds.

    2003/7-15 -- Version 2.25
	Fixed bug in file menu which would crash without title
	Disable center wavelength window unless grating is active
	Removing some extra comments

    2003/7-14 -- Version 2.20
	Change names short/long to f/2, f/4
	Remove global slit tilt window
	Add flag for extra order spectra

    2003/6-16 -- Version 2.15
	Add automatic save of file if changed

    2003/6-14 -- Version 2.10
	Synchronize file name to save/saveas write and read

    2003/6-6 -- Version 2.00
	Add KDC_DialogBox to ask for text information
	Introduce Save and SaveAs functionality in main menu

    2003/5-22 -- Version 1.90
	Add "WDEBUG" status for debug appearance of window, with
	alignment lines, test button, etc...

    2003/5-21 -- Version 1.85
	Add window for center wavelength
	Adjust horizontal spacing again

    2003/5-14 -- Version 1.80
	Add button for object file list; and button list

    2003/5-8 -- Version 1.75
	Add optional background fiducials/grid as debug alignment aid

    2003/5-7 -- Version 1.70
	Add font size as command line option.

    2003/5-6 -- Version 1.65
	Change "Xsize" to take horizontal units rather than characters

    2003/5-5 -- Version 1.60
	Change horizontal location/size values to dynamic

    2003/4-30 -- Version 1.50
	Change to dynamic location values for all positions
	Starting with vertical positions...

    2003/4-30 -- Version 1.20
	Add reference slit size window as automatic

    2003/4-30 -- Version 1.10
	Cleaning up source during conversion to new CBX library
	Moving function prototypes to intgui.h; intend to move
	all of intgui.h back to this file later, maybe.

    2003/4-22+ -- Version 1.00
	Changing from old to new version of CCB X-library; changes
	most/all references to library functions...

    2003/4-18 -- Version 0.80
	Added generalized window handling for enter,leave,keypress

    2003/4-9 -- Version 0.72
	Adding some items to the file output

    2003/4-4 -- Version 0.70
	Adding filter menu and passband windows
	Also more slit specifications

    2003/4-3 -- Version 0.65
	Add decimal point count in window structure; combine
	the DMS with HMS window types using scale and dp.

    2003/4-1 -- Version 0.60
	Add scale factor to window structure
	Adding slit specifications

    2003/3-31 -- Version 0.50
	Remove some extra/redundant comments
	Add slit position angle

    2003/3-28 -- Version 0.40
	Add selection of instrument, disperser

    2003/3-27 -- Version 0.35
	Add instrument, disperser name menus
	Add generalized menus for inst, disp menu activity

    2003/3-25 -- Version 0.30
	New window block to generalize window updating
	Testing of output to file

    2003/3-24 -- Version 0.25
	Adding more windows to gui...

    2003/3-21 -- Version 0.20
	Add the GUI window itself

    2003/3-20 -- Version 0.15
	Add subroutine to get the optic data

    2003/3-19 -- Version 0.11
	Restart of debug version
	Getting input and argument logic right.

    2002/5-13 -- Version 0.10
	Initial debug version. 

..*/


/* SIMPLIFIED  input logic --
  All flag arguments are put into the "opts" structure, defined
  globally for this application.  The optional file name argument is
  captured in "infile", and only the first such argument is used.  */






